CC=mpCC

GENERIC_LFLAGS =\
-L../../hypre_sp5/lib \
  -lHYPRE\
  -lHYPRE_DistributedMatrix\
  -lHYPRE_DistributedMatrixPilutSolver\
  -lHYPRE_Euclid\
  -lHYPRE_FEI\
  -lHYPRE_IJ_mv\
  -lHYPRE_krylov\
  -lHYPRE_LSI\
  -lHYPRE_MatrixMatrix\
  -lHYPRE_mli\
  -lHYPRE_multivector\
  -lHYPRE_ParaSails\
  -lHYPRE_parcsr_block_mv\
  -lHYPRE_parcsr_ls\
  -lHYPRE_parcsr_mv\
  -lHYPRE_seq_mv\
  -lHYPRE_sstruct_ls\
  -lHYPRE_sstruct_mv\
  -lHYPRE_struct_ls\
  -lHYPRE_struct_mv\
  -lHYPRE_superlu\
  -lHYPRE_utilities\
  -lm

 CFLAGS= -I.\
  -D_FLOAT_EXPAND  -DMY_HYPRE\
  -DHYPRE_TIMING \
 -I../..//hypre_sp5/include\
 -D_MY_CPP -D__MYPI__ -O4 -qipa=inline -qmaxmem=8000000

all: sparpcg spelast

smympi.o: mympicxx.cpp mympicxx.hpp
	$(CC) -c -o smympi.o mympicxx.cpp

sparpcg: srest.o scg.o sfunctors.o ssysgen.o svisitors.o svectors.o ssolver.o smatrix.o sprecon.o smympi.o
	$(CC) $(CFLAGS) srest.o scg.o sfunctors.o ssysgen.o svectors.o ssolver.o smatrix.o sprecon.o smympi.o -o sparpcg $(GENERIC_LFLAGS)

spelast: selast.o ssparcemmat.o selastgen.o ssysgen.o svectors.o selastsolver.o ssolver.o sprecon.o svisitors.o sfunctors.o smympi.o
	$(CC) $(CFLAGS) -o spelast selast.o ssparcemmat.o selastgen.o  ssysgen.o svectors.o selastsolver.o ssolver.o sprecon.o svisitors.o sfunctors.o smympi.o $(GENERIC_LFLAGS)

svisitors.o:visitors.cpp visitors.hpp visitors.h index.h vectors.h \
  small.h
	$(CC) $(CFLAGS) -o svisitors.o -c visitors.cpp

svectors.o:vectors.cpp vectors.hpp vectors.h small.h index.h useful.h
	$(CC) $(CFLAGS) -o svectors.o -c vectors.cpp

sprecon.o:precon.h precon.hpp precon.cpp
	$(CC) $(CFLAGS) -o sprecon.o -c precon.cpp

smatrix.o:  matrix.cpp matrix.hpp matrix.h solver.h useful.h small.h \
  vectors.h index.h functors.h mmpl.hpp ttl/meta/typelist.hpp \
  ttl/config.hpp ttl/exception.hpp ttl/config.hpp ttl/macro_params.hpp \
  ttl/macro_repeat.hpp ttl/macro_misc.hpp ttl/detail/macro_counter.hpp \
  ttl/macro_misc.hpp ttl/equivalent_types.hpp ttl/selector.hpp visitors.h \
  visitors.hpp
	$(CC) $(CFLAGS) -o smatrix.o -c matrix.cpp

ssolver.o: solver.cpp solver.hpp solver.h useful.h small.h functors.h \
  mmpl.hpp ttl/meta/typelist.hpp ttl/config.hpp ttl/exception.hpp \
  ttl/config.hpp ttl/macro_params.hpp ttl/macro_repeat.hpp \
  ttl/macro_misc.hpp ttl/detail/macro_counter.hpp ttl/macro_misc.hpp \
  ttl/equivalent_types.hpp ttl/selector.hpp visitors.h index.h vectors.h
	$(CC) $(CFLAGS) -o ssolver.o -c solver.cpp

scg.o:	cg.cpp cg.hpp cg.h vectors.h small.h index.h sysgen.h ntransform.h \
  multipoly.h base.h calc.h poly.h mirorer.h matrix.h solver.h useful.h \
  functors.h mmpl.hpp ttl/meta/typelist.hpp ttl/config.hpp \
  ttl/exception.hpp ttl/config.hpp ttl/macro_params.hpp \
  ttl/macro_repeat.hpp ttl/macro_misc.hpp ttl/detail/macro_counter.hpp \
  ttl/macro_misc.hpp ttl/equivalent_types.hpp ttl/selector.hpp visitors.h \
  visitors.hpp precon.h
	$(CC) $(CFLAGS) -o scg.o -c cg.cpp

ssysgen.o: sysgen.cpp sysgen.hpp sysgen.h small.h ntransform.h multipoly.h \
  base.h calc.h poly.h mirorer.h useful.h index.h vectors.h
	$(CC) $(CFLAGS) -o ssysgen.o -c sysgen.cpp

sfunctors.o:  functors.cpp functors.hpp visitors.h index.h vectors.h \
  small.h functors.h mmpl.hpp ttl/meta/typelist.hpp ttl/config.hpp \
  ttl/exception.hpp ttl/config.hpp ttl/macro_params.hpp \
  ttl/macro_repeat.hpp ttl/macro_misc.hpp ttl/detail/macro_counter.hpp \
  ttl/macro_misc.hpp ttl/equivalent_types.hpp ttl/selector.hpp useful.h \
  visitors.hpp
	$(CC) $(CFLAGS) -o sfunctors.o -c functors.cpp

srest.o:parpcg.cpp small.h sysgen.h ntransform.h multipoly.h base.h \
  calc.h poly.h mirorer.h index.h cg.h vectors.h matrix.h solver.h \
  useful.h util.h functors.h mmpl.hpp ttl/meta/typelist.hpp \
  ttl/config.hpp ttl/exception.hpp ttl/config.hpp ttl/macro_params.hpp \
  ttl/macro_repeat.hpp ttl/macro_misc.hpp ttl/detail/macro_counter.hpp \
  ttl/macro_misc.hpp ttl/equivalent_types.hpp ttl/selector.hpp visitors.h \
  fc.h
	$(CC) $(CFLAGS)  -o srest.o -c parpcg.cpp 



selastgen.o: elastgen.cpp elastgen.hpp elastgen.h small.h sysgen.h \
  ntransform.h multipoly.h base.h calc.h poly.h mirorer.h sparcemmat.h
	$(CC) $(CFLAGS) -c -o selastgen.o elastgen.cpp

selast.o: pelast.cpp elastgen.h small.h sysgen.h ntransform.h multipoly.h \
  base.h calc.h poly.h mirorer.h sparcemmat.h util.h elastsolver.h \
  vectors.h index.h solver.h useful.h
	$(CC) $(CFLAGS) -c -o  selast.o pelast.cpp 

ssparcemmat.o:  sparcemmat.cpp sparcemmat.hpp sparcemmat.h small.h
	$(CC) $(CFLAGS) -c -o ssparcemmat.o sparcemmat.cpp

selastsolver.o:  elastsolver.cpp elastsolver.hpp elastsolver.h vectors.h \
  small.h index.h elastgen.h sysgen.h ntransform.h multipoly.h base.h \
  calc.h poly.h mirorer.h sparcemmat.h solver.h useful.h precon.h \
  functors.h mmpl.hpp ttl/meta/typelist.hpp ttl/config.hpp \
  ttl/exception.hpp ttl/config.hpp ttl/macro_params.hpp \
  ttl/macro_repeat.hpp ttl/macro_misc.hpp ttl/detail/macro_counter.hpp \
  ttl/macro_misc.hpp ttl/equivalent_types.hpp ttl/selector.hpp visitors.h
	$(CC) $(CFLAGS) -c -o selastsolver.o elastsolver.cpp

clean:
	rm s*.o sparpcg spelast

