#include <set>
#include <map>

#include "mirorer.h"
#include "functionlin.h"
#include "functionpol.h"			
#include "functiontricubic.h"			
#include <tr1/functional>

#include "mcubes.h"
#include "mpt.h"



template<class Iv, class Ov>
void Renumber(const Iv&iv, Ov&ov, std::vector<int> const&num)
{
  ov.SetSize(iv.sizes[0], iv.sizes[1], iv.sizes[2],
	     iv.dim[0], iv.dim[1], iv.dim[2]);
  typedef typename  Iv::size_type size_type;
    for(size_type i = 0; i < iv.sizes[0]; ++i)
      for(size_type j = 0; j < iv.sizes[1]; ++j)
	for(size_type k = 0; k < iv.sizes[2]; ++k)
	  ov(i,j,k) = std::binary_search(num.begin(), num.end(),iv(i, j, k));
}

template<class T>
void Zeroify(vector3d<T> &v, int I, int J, int K)
{
}


LinearFunction3d<double, double, double> linfunc;

TricubicFunction<double, double, double> trifunc;




double TriFunc(double x, double y, double z)
{
  if ((x < 0)
    ||(y < 0)
    || (z < 0)
    ||(x >= trifunc.range[0])
    ||(y >= trifunc.range[1])
    ||(z >= trifunc.range[2])
      )
  {
    return 0;
  }
  return trifunc(x, y, z);
}
double LinFunc(double x, double y, double z)
{
  if ((x < 0)
    ||(y < 0)
    || (z < 0)
    ||(x >= linfunc.range[0])
    ||(y >= linfunc.range[1])
    ||(z >= linfunc.range[2])
      )
  {
    return 0;
  }
  return linfunc(x, y, z);
}

typedef std::tr1::function<double(double, double, double)> Function;
Function  valuefunc = LinFunc;


const int NT=2;

template<class R, class A, class B, class C>
void Convolute3dPar(vector3d<R> &res, vector3d<A> const &a, vector3d<B> const &b, C mul)
{
  res.SetSize(a.sizes[0],
	      a.sizes[1],
	      a.sizes[2],
	      a.dim[0],
	      a.dim[1],
	      a.dim[2]);
  mpt::Thread threads[NT];
  for (int i = 0; i < NT; ++i)
      threads[i] = mpt::StartThread(
				    std::tr1::bind(Convolute3d<R, A, B, C>
						   ,
						   std::tr1::ref(res),
						   std::tr1::ref(a),
						   std::tr1::ref(b),
						   mul,
						   i, NT)
				    );
  for(int i = 0; i < NT; ++i)
    threads[i].Join();
}

template<class Double>
void Normalize(Double *vec)
{
  Double size = vec[0]*vec[0] + vec[1]*vec[1] + vec[2]*vec[2];
  size = sqrt(size);
  for (int i = 0; i < 3; ++i)
    vec[i]/=size;
}


template<class Double>
int Order(const Double &a, const Double &b)
{
  if (fabs(a-b)/fabs(a+b) < 1e-15)
    return 0;
  if (a<b)
    return 1;
  return -1;
  //  return int(a<b)*2-1;
}

template<class Double>
class VertexInfo
{
public:
  Double vertex[3];
  Double normal[3];
  bool operator<(const VertexInfo&b)const
  {
    for(int i = 0; i < 3; ++i)
    {
      int r = Order(vertex[i], b.vertex[i]);
      if (r==1)
	return true;
      if (r==-1)
	return false;
    }
    return false;
  }
      
};
template<class Double>
class VertexInfo1
{
public:
  Double vertex[3];
  bool operator<(const VertexInfo1&b)const
  {
    for(int i = 0; i < 3; ++i)
    {
      int r = Order(vertex[i], b.vertex[i]);
      if (r==1)
	return true;
      if (r==-1)
	return false;
    }
    return false;
  }
      
};

class VertexMap:public mpt::Mutex
{
public:
  typedef std::set<VertexInfo<double> > map_type;
  typedef  map_type::iterator iterator_type;
private:
  map_type map;
public:
  void AddVert(VertexInfo<double> const &a)
  {
    GUARD(this);
    iterator_type it = map.find(a);
    if (it == map.end())
    {
      map.insert(a);
    }
    else
    {
      for(int i = 0; i < 3; ++i)
      	const_cast<double&>((*it).normal[i])
      	  +=a.normal[i];
    }
  }
  void ExecNormal(double const *a)
  {
    VertexInfo<double> vi;
    for(int i = 0; i < 3; ++i)
      vi.vertex[i] = a[i];
    iterator_type it = map.find(vi);
    if (it == map.end())
    {
      std::cerr<<"mamamata si e ebalo..." << std::endl;
      for (int i = 0; i <3; ++i)
	std::cerr << vi.vertex[i] << " ";
      std::cerr << std::endl;
    }
    else
    {
      double vals[4];
      std::copy(it->normal, it->normal+4, vals);
      Normalize(vals);
      //      ::glNormal3dv(vals);
    }
  }
  void Clear()
  {
    map.clear();
  }
    
};

VertexMap vertex_map;


enum Sides {
  LeftSide = 0x1,
  RightSide = 0x2,
  TopSide = 0x4,
  BottomSide = 0x8,
  FrontSide = 0x10,
  BackSide = 0x20
};




double bonethres = 0.5;//128;//0.15;









double Cos(double v1[3], double v2[3], double v3[3])
{
  for(int i = 0; i < 3;++i)
  {
    v1[i] -= v3[i];
    v2[i] -= v3[i];
  }
  double l1 = 0;
  double l2 = 0;
  for (int i = 0; i < 3; ++i)
  {
    l1 += v1[i]*v1[i];
    l2 += v2[i]*v2[i];
  }
  l1 = sqrt(l1);
  l2 = sqrt(l2);
  double cos = 0;
  for (int i = 0; i < 3; ++i)
  {
    cos += v1[i]*v2[i];
  }
  cos /= l1*l2;
  return cos;    
}

template<class T>
struct Quad
{
  T w,x,y,z;
  Quad():w(), x(), y(), z(){}
  Quad(T ax, T ay, T az, T angle):
    w(cos(angle/2)),
    x(ax*sin(angle/2)),
    y(ay*sin(angle/2)),
    z(az*sin(angle/2))
  {}
  void GetMatrix(T (&A)[3][3])
  {
    A[0][0] = w*w + x*x - y*y - z*z; A[0][1] = 2*x*y - 2*w*z;        A[0][2] = 2*x*z + 2*w*y;
    A[1][0] = 2*x*y + 2*w*z;         A[1][1] = w*w - x*x + y*y - z*z;A[1][2] = 2*y*z - 2*w*x;
    A[2][0] = 2*x*z - 2*w*y;         A[2][1] = 2*y*z + 2*w*x;        A[2][2] = w*w - x*x - y*y + z*z;
  }
};
template<class T>
Quad<T> operator*(const Quad<T> &a, const Quad<T>&b)
{
  Quad<T> r;
  r.w = a.w*b.w - a.x*b.x - a.y*b.y - a.z*b.z;
  r.x = a.w*b.x + a.x*b.w + a.y*b.z - a.z*b.y;
  r.y = a.w*b.y + a.y*b.w + a.z*b.x - a.x*b.z;
  r.z = a.w*b.z + a.z*b.w + a.x*b.y - a.y*b.x;
  return r;
}

  

using namespace std;

double vsx, vsy, vsz;
double minx, miny, minz, maxx, maxy, maxz;
double xc, yc, zc;

double angle=atan(1)*2,ratio = 1;
bool usecube = false;
bool swapp=true;
double angle1=0;
double angle2=0;
double dist = 2.5;

double
bmin[2],  bmax[2];
double amin, amax;
std::vector<double> Dmins[3][2];
std::vector<double> Dmaxs[3][2];



template<class T>
void MulSmallMat(T (&res)[3][3],const T (&A)[3][3],const T(&B)[3][3])
{
  for(int i = 0; i < 3; ++i)
    for(int j = 0; j < 3; ++j)
    {
      T sum = 0;
      for(int k = 0; k < 3; ++k)
	sum += A[i][k] * B[k][j];
      res[i][j]=sum;
    }
}
template<class T>
void MulSmallMatVec(T (&res)[3],const T (&A)[3][3],const T(&x)[3])
{
  for(int i = 0; i < 3; ++i)
  {
    double sum = 0;
    for(int k = 0; k < 3; ++k)
      sum += A[i][k] * x[k];
    res[i]=sum;
  }
}
template<class T>
void Trans(const T (&A)[3][3], T(&x)[3])
{
  T t[3];t[0] = x[0];t[1] = x[1]; t[2] = x[2];
  MulSmallMatVec(x, A, t);
}

std::vector<double> surfpa[NT];


struct P3D
{
  double X[3];
};



bool IntersectLine(P3D p1, P3D p2, double v1, double v2, double val, P3D &res)
{
  double f1 = min(v1, v2);
  double f2 = max(v1, v2);
  if ((val <f1) || (val >f2))
    return false;
  if (v2==v1)
    return false;
  double k = (val - v1)/(v2-v1);
  double k1  = 1 - k;
  for(int i = 0; i <3; ++i)
    res.X[i] = p1.X[i]*k1 + p2.X[i]*k;
  return true;
}


template <class T>
inline bool In(const T&a, const T&min, const T &max)
{
  return (a >=min) && (a<max);
}
template<class T, class I>
bool check(T &vec, I i, I j, I k)
{
  ptrdiff_t szs[3];
  bool ret = true;
  if (i>0)
    ret = ret && vec(i-1, j, k);
  else
    ret = false;
  if (j>0)
    ret = ret && vec(i, j-1, k);
  else
    ret = false;
  if (k>0)
    ret = ret && vec(i, j, k-1);
  else
    ret = false;
  if (i<vec.sizes[0]-1)
    ret  = ret && vec(i+1, j, k);
  else
    ret = false;
  if (j<vec.sizes[1]-1)
    ret  = ret && vec(i, j+1, k);
  else
    ret = false;
  if (k<vec.sizes[2]-1)
    ret = ret && vec(i, j, k+1);
  else
    ret = false;
  bool tmp = vec(i,j,k);
  int sz1, sz2, sz3;
  sz1 = vec.sizes[0]/2;
  sz2 = vec.sizes[1]/2;
  sz3 = vec.sizes[2]/2;
  //     if ((i<sz1)&&(j<sz2) && (k <sz3))
  //       return false;
  if ((i==sz1)||(j==sz2) || (k==sz3))
    return vec(i,j,k);
  return tmp^ret;
}

struct color4
{
  double r,g,b,a;
  color4():r(), g(), b(), a(){}
  color4(double r_, double g_, double b_, double a_)
  {
    r = r_;
    g = g_;
    b = b_;
    a = a_;
  }
  void Set(double r_, double g_, double b_, double a_)
  {
    r = r_;
    g = g_;
    b = b_;
    a = a_;
  }
  void operator *=(const color4&a)
  { 
    r *= a.r;
    g *= a.g;
    b *= a.b;
    this->a *= a.a;
    return void(*this);//!!!
  }
};

std::vector<color4> cols, colwhite;

void CalcNormal(SmallVector<double, 3> &n, SmallVector<double, 3> const&p)
{
  //  double c = valuefunc(p[0], p[1], p[2]);
  double h = -0.01;
  double sign = -1;
  n[0] = sign*(valuefunc(p[0]-h, p[1], p[2]) - valuefunc(p[0]+h, p[1], p[2]));
  n[1] = sign*(valuefunc(p[0], p[1]-h, p[2]) - valuefunc(p[0], p[1]+h, p[2]));
  n[2] = sign*(valuefunc(p[0], p[1], p[2]-h) - valuefunc(p[0], p[1], p[2]+h));
  
//   std::cout << n[0] << " " << n[1] << " "<< n[2] << std::endl;
//   std::cout << c << " " <<n[0] << " " << n[1] << " "<< n[2] << std::endl;
}


double Sqr(double a)
{
  return a*a;
}

double Length(SmallVector<double, 3> *p1, SmallVector<double, 3> *p2)
{
  SmallVector<double, 3>  &P1 = *p1;
  SmallVector<double, 3>  &P2 = *p2;

  return sqrt(Sqr(P1[0]-P2[0])+Sqr(P1[1]-P2[1])+Sqr(P1[2]-P2[2]));
}

double Area(SmallVector<double, 3> **p)
{
  double l1 = Length(p[0], p[1]);
  double l2 = Length(p[1], p[2]);
  double l3 = Length(p[2], p[0]);
  double s = (l1 + l2 + l3)/2;
  double S = s*(s-l1)*(s-l2)*(s-l3)/2.;
  return sqrt(S);
}
    
  

void DrawSide3(SmallVector<double, 3> &p1,
	      SmallVector<double, 3> &p2,
	      SmallVector<double, 3> &p3,
	      bool isbone,
	       double mul,
	       int thread
	      )
{
  SmallVector<double, 3> n, v1, v2;
  v2 = p3;
  v1 = p2;
  v2 -= p1;
  v1 -= p1;
  n[0] = v1[1]*v2[2] - v1[2]*v2[1];
  n[1] = v1[2]*v2[0] - v1[0]*v2[2];
  n[2] = v1[0]*v2[1] - v1[1]*v2[0];
  double Sine = sqrt(Sqr(n[0])+Sqr(n[1]) +Sqr(n[2]));
  n/=Sine;
  // surfp.push_back(n[0]);
  // surfp.push_back(n[1]);
  // surfp.push_back(n[1]);

  
  surfpa[thread].push_back(p1[0]);
  surfpa[thread].push_back(p1[1]);
  surfpa[thread].push_back(p1[2]);
  surfpa[thread].push_back(p2[0]);
  surfpa[thread].push_back(p2[1]);
  surfpa[thread].push_back(p2[2]);
  surfpa[thread].push_back(p3[0]);
  surfpa[thread].push_back(p3[1]);
  surfpa[thread].push_back(p3[2]);
}


void DrawHexMC(double x1, double y1, double z1,
	     double o1, double o2, double o3,
	       int thread
	     )
{
  double o = 0;
  double v[8];
  P3D p[8] = {
    {{x1 + o , y1 + o , z1 + o3}}, 
    {{x1 + o1, y1 + o , z1 + o3}}, 
    {{x1 + o1, y1 + o , z1 + o }}, 
    {{x1 + o , y1 + o , z1 + o }}, 
    {{x1 + o , y1 + o2, z1 + o3}}, 
    {{x1 + o1, y1 + o2, z1 + o3}}, 
    {{x1 + o1, y1 + o2, z1 + o }}, 
    {{x1 + o , y1 + o2, z1 + o }}
  };
  int edges[12][2] =    {
      {0,1},
      {1,2},
      {2,3},
      {3,0},
      {4,5},
      {5,6},
      {6,7},
      {7,4},
      {0,4},
      {1,5},
      {2,6},
      {3,7}
  };

  P3D poe[12];
  //  P3D npoe[12];

  double bt = bonethres;
  int ind1 = 0;
  for (int i = 0; i < 8; ++i)
  {
    double val = valuefunc(p[i].X[0], p[i].X[1], p[i].X[2]);
//     double x = p[i].X[0]-0.5;
//     double y = p[i].X[1]-0.5;
//     double z = p[i].X[2]-0.5;
//     val = x*x + y*y+z*z;
    //    val *= p[i].X[0]*(1-p[i].X[0])*p[i].X[1]*(1-p[i].X[1])*p[i].X[2]*(1-p[i].X[2]);
    v[i] = val;
    int cond = val >bt;
    ind1 += (cond << i);
  }

  int ind2 = ET[ind1];
  for (int i = 0; i < 12; ++i)
    if (ind2 & (1 << i))
    {
      P3D &p1 = p[edges[i][0]];
      P3D &p2 = p[edges[i][1]];
      double &v1 = v[edges[i][0]];
      double &v2 = v[edges[i][1]];
      double x = (bt - v2)/(v1-v2);
      for (int j = 0; j <3; ++j)             //interpolation
	poe[i].X[j] = x * p1.X[j] + (1-x)*p2.X[j];
    }
  for (int i = 0; i < 15; i+=3)
  {
    if (MC[ind1][i] <0)
      break;
    SmallVector<double, 3> p1, p2, p3;
    for (int j = 0; j < 3; ++j)
    {
      p1[j] = poe[MC[ind1][i]].X[j];
      p2[j] = poe[MC[ind1][i+1]].X[j];
      p3[j] = poe[MC[ind1][i+2]].X[j];
    }
    DrawSide3(p1, p2, p3,
	      true,
	      1.0,
	      thread
	      );
    
    
  }
  
}

size_t N1, N2, N3;
double bsx, bsy, bsz;
double sx, sy, sz;
void GenerateMCSolidList1(int start, int end, int thread)
{
  
  for(int i = start; i < end; ++ i)
  {
    if (!thread)
      std::cout <<"\r" << 100.*(double)(i-start)/(end-start) << "%                 " << std::flush;
    for(unsigned j = 0; j < N2; ++j)
    {
      for (unsigned  k = 0; k < N3; ++k)
      {
	  DrawHexMC( sx + i*bsx, sy + j*bsy, sz + k*bsz, bsx, bsy, bsz,
		     thread);
      }
    }
  }
}
void GenerateMCSolidList()
{
  std::cout << "Solid list start " << std::endl;
  const int N = N1;
  //  const int NT = 1;
  mpt::Thread threads[NT];
  for (int i = 0; i < NT; ++i)
      threads[i] = mpt::StartThread(
				    std::tr1::bind(GenerateMCSolidList1
						   ,
						   i*N/NT,
						   (i+1)*N/NT,
						   i
						   )
				    );
  for(int i = 0; i < NT; ++i)
    threads[i].Join();
  for (int i = 1; i <NT; ++i)
  {
    surfpa[0].insert(surfpa[0].end(), surfpa[i].begin(), surfpa[i].end());
    std::vector<double> empty;
    std::swap(empty, surfpa[i]);
  }
      
  std::cout <<"\r" <<  "100%                 ";
  //  GenerateMCSolidList1(i*N/NT, (i+1)*N/NT, i);
  //  std::for_each(vecs.begin(), vecs.end(), std::mem_fun_ref(&GlObjectd::Execute));
}




inline double razt(double x1, double y1, double z1,
		   double x2, double y2, double z2)
{
  x1-=x2;
  y1-=y2;
  z1-=z2;
  return sqrt(x1*x1 + y1*y1 + z1*z1);
}




template<class T>
T Cast(const std::string&s)
{
  std::istringstream is(s);
  T res;
  is >> res;
  return res;
}


int main(int argc, char **argv)
{
  std::string ifname="file.ihr";
  std::string ofname="file";
  if (argc>1)
    ifname=argv[1];
  if (argc>2)
    ofname = argv[2];
  std::vector<int> which;

  bonethres = Cast<double>(argv[3]);
  for (int i = 4; i < argc; ++i)
  {
    int a = Cast<int>(argv[i]);
    which.push_back(a);
  }
  std::sort(which.begin(), which.end());
  vector3d<unsigned char> ivec;
  vector3d<double> ovec;
  vector3d<double> resvec;
  vector3d<double> gauss;
  ivec.ReadInr(ifname);
  std::cout << ifname << " read." << std::endl;
  Renumber(ivec, ovec, which);
  ivec.Clear();
  resvec.SetSize(ovec.sizes[0], ovec.sizes[1], ovec.sizes[2],
	     ovec.dim[0], ovec.dim[1], ovec.dim[2]);
  std::cout << "Renumbering Done" << std::endl;
  double scale = 0.95;//0.8;
  GaussianConvMat(gauss, 5, 1.3, ovec.dim[0]*scale, ovec.dim[1]*scale, ovec.dim[2]*scale);
  Convolute3dPar(resvec, ovec, gauss, 1.3);
  std::cout << "Convolution Done" << std::endl;
  ovec.Clear();
  SmallVector<double, 3> min, max;
  min = 0;
  max[0] = resvec.sizes[0] * resvec.dim[0];
  max[1] = resvec.sizes[1] * resvec.dim[1];
  max[2] = resvec.sizes[2] * resvec.dim[2];
  //  trifunc.Init(resvec, min, max, false);
  linfunc.Init(resvec, min, max, false);
  N1 = resvec.sizes[0]+2;
  N2 = resvec.sizes[1]+2;
  N3 = resvec.sizes[2]+2;
  // N1 = 258;
  // N2 = 258;
  // N3 = 258;
  max[0] += resvec.dim[0];
  max[1] += resvec.dim[1];
  max[2] += resvec.dim[2];
  sx = - resvec.dim[0];
  sy = - resvec.dim[1];
  sz = - resvec.dim[2];
  bsx = max[0]/N1;
  bsy = max[1]/N2;
  bsz = max[2]/N3;
  GenerateMCSolidList();

  typedef std::map<VertexInfo1<double>, int > map_type;
  typedef typename map_type::iterator iter;
  map_type map;
  int ind = 0;
  for (size_t i = 0; i < surfpa[0].size(); i+=3)
  {
    VertexInfo1<double> a;
    std::copy(&surfpa[0][i], &surfpa[0][i+3], a.vertex);
    if (map.find(a)==map.end())
    {
      map.insert(std::make_pair(a, ind++));
    }
  }

  //output off format
  std::ofstream f((ofname+".off").c_str());
  f.precision(16);
  f << "OFF\n";
  f <<  map.size() << " " << surfpa[0].size()/9 << " 0\n\n";
		  
  ind = 0;
  std::cout << map.size() << " distinct verts" << std::endl;
  //renumber & output indeses
  for (iter it = map.begin(); it!=map.end(); ++it)
  {
    it->second=ind++;
    double tsize = 1.;
    f << it->first.vertex[0]*tsize << " " << it->first.vertex[1]*tsize << " " << it->first.vertex[2]*tsize << "\n";
  }
  
  for (size_t i = 0; i < surfpa[0].size(); i+=9)
    {
      VertexInfo1<double> a, b,c;
      std::copy(&surfpa[0][i], &surfpa[0][i+3], a.vertex);
      std::copy(&surfpa[0][i+3], &surfpa[0][i+6], b.vertex);
      std::copy(&surfpa[0][i+6], &surfpa[0][i+9], c.vertex);
      f <<"3 " << map[a] << " " << map[c] << " " << map[b] << "\n";
    }

  f << std::endl;  
}
