// Copyright (c) 2011  GeometryFactory Sarl (France)
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; version 2.1 of the License.
// See the file LICENSE.LGPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $URL: https://geometryfactory.svn.codebasehq.com/hutchinson/code/mesh_generation/angle_3_extension.h $
// $Id: angle_3_extension.h 18 2011-10-05 07:35:10Z lrineau $
// 
//
// Author(s)     : Laurent Rineau

#ifndef CGAL_ANGLE_3_EXTENSION_H
#define CGAL_ANGLE_3_EXTENSION_H

#include <CGAL/number_utils.h>

namespace CGAL {

// for Cartesian_kernel/include/CGAL/predicates/kernel_ftC3.h

template < class FT >
inline
typename Same_uncertainty_nt<Angle, FT>::type
angleC3(const FT &px, const FT &py, const FT &pz,
        const FT &qx, const FT &qy, const FT &qz,
        const FT &ux, const FT &uy, const FT &uz, bool)
{
  return enum_cast<Angle>(CGAL_NTS sign((qx-px)*ux+
	                                (qy-py)*uy+
				        (qz-pz)*uz));
}



// for Cartesian_kernel/include/CGAL/Cartesian/function_objects.h

  template <typename K>
  class Angle_3_ext
  {
    typedef typename K::Point_3 Point_3;
    typedef typename K::Vector_3 Vector_3;
  public:
    typedef typename K::Angle   result_type;

    result_type
    operator()(const Point_3& p, const Point_3& q, const Point_3& r) const
    {
      return angleC3(p.x(), p.y(), p.z(),
		     q.x(), q.y(), q.z(),
		     r.x(), r.y(), r.z());
    }

    result_type
    operator()(const Point_3& p, const Point_3& q, const Vector_3& u) const
    {
      return angleC3(p.x(), p.y(), p.z(),
		     q.x(), q.y(), q.z(),
		     u.x(), u.y(), u.z(), true); // vector
    }
  };



template <typename FK>
class Filtered_angle_3_ext 
  : public Filtered_predicate< Angle_3_ext<typename FK::Exact_kernel>,
                               Angle_3_ext<typename FK::Approximate_kernel>,
                               typename FK::C2E,
                               typename FK::C2F >
{};

} // end namespace CGAL

#endif // CGAL_ANGLE_3_EXTENSION_H
