#include "mirorer.h"
#include "functionlin.h"
#include "functionpol.h"			
#include "functiontricubic.h"			
#include <tr1/functional>

#include <CGAL/trace.h>
#include <CGAL/Exact_predicates_inexact_constructions_kernel.h>
#include <CGAL/Polyhedron_3.h>
#include <CGAL/IO/Polyhedron_iostream.h>
#include <CGAL/Surface_mesh_default_triangulation_3.h>
#include <CGAL/make_surface_mesh.h>
#include <CGAL/Implicit_surface_3.h>
#include <CGAL/IO/output_surface_facets_to_polyhedron.h>
//#include <CGAL/Point_with_normal_3.h>
#include <CGAL/property_map.h>
#include <CGAL/IO/read_xyz_points.h>
#include <CGAL/compute_average_spacing.h>

template<class Iv, class Ov>
void Renumber(const Iv&iv, Ov&ov, int num)
{
  ov.SetSize(iv.sizes[0], iv.sizes[1], iv.sizes[2],
	     iv.dim[0], iv.dim[1], iv.dim[2]);
  typedef typename  Iv::size_type size_type;
    for(size_type i = 0; i < iv.sizes[0]; ++i)
      for(size_type j = 0; j < iv.sizes[1]; ++j)
	for(size_type k = 0; k < iv.sizes[2]; ++k)
	  ov(i,j,k) = (iv(i, j, k)==num)*2 - 1;
}

typedef CGAL::Exact_predicates_inexact_constructions_kernel K;
typedef std::tr1::function<double(typename K::Point_3 const&)> Function ;
typedef K::FT FT;
typedef K::Point_3 Point;
typedef K::Sphere_3 Sphere;
typedef CGAL::Polyhedron_3<K> Polyhedron;
typedef CGAL::Surface_mesh_default_triangulation_3 STr;
typedef CGAL::Surface_mesh_complex_2_in_triangulation_3<STr> C2t3;
typedef CGAL::Implicit_surface_3<K, Function> Surface_3;
LinearFunction3d<double, double, double> linfunc;

TricubicFunction<double, double, double> trifunc;

double sphere_function1 ( typename K::Point_3 const& p)
{ return CGAL::squared_distance(p, typename K::Point_3(0, 0, 0))-1; }
double sign = 1.;
double TriFunc(Point const &p)
{
  double x, y, z;
  x = p.x();
  y = p.y();
  z = p.z();
  if ((x < 0)
    ||(y < 0)
    || (z < 0)
    ||(x >= trifunc.range[0])
    ||(y >= trifunc.range[1])
    ||(z >= trifunc.range[1])
      )
  {
    return -1*sign;
  }
  return trifunc(p.x(), p.y(), p.z())*sign;
}

int main(int argc, char **argv)
{
  std::string ifname="file.ihr";
  std::string ofname="file";
  int which = 1;
  if (argc>1)
    ifname=argv[1];
  if (argc>2)
  {
    std::istringstream is(argv[2]);
    is >> which;
  }
  if (argc>3)
    ofname = argv[3];
  vector3d<unsigned char> ivec;
  vector3d<double> ovec;
  vector3d<double> &resvec = ovec;
  vector3d<double> gauss;
  ivec.ReadInr(ifname);
  std::cout << ifname << " read." << std::endl;
  Renumber(ivec, ovec, which);
  ivec.Clear();
  resvec.SetSize(ovec.sizes[0], ovec.sizes[1], ovec.sizes[2],
	     ovec.dim[0], ovec.dim[1], ovec.dim[2]);
  std::cout << "Renumbering Done" << std::endl;
  GaussianConvMat(gauss, 5, 1.3, ovec.dim[0], ovec.dim[1], ovec.dim[2]);
  //  Convolute3d(resvec, ovec, gauss, 1.0);
  std::cout << "Convolution Done" << std::endl;
  //  ovec.Clear();
  SmallVector<double, 3> min, max;
  min = 0;
  max[0] = resvec.sizes[0] * resvec.dim[0];
  max[1] = resvec.sizes[1] * resvec.dim[1];
  max[2] = resvec.sizes[2] * resvec.dim[2];
  trifunc.Init(resvec, min, max, false);
  // resvec.SetPars(0,
  // 		 0, resvec.sizes[0] * resvec.dim[0],
  // 		 0, resvec.sizes[1] * resvec.dim[1],
  // 		 0, resvec.sizes[2] * resvec.dim[2]);


  FT sm_angle = 20.0; // Min triangle angle in degrees.
  FT sm_radius = 4; // Max triangle size w.r.t. point set average spacing.
  FT sm_distance = 4; // Surface Approximation error w.r.t. point set average spacing.


  // Computes average spacing
  Point inner_point =  Point( resvec.sizes[0] * resvec.dim[0] *0.5,
  			      resvec.sizes[1] * resvec.dim[1] *0.5,
  			      resvec.sizes[2] * resvec.dim[2] *0.5);
  Function func = TriFunc;
  if (func(inner_point)>=0)
    sign*=-1;
  // Defines the implicit surface: requires defining a
  // conservative bounding sphere centered at inner point.
  //  Function func = sphere_function1;
  Surface_3 surface(func,
		    Sphere(inner_point,1000));

  // Defines surface mesh generation criteria
  CGAL::Surface_mesh_default_criteria_3<STr> criteria(sm_angle,  // Min triangle angle (degrees)
						      sm_radius,  // Max triangle size
						      sm_distance); // Approximation error

  // Generates surface mesh with manifold option
  STr tr; // 3D Delaunay triangulation for surface mesh generation
  C2t3 c2t3(tr); // 2D complex in 3D Delaunay triangulation
  CGAL::make_surface_mesh(c2t3,                                 // reconstructed mesh
			  surface,                              // implicit surface
			  criteria,                             // meshing criteria
			  CGAL::Manifold_with_boundary_tag());  // require manifold mesh

  if(tr.number_of_vertices() == 0)
    return EXIT_FAILURE;

  // saves reconstructed surface mesh
  std::ofstream out((ofname+".off").c_str());
  Polyhedron output_mesh;
  CGAL::output_surface_facets_to_polyhedron(c2t3, output_mesh);
  out << output_mesh;
  
    
}
