#define CGAL_MESH_3_IO_VERBOSE

#include <CGAL/Exact_predicates_inexact_constructions_kernel.h>

#include <CGAL/Mesh_triangulation_3.h>
#include <CGAL/Mesh_complex_3_in_triangulation_3.h>
#include <CGAL/Mesh_criteria_3.h>

#include <CGAL/Labeled_image_mesh_domain_3.h>
#include <CGAL/make_mesh_3.h>
#include <CGAL/Image_3.h>

// Domain
typedef CGAL::Exact_predicates_inexact_constructions_kernel K;
typedef CGAL::Labeled_image_mesh_domain_3<CGAL::Image_3,K> Mesh_domain;

// Triangulation
typedef CGAL::Mesh_triangulation_3<Mesh_domain>::type Tr;
typedef CGAL::Mesh_complex_3_in_triangulation_3<Tr> C3t3;

// Criteria
typedef CGAL::Mesh_criteria_3<Tr> Mesh_criteria;
typedef CGAL::Mesh_constant_domain_field_3<Mesh_domain::R,
                                           Mesh_domain::Index> Sizing_field;

// To avoid verbose function and named parameters call
using namespace CGAL::parameters;

typedef K::FT FT;
typedef K::Point_3 Point;
struct Spherical_sizing_field
{
  typedef ::FT FT;
  typedef Point Point_3;
  typedef Mesh_domain::Index Index;

  FT operator()(const Point_3& p, const int, const Index&) const
  {
    FT cx, cy, cz;
    cx = 118;
    cy = 207;
    cz = 110;
    Point P(cx, cy, cz);
    cz = 110;
    Point P1(cx, cy-10, cz);
    FT sq_d_to_origin = CGAL::squared_distance(p, P);
    FT sq_d_to_origin1 =std::min( CGAL::squared_distance(p, P1),sq_d_to_origin) ;
    FT dist = CGAL::abs( CGAL::sqrt(sq_d_to_origin1)) / 30. + 0.01;
    if (dist >3)
      dist = 3;
    return dist;
  }
};




int main(int argc, char**argv)
{
  std::string fname = "mesh.inr";
  if (argc>1)
    fname = argv[1];
  // Loads image
  CGAL::Image_3 image;
  image.read(fname.c_str());

  // Domain
  Mesh_domain domain(image);





  // Sizing field: set global size to 8 and kidney size (label 127) to 3
  double needle_size = 0.2;
  int volume_dimension = 3;
  Sizing_field size1(5);
  needle_size = 0.2;
  size1.set_size(needle_size, volume_dimension, 
                domain.index_from_subdomain_index(1));
  size1.set_size(needle_size, volume_dimension, 
                domain.index_from_subdomain_index(2));
  Spherical_sizing_field sss;
  
  // Mesh criteria
  Mesh_criteria criteria(facet_angle=25, facet_size=size1, facet_distance=0.15,
                         cell_radius_edge_ratio=3, cell_size=sss);
  // Mesh_criteria criteria(facet_angle=25, facet_size=size1, facet_distance=0.15,
  //                        cell_radius_edge_ratio=3, cell_size=size);




  

  // Meshing
  C3t3 c3t3 = CGAL::make_mesh_3<C3t3>(domain, criteria);//, features(domain));//, lloyd(), odt(), perturb(), exude());

  // Output
  std::ofstream medit_file((fname+".mesh").c_str());
  c3t3.output_to_medit(medit_file, true, true); // rebind?=true, //show surface patches = true

  return 0;
}
