#define CGAL_MESH_3_IO_VERBOSE

#define CGAL_CFG_ARRAY_MEMBER_INITIALIZATION_BUG

#include <CGAL/Exact_predicates_inexact_constructions_kernel.h>
#include <CGAL/Exact_predicates_exact_constructions_kernel_with_sqrt.h>

#include <CGAL/Simple_cartesian.h>
#include <CGAL/AABB_tree.h>
#include <CGAL/AABB_traits.h>
#include <CGAL/Polyhedron_3.h>
#include <CGAL/AABB_polyhedron_triangle_primitive.h>

#include <CGAL/Mesh_triangulation_3.h>
#include <CGAL/Mesh_polyhedron_3.h>
#include <CGAL/Mesh_complex_3_in_triangulation_3.h>
#include <CGAL/Mesh_criteria_3.h>

#include <CGAL/Labeled_image_mesh_domain_3.h>
#include <CGAL/make_mesh_3.h>
#include <CGAL/Image_3.h>

#include <tr1/functional>

#include <CGAL/IO/Polyhedron_iostream.h>
#include <CGAL/Polyhedral_mesh_domain_3.h>

#include "Multi_domain_polyhedral_mesh_domain_3.h"



typedef CGAL::Exact_predicates_inexact_constructions_kernel K;
typedef CGAL::Mesh_polyhedron_3<K>::type Polyhedron;
typedef CGAL::AABB_polyhedron_triangle_primitive<K,Polyhedron> Primitive;
typedef CGAL::AABB_traits<K, Primitive> Traits;
typedef CGAL::AABB_tree<Traits> Tree;
typedef CGAL::Multi_domain_polyhedral_mesh_domain_3<Polyhedron, K> Mesh_domain;



// Triangulation
typedef CGAL::Mesh_triangulation_3<Mesh_domain>::type Tr;
typedef CGAL::Mesh_complex_3_in_triangulation_3<Tr> C3t3;

// Criteria
typedef CGAL::Mesh_criteria_3<Tr> Mesh_criteria;
typedef CGAL::Mesh_constant_domain_field_3<Mesh_domain::R,
                                           Mesh_domain::Index> Sizing_field;
typedef K::Point_3 Point;
struct Spherical_sizing_field
{
  typedef double FT;
  typedef Point Point_3;
  typedef Mesh_domain::Index Index;
  Point where;
  FT dist;
  FT size1, size2;
  FT operator()(const Point_3& p, const int, const Index&) const
  {
    FT cx, cy, cz;
    cx = 118;
    cy = 207;
    cz = 110;
    Point P(cx, cy, cz);
    cz = 110;
    Point P1(cx, cy-10, cz);
    FT sq_d_to_origin = CGAL::squared_distance(p, P);
    FT sq_d_to_origin1 =std::min( CGAL::squared_distance(p, P1),sq_d_to_origin) ;
    FT dist = CGAL::abs( CGAL::sqrt(sq_d_to_origin1)) / 30. + 0.01;
    if (dist >3)
      dist = 3;
    return dist;
  }
};


using namespace CGAL::parameters;

int main(int argc, char**argv)
{
  std::string fname = "mesh.inr";
  if (argc>1)
    fname = argv[1];
  // Loads image

  std::vector<Polyhedron> vecp;
  Mesh_domain domain;
  for(int i = 2; i < argc; ++i)
  {
    Polyhedron a;
    vecp.push_back(a);
  }
  for(unsigned i = 0; i < vecp.size(); ++i)
  {
    std::ifstream f(argv[i+2]);
    f >> vecp[i];
    std::cout << vecp[i].bytes() << std::endl;
    domain.add_subdomain(vecp[i], i+1, 0.0);
  }
  std::cout << "All read" << std::endl;
  domain.build();
  std::cout << "Domain constructed" << std::endl;
  


  Point p(75+43, 75+132, 75+35);
  Spherical_sizing_field volume_size;
  Sizing_field face_size(3);
			      
  
  // // Mesh criteria
  Mesh_criteria criteria(facet_angle=20, facet_size=face_size, facet_distance=0.05,
			 cell_radius_edge=3, cell_size=volume_size);

  // Meshing
  C3t3 c3t3 = CGAL::make_mesh_3<C3t3>(domain, criteria);//, features(domain));//, lloyd(), odt(), perturb(), exude());
  std::cout << "Mesh constructed" << std::endl;

  // Output


  std::ofstream medit_file((fname+".mesh").c_str());
  c3t3.output_to_medit(medit_file, true, true); // rebind?=true, //show surface patches = true

  return 0;
}
