#define CGAL_CFG_ARRAY_MEMBER_INITIALIZATION_BUG
#define CGAL_MESH_3_IO_VERBOSE
#define NDEBUG
#include <CGAL/Exact_predicates_inexact_constructions_kernel.h>
#include <CGAL/Exact_predicates_exact_constructions_kernel_with_sqrt.h>

#include <CGAL/Simple_cartesian.h>
#include <CGAL/AABB_tree.h>
#include <CGAL/AABB_traits.h>
#include <CGAL/Polyhedron_3.h>
#include <CGAL/AABB_polyhedron_triangle_primitive.h>

#include <CGAL/Mesh_triangulation_3.h>
#include <CGAL/Mesh_complex_3_in_triangulation_3.h>
#include <CGAL/Mesh_criteria_3.h>

#include <CGAL/Labeled_image_mesh_domain_3.h>
#include <CGAL/make_mesh_3.h>
#include <CGAL/Image_3.h>

#include <tr1/functional>

#include <CGAL/Polyhedron_3.h>
#include <CGAL/IO/Polyhedron_iostream.h>
#include <CGAL/Polyhedral_mesh_domain_3.h>

double cx, cy, cz, R;


namespace CGAL {
  namespace Mesh_3 {
    template<class Function_, class BGT>
    class Function_vector_to_labeled_function_wrapper
    {
    public:
      // Types
      typedef int return_type;
      typedef std::vector<Function_> Function_vector;
      typedef typename BGT::Point_3 Point_3;
      /// Constructor

      Function_vector_to_labeled_function_wrapper(Function_vector& v):
	function_vector_(v) {}
      /// Destructor
      ~Function_vector_to_labeled_function_wrapper() {}

      /// Operator ()
      return_type operator()(const Point_3& p, const bool = true)
	const
      {
	int nb_func = function_vector_.size();
	if ( nb_func > 8 )
	{
	  CGAL_error_msg("We support at most 8 functions !");
	}

	char bits = 0;
	for ( int i = 0 ; i < nb_func ; ++i ) {
	  // Insert value into bits : we compute fi(p) and 
	  //insert
	  //result at bit i of bits
	  bits |= ((
		    function_vector_[i](p) < 0) << i );
	}
	return ( static_cast<return_type>(bits) );
      }
    private:
      /// Functions to wrap
      Function_vector function_vector_;
    };



    template<class Tree_, class BGT>
    class Tree_to_labeled_function_wrapper
    {
    public:
      // Types
      typedef int return_type;
      typedef std::vector<Tree_*> Function_vector;
      typedef typename BGT::Point_3 Point_3;
      typedef typename BGT::Ray_3 Ray_3;
      typedef typename BGT::Vector_3 Vector_3;
      /// Constructor

      Tree_to_labeled_function_wrapper(Function_vector& v):
	function_vector_(v) {}
      /// Destructor
      ~Tree_to_labeled_function_wrapper() {
      }

      /// Operator ()
      return_type operator()(const Point_3& p, const bool = true)
	const
      {
	if (CGAL::squared_distance(p, Point_3(cx, cy, cz))>=R)
	  return 0;
	Vector_3 v(1,0,0);
	//	std::cout << p.x() << " " << p.y() << " " << p.z() << "\n";
	for ( size_t i = function_vector_.size();i--;)
	  //	for ( size_t i = 0; i < function_vector_.size(); ++i)
	{
	  Ray_3 ray(p, v);
	  //	  std::cout << i << std::endl;
	  //	  if ( function_vector_[i]->is_in_domain_object()(p)>0)
	  if(!function_vector_[i]->do_intersect(ray))
	    return 0;
	  int inter = function_vector_[i]->number_of_intersected_primitives(ray);
	  //          std::cout << inter << " yeee  " << std::endl;
	  if (inter&1)
	  {
	    return (static_cast<return_type>(i+1)) ;
	  }
        }
	return 0;
      }
    private:
      /// Functions to wrap
      Function_vector &function_vector_;
    };
  }
}




typedef CGAL::Exact_predicates_inexact_constructions_kernel K;
//typedef CGAL::Simple_cartesian<double> K;
typedef CGAL::Polyhedron_3<K> Polyhedron;
typedef CGAL::AABB_polyhedron_triangle_primitive<K,Polyhedron> Primitive;
typedef CGAL::AABB_traits<K, Primitive> Traits;
typedef CGAL::AABB_tree<Traits> Tree;
typedef CGAL::Polyhedral_mesh_domain_3<Polyhedron, K> Mesh_domain_poly;
typedef std::tr1::function<double(K::Point_3 const&)> Function ;
typedef
CGAL::Mesh_3::Tree_to_labeled_function_wrapper<Tree,
K> Function_wrapper;
typedef Function_wrapper::Function_vector 
Function_vector;
typedef CGAL::Mesh_3::Labeled_mesh_domain_3<Function_wrapper, K> 
Mesh_domain;



// Triangulation
typedef CGAL::Mesh_triangulation_3<Mesh_domain>::type Tr;
typedef CGAL::Mesh_complex_3_in_triangulation_3<Tr> C3t3;

// Criteria
typedef CGAL::Mesh_criteria_3<Tr> Mesh_criteria;
typedef CGAL::Mesh_constant_domain_field_3<Mesh_domain::R,
                                           Mesh_domain::Index> Sizing_field;

// // To avoid verbose function and named parameters call
// double sphere_function ( typename K::Point_3 const& p)
// { return CGAL::squared_distance(p, typename K::Point_3(5, 0, 0))-1; }
// double sphere_function1 ( typename K::Point_3 const& p)
// { return CGAL::squared_distance(p, typename K::Point_3(0, 0, 0))-1; }
using namespace CGAL::parameters;

typedef K::Point_3 Point;
struct Spherical_sizing_field
{
  typedef double FT;
  typedef Point Point_3;
  typedef Mesh_domain::Index Index;
  Point where;
  FT dist;
  FT size1, size2;
  FT operator()(const Point_3& p, const int, const Index&) const
  {
    FT cx, cy, cz;
    cx = 118;
    cy = 207;
    cz = 110;
    Point P(cx, cy, cz);
    cz = 110;
    Point P1(cx, cy-10, cz);
    FT sq_d_to_origin = CGAL::squared_distance(p, P);
    FT sq_d_to_origin1 =std::min( CGAL::squared_distance(p, P1),sq_d_to_origin) ;
    FT dist = CGAL::abs( CGAL::sqrt(sq_d_to_origin1)) / 30. + 0.01;
    if (dist >3)
      dist = 3;
    return dist;
  }
};

int main(int argc, char**argv)
{
  std::string fname = "mesh.inr";
  if (argc>1)
    fname = argv[1];


  Function_vector vec;
  std::vector<Polyhedron> vecp;
  for(int i = 2; i < argc; ++i)
  {
    vecp.push_back(Polyhedron());
    std::ifstream f(argv[i]);
    std::cout << "Reading " << argv[i] << std::endl;
    size_t j = vecp.size()-1;
    scan_OFF(f, vecp[j], true);//f >> vecp[j];
    Tree  *md = new Tree(vecp[j].facets_begin(), vecp[j].facets_end());
    std::cout << "Size of the polyhedron in mem " << vecp[j].bytes() << std::endl;
    std::cout << argv[i] << " " << md->size() << std::endl;
    md->accelerate_distance_queries();
    vec.push_back(md);
  }
  std::cout << "everithing is read " << std::endl;
  // center of the mesh;
  cx = 37+75;
  cy = 45+75;
  cz = 10+75;
  R = 1000*1000; //radius squared
  
  
  Mesh_domain domain(vec, K::Sphere_3(K::Point_3(75, 75, 75), 1000));


  std::cout << "Domain constructed" << std::endl;
  

  Point p(75+43, 75+132, 75+35);
  Spherical_sizing_field volume_size;
  Sizing_field face_size(3);

  
  Mesh_criteria criteria(facet_angle=20, facet_size=face_size, facet_distance=0.5,
			 cell_radius_edge=3, cell_size=volume_size);


  // Meshing
  C3t3 c3t3 = CGAL::make_mesh_3<C3t3>(domain, criteria);//, features(domain));//, lloyd(), odt(), perturb(), exude());
  std::cout << "Mesh constructed" << std::endl;

  // Output

  for (size_t i = 0; i < vec.size(); ++i)
    delete vec[i];

  std::ofstream medit_file((fname+".mesh").c_str());
  c3t3.output_to_medit(medit_file, true, true); // rebind?=true, //show surface patches = tru

  return 0;
}
