#include <iostream>
#include "mirorer.h"
#include <string>
#include <sstream>
#include <fstream>
#include <map>
#include <algorithm>
#include <vector>

int dimx, dimy, dimz;
double vx, vy, vz;

void Init()
{
  dimx = dimy = dimz = 1;
  vx = vy = vz = 1.0;
}

std::string GetName()
{
  std::ostringstream os;
  os << dimx << "x" << dimy << "x" << dimz << ".bonec";
  std::cout << "writing " << os.str() << std::endl;
  return os.str();
}

void CheckDimension(const std::string &line)
{
  std::istringstream is(line);
  std::string dim;
  is >>dim;
  if (dim == "DIMENSIONS")
    is >> dimx >> dimy >> dimz;
    
}
void CheckSpacing(const std::string &line)
{
  std::istringstream is(line);
  std::string dim;
  is >>dim;
  if (dim == "SPACING")
    is >> vx >> vy >> vz;
    
}

bool CheckExit(const std::string &line)
{
  std::istringstream is(line);
  std::string a;
  is >>a;
  return a != "COLOR_SCALARS";
}
 

template<class T>
void Renumber(vector3d<T> &in, vector3d<T> &out)
{
  std::map<T, T> map;
  typename std::map<T, T>::iterator it;
  
  out.SetSize(in.sizes[0], in.sizes[1], in.sizes[2]);
  for(size_t i = 0; i < in.sizes[0]; ++i)
    for(size_t j = 0; j < in.sizes[1]; ++j)
      for(size_t k = 0; k < in.sizes[2]; ++k)
      {
	T val = in(i, j, k);
	map[val] = val;
      }
  T cur = 0;
  for (it = map.begin(); it!=map.end(); ++it)
  {
    it->second = cur++;
  }
  for(size_t i = 0; i < in.sizes[0]; ++i)
    for(size_t j = 0; j < in.sizes[1]; ++j)
      for(size_t k = 0; k < in.sizes[2]; ++k)
	out(i, j, k) = map[in(i, j, k)];    
}

struct MaxComparator{
  template<class T>
  bool  operator()(const T& a, const T &b)
  {
    return a.second<b.second;
  }
};



template<class T>
void Resample(vector3d<T> const &in, vector3d<T> &out,
	      size_t ox, size_t oy, size_t oz,
	      double (&iv)[3],
	      double (&ov)[3],
	      double (&off)[3], int median)
{
  out.SetSize(ox, oy, oz);
  size_t I[3];
  size_t J[3];
  size_t K[3];
  size_t L[3];
  
  double coord[3];
  for (I[0]=0; I[0] < ox; ++I[0])
    for (I[1]=0; I[1] < oy; ++I[1])
      for (I[2]=0; I[2] < oz; ++I[2])
      {
	std::vector<T> vec;
	size_t res = 2*median+1;
	vec.reserve(res*res*res);
	for (K[0]=0; K[0] < res; ++K[0])
	  for (K[1]=0; K[1] < res; ++K[1])
	    for (K[2]=0; K[2] < res; ++K[2])
	    {
	      for (int i = 0; i < 3; ++i)
	      {
		L[i] = I[i]+K[i]-median;
		coord[i] = off[i] + L[i]*ov[i]+ov[i]*0.5;
		J[i] = coord[i]/iv[i];
		if (J[i]>=in.sizes[i])
		  J[i] = in.sizes[i]-1;
		if (coord[i]<0)
		  J[i] = 0;
	      }
	      vec.push_back(in(J[0], J[1], J[2]));
	      //	out(I[0], I[1], I[2]) = in(J[0], J[1], J[2]);
	    }
	std::sort(vec.begin(), vec.end());
	out(I[0], I[1], I[2]) = vec[vec.size()/2];
      }
}
  
       

void Process(const std::string &fname)
{
  std::ifstream file;
  file.exceptions(std::ios::failbit | std::ios::badbit |std::ios::eofbit);
  file.open(fname.c_str());
  std::string line;
  do{
    getline(file, line);
    CheckDimension(line);
    CheckSpacing(line);
  }while(CheckExit(line));
  std::cout << sizeof(unsigned char) << std::endl;
  vector3d<unsigned char> vec, veco;
  vec.SetSize(dimx, dimy, dimz);
  for(size_t k = 0; k < vec.sizes[2]; ++k)
    for(size_t j = 0; j < vec.sizes[1]; ++j)
      for(size_t i = 0; i < vec.sizes[0]; ++i)
      {
	char chars[2];
	file.read(chars, 1);
	vec(i,j, k) =  chars[0];
      }
  //  Renumber(vec, veco);
  double iv[3], ov[3], off[3];
  iv[0] = vx;
  iv[1] = vy;
  iv[2] = vz;

  dimx =vec.sizes[0];
  dimy=vec.sizes[1];
  dimz=vec.sizes[2];
  ov[0] = iv[0]*vec.sizes[0]/dimx;
  ov[1] = iv[1]*vec.sizes[1]/dimy;
  ov[2] = iv[2]*vec.sizes[2]/dimz;
  off[0] = off[1] = off[2] = 0.;
  vx = ov[0];
  vy = ov[1];
  vz = ov[2];
  // Resample(veco, vec, dimx, dimy, dimz,
  //  	   iv, ov, off, 1);
  vec.WriteNormalOrder(GetName());  
  vec.WriteInr(GetName()+".inr", ov[0], ov[1], ov[2]);  
}
    


//reads pint array vtk type short and wrirtes a bonec
int main(int argc, char ** argv)
{
  try{
    Init();
    std::string fname = "file.vtk";
    if (argc >1)
      fname = argv[1];
    Process(fname);
  }catch(std::exception const &e)
  {
    std::cout << e.what() << std::endl;
    throw;
  }
}
  
