#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <iomanip>
#include "mirorer.h"
#include "small.h"
#include "smallops.h"


const double PI = 4.*atan(1);

typedef SmallVector<double, 3> Point3d;
template<class T>
T sqr(T a)
{
  return a*a;
}
void Normalize(double &angle)
{
  const double p2 = 2*PI;
  while(angle < 0)
    angle += p2;
  while (angle >=p2)
    angle -= p2;
}
void CartToCyl(Point3d const &cc, Point3d const &center, Point3d &cyl)
{
  Point3d s = cc;
  s-=center;
  cyl[2] = s[2];// Z ready;
  double angle = atan2(s[1], s[0]);
  double r = sqrt(sqr(s[0])+sqr(s[1]));
  Normalize(angle);
  cyl[0] = angle;
  cyl[1] = r;

}

inline double DegToRad(double x)
{
  return x/180*PI;
}
inline double RadToDeg(double x)
{
  return x/PI*180;
}


void CylToCart(Point3d const &cyl, Point3d const &center, Point3d &cart)
{
  cart[2] = cyl[2];
  cart[0] = cyl[1]*sin(cyl[0]);
  cart[1] = cyl[1]*cos(cyl[0]);
  cart += center;
}

int GoodZ=4;

void PeriodicizeExact(double &angle, const double &AnglePeriod)
{
  int quadrant = angle/AnglePeriod;
  double rest = angle- quadrant*AnglePeriod;
  if (quadrant&1)
    rest = AnglePeriod-rest;
  angle = rest;
}
void PeriodicizeAny(double &angle, const double &AnglePeriod)
{
  if (angle >PI)
    angle = 2*PI-angle;
  PeriodicizeExact(angle, AnglePeriod);
}

double min_r, max_r;
double min_z, max_z;
int thres = 130;
template<class T, class V>
T ClippedValue(vector3d<T>const &vec, SmallVector<V, 3>const &x, double r, bool clip=false)
{

  //  std::cout << x[0] << " " << x[1] << " " << x[2] << std::endl;
  int X[3];
  for (int i = 0; i < 3; ++i)
  {
    X[i] = x[i]/vec.dim[i];//-0.5;
    if (X[i]<0)
      return 0;
    if (X[i]>=vec.sizes[i])
      return 0;
  }

  if (X[2]<=GoodZ)
    return 3;
  T val = vec(X[0], X[1], X[2]);
  if (clip)
  {
    if (X[2]>=max_z && (val <thres))
      return 4;

    if (r>max_r)
      return 1;
    if (r<min_r)
      return 2;
  }
  
  if (val <thres )
  {
    return 40; // this is air
  }
  
  return 255;
}


int main(int argc, char **argv)
{
  using namespace std;
  if (argc < 2)
    return -1;
  std::string fame = "132";
  if (argc >2)
    fame = argv[2];
  std::istringstream is1(fame);
  is1 >> thres;
    
  string fname = argv[1];

  vector3d<unsigned char> vec;
  vector3d<unsigned char>vec1;
  vec.ReadInr(fname);
  int N1 = vec.sizes[0];
  int N2 = vec.sizes[1];
  int N3 = vec.sizes[2];
  double d1 = vec.dim[0];
  double d2 = vec.dim[1];
  double d3 = vec.dim[2];
  GoodZ = 5*0.0009528/vec.dim[2] + 0.5;
  double Rm = 120.0*1e-3;
  max_r = 0;//Rm*0.95;
  min_r =  Rm;
  min_z = max_z = N3/2;//*d3/2;
  double R = Rm/d1;//100;
  std::cout << "The radius is "<< R*d1 << "m." << std::endl;
  int M1 = R*2+1+0.5;
  int M2 = R*2+1+0.5;
  int M3 = N3;
  vec1.SetSize(M1, M2, M3, d1, d2, d3);
  Point3d c,oc;
  c[0] = (R+0.5)*d1;
  c[1] = (R+0.5)*d2;
  c[2] = 0;
  std::cout << "center is " << c << std::endl;
  oc[0] = (N1+1)*d1*1.09;//152;//0.95;
  oc[1] = (N2+1)*d2/2;
  oc[2] = 0;
  double rotangle = DegToRad(90);
  for (int i = 0; i < M1; ++i)
    for (int j = 0; j < M2; ++j)
      for (int k = 0; k < M3; ++k)
      {
	Point3d p, cyl, cart;
	p[0] = (i+0.5)*d1;
	p[1] = (j+0.5)*d2;
	p[2] = (k+0.5)*d3;
	CartToCyl(p, c, cyl);
	PeriodicizeAny(cyl[0], rotangle);
	cyl[0]+=3*PI/2-rotangle/2;
	Normalize(cyl[0]);
	CylToCart(cyl, oc, cart);
	unsigned char ch =  ClippedValue(vec, cart, cyl[1]);
	if ((ch==255)&&(cyl[1]<min_r))
	  min_r = cyl[1];
	if ((ch==255)&&(cyl[1]>max_r))
	  max_r = cyl[1];
	if ((ch==255)&&(k>max_z))
	  max_z = k;
	if ((ch==255)&&(k<min_z))
	  min_z = k;
      }
  std::cout << "Min_r = " << min_r << ", max_r = "  << max_r << std::endl;
  double width = 4;
  min_r -= width*d2;
  max_r -= width*d2;
  max_z -= 1;
  min_z += 1;
  std::cout << "Min_r = " << min_r << ", max_r = "  << max_r << std::endl;
  for (int i = 0; i < M1; ++i)
    for (int j = 0; j < M2; ++j)
      for (int k = 0; k < M3; ++k)
      {
	Point3d p, cyl, cart;
	p[0] = (i+0.5)*d1;
	p[1] = (j+0.5)*d2;
	p[2] = (k+0.5)*d3;
	CartToCyl(p, c, cyl);
	PeriodicizeAny(cyl[0], rotangle);
	cyl[0]+=3*PI/2-rotangle/2;
	Normalize(cyl[0]);
	CylToCart(cyl, oc, cart);
	vec1(i,j,k) = ClippedValue(vec, cart, cyl[1], true);
      }
    
//   vec1(0,0,0) = 255;
//   vec1(M1/2,0,0) = 255;
  std::ostringstream fn;
  fn <<M1 << "x" << M2 << "x" <<M3 << ".cyclo." << fname;
  std::string ofname = fn.str();
  if (argc>3)
    ofname = argv[3];
  vec1.WriteInr(ofname);
//   std::ostringstream fn;
//   fn <<M1 << "x" << M2 << "x" <<M3 << ".cyclo.bonec";
//   vec1.Write(fn.str());
  return 0;
}
	  
    
  
