# include <vector>
# include <map>
# include <cmath>
# include <iostream>
# include <iomanip>
# include <fstream>
# include <string>
# include <sstream>
# include <stdexcept>
# include <GL/gl.h>
# include <GL/glu.h>
# include <GL/glut.h>
#include "functionlin.h"
#include "functionpol.h"			
extern "C" {
# include "gl2ps.h"
}
#include <sstream>
#include "mirorer.h"
#include <limits>


enum menu_items {
  SAVE_EPS,
  SAVE_PNG,
  SAVE_AVI,
  QUIT
};


enum Sides {
  LeftSide = 0x1,
  RightSide = 0x2,
  TopSide = 0x4,
  BottomSide = 0x8,
  FrontSide = 0x10,
  BackSide = 0x20
};



std::string filename;
std::string filename1;
std::string filenameI[3];

double movedistance = 0.0000;
double movesize = 0.0000330;
bool DirUp=true;
int CurMoveCount = 0;
int MaxMoveCount = 45;
double nceps = 1e-10;
double v1[3], v2[3], v3[3];
double ox[3] = {1,0,0}, oy[3]={0,1,0}, oz[3] = {0,0,1};
double xob[3] = {1, 0, 0}, xu[3]={0,1,0};
double bonethres = 0.5;
double minfixer = 0;
double maxfixer = 0;
double GrayCo = 0.75;
bool DrawSolid = true;
bool GrayBone = false;
bool Shading = true;
bool DrawIso = false;
bool ShowPlates = false;
bool ShowPalette = false;
bool RemoveColor = false;
bool UseLighting = false;
bool SolidColor = false;
int isobehav = 3;
int WireFrame = 0;
int CurLine=0;
int il = 20;
double trlines = 0.8;
double isocol = 0.050;
enum Dirs{Ndir, Xdir, Ydir, Zdir, Cutdir};
Dirs CurDir = Ndir;
int Resolution;
double mypol=1;//0.5;
  static  vector3d<unsigned char > vec1;
double gmax, gmin;



class DisplayList
{
  int ln;
public:
  DisplayList():ln(0)
  {
  }
  ~DisplayList()
  {
    if(ln)
      glDeleteLists(ln,1);
  }
  void CompileStart()
  {
    if(ln)
      glDeleteLists(ln,1);
    ln = glGenLists(1);
    glNewList(ln, GL_COMPILE);
  }
  void CompileFinish()
  {
    glEndList();
  }
  operator int ()
  {
    return ln;
  }
};



DisplayList IsoList;
DisplayList BoneList;
DisplayList SolidList;

LinearFunction3d<double, double, bool> bone;
LinearFunction3d<double, double, double> values[3];
PolyFunction3d<double, double> valuespol[3];
vector3d<unsigned char> SolidPart;



LinearFunction3d<double, double, double> valuess;
PolyFunction3d<double, double> valuespols;

double (*valuefuncs)(double, double, double);


double (*bonefunc)(double, double, double)=NULL;;
double (*valuefunc[3])(double, double, double);

double ConstantFunc(double, double, double)
{
  return 0.;
}

  
double LinearBone(double x, double y, double z)
{
  return bone(x, y, z);
}
double LinearFuncs(double x, double y, double z)
{
  assert(bonefunc);
  return valuess(x, y, z);//*(1-bonefunc(x,y,z));
}

template<int num>
double LinearFunc(double x, double y, double z)
{
  assert(bonefunc);
  return values[num](x, y, z);//*(1-bonefunc(x,y,z));
}
template<int num>
double PolynomFunc(double x, double y, double z)
{
  return valuespol[num](x, y, z);
}
double PolynomFuncs(double x, double y, double z)
{
  return valuespols(x, y, z);
}





template <class T>
void RemoveExtra(vector3d<T> &vec)
{
  for(int i = 0; i < vec.sizes[0]; i++)
    for(int j = 0; j < vec.sizes[1]; j++)
      for(int k = 0; k < vec.sizes[2]; k++)
      {
	if (i >0)
	{
	  bool lp = vec(i-1,j,k) & RightSide;
	  bool rp = vec(i, j, k) & LeftSide;
	  lp = lp and rp;
	  if (lp)
	  {
	    vec(i-1, j, k) &= ~ RightSide;
	    vec(i, j, k) &= ~LeftSide;
	  }
	}
	if (j >0)
	{
	  bool lp = vec(i,j-1,k) & BackSide;
	  bool rp = vec(i, j, k) & FrontSide;
	  lp = lp and rp;
	  if (lp)
	  {
	    vec(i, j-1, k) &= ~ BackSide;
	    vec(i, j, k) &= ~FrontSide;
	  }
	}
	if (k >0)
	{
	  bool lp = vec(i,j,k-1) & TopSide;
	  bool rp = vec(i, j, k) & BottomSide;
	  lp = lp and rp;
	  if (lp)
	  {
	    vec(i, j, k-1) &= ~TopSide;
	    vec(i, j, k) &= ~BottomSide;
	  }
	    
	}
      }
}


void GenerateSolid()
{
  
  vector3d<unsigned char> &vec = SolidPart;
  int &sz = Resolution;
  int &N = Resolution;
  SolidPart.SetSize(sz, sz, sz);
  switch(CurDir)
  {
  case Cutdir:
    for(int i = 0; i < sz; ++i)
      for(int j = 0; j<sz; ++j)
	for(int k=0; k < sz; ++k)
	{
	  if (!((i<9*sz/18)&&(j<sz*4/7)&&(k>sz/4)))
	    vec(i,j,k)= 255;
	  else
	    vec(i,j,k) = 0;
	}
    break;
  case Ndir:
    for(int i = 0; i < N; ++i)
      for(int j = 0; j< N; ++j)
	for(int k=0; k < N; ++k)
	{
	  double h = 0.5/N;
	  double on = 1./N;
	  double hi = i*on+h;
	  double hj = j*on+h;
	  double hk = k*on+h;
	  vec(i,j,k) = (bonefunc(hi, hj, hk)>=bonethres)?255:0;
	}
    break;
  }
  if((CurDir!=Ndir)&&(CurDir!=Cutdir))
  {
    for(int i = 0; i < N; ++i)
      for(int j = 0; j < N; ++j)
	for(int k = 0; k < N; ++k)
	  vec(i,j,k)=0;

  }
  if (CurDir!=Cutdir)
  {
    int Line = CurLine;
    for(int i = 0; i < N; ++i)
      for(int j = 0; j < N; ++j)
      {
	switch(CurDir)
	{
	case Ndir:
	  if(ShowPlates)
	  {
	    vec(i,j, 0) = 255;
	    vec(i,j,N-1)=255;
	  }
	  break;
	case Xdir:
	  vec(Line, i,j) = 255;
	  break;
	case Ydir:
	  vec(i,Line, j) = 255;
	  break;
	case Zdir:
	  vec(i,j,Line) = 255;
	  break;
	}
      }
  }

  RemoveExtra(vec);  
}


template<int num>
void ReadValues()
{
  SmallVector<double, 3> min, max;
  min = 0;
  max = 1;
  int N;
  int k;
  char type;
  {
    char ch = 'z';
    std::istringstream is(filenameI[num]);
    is >> N;
    is >> ch >> type >>ch >>k;
    N *=k;
  }
  if (type == 'f')
  {
    vector3d<double> im;
    im.SetSize(N, N, N);
    try{
      im.Read(filenameI[num], 0);
    }catch (...)
    {
      std::cout << N << " " <<  std::endl;
      int sz = 2;
      im.SetSize(sz,sz,sz);
      for(int i = 0; i < sz; ++i)
	for(int j = 0; j<sz; ++j)
	  for(int k=0; k < sz; ++k)
	  {
	    im(i,j,k) = 1.;
	  }
    }
    values[num].Init(im, min, max, false, true);
    valuefunc[num] = LinearFunc<num>;
  }
  else  if (type == 'g')
  {
    valuespol[num].Init(filenameI[num], min, max);
    valuefunc[num] = PolynomFunc<num>;
  }else
    valuefunc[num] = ConstantFunc;
}

void ReadData()
{
  int sz;
  SmallVector<double, 3> min, max;
  min = 0;
  max = 1;
  {
    std::istringstream is(filename);
    is >> sz;
  }
  vector3d<unsigned char > vec;
  vec.SetSize(sz, sz, sz);
  try{
    vec.Read(filename, 0);
    if (Resolution == 0)
      Resolution = sz;
  }catch(...)
  {
    sz = 3;
    vec.SetSize(sz,sz,sz);
      for(int i = 0; i < sz; ++i)
	for(int j = 0; j<sz; ++j)
	  for(int k=0; k < sz; ++k)
	  {
	      vec(i,j,k) = 0;
	  }
  }

  bone.Init(vec, min, max, false, false); //no inter- & exter-polation for the bone
  bonefunc = LinearBone;
  // bone read finished

  ReadValues<0>();
  ReadValues<1>();
  ReadValues<2>();

  int N;
  int k;
  char type;
  {
    char ch = 'z';
    std::istringstream is(filename1);
    is >> N;
    is >> ch >> type >>ch >>k;
    N *=k;
  }
  if (type == 'f')
  {
    vector3d<double> im;
    im.SetSize(N, N, N);
    try{
      im.Read(filename1, 0);
      if (Resolution == 0)
	Resolution = N/k;
    }catch (...)
    {
      std::cout << N << " " <<  std::endl;
      sz = 2;
      im.SetSize(sz,sz,sz);
      for(int i = 0; i < sz; ++i)
	for(int j = 0; j<sz; ++j)
	  for(int k=0; k < sz; ++k)
	  {
	    im(i,j,k) = 1.;
	  }
    }
    valuess.Init(im, min, max, false, true);
    valuefuncs = LinearFuncs;
  }
  else  if (type == 'g')
  {
    valuespols.Init(filename1, min, max);
    valuefuncs = PolynomFuncs;
  }else
    valuefuncs = ConstantFunc;
  

}



double Cos(double v1[3], double v2[3], double v3[3])
{
  for(int i = 0; i < 3;++i)
  {
    v1[i] -= v3[i];
    v2[i] -= v3[i];
  }
  double l1 = 0;
  double l2 = 0;
  for (int i = 0; i < 3; ++i)
  {
    l1 += v1[i]*v1[i];
    l2 += v2[i]*v2[i];
  }
  l1 = sqrt(l1);
  l2 = sqrt(l2);
  double cos = 0;
  for (int i = 0; i < 3; ++i)
  {
    cos += v1[i]*v2[i];
  }
  cos /= l1*l2;
  return cos;    
}

template<class T>
struct Quad
{
  T w,x,y,z;
  Quad():w(), x(), y(), z(){}
  Quad(T ax, T ay, T az, T angle):
    w(cos(angle/2)),
    x(ax*sin(angle/2)),
    y(ay*sin(angle/2)),
    z(az*sin(angle/2))
  {}
  void GetMatrix(T (&A)[3][3])
  {
    A[0][0] = w*w + x*x - y*y - z*z; A[0][1] = 2*x*y - 2*w*z;        A[0][2] = 2*x*z + 2*w*y;
    A[1][0] = 2*x*y + 2*w*z;         A[1][1] = w*w - x*x + y*y - z*z;A[1][2] = 2*y*z - 2*w*x;
    A[2][0] = 2*x*z - 2*w*y;         A[2][1] = 2*y*z + 2*w*x;        A[2][2] = w*w - x*x - y*y + z*z;
  }
};
template<class T>
Quad<T> operator*(const Quad<T> &a, const Quad<T>&b)
{
  Quad<T> r;
  r.w = a.w*b.w - a.x*b.x - a.y*b.y - a.z*b.z;
  r.x = a.w*b.x + a.x*b.w + a.y*b.z - a.z*b.y;
  r.y = a.w*b.y + a.y*b.w + a.z*b.x - a.x*b.z;
  r.z = a.w*b.z + a.z*b.w + a.x*b.y - a.y*b.x;
  return r;
}

  

using namespace std;

double MIN_X,MAX_X, MIN_Y, MAX_Y, MIN_Z, MAX_Z;
double vsx, vsy, vsz;
double minx, miny, minz, maxx, maxy, maxz;
double xc, yc, zc;

float angle=atan(1)*2,ratio = 1;
bool usecube = false;
bool swapp=true;
float angle1=0;
float angle2=0;
GLint snowman_display_list;
double dist = 2.5;

double
bmin[2],  bmax[2];
double amin, amax;
std::vector<double> Dmins[3][2];
std::vector<double> Dmaxs[3][2];

void DoMinMax(const double&val, double&min, double &max)
{
  if (val<min)
    min = val;
  if (val>max)
    max = val;
}
double GenMins()
{
  const double MIN = std::numeric_limits<double>::max();
  const double MAX = -std::numeric_limits<double>::max();
  int R = 1;
  int N = Resolution*R;
  for(int i = 0; i < 2; ++i)
  {
    bmin[i] = MIN;
    bmax[i] = MAX;
    for(int j = 0; j < 3; ++j)
    {
      Dmins[j][i].resize(N/R);
      Dmaxs[j][i].resize(N/R);
      for(int k = 0; k < N; ++k)
      {
	Dmins[j][i][k/R] = MIN;
	Dmaxs[j][i][k/R] = MAX;
      }
    }
  }
  double h = 0.5/N;
  double on = 1./N;
  for(int i = 0; i < N; ++i)
    for (int j = 0; j < N; ++j)
      for(int k = 0; k < N; ++k)
      {
	  double hi = i*on+h;
	  double hj = j*on+h;
	  double hk = k*on+h;
	  double val = valuefuncs(hi, hj, hk);
	  bool isb = (bonefunc(hi, hj, hk)>=bonethres);
	  DoMinMax(val, bmin[isb], bmax[isb]);
	  DoMinMax(val, Dmins[0][isb][i/R], Dmaxs[0][isb][i/R]);
	  DoMinMax(val, Dmins[1][isb][j/R], Dmaxs[1][isb][j/R]);
	  DoMinMax(val, Dmins[2][isb][k/R], Dmaxs[2][isb][k/R]);
      }
}

void SetMinMax(int i=6)
{
  if (i==6)
  {
    Dirs olddir = CurDir;
    if(isobehav>=3)
    {
      CurDir = Ndir;
    }      
    SetMinMax(isobehav%3);
    CurDir = olddir;
    
    double size = gmax - gmin;
    amin = gmin += size*minfixer/30;
    amax = gmax += size*maxfixer/30;
    if (gmin>=gmax)
      gmax = gmin + nceps;
    //    std::cout << gmin << " " << gmax << std::endl;
    return;
  }
  if (i==2)
  {
    double tmin, tmax;
    SetMinMax(0);
    tmin = amin;
    tmax = amax;
    SetMinMax(1);
    if (tmin<gmin)
      amin = gmin = tmin;
    if (tmax>gmax)
      amax = gmax = tmax;
  }else
  {
    
    if (CurDir==Ndir)
    {
      amin = bmin[i];
      amax = bmax[i];
    }
    if (CurDir==Cutdir)
    {
      amin = bmin[i];
      amax = bmax[i];
    }
    if (CurDir==Xdir)
    {
      amin = Dmins[0][i][CurLine];
      amax = Dmaxs[0][i][CurLine];
    }
    if (CurDir==Ydir)
    {
      amin = Dmins[1][i][CurLine];
      amax = Dmaxs[1][i][CurLine];
    }
    if (CurDir==Zdir)
    {
      amin = Dmins[2][i][CurLine];
      amax = Dmaxs[2][i][CurLine];
    }
    
  }
  

  gmin = amin;
  gmax = amax;
}


template<class T>
void MulSmallMat(T (&res)[3][3],const T (&A)[3][3],const T(&B)[3][3])
{
  for(int i = 0; i < 3; ++i)
    for(int j = 0; j < 3; ++j)
    {
      T sum = 0;
      for(int k = 0; k < 3; ++k)
	sum += A[i][k] * B[k][j];
      res[i][j]=sum;
    }
}
template<class T>
void MulSmallMatVec(T (&res)[3],const T (&A)[3][3],const T(&x)[3])
{
  for(int i = 0; i < 3; ++i)
  {
    double sum = 0;
    for(int k = 0; k < 3; ++k)
      sum += A[i][k] * x[k];
    res[i]=sum;
  }
}
template<class T>
void Trans(const T (&A)[3][3], T(&x)[3])
{
  T t[3];t[0] = x[0];t[1] = x[1]; t[2] = x[2];
  MulSmallMatVec(x, A, t);
}

void orientMe()
{
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  double x = xc;//(MIN_X + MAX_X)/2;
  double y = yc;//(MIN_Y + MAX_Y)/2;
  double z = zc;// (MIN_Z + MAX_Z)/2;
  //    double matx[3][3], maty[3][3], matz[3][3], tmp[3][3];
  //    static double v1[3];
  //    static double v[3];
  //    static double t[3];
  //    matx[0][0] = 1;           matx[0][1] = 0;             matx[0][2] = 0;
  //    matx[1][0] = 0;           matx[1][1] = cos(angle);    matx[1][2] = -sin(angle);
  //    matx[2][0] = 0;           matx[2][1] = sin(angle);    matx[2][2] = cos(angle);

  //    maty[0][0] = cos(angle1); maty[0][1] = 0;             maty[0][2] = sin(angle1);
  //    maty[1][0] = 0;           maty[1][1] = 1;             maty[1][2] = 0;
  //    maty[2][0] = -sin(angle1);maty[2][1] = 0;             maty[2][2] = cos(angle1);

  //    matz[0][0] = cos(angle2); matz[0][1] =-sin(angle2);   matz[0][2] = 0;
  //    matz[1][0] = sin(angle2); matz[1][1] =cos(angle2);    matz[1][2] = 0;
  //    matz[2][0] = 0;           matz[2][1] = 0;             matz[2][2] = 1;
  //    //   cout << angle << " " << angle1 << " " << angle2 << endl;
  //    //   MulSmallMat(tmp, matx, maty);
  //    //   MulSmallMat(matx, tmp, matz);
  //    double x = xc;//(MIN_X + MAX_X)/2;
  //    double y = yc;//(MIN_Y + MAX_Y)/2;
  //    double z = zc;// (MIN_Z + MAX_Z)/2;
  //    v[0] = dist;
  //    v[1] = 0;
  //    v[2] = 0;
  //    v1[0] = v[0];
  //    v1[1] = v[1];
  //    v1[2] = v[2];
  //    //   MulSmallMatVec(v1, matx, v);
  //    MulSmallMatVec(v, maty, v1);
  //    MulSmallMatVec(v1, matz, v);
   
  //    double xo = v1[0];
  //    double yo = v1[1];
  //    double zo = v1[2];
  //    v[0] = 0;
  //    v[1] = 1;
  //    v[2] = 0;
  //    v1[0] = v[0];
  //    v1[1] = v[1];
  //    v1[2] = v[2];
  //    //   MulSmallMatVec(v1, matx, v);
  //    MulSmallMatVec(v, maty, v1);
  //    MulSmallMatVec(v1, matz, v);
  //    double xu = v1[0];
  //    double yu = v1[1];
  //    double zu = v1[2];
  // //    if (angle1>atan(1)*2)
  // //      angle1=atan(1)*1.999;
  // //    if (angle1<-atan(1)*2)
  // //      angle1=-atan(1)*1.999;

  // //    double xo, yo, zo;
  // //    xo = cos(angle)*dist;
  // //    zo = sin(angle)*dist;

  // //    yo = cos(angle1)*xo + sin(angle1)*zo;
  // //    zo = sin(angle1)*xo + cos(angle1)*zo;
   
  // //    double xo = cos(angle1)*cos(angle)*dist;
  // //    double yo = sin(angle1)*dist;
  // //    double zo = cos(angle1)*sin(angle)*dist;

  static bool firsttime = false;

  Quad<double> q1 = Quad<double>(ox[0], ox[1], ox[2], angle);
  Quad<double> q2 = Quad<double>(oy[0], oy[1], oy[2], angle1);
  Quad<double> q3 = Quad<double>(oz[0], oz[1], oz[2], angle2);
  Quad<double> q4;
  q4 = q1*q2*q3;
  double Mat[3][3];
  q4.GetMatrix(Mat);
  Trans(Mat, xob);
  Trans(Mat, xu);
  if (swapp)
  {
    Trans(Mat, ox);
    Trans(Mat, oy);
    Trans(Mat, oz);
  }
  double xo[3];
  xo[0] = dist*xob[0];
  xo[1] = dist*xob[1];
  xo[2] = dist*xob[2];
  for (int i = 0; i < 3; ++i)
    v1[i]=v2[i] = v3[i] = 0.;
  v2[0] = .2;
  v2[1] = .2;
  v1[0] = x;
  v1[1] = y;
  v1[2] = z;
   
  gluLookAt (x-xo[0], y-xo[1], z-xo[2],  x, y, z, xu[0], xu[1], xu[2]);
  angle = angle1 = angle2 = 0;

}





double atest = 1.;




void color(double d)
{
  d *= 4;
  double r, g, b;
  if (d < 1.) {
    r = 0.; g = d; b = 1.;
  } else if (d < 2.) {
    r = 0.; g = 1.; b = 2. - d;
  } else if (d < 3.) {
    r = d - 2; g = 1.; b = 0.;
  } else {
    r = 1.; g = 4 - d; b = 0.;
  }
  glColor4d(r, g, b, 0.7);

}
void mat(double cr, double cg, double cb, double tr, double mul=1)
{
//   if(RemoveColor)
//   {
//     GLfloat amb[] = {1,1,1,1};    
//     glColor4fv(amb);
//   }
//   else
  {
    GLfloat amb[] = {cr*mul, cg*mul, cb*mul, tr*mul};
    glColor4fv(amb);
  }

//   {
//     GLfloat amb[] = {1,1,1,1};
//     //   glMaterialf(GL_FRONT, GL_SHININESS, 3);
//     {
//       GLfloat amb[] = {0,0,0,1};
//       glMaterialfv(GL_FRONT, GL_AMBIENT, amb);
//     }
//     //   double in =0.2;
//     //   amb[0]*=in;
//     //   amb[1]*=in;
//     //   amb[2]*=in;
//     //    glMaterialfv(GL_FRONT, GL_DIFFUSE, amb);
//     //   glMaterialfv(GL_FRONT, GL_SPECULAR, amb);
//   }
}
struct P3D
{
  double X[3];
};

bool IntersectLine(P3D p1, P3D p2, double v1, double v2, double val, P3D &res)
{
  double f1 = min(v1, v2);
  double f2 = max(v1, v2);
  if ((val <f1) || (val >f2))
    return false;
  if (v2==v1)
    return false;
  double k = (val - v1)/(v2-v1);
  double k1 =  k1 = 1 - k;
  for(int i = 0; i <3; ++i)
    res.X[i] = p1.X[i]*k1 + p2.X[i]*k;
  return true;
}

void DoSide( P3D p[4], double v[4])
{
  for(int i = 0; i < il; ++i)
  {
    
    double curval = gmin + (gmax-gmin)*(i+0.5)/(il);
     double val1 = (curval - gmin)/(gmax-gmin);
    val1 = val1*2 - 1;
    val1 = (val1<0)?
      -pow(fabs(val1), mypol):
      pow(fabs(val1), mypol);
    val1 = (val1+1)/2;
    curval = val1*(gmax-gmin) +gmin;
    
    
    P3D points[4];
    int maxpoint = 0;
    if(IntersectLine(p[0], p[1], v[0], v[1], curval, points[maxpoint]))
      maxpoint++;
    if(IntersectLine(p[2], p[1], v[2], v[1], curval, points[maxpoint]))
      maxpoint++;
    if(IntersectLine(p[2], p[3], v[2], v[3], curval, points[maxpoint]))
      maxpoint++;
    if(IntersectLine(p[0], p[3], v[0], v[3], curval, points[maxpoint]))
      maxpoint++;
    if(maxpoint==0)
        continue;
    glBegin(GL_LINES);
    for (int i = 0; i < maxpoint;++i)
    {
      glColor4f(0,0,0,1);
      glVertex3d(points[i].X[0], points[i].X[1], points[i].X[2]);
    }
    glEnd();
  }
}



template <class T>
inline bool In(const T&a, const T&min, const T &max)
{
  return (a >=min) && (a<max);
}
template<class T, class I>
bool check(T &vec, I i, I j, I k)
{
  ptrdiff_t szs[3];
  bool ret = true;
  if (i>0)
    ret = ret && vec(i-1, j, k);
  else
    ret = false;
  if (j>0)
    ret = ret && vec(i, j-1, k);
  else
    ret = false;
  if (k>0)
    ret = ret && vec(i, j, k-1);
  else
    ret = false;
  if (i<vec.sizes[0]-1)
    ret  = ret && vec(i+1, j, k);
  else
    ret = false;
  if (j<vec.sizes[1]-1)
    ret  = ret && vec(i, j+1, k);
  else
    ret = false;
  if (k<vec.sizes[2]-1)
    ret = ret && vec(i, j, k+1);
  else
    ret = false;
  bool tmp = vec(i,j,k);
  int sz1, sz2, sz3;
  sz1 = vec.sizes[0]/2;
  sz2 = vec.sizes[1]/2;
  sz3 = vec.sizes[2]/2;
  //     if ((i<sz1)&&(j<sz2) && (k <sz3))
  //       return false;
  if ((i==sz1)||(j==sz2) || (k==sz3))
    return vec(i,j,k);
  return tmp^ret;
}

struct color4
{
  double r,g,b,a;
  color4():r(), g(), b(), a(){}
  color4(double r_, double g_, double b_, double a_)
  {
    r = r_;
    g = g_;
    b = b_;
    a = a_;
  }
  void Set(double r_, double g_, double b_, double a_)
  {
    r = r_;
    g = g_;
    b = b_;
    a = a_;
  }
  void operator *=(const color4&a)
  { 
    r *= a.r;
    g *= a.g;
    b *= a.b;
    this->a *= a.a;
    return void(*this);//!!!
  }
};

std::vector<color4> cols, colwhite;

void InitColors()
{
  double ccc = 0.4;
  colwhite.push_back(color4(ccc,ccc,ccc,1));
  colwhite.push_back(color4(ccc,ccc,ccc,1));
  CurLine = Resolution/2;
  double alfa = 1;
  //  cols.push_back(color4(0,   0, 0.5, alfa*1));
  cols.push_back(color4(0,   0, 1, alfa*1));
  cols.push_back(color4(0,   1, 1, alfa*0.1));
  cols.push_back(color4(0,   .9, 0, alfa*0.02));
  cols.push_back(color4(1,   1, 0, alfa*0.1));
  cols.push_back(color4(1,   0, 0, alfa*1));
  //  cols.push_back(color4(0.5,   0, 0, alfa*1));
}

void GetColor(std::vector<color4>&cols, double val, color4&res)
{
  if (isnan(val))
    val = 0.5;
  if (val <0)
    val = 0;
  if (val >1)
    val = 1;
  int size = cols.size();
  if (size <2)
  {
    if (size==1)
      res = cols[0];
    else
      res = color4(.8,.8,.8,1);
    return;
  }
  val *= (size-1);
  int ival = static_cast<int>(trunc(val));
  val -= ival;
  color4 &l=cols[ival];
  if (ival==(size-1))
  {
    res = l;
    return;
  }
  color4 &r=cols[ival+1];
  res.r = l.r*(1-val) + r.r*val;
  res.g = l.g*(1-val) + r.g*val;
  res.b = l.b*(1-val) + r.b*val;
  res.a = l.a*(1-val) + r.a*val;
}





void SetCol(double val, double mul, bool isbone)
{
  double val1 = (val - amin)/(amax-amin);
  val1 = 1-val1;
  val1 = val1*2 - 1;
  val1 = (val1<0)?
    -pow(fabs(val1), mypol):
    pow(fabs(val1), mypol);	  
  val1 = (val1+1)/2;
  color4 c;
  //  std::cout << amin << " " << amax <<" " << val << " "  << val1 << endl;
  GetColor(cols, val1, c);
  if(!Shading)
  {
    c.r = c.g = c.b = c.a = 1;
  }
  if (GrayBone&&isbone)
  {
    c.r = c.g = c.b  = GrayCo;
    c.a = 1;
  }
  mat(c.r*mul, c.g*mul, c.b*mul, c.a);
}
    

void DrawHex(double x1, double x2, double x3,
	     double o1, double o2, double o3, unsigned char sides
	     )
{
  bool isbone;
  isbone = bonefunc(x1 + 0.5*o1, x2 + 0.5*o2, x3 + 0.5*o3)>=bonethres;
  double val;
  val = valuefuncs(x1 + 0.5*o1, x2 + 0.5*o2, x3 + 0.5*o3);
  
  double eps = nceps;
  double vx1, vx2, vx3, vo1, vo2, vo3;
  double nx1, nx2, nx3;
  SmallVector<SmallVector<double, 3>, 8> X, Y;
  for(int i = 0; i < 8; ++i)
  {
    X[i][0] = x1;
    X[i][1] = x2;
    X[i][2] = x3;
  }
  X[1][0] += o1;
  X[3][0] += o1;
  X[5][0] += o1;
  X[7][0] += o1;

  X[2][1] += o2;
  X[3][1] += o2;
  X[6][1] += o2;
  X[7][1] += o2;

  X[4][2] += o3;
  X[5][2] += o3;
  X[6][2] += o3;
  X[7][2] += o3;

  for (int i = 0; i < 8; ++i)
  {
    for(int j = 0; j < 3; ++j)
      Y[i][j] = X[i][j] + valuefunc[j](X[i][0], X[i][1], X[i][2])*movedistance;
  }
  nx1 = x1 + valuefunc[0](x1+o1/2, x2 + o2/2, x3 + o3/2)*movedistance;
  nx2 = x2 + valuefunc[1](x1+o1/2, x2 + o2/2, x3 + o3/2)*movedistance;
  nx3 = x3 + valuefunc[2](x1+o1/2, x2 + o2/2, x3 + o3/2)*movedistance;
  x1 = nx1;
  x2 = nx2;
  x3 = nx3;
  vx1 = x1 + eps;
  vx2 = x2 + eps;
  vx3 = x3 + eps;
  vo1 = o1 - 2*eps;
  vo2 = o2 - 2*eps;
  vo3 = o3 - 2*eps;
  double m0, m1, m2;
  m0 = 1.;
  if (Shading)
  {
    m1 = 0.9;
    m2 = 0.8;
  }
  else
    m1 = m2 = 1.;
  double mul = 0.5+(x1+x2+x3)*0.5/3.;
  if(Shading)
  {
    m1*=mul;
    m0*=mul;
    m2*=mul;
  }
  double div = 3331; //! what a constant!
  int mode = GL_QUADS;//GL_LINE_LOOP;
  
  SolidColor = true;

  if (sides)
  {
    glBegin(mode);

    if(SolidColor)
      SetCol(val, m0, isbone);
    if(sides & BottomSide)
    {
      if (UseLighting) glNormal3d(-o3/div, -o3/div, -o3);
      if (!SolidColor)
      {
	// val = valuefunc(vx1, vx2, vx3);
	SetCol(val, m0, isbone);
      }
      //      glVertex3d(x1, x2, x3);
      {int v = 0; glVertex3d(Y[v][0], Y[v][1], Y[v][2]);}
      if (UseLighting) glNormal3d(-o3/div, o3/div, -o3);
      if (!SolidColor)
      {
	// val = valuefunc(vx1, vx2+vo2, vx3);
	SetCol(val, m0, isbone);
      }
      //glVertex3d(x1, x2+o2, x3);
      {int v = 2; glVertex3d(Y[v][0], Y[v][1], Y[v][2]);}

      if (UseLighting) glNormal3d(o3/div, o3/div, -o3);
      if (!SolidColor)
      {
	// val = valuefunc(vx1+vo1, vx2+vo2, vx3);
	SetCol(val, m0, isbone);
      }
      //glVertex3d(x1+o1, x2+o2, x3);
      {int v = 3; glVertex3d(Y[v][0], Y[v][1], Y[v][2]);}

      if (UseLighting) glNormal3d(o3/div, -o3/div, -o3);
      if (!SolidColor)
      {
	// val = valuefunc(vx1+vo1, vx2, vx3);
	SetCol(val, m0, isbone);
      }
      //      glVertex3d(x1+o1, x2, x3);
      {int v = 1; glVertex3d(Y[v][0], Y[v][1], Y[v][2]);}

      
    }
    if (sides & TopSide)
    {
      
      if (UseLighting) glNormal3d(o3/div, -o3/div, o3);
      if (!SolidColor)
      {
	// val = valuefunc(vx1+vo1, vx2, vx3+vo3);
	SetCol(val, m0, isbone);
      }
      //      glVertex3d(x1+o1, x2,    x3+o3);
      {int v = 5; glVertex3d(Y[v][0], Y[v][1], Y[v][2]);}

      if (UseLighting) glNormal3d(o3/div, o3/div, o3);
      if (!SolidColor)
      {
	// val = valuefunc(vx1+vo1, vx2+vo2, vx3+vo3);
	SetCol(val, m0, isbone);
      }
      //      glVertex3d(x1+o1, x2+o2, x3+o3);
      {int v = 7; glVertex3d(Y[v][0], Y[v][1], Y[v][2]);}

      if (UseLighting) glNormal3d(-o3/div, o3/div, o3);
      if (!SolidColor)
      {
	// val = valuefunc(vx1, vx2+vo2, vx3+vo3);
	SetCol(val, m0, isbone);
      }
      //      glVertex3d(x1,    x2+o2, x3+o3);
      {int v = 6; glVertex3d(Y[v][0], Y[v][1], Y[v][2]);}
      
      if (UseLighting) glNormal3d(-o3/div, -o3/div, o3);
      if (!SolidColor)
      {
	// val = valuefunc(vx1, vx2, vx3+vo3);
	SetCol(val, m0, isbone);
      }
      //glVertex3d(x1,    x2,    x3+o3);
      {int v = 4; glVertex3d(Y[v][0], Y[v][1], Y[v][2]);}


      
    }

    if(SolidColor)
      SetCol(val, m1, isbone);
    if (sides & FrontSide)
    {
      
      if (UseLighting) glNormal3d(-o2/div, -o2, -o2/div);
      if (!SolidColor)
      {
	// val = valuefunc(vx1, vx2, vx3);
	SetCol(val, m1, isbone);
      }
      //glVertex3d(x1,    x2,    x3);
      {int v = 0; glVertex3d(Y[v][0], Y[v][1], Y[v][2]);}
      
      if (UseLighting) glNormal3d(o2/div, -o2, -o2/div);
      if (!SolidColor)
      {
	// val = valuefunc(vx1+vo1, vx2, vx3);
	SetCol(val, m1, isbone);
      }
      //      glVertex3d(x1+o1, x2,    x3);
      {int v = 1; glVertex3d(Y[v][0], Y[v][1], Y[v][2]);}

      if (UseLighting) glNormal3d(o2/div, -o2, o2/div);
      if (!SolidColor)
      {
	// val = valuefunc(vx1+vo1, vx2, vx3+vo3);
	SetCol(val, m1, isbone);
      }
      //      glVertex3d(x1+o1, x2,    x3+o3);
      {int v = 5; glVertex3d(Y[v][0], Y[v][1], Y[v][2]);}

      if (UseLighting) glNormal3d(-o2/div, -o2, o2/div);
      if (!SolidColor)
      {
	// val = valuefunc(vx1, vx2, vx3+vo3);
	SetCol(val, m1, isbone);
      }
      //      glVertex3d(x1,    x2,    x3+o3);
      {int v = 4; glVertex3d(Y[v][0], Y[v][1], Y[v][2]);}
      
    }
    if (sides & BackSide)
    {
      

      if (UseLighting) glNormal3d(o2/div, o2, -o2/div);
      if (!SolidColor)
      {
	// val = valuefunc(vx1+vo1, vx2+vo2, vx3);
	SetCol(val, m1, isbone);
      }
      //      glVertex3d(x1+o1, x2+o2,    x3);
      {int v = 3; glVertex3d(Y[v][0], Y[v][1], Y[v][2]);}


      if (UseLighting) glNormal3d(-o2/div, o2, -o2/div);
      if (!SolidColor)
      {
	// val = valuefunc(vx1, vx2+vo2, vx3);
	SetCol(val, m1, isbone);
      }
      //      glVertex3d(x1,    x2+o2,    x3);
      {int v = 2; glVertex3d(Y[v][0], Y[v][1], Y[v][2]);}
      
      if (UseLighting) glNormal3d(-o2/div, o2, o2/div);
      if (!SolidColor)
      {
	// val = valuefunc(vx1, vx2+vo2, vx3+vo3);
	SetCol(val, m1, isbone);
      }
      //      glVertex3d(x1,    x2+o2,    x3+o3);
      {int v = 6; glVertex3d(Y[v][0], Y[v][1], Y[v][2]);}

      if (UseLighting) glNormal3d(o2/div, o2, o2/div);
      if (!SolidColor)
      {
	// val = valuefunc(vx1+vo1, vx2+vo2, vx3+vo3);
	SetCol(val, m1, isbone);
      }
      //      glVertex3d(x1+o1, x2+o2,    x3+o3);
      {int v = 7; glVertex3d(Y[v][0], Y[v][1], Y[v][2]);}

      
    }
    if(SolidColor)
      SetCol(val, m2, isbone);
    if(sides & LeftSide)
    {
      
      if (UseLighting) glNormal3d(-o1, -o1/div, -o1/div);
      if (!SolidColor)
      {
	// val = valuefunc(vx1, vx2, vx3);
	SetCol(val, m2, isbone);
      }
      //      glVertex3d(x1,    x2,    x3);
      {int v = 0; glVertex3d(Y[v][0], Y[v][1], Y[v][2]);}

      if (UseLighting) glNormal3d(-o1, -o1/div, o1/div);
      if (!SolidColor)
      {
	// val = valuefunc(vx1, vx2, vx3+vo3);
	SetCol(val, m2, isbone);
      }
      //      glVertex3d(x1,    x2,    x3+o3);
      {int v = 4; glVertex3d(Y[v][0], Y[v][1], Y[v][2]);}

      if (UseLighting) glNormal3d(-o1, o1/div, o1/div);
      if (!SolidColor)
      {
	// val = valuefunc(vx1, vx2+vo2, vx3+vo3);
	SetCol(val, m2, isbone);
      }
      //      glVertex3d(x1,    x2+o2,    x3+o3);
      {int v = 6; glVertex3d(Y[v][0], Y[v][1], Y[v][2]);}

      if (UseLighting) glNormal3d(-o1, o1/div, -o1/div);
      if (!SolidColor)
      {
	// val = valuefunc(vx1, vx2+vo2, vx3);
	SetCol(val, m2, isbone);
      }
      //      glVertex3d(x1,    x2+o2,    x3);
      {int v = 2; glVertex3d(Y[v][0], Y[v][1], Y[v][2]);}

      
    }

    if (sides & RightSide)
    {
      

      if (UseLighting) glNormal3d(o1, o1/div, -o1/div);
      if (!SolidColor)
      {
	// val = valuefunc(vx1+vo1, vx2+vo2, vx3);
	SetCol(val, m0, isbone);
      }
      //      glVertex3d(x1+o1,    x2+o2,    x3);
      {int v = 3; glVertex3d(Y[v][0], Y[v][1], Y[v][2]);}

      if (UseLighting) glNormal3d(o1, o1/div, o1/div);
      if (!SolidColor)
      {
	// val = valuefunc(vx1+vo1, vx2+vo2, vx3+vo3);
	SetCol(val, m0, isbone);
      }
      //      glVertex3d(x1+o1,    x2+o2,    x3+o3);
      {int v = 7; glVertex3d(Y[v][0], Y[v][1], Y[v][2]);}

      if (UseLighting) glNormal3d(o1, -o1/div, o1/div);
      if (!SolidColor)
      {
	// val = valuefunc(vx1+vo1, vx2, vx3+vo3);
	SetCol(val, m0, isbone);
      }
      //      glVertex3d(x1+o1,    x2,    x3+o3);
      {int v = 5; glVertex3d(Y[v][0], Y[v][1], Y[v][2]);}

      if (UseLighting) glNormal3d(o1, -o1/div, -o1/div);
      if (!SolidColor)
      {
	// val = valuefunc(vx1+vo1, vx2, vx3);
	SetCol(val, m0, isbone);
      }
      //      glVertex3d(x1+o1,    x2,    x3);
      {int v = 1; glVertex3d(Y[v][0], Y[v][1], Y[v][2]);}
    }
    glEnd();

  }
}

void GenerateSolidList()
{
  SolidList.CompileStart();
  const int N = Resolution;
  double bs = 1./N;
  for(int i = 0; i < N; ++ i)
  {
    for(int j = 0; j < N; ++j)
    {
      for (int k = 0; k < N; ++k)
      {
	  DrawHex( i*bs, j*bs, k*bs, bs, bs, bs,
		   SolidPart(i,j,k));
      }
    }
  }
  SolidList.CompileFinish();
}
void GenerateBoneList()
{
  BoneList.CompileStart();
  if (CurDir != Ndir)
  {
    const int N = Resolution;
    double bs = 1./N;
    for(int i = 0; i < N; ++ i)
    {
      for(int j = 0; j < N; ++j)
      {
	for (int k = 0; k < N; ++k)
	{
	  if (bonefunc(i*bs +bs*0.5, j*bs +bs*0.5, k*bs +bs*0.5)
	      >= bonethres)
	    DrawHex( i*bs, j*bs, k*bs, bs, bs, bs,
		     SolidPart(i,j,k));
	}
      }
    }
  }
  BoneList.CompileFinish();
}




inline double razt(double x1, double y1, double z1,
		   double x2, double y2, double z2)
{
  x1-=x2;
  y1-=y2;
  z1-=z2;
  return sqrt(x1*x1 + y1*y1 + z1*z1);
}



void RenewSolid(bool NewShape = true)
{
  SetMinMax();
  if (NewShape)
  {
    GenerateSolid();
  }
  GenerateSolidList();
  GenerateBoneList();
}

void gfxinit ()
{
  static bool firsttime = true;
  static double min, max;
  glLineWidth(2);
  glEnable(GL_NORMALIZE);
  glEnable(GL_ALPHA_TEST);
  glEnable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);
  glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
  glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
  glDepthFunc(GL_LESS);
  GLfloat lamb[4] = {0.8, 0.8, 0.8, 1};
//     glLightModelfv(GL_LIGHT_MODEL_AMBIENT,lamb);
//  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
//     GLfloat dcol [] = { 1,1 ,1,1};
//     GLfloat scol [] = { 1, 1 ,1,1};
    
//     GLfloat pos1 [] = { 0,5 ,0,1};
//     glLightfv(GL_LIGHT0, GL_POSITION, pos1);
//     GLfloat pos2 [] = { -5, 0 ,0,1};
//     glLightfv(GL_LIGHT1, GL_POSITION, pos2);
//     GLfloat pos3 [] = {0 ,-50 ,0,1};
//     glLightfv(GL_LIGHT2, GL_POSITION, pos3);
//     GLfloat pos4 [] = { 5,0 ,0,1};
//     glLightfv(GL_LIGHT3, GL_POSITION, pos4);
    
//     glLightfv(GL_LIGHT0, GL_DIFFUSE, dcol);
//     glLightfv(GL_LIGHT0, GL_SPECULAR, scol);
//     glLightfv(GL_LIGHT1, GL_DIFFUSE, dcol);
//     glLightfv(GL_LIGHT1, GL_SPECULAR, scol);
//     glLightfv(GL_LIGHT2, GL_DIFFUSE, dcol);
//     glLightfv(GL_LIGHT2, GL_SPECULAR, scol);
//     glLightfv(GL_LIGHT3, GL_DIFFUSE, dcol);
//     glLightfv(GL_LIGHT3, GL_SPECULAR, scol);
  //  glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
  //  glEnable(GL_FOG);
  double ff = 0.99;
  GLfloat fogColor[4] = {ff, ff, ff, 1.0};
  int fogMode = GL_EXP;
  glFogi (GL_FOG_MODE, fogMode);
  glFogfv (GL_FOG_COLOR, fogColor);
  glFogf (GL_FOG_DENSITY, 0.20);
  glHint (GL_FOG_HINT, GL_DONT_CARE);
  glFogf (GL_FOG_START, 2);
  glFogf (GL_FOG_END, 4);
  
  //  glEnable(GL_LIGHTING);
  //    glEnable(GL_LIGHT0);
  //     glEnable(GL_LIGHT1);
  //     glEnable(GL_LIGHT2);
  //     glEnable(GL_LIGHT3);
  
  MAX_X = 1;
  MIN_X = 0;
  MAX_Y = 1;
  MIN_Y = 0;
  MAX_Z = 1;
  MIN_Z = 0;
  xc = .5;
  yc  = .5;
  zc = .5;
}


void display()
{
  //  gl2psBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  gl2psLineWidth(1.25);
  glClearColor(1,1,1,1);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glAlphaFunc(GL_GEQUAL, 1-atest);
  double eps = 1e-3;
  glDepthRange(eps, 1);
  glShadeModel(SolidColor?GL_FLAT:GL_SMOOTH);
  if(DrawSolid)
  {
    glDisable(GL_LIGHTING);
    glDisable(GL_BLEND);
    gl2psDisable(GL2PS_BLEND);
    GLenum mode;
    switch(WireFrame)
    {
    case 0: mode = GL_FILL;break;
    case 1: mode = GL_LINE;break;
    case 2: mode = GL_POINT;break;
    }
    glPolygonMode(GL_FRONT, mode);
    glCullFace(GL_BACK);
    glCallList(SolidList);

    
    if (RemoveColor)
    {
      glEnable (GL_BLEND);
      gl2psEnable(GL2PS_BLEND);
      glEnable(GL_LIGHTING);
      glBlendFunc (GL_SRC_ALPHA ,GL_ONE_MINUS_SRC_ALPHA);
      glPolygonMode (GL_BACK, GL_LINE);
      glCullFace (GL_FRONT);// Don't Draw Any Front-Facing Polygons
      glDepthFunc (GL_LEQUAL);// Change The Depth Mode
      glDepthRange(0.2*eps,1);
      glDepthMask(GL_FALSE);
      glCallList(SolidList);
      glDepthMask(GL_TRUE);
      glDepthFunc (GL_LESS);// Change The Depth Mode
      glCullFace (GL_BACK);// Don't Draw Any Front-Facing Polygons
      glDisable(GL_LIGHTING);
    }else
      glPolygonMode (GL_BACK, GL_FILL);


  }
  glEnable(GL_BLEND);
  gl2psEnable(GL2PS_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  gl2psBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  gl2psLineWidth(0.7);
  glDepthFunc (GL_LEQUAL);// Change The Depth Mode

  glDisable(GL_BLEND);
  gl2psDisable(GL2PS_BLEND);
  if (GrayBone)
  {
    glDepthRange(0, 1-eps);
    glCallList(BoneList);
  }
  glDepthFunc (GL_LESS);// Change The Depth Mode
  glFlush();
  glutSwapBuffers();
}

static int width, height;
void reshape (int w, int h)
{
  width = w;
  height = h;
  glViewport (0, 0, (GLsizei) w, (GLsizei) h); 
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluPerspective(30.0, (GLfloat) w/(GLfloat) h, 1.0, 200.0);
  orientMe();
}

#include <png.h>
int curframe = 0;
void save_png(int frame=curframe)
{
  curframe++;
  FILE *fp;
  ostringstream os;
  os.fill('0');
  os<<setw(4) << frame <<".out" << ".png";
  fp = fopen(os.str().c_str(), "wb");
  glReadBuffer(GL_FRONT);
  GLfloat *pixels = new GLfloat[width*height*4];
  glReadPixels(0,0, width, height, GL_RGBA, GL_FLOAT, pixels);
  //    cout << " " << width << "x" << height << endl;
  try{
    png_structp png_ptr = png_create_write_struct
      (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if (!png_ptr)
      throw(4);
      
    png_infop info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr)
    {
      png_destroy_write_struct(&png_ptr,
			       (png_infopp)NULL);
      throw 5;
    }
      
    if (setjmp(png_jmpbuf(png_ptr)))
    {
      png_destroy_write_struct(&png_ptr, &info_ptr);
      fclose(fp);
      throw 6;
    }
    png_init_io(png_ptr, fp);      
    /* set the zlib compression level */
    png_set_compression_level(png_ptr,
			      Z_BEST_COMPRESSION);

    /* set other zlib parameters */
    png_set_compression_mem_level(png_ptr, 8);
    png_set_compression_strategy(png_ptr,
				 Z_DEFAULT_STRATEGY);
    png_set_compression_window_bits(png_ptr, 15);
    png_set_compression_method(png_ptr, 8);
    png_set_compression_buffer_size(png_ptr, 8192);
    png_set_IHDR(png_ptr, info_ptr, width, height,
		 8, PNG_COLOR_TYPE_RGB_ALPHA, PNG_INTERLACE_NONE,
		 PNG_COMPRESSION_TYPE_DEFAULT,
		 PNG_FILTER_TYPE_DEFAULT);
    png_set_sRGB(png_ptr, info_ptr, PNG_sRGB_INTENT_PERCEPTUAL);
    png_color_8 a;
    a.red = 8;
    a.green=8;
    a.blue=8;
    a.gray=0;
    a.alpha=0;
    //    png_set_sBIT(png_ptr, info_ptr, &a);
    png_text  text[6];
    int  num_text = 0;
    text[num_text].compression = PNG_TEXT_COMPRESSION_NONE;
    text[num_text].key = "Title";
    text[num_text].text = "Kartinka";
    ++num_text;
    text[num_text].compression = PNG_TEXT_COMPRESSION_NONE;
    text[num_text].key = "Author";
    text[num_text].text = "Yavor Vutov";
    ++num_text;
    text[num_text].compression = PNG_TEXT_COMPRESSION_NONE;
    text[num_text].key = "Description";
    text[num_text].text = "3d Rendering";
    ++num_text;
    
    png_set_text(png_ptr, info_ptr, text, num_text);    
    png_set_packing(png_ptr);

    
    
    unsigned char **row_pointers = new unsigned char* [2*height];
    for (int i=0; i<height; i++)
      row_pointers[i]= new unsigned char[4*width];

    for(int i = 0; i < height;++i)
      for(int j = 0; j < width; ++j)
      {
	row_pointers[i][j*4] =  static_cast<unsigned char>(255*pixels[(height-1-i)*width*4 +j*4+0]);
	row_pointers[i][j*4+1] = static_cast<unsigned char>(255*pixels[(height-1-i)*width*4 +j*4+1]);
	row_pointers[i][j*4+2] = static_cast<unsigned char>(255*pixels[(height-1-i)*width*4 +j*4+2]);
	row_pointers[i][j*4+3] = static_cast<unsigned char>(255*pixels[(height-1-i)*width*4 +j*4+3]);
	//	row_pointers[i][j*4+3] = 255;
      }
    
    png_set_rows(png_ptr, info_ptr, row_pointers);

    //    png_write_info(png_ptr, info_ptr);
    
    png_write_png(png_ptr, info_ptr, PNG_TRANSFORM_IDENTITY, NULL);
    png_write_end(png_ptr, info_ptr);

    for (int i=0; i<height; i++)
      delete[] row_pointers[i];
    delete[] row_pointers;
    png_destroy_write_struct(&png_ptr, &info_ptr);
  }catch(int err){
    cerr << "Png write error " << err << endl;}
  delete[]pixels;
  fclose(fp);
}  
void save_eps(int frame=0)
{
  FILE *fp;
  int state = GL2PS_OVERFLOW, buffsize = 16*1024*1024;
  ostringstream os;
  os.fill('0');
  os<<setw(4) << frame <<".out" << ".eps";
  fp = fopen(os.str().c_str(), "wb");
  while(state == GL2PS_OVERFLOW){
    buffsize *= 2;
    gl2psBeginPage("test", "gl2psTestSimple", NULL,
		   GL2PS_EPS,
		   GL2PS_SIMPLE_SORT,
		   //GL2PS_BSP_SORT,
		   //                       GL2PS_DRAW_BACKGROUND |
		   GL2PS_USE_CURRENT_VIEWPORT
		   //		   |GL2PS_NO_BLENDING
		   |GL2PS_OCCLUSION_CULL
		   |GL2PS_SIMPLE_LINE_OFFSET
		   , 
		   GL_RGBA, 0, NULL, 0, 0, 0,  buffsize, fp, "Kartinka");
    display();
    state = gl2psEndPage();
  }
  fclose(fp);
}
void save_avi()
{
  const int numframes = 360*4;
  double step = atan(1)/360.;
  for(int a = 0; a<numframes;++a)
  {
    save_eps(a);
    angle1 = step;
    orientMe();
    glutPostRedisplay();
  }
}
bool active = false;
void avi_png_idle();
void save_avi_png()
{
  active = true;
  curframe = 0;
  
  glutIdleFunc(avi_png_idle);
}

void menu(int item)
{
  glutPostRedisplay();
  switch(item) {
  case SAVE_EPS:
    save_eps();
    break;
  case SAVE_PNG:
    save_png();
    break;
  case SAVE_AVI:
    save_avi();
    break;
  case QUIT:
    exit(0);
  }
}


void inputKey(int key, int x, int y) {
  switch (key) {
  case GLUT_KEY_F1:
    swapp=not swapp;
    if (swapp)
    {
      for(int i = 0; i < 3; ++i)ox[i]=oy[i]=oz[i]=xob[i]=xu[i]=0;
      ox[0] = oy[1] = oz[2] = xob[0] = xu[1]= 1;
    }
    break;
  case GLUT_KEY_PAGE_DOWN:
    angle -=0.02f;
    break;
  case GLUT_KEY_LEFT : 
    angle1 += 0.02f;
    break;
  case GLUT_KEY_RIGHT :
    angle1 -=0.02f;
    break;
  case GLUT_KEY_UP : 
    angle2 +=0.02f;
    break;
  case GLUT_KEY_PAGE_UP:
    dist -= 0.1;
    break;
  case GLUT_KEY_DOWN :
    angle2 -=0.02f;
    break;
  case GLUT_KEY_F5 :
    DrawSolid = !DrawSolid;
    break;
  case GLUT_KEY_F7 :
    ShowPlates = !ShowPlates;
    RenewSolid();
    break;
  case GLUT_KEY_F8 :
    //    ShowPalette = !ShowPalette;
    SolidColor = !SolidColor;
    RenewSolid(false);
    break;
  case GLUT_KEY_F10 :
    RemoveColor = ! RemoveColor;
    break;
  case GLUT_KEY_F3 :
    Shading = !Shading;
    RenewSolid();
    break;
  case GLUT_KEY_F2 :
    GrayBone = !GrayBone;
    RenewSolid(false);
    break;
  case GLUT_KEY_F11 :
    WireFrame = (WireFrame+1)%3;
    break;
  case GLUT_KEY_F12 :
    isobehav = (isobehav+1)%6;
    RenewSolid(false);
    break;
  case GLUT_KEY_F6:
    DrawIso = !DrawIso;
    break;
  case GLUT_KEY_INSERT:
    dist += 0.1;
    break;
  default:
    return;
  }
  orientMe();
  glutPostRedisplay();
	
}
void inputKey1(unsigned char key, int x, int y) {
  switch (key) {
  case 127 : 
    angle += 0.02f;
    break;
  case 'n':
    if (CurDir!=Ndir)
    {
      CurDir = Ndir;
      RenewSolid();
    }
    break;
  case 'c':
    if (CurDir!=Cutdir)
    {
      CurDir = Cutdir;
      RenewSolid();
    }
    break;
  case 'x':
    if (CurDir!=Xdir)
    {
      CurDir = Xdir;
      RenewSolid();
    }
    break;
  case 'y':
    if (CurDir!=Ydir)
    {
      CurDir = Ydir;
      RenewSolid();
    }
    break;
  case 'z':
    if (CurDir!= Zdir)
    {
      CurDir = Zdir;
      RenewSolid();
    }
    break;
  case 'j':
    if (movedistance>=movesize)
    {
      movedistance-=movesize;
      RenewSolid();
    }
    break;
  case 'k':
    movedistance+=movesize;
    RenewSolid();
    break;
  case 'm':
    if(mypol>0.1)
    {
      mypol -=0.08;
      RenewSolid(false);
    }
    break;
  case '1':
    minfixer -=1;
    RenewSolid(false);
    break;
  case '2':
    minfixer +=1;
    RenewSolid(false);
    break;
  case '3':
    maxfixer -=1;
    RenewSolid(false);
    break;
  case '4':
    maxfixer +=1;
    RenewSolid(false);
    break;
  case ',':
    if(mypol<4)
    {
      mypol+=0.08;
      RenewSolid(false);
    }
    break;
  case 'i':
    if(il<1000)
    {
      il++;
    }
    break;
  case 'u':
    if(il>1)
    {
      il--;
    }
    break;
  case 'I':
    if(il<1000)
    {
      il+=50;
    }
    break;
  case 'U':
    if(il>51)
    {
      il-=50;
    }
    break;
	  
  case 's':
    save_png();
    break;
  case 'a':
    save_avi_png();
    break;
  case '-':
    atest -= 1/256.;
    break;
  case '+':
    atest += 1/256.;
    break;
  default:
    return;
  }
  orientMe();
  glutPostRedisplay();
	
}

void avi_png_idle()
{
  const int numframes=360;
  
  save_png(curframe);
  double step = atan(1)*8./360.;
  angle1 = step;
  if (DirUp)
  {
    movedistance+= movesize;
    if(++CurMoveCount>=MaxMoveCount)
    {
      DirUp = !DirUp;
      CurMoveCount = 0;
    }
  }else
  {
    movedistance-= movesize;
    if(++CurMoveCount>=MaxMoveCount)
    {
      DirUp = !DirUp;
      CurMoveCount = 0;
    }
  }
  RenewSolid();
  orientMe();
  glutPostRedisplay();
  cout << curframe-1 << "\r" << flush;
  if(curframe==numframes)
  {
    cout << "                             " <<  "\r" << flush;
    glutIdleFunc(NULL);
  }
}

void MyGlutInit(int &argc, char **&argv)
{
    glutInit(&argc, argv);
    glutInitWindowPosition(-1, -1);
    glutInitWindowSize(800, 800);
  
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA| GLUT_DEPTH|GLUT_ALPHA);
    glutCreateWindow("Bone Drawer");
    glutDisplayFunc(display);
    //  glutIdleFunc(display);
    glutCreateMenu(menu);
    glutAddMenuEntry("Save EPS", SAVE_EPS);
    glutAddMenuEntry("Save PNG", SAVE_PNG);
    glutAddMenuEntry("Save AVI", SAVE_AVI);
    glutAddMenuEntry("Exit", QUIT);
    glutAttachMenu(GLUT_RIGHT_BUTTON);
    glutSpecialFunc(inputKey);
    glutKeyboardFunc(inputKey1);
    glutReshapeFunc(reshape);
    glEnable(GL_LINE_SMOOTH);
    glEnable(GL_POLYGON_SMOOTH);
    //  glutIdleFunc(idle);
    glPointSize(600./Resolution);
}

int main(int argc, char **argv)
  try {
    filename = "64.bonec";
    filename1 = filenameI[0] = filenameI[1] = filenameI[2] = "128.z.0.00";
    if (argc >1)
      filename = argv[1];
    if (argc>2)
      Resolution = atol(argv[2]);
    if (argc >3)
      filenameI[0] = argv[3];
    if (argc >4)
      filenameI[1] = argv[4];
    if (argc >5)
      filenameI[2] = argv[5];
    if (argc >6)
      filename1 = argv[6];

    ReadData();
    GenMins();
    MyGlutInit(argc, argv);
    gfxinit();
    InitColors();

    RenewSolid();

    glutMainLoop();
  } catch (const std::exception & e) {
    std::cerr << e.what() << std::endl;
  }
