# include <vector>
# include <map>
# include <cmath>
# include <iostream>
# include <iomanip>
# include <fstream>
# include <string>
# include <sstream>
# include <stdexcept>
# include <GL/gl.h>
# include <GL/glu.h>
# include <GL/glut.h>
#include "functionlin.h"
extern "C" {
# include "gl2ps.h"
}
#include <sstream>
#include "mirorer.h"


std::string filename;
std::string filename1;


double v1[3], v2[3], v3[3];
double ox[3] = {1,0,0}, oy[3]={0,1,0}, oz[3] = {0,0,1};
double xob[3] = {1, 0, 0}, xu[3]={0,1,0};

bool DrawSolid = true;
bool DrawIso = false;
bool ShowPlates = false;
bool ShowPalette = false;
bool RemoveColor = false;
int isobehav = 0;
int WireFrame = 0;
int CurLine=0;
int il = 2;
double trlines = 1;
double isocol = 0.050;
enum Dirs{Ndir, Xdir, Ydir, Zdir};
Dirs CurDir = Ndir;
int Resolution;
double mypol=1;//0.5;
  static  vector3d<unsigned char > vec1;
double gmax, gmin;



class DisplayList
{
  int ln;
public:
  DisplayList():ln(0)
  {
  }
  ~DisplayList()
  {
    if(ln)
      glDeleteLists(ln,1);
  }
  void CompileStart()
  {
    if(ln)
      glDeleteLists(ln,1);
    ln = glGenLists(1);
    glNewList(ln, GL_COMPILE);
  }
  void CompileFinish()
  {
    glEndList();
  }
  operator int ()
  {
    return ln;
  }
};



DisplayList IsoList;
DisplayList BoneList;
DisplayList SolidList;

LinearFunction3d<double, double, bool> bone;
LinearFunction3d<double, double, double> values;
vector3d<unsigned char> SolidPart;

void ReadData()
{
  int sz;
  SmallVector<double, 3> min, max;
  min = 0;
  max = 1;
  {
    std::istringstream is(filename);
    is >> sz;
  }
  vector3d<unsigned char > vec;
  vec.SetSize(sz, sz, sz);
  try{
    vec.Read(filename, 0);
  }catch(...)
  {
    sz = 3;
    vec.SetSize(sz,sz,sz);
      for(int i = 0; i < sz; ++i)
	for(int j = 0; j<sz; ++j)
	  for(int k=0; k < sz; ++k)
	  {
	      vec(i,j,k) = 0;
	  }
  }

  bone.Init(vec, min, max, false); //no interpolation for the bone
  // bone read
  int N;
  int k;
  {
    char ch;
    std::istringstream is(filename1);
    is >> N;
    is >> ch >> ch >>ch >>k;
    N *=k;
  }
  vector3d<double> im;
  im.SetSize(N, N, N);
  try{
    im.Read(filename1, 0);
  }catch (...)
  {
    std::cout << N << " " <<  std::endl;
    sz = 2;
    im.SetSize(sz,sz,sz);
    for(int i = 0; i < sz; ++i)
      for(int j = 0; j<sz; ++j)
 	for(int k=0; k < sz; ++k)
 	{
 	  im(i,j,k) = 1.;
 	}
  }
  values.Init(im, min, max);
}



double Cos(double v1[3], double v2[3], double v3[3])
{
  for(int i = 0; i < 3;++i)
  {
    v1[i] -= v3[i];
    v2[i] -= v3[i];
  }
  double l1 = 0;
  double l2 = 0;
  for (int i = 0; i < 3; ++i)
  {
    l1 += v1[i]*v1[i];
    l2 += v2[i]*v2[i];
  }
  l1 = sqrt(l1);
  l2 = sqrt(l2);
  double cos = 0;
  for (int i = 0; i < 3; ++i)
  {
    cos += v1[i]*v2[i];
  }
  cos /= l1*l2;
  return cos;    
}

template<class T>
struct Quad
{
  T w,x,y,z;
  Quad():w(), x(), y(), z(){}
  Quad(T ax, T ay, T az, T angle):
    w(cos(angle/2)),
    x(ax*sin(angle/2)),
    y(ay*sin(angle/2)),
    z(az*sin(angle/2))
  {}
  void GetMatrix(T (&A)[3][3])
  {
    A[0][0] = w*w + x*x - y*y - z*z; A[0][1] = 2*x*y - 2*w*z;        A[0][2] = 2*x*z + 2*w*y;
    A[1][0] = 2*x*y + 2*w*z;         A[1][1] = w*w - x*x + y*y - z*z;A[1][2] = 2*y*z - 2*w*x;
    A[2][0] = 2*x*z - 2*w*y;         A[2][1] = 2*y*z + 2*w*x;        A[2][2] = w*w - x*x - y*y + z*z;
  }
};
template<class T>
Quad<T> operator*(const Quad<T> &a, const Quad<T>&b)
{
  Quad<T> r;
  r.w = a.w*b.w - a.x*b.x - a.y*b.y - a.z*b.z;
  r.x = a.w*b.x + a.x*b.w + a.y*b.z - a.z*b.y;
  r.y = a.w*b.y + a.y*b.w + a.z*b.x - a.x*b.z;
  r.z = a.w*b.z + a.z*b.w + a.x*b.y - a.y*b.x;
  return r;
}

template<class T, class V>
void MinMax(vector3d<T> &vec, vector3d<V> &b, T &min, T &max)
{
  bool firstime = true;

  for (int i = 0; i < vec.sizes[0]; ++i)
    for(int j = 0; j < vec.sizes[1]; ++j)
      for(int k = 0; k < vec.sizes[2]; ++k)
      {
	//	vec(i,j,k) = std::fabs(vec(i,j,k));
	if (((isobehav==1)&&(!b(i,j,k))) ||
	    ((isobehav==2)&&b(i,j,k)) ||
	    ((isobehav==0)))
	{
	  if (firstime)
	  {
	    min = max = vec(i,j,k);
	    firstime = false;
	  }
	  else
	  {
	    T a = vec(i,j,k);
	    min = (min<a)?min:a;
	    max = (max>a)?max:a;
	  }
	}
      }
}
  

using namespace std;

double MIN_X,MAX_X, MIN_Y, MAX_Y, MIN_Z, MAX_Z;
double vsx, vsy, vsz;
double minx, miny, minz, maxx, maxy, maxz;
double xc, yc, zc;

static float angle=atan(1)*2,ratio = 1;
bool usecube = false;
static bool swapp=true;
static float angle1=0;
static float angle2=0;
static GLint snowman_display_list;
double dist = 2.5;

template<class T>
void MulSmallMat(T (&res)[3][3],const T (&A)[3][3],const T(&B)[3][3])
{
  for(int i = 0; i < 3; ++i)
    for(int j = 0; j < 3; ++j)
    {
      T sum = 0;
      for(int k = 0; k < 3; ++k)
	sum += A[i][k] * B[k][j];
      res[i][j]=sum;
    }
}
template<class T>
void MulSmallMatVec(T (&res)[3],const T (&A)[3][3],const T(&x)[3])
{
  for(int i = 0; i < 3; ++i)
  {
    double sum = 0;
    for(int k = 0; k < 3; ++k)
      sum += A[i][k] * x[k];
    res[i]=sum;
  }
}
template<class T>
void Trans(const T (&A)[3][3], T(&x)[3])
{
  T t[3];t[0] = x[0];t[1] = x[1]; t[2] = x[2];
  MulSmallMatVec(x, A, t);
}

void orientMe()
{
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  double x = xc;//(MIN_X + MAX_X)/2;
  double y = yc;//(MIN_Y + MAX_Y)/2;
  double z = zc;// (MIN_Z + MAX_Z)/2;
  //    double matx[3][3], maty[3][3], matz[3][3], tmp[3][3];
  //    static double v1[3];
  //    static double v[3];
  //    static double t[3];
  //    matx[0][0] = 1;           matx[0][1] = 0;             matx[0][2] = 0;
  //    matx[1][0] = 0;           matx[1][1] = cos(angle);    matx[1][2] = -sin(angle);
  //    matx[2][0] = 0;           matx[2][1] = sin(angle);    matx[2][2] = cos(angle);

  //    maty[0][0] = cos(angle1); maty[0][1] = 0;             maty[0][2] = sin(angle1);
  //    maty[1][0] = 0;           maty[1][1] = 1;             maty[1][2] = 0;
  //    maty[2][0] = -sin(angle1);maty[2][1] = 0;             maty[2][2] = cos(angle1);

  //    matz[0][0] = cos(angle2); matz[0][1] =-sin(angle2);   matz[0][2] = 0;
  //    matz[1][0] = sin(angle2); matz[1][1] =cos(angle2);    matz[1][2] = 0;
  //    matz[2][0] = 0;           matz[2][1] = 0;             matz[2][2] = 1;
  //    //   cout << angle << " " << angle1 << " " << angle2 << endl;
  //    //   MulSmallMat(tmp, matx, maty);
  //    //   MulSmallMat(matx, tmp, matz);
  //    double x = xc;//(MIN_X + MAX_X)/2;
  //    double y = yc;//(MIN_Y + MAX_Y)/2;
  //    double z = zc;// (MIN_Z + MAX_Z)/2;
  //    v[0] = dist;
  //    v[1] = 0;
  //    v[2] = 0;
  //    v1[0] = v[0];
  //    v1[1] = v[1];
  //    v1[2] = v[2];
  //    //   MulSmallMatVec(v1, matx, v);
  //    MulSmallMatVec(v, maty, v1);
  //    MulSmallMatVec(v1, matz, v);
   
  //    double xo = v1[0];
  //    double yo = v1[1];
  //    double zo = v1[2];
  //    v[0] = 0;
  //    v[1] = 1;
  //    v[2] = 0;
  //    v1[0] = v[0];
  //    v1[1] = v[1];
  //    v1[2] = v[2];
  //    //   MulSmallMatVec(v1, matx, v);
  //    MulSmallMatVec(v, maty, v1);
  //    MulSmallMatVec(v1, matz, v);
  //    double xu = v1[0];
  //    double yu = v1[1];
  //    double zu = v1[2];
  // //    if (angle1>atan(1)*2)
  // //      angle1=atan(1)*1.999;
  // //    if (angle1<-atan(1)*2)
  // //      angle1=-atan(1)*1.999;

  // //    double xo, yo, zo;
  // //    xo = cos(angle)*dist;
  // //    zo = sin(angle)*dist;

  // //    yo = cos(angle1)*xo + sin(angle1)*zo;
  // //    zo = sin(angle1)*xo + cos(angle1)*zo;
   
  // //    double xo = cos(angle1)*cos(angle)*dist;
  // //    double yo = sin(angle1)*dist;
  // //    double zo = cos(angle1)*sin(angle)*dist;

  static bool firsttime = false;

  Quad<double> q1 = Quad<double>(ox[0], ox[1], ox[2], angle);
  Quad<double> q2 = Quad<double>(oy[0], oy[1], oy[2], angle1);
  Quad<double> q3 = Quad<double>(oz[0], oz[1], oz[2], angle2);
  Quad<double> q4;
  q4 = q1*q2*q3;
  double Mat[3][3];
  q4.GetMatrix(Mat);
  Trans(Mat, xob);
  Trans(Mat, xu);
  if (swapp)
  {
    Trans(Mat, ox);
    Trans(Mat, oy);
    Trans(Mat, oz);
  }
  double xo[3];
  xo[0] = dist*xob[0];
  xo[1] = dist*xob[1];
  xo[2] = dist*xob[2];
  for (int i = 0; i < 3; ++i)
    v1[i]=v2[i] = v3[i] = 0.;
  v2[0] = .2;
  v2[1] = .2;
  v1[0] = x;
  v1[1] = y;
  v1[2] = z;
   
  gluLookAt (x-xo[0], y-xo[1], z-xo[2],  x, y, z, xu[0], xu[1], xu[2]);
  angle = angle1 = angle2 = 0;

}



enum menu_items {
  SAVE_EPS,
  SAVE_PNG,
  SAVE_AVI,
  QUIT
};


enum Sides {
  LeftSide = 0x1,
  RightSide = 0x2,
  TopSide = 0x4,
  BottomSide = 0x8,
  FrontSide = 0x10,
  BackSide = 0x20
};

template <class T>
void RemoveExtra(vector3d<T> &vec)
{
  for(int i = 0; i < vec.sizes[0]; i++)
    for(int j = 0; j < vec.sizes[1]; j++)
      for(int k = 0; k < vec.sizes[2]; k++)
      {
	if (i >0)
	{
	  bool lp = vec(i-1,j,k) & RightSide;
	  bool rp = vec(i, j, k) & LeftSide;
	  lp = lp and rp;
	  if (lp)
	  {
	    vec(i-1, j, k) &= ~ RightSide;
	    vec(i, j, k) &= ~LeftSide;
	  }
	}
	if (j >0)
	{
	  bool lp = vec(i,j-1,k) & BackSide;
	  bool rp = vec(i, j, k) & FrontSide;
	  lp = lp and rp;
	  if (lp)
	  {
	    vec(i, j-1, k) &= ~ BackSide;
	    vec(i, j, k) &= ~FrontSide;
	  }
	}
	if (k >0)
	{
	  bool lp = vec(i,j,k-1) & TopSide;
	  bool rp = vec(i, j, k) & BottomSide;
	  lp = lp and rp;
	  if (lp)
	  {
	    vec(i, j, k-1) &= ~TopSide;
	    vec(i, j, k) &= ~BottomSide;
	  }
	    
	}
      }
}

double atest = 1.;




void color(double d)
{
  d *= 4;
  double r, g, b;
  if (d < 1.) {
    r = 0.; g = d; b = 1.;
  } else if (d < 2.) {
    r = 0.; g = 1.; b = 2. - d;
  } else if (d < 3.) {
    r = d - 2; g = 1.; b = 0.;
  } else {
    r = 1.; g = 4 - d; b = 0.;
  }
  glColor4d(r, g, b, 0.7);

}
void mat(double cr, double cg, double cb, double tr, double mul=1)
{
  if(RemoveColor)
  {
    GLfloat amb[] = {1,1,1,1};    
    glColor4fv(amb);
  }
  else
  {
    GLfloat amb[] = {cr*mul, cg*mul, cb*mul, tr*mul};
    glColor4fv(amb);
  }

//   {
//     GLfloat amb[] = {1,1,1,1};
//     //   glMaterialf(GL_FRONT, GL_SHININESS, 3);
//     {
//       GLfloat amb[] = {0,0,0,1};
//       glMaterialfv(GL_FRONT, GL_AMBIENT, amb);
//     }
//     //   double in =0.2;
//     //   amb[0]*=in;
//     //   amb[1]*=in;
//     //   amb[2]*=in;
//     //    glMaterialfv(GL_FRONT, GL_DIFFUSE, amb);
//     //   glMaterialfv(GL_FRONT, GL_SPECULAR, amb);
//   }
}
struct P3D
{
  double X[3];
};

vector3d<double> isovals;
bool IntersectLine(P3D p1, P3D p2, double v1, double v2, double val, P3D &res)
{
  double f1 = min(v1, v2);
  double f2 = max(v1, v2);
  if ((val <f1) || (val >f2))
    return false;
  if (v2==v1)
    return false;
  double k = (val - v1)/(v2-v1);
  double k1 =  k1 = 1 - k;
  for(int i = 0; i <3; ++i)
    res.X[i] = p1.X[i]*k1 + p2.X[i]*k;
  return true;
}

void DoSide( P3D p[4], double v[4])
{
  for(int i = 0; i < il; ++i)
  {
    
    double curval = gmin + (gmax-gmin)*(i+0.5)/(il);
     double val1 = (curval - gmin)/(gmax-gmin);
    val1 = val1*2 - 1;
    val1 = (val1<0)?
      -pow(fabs(val1), mypol):
      pow(fabs(val1), mypol);
    val1 = (val1+1)/2;
    curval = val1*(gmax-gmin) +gmin;
    
    
    P3D points[4];
    int maxpoint = 0;
    if(IntersectLine(p[0], p[1], v[0], v[1], curval, points[maxpoint]))
      maxpoint++;
    if(IntersectLine(p[2], p[1], v[2], v[1], curval, points[maxpoint]))
      maxpoint++;
    if(IntersectLine(p[2], p[3], v[2], v[3], curval, points[maxpoint]))
      maxpoint++;
    if(IntersectLine(p[0], p[3], v[0], v[3], curval, points[maxpoint]))
      maxpoint++;
    if(maxpoint==0)
        continue;
    glBegin(GL_LINE_LOOP);
    for (int i = 0; i < maxpoint;++i)
    {
      glColor4f(0,0,0,1);
      glVertex3d(points[i].X[0], points[i].X[1], points[i].X[2]);
    }
    glEnd();
  }
}

void DrawHexIso(int I, int J, int K, double bs, int sides)
{
  if(!sides)
    return;
  double x, y, z, x1, y1, z1;
  x = I*bs;
  x1 = x + bs;
  y = J*bs;
  y1 = y + bs;
  z = K*bs;
  z1 = z + bs;
  P3D p[8] = {{x, y,   z}, 
	      {x1, y,  z}, 
	      {x, y1,  z},
	      {x1, y1, z},
	      {x, y,   z1},
	      {x1, y,  z1},
	      {x, y1,  z1},
	      {x1, y1, z1}};
  
  double v[8] = { isovals(I,J,    K),
		  isovals(I+1,J,  K),
		  isovals(I,J+1,  K),
		  isovals(I+1,J+1,K),
		  isovals(I,J,    K+1),
		  isovals(I+1,J,  K+1),
		  isovals(I,J+1,  K+1),
		  isovals(I+1,J+1,K+1)};
  if(sides &BottomSide)
  {
    P3D side[4] = {p[0], p[1], p[3], p[2]};
    double vs[4] ={v[0], v[1], v[3], v[2]};
    DoSide(side, vs);
  }
  if(sides &TopSide)
  {
    P3D side[4] = {p[4], p[5], p[7], p[6]};
    double vs[4] ={v[4], v[5], v[7], v[6]};
    DoSide(side, vs);
  }
  if(sides &LeftSide)
  {
    P3D side[4] = {p[0], p[2], p[6], p[4]};
    double vs[4] ={v[0], v[2], v[6], v[4]};
    DoSide(side, vs);
  }
  if(sides &RightSide)
  {
    P3D side[4] = {p[1], p[3], p[7], p[5]};
    double vs[4] ={v[1], v[3], v[7], v[5]};
    DoSide(side, vs);
  }
  if(sides &FrontSide)
  {
    P3D side[4] = {p[0], p[1], p[5], p[4]};
    double vs[4] ={v[0], v[1], v[5], v[4]};
    DoSide(side, vs);
  }
  if(sides &BackSide)
  {
    P3D side[4] = {p[2], p[3], p[7], p[6]};
    double vs[4] ={v[2], v[3], v[7], v[6]};
    DoSide(side, vs);
  }
    
}


void DrawHex(double x1, double x2, double x3,
	     double o1, double o2, double o3, unsigned char sides,
	     double cr, double cg, double cb, double ct)
{
  double mr, mg, mb;
  double mr1, mg1, mb1;
  mr = mg = mb = 0.90;
  mr1 = mg1 = mb1 = 0.8;
  double div = 331;
  int mode = GL_QUADS;//GL_LINE_LOOP;
  //   if (sides)
  //   {
  //     v3[0] = x1 + o1/2;
  //     v3[1] = x2 + o2/2;
  //     v3[2] = x3 + o3/2;
  //     glBegin(GL_POINTS);

  //     v2[0] = .1;
  //     v2[1] = .7;
  //     v2[2] = .7;
  //     double Int = (Cos(v1, v2, v3));
  //     if (Int <0)Int=0;
  //     v2[0] = -1;
  //     v2[1] = -.2;
  //     v2[2] = 1;
  //     double hhh = (Cos(v1, v2, v3));
  //     if (hhh<0)hhh=0;
  //     Int +=hhh;
  //     //    cout << Int << endl;
  //     Int = Int*0.2*0.5;
  //     mat(cr+Int, cg+Int, cb+Int, ct);
  //     glVertex3d(x1+o1/2, x2+o2/2, x3+o3/2);
  //     glEnd();
  //   }
  //  return;
  //  mat(cr, cg, cb, ct);
  //   glLoadIdentity ();
  //   glTranslated(-x1,-x2,-x3);
  //   glutSolidSphere (o1/2, 3, 2);
  //   return;

  if (sides)
  {
    mat(cr, cg, cb, ct);

    if(sides & BottomSide)
    {
      glBegin(mode);
      //glNormal3d(-o3/div, -o3/div, -o3);
      glVertex3d(x1, x2, x3);
      //glNormal3d(-o3/div, o3/div, -o3);
      glVertex3d(x1, x2+o2, x3);

      //glNormal3d(o3/div, o3/div, -o3);
      glVertex3d(x1+o1, x2+o2, x3);

      //glNormal3d(o3/div, -o3/div, -o3);
      glVertex3d(x1+o1, x2, x3);

      glEnd();
    }
    if (sides & TopSide)
    {
      glBegin(mode);
      //glNormal3d(o3/div, -o3/div, o3);
      glVertex3d(x1+o1, x2,    x3+o3);

      //glNormal3d(o3/div, o3/div, o3);
      glVertex3d(x1+o1, x2+o2, x3+o3);

      //glNormal3d(-o3/div, o3/div, o3);
      glVertex3d(x1,    x2+o2, x3+o3);
      
      //glNormal3d(-o3/div, -o3/div, o3);
      glVertex3d(x1,    x2,    x3+o3);


      glEnd();
    }

    mat(cr*mr, cg*mg, cb*mb, ct);
    if (sides & FrontSide)
    {
      glBegin(mode);
      //glNormal3d(-o2/div, -o2, -o2/div);
      glVertex3d(x1,    x2,    x3);
      
      //glNormal3d(o2/div, -o2, -o2/div);
      glVertex3d(x1+o1, x2,    x3);

      //glNormal3d(o2/div, -o2, o2/div);
      glVertex3d(x1+o1, x2,    x3+o3);

      //glNormal3d(-o2/div, -o2, o2/div);
      glVertex3d(x1,    x2,    x3+o3);
      glEnd();
    }
    if (sides & BackSide)
    {
      glBegin(mode);

      //glNormal3d(o2/div, o2, -o2/div);
      glVertex3d(x1+o1, x2+o2,    x3);


      //glNormal3d(-o2/div, o2, -o2/div);
      glVertex3d(x1,    x2+o2,    x3);
      
      //glNormal3d(-o2/div, o2, o2/div);
      glVertex3d(x1,    x2+o2,    x3+o3);

      //glNormal3d(o2/div, o2, o2/div);
      glVertex3d(x1+o1, x2+o2,    x3+o3);

      glEnd();
    }
    mat(cr*mr1, cg*mg1, cb*mb1, ct);
    if(sides & LeftSide)
    {
      glBegin(mode);
      //glNormal3d(-o1, -o1/div, -o1/div);
      glVertex3d(x1,    x2,    x3);

      //glNormal3d(-o1, -o1/div, o1/div);
      glVertex3d(x1,    x2,    x3+o3);

      //glNormal3d(-o1, o1/div, o1/div);
      glVertex3d(x1,    x2+o2,    x3+o3);

      //glNormal3d(-o1, o1/div, -o1/div);
      glVertex3d(x1,    x2+o2,    x3);

      glEnd();
    }

    if (sides & RightSide)
    {
      glBegin(mode);

      //glNormal3d(o1, o1/div, -o1/div);
      glVertex3d(x1+o1,    x2+o2,    x3);

      //glNormal3d(o1, o1/div, o1/div);
      glVertex3d(x1+o1,    x2+o2,    x3+o3);

      //glNormal3d(o1, -o1/div, o1/div);
      glVertex3d(x1+o1,    x2,    x3+o3);

      //glNormal3d(o1, -o1/div, -o1/div);
      glVertex3d(x1+o1,    x2,    x3);
      glEnd();
    }

  }
}

template <class T>
inline bool In(const T&a, const T&min, const T &max)
{
  return (a >=min) && (a<max);
}
template<class T, class I>
bool check(T &vec, I i, I j, I k)
{
  ptrdiff_t szs[3];
  bool ret = true;
  if (i>0)
    ret = ret && vec(i-1, j, k);
  else
    ret = false;
  if (j>0)
    ret = ret && vec(i, j-1, k);
  else
    ret = false;
  if (k>0)
    ret = ret && vec(i, j, k-1);
  else
    ret = false;
  if (i<vec.sizes[0]-1)
    ret  = ret && vec(i+1, j, k);
  else
    ret = false;
  if (j<vec.sizes[1]-1)
    ret  = ret && vec(i, j+1, k);
  else
    ret = false;
  if (k<vec.sizes[2]-1)
    ret = ret && vec(i, j, k+1);
  else
    ret = false;
  bool tmp = vec(i,j,k);
  int sz1, sz2, sz3;
  sz1 = vec.sizes[0]/2;
  sz2 = vec.sizes[1]/2;
  sz3 = vec.sizes[2]/2;
  //     if ((i<sz1)&&(j<sz2) && (k <sz3))
  //       return false;
  if ((i==sz1)||(j==sz2) || (k==sz3))
    return vec(i,j,k);
  return tmp^ret;
}

struct color4
{
  double r,g,b,a;
  color4():r(), g(), b(), a(){}
  color4(double r_, double g_, double b_, double a_)
  {
    r = r_;
    g = g_;
    b = b_;
    a = a_;
  }
  void Set(double r_, double g_, double b_, double a_)
  {
    r = r_;
    g = g_;
    b = b_;
    a = a_;
  }
  void operator *=(const color4&a)
  { 
    r *= a.r;
    g *= a.g;
    b *= a.b;
    this->a *= a.a;
    return void(*this);//!!!
  }
};

void GetColor(std::vector<color4>&cols, double val, color4&res)
{
  int size = cols.size();
  if (size <2)
  {
    if (size==1)
      res = cols[0];
    else
      res = color4(.8,.8,.8,1);
    return;
  }
  val *= (size-1);
  int ival = static_cast<int>(floor(val));
  val -= ival;
  color4 &l=cols[ival];
  if (ival==(size-1))
  {
    res = l;
    return;
  }
  color4 &r=cols[ival+1];
  res.r = l.r*(1-val) + r.r*val;
  res.g = l.g*(1-val) + r.g*val;
  res.b = l.b*(1-val) + r.b*val;
  res.a = l.a*(1-val) + r.a*val;
}
inline double razt(double x1, double y1, double z1,
		   double x2, double y2, double z2)
{
  x1-=x2;
  y1-=y2;
  z1-=z2;
  return sqrt(x1*x1 + y1*y1 + z1*z1);
}
double count(vector3d<unsigned char>const&im,int i, int j, int k)
{
//   if ((i>=N)||(j>=N) || (k>=N) || (im(i,j,k)==0))
//     return 0
  const int S = 8;
  double c = 0;
  for(int I = -S; I <=S ;++I)
  {
    if (((i+I)<0) || ((i+I)>=N))
      continue;
    for(int J = -S; J <=S; ++J)
    {
      if (((j+J)<0) || ((j+J)>=N))
	continue;
      for(int K = -S; K <= S; ++K)
      {
	if (((k+K)<0) || ((k+K)>=N))
	  continue;
	if (im(i+I, j+J, k+K))
	  c++;
      }
    }
  }
  return c;
      
}
void GenIsoVals1(vector3d<double> const &im)
{

  isovals.SetSize(N+1, N+1, N+1);
  gmin = gmax = 0;
  for(int i = 0; i <= N; ++i)
    for(int j = 0; j <=N; ++j)
      for(int k = 0; k <=N; ++k)
      {
	double a = isovals(i,j,k) = i;razt(i,j,k,N/2, N/2, N/2);
	if (gmin>a)gmin=a;
	if (gmax<a)gmax=a;
      }
}
void GenIsoVals2(vector3d<unsigned char> const &im)
{
  isovals.SetSize(N+1, N+1, N+1);
  gmin = gmax = 0;
  for(int i = 0; i <= N; ++i)
    for(int j = 0; j <=N; ++j)
      for(int k = 0; k <=N; ++k)
      {
	double a = isovals(i,j,k) = count(im, i,j,k);
	if (gmin>a)gmin=a;
	if (gmax<a)gmax=a;
      }
}

// void GenIsoList()
// {
//   double bs = 1./N;

//   IsoList.CompileStart();

//   glColor4d(isocol, isocol, isocol, trlines);

//   for(int i = 0; i < im.sizes[0]; ++ i)
//     for(int j = 0; j < im.sizes[1]
// 	  ; ++j)
//       for (int k = 0; k <im.sizes[2]
// 	     ; ++k)
//       {
	
// 	DrawHexIso(i, j ,k, bs, vec(i,j,k));
//       }
//   glEndList();
// }

// inline void update(vector3d<double> const &im,vector3d<unsigned char> const &vec, double &val, int &cnt, int i,int j, int k, bool at=false)
// {
//   if (at ||
//       ((isobehav==1)&&!vec(i,j,k)) ||
//       ((isobehav==2)&&vec(i,j,k))  ||
//       ((isobehav==0))
//      )
//   {
//     val += im(i,j,k);
//     cnt++;
//   }
// }
// void GenIsoVals(vector3d<double> const &im,vector3d<unsigned char> const &vec)
// {
//   // interpolation
//   isovals.SetSize(N+1, N+1, N+1);
//   gmin = gmax = im(0,0,0);
//   for(int i = 1; i < N; ++i)
//     for(int j = 1; j < N; ++j)
//       for(int k = 1; k < N; ++k)
//       {
// 	double val = 0;
// 	int cnt = 0;
	
// 	update(im, vec, val, cnt, i,   j,   k);
// 	update(im, vec, val, cnt, i-1, j,   k);
// 	update(im, vec, val, cnt, i, j-1,   k);
// 	update(im, vec, val, cnt, i-1,j-1,  k);
// 	update(im, vec, val, cnt, i,   j,   k-1);
// 	update(im, vec, val, cnt, i-1, j,   k-1);
// 	update(im, vec, val, cnt, i, j-1,   k-1);
// 	update(im, vec, val, cnt, i-1,j-1,  k-1);
// 	if (cnt>0)
// 	  isovals(i,j,k) = val/cnt;
// 	else
// 	{
// 	  update(im, vec, val, cnt, i,   j,   k, true);
// 	  update(im, vec, val, cnt, i-1, j,   k, true);
// 	  update(im, vec, val, cnt, i, j-1,   k, true);
// 	  update(im, vec, val, cnt, i-1,j-1,  k, true);
// 	  update(im, vec, val, cnt, i,   j,   k-1, true);
// 	  update(im, vec, val, cnt, i-1, j,   k-1, true);
// 	  update(im, vec, val, cnt, i, j-1,   k-1, true);
// 	  update(im, vec, val, cnt, i-1,j-1,  k-1, true);
// 	  isovals(i,j,k) = val/cnt;
// 	}
// 	if (cnt>0)
// 	{
// 	  double a = isovals(i,j,k);
// 	  if (gmin>a)gmin=a;
// 	  if (gmax<a)gmax=a;
// 	}
//       }

//   //extrapolation
//   for(int i = 0; i <= N; ++i)
//     for(int j = 0; j <=N; ++j)
//     {
//        isovals(i,j,0) = 2*isovals(i,j,1) - isovals(i,j,2);
//        isovals(i,j,N) = 2*isovals(i,j,N-1) - isovals(i,j,N-2);
//     }
//   for(int i = 0; i <= N; ++i)
//     for(int k = 0; k <=N; ++k)
//     {
//       isovals(i,0,k) = 2*isovals(i,1,k) - isovals(i,2,k);
//       isovals(i,N,k) = 2*isovals(i,N-1,k) - isovals(i,N-2,k);
//     }
//   for(int j = 0; j <=N; ++j)
//     for(int k = 0; k <=N; ++k)
//     {
//       isovals(0,j,k) = 2*isovals(1,j,k) - isovals(2,j,k);
//       isovals(N,j,k) = 2*isovals(N-1,j,k) - isovals(N-2, j,k);
//     }
// }


// void gfxinit ()
// {
//   static bool firsttime = true;
//   static double min, max;
//   if (firsttime)
//   {
//     CurList1 = glGenLists(1);
//     glNewList(CurList1, GL_COMPILE);

//     glLineWidth(2);
//     glEnable(GL_NORMALIZE);
//     glEnable(GL_CULL_FACE);
//     glEnable(GL_DEPTH_TEST);
//     glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
//     glDepthFunc(GL_LESS);
//     glShadeModel(GL_FLAT);
//     GLfloat lamb[4] = {0.8, 0.8, 0.8, 1};
// //     glLightModelfv(GL_LIGHT_MODEL_AMBIENT,lamb);
//      glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
// //     GLfloat dcol [] = { 1,1 ,1,1};
// //     GLfloat scol [] = { 1, 1 ,1,1};
    
// //     GLfloat pos1 [] = { 0,5 ,0,1};
// //     glLightfv(GL_LIGHT0, GL_POSITION, pos1);
// //     GLfloat pos2 [] = { -5, 0 ,0,1};
// //     glLightfv(GL_LIGHT1, GL_POSITION, pos2);
// //     GLfloat pos3 [] = {0 ,-50 ,0,1};
// //     glLightfv(GL_LIGHT2, GL_POSITION, pos3);
// //     GLfloat pos4 [] = { 5,0 ,0,1};
// //     glLightfv(GL_LIGHT3, GL_POSITION, pos4);
    
// //     glLightfv(GL_LIGHT0, GL_DIFFUSE, dcol);
// //     glLightfv(GL_LIGHT0, GL_SPECULAR, scol);
// //     glLightfv(GL_LIGHT1, GL_DIFFUSE, dcol);
// //     glLightfv(GL_LIGHT1, GL_SPECULAR, scol);
// //     glLightfv(GL_LIGHT2, GL_DIFFUSE, dcol);
// //     glLightfv(GL_LIGHT2, GL_SPECULAR, scol);
// //     glLightfv(GL_LIGHT3, GL_DIFFUSE, dcol);
// //     glLightfv(GL_LIGHT3, GL_SPECULAR, scol);
//     //  glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
//     //  glEnable(GL_FOG);
//     double ff = 0.99;
//     GLfloat fogColor[4] = {ff, ff, ff, 1.0};
//     int fogMode = GL_EXP;
//     glFogi (GL_FOG_MODE, fogMode);
//     glFogfv (GL_FOG_COLOR, fogColor);
//     glFogf (GL_FOG_DENSITY, 0.20);
//     glHint (GL_FOG_HINT, GL_DONT_CARE);
//     glFogf (GL_FOG_START, 2);
//     glFogf (GL_FOG_END, 4);
    
//     //  glEnable(GL_LIGHTING);
//     //    glEnable(GL_LIGHT0);
// //     glEnable(GL_LIGHT1);
// //     glEnable(GL_LIGHT2);
// //     glEnable(GL_LIGHT3);
    
//     glEndList();//1list
  
//     MAX_X = 1;
//     MIN_X = 0;
//     MAX_Y = 1;
//     MIN_Y = 0;
//     MAX_Z = 1;
//     MIN_Z = 0;
//     xc = (MIN_X + MAX_X)/2;
//     yc = (MIN_Y + MAX_Y)/2;
//     zc = (MIN_Z + MAX_Z)/2;
//     {
//       istringstream is(filename);
//       is >> sz;
//     }
//     {
//       istringstream is(filename1);
//       is >> N;
//     }
//     cout << "size is " << sz << endl;
//     vec.SetSize(N, N, N);
//     im.SetSize(N, N, N);
//     vec1.SetSize(sz,sz,sz);
//     try{
//       vec1.Read(filename, 0);
//     }catch(...)
//     {
//       for(int i = 0; i < sz; ++i)
// 	for(int j = 0; j<sz; ++j)
// 	  for(int k=0; k < sz; ++k)
// 	  {
// 	    //	  if (!((i<N/2)&&(j<N/2)&&(k<N/2)||(i+j+k<.9*N)))
// 	    //	    if ((i>10)&&(i<13))
// 	    int I = i - sz/2;
// 	    int J = j - sz/2;
// 	    int K = k - sz/2;
// 	    //	    if ((I*I + J*J + K*K)<(sz*sz/4))
// 	    if (!((i<sz/2)&&(j<sz/2)&&(k>sz/2)))
// 	      	    vec1(i,j,k)= 255;
// 	    else
// 	      vec1(i,j,k) = 0;
// 	  }
//     }
//     try{
//       im.Read(filename1, 0);
//     }catch (...)
//     {
//       for(int i = 0; i < N; ++i)
// 	for(int j = 0; j< N; ++j)
// 	  for(int k=0; k < N; ++k)
// 	    im(i,j,k) = 0;
//     }

//   } // firsttime
//   else
//   {
//   }
//   glDeleteLists(CurList,1);
//   CurList = glGenLists(1);
//   //  cout << CurList;
//   glNewList(CurList, GL_COMPILE);
//   if(CurDir!=Ndir)
//   {
//     for(int i = 0; i < N; ++i)
//       for(int j = 0; j < N; ++j)
// 	for(int k = 0; k < N; ++k)
// 	  vec(i,j,k)=0;
//   }
//   else
//   {
//     for(int i = 0; i < N; ++i)
//       for(int j = 0; j< N; ++j)
// 	for(int k=0; k < N; ++k)
// 	  if(N!=sz)
// 	    vec(i,j,k) = vec1(i *sz/N, j *sz/N, k*sz/N);
// 	  else
// 	    vec(i,j,k) = vec1(i,j,k);
//     GenIsoVals(im, vec);
//   }
//   for (int Line = CurLine; Line==CurLine; Line+=20)
//   {
//     for(int i = 0; i < N; ++i)
//       for(int j = 0; j < N; ++j)
//       {
// 	switch(CurDir)
// 	{
// 	case Ndir:
// 	  if(ShowPlates)
// 	  {
// 	    vec(i,j, 0) = 255;
// 	    vec(i,j,N-1)=255;
// 	  }
// 	  break;
// 	case Xdir:
// 	  vec(Line, i,j) = 255;
// 	  break;
// 	case Ydir:
// 	  vec(i,Line, j) = 255;
// 	  break;
// 	case Zdir:
// 	  vec(i,j,Line) = 255;
// 	  break;
// 	}
//       }
//   }
//   //  if(firsttime)
//   {
//     MinMax(im, vec, min, max);
//     if (min>=max)
//     {
//       min = -1;max= 1;
//     }
//   }
//   RemoveExtra(vec);
//   double bs = 1./N;
//   vsx = vsy = vsz = bs;
//   //   min=0;
//   if (ShowPalette)
//     for(int i = 0; i < N; ++i)
//     {
//       for (int k = 0; k < 5; ++k)
//       {
//          im(k,i,0) = min + (max-min)*static_cast<double>(i)/(N-1);
//          im(k,i,1) = min + (max-min)*static_cast<double>(i)/(N-1);
//          vec(k,i,0) = 255;
//          vec(k,i,1) = 255;
//       }
//     }
//   std::vector<color4> cols;
//   std::vector<color4> colsg;
//   std::vector<color4> colwhite;
//   double ccc = 0.4;
//   colwhite.push_back(color4(ccc,ccc,ccc,1));
//   colwhite.push_back(color4(ccc,ccc,ccc,1));

//   double alfa = 1;
//   //  cols.push_back(color4(0,   0, 0.5, alfa*1));
//   cols.push_back(color4(0,   0, 1, alfa*1));
//   cols.push_back(color4(0,   1, 1, alfa*0.1));
//   cols.push_back(color4(0,   .9, 0, alfa*0.02));
//   cols.push_back(color4(1,   1, 0, alfa*0.1));
//   cols.push_back(color4(1,   0, 0, alfa*1));
//   //  cols.push_back(color4(0.5,   0, 0, alfa*1));

//   color4 f(1,1,0,1);
//   color4 l(1,0,0,1);
//   double l1 = 1.3;
//   double o1 = 1;
//   double o2 = 0;
//   colsg.push_back(color4(l1, l1,l1,l1));
//   for(int i = 1; i < il; ++i)
//   {
//     double koeff = double(i)/(il-1);
//     double koeff1 = 1-koeff;

//     colsg.push_back(color4(l1, l1,l1,l1));
//     colsg.push_back(color4(l1, l1,l1,l1));
//     colsg.push_back(color4(o1, o1, o1, o1));
//     colsg.push_back(color4(o2, o2, o2, o2));
//     colsg.push_back(color4(o1, o1, o1, o1));
//     colsg.push_back(color4(l1, l1,l1,l1));
//     colsg.push_back(color4(l1, l1,l1,l1));
//   }
  
//   //  cout << "min - " << min << " max - " << max << endl;
//   glPointSize(800./N);
  
//   for(int i = 0; i < im.sizes[0]; ++ i)
//   {
// //         if ((CurDir == Xdir)&&(CurLine!=i))
// //       continue;
//     for(int j = 0; j < im.sizes[1]
// 	  ; ++j)
//     {
// //       if ((CurDir == Ydir)&&(CurLine!=j))
// // 	continue;
//       for (int k = 0; k <im.sizes[2]
// 	     ; ++k)
//       {
	
// 	//	if(check(vec,i, j, k))  //if(!(rand()%8))
// 	{
// // 	  if ((CurDir == Zdir)&&(CurLine!=k))
// // 	    continue;
// 	  int ii = i;
// 	  int jj = j;
// 	  int kk = k;
// 	  double val = im(i,j,k);
// 	  //	  if (val<0) continue;
// 	  double amax = std::max(fabs(max), fabs(min));
// 	  double cr, cg, cb, t;

// 	  bool positive = val>=0;
// 	  //	  double val1,val2;
//  	  double val1 = (val - min)/(max-min);
//  	  val1 = 1-val1;
// 	  double tr = 1;
// 	  double r, g, b;
// 	  double in = 0.5 + double(i)/(N-1)/4 + double(j)/(N-1)/4;
// 	  color4 col, grad;
// 	  GetColor(colsg, val1, grad);
//  	  val1 = val1*2 - 1;
// 	  val1 = (val1<0)?
// 	    -pow(fabs(val1), mypol):
// 	    pow(fabs(val1), mypol);
	  
//  	  val1 = (val1+1)/2;
// 	  GetColor(cols, val1, col);
// 	  GetColor(colsg, val1, grad);
// 	  //	  col*=grad;
// 	  //	  GetColor(colwhite, val1, col);
// 	  cr = col.r*in;
// 	  cb = col.b*in;
// 	  cg = col.g*in;
// 	  tr = col.a*tr;
// 	  // 	  if (tr <0.5)
// 	  // 	    continue;

// 	  DrawHex( i*bs, j*bs, k*bs, bs, bs, bs,
// 		   vec(ii,jj,kk),
// 		   cr, cg, cb, tr);
// 	}
//       }
//     }
//   }
//   xc = bs*vec.sizes[0]/2;
//   yc = bs*vec.sizes[1]/2;
//   zc = bs*vec.sizes[2]/2;

//   glEndList();

//   GenIsoList();

//   firsttime = false;
// }

void display()
{
  //  gl2psBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  gl2psLineWidth(1.25);
  glClearColor(1,1,1,1);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glAlphaFunc(GL_GEQUAL, 1-atest);
  //  gfxinit();
  double eps = 0.001;
  glDepthRange(eps,1);
  //  glCallList(CurList1);
  if(DrawSolid)
  {
    glDisable(GL_LIGHTING);
    glDisable(GL_BLEND);
    gl2psDisable(GL2PS_BLEND);
    GLenum mode;
    switch(WireFrame)
    {
    case 0: mode = GL_FILL;break;
    case 1: mode = GL_LINE;break;
    case 2: mode = GL_POINT;break;
    }
    glPolygonMode(GL_FRONT, mode);
    glCullFace(GL_BACK);
    //    glCallList(CurList);

    //    glDepthRange(0,1-eps);
    
    if (RemoveColor)
    {
      glEnable (GL_BLEND);
      gl2psEnable(GL2PS_BLEND);
      glEnable(GL_LIGHTING);
      glBlendFunc (GL_SRC_ALPHA ,GL_ONE_MINUS_SRC_ALPHA);
      glPolygonMode (GL_BACK, GL_LINE);
      glCullFace (GL_FRONT);// Don't Draw Any Front-Facing Polygons
      glDepthFunc (GL_LEQUAL);// Change The Depth Mode
      //glDepthRange(0, 1-eps);
      //      glCallList(CurList);
      glDepthFunc (GL_LESS);// Change The Depth Mode
    }else
      glPolygonMode (GL_BACK, GL_FILL);


  }
  glEnable(GL_BLEND);
  gl2psEnable(GL2PS_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  gl2psBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  gl2psLineWidth(0.7);
  glDepthFunc (GL_LEQUAL);// Change The Depth Mode
  if(DrawIso)
    glCallList(IsoList);
  glDepthFunc (GL_LESS);// Change The Depth Mode
  glFlush();
  glutSwapBuffers();
}

static int width, height;
void reshape (int w, int h)
{
  width = w;
  height = h;
  glViewport (0, 0, (GLsizei) w, (GLsizei) h); 
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluPerspective(30.0, (GLfloat) w/(GLfloat) h, 1.0, 200.0);
  orientMe();
}

#include <png.h>
int curframe = 0;
void save_png(int frame=curframe)
{
  curframe++;
  FILE *fp;
  ostringstream os;
  os.fill('0');
  os<<setw(4) << frame <<".out" << ".png";
  fp = fopen(os.str().c_str(), "wb");
  glReadBuffer(GL_FRONT);
  GLfloat *pixels = new GLfloat[width*height*4];
  glReadPixels(0,0, width, height, GL_RGBA, GL_FLOAT, pixels);
  //    cout << " " << width << "x" << height << endl;
  try{
    png_structp png_ptr = png_create_write_struct
      (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if (!png_ptr)
      throw(4);
      
    png_infop info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr)
    {
      png_destroy_write_struct(&png_ptr,
			       (png_infopp)NULL);
      throw 5;
    }
      
    if (setjmp(png_jmpbuf(png_ptr)))
    {
      png_destroy_write_struct(&png_ptr, &info_ptr);
      fclose(fp);
      throw 6;
    }
    png_init_io(png_ptr, fp);      
    /* set the zlib compression level */
    png_set_compression_level(png_ptr,
			      Z_BEST_COMPRESSION);

    /* set other zlib parameters */
    png_set_compression_mem_level(png_ptr, 8);
    png_set_compression_strategy(png_ptr,
				 Z_DEFAULT_STRATEGY);
    png_set_compression_window_bits(png_ptr, 15);
    png_set_compression_method(png_ptr, 8);
    png_set_compression_buffer_size(png_ptr, 8192);
    png_set_IHDR(png_ptr, info_ptr, width, height,
		 8, PNG_COLOR_TYPE_RGB_ALPHA, PNG_INTERLACE_NONE,
		 PNG_COMPRESSION_TYPE_DEFAULT,
		 PNG_FILTER_TYPE_DEFAULT);
    png_set_sRGB(png_ptr, info_ptr, PNG_sRGB_INTENT_PERCEPTUAL);
    png_color_8 a;
    a.red = 8;
    a.green=8;
    a.blue=8;
    a.gray=0;
    a.alpha=0;
    //    png_set_sBIT(png_ptr, info_ptr, &a);
    png_text  text[6];
    int  num_text = 0;
    text[num_text].compression = PNG_TEXT_COMPRESSION_NONE;
    text[num_text].key = "Title";
    text[num_text].text = "Kartinka";
    ++num_text;
    text[num_text].compression = PNG_TEXT_COMPRESSION_NONE;
    text[num_text].key = "Author";
    text[num_text].text = "Yavor Vutov";
    ++num_text;
    text[num_text].compression = PNG_TEXT_COMPRESSION_NONE;
    text[num_text].key = "Description";
    text[num_text].text = "3d Rendering";
    ++num_text;
    
    png_set_text(png_ptr, info_ptr, text, num_text);    
    png_set_packing(png_ptr);

    
    
    unsigned char **row_pointers = new unsigned char* [2*height];
    for (int i=0; i<height; i++)
      row_pointers[i]= new unsigned char[4*width];

    for(int i = 0; i < height;++i)
      for(int j = 0; j < width; ++j)
      {
	row_pointers[i][j*4] =  static_cast<unsigned char>(255*pixels[(height-1-i)*width*4 +j*4+0]);
	row_pointers[i][j*4+1] = static_cast<unsigned char>(255*pixels[(height-1-i)*width*4 +j*4+1]);
	row_pointers[i][j*4+2] = static_cast<unsigned char>(255*pixels[(height-1-i)*width*4 +j*4+2]);
	row_pointers[i][j*4+3] = static_cast<unsigned char>(255*pixels[(height-1-i)*width*4 +j*4+3]);
	//	row_pointers[i][j*4+3] = 255;
      }
    
    png_set_rows(png_ptr, info_ptr, row_pointers);

    //    png_write_info(png_ptr, info_ptr);
    
    png_write_png(png_ptr, info_ptr, PNG_TRANSFORM_IDENTITY, NULL);
    png_write_end(png_ptr, info_ptr);

    for (int i=0; i<height; i++)
      delete[] row_pointers[i];
    delete[] row_pointers;
    png_destroy_write_struct(&png_ptr, &info_ptr);
  }catch(int err){
    cerr << "Png write error " << err << endl;}
  delete[]pixels;
  fclose(fp);
}  
void save_eps(int frame=0)
{
  FILE *fp;
  int state = GL2PS_OVERFLOW, buffsize = 16*1024*1024;
  ostringstream os;
  os.fill('0');
  os<<setw(4) << frame <<".out" << ".eps";
  fp = fopen(os.str().c_str(), "wb");
  while(state == GL2PS_OVERFLOW){
    buffsize *= 2;
    gl2psBeginPage("test", "gl2psTestSimple", NULL,
		   GL2PS_EPS,
		   GL2PS_SIMPLE_SORT,
		   //GL2PS_BSP_SORT,
		   //                       GL2PS_DRAW_BACKGROUND |
		   GL2PS_USE_CURRENT_VIEWPORT
		   //		   |GL2PS_NO_BLENDING
		   |GL2PS_OCCLUSION_CULL
		   |GL2PS_SIMPLE_LINE_OFFSET
		   , 
		   GL_RGBA, 0, NULL, 0, 0, 0,  buffsize, fp, "Kartinka");
    display();
    state = gl2psEndPage();
  }
  fclose(fp);
}
void save_avi()
{
  const int numframes = 360;
  double step = atan(1)*8/numframes;
  for(int a = 0; a<numframes;++a)
  {
    save_eps(a);
    angle1 = step;
    orientMe();
    glutPostRedisplay();
  }
}
bool active = false;
void avi_png_idle();
void save_avi_png()
{
  active = true;
  curframe = 0;
  
  glutIdleFunc(avi_png_idle);
}
void gfxinit()
{
}
void menu(int item)
{
  glutPostRedisplay();
  switch(item) {
  case SAVE_EPS:
    save_eps();
    break;
  case SAVE_PNG:
    save_png();
    break;
  case SAVE_AVI:
    save_avi();
    break;
  case QUIT:
    exit(0);
  }
}


void inputKey(int key, int x, int y) {
  switch (key) {
  case GLUT_KEY_F1:
    swapp=not swapp;
    if (swapp)
    {
      for(int i = 0; i < 3; ++i)ox[i]=oy[i]=oz[i]=xob[i]=xu[i]=0;
      ox[0] = oy[1] = oz[2] = xob[0] = xu[1]= 1;
    }
    break;
  case GLUT_KEY_PAGE_DOWN:
    angle -=0.02f;
    break;
  case GLUT_KEY_LEFT : 
    angle1 += 0.02f;
    break;
  case GLUT_KEY_RIGHT :
    angle1 -=0.02f;
    break;
  case GLUT_KEY_UP : 
    angle2 +=0.02f;
    break;
  case GLUT_KEY_PAGE_UP:
    dist -= 0.1;
    break;
  case GLUT_KEY_DOWN :
    angle2 -=0.02f;
    break;
  case GLUT_KEY_F5 :
    DrawSolid = !DrawSolid;
    break;
  case GLUT_KEY_F7 :
    ShowPlates = !ShowPlates;
    gfxinit();
    break;
  case GLUT_KEY_F8 :
    ShowPalette = !ShowPalette;
    gfxinit();
    break;
  case GLUT_KEY_F10 :
    RemoveColor = ! RemoveColor;
    gfxinit();
    break;
  case GLUT_KEY_F11 :
    WireFrame = (WireFrame+1)%3;
    break;
  case GLUT_KEY_F12 :
    isobehav = (isobehav+1)%3;
    gfxinit();
    break;
  case GLUT_KEY_F6:
    DrawIso = !DrawIso;
    break;
  case GLUT_KEY_INSERT:
    dist += 0.1;
    break;
  default:
    return;
  }
  orientMe();
  glutPostRedisplay();
	
}
void inputKey1(unsigned char key, int x, int y) {
  switch (key) {
  case 127 : 
    angle += 0.02f;
    break;
  case 'n':
    CurDir = Ndir;
    gfxinit();
    break;
  case 'x':
    CurDir = Xdir;
    gfxinit();
    break;
  case 'y':
    CurDir = Ydir;
    gfxinit();
    break;
  case 'z':
    CurDir = Zdir;
    gfxinit();
    break;
  case 'j':
    if(CurLine)
    {
      CurLine--;
      gfxinit();
    }
    break;
  case 'k':
    if(CurLine<N-1)
    {
      CurLine++;
      gfxinit();
    }
    break;
  case 'm':
    if(mypol>0.1)
    {
      mypol -=0.08;
      gfxinit();
    }
    break;
  case ',':
    if(mypol<4)
    {
      mypol+=0.08;
      gfxinit();
    }
    break;
  case 'i':
    if(il<1000)
    {
      il++;
      gfxinit();
    }
    break;
  case 'u':
    if(il>1)
    {
      il--;
      gfxinit();
    }
    break;
  case 'I':
    if(il<1000)
    {
      il+=50;
      gfxinit();
    }
    break;
  case 'U':
    if(il>51)
    {
      il-=50;
      gfxinit();
    }
    break;
	  
  case 's':
    save_png();
    break;
  case 'a':
    save_avi_png();
    break;
  case '-':
    atest -= 1/256.;
    break;
  case '+':
    atest += 1/256.;
    break;
  default:
    return;
  }
  orientMe();
  glutPostRedisplay();
	
}

void avi_png_idle()
{
  const int numframes=360;
  
  save_png(curframe);
  double step = atan(1)*8/numframes;
  angle1 = step;
  orientMe();
  glutPostRedisplay();
  cout << curframe-1 << "\r" << flush;
  if(curframe==numframes)
  {
    cout << "                             " <<  "\r" << flush;
    glutIdleFunc(NULL);
  }
}

void MyGlutInit(int &argc, char **&argv)
{
    glutInit(&argc, argv);
    glutInitWindowPosition(-1, -1);
    glutInitWindowSize(800, 800);
  
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA| GLUT_DEPTH|GLUT_ALPHA);
    glutCreateWindow("Bone Drawer");
    glutDisplayFunc(display);
    //  glutIdleFunc(display);
    glutCreateMenu(menu);
    glutAddMenuEntry("Save EPS", SAVE_EPS);
    glutAddMenuEntry("Save PNG", SAVE_PNG);
    glutAddMenuEntry("Save AVI", SAVE_AVI);
    glutAddMenuEntry("Exit", QUIT);
    glutAttachMenu(GLUT_RIGHT_BUTTON);
    glutSpecialFunc(inputKey);
    glutKeyboardFunc(inputKey1);
    glutReshapeFunc(reshape);
    //  glutIdleFunc(idle);
    gfxinit();
}

int main(int argc, char **argv)
  try {

    filename = "64.bonec";
    filename1= "128.f.3.20";
    if (argc >1)
      filename = argv[1];
    if (argc >2)
      filename1 = argv[2];

    ReadData();

    //    MyGlutInit(argc, argv);
    //    glutMainLoop();
  } catch (const std::exception & e) {
    std::cerr << e.what() << std::endl;
  }
