#ifndef __functionlin_h__
#define __functionlin_h__
#include "small.h"
#include "mirorer.h"
#include <cassert>
#include <cmath>

template<class ValueT, class IndexT = ValueT, class StorageT=ValueT>
class LinearFunction3d
{
public:
  typedef ValueT value_type;
  typedef IndexT index_type;
  typedef SmallVector<index_type,3> P3D;
  vector3d<StorageT> buf;
  P3D Imin, Imax;
  P3D unit_size;
  P3D range;
  P3D zero;
  P3D zero1;
  bool interpo;
  template<class T>
  void Init(const vector3d<T>&v,
	    P3D const &min=0, P3D const &max=1,bool extrapolate = true, bool interpolate = true)
  {
    interpo = interpolate;
    Imin = min;
    Imax = max;
    range = Imax;
    range -= Imin;
    unit_size = range;

    for(int i = 0; i < 3; ++i)
      unit_size[i]/=v.sizes[i];
    buf.SetSize(v.sizes[0]+3, v.sizes[1]+3, v.sizes[2]+3);


    zero1 = unit_size;
    
    zero1 /= 2.;
    zero =  min;
    zero -= zero1;
    for(unsigned i = 0; i<buf.sizes[0]; ++i)
      for(unsigned j = 0; j<buf.sizes[1]; ++j)
	for(unsigned k = 0; k <buf.sizes[2];++k)
	  buf(i, j, k) = 0;

    for(unsigned i = 0; i <v.sizes[0]; ++i)
    {
      for(unsigned j = 0; j < v.sizes[1]; ++j)
      {
	for(unsigned k = 0; k < v.sizes[2];++k)
	{
	    buf(i+1, j+1, k+1) = v(i,j,k);
	}
      }
    }
    if(extrapolate)
    {
      for(unsigned i = 0; i < buf.sizes[0]; ++i)
	for(unsigned j = 0; j < buf.sizes[1]; ++j)
	{
	  const unsigned N = buf.sizes[2] - 2;
	  buf(i,j,0) = 2*buf(i,j,1) - buf(i,j,2);
	  buf(i,j,N) = 2*buf(i,j,N-1) - buf(i,j,N-2);
	  buf(i,j,N+1) = buf(i,j,N);
	}
      for(unsigned i = 0; i < buf.sizes[0]; ++i)
	for(unsigned k = 0; k < buf.sizes[2]; ++k)
	{
	  const unsigned N = buf.sizes[1] - 2;
	  buf(i,0,k) = 2*buf(i,1,k) - buf(i,2,k);
	  buf(i,N,k) = 2*buf(i,N-1,k) - buf(i,N-2,k);
	  buf(i,N+1,k) = buf(i, N, k);
	}
      for(unsigned j = 0; j < buf.sizes[1]; ++j)
	for(unsigned k = 0; k < buf.sizes[2]; ++k)
	{
	  const unsigned N = buf.sizes[0] - 2;
	  buf(0,j,k) = 2*buf(1,j,k) - buf(2,j,k);
	  buf(N,j,k) = 2*buf(N-1,j,k) - buf(N-2, j,k);
	  buf(N+1,j,k) = buf(N, j, k);
	}
    }
    else
    {
      for(unsigned i = 0; i < buf.sizes[0]; ++i)
	for(unsigned j = 0; j < buf.sizes[1]; ++j)
	{
	  const unsigned N = buf.sizes[2] - 2;
	  buf(i,j,0) = 0;//1*buf(i,j,1);// - buf(i,j,2);
	  buf(i,j,N) = 0;//1*buf(i,j,N-1);// - buf(i,j,N-2);
	  buf(i,j,N+1) = 0;//buf(i,j,N);
	}
      for(unsigned i = 0; i < buf.sizes[0]; ++i)
	for(unsigned k = 0; k < buf.sizes[2]; ++k)
	{
	  const unsigned N = buf.sizes[1] - 2;
	  buf(i,0,k) = 0;//1*buf(i,1,k);// - buf(i,2,k);
	  buf(i,N,k) = 0;//1*buf(i,N-1,k);// - buf(i,N-2,k);
	  buf(i,N+1,k) = 0;// buf(i, N, k);
	}
      for(unsigned j = 0; j < buf.sizes[1]; ++j)
	for(unsigned k = 0; k < buf.sizes[2]; ++k)
	{
	  const unsigned N = buf.sizes[0] - 2;
	  buf(0,j,k) = 0;//1*buf(1,j,k);// - buf(2,j,k);
	  buf(N,j,k) = 0;//1*buf(N-1,j,k);// - buf(N-2, j,k);
	  buf(N+1,j,k) = 0;// buf(N, j, k);
	}
    }
  }
  value_type RetVal(const SmallVector<int, 3> &ind, const P3D&c)const
  {
    for(int i = 0; i < 3; ++i)
    {
      assert(ind[i]>=0);
      assert((unsigned)ind[i]<buf.sizes[i]-1);
    }

    value_type val = 0;
    //    std::cout << "eh" << (int) buf(ind[0]  , ind[1]  , ind[2]) << std::endl;
    val += buf(ind[0]  , ind[1]  , ind[2])   * c[0]     * c[1]     * c[2];
    val += buf(ind[0]+1, ind[1]  , ind[2])   * (1-c[0]) * (c[1])   * (c[2]);
    val += buf(ind[0]  , ind[1]+1, ind[2])   * (c[0])   * (1-c[1]) * (c[2]);
    val += buf(ind[0]+1, ind[1]+1, ind[2])   * (1-c[0]) * (1-c[1]) * (c[2]);
    val += buf(ind[0]  , ind[1]  , ind[2]+1) * c[0]     * c[1]     * (1-c[2]);
    val += buf(ind[0]+1, ind[1]  , ind[2]+1) * (1-c[0]) * (c[1])   * (1-c[2]);
    val += buf(ind[0]  , ind[1]+1, ind[2]+1) * (c[0])   * (1-c[1]) * (1-c[2]);
    val += buf(ind[0]+1, ind[1]+1, ind[2]+1) * (1-c[0]) * (1-c[1]) * (1-c[2]);
    return val;
  }
  void CalcInd(P3D p, SmallVector<int, 3> &ind, P3D&c)const
  {
    //    std::cout <<p[0] << " " <<p[1] << " " << p[2]<< std::endl;
    //    std::cout <<x[0] << " " <<x[1] << " " << x[2]<< std::endl;
    //    std::cout <<ind[0] << " " <<ind[1] << " " << ind[2]<< std::endl;
    for(int i = 0; i < 3; ++i)
    {
      //      std::cout <<std::setw(10)<< p[i] << std::endl;
      if (p[i] <  (Imin[i] - zero1[i]))
        p[i] = Imin[i];
      if (!(p[i] >  (Imin[i] - zero1[i])))
	throw 1;
      if (p[i] >  (Imax[i] + zero1[i]))
	p[i] = Imax[i];
      if(!(p[i] <  (Imax[i] + zero1[i])))
	throw 2;
    }
/*     std::cout <<p[0] << " " <<p[1] << " " << p[2]<< std::endl; */
/*     std::cout <<Imax[0] << " " <<Imax[1] << " " << Imax[2]<< std::endl; */
/*     std::cout <<std::endl; */
    P3D x;
    x = p;
    x -= zero;
    for (int i = 0; i < 3; ++i)
    {
      ind[i] =(int) trunc(x[i]/unit_size[i]);
      //      std::cout << ind[i] << " ";
      c[i] = 1.-(x[i] - ind[i]*unit_size[i])/unit_size[i];
      //       ind[i]++;
      if(!interpo)
      {
	if (c[i] >0.5)
	  c[i] = 1;
	else
	  c[i] = 0;
      }
    }    
    
  }
  value_type operator()(const P3D &p)const try
  {
    SmallVector<int, 3> ind;
    P3D c;
    CalcInd(p, ind, c);
    return RetVal(ind, c);
  }
  catch(int a)
  {
    return value_type();
  }
    
  value_type operator()(index_type x, index_type y, index_type z)const
  {
    P3D p;
    p[0] = x; p[1] = y; p[2] = z;
    return (*this)(p);
  }
};



#endif

