#ifndef __functionpol_h__
#define __functionpol_h__
#include "small.h"
#include "mirorer.h"
#include "../programm/multipoly.h"
#include <cassert>
#include <cmath>

using My::Base::MultiPoly;

template<class ValueT, class IndexT = ValueT>
class PolyFunction3d
{
public:
  typedef ValueT value_type;
  typedef IndexT index_type;
  typedef SmallVector<index_type,3> P3D;
  typedef MultiPoly<ValueT, 3 > Poly3d;
  vector3d<Poly3d> buf;
  P3D Imin, Imax;
  P3D unit_size;
  P3D range;
  P3D zero;
  index_type eps;
  void Init(const std::string &fname, 
	    P3D const &min=0, P3D const &max=1, double eps_ = 0.000001)
  {
    eps = eps_;
    Imin = min;
    Imax = max;
    range = Imax;
    range -= Imin;
    unit_size = range;
    zero = min;
    std::ifstream f;
    using namespace std;
    f.exceptions(ios::failbit | ios::badbit |ios::eofbit);
    f.open(fname.c_str());
    std::string str;
    getline(f, str);
    str = str.substr(1);
    istringstream is(str);
    int N1, N2, N3;
    is >> N1 >> N2 >> N3 ;

    buf.SetSize(N1, N2, N3);
    for(int i = 0; i < 3; ++i)
      unit_size[i]/=buf.sizes[i];
    for (int k = 0; k < N3; ++k)
      for(int i = 0; i < N1; ++i)
	for (int j = 0; j < N2; ++j)
	{
	  do{
	    getline(f, str);
	  }
	  while((str.length()>0)&&(str[0]=='#'));
	  istringstream is(str);
	  int R;
	  is >> R;
	  buf(i,j,k)[R-1] = 0;
	  for (int r = 0; r < R; ++r)
	    is >> buf(i,j,k)[r];	  
	}

  }
  value_type RetVal(const SmallVector<int, 3> &ind, const P3D&c)const
  {
    assert(ind[0]>=0);
    assert(ind[1]>=0);
    assert(ind[2]>=0);
    assert(ind[0]<buf.sizes[0]);
    assert(ind[1]<buf.sizes[1]);
    assert(ind[2]<buf.sizes[2]); 
    return buf(ind[0], ind[1], ind[2])(c);
  }
  void CalcInd(const P3D &p, SmallVector<int, 3> &ind, P3D&c)const
  {
    P3D x;
    x = p;
    x -= zero;
    for (int i = 0; i < 3; ++i)
    {
      ind[i] =(int) trunc(x[i]/unit_size[i]);
      c[i] = -1+2*(x[i] - ind[i]*unit_size[i])/unit_size[i];
      if (ind[i]>=buf.sizes[i])
      {
	ind[i]--;
	c[i]+=2;
	std::cout << x[i] << " ooout " << i << std::endl;
      }
      if (ind[i] <0)
      {
	std::cout << "index " <<  ind[i] << std::endl;
	ind[i]++;
	c[i]-=2;
      }
	  
    }
  }
  value_type operator()(const P3D &p)const
  {
    SmallVector<int, 3> ind;
    P3D c;
    CalcInd(p, ind, c);
    return RetVal(ind, c);
  }
  value_type operator()(index_type x, index_type y, index_type z)const
  {
    SmallVector<int, 3> ind;
    P3D c;
    P3D p;
    p[0] = x; p[1] = y; p[2] = z;
    CalcInd(p, ind, c);
    return RetVal(ind, c);
  }
};



#endif

