
# include <vector>
# include <map>
# include <cmath>
# include <iostream>
# include <iomanip>
# include <fstream>
# include <string>
# include <sstream>
# include <stdexcept>
# include <GL/gl.h>
# include <GL/glu.h>
# include <GL/glut.h>
extern "C" {
# include "gl2ps.h"
}

#include <sstream>
//#include "meshreader.h"

#include "mirorer.h"

std::string filename;
using namespace std;

double MIN_X,MAX_X, MIN_Y, MAX_Y, MIN_Z, MAX_Z;
double vsx, vsy, vsz;
double minx, miny, minz, maxx, maxy, maxz;
double xc, yc, zc;

static float angle=0.4,ratio = 1;
static GLint snowman_display_list;
double dist = 2;

void orientMe(double ang)
{
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   double x = xc;//(MIN_X + MAX_X)/2;
   double y = yc;//(MIN_Y + MAX_Y)/2;
   double z = zc;// (MIN_Z + MAX_Z)/2;
   double xo = sin(angle)*dist;
   double zo = cos(angle)*dist;
   gluLookAt (x-xo, y, z -zo,  x, y, z, 0.0, 1.0, 0.0);
   GLfloat position [] = { x-xo, y, z-zo, 1};
  glLightfv(GL_LIGHT0, GL_POSITION, position);
  position[0] = position[2] =
  position[1] = 0.5;
  glLightfv(GL_LIGHT0, GL_SPOT_DIRECTION, position);
}



enum menu_items {
    SAVE_EPS,
    QUIT
};



void display()
{
  gl2psBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  gl2psLineWidth(0.25);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glCallList(1);
  glFlush();
  glutSwapBuffers();
}

void inputKey(int key, int x, int y) {
	switch (key) {
	case GLUT_KEY_LEFT : 
	  angle -= 0.01f;
	  break;
	case GLUT_KEY_RIGHT : 
	  angle +=0.01f;
	  break;
	case GLUT_KEY_UP : 
	  dist -= 0.2;
	  break;
	case GLUT_KEY_DOWN : ;
	  dist += 0.2;
	  break;
	default:
	  return;
	}
	orientMe(angle);
	glutPostRedisplay();
	
}



void color(double d)
{
    d *= 4;
    double r, g, b;
    if (d < 1.) {
        r = 0.; g = d; b = 1.;
    } else if (d < 2.) {
        r = 0.; g = 1.; b = 2. - d;
    } else if (d < 3.) {
        r = d - 2; g = 1.; b = 0.;
    } else {
        r = 1.; g = 4 - d; b = 0.;
    }
    glColor4d(r, g, b, 0.7);

}

void DrawHex(double x1, double x2, double x3,
	     double o1, double o2, double o3, double col=1.)
 {
  glBegin(GL_QUADS);
  glColor4d(col, col, col, 1.);
  glNormal3d(0, 0, -o3);
  glVertex3d(x1, x2, x3);
  glVertex3d(x1+o1, x2, x3);
  glVertex3d(x1+o1, x2+o2, x3);
  glVertex3d(x1, x2+o2, x3);

  glNormal3d(0, 0, o3);
  glVertex3d(x1,    x2,    x3+o3);
  glVertex3d(x1+o1, x2,    x3+o3);
  glVertex3d(x1+o1, x2+o2, x3+o3);
  glVertex3d(x1,    x2+o2, x3+o3);

  col*=0.9;
  glColor4d(col, col, col, 1.);
  glNormal3d(0, -o2, 0);
  glVertex3d(x1,    x2,    x3);
  glVertex3d(x1+o1, x2,    x3);
  glVertex3d(x1+o1, x2,    x3+o3);
  glVertex3d(x1,    x2,    x3+o3);

  glNormal3d(0, o2, 0);
  glVertex3d(x1,    x2+o2,    x3);
  glVertex3d(x1+o1, x2+o2,    x3);
  glVertex3d(x1+o1, x2+o2,    x3+o3);
  glVertex3d(x1,    x2+o2,    x3+o3);

  col*=0.9;
  glColor4d(col, col, col, 1.);
  glNormal3d(-o1, 0, 0);
  glVertex3d(x1,    x2,    x3);
  glVertex3d(x1,    x2+o2,    x3);
  glVertex3d(x1,    x2+o2,    x3+o3);
  glVertex3d(x1,    x2,    x3+o3);

  glNormal3d(o1, 0, 0);
  glVertex3d(x1+o1,    x2,    x3);
  glVertex3d(x1+o1,    x2+o2,    x3);
  glVertex3d(x1+o1,    x2+o2,    x3+o3);
  glVertex3d(x1+o1,    x2,    x3+o3);


  glEnd();

}

template <class T>
inline bool In(const T&a, const T&min, const T &max)
{
  return (a >=min) && (a<max);
}
template<class T, class I>
bool check(T &vec, I i, I j, I k)
{
    ptrdiff_t szs[3];
    bool ret = true;
    if (i>0)
      ret = ret && vec(i-1, j, k);
    else
      ret = false;
    if (j>0)
      ret = ret && vec(i, j-1, k);
    else
      ret = false;
    if (k>0)
      ret = ret && vec(i, j, k-1);
    else
      ret = false;
    if (i<vec.sizes[0]-1)
      ret  = ret && vec(i+1, j, k);
    else
      ret = false;
    if (j<vec.sizes[1]-1)
      ret  = ret && vec(i, j+1, k);
    else
      ret = false;
    if (k<vec.sizes[2]-1)
      ret = ret && vec(i, j, k+1);
    else
      ret = false;
    bool tmp = vec(i,j,k);
    int sz1, sz2, sz3;
    sz1 = vec.sizes[0]/2;
    sz2 = vec.sizes[1]/2;
    sz3 = vec.sizes[2]/2;
//     if ((i<sz1)&&(j<sz2) && (k <sz3))
//       return false;
    if ((i==sz1)||(j==sz2) || (k==sz3))
      return vec(i,j,k);
    return tmp^ret;
}
 
void gfxinit ()
{
  glNewList(1, GL_COMPILE);
  
  MAX_X = 1;
  MIN_X = 0;
  MAX_Y = 1;
  MIN_Y = 0;
  MAX_Z = 1;
  MIN_Z = 0;
  xc = (MIN_X + MAX_X)/2;
  yc = (MIN_Y + MAX_Y)/2;
  zc = (MIN_Z + MAX_Z)/2;
  istringstream is(filename);
  int sz;
  is >> sz;
  cout << "size is " << sz << endl;
  vector3d<unsigned char > vec;
  vector3d<unsigned char > vec1;
  vec.SetSize(sz,sz,sz);
  vec1.SetSize(sz,sz,sz);
  vec1.Read(filename, 0);
  for(int i = 0; i < sz; ++i)
    for(int j = 0; j < sz; ++j)
      {
	vec1(i,j, 0) = 255;
	vec1(i,j,sz-1)=255;
      }
  for(int i = 0; i < sz; ++i)
    for(int j = 0; j< sz; ++j)
      for(int k=0; k < sz; ++k)
	vec(i,j,k) = vec1(sz-1-i,k,j);
  cout << "read" << endl;
  double bs = 1./sz;
  vsx = vsy = vsz = bs;
  //  DrawHex(0.25, 0.25, 0.25, 0.75, 0.75, 0.75);
  glColor4d(1.0, 1.0, 1.0, 1.);
  for(int i = 0; i < vec.sizes[0]; ++ i)
    for(int j = 0; j < vec.sizes[1]
	  ; ++j)
      for (int k = 0; k <vec.sizes[2]
	     ; ++k)
      {
	if(check(vec,i, j, k))  //if(!(rand()%8))
	{
	  double color = vec(i,j,k)/255.;
	  color = 1.5 - color;
	  	  color = 1;
	  color *=  ((double)i+k)/vec.sizes[0]/1.4143 ;
	  //	  if(!(rand()%1))
	  DrawHex( i*bs, j*bs, k*bs, bs, bs, bs, 0.62 + 0.35*color);
	}
      }

  xc = bs*vec.sizes[0]/2;
  yc = bs*vec.sizes[1]/2;
  zc = bs*vec.sizes[2]/2;

  glEndList();

  glLineWidth(0.001);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_POINT_SMOOTH);
  glEnable(GL_BLEND);
  glShadeModel(GL_FLAT);
  glEnable(GL_LINE_SMOOTH);
  glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glClearColor(1,1,1,0);
  GLfloat position [] = { 0.5, 0.5, -10., 1.0 };
  //  glMatrixMode(GL_MODELVIEW);
  //  glLoadIdentity();

    glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 50.);
  glColorMaterial(GL_FRONT_AND_BACK, GL_SPECULAR);
  glEnable(GL_COLOR_MATERIAL);
  GLfloat amb[] = {1, 1, 1, 1.0};
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, amb);
  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, amb);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, amb);

//  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, 0);
  glLightfv(GL_LIGHT0, GL_POSITION, position);
  amb[1]=0;
  glLightfv(GL_LIGHT0, GL_SPECULAR, amb);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, amb);
  glLightf(GL_LIGHT0,GL_SPOT_EXPONENT, 40.);
  glLightf(GL_LIGHT0, GL_QUADRATIC_ATTENUATION, 1.2);
  amb[0]=amb[1]=amb[2]=0.3;
  glLightfv(GL_LIGHT0, GL_AMBIENT, amb);
  position[0]=position[1]=position[2]=0.5;
  glLightfv(GL_LIGHT0, GL_SPOT_DIRECTION, position);
  
  //  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  //  glEnable(GL_LIGHT1);
  using namespace std;
  cout << MIN_X << " " << MAX_X << endl 
       << MIN_Y << " " << MAX_Y << endl
       << MIN_Z << " " << MAX_Z << endl;
  cout << (MAX_X - MIN_X)/vsx << endl
       << (MAX_Y - MIN_Y)/vsy << endl
       << (MAX_Z - MIN_Z)/vsz << endl;

}

void reshape (int w, int h)
{
   glViewport (0, 0, (GLsizei) w, (GLsizei) h); 
   glMatrixMode (GL_PROJECTION);
   glLoadIdentity ();
   gluPerspective(30.0, (GLfloat) w/(GLfloat) h, 1.0, 200.0);

   orientMe(angle);
}

void save_eps()
{
    FILE *fp;
    int state = GL2PS_OVERFLOW, buffsize = 16*1024*1024;
    
    fp = fopen("out.eps", "wb");
    while(state == GL2PS_OVERFLOW){
        buffsize *= 2;
        gl2psBeginPage("test", "gl2psTestSimple", NULL,
                       GL2PS_EPS,
		       GL2PS_SIMPLE_SORT
		       /*GL2PS_BSP_SORT*/, 
		       //                       GL2PS_DRAW_BACKGROUND |
		       GL2PS_USE_CURRENT_VIEWPORT
		       |GL2PS_OCCLUSION_CULL
		       , 
                       GL_RGBA, 0, NULL, 0, 0, 0,  buffsize, fp, "out.eps");
        display();
        state = gl2psEndPage();
    }
    fclose(fp);
}

void menu(int item)
{
    switch(item) {
    case SAVE_EPS:
        save_eps();
        break;
    case QUIT:
        exit(0);
    }
    glutPostRedisplay();
}



int main(int argc, char * argv[])
try {

  if (argc >1)
    filename = argv[1];
  //  ReadMesh("zv_s06-66_fe.mesh");
  
  glutInit(&argc, argv);
  glutInitWindowPosition(-1, -1);
  glutInitWindowSize(500, 500);
  
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
  glutCreateWindow("Bone Drawer");
  glutDisplayFunc(display);
  //  glutIdleFunc(display);
  glutCreateMenu(menu);
  glutAddMenuEntry("Save EPS", SAVE_EPS);
  glutAddMenuEntry("Exit", QUIT);
  glutAttachMenu(GLUT_RIGHT_BUTTON);
  glutSpecialFunc(inputKey);
  glutReshapeFunc(reshape);
  gfxinit();
  glutMainLoop();
} catch (const std::exception & e) {
    std::cerr << e.what() << std::endl;
}
