#include "mirorer.h"
#include <vector>
#include <map>
#include "mutils.h"
#include <fstream>
#include <set>


#define BCOMPRESS
#ifdef BCOMPRESS
#include "boost/iostreams/filtering_streambuf.hpp"
#include "boost/iostreams/filtering_stream.hpp"
#include "boost/iostreams/filter/bzip2.hpp"
#endif

std::ofstream ff;
#ifdef BCOMPRESS
boost::iostreams::filtering_stream<boost::iostreams::output> bz;
#endif
std::ostream *of;


typedef int index_type;
int G1, G2, G3;
int N1, N2, N3;
int l1;
int l2;
int l3;
int s1;
int s2;
int s3;
int size1;
int size2;
int size3;
index_type start_point;
index_type end_point;
index_type voxels = 0;
index_type boundaryes=0;
index_type points=2;
index_type locals[8];
double r1, r2, r3;
std::vector<index_type> indexes;
vector3d<unsigned char > vec;
const unsigned char good_thres = 10;
std::string fname_base;
enum Faces{Left, Right, Front, Back, Down, Up};
template<class Int>
inline index_type GetInd(Int s1, Int s2, Int s3, Int i, Int j, Int k, Int o1=1, Int o2=1, Int o3=1)
{
  return (index_type)(s1+o1)*(s2+o2)*k + (index_type)(s1+o1)*j +i;
}

typedef std::map<std::pair<unsigned char, unsigned char>, int> b_type;
b_type boundaries;
typedef b_type::iterator miter;

typedef std::set<std::pair<unsigned char, unsigned char> > intrested_in_type;
intrested_in_type intrested_in;
miter  Makeit(Faces f)
{
  std::pair<unsigned char, unsigned char> ind(f, f);
  int r = f;
  std::pair<miter, bool> it = boundaries.insert(std::make_pair(ind, r) );
  return it.first;
}
miter GetNeighbour(int i, int j, int k, Faces which)
{
  static const int I[6]={-1, +1,  0,  0,  0, 0};
  static const int J[6]={ 0,  0, -1, +1,  0, 0};
  static const int K[6]={ 0,  0,  0,  0, -1, +1};
  static int first_free=6;
  int ii = I[which]+i;
  int jj = J[which]+j;
  int kk = K[which]+k;
  if (ii<0) return Makeit(Left);
  if (ii>=N1) return Makeit(Right);
  if (jj<0) return Makeit(Front);
  if (jj>=N2) return Makeit(Back);
  if (kk<0) return Makeit(Down);
  if (kk>=N3) return Makeit(Up);
  std::pair<unsigned char, unsigned char> ind(vec(i,j,k), vec(ii,jj,kk));
  if (ind.first==ind.second)
    return boundaries.end();
  miter it = boundaries.find(ind);
  if (it == boundaries.end())
  {
    int ret = first_free++;
    std::pair<miter, bool> it1 = boundaries.insert(std::make_pair(ind, ret) );
    it = it1.first;
  }
  return it;
}

miter GetNeighbourReal(int i, int j, int k, Faces which)
{
  miter it = GetNeighbour(i,j,k,which);
  if (intrested_in.find(it->first)==intrested_in.end())
    return boundaries.end();
  else
    return it;     
}
void FillLocals(int i, int j, int k)
{
  locals[0] = indexes[GetInd(N1, N2, N3, i  , j  , k  )]-2;
  locals[1] = indexes[GetInd(N1, N2, N3, i  , j  , k+1)]-2;
  locals[2] = indexes[GetInd(N1, N2, N3, i  , j+1, k  )]-2;
  locals[3] = indexes[GetInd(N1, N2, N3, i  , j+1, k+1)]-2;
  locals[4] = indexes[GetInd(N1, N2, N3, i+1, j  , k  )]-2;
  locals[5] = indexes[GetInd(N1, N2, N3, i+1, j  , k+1)]-2;
  locals[6] = indexes[GetInd(N1, N2, N3, i+1, j+1, k  )]-2;
  locals[7] = indexes[GetInd(N1, N2, N3, i+1, j+1, k+1)]-2;
}
void CountGoodVoxels()
{
  for (int K = 0; K < size3; ++K)
  {
    for (int I = 0; I < size1; ++I)
      for (int J= 0; J < size2; ++J)
      {
	int i = I+s1;
	int j = J+s2;
	int k = K+s3;
	if (vec(i,j,k)>good_thres)
	{
	  indexes[GetInd(N1, N2, N3, i  , j  , k  )] = 1;
	  indexes[GetInd(N1, N2, N3, i  , j  , k+1)] = 1;
	  indexes[GetInd(N1, N2, N3, i  , j+1, k  )] = 1;
	  indexes[GetInd(N1, N2, N3, i  , j+1, k+1)] = 1;
	  indexes[GetInd(N1, N2, N3, i+1, j  , k  )] = 1;
	  indexes[GetInd(N1, N2, N3, i+1, j  , k+1)] = 1;
	  indexes[GetInd(N1, N2, N3, i+1, j+1, k  )] = 1;
	  indexes[GetInd(N1, N2, N3, i+1, j+1, k+1)] = 1;
	  voxels++;
	  for (int face = 0; face<6; ++face)
	  {
	    if (GetNeighbourReal(i,j,k,static_cast<Faces>(face))!=boundaries.end())
	      boundaryes++;
	  }
	}
      }
  }
}

void SaveBoundaries()
{
  std::ofstream of("boundaries.txt.dat");
  for(miter it = boundaries.begin(); it!=boundaries.end(); it++)
    if(intrested_in.find(it->first)!=intrested_in.end())
      of << "(" <<(int)it->first.first << ", " << (int)it->first.second << ") --> " << it->second << "\n";
}

void CountGoodVoxels2()
{
  for (int K = 0; K < size3; ++K)
  {
    for (int I = 0; I < size1; ++I)
      for (int J= 0; J < size2; ++J)
      {
	int i = I+s1;
	int j = J+s2;
	int k = K+s3;
	if (vec(i,j,k)>good_thres)
	{
	  voxels++;
	  for (int face = 0; face<6; ++face)
	  {
	    if (GetNeighbourReal(i,j,k,static_cast<Faces>(face))!=boundaries.end())
	      boundaryes++;
	  }
	}
      }
  }
}

void OutputFace(int face)
{
  static int inds[36] = {
    0, 1, 3,
    3, 2, 0,
    7, 5, 4,
    4, 6, 7,
    0, 4, 5,
    5, 1, 0,
    7, 6, 2,
    2, 3, 7,
    1, 5, 7,
    7, 3, 1,
    6, 4, 0,
    0, 2, 6};
  *of << locals[inds[face*3 + 0]] << " " << locals[inds[face*3 + 1]] << " " <<locals[inds[face*3 + 2]];
}

void OutputTet(int tet)
{
  static int tets[24] = {
    0, 5, 1, 7,
    5, 0 ,4, 7,
    0, 3, 2, 7,
    0, 1, 3, 7,
    6, 0, 2, 7,
    0, 6, 4, 7};
  *of << locals[tets[tet*4 + 0]] << " " << locals[tets[tet*4 + 1]] << " " <<locals[tets[tet*4 + 2]]<< " " <<locals[tets[tet*4+3]];
}
void OutputBounds()
{
  for (int K = 0; K < size3; ++K)
  {
    for (int I = 0; I < size1; ++I)
      for (int J= 0; J < size2; ++J)
      {
	int i = I+s1;
	int j = J+s2;
	int k = K+s3;
	if (vec(i,j,k)>good_thres)
	{
	  FillLocals(i,j,k);
	  for (int face = 0; face<6; ++face)
	  {
	    miter it;
	    if ((it = GetNeighbourReal(i,j,k,static_cast<Faces>(face)))!=boundaries.end())
	    {
	      OutputFace(face*2);
	      *of << " " << it->second << " " << (int)it->first.first << " " << (int)it->first.second <<"\n";
	      OutputFace(face*2+1);
	      *of << " " << it->second << " " << (int)it->first.first << " " << (int)it->first.second <<"\n";
	    }
	  }
	}
      }
  }
}
void OutputTets()
{
  for (int K = 0; K < size3; ++K)
  {
    for (int I = 0; I < size1; ++I)
      for (int J= 0; J < size2; ++J)
      {
	int i = I+s1;
	int j = J+s2;
	int k = K+s3;
	if (vec(i,j,k)>good_thres)
	{
	  FillLocals(i,j,k);
	  for (int tet = 0; tet<6; ++tet)
	  {
	    *of << (int)vec(i,j,k) << " ";
	    OutputTet(tet);
	    *of << "\n";
	  }
	}
      }
  }
}


void CloseFile()
{
#if defined(BCOMPRESS)
  bz.flush();
  bz.pop();
  bz.pop(); //closes the file ff
  ff.close();
#else
  ff.flush();
  ff.close();
#endif
}
void OpenFile(int gi, int gj, int gk)
{
  int rank = GetInd(G1, G2, G3, gi, gj, gk, 0, 0, 0);
  int size = G1*G2*G3;
    
  std::string gfname = fname_base;
  std::string path = SlashifyPath(gfname);
  if (rank==0)
  {
    MkDir(gfname);
    std::ofstream f((path+"pmesh").c_str());
    f <<size << "\n";
    for (int i = 0; i < size; ++i)
    {
      f << i << " " << FnameGen(size, i, /*path*/"", ".part_mesh") << "\n";
      if (i < 16)
      {
	std::string fnamed = FnameGen(size, i, path, ".part_mesh");
	MkDir(GetPath(fnamed));
      }
    }
  }
  std::string fname = FnameGen(size, rank, path, ".part_mesh");
#if defined(BCOMPRESS)
  fname+=".bz2";
#endif
  ff.open(fname.c_str());
#if defined(BCOMPRESS)
  bz.push(boost::iostreams::bzip2_compressor());
  bz.push(ff);
  of = &bz;//(&bz);
#else
  of = &ff;
#endif
}

void SetPoints()
{
  start_point = points;
  for (int K = 0; K < size3+l3; ++K)
    for (int I = 0; I < size1+l1; ++I)
      for (int J= 0; J < size2+l2; ++J)
      {
	int i = I+s1;
	int j = J+s2;
	int k = K+s3;
	if(indexes[GetInd(N1, N2, N3, i,j,k)]==1)
	  indexes [GetInd(N1, N2, N3, i,j,k)]=points++;
      }
  end_point = points-1;
}
void CountPoints()
{
  start_point = points;
  for (int K = 0; K < size3+l3; ++K)
    for (int I = 0; I < size1+l1; ++I)
      for (int J= 0; J < size2+l2; ++J)
      {
	int i = I+s1;
	int j = J+s2;
	int k = K+s3;
	if(indexes[GetInd(N1, N2, N3, i,j,k)]>1)
	  points++;
      }
  end_point = points-1;
}
index_type CountPoints1()
{
  index_type ret = 0;
  for (int K = 0; K < size3+1; ++K)
    for (int I = 0; I < size1+1; ++I)
      for (int J= 0; J < size2+1; ++J)
      {
	int i = I+s1;
	int j = J+s2;
	int k = K+s3;
	if(indexes[GetInd(N1, N2, N3, i,j,k)]>1)
	  ret++;
      }
  return ret;
}
std::vector<int> starts[3];
void InitPieces()
{
  starts[0].resize(G1+1);
  starts[1].resize(G2+1);
  starts[2].resize(G3+1);
  
  for (int i = 0; i <G3; ++i)
    starts[2][i] = (N3+1)*i/G3;
  starts[0][0] = 1.5*(N1+1)/G1;
  starts[0][G1] = 1.5*(N1+1)/G1;
  for (int i = 1; i < G1; ++i)
  {
    starts[0][i] = (N1 - 2*starts[0][0]+1)/(G1-2);
  }
  
  starts[1][0] = 1.5*(N2+1)/G2;
  starts[1][G2] = 1.5*(N2+1)/G2;
  for (int i = 1; i < G2; ++i)
  {
    starts[1][i] = (N2 - 2*starts[1][0]+1)/(G2-2);
  }

  
  int current = 0;
  for (int i = 0; i <= G1; ++i)
  {
    int next = starts[0][i];
    starts[0][i] = current;
    current += next;
  }
  current = 0;
  for (int i = 0; i <= G2; ++i)
  {
    int next = starts[1][i];
    starts[1][i] = current;
    current += next;
  }
  starts[0][G1] = N1;
  starts[1][G2] = N2;
  starts[2][G3] = N3;
}

void SetBigPiece1(int g1, int g2, int g3)
{
  l1 = g1+1==G1;
  l2 = g2+1==G2;
  l3 = g3+1==G3;
  s1 = starts[0][g1];
  s2 = starts[1][g2];
  s3 = starts[2][g3];
  size1 = starts[0][g1+1];
  size2 = starts[1][g2+1];
  size3 = starts[2][g3+1];
  size1-=s1;
  size2-=s2;
  size3-=s3;
}
void SetBigPiece(int g1, int g2, int g3)
{
  l1 = g1+1==G1;
  l2 = g2+1==G2;
  l3 = g3+1==G3;
  s1 = (N1+1)*g1/G1;
  s2 = (N2+1)*g2/G2;
  s3 = (N3+1)*g3/G3;
  size1 = (N1+1-l1)*(g1+1)/G1;
  size2 = (N2+1-l2)*(g2+1)/G2;
  size3 = (N3+1-l3)*(g3+1)/G3;
  size1-=s1;
  size2-=s2;
  size3-=s3;
}
 
void CountBigPiece(int g1, int g2, int g3)
{
  CountGoodVoxels();
}

void MakeNumbers(int g1, int g2, int g3)
{
  SetPoints();
}

void SaveMesh(int g1, int g2, int g3)
{
  OpenFile(g1, g2, g3);
  CountPoints();
  *of << (start_point-2) << " " << (end_point-2) << "\n";
  *of << CountPoints1() << "\n";
  for (int K = 0; K < size3+1; ++K)
    for (int I = 0; I < size1+1; ++I)
      for (int J= 0; J < size2+1; ++J)
      {
	int i = I+s1;
	int j = J+s2;
	int k = K+s3;
	index_type ind;
	if((ind = indexes[GetInd(N1, N2, N3, i,j,k)])>1)
	  *of << (ind-2) << " " << r1*i << " " << r2*j << " " << r3*k << "\n";
      }
  voxels = 0;
  boundaryes = 0;
  CountGoodVoxels2();
  *of << boundaryes*2 << "\n";
  OutputBounds();
  *of << voxels*6 << "\n";
  OutputTets();
  CloseFile();
}

void TraverseBig(void (*func)(int, int, int))
{
  for (int k = 0; k< G3; ++k)
    for (int j = 0; j < G2; ++j)
      for (int i = 0; i < G1; ++i)
      {
	SetBigPiece(i, j, k);
	(*func)(i,j,k);
      }
}

void MakeIntresting()
{
  intrested_in.insert(std::make_pair(255, 1));
  intrested_in.insert(std::make_pair(255, 2));
  intrested_in.insert(std::make_pair(255, 3));
  intrested_in.insert(std::make_pair(255, 4));

  intrested_in.insert(std::make_pair(40, 1));
  intrested_in.insert(std::make_pair(40, 2));
  intrested_in.insert(std::make_pair(40, 3));
  intrested_in.insert(std::make_pair(40, 4));

  intrested_in.insert(std::make_pair(0, 0));
  intrested_in.insert(std::make_pair(1, 1));
  intrested_in.insert(std::make_pair(2, 2));
  intrested_in.insert(std::make_pair(3, 3));
  intrested_in.insert(std::make_pair(4, 4));
  intrested_in.insert(std::make_pair(5, 5));
}

int main(int argc, char ** argv)
{
  std::string filename = "default";
  if (argc >1)
    filename = argv[1];
  fname_base = "omesh";
  if (argc >2)
    fname_base = argv[2];
  vec.ReadInr(filename);
  std::cout << "read!" << std::endl;
  N1 = vec.sizes[0];
  N2 = vec.sizes[1];
  N3 = vec.sizes[2];
  r1 = vec.dim[0];
  r2 = vec.dim[1];
  r3 = vec.dim[2];
  G1=8;
  G2=8;
  G3=16;

  InitPieces();
  MakeIntresting();
  // N1=N2=N3=2;
  // G1=G2=G3=2;
  // vec(1,1,0) = 255;
  // vec(1,1,1) = 255;
  // vec(1,0,0) = 255;
  // vec(1,0,1) = 255;
  // vec(0,1,0) = 255;
  // vec(0,1,1) = 255;
  // vec(0,0,0) = 1;
  // vec(0,0,1) = 255;
  
  indexes.resize((size_t)(N1+1)*(N2+1)*(N3+1), 0);
  std::cout << "Starting" << std::endl;
  voxels=0;
  boundaryes = 0;
  TraverseBig(&CountBigPiece);
  points=2;
  TraverseBig(&MakeNumbers);
  std::cout << voxels << " voxels, "<< boundaryes << " faces and "<< points-2 << " points" << std::endl;
  points=2;
  TraverseBig(&SaveMesh);
  SaveBoundaries();
}
