#ifndef __MESHREADER_H__
#define __MESHREADER_H__
#include <vector>
#include <fstream>
#include <string>
#include "small.h"

template <class T>class Point3d:public SmallVector<T, 3>{};
template <class Int>class Hexagon:public SmallVector<Int, 8>{};

std::vector<Point3d<double> > nodes;
std::vector<Hexagon<unsigned> > ellements;

void GetOffs(Hexagon<unsigned> const &  el, Point3d<double> &p)
{
  p[0] = -nodes[el[0]][0] + nodes[el[6]][0];
  p[1] = -nodes[el[0]][1] + nodes[el[6]][1];
  p[2] = -nodes[el[0]][2] + nodes[el[6]][2];
}

template<class F>
void SkipComments(F &f)
{
  char ch;
  do{
    f >> ch;
    if (!f)
      return;
    f.putback(ch);
    std::string s;
    if(ch=='#')
    {
      std::getline(f, s);
      //      std::cout << s << std::endl;
    }
  }while(ch=='#');
}

void ReadMesh(std::string fname)
{
  std::ifstream f(fname.c_str());
  SkipComments(f);
  using namespace std;
  string eltype;
  f >> eltype;
  int nodecount, elcount;
  double tmp;
  extern double vsx, vsy, vsz;

  f >> elcount >> nodecount ;
  f >> tmp >> tmp >> tmp >> tmp >> tmp ;
  SkipComments(f);
  f >> vsx >> vsy >> vsz;
  SkipComments(f);
  f >> tmp >>tmp;
  SkipComments(f);
  f >> tmp >>tmp;
  SkipComments(f);
  f >> tmp >>tmp;
  SkipComments(f);
  while(nodecount--)
    {
      if(!f)
	break;
      SkipComments(f);
      Point3d<double> pt;
      f >> pt[0] >> pt[1] >> pt[2];
      nodes.push_back(pt);
      //      cout << pt[0] << "\t"  << pt[1] <<"\t"<< pt[2] << endl;
    }
  while(elcount--)
    {
      if(!f)
	break;
      SkipComments(f);
      Hexagon<unsigned> el;
      for(int i = 0; i < 8; ++i)
	{
	  f >> el[i];
	  --el[i];
	}
      Point3d<double> pt;
      //      GetOffs(el, pt);
      //      cout << pt[0] << "\t"  << pt[1] <<"\t"<< pt[2] << endl;
      ellements.push_back(el);
    }
  
}



#endif
