#ifndef __mutils_h__
#define __mutils_h__
#include <string>
#include <sstream>
#include <iostream>
#include <cassert>
#include <utility>
#include <iomanip>
#include "mkdir.h"

template<class a, class b>
std::ostream& operator << (std::ostream &os, std::pair<a, b> const &p)
{
  os << "(" << p.first << ", " << p.second << ") ";
  return os;
}

template<class T>
std::string ToS(const T &a)
{
  std::ostringstream os;
  os << a;
  return os.str();
}
template<class T>
std::string ToS0(const T &a)
{
  std::ostringstream os;
  os.fill('0');
  os << std::setw(3) << a;
  return os.str();
}


inline std::string GetFname(const std::string &n)
{
  size_t pos = n.find_last_of('/');
  if (pos==n.npos)
    return n;
  else
    return n.substr(pos+1);
}
inline std::string GetPath(const std::string &n)
{
  size_t pos = n.find_last_of('/');
  if (pos==n.npos)
    return "./";
  else
    return n.substr(0,pos+1);
}

inline std::string SlashifyPath(const std::string &n)
{
  if(n.length()==0)
    return "./";
  if(n[n.length()-1]!='/')
    return n+"/";
  return n;
}

inline std::string UnSlashifyPath(const std::string &n)
{
  std::string f = SlashifyPath(n);
  return f.substr(0, f.length()-1);
}

inline std::string FnameGen(int size, int rank, const std::string &base, const std::string &suff)
{
  std::ostringstream os;
  os.fill('0');
  os << base << std::setw(2)<<(rank&15) <<"/"  << std::setw(5) << rank << suff;
  return os.str();
}

inline std::string FnameGen1(int size, int rank, const std::string &base, const std::string &suff)
{
  std::ostringstream os;
  os.fill('0');
  os << base << "." << size << suff;
  return os.str();
}

template<class T, class V, class F>
void SlowSortTwoVectors(T &a, V &b, F f = std::less<T>())
{
  assert(a.size()==b.size());
  typedef typename T::value_type val;
  for (size_t i = 0; i <a.size(); ++i)
  {
    val max = a[i];
    size_t maxin=i;
    for (size_t j = i+1; j < a.size(); ++j)
      if (f(a[j],max))
      {
	max = a[j];
	maxin = j;
      }
    if (maxin!=i)
    {
      using std::swap;
      swap(a[i], a[maxin]);
      swap(b[i], b[maxin]);
    }
  }
}
    





#endif
