#include <iostream>
#include <vector>
#include <fstream>
#include "small.h"
#include <sstream>
#include <string>
#include <cmath>
#include <cassert>
template<class T>
T Cast(const std::string&s)
{
  std::istringstream is(s);
  T res;
  is >> res;
  return res;
}

double scale;

typedef SmallVector<double, 3> P3D;

P3D rc, trans, ang;

std::vector<P3D> points;
std::vector<int> vertices;

void Read(const std::string &fname)
{
  std::ifstream f(fname.c_str());
  std::string line;
  getline(f, line);
  int point_count;
  f >> point_count;
  for (int i = 0; i < point_count; ++i)
  {
    P3D p;
    f >> p[0] >> p[1] >>
      p[2];
    points.push_back(p);
  }
  int tet_count;
  f >> tet_count;
  for (int i = 0; i < tet_count; ++i)
  {
    for (int i = 0; i < 3; ++i)
    {
      int num;
      f >> num;
      --num;
      vertices.push_back(num);
    }
  }
}

void Write(const std::string &fname)
{
  std::ofstream f(fname.c_str());
  f << "OFF\n";
  f.precision(16);
  f << points.size() << " " <<(vertices.size()/3) <<" 0\n\n";
  for (size_t i = 0; i < points.size(); ++i)
    f << points[i] << "\n";
  for (size_t i = 0; i < vertices.size(); i+=3)
    f << "3 " <<vertices[i] << " " << vertices[i+1] << " " << vertices[i+2] << "\n";
}
    
void ScaleTranslate()
{
  for (size_t i = 0; i < points.size(); ++i)
  {
    points[i]-=rc;
    points[i]*=scale;
    points[i]+=rc;
    points[i]+=trans;
  }  
}
SmallMatrix<double, 3, 3> rx, ry, rz, mat;

template<class T>
void MulMat(T &res, const T &a, const T &b)
{
  res = 0;
  for (int i = 0; i < 3; ++i)
    for (int j = 0; j < 3; ++j)
      for (int k = 0; k < 3; ++k)
	res[i][j] += a[i][k]*b[k][j];
}

void GenerateMatrices()
{
  using std::sin;
  using std::cos;
  double PI = std::atan(1.)*4;
  double ax = ang[0]/180*PI;
  double ay = ang[1]/180*PI;
  double az = ang[2]/108*PI;
  rx[0][0] = 1;       rx[0][1] = 0;        rx[0][2] = 0;
  rx[1][0] = 0;       rx[1][1] = cos(ax);  rx[1][2] = -sin(ax);
  rx[2][0] = 0;       rx[2][1] = sin(ax);  rx[2][2] = cos(ax);

  ry[0][0] = cos(ay); ry[0][1] = 0;        ry[0][2] = sin(ay);
  ry[1][0] = 0;       ry[1][1] = 1;        ry[1][2] = 0;
  ry[2][0] = -sin(ay);ry[2][1] = 0;        ry[2][2] = cos(ay);

  rz[0][0] = cos(az); rz[0][1] = -sin(az); rz[0][2] = 0;
  rz[1][0] = sin(az); rz[1][1] = cos(az);  rz[1][2] = 0;
  rz[2][0] = 0;       rz[2][1] = 0;        rz[2][2] = 1;
  MulMat(mat, rx, ry);
  MulMat(rx, mat, rz);
}

void Rotate()
{
  GenerateMatrices();
  for (size_t i = 0; i < points.size(); ++i)
  {
    P3D p = points[i];
    p -= rc;
    P3D r;

    {
      r = 0;
      for (int i = 0; i < 3; ++i) // mul x
	for (int j = 0; j <3; ++j)
	  r[i] += rx[i][j]*p[j];
      // p = r;
      // r = 0;
      // for (int i = 0; i < 3; ++i) // mul y
      // 	for (int j = 0; j <3; ++j)
      // 	  r[i] += ry[i][j]*p[j];
      // p = r;
      // r = 0;
      // for (int i = 0; i < 3; ++i) // mul z
      // 	for (int j = 0; j <3; ++j)
      // 	  r[i] += rz[i][j]*p[j];
      r += rc;
    }
    points[i]=r;
  }  
}  
int main(int argc, char ** argv)
{
  // ifname ofname sclae rc1 rc2 rc3  ang1 ang2 ang3 trans1 trans2 trans3
  if (argc<2)
    return 1;
  int c = 1;
  std::string ifname = argv[c++];
  std::string ofname = ifname + ".facet";
  if (argc>2)
    ofname = argv[c++];
  std::cout << ifname << " " << ofname  << std::endl;
  std::string off;
  std::ifstream iff(ifname.c_str());
  std::ofstream f(ofname.c_str());
  f.precision(16);

  using std::ios;
  iff.exceptions(ios::failbit | ios::badbit |ios::eofbit);
  iff >> off;
  assert(off=="OFF");
  size_t point_num, surf_num, dummy;

  iff >> point_num >> surf_num >> dummy;
  f << point_num << " " << surf_num << "\n";
  for (size_t i = 0; i < point_num; ++i)
  {
    double x, y, z;
    iff >> x >> y >>z;
    f << (i+1) << " " << x << " " << y << " " << z << "\n";
  }
  for (size_t i = 0; i < surf_num; ++i)
  {
    size_t a, b, c, dummy;
    iff >> dummy >>  a >>  b >>  c;
    ++a;++b;++c;
    assert(dummy==3);
    f << (i+1) << " " << a << " " << b << " " << c << "\n";    
  }
  f<< std::flush;
    
}
