#include <cstdio>
#include <iostream>
#include <vector>
#include <string>
#include <sstream>
#include <png.h>

#include "mirorer.h"


template<class PixelType>
class basic_picture
{
  std::vector<PixelType> storage;
public:
  int height, width;
  
  void Set(int width_, int height_, png_bytep *ptr )
  {
    width = width_;
    height = height_;
    storage.resize(width*height);
    for(int j = 0; j < height; j++)
      for(int i = 0; i < width; i++)
	storage[j*width + i] = ptr[j][i];
  }
  PixelType operator()(int i, int j)
  {
    return storage[j*width + i];
  }

  int Height(){return height;}
  int Width(){return width;}
};

typedef basic_picture<unsigned char> Picture;

int read_chunk_callback(png_struct* ptr,
         png_unknown_chunkp chunk)
    {
       /* The unknown chunk structure contains your
          chunk data: */
           png_byte name[5];
           png_byte *data;
           png_size_t size;
       /* Note that libpng has already taken care of
          the CRC handling */

       /* put your code here.  Return one of the
          following: */

	   //       return (-n); /* chunk had an error */
       return (0); /* did not recognize */
       //       return (n); /* success */
    }

void read_row_callback(png_struct* ptr, png_uint_32 row,
		       int pass)
{
      /* put your code here */
}



Picture * ReadFile(std::string fname)
{
  using namespace std;

  const int number = 8;
  char header[number];
  FILE *fp = fopen(fname.c_str(), "rb");
  if (!fp)
    {
      throw "No such file";
    }
  try
    {
      if(fread(header, 1, number, fp)!=number)
	throw "File too short.";
      
      bool is_png = (bool) !png_sig_cmp((png_byte*)header, 0, number);
      if (!is_png)
	{
	  throw "Not a png file.";
	}
      
      png_structp png_ptr = png_create_read_struct(
						   PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
      if (!png_ptr)
	throw "Cannot init png library.";
      
      png_infop info_ptr = png_create_info_struct(png_ptr);
      if (!info_ptr)
	{
	  png_destroy_read_struct(&png_ptr,
				  (png_infopp)NULL, (png_infopp)NULL);
	  throw "Cannot create struct.";
	}
      
    
      png_infop end_info = png_create_info_struct(png_ptr);
      if (!end_info)
	{
	  png_destroy_read_struct(&png_ptr, &info_ptr,
				  (png_infopp)NULL);
	  throw "Cannot create struct.";	  
	}


    if (setjmp(png_jmpbuf(png_ptr)))
    {
        png_destroy_read_struct(&png_ptr, &info_ptr,
           &end_info);
	throw "Some kind of error... ";
    }

    png_init_io(png_ptr, fp);

    png_set_sig_bytes(png_ptr, number);

    png_set_read_user_chunk_fn(png_ptr, NULL,
        read_chunk_callback);

    png_set_read_status_fn(png_ptr, read_row_callback);


    png_read_png(png_ptr, info_ptr,  PNG_TRANSFORM_STRIP_16 , NULL);

    png_uint_32 width, height;


    int bit_depth, color_type, interlace_type, compression_type, filter_method;

    png_get_IHDR(png_ptr, info_ptr, &width, &height,
		 &bit_depth, &color_type, &interlace_type,
		 &compression_type, &filter_method);


    //    png_read_end(png_ptr, end_info);

    //    cout << width << "x" << height << "x" << bit_depth << endl;
    png_bytep *row_pointers;
    row_pointers = png_get_rows(png_ptr, info_ptr);
    Picture *pic = new Picture();
    pic->Set(width, height, row_pointers);

    png_destroy_read_struct(&png_ptr, &info_ptr,
			    &end_info);




    return pic;

    }
  catch (...)
    {
      fclose(fp);
      throw;
    }

}


int main(int argc, char**argv)
{
  if(argc != 2)
    exit (1);
  std::istringstream is(argv[1]);
  int sz;
  is >> sz;
  int N1 = sz;
  int N2 = N1;
  int N3 = N1;
  std::cout << sz << std::endl;
  vector3d<char> image;
  image.SetSize(N1, N2, N3);
  
  for(int k = 0; k < N3; k++)
    {
      std::cout << k << "   \r" <<std::flush;
      std::string fname = "/media/cdrom/data/ESA29-99-L3-Byte.0";
      std::stringstream a;
      a << fname << (k+413) << ".png";
      fname = a.str();

	
      try{
	Picture *p  = ReadFile(fname.c_str());
	
	Picture &pic = *p;
	int hei = pic.height;
	int wid = pic.width;

	for (int j = 0; j < N2; j++)
	  {
	    for(int i = 0; i < N1; i++)
	      {
		unsigned char ch = pic(i+wid/2-N1/2,j+hei/2-N2/2);
		image(i, j, k) = ch;
	      }
	  }
	delete p;
      }catch(const char *p)
	{
	  std::cout << p << std::endl;
	}
      std::ostringstream os;
      os << sz << ".bone";
      image.Write(os.str());
    }
  return 0;
}
