#include <fstream>
#include <iostream>
#include <set>
#include <sstream>
#include <cmath>
#include <map>
#include <string>
#include <functional>
#include <vector>
#include <stdexcept>
#include <tr1/functional>
#include "small.h"
#include "smallops.h"

void Split(const std::string &what, std::string &l, std::string &r, const std::string &delimiter)
{
  std::string::size_type p;
  p = what.rfind(delimiter);
  if (p==std::string::npos)
  {
    r = "";
    l = what;
  }else
  {
    l = what.substr(0, p);
    r = what.substr(p+delimiter.length());
  }
}



void Trim(std::string &s, const std::string &trims = "\t\n ")
{
  size_t startpos = s.find_first_not_of(trims.c_str());
  size_t endpos = s.find_last_not_of(trims.c_str());
  if(( std::string::npos == startpos ) || (std::string::npos == endpos))
  {
    s = "";
  }
  else
    s = s.substr(startpos, endpos-startpos +1);
}


template<class Double>
inline int Order(const Double &a, const Double &b, const Double prec = 1e-6)
{
  if (fabs(a-b) < prec)
    return 0;
  if (a<b)
    return 1;
  return -1;
}

template<class T>
T Sqr(T a)
{
  return a*a;
}

template<class Double>
class Vertex
{
public:
  //  Double coord[3];
  SmallVector<Double, 3> coord;
  template<class V>
  Vertex(Vertex<V> const &v)
  {
    for (int i = 0; i < 3; ++i)
      coord[i] = v.coord[i];
  }
  Vertex(){}
};


template<class Double>
struct PointCompareLess
{
  Double prec;
  PointCompareLess(Double p):prec(p){}
  template<class T>
  bool operator()(T &a, T &b)const
  {
    for(int i = 0; i < 3; ++i)
    {
      int r = Order(a[i], b[i], prec);
      if (r==1)
	return true;
      if (r==-1)
	return false;
    }
    return false;
  }
  bool operator()(Vertex<Double> const &a, Vertex<Double> const &b)const
  {
    return (*this)(a.coord, b.coord);
  }
};




template<class Double>
class VertexMap
{
public:
  typedef std::map<Vertex<Double>, double, PointCompareLess<Double> > map_type;
  typedef  typename map_type::iterator iterator_type;
private:
  map_type map;
public:
  VertexMap(const Double prec=1e-6):map(PointCompareLess<Double>(prec)){}
  size_t AddVert(Vertex<Double> const &a)
  {
    iterator_type it = map.find(a);
    if (it == map.end())
    {
      size_t sz = map.size();
      map.insert(std::make_pair(a,sz));
      return sz;
    }
    else
    {
      return it->second;
    }
  }
  void Clear()
  {
    map.clear();
  }
    
};

template<class T>
void MakeMin(T &a, const T&b)
{
  if (b<a)
    a = b;
}

template<class Double>
struct Surface{
  std::vector<Vertex<Double> > points;
  std::vector<size_t>  triangles;
  void Clear()
  {
    points.clear();
    triangles.clear();
  }
  SmallVector<Double, 3> GetNormal(size_t i)const
  {
    const Vertex<double> &p1 = points[triangles[i]];
    const Vertex<double> &p2 = points[triangles[i+1]];
    const Vertex<double> &p3 = points[triangles[i+2]];
    SmallVector<Double, 3> v1, v2;
    v1 = p2.coord - p1.coord;
    v2 = p3.coord - p1.coord;
    v1 = VecCross(v1, v2);
    Double norm = sqrt(VecDot(v1, v1));
    v1/=norm;
    return v1;
  }
  void MakePositiv()
  {
    Double min[3];
    for (size_t i = 0; i < points.size(); ++i)
    {
      for (int j = 0; j < 3; ++j)
	MakeMin(min[j], points[i].coord[j]);
    }
    for (size_t i = 0; i < points.size(); ++i)
    {
      for (int j = 0; j < 3; ++j)
	points[i].coord[j] -= min[j];
    }
  }
};



template<class Double>
struct SurfaceBuilder//:std::unary_function<void, Vertex<Double> const *>
{
  Surface<Double> &surface;
  VertexMap<Double> map;
  SurfaceBuilder(Surface<Double> &a, Double prec=1e-6):surface(a), map(prec){}
  //  template<class T>
  void operator()(Vertex<Double> const a[3])
  {
    size_t r;
    for(int i = 0; i < 3; ++i)
    {
      Vertex<Double> v(a[i]);
      r = map.AddVert(v);
      surface.triangles.push_back(r);
      if (surface.points.size()<=r) // new point
	surface.points.push_back(v);
    }	
  }
};


  


bool IsBinary(std::istream &f)
{
  char buf[81];
  buf[80] = '0';
  f.seekg(0);
  f.read(buf, 80);
  std::string s(buf, f.gcount());
  std::string r;
  std::istringstream is(s);
  is >>r;
  f.seekg(0);
  if (r == "solid")
    return false;
  else return true;
}


template<class Builder>
void ReadBinaryStl(std::istream &f, Builder &build)
{
  float buf[12];
  unsigned int size;
  f.seekg(80);
  f.read(reinterpret_cast<char*>(&size), 4);
  std::cout << size << " Triangles" << std::endl;
  for (unsigned int i = size; i--; )
  {
    char buf1[2];
    f.read(reinterpret_cast<char*>(buf), 12*4);
    f.read(reinterpret_cast<char*>(buf1), 2); // skip the integer
    Vertex<float> v[3];
    for (int i = 0; i < 3; ++i)
    {
      for (int j = 0; j < 3; ++j)
      {
	v[i].coord[j] = buf[i*3+j +3];
      }
    }
    build(v);
    
  }
}

template<class Double>
void ReadVertex(std::istream &f, Vertex<Double> &v)
{
  std::string s;
  std::getline(f, s);
  std::istringstream is(s);
  std::string t;
  is >> t;
  for (int i = 0; i < 3; ++i)
    is >> v.coord[i];
}

template<class Builder>
void ReadAsciiStl(std::istream &f, Builder &build)
{
  f.exceptions(std::ios::failbit | std::ios::badbit |std::ios::eofbit);
  //  f.seekg(0);

  std::string s;
  try{
    std::getline(f, s);
    while(true) {
      std::getline(f, s);// facet normal xxxxx
      std::getline(f, s);// outer loop
      Vertex<float> v[3];
      for (int i = 0; i < 3; ++i)
	ReadVertex(f, v[i]);
      build(v);
      std::getline(f, s);// endloop
      std::getline(f, s);// endfacet
    }
  }catch(...)
  {
  }
}

template<class Double>
void ReadStl(const std::string &fname, Surface<Double> &surf)
{
  surf.Clear();
  std::ifstream f(fname.c_str());
  if (!f)
  {
    throw std::runtime_error("no such file");
  }
  SurfaceBuilder<Double> build(surf, 1e-6);
  if (IsBinary(f))
    ReadBinaryStl(f, build);
  else
    ReadAsciiStl(f, build);    
}

template<class Double>
void WriteOff(const std::string &fname, Surface<Double> const&surf)
{
  std::ofstream f(fname.c_str());
  f << "OFF\n";
  f.precision(sizeof(Double)==4?10:18);
  f.setf(std::ios::scientific);
  f << surf.points.size() << " " <<(surf.triangles.size()/3) <<" 0\n\n";
  for (size_t i = 0; i < surf.points.size(); ++i)
  {
    f << surf.points[i].coord[0] << " ";
    f << surf.points[i].coord[1] << " ";
    f << surf.points[i].coord[2] << "\n";
  }
  for (size_t i = 0; i < surf.triangles.size(); i+=3)
  {
    f << "3 " <<surf.triangles[i] << " " << surf.triangles[i+1] << " " << surf.triangles[i+2] << "\n";
  }
}

template<class Double>
void ReadOff(const std::string &fname, Surface<Double> &surf)
{
  surf.Clear();
  std::ifstream f(fname.c_str());
  std::string line;
  getline(f, line);
  size_t point_count;
  size_t triag_count;
  size_t dummy;
  f >> point_count >> triag_count >>dummy;
  for (int i = 0; i < point_count; ++i)
  {
    Vertex<Double> p;
    f >> p.coord[0] >> p.coord[1] >>
      p.coord[2];
    surf.points.push_back(p);
  }
  for (int i = 0; i < triag_count; ++i)
  {
    size_t num;
    f >> num;
    for (int i = 0; i < 3; ++i)
    {
      f >> num;
      surf.triangles.push_back(num);
    }
  }
}

template <class Double>
void WriteBinaryStl(const std::string &fname, Surface<Double> const&surf)
{
  std::ofstream f(fname.c_str(), std::ios::binary);
  for (int i = 0; i < 40; ++i)
    f << " \n";
  f.seekp(80);
  unsigned int sz = surf.triangles.size()/3;
  f.write(reinterpret_cast<char*>(&sz), 4);
  for (size_t i = 0; i < surf.triangles.size(); i+=3)
  {
    float buf[12];
    SmallVector<Double,3 > normal = surf.GetNormal(i);
    buf[0] = normal[0];
    buf[1] = normal[1];
    buf[2] = normal[2];
    char buf1[2];
    for (int ii = 0; ii < 3; ++ii)
    {
      for (int j = 0; j < 3; ++j)
      {
	buf[ii*3+j +3] = surf.points[surf.triangles[i+ii]].coord[j];
      }
    }

    f.write(reinterpret_cast<char*>(buf), 12*4);
    int zero = 0;
    f.write(reinterpret_cast<char*>(&zero), 2); // skip the integer
  }
}


template <class Double>
void WriteAsciiStl(const std::string &fname, Surface<Double> const&surf)
{
  std::ofstream f(fname.c_str());
  f << "solid  tqlo\n";
  for (size_t i = 0; i < surf.triangles.size(); i+=3)
  {
    SmallVector<Double,3 > normal = surf.GetNormal(i);
    f << "facet normal";
    for (int l = 0; l < 3; ++l)
      f << " " << normal[l];
    f << "\n";
    f << std::flush;
    f << "    outer loop\n";
    f << std::flush;
    for (int ii = 0; ii < 3; ++ii)
    {
      f << "        vertex";
      for (int j = 0; j < 3; ++j)
      {
	f << " " << surf.points[surf.triangles[i+ii]].coord[j];
      }
      f << "\n";
    }
    f << std::flush;
    f << "    endloop\n";
    f << "endfacet\n";
      
    f << std::flush;
      
  }
    
}

int main(int argc, char **argv)
{
  if (argc<2)
  {
    std::cout << "stl_to_off filename [ascii]\n"
	      <<"\t converts filename from .off to .stl and vice versa.\n"
	      <<"\t if any second parameter is specified the written stl file is ascii, otherwise it is binary\n";
      
    return 1;
  }
  Surface<float> surf;
  std::string name, suff;
  Split(argv[1], name, suff, ".");
  std::string offname = name + ".off";
  std::string stlname = name + ".stl";
  bool ascii = argc>2;
  if (suff == "off")
  {
    std::cout << "Reading " << offname << std::endl;
    ReadOff(offname, surf);

    std::cout << "Writing " << (ascii?"ASCII ":"binary ") << stlname << std::endl;
    surf.MakePositiv();
    if (ascii)
      WriteAsciiStl(stlname, surf);
    else
      WriteBinaryStl(stlname, surf);
  }
  if (suff == "stl")
  {
    std::cout << "Reading " <<  stlname << std::endl;
    ReadStl(stlname, surf);
    std::cout << "Writing " <<  offname << std::endl;
    WriteOff(offname, surf);
  }
}
