//#define NDEBUG
# include <vector>
# include <map>
# include <cmath>
# include <iostream>
# include <iomanip>
# include <fstream>
# include <string>
# include <sstream>
# include <stdexcept>
# include <GL/gl.h>
# include <GL/glu.h>
# include <GL/glut.h>
# include <set>
#include "functionlin.h"
#include "functionpol.h"			
#include "functiontricubic.h"			
#include "mcubes.h"
extern "C" {
# include "gl2ps.h"
}
#include <sstream>
#include "mirorer.h"
#include <limits>
#include "mpt.h"
#include <algorithm>

enum menu_items {
  SAVE_EPS,
  SAVE_PNG,
  SAVE_AVI,
  QUIT
};





const int NT=1;

template<class R, class A, class B, class C>
void Convolute3dPar(vector3d<R> &res, vector3d<A> const &a, vector3d<B> const &b, C mul)
{
  res.SetSize(a.sizes[0],
	      a.sizes[1],
	      a.sizes[2]);
  mpt::Thread threads[NT];
  for (int i = 0; i < NT; ++i)
      threads[i] = mpt::StartThread(
				    std::tr1::bind(Convolute3d<R, A, B, C>
						   ,
						   std::tr1::ref(res),
						   std::tr1::ref(a),
						   std::tr1::ref(b),
						   mul,
						   i, NT)
				    );
  for(int i = 0; i < NT; ++i)
    threads[i].Join();
}

template<class Double>
void Normalize(Double *vec)
{
  Double size = vec[0]*vec[0] + vec[1]*vec[1] + vec[2]*vec[2];
  size = sqrt(size);
  for (int i = 0; i < 3; ++i)
    vec[i]/=size;
}


template<class Double>
int Order(const Double &a, const Double &b)
{
  if (fabs(a-b) < 1e-6)
    return 0;
  if (a<b)
    return 1;
  return -1;
  //  return int(a<b)*2-1;
}

template<class Double>
class VertexInfo
{
public:
  Double vertex[3];
  Double normal[3];
  bool operator<(const VertexInfo&b)const
  {
    for(int i = 0; i < 3; ++i)
    {
      int r = Order(vertex[i], b.vertex[i]);
      if (r==1)
	return true;
      if (r==-1)
	return false;
    }
    return false;
  }
      
};
template<class Double>
class VertexInfo1
{
public:
  Double vertex[3];
  bool operator<(const VertexInfo1&b)const
  {
    for(int i = 0; i < 3; ++i)
    {
      int r = Order(vertex[i], b.vertex[i]);
      if (r==1)
	return true;
      if (r==-1)
	return false;
    }
    return false;
  }
      
};

class VertexMap:public mpt::Mutex
{
public:
  typedef std::set<VertexInfo<float> > map_type;
  typedef  map_type::iterator iterator_type;
private:
  map_type map;
public:
  void AddVert(VertexInfo<float> const &a)
  {
    GUARD(this);
    iterator_type it = map.find(a);
    if (it == map.end())
    {
      map.insert(a);
    }
    else
    {
      for(int i = 0; i < 3; ++i)
      	const_cast<float&>((*it).normal[i])
      	  +=a.normal[i];
    }
  }
  void ExecNormal(float const *a)
  {
    VertexInfo<float> vi;
    for(int i = 0; i < 3; ++i)
      vi.vertex[i] = a[i];
    iterator_type it = map.find(vi);
    if (it == map.end())
    {
      std::cerr<<"mamamata si e ebalo..." << std::endl;
      for (int i = 0; i <3; ++i)
	std::cerr << vi.vertex[i] << " ";
      std::cerr << std::endl;
    }
    else
    {
      double vals[4];
      std::copy(it->normal, it->normal+4, vals);
      Normalize(vals);
      ::glNormal3dv(vals);
    }
  }
  void Clear()
  {
    map.clear();
  }
    
};

VertexMap vertex_map;

void ReadNumbersFromName(std::string name, int &N1, int &N2, int &N3)
{
  name = name.substr(0, name.find('.'));
  std::replace(name.begin(), name.end(), 'x', ' ');

  std::istringstream is(name);
  is >> N1 >> N2 >> N3;
  if (  std::count(name.begin(), name.end(), ' ')!=2)
    N2 = N3 = N1;
  
}
void ReadNumbersFromNameAndk(std::string name, int &N1, int &N2, int &N3, int &k, char &type)
{
  char ch;
  //  name = name.substr(0, name.find('.'));
  std::replace(name.begin(), name.end(), 'x', ' ');

  std::istringstream is(name);
  if (  std::count(name.begin(), name.end(), ' ')==2)
  {
    is >> N1 >> N2 >> N3;
    is >> ch >> type >>ch >>k;
    N1 *= k;
    N2 *= k;
    N3 *= k;
  }
  else
  {
    is >> N1;
    is >> ch >> type >>ch >>k;
    N1 *=k;
    N2 = N3 = N1;
  }
  
}

enum Sides {
  LeftSide = 0x1,
  RightSide = 0x2,
  TopSide = 0x4,
  BottomSide = 0x8,
  FrontSide = 0x10,
  BackSide = 0x20
};



std::string filename;
std::string filename1;

double movedistance = 0.0000;
double movesize = 2e-4;//3300000;
//double movesize = 0.0000330;
bool DirUp=true;
int CurMoveCount = 0;
int MaxMoveCount = 255;
bool Crashed = false;
bool UseMc = true;
int colourset = 0;
int maxcolourset = 2;
double nceps = 1e-8;
double v1[3], v2[3], v3[3];
double ox[3] = {1,0,0}, oy[3]={0,1,0}, oz[3] = {0,0,1};
double xob[3] = {1, 0, 0}, xu[3]={0,1,0};
double bonethres = 40;//0.15;
double minfixer = 0;
double maxfixer = 0;
double GrayCo = 0.75;
bool DrawSolid = true;
bool GrayBone = false;
bool Shading = true;
bool DrawIso = false;
bool ShowPlates = false;
bool ShowPalette = false;
bool RemoveColor = false;
bool UseLighting = false;
bool SolidColor = true;
int isobehav = 3;
int WireFrame = 0;
int CurLine=0;
int il = 5;
double trlines = 0.8;
double isocol = 0.8050;
enum Dirs{Ndir, Xdir, Ydir, Zdir, Cutdir};
Dirs CurDir = Ndir;
int Resolution;
double mypol=1;//0.5;
  static  vector3d<unsigned char > vec1;
double gmax, gmin;

double atest = 1.;




class DisplayList
{
  int ln;
public:
  DisplayList():ln(0)
  {
  }
  ~DisplayList()
  {
    if(ln)
      glDeleteLists(ln,1);
  }
  void CompileStart()
  {
    if(ln)
      glDeleteLists(ln,1);
    ln = glGenLists(1);
    glNewList(ln, GL_COMPILE);
  }
  void CompileFinish()
  {
    glEndList();
  }
  operator int ()
  {
    return ln;
  }
};



DisplayList IsoList;
DisplayList BoneList;
DisplayList SolidList;

LinearFunction3d<double, double, double> bone;
TricubicFunction<double, double, double> bone3cubic;

LinearFunction3d<double, double, double> values;
PolyFunction3d<double, double> valuespol;
vector3d<unsigned char> SolidPart;
LinearFunction3d<double, double, double> disps[3];
PolyFunction3d<double, double> dispspol[3];

double (*bonefunc)(double, double, double)=NULL;;
double (*valuefunc)(double, double, double);

double (*dispfunc[3])(double, double, double);

double ConstantFunc0(double, double, double)
{
  return 0.;
}
template<int num>
double DispFunc(double x, double y, double z)
{
  assert(bonefunc);
  return disps[num](x, y, z);//*(1-bonefunc(x,y,z));
}
template<int num>
double DispPolFunc(double x, double y, double z)
{
  return dispspol[num](x, y, z);
}

std::string filenameI[3];

template<int num>
void ReadValues()
{
  SmallVector<double, 3> min, max;
  min = 0;
  max = 1;
  int N1, N2, N3;
  int k;
  char type;
  ReadNumbersFromNameAndk(filenameI[num], N1, N2, N3, k, type);
  if (type == 'f')
  {
    vector3d<double> im;
    im.SetSize(N1, N2, N3);
    try{
      im.Read(filenameI[num], 0);
    }catch (...)
    {
      std::cout << N1 << " " <<  std::endl;
      int sz = 2;
      im.SetSize(sz,sz,sz);
      for(int i = 0; i < sz; ++i)
	for(int j = 0; j<sz; ++j)
	  for(int k=0; k < sz; ++k)
	  {
	    im(i,j,k) = 1.;
	  }
    }
    disps[num].Init(im, min, max, false, true);
    dispfunc[num] = DispFunc<num>;
  }
  else  if (type == 'g')
  {
    dispspol[num].Init(filenameI[num], min, max);
    dispfunc[num] = DispPolFunc<num>;
  }else
    dispfunc[num] = ConstantFunc0;
}


double ConstantFunc(double, double, double)
{
  return 1.;
}

void UpdateIsoFunc(double x, double y, double z, double &va, double &nv, int &vc, int &nvc)
{
  double b = bonefunc(x,y,z);
  double v = valuefunc(x,y,z);
  int f = (b>=bonethres);
  va += f*v;
  nv += (1-f)*v;
  vc += b >= bonethres;
  nvc += b < bonethres;
}
double IsoFunc(double x, double y, double z)
{
  //  return valuefunc(x, y, z);
  double valb = 0;
  double valnb = 0;
  int vc = 0;
  int nvc = 0;
  UpdateIsoFunc(x + nceps, y + nceps, z + nceps, valb, valnb, vc, nvc);
  UpdateIsoFunc(x - nceps, y + nceps, z + nceps, valb, valnb, vc, nvc);
  UpdateIsoFunc(x + nceps, y - nceps, z + nceps, valb, valnb, vc, nvc);
  UpdateIsoFunc(x - nceps, y - nceps, z + nceps, valb, valnb, vc, nvc);
  UpdateIsoFunc(x + nceps, y + nceps, z - nceps, valb, valnb, vc, nvc);
  UpdateIsoFunc(x - nceps, y + nceps, z - nceps, valb, valnb, vc, nvc);
  UpdateIsoFunc(x + nceps, y - nceps, z - nceps, valb, valnb, vc, nvc);
  UpdateIsoFunc(x - nceps, y - nceps, z - nceps, valb, valnb, vc, nvc);
  if (vc == 0)
    return valnb/nvc;
  if (nvc == 0)
    return valb/vc;
  if ((isobehav%3)==0)
    return valnb/nvc;
  else
    return valb/vc;

}
  
double LinearBone(double x, double y, double z)
{
  return bone(x, y, z);
}
double TricubicBone(double x, double y, double z)
{
  return bone3cubic(x, y, z);
}

double LinearFunc(double x, double y, double z)
{
  assert(bonefunc);
  return values(x, y, z);//*(1-bonefunc(x,y,z));
}
double PolynomFunc(double x, double y, double z)
{
  return valuespol(x, y, z);
}

template <class T>
void RemoveExtra1(vector3d<T> *v, int start, int end, int odd)
{
  vector3d<T> &vec = *v;
  int i = start;
  if (((i&1)&&(!odd))||((!(i&1))&&odd))
    ++i;
  for(; i < end; i+=2)
    for(int j = 0; j < vec.sizes[1]; j++)
      for(int k = 0; k < vec.sizes[2]; k++)
      {
	if (i >0)
	{
	  bool lp = vec(i-1,j,k) & RightSide;
	  bool rp = vec(i, j, k) & LeftSide;
	  lp = lp and rp;
	  if (lp)
	  {
	    vec(i-1, j, k) &= ~ RightSide;
	    vec(i, j, k) &= ~LeftSide;
	  }
	}
	if (j >0)
	{
	  bool lp = vec(i,j-1,k) & BackSide;
	  bool rp = vec(i, j, k) & FrontSide;
	  lp = lp and rp;
	  if (lp)
	  {
	    vec(i, j-1, k) &= ~ BackSide;
	    vec(i, j, k) &= ~FrontSide;
	  }
	}
	if (k >0)
	{
	  bool lp = vec(i,j,k-1) & TopSide;
	  bool rp = vec(i, j, k) & BottomSide;
	  lp = lp and rp;
	  if (lp)
	  {
	    vec(i, j, k-1) &= ~TopSide;
	    vec(i, j, k) &= ~BottomSide;
	  }
	    
	}
      }
}

template<class T>
void RemoveExtra(vector3d<T> &vec)
{
  int N = vec.sizes[0];
  mpt::Thread threads[NT];
  for(int odd = 0; odd <=1; ++odd)
  {
    for (int i = 0; i < NT; ++i)
      threads[i] = mpt::StartThread(
				    std::tr1::bind(RemoveExtra1<T>
					      , &vec, i*N/NT, (i+1)*N/NT, odd));
    for(int i = 0; i < NT; ++i)
      threads[i].Join();
  }
}


void GenerateSolid(int start, int end)
{
  
  vector3d<unsigned char> &vec = SolidPart;
  int &sz = Resolution;
  int &N = Resolution;
  switch(CurDir)
  {
  case Cutdir:
    for(int i = start; i < end; ++i)
      for(int j = 0; j<sz; ++j)
	for(int k=0; k < sz; ++k)
	{
	  if (!((i<9*sz/18)&&(j<sz*4/7)&&(k>sz/4)))
	    vec(i,j,k)= 255;
	  else
	    vec(i,j,k) = 0;
	}
    break;
  case Ndir:
    for(int i = start; i < end; ++i)
      for(int j = 0; j< N; ++j)
	for(int k=0; k < N; ++k)
	{
	  double h = 0.5/N;
	  double on = 1./N;
	  double hi = i*on+h;
	  double hj = j*on+h;
	  double hk = k*on+h;
	  vec(i,j,k) = (bonefunc(hi, hj, hk)>=bonethres)?255:0;
	}
    break;
  }
  if((CurDir!=Ndir)&&(CurDir!=Cutdir))
  {
    for(int i = start; i < end; ++i)
      for(int j = 0; j < N; ++j)
	for(int k = 0; k < N; ++k)
	  vec(i,j,k)=0;

  }
  if (CurDir!=Cutdir)
  {
    int Line = CurLine;
    for(int i = 0; i < N; ++i)
      for(int j = 0; j < N; ++j)
      {
	switch(CurDir)
	{
	case Ndir:
	  if(ShowPlates) if ((i>=start)&&(i<end))
	  {
	    vec(i,j, 0) = 255;
	    vec(i,j,N-1)=255;
	  }
	  break;
	case Xdir:
	  if ((Line>=start)&&(Line<end))
	    vec(Line, i,j) = 255;
	  break;
	case Ydir:
	  if ((i>=start)&&(i<end))
	    vec(i,Line, j) = 255;
	  break;
	case Zdir:
	  if ((i>=start)&&(i<end))
	    vec(i,j,Line) = 255;
	  break;
	}
      }
  }

}
void GenerateSolid()
{
  
  int &sz = Resolution;
  int &N = Resolution;
  SolidPart.SetSize(sz, sz, sz);
  mpt::Thread threads[NT];
  for (int i = 0; i < NT; ++i)
    threads[i] = mpt::StartThread(
				  std::tr1::bind((void(*)(int, int))&GenerateSolid
					      ,i*N/NT, (i+1)*N/NT));
  for(int i = 0; i < NT; ++i)
    threads[i].Join();

  
  vector3d<unsigned char> &vec = SolidPart;
  RemoveExtra(vec);  
}



void ReadData()
{
  mpt::Thread t1, t2, t3;
  t1 = mpt::StartThread(ReadValues<0>);
  
  SmallVector<double, 3> min, max;
  min = 0;
  max = 1;
  int sz1, sz2, sz3;
  ReadNumbersFromName(filename, sz1, sz2, sz3);
  vector3d<unsigned char > vec;
  vec.SetSize(sz1, sz2, sz3);
  try{
    vec.Read(filename, 0);
    if (Resolution == 0)
      Resolution = sz1;
  }catch(...)
  {
    int sz = 3;
    vec.SetSize(sz,sz,sz);
      for(int i = 0; i < sz; ++i)
	for(int j = 0; j<sz; ++j)
	  for(int k=0; k < sz; ++k)
	  {
	      vec(i,j,k) = 0;
	  }
  }

  vector3d<double > vec1;
  vector3d<double > vec2;
  vector3d<double > conv;
  vector3d<double > der;
  CreateConv3x3x3(der);
  GaussianConvMat(conv, 5, 1.3);
  Convolute3dPar(vec1, vec, conv, 1.0);
  //  Convolute3dPar(vec2, vec1, der, 1.0);
  vec1.RemoveOuter();
  // bone.Init(vec1, min, max, false, true);//extrapolate, interpolate;
  // bonefunc = LinearBone;
  bone3cubic.Init(vec1, min, max, false);//extrapolate
  bonefunc = TricubicBone;
  // bone read

  t2 = mpt::StartThread( ReadValues<1>);
  t3 = mpt::StartThread( ReadValues<2>);
  
  
  int N1, N2, N3;
  int k;
  char type='x';
  ReadNumbersFromNameAndk(filename1, N1, N2, N3, k, type);
  if (type == 'f')
  {
    vector3d<double> im;
    im.SetSize(N1, N2, N3);
    try{
      im.Read(filename1, 0);
      if (Resolution == 0)
	Resolution = N1/k;
      values.Init(im, min, max, false, false );
      valuefunc = LinearFunc;
    }catch (...)
    {
      //      std::cout << N << " " <<  std::endl;
      int sz = 2;
      im.SetSize(sz,sz,sz);
      for(int i = 0; i < sz; ++i)
	for(int j = 0; j<sz; ++j)
	  for(int k=0; k < sz; ++k)
	  {
	    im(i,j,k) = 1.;
	  }
      valuefunc = bonefunc;
    }
  }
  else  if (type == 'g')
  {
    valuespol.Init(filename1, min, max);
    valuefunc = PolynomFunc;
  }else
    valuefunc = bonefunc;//ConstantFunc;
  t1.Join();
  t2.Join();
  t3.Join();
      
}



double Cos(double v1[3], double v2[3], double v3[3])
{
  for(int i = 0; i < 3;++i)
  {
    v1[i] -= v3[i];
    v2[i] -= v3[i];
  }
  double l1 = 0;
  double l2 = 0;
  for (int i = 0; i < 3; ++i)
  {
    l1 += v1[i]*v1[i];
    l2 += v2[i]*v2[i];
  }
  l1 = sqrt(l1);
  l2 = sqrt(l2);
  double cos = 0;
  for (int i = 0; i < 3; ++i)
  {
    cos += v1[i]*v2[i];
  }
  cos /= l1*l2;
  return cos;    
}

template<class T>
struct Quad
{
  T w,x,y,z;
  Quad():w(), x(), y(), z(){}
  Quad(T ax, T ay, T az, T angle):
    w(cos(angle/2)),
    x(ax*sin(angle/2)),
    y(ay*sin(angle/2)),
    z(az*sin(angle/2))
  {}
  void GetMatrix(T (&A)[3][3])
  {
    A[0][0] = w*w + x*x - y*y - z*z; A[0][1] = 2*x*y - 2*w*z;        A[0][2] = 2*x*z + 2*w*y;
    A[1][0] = 2*x*y + 2*w*z;         A[1][1] = w*w - x*x + y*y - z*z;A[1][2] = 2*y*z - 2*w*x;
    A[2][0] = 2*x*z - 2*w*y;         A[2][1] = 2*y*z + 2*w*x;        A[2][2] = w*w - x*x - y*y + z*z;
  }
};
template<class T>
Quad<T> operator*(const Quad<T> &a, const Quad<T>&b)
{
  Quad<T> r;
  r.w = a.w*b.w - a.x*b.x - a.y*b.y - a.z*b.z;
  r.x = a.w*b.x + a.x*b.w + a.y*b.z - a.z*b.y;
  r.y = a.w*b.y + a.y*b.w + a.z*b.x - a.x*b.z;
  r.z = a.w*b.z + a.z*b.w + a.x*b.y - a.y*b.x;
  return r;
}

  

using namespace std;

double MIN_X,MAX_X, MIN_Y, MAX_Y, MIN_Z, MAX_Z;
double vsx, vsy, vsz;
double minx, miny, minz, maxx, maxy, maxz;
double xc, yc, zc;

float angle=atan(1)*2,ratio = 1;
bool usecube = false;
bool swapp=true;
float angle1=0;
float angle2=0;
GLint snowman_display_list;
double dist = 2.5;

double
bmin[2],  bmax[2];
double amin, amax;
std::vector<double> Dmins[3][2];
std::vector<double> Dmaxs[3][2];

void DoMinMax(const double&val, double&min, double &max)
{
  if (val<min)
    min = val;
  if (val>max)
    max = val;
}
double GenMins(int start, int end,
	       double bmin[2], double  bmax[2],
	       std::vector<double> (*Dmins1)[3][2],
	       std::vector<double> (*Dmaxs1)[3][2]
	       )
{
  std::vector<double> (&Dmins)[3][2] = *Dmins1;
  std::vector<double> (&Dmaxs)[3][2] = *Dmaxs1;
  const double MIN = std::numeric_limits<double>::max();
  const double MAX = -std::numeric_limits<double>::max();

  int R = 1;
  int N = Resolution*R;
  for(int i = 0; i < 2; ++i)
  {
    bmin[i] = MIN;
    bmax[i] = MAX;
    for(int j = 0; j < 3; ++j)
    {
      Dmins[j][i].resize(N/R);
      Dmaxs[j][i].resize(N/R);
      for(int k = 0; k < N; ++k)
      {
	Dmins[j][i][k/R] = MIN;
	Dmaxs[j][i][k/R] = MAX;
      }
    }
  }
  double h = 0.5/N;
  double on = 1./N;
  for(int i = start; i < end; ++i)
    for (int j = 0; j < N; ++j)
      for(int k = 0; k < N; ++k)
      {
	  double hi = i*on+h;
	  double hj = j*on+h;
	  double hk = k*on+h;
	  double val = valuefunc(hi, hj, hk);
	  for (int isb = 0; isb<2; isb++)
	  {
	    DoMinMax(val, bmin[isb], bmax[isb]);
	    bool sb = (bonefunc(hi, hj, hk)>=bonethres);
	    if (!sb && isb )
	      continue;
	    DoMinMax(val, Dmins[0][isb][i/R], Dmaxs[0][isb][i/R]);
	    DoMinMax(val, Dmins[1][isb][j/R], Dmaxs[1][isb][j/R]);
	    DoMinMax(val, Dmins[2][isb][k/R], Dmaxs[2][isb][k/R]);
	  }
      }
}
double GenMins()
{
  const double MIN = std::numeric_limits<double>::max();
  const double MAX = -std::numeric_limits<double>::max();
  int R = 1;
  int N = Resolution*R;
  for(int i = 0; i < 2; ++i)
  {
    bmin[i] = MIN;
    bmax[i] = MAX;
    for(int j = 0; j < 3; ++j)
    {
      Dmins[j][i].resize(N/R);
      Dmaxs[j][i].resize(N/R);
      for(int k = 0; k < N; ++k)
      {
	Dmins[j][i][k/R] = MIN;
	Dmaxs[j][i][k/R] = MAX;
      }
    }
  }
  double h = 0.5/N;
  double on = 1./N;
  mpt::Thread threads[NT];
  std::vector<double> Dminss[NT][3][2];
  std::vector<double> Dmaxss[NT][3][2];
  double bmins[NT][2],  bmaxs[NT][2];

  for (int i = 0; i < NT; ++i)
    threads[i] = mpt::StartThread(
				  std::tr1::bind((double (*)(int start, int end,
	       double bmin[2], double  bmax[2],
	       std::vector<double> (*Dmins1)[3][2],
	       std::vector<double> (*Dmaxs1)[3][2]
	       ))&GenMins
					      , i*N/NT, (i+1)*N/NT,
					      bmins[i], bmaxs[i],
					      Dminss+i, Dmaxss+i
					      ));
  for(int i = 0; i < NT; ++i)
    threads[i].Join();
  for(int b = 0; b < 2; ++b)
    for(int k = 0; k < 3; ++k)
    {
      bmin[b] = bmins[0][b];
      bmax[b] = bmaxs[0][b];
      for(int i = 0; i < N; ++i)
      {
	Dmins[k][b] = Dminss[0][k][b];
	Dmaxs[k][b] = Dmaxss[0][k][b];
      }
      for(int j = 1; j < NT; ++j)
      {
	bmin[b] = std::min(bmin[b], bmins[j][b]);
	bmax[b] = std::max(bmax[b], bmaxs[j][b]);
	for(int i = 0; i < N; ++i)
	{
	  Dmins[k][b][i] = std::min(Dmins[k][b][i], Dminss[j][k][b][i]);
	  Dmaxs[k][b][i] = std::max(Dmaxs[k][b][i], Dmaxss[j][k][b][i]);
	}
      }
    }
  std::cout << std::max(Dmaxs[2][0][N-1], Dmaxs[2][1][N-1]) << std::endl;
  std::cout << std::min(Dmins[2][0][N-1], Dmins[2][1][N-1]) << std::endl;
  //  std::cout << valuefunc(0.5, 0.5, 1) << std::endl;
}


template<class T>
void MulSmallMat(T (&res)[3][3],const T (&A)[3][3],const T(&B)[3][3])
{
  for(int i = 0; i < 3; ++i)
    for(int j = 0; j < 3; ++j)
    {
      T sum = 0;
      for(int k = 0; k < 3; ++k)
	sum += A[i][k] * B[k][j];
      res[i][j]=sum;
    }
}
template<class T>
void MulSmallMatVec(T (&res)[3],const T (&A)[3][3],const T(&x)[3])
{
  for(int i = 0; i < 3; ++i)
  {
    double sum = 0;
    for(int k = 0; k < 3; ++k)
      sum += A[i][k] * x[k];
    res[i]=sum;
  }
}
template<class T>
void Trans(const T (&A)[3][3], T(&x)[3])
{
  T t[3];t[0] = x[0];t[1] = x[1]; t[2] = x[2];
  MulSmallMatVec(x, A, t);
}

void orientMe()
{
  // glMatrixMode(GL_MODELVIEW);
  // glLoadIdentity();
  double x = xc;//(MIN_X + MAX_X)/2;
  double y = yc;//(MIN_Y + MAX_Y)/2;
  double z = zc;// (MIN_Z + MAX_Z)/2;

  static bool firsttime = false;

  Quad<double> q1 = Quad<double>(ox[0], ox[1], ox[2], angle);
  Quad<double> q2 = Quad<double>(oy[0], oy[1], oy[2], angle1);
  Quad<double> q3 = Quad<double>(oz[0], oz[1], oz[2], angle2);
  Quad<double> q4;
  q4 = q1*q2*q3;
  double Mat[3][3];
  q4.GetMatrix(Mat);
  Trans(Mat, xob);
  Trans(Mat, xu);
  if (swapp)
  {
    Trans(Mat, ox);
    Trans(Mat, oy);
    Trans(Mat, oz);
  }
  double xo[3];
  xo[0] = dist*xob[0];
  xo[1] = dist*xob[1];
  xo[2] = dist*xob[2];
  for (int i = 0; i < 3; ++i)
    v1[i]=v2[i] = v3[i] = 0.;
  v2[0] = .2;
  v2[1] = .2;
  v1[0] = x;
  v1[1] = y;
  v1[2] = z;
   
  //  gluLookAt (x-xo[0], y-xo[1], z-xo[2],  x, y, z, xu[0], xu[1], xu[2]);
  angle = angle1 = angle2 = 0;

}





void color(double d)
{
  d *= 4;
  double r, g, b;
  if (d < 1.) {
    r = 0.; g = d; b = 1.;
  } else if (d < 2.) {
    r = 0.; g = 1.; b = 2. - d;
  } else if (d < 3.) {
    r = d - 2; g = 1.; b = 0.;
  } else {
    r = 1.; g = 4 - d; b = 0.;
  }
  glColor4d(r, g, b, 0.7);

}
std::vector<float> surfp;

class GlObjectd
{
public:
  float  vals[4];
  enum Type{COLOR4, VERTEX, NORMAL, NOP, MAT_AMBIENT, MAT_DIFFUSE, MAT_SPECULAR, MAT_SHININESS, NORMAL_MAP};
  Type type;
  void Nop()
  {
    type = NOP;
  }
  void glVertex3d(double x, double y, double z)
  {
    type = VERTEX;
    vals[0] = x;vals[1] = y, vals[2] = z;
  }
  void glColor4d(double x, double y, double z, double a)
  {
    type = COLOR4;
    vals[0] = x;vals[1] = y; vals[2] = z;vals[3] = a;
  }
  void glNormal3d(double x, double y, double z)
  {
    type = NORMAL;
    vals[0] = x;vals[1] = y, vals[2] = z;
    Normalize(vals);
  }
  void glNormal3dmap(double x, double y, double z)
  {
    type = NORMAL_MAP;
    vals[0] = x;
    vals[1] = y;
    vals[2] = z;
  }
  void glMateriald(GLint what, double a)
  {
    if (what==GL_SHININESS)
      type = MAT_SHININESS;
    else
      type = NOP;
    vals[0] = a;
  }
  void glMaterialdv(GLint what, double a[4])
  {
    switch ( what)
    {
    case GL_AMBIENT:
      type = MAT_AMBIENT;break;
    case GL_DIFFUSE:
      type = MAT_DIFFUSE;break;
    case GL_SPECULAR:
      type = MAT_SPECULAR;break;
    default:
      type = NOP;break;
    }
    std::copy(a, a+4, vals);
  }
  void Execute()const
  {
    float valsf[4];
    switch(type)
    {
    case MAT_SHININESS:
    case MAT_AMBIENT:
    case MAT_DIFFUSE:
    case MAT_SPECULAR:
    case VERTEX:
      surfp.push_back(vals[0]);
      surfp.push_back(vals[1]);
      surfp.push_back(vals[2]);
      ::glVertex3fv(vals);break;
    case COLOR4:
    case NORMAL:
    case NORMAL_MAP:;
    }
  }
};


void MyTransform(GlObjectd&a)
{
  if (a.type == GlObjectd::VERTEX)
  {
    double x = a.vals[0];
    double y = a.vals[1];
    double z = a.vals[2];
    
    a.vals[0] = x + dispfunc[0](x, y, z)*movedistance;
    a.vals[1] = y + dispfunc[1](x, y, z)*movedistance;
    a.vals[2] = z + dispfunc[2](x, y, z)*movedistance;
  }
}

void mat(double cr, double cg, double cb, double tr, double mul,std::vector<GlObjectd>&vec)
{
  tr = 1;
  GlObjectd v;
  if(1)//!UseLighting)
  {
    v.glColor4d(cr*mul, cg*mul, cb*mul, tr);
    vec.push_back(v);
  }else
    // mul = 1;
    // v.glColor4d(cr*mul, cg*mul, cb*mul, tr);
    // vec.push_back(v);
  {
    v.glMateriald(GL_SHININESS, 4.);
    vec.push_back(v);
    {
      double c = 0.3;
      GLdouble amb[] = {c*cr, c*cg, c*cb, tr};
      v.glMaterialdv( GL_AMBIENT, amb);
      vec.push_back(v);
    }
    {
      double c = 0.6;
      GLdouble amb[] = {c*cr, c*cg, c*cb, tr};
      v.glMaterialdv( GL_DIFFUSE, amb);
      vec.push_back(v);
    }
    {
      double c = 0.3;
      GLdouble amb[] = {c,c,c,tr};
      v.glMaterialdv( GL_SPECULAR, amb);
      vec.push_back(v);
    }

  }
}
struct P3D
{
  double X[3];
};

bool IntersectLine(P3D p1, P3D p2, double v1, double v2, double val, P3D &res)
{
  double f1 = min(v1, v2);
  double f2 = max(v1, v2);
  if ((val <f1) || (val >f2))
    return false;
  if (v2==v1)
    return false;
  double k = (val - v1)/(v2-v1);
  double k1 =  k1 = 1 - k;
  for(int i = 0; i <3; ++i)
    res.X[i] = p1.X[i]*k1 + p2.X[i]*k;
  return true;
}

void DoSide( P3D p[4], double v[4], std::vector<GlObjectd>&vec)
{
  for(int i = 0; i < il; ++i)
  {
    
    double curval = gmin + (gmax-gmin)*(i+0.5)/(il);
     double val1 = (curval - gmin)/(gmax-gmin);
    val1 = val1*2 - 1;
    val1 = (val1<0)?
      -pow(fabs(val1), mypol):
      pow(fabs(val1), mypol);
    val1 = (val1+1)/2;
    curval = val1*(gmax-gmin) +gmin;
    //    std::cout << curval << std::endl;
    
    P3D points[4];
    int maxpoint = 0;
    if(IntersectLine(p[0], p[1], v[0], v[1], curval, points[maxpoint]))
      maxpoint++;
    if(IntersectLine(p[2], p[1], v[2], v[1], curval, points[maxpoint]))
      maxpoint++;
    if(IntersectLine(p[2], p[3], v[2], v[3], curval, points[maxpoint]))
      maxpoint++;
    if(IntersectLine(p[0], p[3], v[0], v[3], curval, points[maxpoint]))
      maxpoint++;
    if(maxpoint==0)
        continue;
    GlObjectd v;
    for (int i = 0; i < maxpoint;++i)
    {
      v.glVertex3d(points[i].X[0], points[i].X[1], points[i].X[2]);
      vec.push_back(v);
    }
  }
}

void DrawHexIso(int I, int J, int K, double bs, int sides, std::vector<GlObjectd> &vec)
{
  if(!sides)
    return;
  double x, y, z, x1, y1, z1;
  x = I*bs;
  x1 = x + bs;
  y = J*bs;
  y1 = y + bs;
  z = K*bs;
  z1 = z + bs;
  P3D p[8] = {{x, y,   z}, 
	      {x1, y,  z}, 
	      {x, y1,  z},
	      {x1, y1, z},
	      {x, y,   z1},
	      {x1, y,  z1},
	      {x, y1,  z1},
	      {x1, y1, z1}};
  
  double v[8];
  for (int i = 0; i < 8; ++i)
    v[i] = IsoFunc(p[i].X[0], p[i].X[1], p[i].X[2]);
  
  if(sides &BottomSide)
  {
    P3D side[4] = {p[0], p[1], p[3], p[2]};
    double vs[4] ={v[0], v[1], v[3], v[2]};
    DoSide(side, vs, vec);
  }
  if(sides &TopSide)
  {
    P3D side[4] = {p[4], p[5], p[7], p[6]};
    double vs[4] ={v[4], v[5], v[7], v[6]};
    DoSide(side, vs, vec);
  }
  if(sides &LeftSide)
  {
    P3D side[4] = {p[0], p[2], p[6], p[4]};
    double vs[4] ={v[0], v[2], v[6], v[4]};
    DoSide(side, vs, vec);
  }
  if(sides &RightSide)
  {
    P3D side[4] = {p[1], p[3], p[7], p[5]};
    double vs[4] ={v[1], v[3], v[7], v[5]};
    DoSide(side, vs, vec);
  }
  if(sides &FrontSide)
  {
    P3D side[4] = {p[0], p[1], p[5], p[4]};
    double vs[4] ={v[0], v[1], v[5], v[4]};
    DoSide(side, vs, vec);
  }
  if(sides &BackSide)
  {
    P3D side[4] = {p[2], p[3], p[7], p[6]};
    double vs[4] ={v[2], v[3], v[7], v[6]};
    DoSide(side, vs, vec);
  }
    
}


template <class T>
inline bool In(const T&a, const T&min, const T &max)
{
  return (a >=min) && (a<max);
}
template<class T, class I>
bool check(T &vec, I i, I j, I k)
{
  ptrdiff_t szs[3];
  bool ret = true;
  if (i>0)
    ret = ret && vec(i-1, j, k);
  else
    ret = false;
  if (j>0)
    ret = ret && vec(i, j-1, k);
  else
    ret = false;
  if (k>0)
    ret = ret && vec(i, j, k-1);
  else
    ret = false;
  if (i<vec.sizes[0]-1)
    ret  = ret && vec(i+1, j, k);
  else
    ret = false;
  if (j<vec.sizes[1]-1)
    ret  = ret && vec(i, j+1, k);
  else
    ret = false;
  if (k<vec.sizes[2]-1)
    ret = ret && vec(i, j, k+1);
  else
    ret = false;
  bool tmp = vec(i,j,k);
  int sz1, sz2, sz3;
  sz1 = vec.sizes[0]/2;
  sz2 = vec.sizes[1]/2;
  sz3 = vec.sizes[2]/2;
  //     if ((i<sz1)&&(j<sz2) && (k <sz3))
  //       return false;
  if ((i==sz1)||(j==sz2) || (k==sz3))
    return vec(i,j,k);
  return tmp^ret;
}

struct color4
{
  double r,g,b,a;
  color4():r(), g(), b(), a(){}
  color4(double r_, double g_, double b_, double a_)
  {
    r = r_;
    g = g_;
    b = b_;
    a = a_;
  }
  void Set(double r_, double g_, double b_, double a_)
  {
    r = r_;
    g = g_;
    b = b_;
    a = a_;
  }
  void operator *=(const color4&a)
  { 
    r *= a.r;
    g *= a.g;
    b *= a.b;
    this->a *= a.a;
    return void(*this);//!!!
  }
};

std::vector<color4> cols, colwhite;

void InitColors()
{
  double ccc = 0.4;
  colwhite.push_back(color4(ccc,ccc,ccc,1));
  colwhite.push_back(color4(ccc,ccc,ccc,1));
  cols.clear();
  double alfa = 1;
  switch(colourset)
  {
  case 0:
    cols.push_back(color4(0,   0, 1, alfa*1));
    cols.push_back(color4(0,   1, 1, alfa*0.1));
    cols.push_back(color4(0,   .9, 0, alfa*0.02));
    cols.push_back(color4(1,   1, 0, alfa*0.1));
    cols.push_back(color4(1,   0, 0, alfa*1));
    isocol = 0.050;
    
    break;
  case 1:
    cols.push_back(color4(0.1,   0.1, 0.1, 0.1));
    cols.push_back(color4(0.1,   0.1, 0.1, 0.3));
    cols.push_back(color4(0.2,   0.2, 0.2, 0.4));
    cols.push_back(color4(0.5,   0.5, 0.25, 0.5));
    cols.push_back(color4(1,   1, 0.5, 0.6));
    cols.push_back(color4(1,   1, 0.5, 1));
    isocol = 0.950;
    break;
  }
}

void GetColor(std::vector<color4>&cols, double val, color4&res)
{
  if (isnan(val))
    val = 0.5;
  if (val <0)
    val = 0;
  if (val >1)
    val = 1;
  int size = cols.size();
  if (size <2)
  {
    if (size==1)
      res = cols[0];
    else
      res = color4(.8,.8,.8,1);
    return;
  }
  val *= (size-1);
  int ival = static_cast<int>(trunc(val));
  val -= ival;
  color4 &l=cols[ival];
  if (ival==(size-1))
  {
    res = l;
    return;
  }
  color4 &r=cols[ival+1];
  res.r = l.r*(1-val) + r.r*val;
  res.g = l.g*(1-val) + r.g*val;
  res.b = l.b*(1-val) + r.b*val;
  res.a = l.a*(1-val) + r.a*val;
}





//worst function ever!
void SetMinMax(int i=6)
{
  if (i==6)
  {
    Dirs olddir = CurDir;
    if(isobehav>=3)
    {
      CurDir = Ndir;
    }      
    SetMinMax(isobehav%3);
    CurDir = olddir;
    
    double size = gmax - gmin;
    amin = gmin += size*minfixer/30;
    amax = gmax += size*maxfixer/30;
    
    if (gmin>=gmax)
      gmax = gmin + nceps;

    
    std::cout << gmin << " " << gmax << std::endl;
    return;
  }
  if (i==2)
  {
    double tmin, tmax;
    SetMinMax(0);
    tmin = amin;
    tmax = amax;
    SetMinMax(1);
    if (tmin<amin)
      amin = gmin = tmin;
    if (tmax>amax)
      amax = gmax = tmax;
  }else
  {
    
    if (CurDir==Ndir)
    {
      amin = bmin[i];
      amax = bmax[i];
    }
    if (CurDir==Cutdir)
    {
      amin = bmin[i];
      amax = bmax[i];
    }
    if (CurDir==Xdir)
    {
      amin = Dmins[0][i][CurLine];
      amax = Dmaxs[0][i][CurLine];
    }
    if (CurDir==Ydir)
    {
      amin = Dmins[1][i][CurLine];
      amax = Dmaxs[1][i][CurLine];
    }
    if (CurDir==Zdir)
    {
      amin = Dmins[2][i][CurLine];
      amax = Dmaxs[2][i][CurLine];
    }
    
  }
  

  gmin = amin;
  gmax = amax;
}

void SetCol(double val, double mul, bool isbone, std::vector<GlObjectd>&vec)
{
  double val1 = (val - amin)/(amax-amin);
  //  std::cout << val << std::endl;
  val1 = 1-val1;
  val1 = val1*2 - 1;
  val1 = (val1<0)?
    -pow(fabs(val1), mypol):
    pow(fabs(val1), mypol);	  
  val1 = (val1+1)/2;
  color4 c;
  //  std::cout << amin << " " << amax <<" " << val << " "  << val1 << endl;
  GetColor(cols, val1, c);
  if(!Shading)
  {
    return;
    c.r = c.g = c.b = c.a = 1;
  }
  if (GrayBone&&isbone)
  {
    c.r = c.g = c.b  = GrayCo;
    c.a = 1;
  }
  mat(c.r, c.g, c.b, c.a,mul, vec);
}
void CalcNormal(SmallVector<double, 3> &n, SmallVector<double, 3> const&p)
{
  double c = bonefunc(p[0], p[1], p[2]);
  double h = -0.01;
  double sign = -1;
  n[0] = sign*(bonefunc(p[0]-h, p[1], p[2]) - bonefunc(p[0]+h, p[1], p[2]));
  n[1] = sign*(bonefunc(p[0], p[1]-h, p[2]) - bonefunc(p[0], p[1]+h, p[2]));
  n[2] = sign*(bonefunc(p[0], p[1], p[2]-h) - bonefunc(p[0], p[1], p[2]+h));
  
//   std::cout << n[0] << " " << n[1] << " "<< n[2] << std::endl;
//   std::cout << c << " " <<n[0] << " " << n[1] << " "<< n[2] << std::endl;
}


void DrawSide(SmallVector<double, 3> &p1,
	      SmallVector<double, 3> &p2,
	      SmallVector<double, 3> &p3,
	      SmallVector<double, 3> &p4,
	      bool isbone,
	      double mul,
	      std::vector<GlObjectd>&vec
	      )
{
  SmallVector<double, 3> *p[6] = {&p4, &p1, &p2, &p3, &p4, &p1};
  for(int i = 1; i <=4;++i)
  {
    if (!SolidColor)
    {
      double val = valuefunc((*p[i])[0], (*p[i])[1], (*p[i])[2]);
      SetCol(val, mul, isbone, vec);
    }
    GlObjectd v;
    if (UseLighting)
    {
      SmallVector<double, 3> n, v1, v2;
      v2 = *p[i+1];
      v1 = *p[i-1];
      v2 -= *p[i];
      v1 -= *p[i];
      n[0] = v1[1]*v2[2] - v1[2]*v2[1];
      n[1] = v1[2]*v2[0] - v1[0]*v2[2];
      n[2] = v1[0]*v2[1] - v1[1]*v2[0];
      VertexInfo<float> vi;
      for (int j = 0; j <3; ++j)
      {
	vi.vertex[j] = (*p[i])[j];
	vi.normal[j] = n[j];
      }
      vertex_map.AddVert(vi);
      //      v.glNormal3dmap((*p[i])[0], (*p[i])[1], (*p[i])[2]);
      v.glNormal3d(n[0], n[1], n[2]);
      vec.push_back(v);
//       SmallVector<double, 3> n;
//       CalcNormal(n, *p[i]);
//       v.glNormal3d(n[0], n[1], n[2]);
//       vec.push_back(v);
    }
    v.glVertex3d((*p[i])[0], (*p[i])[1], (*p[i])[2]);
    vec.push_back(v);    
  }
  
}

double Sqr(double a)
{
  return a*a;
}

double Length(SmallVector<double, 3> *p1, SmallVector<double, 3> *p2)
{
  SmallVector<double, 3>  &P1 = *p1;
  SmallVector<double, 3>  &P2 = *p2;

  return sqrt(Sqr(P1[0]-P2[0])+Sqr(P1[1]-P2[1])+Sqr(P1[2]-P2[2]));
}

double Area(SmallVector<double, 3> **p)
{
  double l1 = Length(p[0], p[1]);
  double l2 = Length(p[1], p[2]);
  double l3 = Length(p[2], p[0]);
  double s = (l1 + l2 + l3)/2;
  double S = s*(s-l1)*(s-l2)*(s-l3)/2.;
  return sqrt(S);
}
    
  

void DrawSide3(SmallVector<double, 3> &p1,
	      SmallVector<double, 3> &p2,
	      SmallVector<double, 3> &p3,
	      bool isbone,
	      double mul,
	      std::vector<GlObjectd>&vec
	      )
{
  SmallVector<double, 3> *p[5] = {&p1, &p3, &p2, &p1, &p3};
  if (SolidColor)
  {
    double v[3] = {0,0,0};
    for (int i = 1; i <=3; ++i)
      for(int j = 0; j < 3; ++j)
	v[j] += (*p[i])[j];
    for(int j = 0; j < 3; ++j)
      v[j]/=3.;
    
    double val = valuefunc(v[0], v[1], v[2]);
    SetCol(val, mul, isbone, vec);
  }
  double Ar = Area(p);
  for(int i = 1; i <=3;++i)
  {
    if (!SolidColor)
    {
      double val = valuefunc((*p[i])[0], (*p[i])[1], (*p[i])[2]);
      SetCol(val, mul, isbone, vec);
    }
    GlObjectd v;
    {
      SmallVector<double, 3> n, v1, v2;
      v2 = *p[i+1];
      v1 = *p[i-1];
      v2 -= *p[i];
      v1 -= *p[i];
      n[0] = v1[1]*v2[2] - v1[2]*v2[1];
      n[1] = v1[2]*v2[0] - v1[0]*v2[2];
      n[2] = v1[0]*v2[1] - v1[1]*v2[0];
      double Sine = sqrt(Sqr(n[0])+Sqr(n[1]) +Sqr(n[2]));
      Sine /=Length(p[i], p[i-1])*Length(p[i], p[i+1]);
      if (UseLighting)
      {
	VertexInfo<float> vi;
	for (int j = 0; j <3; ++j)
	{
	  vi.vertex[j] = (*p[i])[j];
	  vi.normal[j] = n[j]*Sine;
	}
	vertex_map.AddVert(vi);
	v.glNormal3dmap((*p[i])[0], (*p[i])[1], (*p[i])[2]);
	//      v.glNormal3d(n[0], n[1], n[2]);
	vec.push_back(v);
      }else
      {
	double val = valuefunc((*p[i])[0], (*p[i])[1], (*p[i])[2]);
	SetCol(val, mul*(fabs(n[0]/4)+fabs(n[1]/2)+fabs(n[2])/2)/sqrt(n[0]*n[0]+n[1]*n[1]+n[2]*n[2]), isbone, vec);
      }
//       SmallVector<double, 3> n;
      
//       double val = valuefunc((*p[i])[0], (*p[i])[1], (*p[i])[2]);
//       //      std::cout << val << std::endl;
//       CalcNormal(n, *(p[i]));

//       VertexInfo<float> vi;
//       for (int j = 0; j <3; ++j)
//       {
//  	vi.vertex[j] = (*p[i])[j];
//  	vi.normal[j] = n[j];
//       }
//       vertex_map.AddVert(vi);
//       //      v.glNormal3d(n[0], n[1], n[2]);
//       v.glNormal3dmap((*p[i])[0], (*p[i])[1], (*p[i])[2]);
      
//       vec.push_back(v);
    }
    v.glVertex3d((*p[i])[0], (*p[i])[1], (*p[i])[2]);
    vec.push_back(v);    
  }
  
}

template<class P>
void movepoint(P &p, double o1, double o2, double o3)
{
  return;
  double os[3] = {o1, o2, o3};
  o1/=2;
  o2/=2;
  o3/=2;
  double vecs[14][3] =
    {
      {1, 1, 1},
      {1, 1, -1},
      {1, -1, 1},
      {1, -1, -1},
      {-1, 1, 1},
      {-1, 1, -1},
      {-1, -1, 1},
      {-1, -1, -1},
      {1, 0, 0},
      {-1, 0, 0},
      {0,1, 0},
      {0,-1, 0},
      {0,0,1},
      {0,0,-1}
    };
  int  bes[14];
  for(int i = 0; i < 8; ++i)
  {
    bes[i] = false;
    for(int k = 0; k < 3; ++k)
    {
      double val = p[k] + vecs[i][k]*os[k];
      if ((val <0)||(val>1))
	goto cont;
    }
    bes[i] = bonefunc(
		       p[0]+vecs[i][0]*o1,
		       p[1]+vecs[i][1]*o2,
		       p[2]+vecs[i][2]*o3)
         >=bonethres;
  cont:;
  }
  bes[8]  = bes[0] + bes[1] + bes[2] + bes[3];
  bes[9]  = bes[4] + bes[5] + bes[6] + bes[7];
  bes[10] = bes[0] + bes[1] + bes[4] + bes[5];
  bes[11] = bes[6] + bes[7] + bes[2] + bes[3];
  bes[12] = bes[0] + bes[2] + bes[4] + bes[6];
  bes[13] = bes[1] + bes[3] + bes[5] + bes[7];

  
  double k,l;
  k = 0.5;
  k/=4;
  for(int i = 0; i <3;++i)
  {
    if (bes[i*2+8]!=bes[i*2+9])
    {
      double s = 0;
      if (bes[i*2+8]==4)
	s=-1;
      if (bes[i*2+8]==0)
	s=1;
      if (bes[i*2+9]==4)
	s=1;
      if (bes[i*2+9]==0)
	s=-1;
//       if (bes[i*2+9]==3 || bes[i*2+8]==3)
// 	return;
      s*=-1;
      p[i]+=os[i]*k*s;
    }
  }
  
//   for (int i = 0; i < 8; ++i)
//     for (int c = 0; c < 3; ++c)
//       p[c] += k*vecs[i][c]*os[c]*(double)not bes[i];
//   for (int i = 8; i < 14; ++i)
//     for (int c = 0; c < 3; ++c)
//       p[c] += l*vecs[i][c]*os[c]*(double)bes[i];
  for (int c = 0; c<3; ++c)
  {
    if (p[c] <0) p[c]=0;
    if (p[c]>1) p[c] = 1-1e-13;
  }
}

void DrawHexMC(double x1, double y1, double z1,
	     double o1, double o2, double o3, unsigned char sides,
	     std::vector<GlObjectd>&vec
	     )
{
  double o = 0;
  double v[8];
  P3D p[8] = {
    {x1 + o , y1 + o , z1 + o3}, 
    {x1 + o1, y1 + o , z1 + o3}, 
    {x1 + o1, y1 + o , z1 + o }, 
    {x1 + o , y1 + o , z1 + o }, 
    {x1 + o , y1 + o2, z1 + o3}, 
    {x1 + o1, y1 + o2, z1 + o3}, 
    {x1 + o1, y1 + o2, z1 + o }, 
    {x1 + o , y1 + o2, z1 + o }
  };
  int edges[12][2] =    {
      {0,1},
      {1,2},
      {2,3},
      {3,0},
      {4,5},
      {5,6},
      {6,7},
      {7,4},
      {0,4},
      {1,5},
      {2,6},
      {3,7}
  };

  P3D poe[12];
  P3D npoe[12];

  double bt = bonethres;
  int ind1 = 0;
  for (int i = 0; i < 8; ++i)
  {
    double val = bonefunc(p[i].X[0], p[i].X[1], p[i].X[2]);
//     double x = p[i].X[0]-0.5;
//     double y = p[i].X[1]-0.5;
//     double z = p[i].X[2]-0.5;
//     val = x*x + y*y+z*z;
    //    val *= p[i].X[0]*(1-p[i].X[0])*p[i].X[1]*(1-p[i].X[1])*p[i].X[2]*(1-p[i].X[2]);
    v[i] = val;
    int cond = val >bt;
    ind1 += (cond << i);
  }

  int ind2 = ET[ind1];
  for (int i = 0; i < 12; ++i)
    if (ind2 & (1 << i))
    {
      P3D &p1 = p[edges[i][0]];
      P3D &p2 = p[edges[i][1]];
      double &v1 = v[edges[i][0]];
      double &v2 = v[edges[i][1]];
      double x = (bt - v2)/(v1-v2);
      for (int j = 0; j <3; ++j)             //interpolation
	poe[i].X[j] = x * p1.X[j] + (1-x)*p2.X[j];
    }
  for (int i = 0; i < 16; i+=3)
  {
    if (MC[ind1][i] <0)
      break;
    SmallVector<double, 3> p1, p2, p3;
    for (int j = 0; j < 3; ++j)
    {
      p1[j] = poe[MC[ind1][i]].X[j];
      p2[j] = poe[MC[ind1][i+1]].X[j];
      p3[j] = poe[MC[ind1][i+2]].X[j];
    }
    DrawSide3(p1, p2, p3,
	      true,
	      1.0,
	      vec
	      );
    
    
  }
  
}


void DrawHex(double x1, double x2, double x3,
	     double o1, double o2, double o3, unsigned char sides,
	     std::vector<GlObjectd>&vec
	     )
{

  if(!sides)
    return; // deal only with real guys;
  GlObjectd v;

  SmallVector<double, 3> p[8];

  for(int i = 0; i < 8; ++i)
  {
    p[i][0] = x1;
    p[i][1] = x2;
    p[i][2] = x3;
  }
  p[1][0] += o1;
  p[3][0] += o1;
  p[5][0] += o1;
  p[7][0] += o1;

  p[2][1] += o2;
  p[3][1] += o2;
  p[6][1] += o2;
  p[7][1] += o2;

  p[4][2] += o3;
  p[5][2] += o3;
  p[6][2] += o3;
  p[7][2] += o3;
  for(int i =0; i < 8; ++i)
    movepoint(p[i], o1, o2, o3);

  double m0, m1, m2;
  m0 = 1.;
  if (Shading||not UseLighting)
  {
    m1 = 0.91;
    m2 = 0.82;
  }
  else
    m1 = m2 = 1.;
  double mul = 0.5+(x1+x2+x3)*0.5/3.;
  if(Shading)
  {
    m1*=mul;
    m0*=mul;
    m2*=mul;
  }
  //  m1 = m0 = m2 = 1.;
  double val;
  bool isbone;
  val = valuefunc(x1 + 0.5*o1, x2 + 0.5*o2, x3 + 0.5*o3);
  //  std::cout << "val" << val << " " << bonefunc(x1 + 0.5*o1, x2 + 0.5*o2, x3 + 0.5*o3) << std::endl;
  isbone = bonefunc(x1 + 0.5*o1, x2 + 0.5*o2, x3 + 0.5*o3)>=bonethres;

  if (sides & (BottomSide|TopSide))
  {
    if(SolidColor)
      SetCol(val, m0, isbone,vec);

    if(sides & BottomSide)
      DrawSide(p[0], p[2], p[3], p[1], isbone, m0, vec);
    if (sides &TopSide)
      DrawSide(p[5], p[7], p[6], p[4], isbone, m0, vec);
  }
  if (sides & (FrontSide|BackSide))
  {
    if(SolidColor)
      SetCol(val, m1, isbone,vec);

    if(sides & FrontSide)
      DrawSide(p[0], p[1], p[5], p[4], isbone, m1, vec);
    if (sides &BackSide)
      DrawSide(p[3], p[2], p[6], p[7], isbone, m1, vec);
  }
  if (sides & (LeftSide|RightSide))
  {
    if(SolidColor)
      SetCol(val, m2, isbone,vec);

    if(sides & LeftSide)
      DrawSide(p[0], p[4], p[6], p[2], isbone, m2, vec);
    if (sides & RightSide)
      DrawSide(p[3], p[7], p[5], p[1], isbone, m2, vec);
  }
}

void GenerateSolidList1(int start, int end, std::vector<GlObjectd>*vec)
{
  const int N = Resolution;
  double bs = 1./N;
  for(int i = start; i < end; ++ i)
  {
    for(int j = 0; j < N; ++j)
    {
      for (int k = 0; k < N; ++k)
      {
	  DrawHex( i*bs, j*bs, k*bs, bs, bs, bs,
		   SolidPart(i,j,k), *vec);
      }
    }
  }
  std::for_each(vec->begin(), vec->end(), MyTransform);
}
void GenerateSolidList()
{
  const int N = Resolution;
  mpt::Thread threads[NT];
  std::vector<GlObjectd> vecs[NT];
  for (int i = 0; i < NT; ++i)
    threads[i] = mpt::StartThread(
				  std::tr1::bind(GenerateSolidList1
					      , i*N/NT, (i+1)*N/NT, vecs+i));
  for(int i = 0; i < NT; ++i)
    threads[i].Join();
  SolidList.CompileStart();
  glBegin(GL_QUADS);
   for (int i = 0; i < NT; ++i)
     std::for_each(vecs[i].begin(), vecs[i].end(), std::mem_fun_ref(&GlObjectd::Execute));
  glEnd();
  SolidList.CompileFinish();
}
void GenerateMCSolidList1(int start, int end, std::vector<GlObjectd>*vec)
{
  
  const int N = Resolution;
  double bs = 1./N;
  std:: cout << bs << " yeeeeeeeeeeeeeeee" << std::endl;
  for(int i = start; i < end; ++ i)
  {
    for(int j = 0; j < N; ++j)
    {
      for (int k = 0; k < N; ++k)
      {
	  DrawHexMC( i*bs, j*bs, k*bs, bs, bs, bs,
		   255, *vec);
      }
    }
  }
  std::for_each(vec->begin(), vec->end(), MyTransform);
}
void GenerateMCSolidList()
{
  std::cout << "Solid list start " << std::endl;
  const int N = Resolution;
  mpt::Thread threads[NT];
  std::vector<GlObjectd> vecs;
  int i = 0;
  GenerateMCSolidList1(i*N/NT, (i+1)*N/NT, &vecs);
    SolidList.CompileStart();
  //  glBegin(GL_TRIANGLES);
  std::for_each(vecs.begin(), vecs.end(), std::mem_fun_ref(&GlObjectd::Execute));
   //  glEnd();
  SolidList.CompileFinish();
}
void GenerateBoneList1(int start, int end, std::vector<GlObjectd>*vec)
{
  const int N = Resolution;
  if (CurDir != Ndir)
  {
    const int N = Resolution;
    double bs = 1./N;
    for(int i = start; i < end; ++ i)
    {
      for(int j = 0; j < N; ++j)
      {
	for (int k = 0; k < N; ++k)
	{
	  if (bonefunc(i*bs +bs*0.5, j*bs +bs*0.5, k*bs +bs*0.5)
	      >= bonethres)
	    DrawHex( i*bs, j*bs, k*bs, bs, bs, bs,
		     SolidPart(i,j,k), *vec);
	}
      }
    }
  }
  std::for_each(vec->begin(), vec->end(), MyTransform);
}
void GenerateBoneList()
{
  const int N = Resolution;

  mpt::Thread threads[NT];
  std::vector<GlObjectd> vecs[NT];
  for (int i = 0; i < NT; ++i)
    threads[i] = mpt::StartThread(
				  std::tr1::bind(
					      GenerateBoneList1, i*N/NT, (i+1)*N/NT, vecs+i));
  for(int i = 0; i < NT; ++i)
    threads[i].Join();
  BoneList.CompileStart();
  glBegin(GL_QUADS);
  for (int i = 0; i < NT; ++i)
    std::for_each(vecs[i].begin(), vecs[i].end(), mem_fun_ref(&GlObjectd::Execute));
  glEnd();
  BoneList.CompileFinish();
}




inline double razt(double x1, double y1, double z1,
		   double x2, double y2, double z2)
{
  x1-=x2;
  y1-=y2;
  z1-=z2;
  return sqrt(x1*x1 + y1*y1 + z1*z1);
}


void GenIsoList1(int start, int end, std::vector<GlObjectd> *vec)
{
  const int p = 1;
  const int N = Resolution*p;
  double bs = 1./N;
  GlObjectd v;
  v.glColor4d(isocol, isocol, isocol, 0.9);
  vec->push_back(v);


  for(int i = start; i < end; ++ i)
    for(int j = 0; j < N; ++j)
      for (int k = 0; k < N; ++k)
	DrawHexIso(i, j ,k, bs, SolidPart(i/p,j/p,k/p), *vec);
  std::for_each(vec->begin(), vec->end(), MyTransform);
}
void GenIsoList()
{
  const int p = 1;
  const int N = Resolution*p;
  mpt::Thread threads[NT];
  std::vector<GlObjectd> vecs[NT];
  for (int i = 0; i < NT; ++i)
    threads[i] = mpt::StartThread(
				  std::tr1::bind(
					      GenIsoList1, i*N/NT, (i+1)*N/NT, vecs+i));
  for(int i = 0; i < NT; ++i)
    threads[i].Join();
  IsoList.CompileStart();
  glBegin(GL_LINES);
  for (int i = 0; i < NT; ++i)
    std::for_each(vecs[i].begin(), vecs[i].end(), mem_fun_ref(&GlObjectd::Execute));
  glEnd();
  IsoList.CompileFinish();
}

void RenewIso()
{
  if(DrawIso)
    GenIsoList();
}
void   RenewLighting()
{
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  {
    double c = 0.6;
    glColor4d(c, c, c, 1);
  }
  if (UseLighting)
  {
    
    GLfloat lamb[4] = {0.2, 0.2, 0.2, 1};
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT,lamb);
    //    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
    glEnable(GL_COLOR_MATERIAL);
    glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
    GLfloat dcol [] = { 1,1 ,1,1};
    GLfloat scol [] = { 1, 1 ,1,1};
    GLfloat acol [] = { 0, 0 ,0,1};

    {
      float c = 0.15;
      float amb[] = {c,c,c,1};
      glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, amb);
    }

    GLfloat pos1 [] = { 0.5, -1, -4.0};
    GLfloat pos2 [] = { 0.5, -4, -4.0};
    GLfloat pos3 [] = { 4, 0.5, 0.5};
    GLfloat pos4 [] = { -4, 0.5, 0.5};
    GLfloat pos5 [] = {-4, 0.5, 4};
    GLfloat pos6 [] = { -2, 0.5, -4};
    glLightfv(GL_LIGHT0, GL_POSITION, pos1);
    glLightfv(GL_LIGHT1, GL_POSITION, pos2);
    glLightfv(GL_LIGHT2, GL_POSITION, pos3);
    glLightfv(GL_LIGHT3, GL_POSITION, pos4);
    glLightfv(GL_LIGHT4, GL_POSITION, pos5);
    glLightfv(GL_LIGHT5, GL_POSITION, pos6);
    
    glLightfv(GL_LIGHT0, GL_AMBIENT, acol);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, dcol);
    glLightfv(GL_LIGHT0, GL_SPECULAR, scol);
    glLightfv(GL_LIGHT1, GL_AMBIENT, acol);
    glLightfv(GL_LIGHT1, GL_DIFFUSE, dcol);
    glLightfv(GL_LIGHT1, GL_SPECULAR, scol);
    glLightfv(GL_LIGHT2, GL_AMBIENT, acol);
    glLightfv(GL_LIGHT2, GL_DIFFUSE, dcol);
    glLightfv(GL_LIGHT2, GL_SPECULAR, scol);
    glLightfv(GL_LIGHT3, GL_AMBIENT, acol);
    glLightfv(GL_LIGHT3, GL_DIFFUSE, dcol);
    glLightfv(GL_LIGHT3, GL_SPECULAR, scol);
    glLightfv(GL_LIGHT4, GL_AMBIENT, acol);
    glLightfv(GL_LIGHT4, GL_DIFFUSE, dcol);
    glLightfv(GL_LIGHT4, GL_SPECULAR, scol);
    glLightfv(GL_LIGHT5, GL_AMBIENT, acol);
    glLightfv(GL_LIGHT5, GL_DIFFUSE, dcol);
    glLightfv(GL_LIGHT5, GL_SPECULAR, scol);
    glEnable(GL_LIGHT0);
    //    glEnable(GL_LIGHT1);
    // glEnable(GL_LIGHT2);
    // glEnable(GL_LIGHT3);
    // glEnable(GL_LIGHT4);
    // glEnable(GL_LIGHT5);
    glEnable(GL_LIGHTING);
  }else
  {
    glDisable(GL_LIGHTING);
    glDisable(GL_LIGHT0);
    glDisable(GL_LIGHT1);
    glDisable(GL_LIGHT2);
    glDisable(GL_LIGHT3);
    glDisable(GL_LIGHT4);
    glDisable(GL_LIGHT5);
  }
  
  glPopMatrix();
}

void RenewSolid(bool NewShape = true)
{
  InitColors();
  RenewLighting();
  SetMinMax();
  vertex_map.Clear();
  if (NewShape and not UseMc)
  {
    GenerateSolid();
  }
  if (!UseMc)
  {
    GenerateSolidList();
  }else
    GenerateMCSolidList();
  // GenerateBoneList();
  RenewIso();
}

void gfxinit ()
{
  static bool firsttime = true;
  static double min, max;
  glLineWidth(2);
  //  glEnable(GL_NORMALIZE);
  glDisable(GL_NORMALIZE);
  //  glEnable(GL_ALPHA_TEST);
  glEnable(GL_CULL_FACE);
  //  glDisable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);
  glHint(GL_POINT_SMOOTH_HINT, GL_FASTEST);
  glHint(GL_LINE_SMOOTH_HINT, GL_FASTEST);
  // glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
  // glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
  glDepthFunc(GL_LESS);
  glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
  //  glEnable(GL_FOG);
  double ff = 0.99;
  GLfloat fogColor[4] = {ff, ff, ff, 1.0};
  int fogMode = GL_EXP;
  glFogi (GL_FOG_MODE, fogMode);
  glFogfv (GL_FOG_COLOR, fogColor);
  glFogf (GL_FOG_DENSITY, 0.20);
  glHint (GL_FOG_HINT, GL_DONT_CARE);
  glFogf (GL_FOG_START, 2);
  glFogf (GL_FOG_END, 4);
  
  
  MAX_X = 1;
  MIN_X = 0;
  MAX_Y = 1;
  MIN_Y = 0;
  MAX_Z = 1;
  MIN_Z = 0;
  xc = .5;
  yc  = .5;
  zc = .5;
}


void display()
{
  gl2psBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glDisable(GL_BLEND);
  gl2psLineWidth(1.25);
  glLineWidth(1.25);
  glClearColor(1,1,1,1);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glAlphaFunc(GL_GEQUAL, 1-atest);
  glShadeModel(SolidColor?GL_FLAT:GL_SMOOTH);
 double eps = 1e-3;
  glDepthRange(eps, 1);
  glShadeModel(SolidColor?GL_FLAT:GL_SMOOTH);
  if(DrawSolid)
  {
    RenewLighting();
    glDisable(GL_BLEND);
    gl2psDisable(GL2PS_BLEND);
    GLenum mode;
    switch(WireFrame)
    {
    case 0: mode = GL_FILL;break;
    case 1: mode = GL_LINE;break;
    case 2: mode = GL_POINT;break;
    }
    glPolygonMode(GL_FRONT, mode);
    glCullFace(GL_BACK);
    glCallList(SolidList);




    if (RemoveColor)
    {
      glEnable (GL_BLEND);
      gl2psEnable(GL2PS_BLEND);
      bool olduse = UseLighting;
      UseLighting = false;
      RenewLighting();
      glEnable(GL_LIGHTING);
      glBlendFunc (GL_SRC_ALPHA ,GL_ONE_MINUS_SRC_ALPHA);
      glPolygonMode (GL_BACK, GL_LINE);
      glCullFace (GL_FRONT);// Don't Draw Any Front-Facing Polygons             
      glDepthFunc (GL_LEQUAL);// Change The Depth Mode                          
      glDepthRange(0.2*eps,1);
      glDepthMask(GL_FALSE);
      glCallList(SolidList);
      glDepthMask(GL_TRUE);
      glDepthFunc (GL_LESS);// Change The Depth Mode                            
      glCullFace (GL_BACK);// Don't Draw Any Front-Facing Polygons              
      UseLighting = olduse;
    }
    // else                                                                     
    //   glPolygonMode (GL_BACK, GL_FILL);
  }
  glDisable(GL_LIGHTING);
  glEnable(GL_BLEND);
  gl2psEnable(GL2PS_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  gl2psBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  gl2psLineWidth(2);
  glDepthFunc (GL_LEQUAL);// Change The Depth Mode
  if(DrawIso)
  {
    glDepthMask(GL_FALSE);
    glCallList(IsoList);
    glDepthMask(GL_TRUE);
  }
  glDisable(GL_BLEND);
  gl2psDisable(GL2PS_BLEND);
  if (GrayBone)
  {
    glDepthRange(0, 1-eps);
    glCallList(BoneList);
  }
  glDepthFunc (GL_LESS);// Change The Depth Mode
  glFlush();
  RenewLighting();
  glutSwapBuffers();
}

static int width, height;
void reshape (int w, int h)
{
  width = w;
  height = h;
  glViewport (0, 0, (GLsizei) w, (GLsizei) h); 
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluPerspective(30.0, (GLfloat) w/(GLfloat) h, 1.0, 200.0);
  orientMe();
}

#include <png.h>
int curframe = 0;
void save_png(int frame=curframe)
{
  curframe++;
  FILE *fp;
  ostringstream os;
  os.fill('0');
  os<<setw(4) << frame <<".out" << ".png";
  fp = fopen(os.str().c_str(), "wb");
  glReadBuffer(GL_FRONT);
  GLfloat *pixels = new GLfloat[width*height*4];
  glReadPixels(0,0, width, height, GL_RGBA, GL_FLOAT, pixels);
  //    cout << " " << width << "x" << height << endl;
  try{
    png_structp png_ptr = png_create_write_struct
      (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if (!png_ptr)
      throw(4);
      
    png_infop info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr)
    {
      png_destroy_write_struct(&png_ptr,
			       (png_infopp)NULL);
      throw 5;
    }
      
    if (setjmp(png_jmpbuf(png_ptr)))
    {
      png_destroy_write_struct(&png_ptr, &info_ptr);
      fclose(fp);
      throw 6;
    }
    png_init_io(png_ptr, fp);      
    /* set the zlib compression level */
    png_set_compression_level(png_ptr,
			      Z_BEST_COMPRESSION);

    /* set other zlib parameters */
    png_set_compression_mem_level(png_ptr, 8);
    png_set_compression_strategy(png_ptr,
				 Z_DEFAULT_STRATEGY);
    png_set_compression_window_bits(png_ptr, 15);
    png_set_compression_method(png_ptr, 8);
    png_set_compression_buffer_size(png_ptr, 8192);
    png_set_IHDR(png_ptr, info_ptr, width, height,
		 8, PNG_COLOR_TYPE_RGB_ALPHA, PNG_INTERLACE_NONE,
		 PNG_COMPRESSION_TYPE_DEFAULT,
		 PNG_FILTER_TYPE_DEFAULT);
    png_set_sRGB(png_ptr, info_ptr, PNG_sRGB_INTENT_PERCEPTUAL);
    png_color_8 a;
    a.red = 8;
    a.green=8;
    a.blue=8;
    a.gray=0;
    a.alpha=0;
    //    png_set_sBIT(png_ptr, info_ptr, &a);
    png_text  text[6];
    int  num_text = 0;
    text[num_text].compression = PNG_TEXT_COMPRESSION_NONE;
    text[num_text].key = const_cast<char*>("Title");
    text[num_text].text = const_cast<char*>("Kartinka");
    ++num_text;
    text[num_text].compression = PNG_TEXT_COMPRESSION_NONE;
    text[num_text].key = const_cast<char*>("Author");
    text[num_text].text = const_cast<char*>("Yavor Vutov");
    ++num_text;
    text[num_text].compression = PNG_TEXT_COMPRESSION_NONE;
    text[num_text].key = const_cast<char*>("Description");
    text[num_text].text = const_cast<char*>("3d Rendering");
    ++num_text;
    
    png_set_text(png_ptr, info_ptr, text, num_text);    
    png_set_packing(png_ptr);

    
    
    unsigned char **row_pointers = new unsigned char* [2*height];
    for (int i=0; i<height; i++)
      row_pointers[i]= new unsigned char[4*width];

    for(int i = 0; i < height;++i)
      for(int j = 0; j < width; ++j)
      {
	row_pointers[i][j*4] =  static_cast<unsigned char>(255*pixels[(height-1-i)*width*4 +j*4+0]);
	row_pointers[i][j*4+1] = static_cast<unsigned char>(255*pixels[(height-1-i)*width*4 +j*4+1]);
	row_pointers[i][j*4+2] = static_cast<unsigned char>(255*pixels[(height-1-i)*width*4 +j*4+2]);
	row_pointers[i][j*4+3] = static_cast<unsigned char>(255*pixels[(height-1-i)*width*4 +j*4+3]);
	row_pointers[i][j*4+3] = 255; // no transparency
      }
    
    png_set_rows(png_ptr, info_ptr, row_pointers);

    //    png_write_info(png_ptr, info_ptr);
    
    png_write_png(png_ptr, info_ptr, PNG_TRANSFORM_IDENTITY, NULL);
    png_write_end(png_ptr, info_ptr);

    for (int i=0; i<height; i++)
      delete[] row_pointers[i];
    delete[] row_pointers;
    png_destroy_write_struct(&png_ptr, &info_ptr);
  }catch(int err){
    cerr << "Png write error " << err << endl;}
  delete[]pixels;
  fclose(fp);
}  
void save_eps(int frame=0)
{
  FILE *fp;
  int state = GL2PS_OVERFLOW, buffsize = 16*1024*1024;
  ostringstream os;
  os.fill('0');
  os<<setw(4) << frame <<".out" << ".eps";
  fp = fopen(os.str().c_str(), "wb");
  while(state == GL2PS_OVERFLOW){
    buffsize *= 2;
    gl2psBeginPage("test", "gl2psTestSimple", NULL,
		   GL2PS_EPS,
		   //		   GL2PS_SIMPLE_SORT,
		   GL2PS_BSP_SORT,
		   //                       GL2PS_DRAW_BACKGROUND |
		   GL2PS_USE_CURRENT_VIEWPORT
		   //		   |GL2PS_NO_BLENDING
		   |GL2PS_OCCLUSION_CULL
		   |GL2PS_SIMPLE_LINE_OFFSET
		   , 
		   GL_RGBA, 0, NULL, 0, 0, 0,  buffsize, fp, "Kartinka");
    display();
    state = gl2psEndPage();
  }
  fclose(fp);
}
void save_avi()
{
  const int numframes = 360;
  double step = atan(1)*8/numframes;
  for(int a = 0; a<numframes;++a)
  {
    if (DirUp)
    {
      movedistance+= movesize;
      if(++CurMoveCount>=MaxMoveCount)
      {
	DirUp = !DirUp;
	CurMoveCount = 0;
      }
    }else
    {
      movedistance-= movesize;
      if(++CurMoveCount>=MaxMoveCount)
      {
	DirUp = !DirUp;
	CurMoveCount = 0;
      }
    }
    RenewSolid();
    
    save_eps(a);
    angle1 = step;
    orientMe();
    glutPostRedisplay();
  }
}
bool active = false;
void avi_png_idle();
void save_avi_png()
{
  active = true;
  curframe = 0;
  
  glutIdleFunc(avi_png_idle);
}

void menu(int item)
{
  glutPostRedisplay();
  switch(item) {
  case SAVE_EPS:
    save_eps();
    break;
  case SAVE_PNG:
    save_png();
    break;
  case SAVE_AVI:
    save_avi();
    break;
  case QUIT:
    exit(0);
  }
}


void inputKey(int key, int x, int y) {
  switch (key) {
  case GLUT_KEY_F1:
    swapp=not swapp;
    if (swapp)
    {
      for(int i = 0; i < 3; ++i)ox[i]=oy[i]=oz[i]=xob[i]=xu[i]=0;
      ox[0] = oy[1] = oz[2] = xob[0] = xu[1]= 1;
    }
    break;
  case GLUT_KEY_PAGE_DOWN:
    angle -=0.02f;
    break;
  case GLUT_KEY_LEFT : 
    angle1 += 0.02f;
    break;
  case GLUT_KEY_RIGHT :
    angle1 -=0.02f;
    break;
  case GLUT_KEY_UP : 
    angle2 +=0.02f;
    break;
  case GLUT_KEY_PAGE_UP:
    dist -= 0.1;
    break;
  case GLUT_KEY_DOWN :
    angle2 -=0.02f;
    break;
  case GLUT_KEY_F5 :
    DrawSolid = !DrawSolid;
    break;
  case GLUT_KEY_F7 :
    ShowPlates = !ShowPlates;
    RenewSolid();
    break;
  case GLUT_KEY_F8 :
    //    ShowPalette = !ShowPalette;
    SolidColor = !SolidColor;
    RenewSolid(false);
    break;
  case GLUT_KEY_F10 :
    RemoveColor = ! RemoveColor;
    break;
  case GLUT_KEY_F3 :
    Shading = !Shading;
    RenewSolid();
    break;
  case GLUT_KEY_F2 :
    GrayBone = !GrayBone;
    RenewSolid(false);
    break;
  case GLUT_KEY_F11 :
    WireFrame = (WireFrame+1)%3;
    break;
  case GLUT_KEY_F12 :
    isobehav = (isobehav+1)%6;
    RenewSolid(false);
    break;
  case GLUT_KEY_F6:
    DrawIso = !DrawIso;
    break;
  case GLUT_KEY_INSERT:
    dist += 0.1;
    break;
  default:
    return;
  }
  orientMe();
  glutPostRedisplay();
	
}
void inputKey1(unsigned char key, int x, int y) {
  switch (key) {
  case 127 : 
    angle += 0.02f;
    break;
  case '`':
    Crashed = !Crashed;
    RenewSolid(false);
    break;
  case 'l':
    UseLighting = !UseLighting;
    RenewSolid();
    break;
  case '!':
    //    if (bonethres >-0.95)
      bonethres -=1;
    RenewSolid();
    break;
  case '@':
    //    if (bonethres <0.95)
      bonethres +=1;
    RenewSolid();
    break;
  case 'C':
    colourset = (colourset+1)%maxcolourset;
    RenewSolid();
    break;
  case 'n':
    if (CurDir!=Ndir)
    {
      CurDir = Ndir;
      RenewSolid();
    }
    break;
  case 'c':
    if (CurDir!=Cutdir)
    {
      CurDir = Cutdir;
      RenewSolid();
    }
    break;
  case 'x':
    if (CurDir!=Xdir)
    {
      CurDir = Xdir;
      RenewSolid();
    }
    break;
  case 'y':
    if (CurDir!=Ydir)
    {
      CurDir = Ydir;
      RenewSolid();
    }
    break;
  case 'z':
    if (CurDir!= Zdir)
    {
      CurDir = Zdir;
      RenewSolid();
    }
    break;
  case 'M':
    if (CurDir!= Zdir)
    {
      UseMc = !UseMc;
      RenewSolid();
    }
    break;
  case '5':
    if (movedistance>=movesize)
    {
      movedistance-=movesize;
      RenewSolid();
    }else if (movedistance>0)
    {
      movedistance = 0;
      RenewSolid();
    }
    break;
  case '6':
    movedistance+=movesize;
    RenewSolid();
    break;
  case '7':
    movesize/=2;
    break;
  case '8':
    movesize*=2;
    break;
    
  case 'j':
    if(CurLine)
    {
      CurLine--;
      if (CurDir!=Ndir || CurDir != Cutdir)
      {
	RenewSolid();
      }
    }
    break;
  case 'k':
    if(CurLine<Resolution-1)
    {
      CurLine++;
      if (CurDir!=Ndir || CurDir != Cutdir)
      {
	RenewSolid();
      }
    }
    break;
  case 'm':
    if(mypol>0.1)
    {
      mypol -=0.08;
      RenewSolid(false);
    }
    break;
  case '1':
    minfixer -=1;
    RenewSolid(false);
    break;
  case '2':
    minfixer +=1;
    RenewSolid(false);
    break;
  case '3':
    maxfixer -=1;
    RenewSolid(false);
    break;
  case '4':
    maxfixer +=1;
    RenewSolid(false);
    break;
  case ',':
    if(mypol<4)
    {
      mypol+=0.08;
      RenewSolid(false);
    }
    break;
  case 'i':
    if(il<1000)
    {
      il++;
      RenewIso();
    }
    break;
  case 'u':
    if(il>1)
    {
      il--;
      RenewIso();
    }
    break;
  case 'I':
    if(il<1000)
    {
      il+=50;
      RenewIso();
    }
    break;
  case 'U':
    if(il>51)
    {
      il-=50;
      RenewIso();
    }
    break;
	  
  case 's':
    save_png();
    break;
  case 'a':
    save_avi_png();
    break;
  case '-':
    atest -= 1/256.;
    break;
  case '+':
    atest += 1/256.;
    break;
  default:
    return;
  }
  orientMe();
  glutPostRedisplay();
	
}

void avi_png_idle()
{
  const int numframes=255*2;//360;
  
  save_png(curframe);
  double step = atan(1)*8/numframes;
  //  angle1 = step;

  if (DirUp)
  {
    bonethres+=1;
    movedistance+= movesize;
    if(++CurMoveCount>=MaxMoveCount)
    {
      DirUp = !DirUp;
      CurMoveCount = 0;
    }
  }else
  {
    bonethres-=1;
    movedistance-= movesize;
    if(++CurMoveCount>=MaxMoveCount)
    {
      DirUp = !DirUp;
      CurMoveCount = 0;
    }
  }
  RenewSolid();
  
  orientMe();
  glutPostRedisplay();
  cout << curframe-1 << "\r" << flush;
  if(curframe==numframes)
  {
    cout << "                             " <<  "\r" << flush;
    glutIdleFunc(NULL);
  }
}

void MyGlutInit(int &argc, char **&argv)
{
    glutInit(&argc, argv);
    glutInitWindowPosition(-1, -1);
    glutInitWindowSize(800, 800);
  
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA| GLUT_DEPTH|GLUT_ALPHA);
    glutCreateWindow("Bone Drawer");
    glutDisplayFunc(display);
    //glutIdleFunc(display);
    glutCreateMenu(menu);
    glutAddMenuEntry("Save EPS", SAVE_EPS);
    glutAddMenuEntry("Save PNG", SAVE_PNG);
    glutAddMenuEntry("Save AVI", SAVE_AVI);
    glutAddMenuEntry("Exit", QUIT);
    glutAttachMenu(GLUT_RIGHT_BUTTON);
    glutSpecialFunc(inputKey);
    glutKeyboardFunc(inputKey1);
    glutReshapeFunc(reshape);
    glEnable(GL_LINE_SMOOTH);
    glEnable(GL_POLYGON_SMOOTH);
    //  glutIdleFunc(idle);
    glPointSize(600./Resolution);
}

int main(int argc, char **argv)
  try {

    filename = "64.bonec";
    filename1= "128.z.3.20";
    if (argc >1)
      filename = argv[1];
    if (argc >2)
      filename1 = argv[2];
    if (argc>3)
      Resolution = atol(argv[3]);
    filenameI[0] = filenameI[1] = filenameI[2] = "128.z.0.00";

    if (argc >4)
      filenameI[0] = argv[4];
    if (argc >5)
      filenameI[1] = argv[5];
    if (argc >6)
      filenameI[2] = argv[6];

    
    ReadData();
    GenMins();
    RenewSolid();
    std::cout << surfp.size() << std::endl;

    // std::ofstream f("f.stl");
    // f << "sold blood\n";
    // for (int i = 0; i < surfp.size(); i++)
    //   surfp[i]*=150;
    // for (int i = 0; i < surfp.size(); i+=9)
    // {
    //   f << "facet normal 0 0 0\n";
    //   f << "  outer loop\n";
    //   f << "    vertex " << surfp[i+2] << " " << surfp[i+0] << " " << surfp[i+1] << "\n";
    //   f << "    vertex " << surfp[i+5] << " " << surfp[i+3] << " " << surfp[i+4] << "\n";
    //   f << "    vertex " << surfp[i+8] << " " << surfp[i+6] << " " << surfp[i+7] << "\n";
    //   f << "  endloop\n";
    //   f << "endfacet\n";
    // }    
    
    typedef std::map<VertexInfo1<float>, int > map_type;
    typedef typename map_type::iterator iter;
    map_type map;
    int ind = 1;
    for (int i = 0; i < surfp.size(); i+=3)
    {
      VertexInfo1<float> a;
      std::copy(&surfp[i], &surfp[i+3], a.vertex);
      if (map.find(a)==map.end())
      {
    	map.insert(std::make_pair(a, ind++));
      }
    }
    std::cout << map.size() << " distinct verts" << std::endl;

    std::ofstream f("f.geo");
    f << "algebraic3d\n";
    f << "solid blood = polyhedron(\n";
    typedef std::map<int, VertexInfo1<float> > map_type1;
    typedef typename map_type1::iterator iter1;
    map_type1 map1;
    for (iter it = map.begin(); it != map.end(); ++it)
      map1.insert(std::make_pair(it->second, it->first));
    double tsize = 150.;
    for (iter1 it = map1.begin(); it!=map1.end(); ++it)
    {
      f << it->second.vertex[0]*tsize << ", " << it->second.vertex[1]*tsize << ", " << it->second.vertex[2]*tsize << ";\n";
    }
    f << ";\n";
      
    for (int i = 0; i < surfp.size(); i+=9)
    {
      VertexInfo1<float> a, b,c;
      std::copy(&surfp[i], &surfp[i+3], a.vertex);
      std::copy(&surfp[i+3], &surfp[i+6], b.vertex);
      std::copy(&surfp[i+6], &surfp[i+9], c.vertex);
      f << map[a] << ", " << map[b] << ", " << map[c] ;
      if (i+9 < surfp.size())
    	f << ";\n";
    }
    f << ");\n";
      
    
  } catch (const std::exception & e) {
    std::cerr << e.what() << std::endl;
  }
