#include "common.h"
#include "visual.h"

void display()
{
    gl2psBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    gl2psLineWidth(0.25);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glCallList(1);
    glFlush();
    glutSwapBuffers();
}

void color(double d)
{
    d *= 4;
    double r, g, b;
    if (d < 1.) {
        r = 0.; g = d; b = 1.;
    } else if (d < 2.) {
        r = 0.; g = 1.; b = 2. - d;
    } else if (d < 3.) {
        r = d - 2; g = 1.; b = 0.;
    } else {
        r = 1.; g = 4 - d; b = 0.;
    }
    glColor3d(r, g, b);
}

inline bool between(double a, double b, double x)
{
    const double eps = 1./256.;
    return (x > std::min(a, b) + eps) && (x < std::max(a, b) - eps);
}

void c_tri(double x1, double y1, double c1,
           double x2, double y2, double c2,
           double x3, double y3, double c3)
{
    double x4, y4;
#define CHECK(i, j, k, val) \
    if (between(c##i, c##j, val)) { \
        x4 = ((val - c##i)*x##j + (c##j - val)*x##i)/(c##j - c##i); \
        y4 = ((val - c##i)*y##j + (c##j - val)*y##i)/(c##j - c##i); \
        c_tri(x##j, y##j, c##j, x##k, y##k, c##k, x4, y4, val); \
        c_tri(x##k, y##k, c##k, x##i, y##i, c##i, x4, y4, val); \
    }
#define CHECK_ALL(i, j, k) \
    CHECK(i, j, k, 0.5) \
    else \
    CHECK(i, j, k, 0.25) \
    else \
    CHECK(i, j, k, 0.75) \
    else 
    CHECK_ALL(1, 2, 3)
    CHECK_ALL(3, 1, 2)
    CHECK_ALL(2, 3, 1)
        {
        glBegin(GL_TRIANGLES);
        color(c1); glVertex2d(x1, y1);
        color(c2); glVertex2d(x2, y2);
        color(c3); glVertex2d(x3, y3);
        glEnd();
    }
}

void draw_tri(double x1, double y1, double c1,
              double x2, double y2, double c2,
              double x3, double y3, double c3,
              int N)
{
    c_tri(x1, y1, c1, x2, y2, c2, x3, y3, c3);
    double r1[N], r2[N], r3[N];
    glColor3d(0., 0., 0.);
    glLineWidth(.5);
    for (int i = N + 1; i--;) {
        double x = double(i)/(N + 1);
        r1[i] = (x - c2)/(c1 - c2);
        r2[i] = (x - c3)/(c2 - c3);
        r3[i] = (x - c1)/(c3 - c1);

        glBegin(GL_LINES);
        if ((r1[i] >= 0.) && (r1[i] <= 1.))
            glVertex2d(r1[i]*x1 + (1 - r1[i])*x2, r1[i]*y1 + (1 - r1[i])*y2);
        if ((r2[i] >= 0.) && (r2[i] <= 1.))
            glVertex2d(r2[i]*x2 + (1 - r2[i])*x3, r2[i]*y2 + (1 - r2[i])*y3);
        if ((r3[i] >= 0.) && (r3[i] <= 1.))
            glVertex2d(r3[i]*x3 + (1 - r3[i])*x1, r3[i]*y3 + (1 - r3[i])*y1);
        glEnd();
    }
}

void save_eps()
{
    FILE *fp;
    int state = GL2PS_OVERFLOW, buffsize = 4*4*1024*1024;
    
    fp = fopen("out.eps", "wb");
    while(state == GL2PS_OVERFLOW){
        buffsize *= 2;
        gl2psBeginPage("test", "gl2psTestSimple", NULL,
                       GL2PS_EPS, GL2PS_BSP_SORT, 
                       GL2PS_DRAW_BACKGROUND | GL2PS_USE_CURRENT_VIEWPORT |
                       GL2PS_OCCLUSION_CULL, 
                       GL_RGBA, 0, NULL, 0, 0, 0,  buffsize, fp, "out.eps");
        display();
        state = gl2psEndPage();
    }
    fclose(fp);
}

void menu(int item)
{
    switch(item) {
    case SAVE_EPS:
        save_eps();
        break;
    case QUIT:
        exit(0);
    }
    glutPostRedisplay();
}

