#ifndef __VISUAL_H__
# define __VISUAL_H__

# include "mesh.h"


void menu(int item);

void draw_tri(double x1, double y1, double c1,
              double x2, double y2, double c2,
              double x3, double y3, double c3,
              int N = 10);

template <class T> T min_angle(T x1, T y1, T x2, T y2, T x3, T y3)
{
    T a1 = std::abs(std::atan2(y2 - y1, x2 - x1) -
                    std::atan2(y3 - y1, x3 - x1));
    T a2 = std::abs(std::atan2(y1 - y2, x1 - x2) -
                    std::atan2(y3 - y2, x3 - x2));
    T a3 = std::abs(std::atan2(y1 - y3, x1 - x3) -
                    std::atan2(y2 - y3, x2 - x3));
    if (a1 > pi)
        a1 = 2*pi - a1;
    if (a2 > pi)
        a2 = 2*pi - a2;
    if (a3 > pi)
        a3 = 2*pi - a3;
    return std::min(a1, std::min(a2, a3))*57.295779513082320877;
}

template <class T> T max_angle(T x1, T y1, T x2, T y2, T x3, T y3)
{
    T a1 = std::abs(std::atan2(y2 - y1, x2 - x1) -
                    std::atan2(y3 - y1, x3 - x1));
    T a2 = std::abs(std::atan2(y1 - y2, x1 - x2) -
                    std::atan2(y3 - y2, x3 - x2));
    T a3 = std::abs(std::atan2(y1 - y3, x1 - x3) -
                    std::atan2(y2 - y3, x2 - x3));
    if (a1 > pi)
        a1 = 2*pi - a1;
    if (a2 > pi)
        a2 = 2*pi - a2;
    if (a3 > pi)
        a3 = 2*pi - a3;
    return std::max(a1, std::max(a2, a3))*57.295779513082320877;
}

template <class T> void gfxinit(const mesh_2d<T> & mesh,
                                const vector<T> & sol)
{
    glNewList(1, GL_COMPILE);
    T min_x = T(), min_y = T(), max_x = T(), max_y = T();
    T min_a = 360., max_a = 0.;
    T min_v = sol(0), max_v = sol(0);
    for (size_t i = sol.size(); i--;) {
        min_v = std::min(min_v, sol(i));
        max_v = std::max(max_v, sol(i));
    }
    std::cout << "Min value: " << min_v << std::endl
              << "Max value: " << max_v << std::endl;
    
    foreach(const typename mesh_2d<T>::element_set::value_type & e,
                  mesh.elements) {
        T x[3], y[3];
        size_t num[3], bound[3];
        for (int i = 3; i--;) {
            mesh_node_2d<T> n = mesh.node(e.nodes[i]);
            x[i] = n.x; y[i] = n.y; num[i] = n.number;
            bound[i] = n.boundary;
            min_x = std::min(min_x, n.x);
            min_y = std::min(min_y, n.y);
            max_x = std::max(max_x, n.x);
            max_y = std::max(max_y, n.y);
        }
        draw_tri(x[0], y[0], (sol(num[0]) - min_v)/(max_v - min_v),
                 x[1], y[1], (sol(num[1]) - min_v)/(max_v - min_v),
                 x[2], y[2], (sol(num[2]) - min_v)/(max_v - min_v),
                 12
        );
        glBegin(GL_LINES);
        for (int i = 3; i--;)
            if (mesh.boundary(e.nodes[(i + 1)%3], e.nodes[i]) > 1) {
                glVertex2d(x[i], y[i]);
                glVertex2d(x[(i + 1)%3], y[(i + 1)%3]);
            }
        glEnd();
        min_a = std::min(min_a,
                         min_angle(x[0], y[0], x[1], y[1], x[2], y[2]));
        max_a = std::max(max_a,
                         max_angle(x[0], y[0], x[1], y[1], x[2], y[2]));
//         glLineWidth(0.25);
//         glColor3d(0, 0, 0);
//         glBegin(GL_LINE_LOOP);
//         glVertex2d(x[0], y[0]);
//         glVertex2d(x[1], y[1]);
//         glVertex2d(x[2], y[2]);
//         glEnd();
    }

    std::cout << "MESH STATISTICS" << std::endl
              << "Min angle: " << min_a << std::endl
              << "Max angle: " << max_a << std::endl
              << "Elements: " << mesh.elements.size() << std::endl
              << "Nodes: " << mesh.nodes.size() << std::endl
              << std::endl;
    
    T w = max_x - min_x;
    T h = max_y - min_y;
    min_x -= .0625*w;
    max_x += .0625*w;
    min_y -= .0625*h;
    max_y += .0625*h;
    glEndList();

    glDisable(GL_DEPTH_TEST);
    glMatrixMode(GL_PROJECTION);
    glOrtho(min_x, max_x, min_y, max_y, -1., 1.);
    glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
    glEnable(GL_LINE_SMOOTH);
    glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
    glEnable(GL_POINT_SMOOTH);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glClearColor(1.0, 1.0, 1.0, 0.0);
}

#endif
