#ifndef __elastgen_h__
#define __elastgen_h__
#include "basic.h"
#include "small.h"
#include "sysgen.h"
#include "sparcemmat.h"
#include "useful.h"
EXPORT
template <class Double>
class ElastDiscretizator:public Discretizator<Double>
{
protected :
  typedef typename Discretizator<Double>::Integral Integral;
  using Discretizator<Double>::integral;
  typedef typename Discretizator<Double>::Integral2d Integral2d;
 public:
  typedef typename Discretizator<Double>::Poly3d Poly3d;
 protected:
  using Discretizator<Double>::integral2d;
  using Discretizator<Double>::tri;
  using Discretizator<Double>::disDPsiDt;
  using Discretizator<Double>::ShapeFun;
  using Discretizator<Double>::disShapeFun;
  using Discretizator<Double>::disJdet;
  using Discretizator<Double>::disJrev;
  using Discretizator<Double>::disJrevDPsiDt;

public:
  typedef SparceMacroMatrix<Double, 108, 21> SparceMacroMatrix_type;
  typedef typename SparceMacroMatrix_type::vector_type small_vector_type;
  ElastDiscretizator(DiscretizationType t):Discretizator<Double>(t)
  {
  }
  typedef SmallMatrix<Double, 108, 108, HeapStorage> DenseMacroMatrix_type;
  void GenElastElMat(SparceMacroMatrix_type &GSM1,
		 SparceMacroMatrix_type &GSM2,
		 SmallVector<Double, 3> &h
		 );
  void
  GenRhs(small_vector_type &rhs,
	 SmallMatrix<Double, 8, 3> elpointsFirst,
	 const SmallVector<Double, 3> &h, Double lambda, Double mu );
  
  void BoundCondSet(SparceMacroMatrix_type &res, SparceMacroMatrix_type const&src, BoundaryConditions bc, int which = 7);
  void BoundCondSetRhs(SparceMacroMatrix_type const &mat,
		       small_vector_type &rhs,
		       small_vector_type const &bvalues,
		       BoundaryConditions bc, int which = 7);
  void NoimanBoundCondSet(small_vector_type &rhs, SmallVector<Double, 3> &point, SmallVector<Double, 3> &h,int boundary);
  void NoimanSetOnSmall(SmallVector<Double, 18> &res, SmallVector<Double, 3> &point, SmallVector<Double, 3> &h, int boundary);
  void CalculateStressesAndStrains(SmallVector<Double, 3> &displacements,
				   SmallMatrix<Double, 3, 3> &stresses,
				   SmallMatrix<Double, 3, 3> &strains, 
				   SmallVector<Double, 18>&x, Double lambda, Double mu,
				   const SmallVector<Double, 3>& p=0);

  void CalculateStressesAndStrains(SmallVector<Poly3d, 3> &displacements,
				   SmallMatrix<Poly3d, 3, 3> &stresses,
				   SmallMatrix<Poly3d, 3, 3> &strains, 
				   SmallVector<Double, 18>&x, Double lambda, Double mu);
};
EXPORT
template<class Double>
SmallVector<Double, 3>
FuncRhs(const SmallVector<Double, 3> &p);
EXPORT
template<class Double>
SmallVector<Double, 3>
FuncSolution(const SmallVector<Double, 3> &p);
EXPORT
template<class Double>
SmallMatrix<Double, 3, 3>
FuncNeuman(const SmallVector<Double, 3> &p);

#endif
#ifdef INCLUSION
#include "elastgen.hpp"
#endif
