#ifndef __elastgen_hpp__
#define __elastgen_hpp__

#include "elastgen.h"
#include "sparcemmat.h"

 

template<class Double, int PointsNumber, class Func>
void
Discretize3(SmallVector<SmallVector<Double, 3>, PointsNumber> &disF,
	    SmallVector<SmallVector<Double, 3>, PointsNumber> &points,
	    Func &F)
{
  for(int i = 0; i < PointsNumber; ++i)
    disF[i] = F(points[i]);
  //  std::cout << "pt " <<points[0][0]<< " " << points[0][1] << " " << points[0][2] << std::endl;
}


template<class Double>
void
ElastDiscretizator<Double>::
CalculateStressesAndStrains(SmallVector<Double, 3> &displacements,
			    SmallMatrix<Double, 3, 3> &stresses,
			    SmallMatrix<Double, 3, 3> &strains, 
			    SmallVector<Double, 18>&x, Double lambda, Double mu,
			    const SmallVector<Double, 3>& p)
{
  SmallVector<Poly3d, 3> U;
  SmallMatrix<Poly3d, 3,3> NablaU;
  SmallMatrix<Poly3d, 3,3> NablaDotUI;
  for(int i = 0; i < 6; ++i)
    for(int j = 0; j < 3; ++j)
      U[j] += ShapeFun[i] * x[6*j+i];
  for(int i = 0; i < 3; ++i)
  {
    for(int j = 0; j <3; ++j)
      NablaDotUI[i][i] = NablaDotUI[i][i] + U[j].D(j);
    for(int j = 0; j < 3; ++j)
      NablaU[i][j] = U[j].D(i);
  }
  for(int i = 0; i < 3; ++i)
  {
    displacements[i] = U[i](p);
    for(int j = 0; j < 3; ++j)
    {
      strains[i][j] = 0.5 * (NablaU[i][j](p) + NablaU[j][i](p));
      stresses[i][j] = lambda * NablaDotUI[i][j](p) + 2.*mu *strains[i][j];
    }
  }
  
  
}
template<class Double>
void
ElastDiscretizator<Double>::
CalculateStressesAndStrains(SmallVector<Poly3d, 3> &displacements,
			    SmallMatrix<Poly3d, 3, 3> &stresses,
			    SmallMatrix<Poly3d, 3, 3> &strains, 
			    SmallVector<Double, 18>&x, Double lambda, Double mu)

{
  SmallVector<Poly3d, 3> &U = displacements;
  SmallMatrix<Poly3d, 3,3> NablaU;
  SmallMatrix<Poly3d, 3,3> NablaDotUI;
  for(int i = 0; i < 6; ++i)
    for(int j = 0; j < 3; ++j)
      U[j] += ShapeFun[i] * x[6*j+i];
  for(int i = 0; i < 3; ++i)
  {
    for(int j = 0; j <3; ++j)
      NablaDotUI[i][i] = NablaDotUI[i][i] + U[j].D(j);
    for(int j = 0; j < 3; ++j)
      NablaU[i][j] = U[j].D(i);
  }
  for(int i = 0; i < 3; ++i)
  {
    for(int j = 0; j < 3; ++j)
    {
      strains[i][j] = 0.5 * (NablaU[i][j] + NablaU[j][i]);
      stresses[i][j] = lambda * NablaDotUI[i][j] + 2.*mu *strains[i][j];
    }
  }
  
  
}

template<class Etype, class Double>
inline  void CalcE(Etype &E, Double lambda, double mu)
{
  for(int i = 0; i < 3; ++i)
    for(int j = 0; j < 3; ++j)
      for( int k = 0; k < 3; ++k)
	for( int l = 0; l < 3; ++l)
	  E[i][j][k][l] = Double();

  for(int i = 0; i < 3; ++i)
    for(int j = 0; j < 3; ++j)
      E[i][j][i][j] = E[i][j][j][i] = mu;
  for(int i = 0; i < 3; ++i)
    for(int j = 0; j < 3; ++j)
      E[i][i][j][j] = lambda;
  for(int i = 0; i < 3; ++i)
    E[i][i][i][i] = lambda + 2. * mu;
//   for(int i = 0; i < 3; ++i)
//     for(int j = 0; j < 3; ++j)
//     {
//       for( int k = 0; k < 3; ++k)
// 	for( int l = 0; l < 3; ++l)
// 	  std::cout << E[i][j][k][l] << " ";
//       std::cout << endl;
//     }
}

template<class Double>
void
ElastDiscretizator<Double>::
GenRhs(small_vector_type &rhs,
       SmallMatrix<Double, 8, 3> elpointsFirst,
       const SmallVector<Double, 3> &h, Double lambda, Double mu )
{
  
  static int itop[6][8] = {{1,2,3,4,17,18,19,20},
		{17,18,19,20,33,34,35,36},
		{5,6,10,11,21,22,26,27},
		{10,11,15,16,26,27,31,32},
		{7,8,12,13,23,24,28,29},
		{8,9,13,14,24,25,29,30}};
  // Stollen from the fortran program, so we decrement indexes
  static bool firsttime = true;
  static SmallVector<SmallVector<Double,  3>, 8> allelpointsoff;
  static SmallMatrix<Double, 8, 3> curcube;
  static SmallMatrix<Double, 8, 3> elpoints;

  if(firsttime)
  {
    for(int i = 0; i < 6; ++i)
      for(int j = 0; j < 8; ++j)
	--itop[i][j];

    for(int i = 0; i < 8; ++i)
      allelpointsoff[i] = Double(0);

    for(int i = 0; i < 4; ++i)
      allelpointsoff[i+4][0]  = h[0];

    allelpointsoff[1][2] = h[2];
    allelpointsoff[3][2] = h[2];
    allelpointsoff[5][2] = h[2];
    allelpointsoff[7][2] = h[2];

    allelpointsoff[2][1] = h[1];
    allelpointsoff[3][1] = h[1];
    allelpointsoff[6][1] = h[1];
    allelpointsoff[7][1] = h[1];
  }
  static Double E[3][3][3][3];
  if (Options::GetOptions().hom)  // get elastisity coeffs, from lambda & mu
    CalcE(E, lambda, mu);

    
  SmallVector<SmallVector<Double, 3>, 6> smallrhs;
  smallrhs = Double();
  rhs = Double();
  for(int cube = 0; cube < 8; ++cube)
  {
    SmallVector<SmallVector<Double, 3>, Integral::PointsNumber> disF, points;

    SmallVector<Double, 3> &start = elpointsFirst[cube];
      
  
    for(int p = 0; p < Integral::PointsNumber; ++p)
    {
      SmallVector<Double, 3> &pt = integral.points[p];
          
      points[p] = pt;
      points[p][0] += 1.;
      points[p][1] += 1.;
      points[p][2] += 1.;
      points[p]*=.5;
      points[p]*=h;
      points[p]+=start;    
    }

    if (!Options::GetOptions().hom)  // normal rhs
    {
      Discretize3(disF, points, FuncRhs<Double>);
  

      // Integration itself
      smallrhs = Double();
      for(int f = 0; f < Integral::PointsNumber; f++)
	for(int of = 0; of < 6; of++)
	{
	  smallrhs[of] += disF[f] * (disShapeFun[f][0][of]*integral.coeff[f]*disJdet[f]) ;
	}
      // we have now smallrhs and we should add it tho rhs
    }
    else
    { // homogenization
      if(firsttime)
      { //inicialization
	//SmallVector<Double, 3> &start = elpointsFirst[cube];
	//	curcube = allelpointsoff[cube];
	elpoints[0][0] = 0;
	elpoints[0][1] = 0;
	elpoints[0][2] = 0;
	
	elpoints[3] = elpoints[0];
	elpoints[3][1] += h[1];
	
	elpoints[1] = elpoints[0];
	elpoints[1][0] += h[0];
	
	elpoints[2] = elpoints[3];
	elpoints[2][0] += h[0];
	
	for(int jj = 0; jj <4; jj++)
	{
	  elpoints[4+jj] = elpoints[jj];
	  elpoints[4+jj][2] += h[2];
	}
	
	tri.SetDest(elpoints);
	for(int f = 0; f < Integral::PointsNumber; f++)
	{
	  tri.GetJacobianAndReverseAtPoint(disJrev[f], disJdet[f], integral.points[f]);
	  disJrev[f]/=disJdet[f];
// 	  std::cout <<  std::endl;
// 	  for(int i = 0; i < 3; ++i)
// 	  {
// 	    for (int j = 0; j < 3; ++j)
// 	      std::cout <<std::setw(10) <<  disJrev[f][i][j];
// 	    std::cout <<  std::endl;
// 	  }
	      
	  MatrixMultiply(disJrevDPsiDt[f], disJrev[f], disDPsiDt[f]);
// 	  std::cout <<  std::endl;
// 	  for(int i = 0; i < 3; ++i)
// 	  {
// 	    for (int j = 0; j < 6; ++j)
// 	      std::cout <<std::setw(10) <<  disDPsiDt[f][i][j];
// 	    std::cout <<  std::endl;
// 	  }
//  	  for(int i = 0; i < 3; ++i)
//  	    for(int j = 0; j < 3; ++j)
//  	      std:: cout << f << " " << i << " " << j << " " << disJrevDPsiDt[f][i][j] << std::endl;
// 	  for(int i = 0; i < 3; ++i)
// 	    for(int j = 0; j < 6; ++j)
// 	      std:: cout << f << " " << i << " " << j << " " <<  disJrevDPsiDt[f][i][j] << endl;
	}

      }

      //      std::vector<SmallMatrix<Double, 3, 6> > disJrevDPsiDt;
      // computation itself
      int hk = Options::GetOptions().homk;
      int hl = Options::GetOptions().homl;
      smallrhs = Double();
      for(int f = 0; f < Integral::PointsNumber; f++)
	for(int of = 0; of < 6; of++)
	{
	  for(int i = 0; i < 3; ++i)
	    for(int j = 0; j < 3; ++j)
	    {
	      //	      std::cout << " " << f << " " << j << " " << of << " " << disJrevDPsiDt[f][j][of]  << std::endl;
	      
	      smallrhs[of][i] +=  E[i][j][hk][hl] * (disJrevDPsiDt[f][j][of] * integral.coeff[f]*disJdet[f]);
	    }
	}
    }
    for(int i = 0; i < 6; ++i)
    {

      //      std:: cout << i << "\t" << smallrhs[i][0] << "\t" << smallrhs[i][1] << "\t" << smallrhs[i][2] << std::endl;
      rhs[ 0 + itop[i][cube]] += smallrhs[i][0];
      rhs[36 + itop[i][cube]] += smallrhs[i][1];
      rhs[72 + itop[i][cube]] += smallrhs[i][2];
      // std::cout
      // 	<< std::setw(15) << smallrhs[i][0]
      // 	<< std::setw(15) << smallrhs[i][1]
      // 	<< std::setw(15) << smallrhs[i][2]
      //  	<< std::endl;
    }
    //    std::cout << std::endl;
  }
   // for (int i = 0 ; i < 108; ++i )
   //   std::cout << i  << "\t" << rhs[i] << std::endl;
  firsttime = false;
}


template<class Double>
void
ElastDiscretizator<Double>::
NoimanBoundCondSet(small_vector_type &rhs, SmallVector<Double, 3> &point, SmallVector<Double, 3> &h,int boundary)
{
  static bool firsttime = true;
  static SmallVector<SmallVector<Double,  3>, 8> allelpointsoff;
  static int itop[6][8] = {{1,2,3,4,17,18,19,20},
		{17,18,19,20,33,34,35,36},
		{5,6,10,11,21,22,26,27},
		{10,11,15,16,26,27,31,32},
		{7,8,12,13,23,24,28,29},
		{8,9,13,14,24,25,29,30}};
  static int cubes[BSize][4];

  if(firsttime)
  {
    firsttime = false;

    cubes[BLeft][0] = 0;
    cubes[BLeft][1] = 1;
    cubes[BLeft][2] = 2;
    cubes[BLeft][3] = 3;

    cubes[BRight][0] = 4;
    cubes[BRight][1] = 5;
    cubes[BRight][2] = 6;
    cubes[BRight][3] = 7;

    cubes[BFront][0] = 0;
    cubes[BFront][1] = 1;
    cubes[BFront][2] = 4;
    cubes[BFront][3] = 5;

    cubes[BBack][0] = 2;
    cubes[BBack][1] = 3;
    cubes[BBack][2] = 6;
    cubes[BBack][3] = 7;

    cubes[BDown][0] = 0;
    cubes[BDown][1] = 2;
    cubes[BDown][2] = 4;
    cubes[BDown][3] = 6;

    cubes[BUp][0] = 1;
    cubes[BUp][1] = 3;
    cubes[BUp][2] = 5;
    cubes[BUp][3] = 7;
    

    
    for(int i = 0; i < 6; ++i)
      for(int j = 0; j < 8; ++j)
	--itop[i][j];

    for(int i = 0; i < 8; ++i)
      allelpointsoff[i] = Double(0);

    for(int i = 0; i < 4; ++i)
      allelpointsoff[i+4][0]  = h[0];

    allelpointsoff[1][2] = h[2];
    allelpointsoff[3][2] = h[2];
    allelpointsoff[5][2] = h[2];
    allelpointsoff[7][2] = h[2];

    allelpointsoff[2][1] = h[1];
    allelpointsoff[3][1] = h[1];
    allelpointsoff[6][1] = h[1];
    allelpointsoff[7][1] = h[1];
  }
  SmallMatrix<Double, 8, 3> elpoints;
  for(int i = 0; i < 8; ++i)
  {
    elpoints[i] = point;
    elpoints[i] += allelpointsoff[i];
  }

//     std::cout << "--" ;
//     for(int i = 0; i < 3; ++i)
//       std::cout << elpoints[cube][i] << " ";
//     std::cout << std::endl;
//  std::cout << std::endl;
  
  for(int c = 0; c < 4; c++)
  {
    int cube = cubes[boundary][c];

    SmallVector<Double, 18> res;
    res = Double();
    NoimanSetOnSmall(res, elpoints[cube], h, boundary);
    for(int i = 0; i < 6; ++i)
    {
      //      std:: cout << "hmm " << i << " " << res[0 + i] << " " << res[6 + i] << " " << res[12+i] << endl;
      rhs[0  + itop[boundary][cube]] += res[0  + i];
      rhs[36 + itop[boundary][cube]] += res[6  + i];
      rhs[72 + itop[boundary][cube]] += res[12 + i];
    }
  }
}

template<class Double>
void
ElastDiscretizator<Double>::
NoimanSetOnSmall(SmallVector<Double, 18> &res, SmallVector<Double, 3> &basepoint, SmallVector<Double, 3> &h, int boundary)
{
  SmallMatrix<Double, 3, 3> Hp;
  SmallVector<Double, 3> &normal = this->normals[boundary];
  res = Double();

  Double dvakob = 0;
  if (boundary == BLeft)
    dvakob = h[1]*h[2];
  if (boundary == BRight)
    dvakob = h[1]*h[2];
  if (boundary == BFront)
    dvakob = h[0]*h[2];
  if (boundary == BBack)
    dvakob = h[0]*h[2];
  if (boundary == BUp)
    dvakob = h[0]*h[1];
  if (boundary == BDown)
    dvakob = h[0]*h[1];

  dvakob *= 0.25;

  for(int f = 0; f < Integral2d::PointsNumber; ++f)
  {
    SmallVector<Double, 3> point;
    point = this->pointsOnBoundary[boundary][f];
    point[0] += 1;
    point[1] += 1;
    point[2] += 1;
    point *= 0.5;
    point *= h;
    point += basepoint;

    Hp = FuncNeuman(point);


    SmallVector<Double, 3> Hdisc;
    Hdisc = Double();
    for(int i = 0; i < 3; ++i)
      for(int j = 0; j < 3; ++j)
	Hdisc[i] += Hp[i][j] * normal[j];

//     for(int i = 0; i < 3; ++i)
//       std::cout << point[i] << " ";
//     std::cout << std::endl;

//    std::cout << point[0] << " " << point[1] << " " << point[2] << " " << Hdisc[2] << std::endl;

    for(int fun = 0; fun < 6; fun++)
      //    int fun = boundary;
    {
      res[0  + fun]  += integral2d.coeff[f] * this->disPsiBoundary[fun][boundary][f] * Hdisc[0] * dvakob;
      res[6  + fun]  += integral2d.coeff[f] * this->disPsiBoundary[fun][boundary][f] * Hdisc[1] * dvakob;
      res[12 + fun]  += integral2d.coeff[f] * this->disPsiBoundary[fun][boundary][f] * Hdisc[2] * dvakob;
    }
  }
}
 
template<class T>
inline void err_sum(T&Sum, T&Errsum,  double b)
{
   Sum += b;
//     T a = Sum;
//     T s = a + b;
//     T bb = s - a;
//     T err = (a - (s - bb)) + (b - bb);
//     Sum = s;
//     Errsum +=err;
} 
template<class Double>
void
ElastDiscretizator<Double>::
GenElastElMat(SparceMacroMatrix_type &GSM1_,
		 SparceMacroMatrix_type &GSM2_,
		 SmallVector<Double, 3> &h
		 )
{
  SmallMatrix<Double, 8, 3> elpoints;

  SmallMatrix<SmallMatrix<Double, 6, 6>, 3, 3, HeapStorage>
    a, a_err;
  SmallVector<Double, 7> t, t_err;
  SmallMatrix<Double, 3, 36, HeapStorage> tt;
  SmallMatrix<Double, 18, 18, HeapStorage> esm;
  
  DenseMacroMatrix_type GSM11, GSM12, GSM1, GSM2;
  int itop[6][8] = {{1,2,3,4,17,18,19,20},
		{17,18,19,20,33,34,35,36},
		{5,6,10,11,21,22,26,27},
		{10,11,15,16,26,27,31,32},
		{7,8,12,13,23,24,28,29},
		{8,9,13,14,24,25,29,30}};
  //projection indexes
  int itot[36][2] = {
    {0,6}, {0,6}, {0,6}, {0, 6}, // 0 - 3
    {2, 6}, {2, 6},              // 4 -5
    {4, 6}, {4, 5}, {5, 6},      // 6-8
    {2, 3}, {2,3},               // 9-10
    {4, 6}, {4, 5}, {5, 6},      // 11-13
    {3, 6}, {3, 6},              // 14-15
    {0,1}, {0,1}, {0,1}, {0,1},   // 16-19
    {2, 6}, {2, 6},              // 20-21
    {4, 6}, {4, 5}, {5, 6},      // 22-24
    {2, 3}, {2,3},               // 25-26
    {4, 6}, {4, 5}, {5, 6},      // 27-29
    {3, 6}, {3, 6},              // 30-31
    {1,6}, {1,6}, {1,6}, {1,6}};  //32-35

    
    
  // Stollen from the fortran program, so we decrement indexes
  for(int i = 0; i < 6; ++i)
    for(int j = 0; j < 8; ++j)
    --itop[i][j];

		  
  
  GSM11 = Double(); GSM12 = Double();
  GSM1 = Double(); GSM2 = Double();
  for(int i = 0; i < 3; ++i)
    for(int j = 0; j <3; ++j)
    {
      a[i][j] = Double();
      a_err[i][j] = Double();
    }
  
  elpoints[0][0] = 0;
  elpoints[0][1] = 0;
  elpoints[0][2] = 0;
     
  elpoints[3] = elpoints[0];
  elpoints[3][1] += h[1];
     
  elpoints[1] = elpoints[0];
  elpoints[1][0] += h[0];
     
  elpoints[2] = elpoints[3];
  elpoints[2][0] += h[0];
     
  for(int jj = 0; jj <4; jj++)
  {
    elpoints[4+jj] = elpoints[jj];
    elpoints[4+jj][2] += h[2];
  }

  //initialize transformation
  this->tri.SetDest(elpoints);

  for(int f = 0; f < Integral::PointsNumber; f++)
  {
    tri.GetJacobianAndReverseAtPoint(disJrev[f], disJdet[f], integral.points[f]);
    disJrev[f]/=disJdet[f];
    MatrixMultiply(disJrevDPsiDt[f], disJrev[f], disDPsiDt[f]);
  }

  // nonprojected strain part ->GSM11
  for (int I = 0; I < 3; ++I)
    for (int J = 0; J<3; ++J)
      for (int i = 0; i < 6; ++i)
	for (int j = 0; j < 6; ++j)
	  for(int f = 0; f < Integral::PointsNumber; ++f)
	  {
	    err_sum(
		a[I][J][i][j], 
		a_err[I][J][i][j], 
		integral.coeff[f] * disJrevDPsiDt[f][I][i] * disJrevDPsiDt[f][J][j] *disJdet[f]);
	  }
  //  std::cout << " det "<< disJdet[0] << std::endl;
  for (int I = 0; I < 3; ++I)
    for (int J = 0; J<3; ++J)
      for (int i = 0; i < 6; ++i)
	for (int j = 0; j < 6; ++j)
	    a[I][J][i][j] +=a_err[I][J][i][j];
  
  for(int el = 0; el < 8; ++el)
    for(int i = 0; i < 6; ++i)
    {
      int ii = itop[i][el];
      for(int j = 0; j < 6; ++j)
      {
	int jj = itop[j][el];
	Double sum = Double();
	for (int k = 0; k < 3; ++k)
	  sum += a[k][k][i][j];
			       
	for (int I = 0; I<3; ++I)
	  {
	    GSM11[I*36+ii][I*36 + jj] += sum;
	  }
      }
    }
  // Done GSM11
  // projected strain part -> GSM12
  for (int k = 0; k < 3; ++k)  // dimension
  {
    //simple integral calculation
    t = 0; // initialize integrals
    t_err = 0;
    for (int i = 0; i < 6; ++i)
    {
      for (int f = 0; f < Integral::PointsNumber; ++f)
      {
	err_sum(t[i],t_err[i], integral.coeff[f] * disJrevDPsiDt[f][k][i] * disJdet[f]);
	//	std::cout << "deter " << disJdet[f] <<std::endl;
      }
      t[i]+=t_err[i];
      t[i]/=8;
    }
    for(int i = 0; i < 36; ++i)
    {
      tt[k][i] = t[itot[i][0]] + t[itot[i][1]];
      if (isnan(tt[k][i]))
	std::cerr << "ttnan " << t[itot[i][0]] << " "  <<t[itot[i][1]] << std::endl;
      using namespace std;
      //      cout << setw(2) << k << setw(3) << i  << " " << fixed << tt[k][i] << endl;
    }
  }
  // tt now contains the projections

  for(int i = 0; i < 36; ++i)
    for(int j = 0; j < 36; ++j)
    {
      GSM12[i][j]         = tt[1][i]*tt[1][j] + tt[2][i]*tt[2][j];
      GSM12[i][j+36]      = - tt[1][i] * tt[0][j];
      GSM12[i][j+72]      = - tt[2][i] * tt[0][j];
      GSM12[i+36][j]      = - tt[0][i] * tt[1][j];
      GSM12[i+36][j+36]   = tt[0][i]*tt[0][j] + tt[2][i]*tt[2][j];
      GSM12[i+36][j+72]   = - tt[2][i] * tt[1][j];
      GSM12[i+72][j]      = - tt[0][i] * tt[2][j];
      GSM12[i+72][j+36]   = - tt[1][i] * tt[2][j];
      GSM12[i+72][j+72]   = tt[0][i]*tt[0][j] + tt[1][i]*tt[1][j];
    }
  for(int i = 0; i < 108; ++i)
    for(int j = 0; j < 108; ++j)
    {
      GSM12[i][j] /= -0.25*h[0]*h[1]*h[2]; // h^3/8 - j // -1/2 - formulas original
      //      GSM12[i][j] /=8.;
      //GSM12[i][j] /= -16.*h[0]*h[1]*h[2]/8; // h^3/8 - j // -1/2 - formulas
      //      GSM12[i][j] /= 0.5;
      //      GSM12[i][j] *= (-1./8)/(2.*8.*h[0]*h[1]*h[2]); // fortran
      //      GSM12[i][j] /= -2.*h[0]*h[1]*h[2]/8; // h^3/8 - j // -1/2 - formulas original
    }

  // Done GSM12
  for(int i = 0; i<108; ++i)
    for(int j = 0; j<108; ++j)
    {
      GSM1[i][j] = GSM11[i][j] + GSM12[i][j];
//       if(isnan(GSM1[i][j]))
// 	std::cerr << "1=nan " <<GSM11[i][j] << " " <<  GSM12[i][j] << std::endl;
    }

  //  assembling divergence part

  for(int I = 0; I < 3; ++I)
    for(int J = 0; J<3; ++J)
      for(int el = 0; el < 8; ++el)
	for(int i = 0; i < 6; ++i)
	{
	  int ii = itop[i][el];
	  for(int j = 0; j < 6; ++j)
	  {
	    int jj = itop[j][el];
	    
	    GSM2[36*I +ii][36*J + jj] += a[I][J][i][j];
	  }
	}
  using namespace std;
  //  int maxcolumn = 0;
  typedef SparceMacroMatrix<Double, 108, 21> SMT;
  typename SMT::index_type &index = SMT::GetIndex();
  int count = 0;
  for(int i = 0; i < 108; ++i)
  {
    int column = 0;
    
    for(int j = 0; j < 108; ++j)
    {
      //      cout.precision(8);
      const Double threshold = 1e-7;
//       if(fabs(GSM1[i][j])<threshold*fabs(GSM1[i][i]))
// 	GSM1[i][j] = 0;
//       if(fabs(GSM2[i][j])<threshold*fabs(GSM2[i][i]))
// 	GSM2[i][j] = 0;
      Double z = GSM1[i][j]*2 + GSM2[i][j];
      //      std::cout << fabs(z) << " " << threshold*fabs(GSM1[0][0]) << std::endl;
      if (fabs(z)>threshold*fabs(GSM1[0][0]))
      {
	index[i][column++] = j;
	if (j==i)
	  std::swap(index[i][0], index[i][column-1]);
      }else  if(isnan(z))
	std::cerr << "z=nan " <<GSM1[i][j] << " " <<  GSM2[i][j] << std::endl;
      //      cout << setw(4) << i <<setw(4) << j << " " << z << endl;
    }
    count += index[i][21] = column;
     if(column >21)
       throw "Too many columns ";
  }
//  GSM2 = 0.;
//  std::cout << count << std::endl;
//   for(int i = 0; i < 108; ++i)
//     std::cout << GSM1[i][i] << " " << GSM2[i][i] << std::endl;
  GSM1_ = GSM1;
  GSM2_ = GSM2;
//   for (int i = 0; i < 108; ++i)
//   {
//     int c = index[i][21];
//     std:: cout << "Columns " << c << std::endl;
//     cout.precision(15);
//     for (int j = 0; j < c; ++j)
//       std::cout << GSM1_[i][j] << " " << GSM2_[i][j] << std::endl;
//       std::cout << std::endl;
//   }
//   exit(1);
}
#include <iomanip>
template<class Double>
void
ElastDiscretizator<Double>::
BoundCondSet(SparceMacroMatrix_type &res, SparceMacroMatrix_type const&src, BoundaryConditions bc, int which)
{
  res = src;
  //  std::cout << std::hex<<(unsigned) bc << " " << which << std::endl;
  if(bc & bcLeft)
  {
    if (which & 1)
    {
      res.RemoveIndex(0);
      res.RemoveIndex(1);
      res.RemoveIndex(2);
      res.RemoveIndex(3);
    }


    if (which & 2)
    {
      res.RemoveIndex(0 + 36);
      res.RemoveIndex(1 + 36);
      res.RemoveIndex(2 + 36);
      res.RemoveIndex(3 + 36);
    }

    if (which & 4)
    {
      res.RemoveIndex(0 + 72);
      res.RemoveIndex(1 + 72);
      res.RemoveIndex(2 + 72);
      res.RemoveIndex(3 + 72);
    }
    
  }
  if(bc & bcRight)
  {
    if (which & 1)
    {
      res.RemoveIndex(32);
      res.RemoveIndex(33);
      res.RemoveIndex(34);
      res.RemoveIndex(35);
    }

    if (which & 2)
    {
      res.RemoveIndex(32 + 36);
      res.RemoveIndex(33 + 36);
      res.RemoveIndex(34 + 36);
      res.RemoveIndex(35 + 36);
    }

    if (which & 4)
    {
      res.RemoveIndex(32 + 72);
      res.RemoveIndex(33 + 72);
      res.RemoveIndex(34 + 72);
      res.RemoveIndex(35 + 72);
    }
    
  }
  if(bc & bcFront)
  {
    if (which & 1)
    {
      res.RemoveIndex(4 );
      res.RemoveIndex(5 );
      res.RemoveIndex(20);
      res.RemoveIndex(21);
    }

    if (which & 2)
    {
      res.RemoveIndex(4 + 36);
      res.RemoveIndex(5 + 36);
      res.RemoveIndex(20 + 36);
      res.RemoveIndex(21 + 36);
    }

    if (which & 4)
    {
      res.RemoveIndex(4 + 72);
      res.RemoveIndex(5 + 72);
      res.RemoveIndex(20 + 72);
      res.RemoveIndex(21 + 72);
    }
    
  }
  if(bc & bcBack)
  {
    if (which & 1)
    {
      res.RemoveIndex(14 );
      res.RemoveIndex(15 );
      res.RemoveIndex(30);
      res.RemoveIndex(31);
    }

    if (which & 2)
    {
      res.RemoveIndex(14 + 36);
      res.RemoveIndex(15 + 36);
      res.RemoveIndex(30 + 36);
      res.RemoveIndex(31 + 36);
    }

    if (which & 4)
    {
      res.RemoveIndex(14 + 72);
      res.RemoveIndex(15 + 72);
      res.RemoveIndex(30 + 72);
      res.RemoveIndex(31 + 72);
    }
    
  }
  if(bc & bcDown)
  {
    if (which & 1)
    {
      res.RemoveIndex(6 );
      res.RemoveIndex(11 );
      res.RemoveIndex(22);
      res.RemoveIndex(27);
    }

    if (which & 2)
    {
      res.RemoveIndex(6 + 36);
      res.RemoveIndex(11 + 36);
      res.RemoveIndex(22 + 36);
      res.RemoveIndex(27 + 36);
    }

    if (which & 4)
    {
      res.RemoveIndex(6 + 72);
      res.RemoveIndex(11+ 72);
      res.RemoveIndex(22 + 72);
      res.RemoveIndex(27 + 72);
    }
    
  }
  if(bc & bcUp)
  {
    if (which & 1)
    {
      res.RemoveIndex(8 );
      res.RemoveIndex(13);
      res.RemoveIndex(24);
      res.RemoveIndex(29);
    }

    if (which & 2)
    {
      res.RemoveIndex(8 + 36);
      res.RemoveIndex(13+ 36);
      res.RemoveIndex(24 + 36);
      res.RemoveIndex(29 + 36);
    }

    if (which & 4)
    {
      res.RemoveIndex(8+ 72);
      res.RemoveIndex(13 + 72);
      res.RemoveIndex(24 + 72);
      res.RemoveIndex(29 + 72);
    }
  }

  if (which == 0)
  { 
    if (bc & bcDown)
    {
      res.RemoveIndex(7     );	
      res.RemoveIndex(7 + 36);	
      res.RemoveIndex(7 + 72);

      res.RemoveIndex(9     );	
      res.RemoveIndex(9 + 36);	
      res.RemoveIndex(9 + 72);

      res.RemoveIndex(16     );	
      res.RemoveIndex(16 + 36);	
      res.RemoveIndex(16 + 72);

      res.RemoveIndex(23     );	
      res.RemoveIndex(23 + 36);	
      res.RemoveIndex(23 + 72);
      
    }
    if (bc & bcLeft)
    {
    }
    if (bc & bcFront)
    {
    }
  }
    
  
}

template<class Double>
void
ElastDiscretizator<Double>::
BoundCondSetRhs(SparceMacroMatrix_type const&mat,
		small_vector_type &rhs,
		small_vector_type const&bvalues,
		BoundaryConditions bc, int which)
{
  typedef SparceMacroMatrix_type SMT;
  typename SMT::index_type &index = SMT::GetIndex();
  //  std::cout << hex << (int)bc << std::endl;
  if (bc==bcNone)
    return;

  std::vector<int> nodes;
  if(bc & bcLeft)
  {
    if (which & 1)
    {
      nodes.push_back(0);
      nodes.push_back(1);
      nodes.push_back(2);
      nodes.push_back(3);
    }
	  	    
    if (which & 2)
    {
      nodes.push_back(0 + 36);
      nodes.push_back(1 + 36);
      nodes.push_back(2 + 36);
      nodes.push_back(3 + 36);
    }
	  	    
    if (which & 4)
    {
      nodes.push_back(0 + 72);
      nodes.push_back(1 + 72);
      nodes.push_back(2 + 72);
      nodes.push_back(3 + 72);
    }
    
  }
  if(bc & bcRight)
  {
    if (which & 1)
    {
      nodes.push_back(32);
      nodes.push_back(33);
      nodes.push_back(34);
      nodes.push_back(35);
    }
	  	    
    if (which & 2)
    {
      nodes.push_back(32 + 36);
      nodes.push_back(33 + 36);
      nodes.push_back(34 + 36);
      nodes.push_back(35 + 36);
    }
	  	    
    if (which & 4)
    {
      nodes.push_back(32 + 72);
      nodes.push_back(33 + 72);
      nodes.push_back(34 + 72);
      nodes.push_back(35 + 72);
    }
    
  }
  if(bc & bcFront)
  {
    if (which & 1)
    {
      nodes.push_back(4 );
      nodes.push_back(5 );
      nodes.push_back(20);
      nodes.push_back(21);
    }
	  	    
    if (which & 2)
    {
      nodes.push_back(4 + 36);
      nodes.push_back(5 + 36);
      nodes.push_back(20 + 36);
      nodes.push_back(21 + 36);
    }
	  	    
    if (which & 4)
    {
      nodes.push_back(4 + 72);
      nodes.push_back(5 + 72);
      nodes.push_back(20 + 72);
      nodes.push_back(21 + 72);
    }
    
  }
  if(bc & bcBack)
  {
    if (which & 1)
    {
      nodes.push_back(14 );
      nodes.push_back(15 );
      nodes.push_back(30);
      nodes.push_back(31);
    }
	  	    
    if (which & 2)
    {
      nodes.push_back(14 + 36);
      nodes.push_back(15 + 36);
      nodes.push_back(30 + 36);
      nodes.push_back(31 + 36);
    }
	  	    
    if (which & 4)
    {
      nodes.push_back(14 + 72);
      nodes.push_back(15 + 72);
      nodes.push_back(30 + 72);
      nodes.push_back(31 + 72);
    }
    
  }
  if(bc & bcDown)
  {
    if (which & 1)
    {
      nodes.push_back(6 );
      nodes.push_back(11 );
      nodes.push_back(22);
      nodes.push_back(27);
    }
	  	    
    if (which & 2)
    {
      nodes.push_back(6 + 36);
      nodes.push_back(11 + 36);
      nodes.push_back(22 + 36);
      nodes.push_back(27 + 36);
    }
	  	    
    if (which & 4)
    {
      nodes.push_back(6 + 72);
      nodes.push_back(11+ 72);
      nodes.push_back(22 + 72);
      nodes.push_back(27 + 72);
    }
    
  }
  if(bc & bcUp)
  {
    if (which & 1)
    {
      nodes.push_back(8 );
      nodes.push_back(13);
      nodes.push_back(24);
      nodes.push_back(29);
    }
	  	    
    if (which & 2)
    {
      nodes.push_back(8 + 36);
      nodes.push_back(13+ 36);
      nodes.push_back(24 + 36);
      nodes.push_back(29 + 36);
    }
	  	    
    if (which & 4)
    {
      nodes.push_back(8+ 72);
      nodes.push_back(13 + 72);
      nodes.push_back(24 + 72);
      nodes.push_back(29 + 72);
    }
    
  }

  if (which == 0)
  { 
    if (bc & bcDown)
    {
      nodes.push_back(7     );	
      nodes.push_back(7 + 36);	
      nodes.push_back(7 + 72);

      nodes.push_back(9     );	
      nodes.push_back(9 + 36);	
      nodes.push_back(9 + 72);

      nodes.push_back(16     );	
      nodes.push_back(16 + 36);	
      nodes.push_back(16 + 72);

      nodes.push_back(23     );	
      nodes.push_back(23 + 36);	
      nodes.push_back(23 + 72);
       
    }
  }
  
  
  if(!Options::GetOptions().SmallDirichlet2)
    for(unsigned int k = 0; k < nodes.size(); ++k)
    {
      int i = nodes[k];
      for(int j = 0; j < index[i][21];j++)
      {
	int col = index[i][j];
	Double val = bvalues[i];
	if (i!=col)
	{
	  rhs[col]-=mat[i][j]*val;
	}
      }
    }

  //  std::cout << nodes.size() <<  std::endl;
  for(unsigned int k = 0; k < nodes.size(); ++k)
  {
    int i = nodes[k];
    rhs[i] = bvalues[i];
  }
}



template<class Double>
SmallVector<Double, 3>
FuncRhs(const SmallVector<Double, 3> &p)
{
  Double const &x = p[0];
  Double const &y = p[1];
  Double const &z = p[2];

  SmallVector<Double, 3> res;

  // variant 5;
  res = Double();
  switch (Options::GetOptions().variant)
  {
  case 8:
  case 5:
  case 0:
  case 9: // dirichlet force
    break;
  case 4:
  // variant 4;
    res[0] = -(1.5*(6*x*y*z - 3*sin(x + y + z)) +
	       2.5*(6*x*y*z + 3*y*y*z +
		    3*y*z*z - cos(x + y + z) +
		    sin(x + y - z) - sin(x + y + z)));
    res[1] = -(1.5*(6*x*y*z - 3*cos(x + y + z)) +
	       2.5*(3*x*x*z + 6*x*y*z +
		    3*x*z*z - cos(x + y + z) +
		    sin(x + y - z) - sin(x + y + z)));
    res[2] = -(1.5*(6*x*y*z - 3*sin(x + y - z)) +
	       2.5*(3*x*x*y +
		    3*x*y*y + 6*x*y*z -
		    cos(x + y + z) - sin(x + y - z) -
		    sin(x + y + z)));
    break;
  case 3:
    res[0] = -24 * x;
    res[1] = -24 * y;
    res[2] = -24 * z;
    break;
  case 2:
    res[0] = - z * ( 3 * x * x * x + 7.5 * y * y * z + x * y * (24. * y + 15. * z));
    res[1] = - x * ( 3 * y * y * y + 7.5 * x * z * z + y * z * (24. * z + 15. * x));
    res[2] = - y * ( 3 * z * z * z + + 15 * x * y * z + + x * x * (7.5 * y + 24. *z));
    break;
  case 6:
    res[0] =   y*(3.*(x*x) - 10.*x*(-1.4472135954999585 + z)*
		  (-0.5527864045000423 + z) +
		  13.*(-1.2280149337378508 + z)*(-0.15660045087753388 + z)) +
      (-2.3051586491408833 + x)*(-0.36150801752578327 + x)*
      (3. - 3.*z)*z + (y*y)*
      (-2.5 - 3.*(-2.6666666666666665 + x)*x +
       12.999999999999996*z - 10.*x*z - 8.*(z*z));
    
    res[1] = x*(3.*(y*y) - 10.*y*(-1.4472135954999585 + z)*
		(-0.5527864045000423 + z) +
		13.*(-1.2280149337378508 + z)*(-0.15660045087753388 + z)) +
      (-2.3051586491408833 + y)*(-0.36150801752578327 + y)*
      (3. - 3.*z)*z + (x*x)*
      (-2.5 - 3.*(-2.6666666666666665 + y)*y +
       12.999999999999996*z - 10.*y*z - 8.*(z*z));
    
    res[2] =(3. - 3.*y)*y*(-2.305158649140883 + z)*
      (-0.3615080175257833 + z) +
      x*x*(-2.5 -
	   8.*(-1.625 + y)*y +
	         7.999999999999998*z - 10.*y*z -
	   3.*(z*z)) +
      x*(13.*(-1.2280149337378508 + y)*
	 (-0.15660045087753388 + y) -
	 10.*(-1.447213595499958 + y)*
	 (-0.5527864045000419 + y)*z +
	 3.*z*z);
    
    break;
  case 1:
    res[0] = -8;
    res[1] = -8;
    res[2] = -8;
    break;
  }

  return res;
}


//Dirichlet conditions
template<class Double>
SmallVector<Double, 3>
FuncSolution(const SmallVector<Double, 3> &p)
{
  Double const &x = p[0];
  Double const &y = p[1];
  Double const &z = p[2];

  
  
  SmallVector<Double, 3> res;

  // variant 5;
  res = Double();
  switch (Options::GetOptions().variant)
  {
  case 9: //
    res[2] = -z + z*0.9;
    res[0] = 0;
    res[1] = 0;
  case 10: // pilot
  case 0:
    break;
  case 8:
    break;
  case 5:
    res[0] = x + y + z;
    res[1] = x - y + z;
    res[2] = x + y - z;
    break;
  case 4:
    res[0] = x*x*x*y*z + sin(x + y + z);
    res[1] = x*y*y*y*z + cos(x + y + z);
    res[2] = x*y*z*z*z + sin(x + y - z);
    break;
  case 3:
    res[0] = x * x * x;
    res[1] = y * y * y;
    res[2] = z * z * z;
    break;
  case 2:
    res[0] = x * x * x * y * y * z;
    res[1] = x * y * y * y * z * z;
    res[2] = x * x * y * z * z * z;
    break;
  case 6:
    res[0] = -x*(1-x)*y*(1-y)*z*(1-z);
    res[1] = -x*(1-x)*y*(1-y)*z*(1-z);
    res[2] = -x*(1-x)*y*(1-y)*z*(1-z);
    break;
  case 1:
    res[0] = x*x;
    res[1] = y*y;
    res[2] = z*z;
    break;
  }
  return res;
}


//neuman conditions
template<class Double>
SmallMatrix<Double, 3, 3>
FuncNeuman(const SmallVector<Double, 3> &p) // returns sigma
{
  Double  x = p[0];
  Double  y = p[1];
  Double  z = p[2];

  x/= Options::GetOptions().DM1;
  y/= Options::GetOptions().DM2;
  z/= Options::GetOptions().DM3;

  SmallMatrix<Double, 3, 3> res;

  res = Double();
  switch (Options::GetOptions().variant)
  {
  case 8:
    
    if ((x>=(15/31.))&&(x <=(16/31.))&&(y>=(13/31.))&&(y<=(14/31.)))
    {
      res[1][1] = 150e3;
      res[2][2] =  -2000e3;
    }
    if ((x>=(15/31.))&&(x <=(16/31.))&&(y>=(17/31.))&&(y<=(18/31.)))
    {
      res[1][1] = 150e3;
      res[2][2] =  -4000e3;
    }
    return res;
    break;
  case 0:
    res[2][2] = - Options::GetOptions().force;
    return res;
  case 10:// pilot
    if ((x>=(3/8.))&&(x <=(5/8.))&&(y>=(3/8.))&&(y<=(5/8.)))
    {
      res[2][2] = - Options::GetOptions().force;
      return res;
      //      using namespace std;
      //      cout << res[2][2] << endl;
    }
    break;
  case 5:
    res[0][0] = 2.;
    res[0][1] = 3.;
    res[0][2] = 3.;
    res[1][0] = res[0][1];
    res[1][1] = -4.;
    res[1][2] = 3.;
    res[2][0] = res[0][2];
    res[2][1] = res[1][2];
    res[2][2] = -4.;
    break;
  case 4:
    res[0][0] = x*y*z*(12.*x + 3.*y + 3.*z) - 1.*cos(x + y - z) + 4.*cos(x + y + z) -  1.*sin(x + y + z);
    res[0][1] =  1.5*((x *x * x + y*y*y)*z +  cos(x + y + z) - sin(x + y + z));
    res[0][2] =  1.5*(y*(x * x*x + z*z*z) + 2*cos(x + y)*cos(z));
    res[1][0] = res[0][1];
    res[1][1] =  x*y*z*(3.*x + 12.*y + 3.*z) -  1.*cos(x + y - z) + cos(x + y + z) - 4.*sin(x + y + z);
    res[1][2] =  1.5*(x*(y*y*y + z*z*z) + cos(x + y - z) - sin(x + y + z));
    res[2][0] = res[0][2];
    res[2][1] = res[1][2];
    res[2][2] = x*y*z*(3.*x + 3.*y + 12.*z) - 4.*cos(x + y - z) + cos(x + y + z) - 1.*sin(x + y + z);
    break;
  case 1:
    res[0][0] = 8 * x + 2 * y + 2 * z;
    res[1][1] = 2 * x + 8 * y + 2 * z;
    res[2][2] = 2 * x + 2 * y + 8 * z;
    break;
  case 9:
    break;
  }

  return res;
}


#endif
