#ifndef __elastmatrix_h__
#define __elastmatrix_h__
#include "useful.h"
#include "elastgen.h"
#include "small.h"
#include "../bone/mirorer.h"
#include <map>
#include "solver.h"

EXPORT
template<class Double>
class MacroElementDescriptor
{
 public:
  MacroElementDescriptor(){}
  MacroElementDescriptor(const MacroElementDescriptor&a)
  {
    lambda = a.lambda;
    mu = a.mu;
    mat = a.mat;
    h = a.h;
  }
  typedef typename ElastDiscretizator<Double>::SparceMacroMatrix_type SparceMacroMatrix_type;
  SmallVector<Double,3> h;
  Double lambda, mu;
  void SetCoefficients(Double lambda_, Double mu_)
  {
    lambda = lambda_;
    mu = mu_;
    //    std::cout << E <<" " << ni << " " << lambda << " " << mu << std::endl;
  }

  void SetCoefficients(int i)
  {
    Options &opt = Options::GetOptions();
    lambda = opt.lambda[i];
    mu = opt.mu[i];
  }
  static SparceMacroMatrix_type m1, m2;

  SparceMacroMatrix_type mat;
  
};
  
EXPORT
template<class Double, class Int>
class ElementwiseElastityMatrix
{
public:
  typedef ElastDiscretizator<Double> Discretizator;
  typedef typename Discretizator::small_vector_type small_vector_type;
  typedef typename Discretizator::SparceMacroMatrix_type SparceMacroMatrix_type;
  
  SmallVector<Double,3> h;
  Int N1, N2, N3;
  Int DoubleLines;
  Int StartLine;
  bool FirstFull;
  Discretizator *pdis;
  mutable std::map<int, MacroElementDescriptor<Double> > lookupmap;

  vector3d<unsigned int> coeffs;
  // Bitmask   1f ffff - material
  // Bitmask   E0 ffff - Dirichlet boundary conditions to which displacements
  // Bitmask 3f00 0000 - Dirchlet boundary conditions to which side;
  void Sysgen(Discretizator * pdis, BoundaryConditions bc, BoundaryConditions nc);

  void  SetDirichlet(bool dosmall, BoundaryConditions bc);
  MacroElementDescriptor<Double> &GetAt(Int i, Int j, Int k);  

  BoundaryConditions GetBCAt(Int i, Int j, Int k);
  unsigned int  GetDispAt(Int i, Int j, Int k);
  SparceMacroMatrix_type& GetElMatrixAt(Int i, Int j, Int k)const;
  SparceMacroMatrix_type& GetElMatrixAtNBC(Int i, Int j, Int k)const;
  
  void GenRhsAt(small_vector_type&rhs, small_vector_type const&, 
		Int i, Int j, Int k, Int realk,
		BoundaryConditions bc, BoundaryConditions nc);
  void GenRhsAtd(small_vector_type&rhs, small_vector_type const&, 
		Int i, Int j, Int k, Int realk,
		BoundaryConditions bc, BoundaryConditions nc);
  
};


EXPORT
template<class Vector>
class ElastMatrix:public MatrixBase<Vector>
{
  
public:
  typedef typename Myf<typename Vector::base_type>::base_type Double;
  typedef typename ElastDiscretizator<Double>::SparceMacroMatrix_type SparceMacroMatrix_type;
  typedef typename Vector::Int Int;
  typedef Vector bvector_type;
private:
public:
  MPI::Intracomm &comm;
  Int N1, N2, N3, N3R;
  Int DoubleLines;
  bool FirstFull;
  bool FirstFirst;
  bool LastFull;
  bool LastLast;

  Int Lines;
  Int StartLine;
  int rank, size;
  ElementwiseElastityMatrix<Double, Int> elmentw;

  void MultiplyLines(Vector &res, Vector  &src, int from, int to)const;
  ElastMatrix(MPI::Intracomm &c);
  virtual ~ElastMatrix();
  virtual void Multiply(Vector &res, Vector  &src)const;
  virtual std::string GetName()const
  {
    return "Elast";
  }
  virtual void Sysgen(Vector &rhs, Vector&dissol);

};

EXPORT
template<class Vector>
class ElastMatrixPBC:public ElastMatrix<Vector>
{
 public:
  ElastMatrixPBC(MPI::Intracomm &c);
  ~ElastMatrixPBC();
  virtual void Multiply(Vector &res, Vector  &src)const;
  void AddBoundarys(Vector &res, bool add)const;
  virtual std::string GetName()const
  {
    return "ElastPBC";
  }
  virtual void Sysgen(Vector &rhs, Vector&dissol);

};

EXPORT
template<class Vector, class MVector=Vector>
class JPrecon:public PreconBase<Vector>
{
  ElastMatrix<MVector> *mat;
  void PreconInit(Vector &src)const;
  mutable  Vector J;
  mutable bool firsttime;
public:
  JPrecon(ElastMatrix<MVector> *emat):mat(emat),firsttime(true){}
  virtual void PreconSolve(Vector &res, Vector &src)const;
  virtual std::string GetName()const
  {
    return "Jacoby";
  }
};


#endif
#ifdef INCLUSION
#include "elastmatrix.hpp"
#endif
