#ifndef __elastmatrix_hpp__
#define __elastmatrix_hpp__

#include "elastmatrix.h"
#include "index.h"
#include "functors.h"

// ElementwiseElastistyMatrix //////////////////////////////////////////
 
template<class Double, class Int>
void
ElementwiseElastityMatrix<Double, Int>::
Sysgen(Discretizator *pdis, BoundaryConditions bc, BoundaryConditions nc)
{
  
  Discretizator &dis = *pdis;
  this->pdis = pdis;
  SparceMacroMatrix_type &gsm1 = MacroElementDescriptor<Double>::m1,
    &gsm2 = MacroElementDescriptor<Double>::m2, tmp;

  h[0] = Options::GetOptions().DM1/N1;
  h[1] = Options::GetOptions().DM2/N2;
  h[2] = Options::GetOptions().DM3/Options::GetOptions().N3;
  dis.GenElastElMat(gsm1, gsm2, h);
  //  std::cout << h[0] << " " <<h[1] << " " << h[2] << std::endl;

  Options &opt = Options::GetOptions();
  
  for(Options::cit_type it = opt.lambda.begin(); it !=opt.lambda.end(); ++it)
  {
    int i = it->first;
    //      std::cout << "Material " << i << " " << opt.E[i] << " " << opt.ni[i] << std::endl;
    MacroElementDescriptor<Double> &first = lookupmap[i];
    first.h = h;
    first.SetCoefficients(i);
    tmp = gsm1;
    tmp *= first.mu * 2.;
    first.mat = tmp;
    tmp = gsm2;
    tmp *= first.lambda;
    first.mat += tmp;
    for(int i = 0; i < 108;++i)
      first.mat[i][0]+= Options::GetOptions().Reg * first.mat[i][0];
    //      std::cout << first.mu << " " << first.lambda<< std::endl;

    //      std::cout << std::setw(12) << first.lambda << " " << std::setw(12) << first.mu << std::endl;
    int bcarr[] = {
      bcLeft,
      bcFront,
      bcDown,
      bcRight,
      bcUp,
      bcBack,
      bcRight | bcUp,
      bcRight | bcDown,
      bcRight | bcFront,
      bcRight | bcBack,
      bcRight | bcUp | bcFront,
      bcRight | bcUp | bcBack,
      bcRight | bcDown |bcFront,
      bcRight | bcDown |bcBack,
      bcLeft | bcUp,
      bcLeft | bcDown,
      bcLeft | bcFront,
      bcLeft | bcBack,
      bcLeft | bcUp | bcFront,
      bcLeft | bcUp | bcBack,
      bcLeft | bcDown |bcFront,
      bcLeft | bcDown |bcBack,
      bcDown |bcFront,
      bcDown |bcBack,
      bcUp |bcFront,
      bcUp |bcBack,
      //	bcDown |bcFront |bcRight | bcUp | bcBack,
      bcLeft | bcDown |bcFront |bcRight | bcUp | bcBack
    };

    if (!Options::GetOptions().SmallDirichlet)
    {
      for(unsigned j = 0; j < sizeof(bcarr)/sizeof(BoundaryConditions); j++)
      {
	int curbc = bcarr[j];
	if (!(((curbc|bc)==bc)&&((curbc&bc)==curbc)))
	  continue;
	int index = (curbc <<24) + i+ (7<<21);
	lookupmap[index].SetCoefficients(first.lambda, first.mu);
	// dis.BoundCondSet(lookupmap[index].mat,
	// 		 first.mat, BoundaryConditions(curbc), 1); //only x
	// dis.BoundCondSet(lookupmap[index].mat,
	// 		 first.mat, BoundaryConditions(curbc), 2); //only y
	// dis.BoundCondSet(lookupmap[index].mat,
	// 		 first.mat, BoundaryConditions(curbc), 4); // only z
	dis.BoundCondSet(lookupmap[index].mat,
			 first.mat, BoundaryConditions(curbc), 7); // only z
      }
    }else // A very little Dirichlet conditions
    {
      {
	int curbc = 	bcLeft | bcDown |bcFront;
	int index = (curbc <<24) + i+ (0<<21);
	lookupmap[index].SetCoefficients(first.lambda, first.mu);
	dis.BoundCondSet(lookupmap[index].mat,
			 first.mat, BoundaryConditions(curbc), 0);
      }
      // 	for(unsigned j = 0; j < 3; ++j)
      // 	{
      // 	  int curbc = bcarr[j];
      // 	  int index = (curbc <<24) + i+ (1<<(21+j));
      // 	  lookupmap[index].SetCoefficients(first.lambda, first.mu);
      // 	  dis.BoundCondSet(lookupmap[index].mat,
      // 			   first.mat, BoundaryConditions(curbc), 1 << j);
      // 	}
    }
  }

  coeffs.SetSize(N1/2, N2/2, DoubleLines);
  for(unsigned int i = 0; i < coeffs.sizes[0]; i++)
    for(unsigned  int j = 0; j < coeffs.sizes[1]; j++)
      for(unsigned  int k = 0; k < coeffs.sizes[2];k++)
	coeffs(i,j,k) = 255;
  if (Options::GetOptions().usebone)
  {
    std::ostringstream os;
    //    os << Options::GetOptions().scratchpath << "/"<< N3/2 << ".bonec";
    os << Options::GetOptions().bone_fname;
    try{
      coeffs.ReadBytes(os.str(), StartLine/2);
    }catch(std::exception&)
    {
      std::cerr << "could not open file \"" << os.str() << "\""<< std::endl;
      throw;
    }
    if (Options::GetOptions().variant!=8) //no plates for pilot
      if(Options::GetOptions().useplates)
	for(unsigned int i = 0; i < coeffs.sizes[0]; i++)
	  for(unsigned int j = 0; j < coeffs.sizes[1]; j++)
	    if (coeffs.sizes[2]>0)
	    {
	      int k = 0;
	      Int realk = 2*k + StartLine - !FirstFull;
	      if(realk==0)
		coeffs(i,j,k) = 1;
	      k = coeffs.sizes[2]-1;
	      realk = 2*k + StartLine - !FirstFull ;
	      if((realk+2)==N3)
		coeffs(i,j,k) = 1;
	    }
  }
  //boundary conditions

  SetDirichlet(Options::GetOptions().SmallDirichlet, bc);
  
}

template<class Double, class Int>
void
ElementwiseElastityMatrix<Double, Int>::
SetDirichlet(bool dosmall, BoundaryConditions bc)
{
  if (!dosmall)
  {
    for(unsigned int i = 0; i < coeffs.sizes[0]; ++i)
      for(unsigned int j = 0; j < coeffs.sizes[1];++j)
      {
      
	if ((bc & bcUp)&&(coeffs.sizes[2])&&(StartLine/2+coeffs.sizes[2]==static_cast<size_t>(N3)/2))
	  coeffs(i,j,coeffs.sizes[2]-1)|= (unsigned(bcUp)<<24) + (7 << 21);
	if ((bc & bcDown)&&(coeffs.sizes[2])&&((StartLine/2)==0))
	  coeffs(i,j,0)|= (unsigned(bcDown)<<24) + (7 << 21) ;
      }
    for(unsigned int i = 0; i < coeffs.sizes[0]; ++i)
      for(unsigned int k = 0; k < coeffs.sizes[2];++k)
      {
	if (bc & bcBack)
	  coeffs(i,coeffs.sizes[1]-1,k)|= (unsigned(bcBack)<<24) + (7 << 21);
	if (bc & bcFront)
	  coeffs(i,0,k)|= (unsigned(bcFront)<<24) + (7 << 21);
      }
    for(unsigned int j = 0; j < coeffs.sizes[1];++j)
      for(unsigned int k = 0; k < coeffs.sizes[2];++k)
      {
	if (bc & bcRight)
	  coeffs(coeffs.sizes[0]-1,j,k)|= (unsigned(bcRight)<<24) + (7 << 21);
	if (bc & bcLeft)
	  coeffs(0,j,k)|= (unsigned(bcLeft)<<24) + (7 << 21);
      }
  }else // small dirichlet
  {
    for(unsigned int i = 0; i < coeffs.sizes[0]; ++i)
      for(unsigned int j = 0; j < coeffs.sizes[1];++j)
	for(unsigned int k = 0; k < coeffs.sizes[2];++k)
	  coeffs(i,j,k) &= ~ ((0x3f << 24)| (0x7 << 21));
    if ((coeffs.sizes[2])&&((StartLine/2)==0))
    {
      coeffs(0,0,0)|= (unsigned(bcDown | bcLeft | bcFront)<<24) + (0 << 21) ;
    }
    //     if ((coeffs.sizes[2])&&(StartLine/2+coeffs.sizes[2]==static_cast<size_t>(N3)/2))
    //       coeffs(0,0,coeffs.sizes[2]-1)|= (unsigned(bcUp)<<24) + (7 << 21) ; //the same up
  }
}


template<class Double, class Int>
void
ElementwiseElastityMatrix<Double, Int>::
GenRhsAt(small_vector_type&small_rhs, small_vector_type const &discretizedU, Int i, Int j, Int k, Int realk, BoundaryConditions bc, BoundaryConditions nc)
{
  Discretizator &dis = *pdis;
  SmallMatrix<Double, 8, 3> elpoints;
  elpoints[0][0] = h[0]*(Double)(i);
  elpoints[0][1] = h[1]*(Double)j;
  elpoints[0][2] = h[2]*(Double)(realk);

  elpoints[1][0] = h[0]*(Double)(i);
  elpoints[1][1] = h[1]*(Double)j;
  elpoints[1][2] = h[2]*(Double)(realk+1);
	
  elpoints[2][0] = h[0]*(Double)(i);
  elpoints[2][1] = h[1]*(Double)(j+1);
  elpoints[2][2] = h[2]*(Double)(realk);
	
  elpoints[3][0] = h[0]*(Double)(i);
  elpoints[3][1] = h[1]*(Double)(j+1);
  elpoints[3][2] = h[2]*(Double)(realk+1);

  elpoints[4][0] = h[0]*(Double)(i+1);
  elpoints[4][1] = h[1]*(Double)j;
  elpoints[4][2] = h[2]*(Double)(realk);
	
  elpoints[5][0] = h[0]*(Double)(i+1);
  elpoints[5][1] = h[1]*(Double)j;
  elpoints[5][2] = h[2]*(Double)(realk+1);
	
  elpoints[6][0] = h[0]*(Double)(i+1);
  elpoints[6][1] = h[1]*(Double)(j+1);
  elpoints[6][2] = h[2]*(Double)(realk);
	
  elpoints[7][0] = h[0]*(Double)(i+1);
  elpoints[7][1] = h[1]*(Double)(j+1);
  elpoints[7][2] = h[2]*(Double)(realk+1);
  

  Double lambda = GetAt(i/2, j/2, k/2).lambda;
  Double mu  = GetAt(i/2, j/2, k/2).mu;
  small_rhs=0.;
  dis.GenRhs(small_rhs, elpoints, h, lambda, mu);
	
  // neuman conditions

  //  std::cout << (int)nc << std::endl;
  if ((nc & bcUp)&&
      (coeffs.sizes[2])&&
      (StartLine/2+coeffs.sizes[2]==static_cast<size_t>(N3)/2)&&
      ((static_cast<size_t>(k)+2)==2*coeffs.sizes[2]))
    dis.NoimanBoundCondSet(small_rhs, elpoints[0], h, BUp);
  if ((nc & bcDown)&&(coeffs.sizes[2])&&((StartLine/2)==0)&&(k==0))
    dis.NoimanBoundCondSet(small_rhs, elpoints[0], h, BDown);
	
	
  if (((j+2) == N2)&&(nc & bcBack))
    dis.NoimanBoundCondSet(small_rhs, elpoints[0], h, BBack);
	
  if ((j==0)&&(nc & bcFront))
    dis.NoimanBoundCondSet(small_rhs, elpoints[0], h, BFront);
	
	
  if (((i+2) == N1)&&(nc & bcRight))
    dis.NoimanBoundCondSet(small_rhs, elpoints[0], h, BRight);
	
  if ((i==0)&&(nc & bcLeft))
    dis.NoimanBoundCondSet(small_rhs, elpoints[0], h, BLeft);


  
  //dirichlet
//   SparceMacroMatrix_type  &A = GetElMatrixAtNBC(i/2, j/2, k/2);
//   dis.BoundCondSetRhs(A, small_rhs, discretizedU,
// 		      GetBCAt(i/2, j/2, k/2),
// 		      GetDispAt(i/2, j/2, k/2)
// 		      );
		
  
  
}
template<class Double, class Int>
void
ElementwiseElastityMatrix<Double, Int>::
GenRhsAtd(small_vector_type&small_rhs, small_vector_type const &discretizedU, Int i, Int j, Int k, Int realk, BoundaryConditions bc, BoundaryConditions nc)
{
  //dirichlet
  Discretizator &dis = *pdis;
  SparceMacroMatrix_type  &A = GetElMatrixAtNBC(i/2, j/2, k/2);
  dis.BoundCondSetRhs(A, small_rhs, discretizedU,
		      GetBCAt(i/2, j/2, k/2),
		      GetDispAt(i/2, j/2, k/2)
		      );

}

template<class Double, class Int>
typename ElementwiseElastityMatrix<Double, Int>::SparceMacroMatrix_type& 
ElementwiseElastityMatrix<Double, Int>::
GetElMatrixAt(Int i, Int j, Int k)const
{
//   std::cout <<
//     std::setw(3) << i <<
//     std::setw(3) << j <<
//     std::setw(3) << k <<
//     std::hex << coeffs(i,j,k)
// 		 << std::endl;
  return lookupmap[coeffs(i,j,k) &0x3fffffffu].mat;
}

template<class Double, class Int>
typename ElementwiseElastityMatrix<Double, Int>::SparceMacroMatrix_type& 
ElementwiseElastityMatrix<Double, Int>::
GetElMatrixAtNBC(Int i, Int j, Int k)const
{
//   std::cout <<
//     std::setw(3) << i <<
//     std::setw(3) << j <<
//     std::setw(3) << k <<
//     std::hex << coeffs(i,j,k)
// 		 << std::endl;
  return lookupmap[coeffs(i,j,k) &0x001fffffu].mat;
}

template<class Double, class Int>
BoundaryConditions
ElementwiseElastityMatrix<Double, Int>::
GetBCAt(Int i, Int j, Int k)
{
//   std::cout <<
//     std::setw(3) << i <<
//     std::setw(3) << j <<
//     std::setw(3) << k <<
//     std::hex << coeffs(i,j,k)
// 		 << std::endl;
  return BoundaryConditions(0x3f&(coeffs(i,j,k) >>24));
}

template<class Double, class Int>
unsigned int
ElementwiseElastityMatrix<Double, Int>::
GetDispAt(Int i, Int j, Int k)
{
//   std::cout <<
//     std::setw(3) << i <<
//     std::setw(3) << j <<
//     std::setw(3) << k <<
//     std::hex << coeffs(i,j,k)
// 		 << std::endl;
  return 7&(coeffs(i,j,k) >>21);
}
template<class Double, class Int>
MacroElementDescriptor<Double>&
ElementwiseElastityMatrix<Double, Int>::
GetAt(Int i, Int j, Int k)
{
//   std::cout <<
//     std::setw(3) << i <<
//     std::setw(3) << j <<
//     std::setw(3) << k <<
//     std::hex << coeffs(i,j,k)
// 		 << std::endl;
  return lookupmap[coeffs(i,j,k)];
}

  




// Elast matrix////////////////////////////////////////////////

template<class Vector>
ElastMatrix<Vector>::
ElastMatrix(MPI::Intracomm &c)
     :comm(c),
     N1(Options::GetOptions().N1),
     N2(Options::GetOptions().N2),
     N3R(Options::GetOptions().N3),
     rank(c.Get_rank()), size(c.Get_size())
{
  StartLine = rank*(N3R+1)/size;
  Lines = (rank+1)*(N3R+1)/size - StartLine;

  N3 = Lines;
  
  DoubleLines = Lines/2;
  
  FirstFirst = rank==0;
  LastLast = rank == (size-1);
  FirstFull = ((StartLine &1) == 0);
  
  DoubleLines = (Lines&1)? (Lines+1)/2-Int(FirstFull) : (Lines)/2;

}


template<class Vector>
ElastMatrix<Vector>::
~ElastMatrix()
{
}

template<class Vector>
void
ElastMatrix<Vector>::
Sysgen(Vector &rhs, Vector&dissol)
{
  BoundaryConditions nc =  Options::GetOptions().nc;
  DiscretizationType distype = Options::GetOptions().distype;
  if (distype == RV)
    distype = MV;
  if (distype == RP)
    distype = MP;
  BoundaryConditions bc = Options::GetOptions().bc;
  
  typedef ElastDiscretizator<Double> Discretizator;
  ElastDiscretizator<Double> *pdis;
  //  Discretizator &dis = *pdis;
  typedef typename Discretizator::SparceMacroMatrix_type SparceMacroMatrix_type;
  typedef typename Discretizator::small_vector_type small_vector_type;


  pdis = new Discretizator(distype);


  
  SmallVector<Double, 3> h;
  h[0] = Options::GetOptions().DM1/N1;
  h[1] = Options::GetOptions().DM2/N2;
  h[2] = Options::GetOptions().DM3/N3R;

  

  elmentw.DoubleLines = DoubleLines;
  elmentw.N1 = N1;
  elmentw.N2 = N2;
  elmentw.N3 = N3R;
  elmentw.StartLine = StartLine;
  elmentw.FirstFull = FirstFull;
  elmentw.Sysgen(pdis, bc, nc);

  
  //generating rhs
  ClearVector(rhs, 0);
  ClearVector(dissol,0);
  
  SmallVector<Double, 3> p, pl, pd, pf;
  Int NumberOfLines = Lines+1;
  Int SL = StartLine-1;
  Int SO = -1;
  if (rank == 0)
  {
    SL++;
    SO++;
    NumberOfLines--;
  }
  Index3dBase<Int> ind(N1, N2, Lines);
  for(int i = 0; i <= N1; i++)
  {
    //    cout << "i" << i << endl;
    for(int j = 0; j <= N2; j++)
      for(int k = 0; k < NumberOfLines; k++)
      {
	ind.SetPos(i, j, k + SO);
	
	p[0] = (Double)i *h[0];
	p[1] = (Double)j *h[1];
	p[2] = (Double)(k + SL)*h[2];
	pl = p;
	pd = p;
	pf = p;
	pl[1] += h[1]*0.5;
	pl[2] += h[2]*0.5;
	
	pf[0] += h[0]*0.5;
	pf[2] += h[2]*0.5;
	
	pd[0] += h[0]*0.5;
	pd[1] += h[1]*0.5;

	//	std::cout << i << " " << j << " " << k+SL << " " << pl[0] <<  " " << pl[1] << " " << pl[2] << endl;
	
 	dissol[ind][LeftEl] = FuncSolution(pl);
 	dissol[ind][DownEl] = FuncSolution(pd);
 	dissol[ind][FrontEl] = FuncSolution(pf);
      }
  }
  dissol.StartSendRecv();
  dissol.FinishSendRecv();


  


  for(Int i = 0; i < N1; i+=2)
    for(Int j = 0; j < N2; j+=2)
    {
      for(Int k = 0; k < DoubleLines*2; k+=2)
      {

	int realk = k + StartLine - 1+Int(FirstFull);
	Index3dBase<Int> ind(N1, N2, Lines);
	ind.SetPos(i, j, k - 1 + Int(FirstFull));
	
	small_vector_type small_rhs;
	small_vector_type discretizedU;

	BigToSmallCpy<bvector_type, small_vector_type> b2scopy1(dissol, discretizedU);
	DoForMacroElement(b2scopy1, ind);
	elmentw.GenRhsAt(small_rhs, discretizedU, i, j, k, realk, bc, nc);
	
	SmallToBigAdd<bvector_type, small_vector_type> s2badd(rhs, small_rhs);
	DoForMacroElement(s2badd, ind);
      }
    }
  rhs.SendDown = (StartLine & 1) != 0;
  rhs.SendUp   = (((StartLine+Lines-1)&1) != 0);
//   rhs.StartSendRecvAdd();
//   rhs.FinishSendRecvAdd();

  // dirichlet conditions
    
//   rhs.StartSendRecv();
//   rhs.FinishSendRecv();
  for(Int i = 0; i < N1; i+=2)
    for(Int j = 0; j < N2; j+=2)
    {
      for(Int k = 0; k < DoubleLines*2; k+=2)
      {
  	int realk = k + StartLine - 1+Int(FirstFull);
  	Index3dBase<Int> ind(N1, N2, Lines);
  	ind.SetPos(i, j, k - 1 + Int(FirstFull));

  	small_vector_type small_rhs;
  	small_vector_type discretizedU;

  	BigToSmallCpy<bvector_type, small_vector_type> b2scopy1(dissol, discretizedU);
  	BigToSmallCpy<bvector_type, small_vector_type> b2scopy2(rhs, small_rhs);
	
  	DoForMacroElement(b2scopy1, ind);
  	DoForMacroElement(b2scopy2, ind);
  	elmentw.GenRhsAtd(small_rhs, discretizedU, i, j, k, realk, bc, nc);
	
  	SmallToBigCpy<bvector_type, small_vector_type> s2bcpy(rhs, small_rhs);
	DoForMacroElement(s2bcpy, ind);
      }
    }
  rhs.StartSendRecvAdd();
  rhs.FinishSendRecvAdd(true);//false); //no adition, just replace, but not inplace


  if (Options::GetOptions().SmallDirichlet2)
    elmentw.SetDirichlet(true, bc);
  delete pdis;
}

template<class Vector>
void
ElastMatrix<Vector>::
MultiplyLines(Vector &res, Vector  &src, int from, int to)const
{
  typedef typename SparceMacroMatrix_type:: vector_type vector_type;
  vector_type small_res, small_src;
  typedef typename Vector::Int Int;
#pragma omp parallel for shared(from, to) private(small_res, small_src)
  for(Int i = 0; i < N1; i+=2)
    for(Int j = 0; j < N2; j+=2)
    {
      for(Int k = from; k < to; k+=2)
      {
	Index3dBase<Int> ind(N1, N2, Lines);
	ind.SetPos(i, j, k - 1 +Int(FirstFull));
	//	std::cout << elmentw.coeffs(i/2, j/2, k/2) << endl;
	SparceMacroMatrix_type const  &A = elmentw.GetElMatrixAt(i/2, j/2, k/2);
	  
	BigToSmallCpy<bvector_type, vector_type> b2scopy1(src, small_src);
	BigToSmallCpy<bvector_type, vector_type> b2scopy2(res, small_res);
	SmallToBigAdd<bvector_type, vector_type> s2badd(res, small_res);
	  
	// src ->small_src;
	DoForMacroElement(b2scopy1, ind);
	// res -> small_res
#pragma omp critical (crit1)
	{
	  DoForMacroElement(b2scopy2, ind);
	}
	  
	A.MulVec(small_res, small_src);
	// small_res -> res
#pragma omp critical (crit1)
	{
	  DoForMacroElement(s2badd, ind);
	}
      }
    }
}

template<class Vector>
void
ElastMatrix<Vector>::
Multiply(Vector &res, Vector  &src)const
{
  TimeCalc1("MatVec");
  ClearVector(res,0);
  res.SendDown = (StartLine & 1) != 0;
  res.SendUp   = (((StartLine+N3-1)&1) != 0);
  src.SendDown = (StartLine & 1) != 0;
  src.SendUp   = (((StartLine+N3-1)&1) != 0);

  src.StartSendRecv();
  src.FinishSendRecv();
  MultiplyLines(res, src, 0, DoubleLines*2);
  res.StartSendRecvAdd();
  res.FinishSendRecvAdd();
}


template<class Vector>
ElastMatrixPBC<Vector>::
ElastMatrixPBC(MPI::Intracomm &c):ElastMatrix<Vector>(c)
{
}

template<class Vector>
ElastMatrixPBC<Vector>::
~ElastMatrixPBC()
{
}
template<class Vector>
void
ElastMatrixPBC<Vector>::
Multiply(Vector &res, Vector  &src)const
{
  src.AddBoundarys(0);
  ElastMatrix<Vector>::Multiply(res, src);
  res.AddBoundarys(1);
  src.AddBoundarys(2);
}
template<class Vector>
void
ElastMatrixPBC<Vector>::
Sysgen(Vector &rhs, Vector&dissol)
{
  ElastMatrix<Vector>::Sysgen(rhs, dissol);
  rhs.AddBoundarys( 1);
  //  rhs.ClearZero();
}



template<class Vector, class MVector>
void
JPrecon<Vector, MVector>::
PreconSolve(Vector &res, Vector &src)const
{
  PreconInit(src);
  VecVecOper<VecCpy>(res, src);
  VecVecOper<VecMul>(res, J);        //res
  
}
template<class Vector, class MVector>
void
JPrecon<Vector, MVector>::
PreconInit(Vector &src)const
{
  if (firsttime)
    firsttime=false;
  else
    return;
  J.SetSize(src.N1, src.N2, src.N3, src.N3R);
  J.SetComm(*src.comm);
  
  typedef typename ElastMatrix<MVector>::SparceMacroMatrix_type:: vector_type vector_type;
  typedef typename ElastMatrix<MVector>::SparceMacroMatrix_type SparceMacroMatrix_type;
  typedef typename MVector::int_type Int;
  ClearVector(J,0);

  vector_type small_src;
  for(Int i = 0; i < mat->N1; i+=2)
    for(Int j = 0; j < mat->N2; j+=2)
      for(Int k = 0; k < (mat->DoubleLines*2); k+=2)
      {
	Index3dBase<Int> ind(mat->N1, mat->N2, mat->Lines);
	ind.SetPos(i, j, k - 1 +Int(mat->FirstFull));
	//	std::cout << i/2 <<" " <<   j/2 << " " <<  k/2 << endl;
	SparceMacroMatrix_type  &A = mat->elmentw.GetElMatrixAt(i/2,j/2, k/2);
	BigToSmallCpy<Vector, vector_type> b2scopy(J, small_src);
	SmallToBigCpy<Vector, vector_type> s2bcopy(J, small_src);
	
	// J ->small_src;
	DoForMacroElement(b2scopy, ind);
	
	//jacoby itself

	for(int b = 0; b <108; b++)
	{
	  small_src[b] +=  A[b][0];
	  if (isnan(small_src[b]))
	    std::cout << i <<" "  << j << " " << k << " " <<b << std::endl;
	  
	}
	// small_src -> j
	DoForMacroElement(s2bcopy, ind);
      }

  J.SendDown = (mat->StartLine & 1) != 0;
  J.SendUp   = (((mat->StartLine + mat->N3-1)&1) != 0);
  J.StartSendRecvAdd();
  J.FinishSendRecvAdd(true);//add!
  if (Options::GetOptions().PBC)
    J.AddBoundarys(1); // add sides
  VecOper<VecRec>(J); // J = 1/J
  if (Options::GetOptions().PBC)
    J.AddBoundarys(2); // zero one part
}



template<class Double>
typename MacroElementDescriptor<Double>::SparceMacroMatrix_type MacroElementDescriptor<Double>::m1;

template<class Double>
typename MacroElementDescriptor<Double>::SparceMacroMatrix_type MacroElementDescriptor<Double>::m2;




#endif
