#ifndef __functors_hpp__
#define __functors_hpp__

#include <iostream>
#include "visitors.h"

#include "functors.h"
#include "useful.h"

#ifdef _OPENMP
#include <omp.h>
#endif

template<
  template<class, class> class Functor, 
  class VecType1,
  class VecType2> 
void 
VecVecOper(VecType1 &vec1, VecType2 &vec2)
{
  TimeCalc1("VecVec");
  Functor<VecType1, VecType2> func(vec1, vec2);
  Index3dBase<typename VecType1::int_type>
    ind(vec1.N1, vec1.N2, vec1.N3);
  Execute<ForAll>(ind, func, vec1.size, vec1.rank);
}
template<
  template<class> class Functor, 
  class VecType>
void 
VecOper(VecType &vec1)
{
  //  TimeCalc1("VecVec");
  Functor<VecType> func(vec1);
  Index3dBase<typename VecType::int_type>
    ind(vec1.N1, vec1.N2, vec1.N3);
  Execute<ForAll>(ind, func, vec1.size, vec1.rank);
}

template<
  template<class, class> class Functor, 
  class VecType1,
  class VecType2,
  class T> 
void 
VecVecOper(VecType1 &vec1, VecType2 &vec2,  T val)
{
  TimeCalc1("VecVecAlfa");
  Functor<VecType1, VecType2> func(vec1, vec2, val);
  Index3dBase<typename VecType1::int_type>
    ind(vec1.N1, vec1.N2, vec1.N3);
  Execute<ForAll>(ind, func, vec1.size, vec2.rank);
}

#
template<
  template<class, class> class Functor, 
  class VecType1,
  class VecType2> 
typename Myf<typename VecType1::base_type>::base_type
VecVecReduce(VecType1 &vec1, VecType2 &vec2)
{
  typedef typename VecType1::base_type Double;
  TimeCalc1("VecDot");
  typename Myf<typename VecType1::base_type>::base_type sum = 0;
#pragma omp parallel reduction(+:sum) 
  {
    Double sump;
    typename Myf<typename VecType1::base_type>::base_type res, tmp;
    Functor<VecType1, VecType2> func(vec1, vec2);
    Index3dBase<typename VecType1::int_type>
      ind(vec1.N1, vec1.N2, vec1.N3);

    Execute<ForAllReduce>(ind, func, vec1.size, vec1.rank);
    sump = func.GetSum();
    tmp =  Myf<Double>::Reduce(sump);
    sum += tmp;
      //      std::cerr << omp_get_thread_num() << std::endl;
    //    std::cout << &sum  << " " << &tmp << std::endl;
  }               // here reduction of sum occurs
  if(!OVERLAP)
    vec1.comm->Barrier();

  typename Myf<typename VecType1::base_type>::base_type res = 0;
  
  {
    TimeCalc1("VecDotComm");
    typename Myf<typename VecType1::base_type>::base_type  tmp1 = sum; 
    vec1.comm->Allreduce(&tmp1, &res, 1, Myf<    typename Myf<typename VecType1::base_type>::base_type>::Datatype(), MPI::SUM);
  }
  return res;
}


template<
  template<class, class> class Functor, 
  class VecType1,
  class VecType2> 
typename Myf<typename VecType1::base_type>::base_type
VecVecReducePBC(VecType1 &vec1, VecType2 &vec2)
{
  
  typedef typename VecType1::base_type Double;
  TimeCalc1("VecDot");
  typename Myf<typename VecType1::base_type>::base_type sum = 0;
#pragma omp parallel reduction(+:sum) 
  {
    Double sump;
    typename Myf<typename VecType1::base_type>::base_type res, tmp;
    Functor<VecType1, VecType2> func(vec1, vec2);
    Index3dBase<typename VecType1::int_type>
      ind(vec1.N1, vec1.N2, vec1.N3);
    ind.ON3 = vec1.ON3;
    Execute<ForAllPBC>(ind, func, vec1.size, vec1.rank);
    sump = func.GetSum();
    tmp =  Myf<Double>::Reduce(sump);
    sum += tmp;
      //      std::cerr << omp_get_thread_num() << std::endl;
    //    std::cout << &sum  << " " << &tmp << std::endl;
  }               // here reduction of sum occurs
  if(!OVERLAP)
    vec1.comm->Barrier();

  typename Myf<typename VecType1::base_type>::base_type res = 0;
  
  {
    TimeCalc1("VecDotComm");
    typename Myf<typename VecType1::base_type>::base_type  tmp1 = sum; 
    vec1.comm->Allreduce(&tmp1, &res, 1, Myf<    typename Myf<typename VecType1::base_type>::base_type>::Datatype(), MPI::SUM);
  }
  return res;
}


template<
  template<class> class Functor, 
  class VecType1
  > 
typename Myf<typename VecType1::base_type>::base_type
VecReduceNorm(VecType1 &vec1)
{
  Functor<VecType1> func(vec1);
  Index3dBase<typename VecType1::int_type>
    ind(vec1.N1, vec1.N2, vec1.N3);
  Execute<ForAll>(ind, func, vec1.size, vec1.rank);


  typedef typename VecType1::base_type Double;
  Double sum = func.GetNorm();

  if(!OVERLAP)
    vec1.comm->Barrier();

  {
    TimeCalc1("VecDotComm");
    typename Myf<typename VecType1::base_type>::base_type res, tmp;
    tmp =  Myf<Double>::Max(sum);
    vec1.comm->Allreduce(&tmp, &res, 1, Myf<    typename Myf<typename VecType1::base_type>::base_type>::Datatype(), MPI::MAX);
    return res;
  }

}


#endif
