#ifndef __hypre_precon_h__
#define __hypre_precon_h__
#include "mpi.h"
#include "solver.h"
#include "useful.h"
//hypre headers
extern "C" {
#include <HYPRE.h>
#include <HYPRE_IJ_mv.h>
#include <HYPRE_parcsr_mv.h>
#include <HYPRE_parcsr_ls.h>
#include <HYPRE_krylov.h>
}
inline void CalcIndeces(SmallVector<int, 6>& nv, int i_, int j_, int k_, int n1_, int n2_, int n3_)
{
  int &i = k_;
  int &j = i_;
  int &k = j_;
  int &n1 = n3_;
  int &n2 = n1_;
  int &n3 = n2_;
  i++, j++, k++;
  nv[0] = (i-1)*(3*n2*n3 + n2 + n3) + n2*n3 + (j-1)*(2*n3+1) + k; //Left
  nv[1] = (i-1)*(3*n2*n3 + n2 + n3) + n2*n3 + j*(2*n3+1) + k;     //Right
  nv[2] = (i-1)*(3*n2*n3 + n2 + n3)+n2*n3+(j-1)*(2*n3+1)+n3+k;    //front
  nv[3] = nv[2] + 1;                                              //Back
  nv[4] = (i-1)*(3*n2*n3 + n2 + n3) + (j-1)*n3 + k;               //Down
  nv[5] = i*(3*n2*n3 + n2 + n3) + (j-1)*n3 + k;                   //Up
  for(int i = 0; i < 6; i++)nv[i]--; //stollen!
//   using namespace std;
//   string a;
//   cin >> a;
//   cerr << i_ << " " << j_ << " " << k_ << endl;
//   for(int i = 0; i < 6; i++) cerr  << nv[i] << " ";
}
inline void CalcIndecesBound(SmallVector<int, 6>& nv, int i, int j, int k, int n1, int n2, int n3)
{
  SmallVector<int, 6> tmp;
  CalcIndeces(nv, i, j, k, n1, n2, n3);
  if(i==(n1-1))
  {
    CalcIndeces(tmp, 0, j, k, n1, n2, n3);
    nv[1] = tmp[0];
  }
  if(j==(n2-1))
  {
    CalcIndeces(tmp, i, 0, k, n1, n2, n3);
    nv[3] = tmp[2];
  }
  if(k==(n3-1))
  {
    CalcIndeces(tmp, i, j, 0, n1, n2, n3);
    nv[5] = tmp[4];
  }
}


inline int GetLb(int N1, int N2, int curline)
{
  int fp, sp, hp;
  fp = N1*N2;
  sp = N1*(N2+1) + N2*(N1+1);
  hp = fp + sp;
  int ret;
  ret = (curline)*hp;// + (curline&1)*fp;
  return ret;
}
inline int GetNum(int N1, int N2, int N3)
{
  return N1*N2*N3*3 + N1*N2 + N2*N3 + N1*N3;
}

class HypreIJRealBase{
protected:
  int il, iu;
  int size;
  int rank;
  MPI_Comm comm;
 public:
  HypreIJRealBase(){};
  HypreIJRealBase(MPI_Comm &comm, int N)
  {
    MPI_Comm_size(comm, &size);
    MPI_Comm_rank(comm, &rank);
    il = (N*rank)/size;
    iu = N*(rank+1)/size -1;
  }
};

class HypreIJbase:public HypreIJRealBase
{
protected:

  int Lines;
public:
  int N1, N2, N3, N3R;
  int StartLine;
  HypreIJbase(){};
  HypreIJbase(MPI_Comm &comm, int N):HypreIJRealBase(comm, N){}
  void SetComm(MPI_Comm &comm_, int N1, int N2, int N3)
  {
    comm = comm_;
    MPI_Comm_size(comm, &size);
    MPI_Comm_rank(comm, &rank);
    StartLine = rank*(N3+1)/size;
    Lines = (rank+1)*(N3+1)/size - StartLine;
    //    cout << "rank "<< rank <<" " <<  N3 << " " <<StartLine <<endl;
    il = GetLb(N1, N2, StartLine);
    using namespace std;
    iu = GetLb(N1, N2, StartLine + Lines) -1;
    //    cout << "il iu" << il << " " << iu << endl;
    if (iu >= GetNum(N1, N2, N3))
      iu = GetNum(N1, N2, N3)-1;
  }
  ~HypreIJbase()
  {
  }
};
class HypreIJVector:public HypreIJbase
{
  HYPRE_IJVector vec;
public:
  typedef double base_type;
  typedef int Int;
  operator HYPRE_ParVector ()
  {
    HYPRE_ParVector  obj;
    void *ptr;
    HYPRE_IJVectorGetObject(vec, &ptr);
    obj = static_cast<HYPRE_ParVector>(ptr);
    return obj;
  }
  HypreIJVector():HypreIJbase(){}
  HypreIJVector(MPI_Comm comm, int N):HypreIJbase(comm, N)
  {
    //    cout << il <<  " " << iu << endl;
    HYPRE_IJVectorCreate(comm, il, iu, &vec);
    HYPRE_IJVectorSetObjectType(vec, HYPRE_PARCSR);
    HYPRE_IJVectorInitialize(vec);
  }
  void Create(MPI_Comm comm, int N1_, int N2_, int N3_)
  {
    N3R = N3_;
    N1 = N1_; N2 = N2_, N3 = N3_;
    SetComm(comm, N1, N2, N3);
    N3 = Lines;
    //    std::cout << "vec" << rank << " "<< size << " "<< il << " " << iu  << " " << N3 << std::endl;
    HYPRE_IJVectorCreate(comm, il, iu, &vec);
    HYPRE_IJVectorSetObjectType(vec, HYPRE_PARCSR);
    HYPRE_IJVectorInitialize(vec);
  }
//   rhs.SetSize(N1, N2, Lines, N3R);
//   rhs.SetComm(comm);
  
  template<int S>
  void AddTo( const SmallVector<double, S>&elrhs, const SmallVector<int, S> &inds)
  {
    for(int i = 0; i< S; i++)
      if((inds[i]>=il)&&(inds[i]<=iu))
	HYPRE_IJVectorAddToValues(vec, 1, &inds[i], &elrhs[i]);
  }
  template<int S>
  void Set( const SmallVector<double, S>&elrhs, const SmallVector<int, S> &inds)
  {
    for(int i = 0; i< S; i++)
      if((inds[i]>=il)&&(inds[i]<=iu))
	HYPRE_IJVectorSetValues(vec, 1, &inds[i], &elrhs[i]);
  }
  template<int S>
  void Get(SmallVector<double, S>&elrhs, const SmallVector<int, S> &inds)const
  {
    for(int i = 0; i< S; i++)
      if((inds[i]>=il)&&(inds[i]<=iu))
	HYPRE_IJVectorGetValues(vec, 1, &inds[i], &elrhs[i]);
  }
  void Assemble()
  {
    HYPRE_IJVectorAssemble(vec);
  }
  ~HypreIJVector()
  {
    HYPRE_IJVectorDestroy(vec);
  }
};

class HypreIJMatrix:public HypreIJbase
{
  HYPRE_IJMatrix mat;
public:
  typedef int int_type;
  int N1, N2, N3, N3R;
  operator HYPRE_ParCSRMatrix ()
  {
    HYPRE_ParCSRMatrix  obj;
    void *ptr;
    HYPRE_IJMatrixGetObject(mat, &ptr);
    obj = static_cast<HYPRE_ParCSRMatrix>(ptr);
    return obj;
  }
  HypreIJMatrix():HypreIJbase(){}
  void Create(MPI_Comm comm, int N1_, int N2_, int N3_)
  {
    N3R = N3_;
    N1 = N1_; N2 = N2_, N3 = N3_;
    SetComm(comm, N1, N2, N3);
    HYPRE_IJMatrixCreate(comm, il, iu, il, iu, &mat);
    HYPRE_IJMatrixSetObjectType(mat, HYPRE_PARCSR);
    HYPRE_IJMatrixSetMaxOffProcElmts(mat, 0);
    std::vector<int> vdiag, vofdiag;
    int N=GetNum(N1, N2, N3);
    vdiag.resize(N);
    //    vofdiag.resize(N);
    for(int i = 0; i <N; ++i)
    {
      vdiag[i] = 11;
      //      vofdiag[i] = 16;
    }
    
    HYPRE_IJMatrixSetRowSizes(mat, &vdiag[0]);
    HYPRE_IJMatrixInitialize(mat);
    N3 = Lines;
  }
  void AddBoundaries(int add)
  {
    if (add)
    {
      HYPRE_IJMatrixInitialize(mat);
      SmallVector<int, 6> nv1, nv2, nvz;

      SmallMatrix<double, 6,6> m1, m2, mz;

      for(int i = 0; i < N1; ++i)
	for(int k = 0; k < N3; ++k)
	{
	  int realk = k + StartLine;
	  if ((realk<0)||(realk>=N3R))
	    continue;
	  CalcIndeces(nv1,i,0,realk, N1, N2, N3R);
	  CalcIndeces(nv2,i,N2-1,realk, N1, N2, N3R);
	  //	  Get(m1, nv1);
	  Get(m2, nv2);
	  mz = 0.0;

	  //	  m1[0][0] += m2[1][1];                //front +a back
	  for (int l = 0; l < 6; ++l)
	  {
	    mz[l][3] = mz[3][l] = m2[3][l];
	    m2[l][3] = m2[3][l] = (double) (l==3);
	  }
	  //	  Set(m1, nv1);
	  Set(m2, nv2);
	  CalcIndecesBound(nv2,i,N2-1,realk, N1, N2, N3R);
	  AddTo(mz, nv2);
	}
      Assemble();
      HYPRE_IJMatrixInitialize(mat);
      for(int j = 0; j < N2; ++j)
	for(int k = 0; k < N3; ++k)
	{
	  int realk = k + StartLine;
	  if ((realk<0)||(realk>=N3R))
	    continue;
	  CalcIndeces(nv1,0,j,realk, N1, N2, N3R);
	  CalcIndeces(nv2,N1-1,j,realk, N1, N2, N3R);
	  //	  Get(m1, nv1);
	  Get(m2, nv2);

	  //	  m1[4][4] += m2[5][5];                //left += right
	  mz = 0.0;
	  for (int l = 0; l < 6; ++l)
	  {
	    mz[1][l] = mz[l][1] = m2[1][l];
	    m2[l][1] = m2[1][l] = (double) (l==1);
	  }

	  //	  Set(m1, nv1);
	  Set(m2, nv2);
	  CalcIndecesBound(nv2,N1-1,j,realk, N1, N2, N3R);
	  AddTo(mz, nv2);
	}
      Assemble();
      HYPRE_IJMatrixInitialize(mat);
      std::vector<double > buf;
      buf.resize(N1*N2*6);
      int ind = 0;
      if (rank == (size-1))
      {
	for(int i = 0; i < N1; ++i)
	  for(int j = 0; j < N2; ++j)
	  {
	    int realk1 = 0;
	    CalcIndeces(nv1,i,j,realk1, N1, N2, N3R);
	    int realk2 = N3R-1;
	    CalcIndeces(nv2,i,j,realk2, N1, N2, N3R);
	    Get(m2, nv2);
	    mz = 0.0;
	    for (int l = 0; l < 6; ++l)
	    {
	      buf[ind++] = m2[5][l];
	      m2[l][5] = m2[5][l] = (double) (l==5);
	    }
	    
	    Set(m2, nv2);

	  }
      }
      MPI_Bcast(&buf[0], N1*N2*6, MPI_DOUBLE, size-1, comm);
      ind = 0;
      for(int i = 0; i < N1; ++i)
	for(int j = 0; j < N2; ++j)
	{
	  int realk1 = 0;
	  CalcIndeces(nv1,i,j,realk1, N1, N2, N3R);
	  int realk2 = N3R-1;
	  CalcIndeces(nv2,i,j,realk2, N1, N2, N3R);
	  //	    Get(m1, nv1);
	  mz = 0.0;
	  for (int l = 0; l < 6; ++l)
	  {
	    mz[l][5] = mz[5][l] = buf[ind++];
	  }	    
	  CalcIndecesBound(nv2,i,j,realk2, N1, N2, N3R);
	  AddTo(mz, nv2);
	}
      Assemble();
    }
    
  }
  void AddZeros()
  {
    if (!Options::GetOptions().PBC)
      return;
    SmallVector<int, 6> nv1, nv2;

    SmallMatrix<double, 6,6> mz;

    mz = 0.0;

    for(int i = 0; i < N1; ++i)
      for(int k = 0; k < N3; ++k)
	{
	  int realk = k + StartLine;
	  if ((realk<0)||(realk>=N3R))
	    continue;
	  CalcIndecesBound(nv2,i,N2-1,realk, N1, N2, N3R);
	  AddTo(mz, nv2);
	}

    for(int j = 0; j < N2; ++j)
      for(int k = 0; k < N3; ++k)
	{
	  int realk = k + StartLine;
	  if ((realk<0)||(realk>=N3R))
	    continue;
	  CalcIndecesBound(nv2,N1-1,j,realk, N1, N2, N3R);
	  AddTo(mz, nv2);
	}
    //    if ((rank==0) ||(rank==size-1))
      {
	for(int i = 0; i < N1; ++i)
	  for(int j = 0; j < N2; ++j)
	    {
	      int realk1 = 0;
	      int realk2 = N3R-1;
	      CalcIndecesBound(nv2,i,j,realk2, N1, N2, N3R);
	      AddTo(mz, nv2);
	    }
      }
  }
    
  HypreIJMatrix(MPI_Comm comm, int N):HypreIJbase(comm, N)
  {
    HYPRE_IJMatrixCreate(comm, il, iu, il, iu, &mat);
    HYPRE_IJMatrixSetObjectType(mat, HYPRE_PARCSR);
    //    HYPRE_IJMatrixSetMaxOffProcElmts(mat, 0);
    std::vector<int> vdiag, vofdiag;
    vdiag.resize(N);
    //    vofdiag.resize(N);
    for(int i = 0; i <N; ++i)
    {
      vdiag[i] = 11;
      //      vofdiag[i] = 16;
    }
    
    //    HYPRE_IJMatrixSetRowSizes(mat, &vdiag[0]);
    HYPRE_IJMatrixInitialize(mat);
  }
  template<int S>
  void AddTo( const SmallMatrix<double, S, S> &elmat1, const SmallVector<int, S> &inds)
  {
    SmallMatrix<double, S, S> elmat;
    elmat = elmat1;
    int ncols = S;

    for(int i = 0; i< S; i++)
    {
      elmat[i][i]*=Options::GetOptions().pfactor;
      if((inds[i]>=il)&&(inds[i]<=iu))
	HYPRE_IJMatrixAddToValues(mat, 1, &ncols, &inds[i], &inds[0], &elmat[i][0]);
    }
  }
  template<int S>
  void Set( const SmallMatrix<double, S, S> &elmat1, const SmallVector<int, S> &inds)
  {
    SmallMatrix<double, S, S> elmat;
    elmat = elmat1;
    int ncols = S;

    for(int i = 0; i< S; i++)
    {
      if((inds[i]>=il)&&(inds[i]<=iu))
	HYPRE_IJMatrixSetValues(mat, 1, &ncols, &inds[i], &inds[0], &elmat[i][0]);
    }
  }
  template<int S>
  void Get(  SmallMatrix<double, S, S> &elmat, const SmallVector<int, S> &inds1)
  {
    int ncols = S;
    SmallVector<int, S> inds;
    inds = inds1;
    for(int i = 0; i< S; i++)
    {
      elmat[i][i]*=Options::GetOptions().pfactor;
      if((inds[i]>=il)&&(inds[i]<=iu))
	HYPRE_IJMatrixGetValues(mat, 1, &ncols, &inds[i], &inds[0], &elmat[i][0]);
    }
  }
  void Assemble()
  {
    HYPRE_IJMatrixAssemble(mat);
  }
  ~HypreIJMatrix()
  {
    HYPRE_IJMatrixDestroy(mat);
  }
};

template<int displacement, class Vector>
void Copy(HypreIJVector&hvec, const Vector &vec)
{
  int N1, N2, N3;
  N1 = hvec.N1; N2 = hvec.N2; N3 = hvec.N3;
  int N3R = hvec.N3R;
  Index3dBase<int> ind(N1, N2, N3);
  for(int i = 0; i < N1; ++i)
    for(int j = 0; j < N2; ++j)
      for(int k = 0; k < N3; ++k)
      {
	ind.SetPos(i, j, k);
	SmallVector<int, 6> nv;
	SmallVector<double, 6> values;
	CalcIndeces(nv,i,j,k + hvec.StartLine, N1, N2, N3R);
	values[0] = vec[ind][LeftEl][displacement]      ;
	values[1] = vec.Right(ind)[LeftEl][displacement];
	values[2] = vec[ind][FrontEl][displacement]     ;
	values[3] = vec.Back(ind)[FrontEl][displacement];
	values[4] = vec[ind][DownEl][displacement]      ;
	values[5] = vec.Up(ind)[DownEl][displacement]   ;
	hvec.Set(values, nv);
      }
  //  hvec.Assemble();
}

template<class Vector>
void Copy(HypreIJVector&hvec, const Vector &vec)
{
  int N1, N2, N3;
  N1 = hvec.N1; N2 = hvec.N2; N3 = hvec.N3;
  int N3R = hvec.N3R;
  Index3dBase<int> ind(N1, N2, N3);
  for(int i = 0; i < N1; ++i)
    for(int j = 0; j < N2; ++j)
      for(int k = 0; k < N3; ++k)
      {
	ind.SetPos(i, j, k);
	SmallVector<int, 6> nv;
	SmallVector<double, 6> values;
	CalcIndeces(nv,i,j,k + hvec.StartLine, N1, N2, N3R);
	values[0] = vec[ind][LeftEl]      ;
	values[1] = vec.Right(ind)[LeftEl];
	values[2] = vec[ind][FrontEl]     ;
	values[3] = vec.Back(ind)[FrontEl];
	values[4] = vec[ind][DownEl]      ;
	values[5] = vec.Up(ind)[DownEl]   ;
	hvec.Set(values, nv);
      }
  //  hvec.Assemble();
}


template<int displacement, class Vector>
void Copy(const Vector &vec, const HypreIJVector&hvec)
{
  int N1, N2, N3;
  N1 = hvec.N1; N2 = hvec.N2; N3 = hvec.N3 ;
  int N3R = hvec.N3R;
  //  std::cout  << N1 << " " << N2 << " " << N3 << std::endl;
  Index3dBase<int> ind(N1, N2, N3);
  for(int i = 0; i < N1; ++i)
    for(int j = 0; j < N2; ++j)
      for(int k = 0; k < N3; ++k)
      {
	ind.SetPos(i, j, k);
	SmallVector<int, 6> nv;
	SmallVector<double, 6> values;
	CalcIndeces(nv,i,j,k+hvec.StartLine, N1, N2, N3R);
	hvec.Get(values, nv);
	vec[ind][LeftEl][displacement]       = values[0];
	vec.Right(ind)[LeftEl][displacement] = values[1];
	vec[ind][FrontEl][displacement]      = values[2];
	vec.Back(ind)[FrontEl][displacement] = values[3];
	vec[ind][DownEl][displacement]       = values[4];
	vec.Up(ind)[DownEl][displacement]    = values[5];
      }
}

template<class Vector>
void Copy(const Vector &vec, const HypreIJVector&hvec)
{
  int N1, N2, N3;
  N1 = hvec.N1; N2 = hvec.N2; N3 = hvec.N3 ;
  int N3R = hvec.N3R;
  //  std::cout  << N1 << " " << N2 << " " << N3 << std::endl;
  Index3dBase<int> ind(N1, N2, N3);
  for(int i = 0; i < N1; ++i)
    for(int j = 0; j < N2; ++j)
      for(int k = 0; k < N3; ++k)
      {
	ind.SetPos(i, j, k);
	SmallVector<int, 6> nv;
	SmallVector<double, 6> values;
	CalcIndeces(nv,i,j,k+hvec.StartLine, N1, N2, N3R);
	hvec.Get(values, nv);
	vec[ind][LeftEl]       = values[0];
	vec.Right(ind)[LeftEl] = values[1];
	vec[ind][FrontEl]      = values[2];
	vec.Back(ind)[FrontEl] = values[3];
	vec[ind][DownEl]       = values[4];
	vec.Up(ind)[DownEl]    = values[5];
      }
}


template<int displacement, class Vector>
void Copy(const Vector &vec, const Vector &svec)
{
  int N1, N2, N3;
  N1 = 8; N2 = 8; N3 = 8;
  //  std::cout  << N1 << " " << N2 << " " << N3 << std::endl;
  Index3dBase<int> ind(N1, N2, N3+1);
  for(int i = 0; i < N1; ++i)
    for(int j = 0; j < N2; ++j)
      for(int k = 0; k < N3; ++k)
      {
	ind.SetPos(i, j, k);
	vec[ind][LeftEl][displacement]       = svec[ind][LeftEl][displacement]       ;
	vec.Right(ind)[LeftEl][displacement] = svec.Right(ind)[LeftEl][displacement] ;
	vec[ind][FrontEl][displacement]      = svec[ind][FrontEl][displacement]      ;
	vec.Back(ind)[FrontEl][displacement] = svec.Back(ind)[FrontEl][displacement] ;
	vec[ind][DownEl][displacement]       = svec[ind][DownEl][displacement]       ;
	vec.Up(ind)[DownEl][displacement]    = svec.Up(ind)[DownEl][displacement]    ;
      }
}

template<class Vector>
class HypreDDPrecon:public PreconBase<Vector>
{
public:
  mutable HypreIJMatrix mat;
  mutable HypreIJVector hsrc;
  mutable HypreIJVector hdst;
  mutable HYPRE_Solver solver;
  mutable HYPRE_Solver PCGsolver;

  mutable MPI_Comm comm;
  void Init(MPI_Comm comm_, int N1, int N2, int N3)
  {
    comm=comm_;
    mat.Create(comm, N1, N2, N3);
    hsrc.Create(comm, N1, N2, N3);
    hdst.Create(comm, N1, N2, N3);
  }
  void Init2()
  {
    TimeCalc1("PreconditionerCreation");
    hsrc.Assemble();
    hdst.Assemble();
    mat.AddZeros();
    mat.Assemble();
    mat.AddBoundaries(Options::GetOptions().PBC);
//     HYPRE_BoomerAMGCreate(&solver);
//     HYPRE_BoomerAMGSetPrintLevel(solver, 1);
//     HYPRE_BoomerAMGSetSym(solver, 1);
//     HYPRE_BoomerAMGSetMaxIter(solver, Options::GetOptions().mit);
//     HYPRE_BoomerAMGSetMaxLevels(solver, 25);
//     HYPRE_BoomerAMGSetStrongThreshold(solver,Options::GetOptions().thres);
//     //        HYPRE_BoomerAMGSetRelaxWt(solver, Options::GetOptions().rwt);
//     //    HYPRE_BoomerAMGSetCoarsenType(solver, 3);
//     //  HYPRE_BoomerAMGSetMeasureType(solver, 0);
//     HYPRE_BoomerAMGSetTol(solver, 1e-14);
//     HYPRE_BoomerAMGSetup(solver, mat, hsrc, hdst);
  HYPRE_ParCSRPCGCreate(comm, &PCGsolver);
  HYPRE_PCGSetTol(PCGsolver, 1e-9);
  HYPRE_PCGSetMaxIter(PCGsolver, Options::GetOptions().mit);
  HYPRE_BoomerAMGCreate(&solver);
  HYPRE_BoomerAMGSetSym(solver, 1);
  HYPRE_BoomerAMGSetPrintLevel(solver,1);
  HYPRE_PCGSetPrintLevel(PCGsolver, 1);
  HYPRE_BoomerAMGSetMaxLevels(solver, 25);
  HYPRE_BoomerAMGSetAggNumLevels(solver, Options::GetOptions().agr);
  HYPRE_BoomerAMGSetStrongThreshold(solver,Options::GetOptions().thres);

  HYPRE_BoomerAMGSetCoarsenType(solver, Options::GetOptions().HypreCoarsType);
  HYPRE_BoomerAMGSetInterpType(solver, Options::GetOptions().HypreInterpType);
  HYPRE_BoomerAMGSetPMaxElmts(solver, Options::GetOptions().HyprePMaxElmts);
  HYPRE_BoomerAMGSetRelaxType(solver, Options::GetOptions().HypreRelaxType);
  HYPRE_BoomerAMGSetOuterWt(solver, Options::GetOptions().HypreOuterWt);

  HYPRE_BoomerAMGSetTol(solver, 0);
  HYPRE_PCGSetPrecond(PCGsolver,(HYPRE_PtrToSolverFcn) HYPRE_BoomerAMGSolve,
                          (HYPRE_PtrToSolverFcn) HYPRE_BoomerAMGSetup ,solver);
  HYPRE_ParCSRPCGSetup(PCGsolver, mat, hsrc, hdst);


  }
  void PreconSolve(Vector &dst, Vector &src)const
  {
    TimeCalc1("PreconSolve");
    {
      TimeCalc1("PreconSolveComm");
      src.StartSendRecvh1();
      src.FinishSendRecvh1();
      if(Options::GetOptions().PBC)
      {
	src.AddBoundarys(2);
      }
    }

    Copy<0>(hsrc, src);
//     HYPRE_BoomerAMGSolve(solver, mat, hsrc, hdst);
    HYPRE_ParCSRPCGSolve(PCGsolver, mat, hsrc, hdst);  
    Copy<0>(dst, hdst);
    Copy<1>(hsrc, src);
//     HYPRE_BoomerAMGSolve(solver, mat, hsrc, hdst);
    HYPRE_ParCSRPCGSolve(PCGsolver, mat, hsrc, hdst);  
    Copy<1>(dst, hdst);
    Copy<2>(hsrc, src);
//     HYPRE_BoomerAMGSolve(solver, mat, hsrc, hdst);
    HYPRE_ParCSRPCGSolve(PCGsolver, mat, hsrc, hdst);  
    Copy<2>(dst, hdst);

    {
      TimeCalc1("PreconSolveComm");
      if(Options::GetOptions().PBC)
      {
	src.AddBoundarys(2);
	dst.AddBoundarys(2);
      }
      dst.StartSendRecvh2();
      dst.FinishSendRecvh2();

    }
  }
  ~HypreDDPrecon()
  {
    HYPRE_BoomerAMGDestroy(solver);
    HYPRE_ParCSRPCGDestroy(PCGsolver);
    
  }


  std::string GetName()const
  {
    return "BoomerDD";
  };
  
};


template<class Vector>
class HyprePrecon:public PreconBase<Vector>
{
public:
  mutable HypreIJMatrix mat;
  mutable HypreIJVector hsrc;
  mutable HypreIJVector hdst;
  mutable HYPRE_Solver solver;
  mutable HYPRE_Solver PCGsolver;

  mutable MPI_Comm comm;
  void Init(MPI_Comm comm_, int N1, int N2, int N3)
  {
    comm=comm_;
    mat.Create(comm, N1, N2, N3);
    hsrc.Create(comm, N1, N2, N3);
    hdst.Create(comm, N1, N2, N3);
  }
  void Init2()
  {
    TimeCalc1("PreconditionerCreation");
    hsrc.Assemble();
    hdst.Assemble();
    mat.AddZeros();
    mat.Assemble();
    mat.AddBoundaries(Options::GetOptions().PBC);
//     HYPRE_BoomerAMGCreate(&solver);
//     HYPRE_BoomerAMGSetPrintLevel(solver, 1);
//     HYPRE_BoomerAMGSetSym(solver, 1);
//     HYPRE_BoomerAMGSetMaxLevels(solver, 25);
//     HYPRE_BoomerAMGSetStrongThreshold(solver,Options::GetOptions().thres);
//     //        HYPRE_BoomerAMGSetRelaxWt(solver, Options::GetOptions().rwt);
//     //    HYPRE_BoomerAMGSetCoarsenType(solver, 3);
//     //  HYPRE_BoomerAMGSetMeasureType(solver, 0);
//     HYPRE_BoomerAMGSetTol(solver, 1e-14);
//     HYPRE_BoomerAMGSetup(solver, mat, hsrc, hdst);
  HYPRE_ParCSRPCGCreate(comm, &PCGsolver);
  HYPRE_PCGSetTol(PCGsolver, Options::GetOptions().InnerTol);
  HYPRE_PCGSetMaxIter(PCGsolver, Options::GetOptions().mit);
  HYPRE_BoomerAMGCreate(&solver);
  HYPRE_BoomerAMGSetMaxIter(solver, 1);//Options::GetOptions().mit);
  HYPRE_BoomerAMGSetSym(solver, 1);
  HYPRE_BoomerAMGSetPrintLevel(solver,1);
  HYPRE_PCGSetPrintLevel(PCGsolver, 1);
  HYPRE_BoomerAMGSetMaxLevels(solver, 25);
  HYPRE_BoomerAMGSetAggNumLevels(solver, Options::GetOptions().agr);
  HYPRE_BoomerAMGSetStrongThreshold(solver,Options::GetOptions().thres);

  HYPRE_BoomerAMGSetCoarsenType(solver, Options::GetOptions().HypreCoarsType);
  HYPRE_BoomerAMGSetInterpType(solver, Options::GetOptions().HypreInterpType);
  HYPRE_BoomerAMGSetPMaxElmts(solver, Options::GetOptions().HyprePMaxElmts);
  HYPRE_BoomerAMGSetRelaxType(solver, Options::GetOptions().HypreRelaxType);
  HYPRE_BoomerAMGSetOuterWt(solver, Options::GetOptions().HypreOuterWt);

  HYPRE_BoomerAMGSetTol(solver, 0);
  HYPRE_PCGSetPrecond(PCGsolver,(HYPRE_PtrToSolverFcn) HYPRE_BoomerAMGSolve,
                          (HYPRE_PtrToSolverFcn) HYPRE_BoomerAMGSetup ,solver);
  HYPRE_ParCSRPCGSetup(PCGsolver, mat, hsrc, hdst);


  }
  void PreconSolve(Vector &dst, Vector &src)const
  {
    TimeCalc1("PreconSolve");
    {
      TimeCalc1("PreconSolveComm");
      src.StartSendRecvh1();
      src.FinishSendRecvh1();
    }

    Copy(hsrc, src);
    //    HYPRE_BoomerAMGSolve(solver, mat, hsrc, hdst);
    HYPRE_ParCSRPCGSolve(PCGsolver, mat, hsrc, hdst);
    int lastiters=0;
    HYPRE_PCGGetNumIterations(PCGsolver, &lastiters);

    ManageTimes::Get("AMG iterations")+=lastiters;
    Copy(dst, hdst);
  }
  ~HyprePrecon()
  {
    HYPRE_BoomerAMGDestroy(solver);
    HYPRE_ParCSRPCGDestroy(PCGsolver);
    
  }


  std::string GetName()const
  {
    return "Boomer";
  };
  
}
;


inline void ClearMatrix(HypreIJMatrix&){}
inline void ClearVector(HypreIJVector&){}

#endif
