#ifndef __jaged_diagonal_h_included__
#define __jaged_diagonal_h_included__
#include "small.h"
#include <vector>
#include <cmath>

inline void rot(int n, int *x, int *y, int rx, int ry) {
  if (ry == 0) {
    if (rx == 1) {
      *x = n-1 - *x;
      *y = n-1 - *y;
    }
    
    //Swap x and y
    int t  = *x;
    *x = *y;
    *y = t;
  }
}

//convert (x,y) to d
inline int xy2d (int n, int x, int y) {
  int rx, ry, s, d=0;
  for (s=n/2; s>0; s/=2) {
    rx = (x & s) > 0;
    ry = (y & s) > 0;
    d += s * s * ((3 * rx) ^ ry);
    rot(s, &x, &y, rx, ry);
  }
  return d;
}
 
//convert d to (x,y)
inline void d2xy(int n, int d, int *x, int *y) {
  int rx, ry, s, t=d;
  *x = *y = 0;
  for (s=1; s<n; s*=2) {
    rx = 1 & (t/2);
    ry = 1 & (t ^ rx);
    rot(s, x, y, rx, ry);
    *x += s * rx;
    *y += s * ry;
    t /= 4;
  }
}
template<class Double, class Int=int>
class IJ_Matrix
{
  struct str{
    Double val;
    Int i;
    int j;
  };
  std::vector<str> svector;
  bool symmetric;
  Int height;
  Int width;
public:
  IJ_Matrix(bool sym=true):height(0), width(0),symmetric(sym)
  {
  }
  void SetSymmetric(bool sym=true){symmetric = sym;}
  void Normalize()
  {
    int n = 1;
    Int h = height;
    if (width>height)
      h = width;
    while (n<h)
      n*=2;
    
    for (Int i = 0; i < svector.size(); ++i)
      for (Int j = i+1; j < svector.size(); ++j)
      {
	if (xy2d(n, I(i), J(i)) > xy2d(n, I(j), J(j)))
	{
	  std::swap(svector[i], svector[j]);
	  // std::swap(I(i), I(j));
	  // std::swap(J(i), J(j));
	  // std::swap(values[i], values[j]);
}
      }
  }
  Int I(Int i)const
  {
    return svector[i].i;
  }
  Int J(Int j)const
  {
    return  svector[j].j;
  }
  Double& Value(Int i)
  {
    return svector[i].val;
  }
  Double Value(Int i)const
  {
    return svector[i].val;
  }
  void MulAddSym(Double *res, const Double *x)const 
  {
    for (Int i = Size();i--;)
    {
      res[I(i)] += Value(i)*x[J(i)];
      res[J(i)] += Value(i)*x[I(i)];
    }
  }
  void MulAddNoSym(Double *res, const Double *x)const 
  {
    for (Int i = Size();i--;)
    {
      res[I(i)] += Value(i)*x[J(i)];
    }
  }

  void MulAddT(Double *res, const Double *x)const 
  {
    for (Int i = Size();i--;)
    {
      res[J(i)] += Value(i)*x[I(i)];
    }
  }
  
  void MulAdd(Double *res, const Double *x)const 
  {
    if(symmetric)
      MulAddSym(res, x);
    else
      MulAddNoSym(res, x);
  }
  Int Size()const
  {
    return svector.size();
  }
  void RemoveIndex(Int ind)
  {
    for (int i = 0; i < Size();++i)
      if ((I(i) == ind)||(J(i)==ind))
	Value(i) = (I(i)==J(i))/(1.+symmetric);
  }
  void RemoveDiffIndex(Int indI, Int indJ)
  {
    for (int i = 0; i < Size();++i)
      if ((I(i) == indI)||(J(i)==indJ))
	Value(i) = ((I(i)==indI)&&(indJ==J(i)))/(1.+symmetric);
  }
  void RemoveRow(Int ind)
  {
    for (int i = 0; i < Size();++i)
      if ((I(i) == ind))
	Value(i) = 0.;
  }
  void RemoveColumn(Int ind)
  {
    for (int i = 0; i < Size();++i)
      if ((J(i) == ind))
	Value(i) = 0.;
  }
  template<class T>
  IJ_Matrix const & operator = (const T &mat)
  {
    
    height = mat.Height();
    width = mat.Width();
    for (Int i = 0; i < height; ++i)
      for (Int j = (symmetric?i:0); j < width; ++j)
      {
	Double val = mat[i][j];
	if ( std::fabs(val) >1e-8)
	{
	  
	  if (symmetric && (i==j))
	    val/=2;
	  str a;
	  a.val = val;
	  a.i = i;
	  a.j = j;
	  svector.push_back(a);
	}
      }
    Normalize();
    return *this;
  }
  Int Height()const {return height;}
  Int Width()const {return width;}
  
};
 

template<class Mat, class Double, class Int=int>
class SomeWhereMatrix
{
 public:
  
  SomeWhereMatrix(Int h, Int w, Mat &ref,
		  Int si, Int sj):width(w),height(h),starti(si), startj(sj), mat(ref)
  {
  }
  const Int starti, startj;
  const Int width, height;
  Mat &mat;

  
  Double const * operator[](Int i)
  {
    return  &mat[starti+i][startj];
  }
  Double const* operator[](Int i)const
  {
    return  &mat[starti+i][startj];
  }
  Int Height()const {return height;}
  Int Width()const {return width;}
};

template<class Double, class Int=int>
class MosaicMatrix
{
  Int maxsizeI, maxsizeJ;
  Int width; int height;
  int isize, jsize;
  bool symmetric;
public:
  void Height()const {return height;}
  void Width()const {return width;}
  MosaicMatrix(Int maxI=32, Int maxJ=0):maxsizeI(maxI), maxsizeJ(maxJ)
  {
    if (maxJ==0)
      maxsizeJ = maxsizeI;
    maxsizeI = 32*sizeof(double)/sizeof(Double);
    maxsizeJ = 32*sizeof(double)/sizeof(Double);
    symmetric = false;
  }
  std::vector<IJ_Matrix<Double,Int> > matrices;


  void MulAdd(Double *res, const Double *x)const 
  {
    //    for (Int i = 0; i < isize; ++i)
    //for (Int j = (symmetric?i:0); j<jsize; ++j)
    for (Int j = 0; j<jsize; ++j)
      for (Int i = 0; i < (symmetric?j+1:isize); ++i)
       {
	IJ_Matrix<Double,Int>  const &mat = (*this)(i,j);
	mat.MulAdd(res+i*maxsizeI, x+j*maxsizeJ);
	if (symmetric && i!=j)
	  mat.MulAddT(res+j*maxsizeJ, x+i*maxsizeI);
      }
  }

  IJ_Matrix<Double,Int> & operator()(Int i, Int j)
  {
    return matrices[i*jsize+j];
  }
  IJ_Matrix<Double,Int> const & operator()(Int i, Int j)const
  {
    return matrices[i*jsize+j];
  }
  
  void RemoveIndex(Int ind)
  {
    Int mi = ind/maxsizeI;
    Int mj = ind/maxsizeJ;

    Int local_i = ind-mi*maxsizeI;
    Int local_j = ind-mj*maxsizeJ;
    for (Int j = (symmetric?mj+1:0); j < jsize; ++j)
    {
      (*this)(mi, j).RemoveRow(local_i);
    }
    for (Int i = 0; i <(symmetric?mi:isize); ++i)
    {
      (*this)(i, mj).RemoveColumn(local_j);
    }
    (*this)(mi,mj).RemoveDiffIndex(local_i, local_j);

  }

  template<class T>
  MosaicMatrix const & operator = (const T &mat)
  {
    width = mat.Width();
    height = mat.Height();
    isize = (height+maxsizeI-1)/maxsizeI;
    jsize = (width+maxsizeJ-1)/maxsizeJ;
    matrices.resize(isize*jsize);
    for (Int i = 0; i < isize; ++i)
      for (Int j = (symmetric?i:0); j<jsize; ++j)
      {
	Int w = width-j*maxsizeJ;
	Int h = height - i*maxsizeI;
	if (w >maxsizeJ)
	  w = maxsizeJ;
	if (h >maxsizeI)
	  h = maxsizeI;

	SomeWhereMatrix<const T, Double, Int> swm(h, w, mat, i*maxsizeI, j*maxsizeJ);
	(*this)(i,j).SetSymmetric(symmetric && (i==j));
	(*this)(i,j) = swm;
      }
    return *this;
  }
};



#endif
