#ifndef __MultiPoly_h__
#define __MultiPoly_h__

#include <vector>
#include <string>
#include <cassert>
#include "small.h"
#ifdef FAST_POLY
#  include "mpt.h"
#  include "pool.h"
#endif

      
namespace My{
  namespace Base{
      
    

    template<class T>
    T Step(const T&a, unsigned step)
    {
      T res = 1;
      for (unsigned int i = 0; i < step; i++)
	res*=a;
      return res;
    }
    

    template<class T>
    class VarArray
    {
      unsigned int size;
      typedef std::vector<T> Storage_t;
      Storage_t *storage;
#ifdef FAST_POLY
      typedef NPool<Storage_t*, 1000> pool_t;
#endif
    public:
      T operator[] (unsigned index)const
      {
	if (index < size)
	  return (*storage)[index];
	else
	  return 0;// good for now
	//			throw range_error("Index out of bounds");
      }
      T& operator[] (unsigned index)
      {
	if (index >= size)
	{
	  storage->resize(index+1);
	  for(unsigned int i = size; i <=index; ++i)
	    (*storage)[i] = 0;
	  size = index+1;
	}
	return (*storage)[index];
		
      }
			
      unsigned Size() const
      {
	return size;
      }
      VarArray():size(0)//, storage(64)
      {
#ifdef FAST_POLY
	::mpt::ThreadOnce<pool_t>::GetRef().Get(storage);
#else
	storage = new Storage_t;
	
#endif					       
      }
      VarArray(VarArray<T>const &a):size(0)
      {
#ifdef FAST_POLY
	::mpt::ThreadOnce<pool_t>::GetRef().Get(storage);
#else
	storage = new Storage_t;
#endif						
	*this = a;
      }
      void Clear()
      {
	storage->resize(0);
	size = 0;
      }
      VarArray<T>& operator=(VarArray<T>  const   &a)
      {
	//	  	  storage = a.storage;

	if(size<a.size)
	  storage->resize(a.size);
	size = a.size;
	std::copy(a.storage->begin(), a.storage->begin()+size, storage->begin());

	return *this;
      }
      ~VarArray()
      {
#ifdef FAST_POLY
	::mpt::ThreadOnce<pool_t>::GetRef().Put(storage);
#else
	delete storage;
#endif						
      }
    };
    /* template<class T> */
    /* class VarArray */
    /* { */
    /* 	unsigned int size; */
    /* 	T *storage; */
    /* public: */
    /* 	T operator[] (unsigned index)const */
    /* 	{ */
    /* 		if (index < size) */
    /* 			return storage[index]; */
    /* 		else */
    /* 			return 0;// good for now */
    /* //			throw range_error("Index out of bounds"); */
    /* 	} */
    /* 	T& operator[] (unsigned index) */
    /* 	{ */
    /* 		if (index < size) */
    /* 			return storage[index]; */
    /* 		T *newstor = new T[index+1]; */
    /* 		if (storage) */
    /* 		{ */
    /* 			unsigned i; */
    /* 			for (i = 0; i < size; i++) */
    /* 				newstor[i] = storage[i]; */
    /* 			for (i = size; i<=index;i++) */
    /* 				newstor[i] = 0; // may be this should be addable from constructor, but... */
    /* 		}else */
    /* 			for(unsigned i = 0; i <=index; i++) */
    /* 				newstor[i] = 0; // the same as above */
	
    /* 		Clear(); */
    /* 		size = index+1; */
    /* 		storage = newstor; */
    /* 		return storage[index]; */
		
    /* 	} */
			
    /* 	unsigned Size() const */
    /* 	{ */
    /* 		return size; */
    /* 	} */
    /* 	VarArray():size(0), storage(NULL) */
    /* 	{ */
    /* 	} */
    /* 	VarArray(VarArray<T>const &a):size(0), storage(NULL) */
    /* 	{ */
    /* 		*this = a; */
    /* 	} */
    /* 	void Clear() */
    /* 	{ */
    /* 		if (storage) */
    /* 			delete[] storage; */
    /* 		storage = NULL; */
    /* 		size = 0; */
    /* 	} */
    /* 	VarArray<T>& operator=(VarArray<T> const &a) */
    /* 	{ */
    /* 		if (size == a.size) */
    /* 		{ */
    /* 			for (int i = 0; i <size; i++) */
    /* 				storage[i] = a.storage[i]; */
    /* 		}else */
    /* 		{ */
    /* 			Clear(); */
    /* 			size = a.size; */
    /* 			if (size) */
    /* 			{ */
    /* 				storage = new T[size]; */
    /* 				for (int i = 0; i < size; i++) */
    /* 					storage[i] = a.storage[i]; */
    /* 			} */
    /* 		} */
    /* 		return *this; */
    /* 	} */
    /* 	~VarArray() */
    /* 	{ */
    /* 		Clear(); */
    /* 	} */
    /* }; */

    template<class T, int varNum>
    class MultiPoly
    {
      VarArray<T> storage;

    public:
      typedef SmallVector<unsigned, varNum> Monom_type;
      typedef SmallVector<T, varNum> Var_type;
      MultiPoly(){}
      static unsigned StepIt(unsigned, unsigned);
      unsigned Size()const
      {
	return storage.Size();
      }
      static unsigned GetInd(Monom_type const &);
      static int NextPerm(Monom_type &, unsigned = 0);

      static T GetMonm(Monom_type const &, Var_type const &);
      T operator()(Var_type const &) const;
      T&operator[](Monom_type const&i)
      {
	return storage[GetInd(i)];
      }
      T&operator[](unsigned int i)
      {
	return storage[i];
      }
      T operator[](unsigned int i)const
      {
	return storage[i];
      }
      void operator = (const T&a)
      {
	storage.Clear();
	storage[0] = a;
      }
      unsigned Vars()const
      {
	return varNum;
      }

      MultiPoly D(int var)
      {
	MultiPoly res;
	Monom_type ar = 0;
	Monom_type br;
	for (unsigned int i = 0; i < Size(); i++)
	{
	  if (ar[var])
	  {
	    br = ar;
	    br[var]--;
	    res[br]+=storage[i]*(T)ar[var];
	  }		  
	  NextPerm(ar);
	}
	return res;
      }

      MultiPoly& operator +=(const MultiPoly&a)
      {
	int mi = std::max(a.Size(), Size());

	for (int i = 0; i< mi; i++)
	  (*this)[i] += a[i];
	return *this;
      }
      MultiPoly& operator *=(const T&a)
      {
	int mi = Size();

	for (int i = 0; i< mi; i++)
	  (*this)[i] *= a;
	return *this;
      }

      template<class Ostream>
      void Print(Ostream &out, const std::string & varname="t")
      {
	out << storage[0];
	Monom_type ar = 0;
	for(int i = 1; i < Size(); i++)
	{
	  NextPerm(ar);
	  if(storage[i]!=0)
	  {
	    out << " +";
	    if(storage[i]!=1)
	      out << " "<< storage[i];
	    for(int vn = 0; vn <varNum; vn++)
	    {
	      if (ar[vn])
		out << " " << varname << vn+1;
	      if (ar[vn]>1)
		out << "^" << ar[vn];
	    }
	  }
	}
      }


    };

    template<class T>
    T Max(T const&a, T const&b)
    {
      return (a<b)?b:a;
    }

    template<class T, int varNum>
    MultiPoly<T, varNum> operator+(MultiPoly<T, varNum> const &a, MultiPoly<T, varNum> const &b)
    {
      int mi = Max(a.Size(), b.Size());
      MultiPoly<T, varNum> res;
      for (int i = 0; i< mi; i++)
	res[i] = a[i] + b[i];
      return res;
    }

    template<class T, int varNum>
    MultiPoly<T, varNum>  operator*(MultiPoly<T, varNum> const &a, MultiPoly<T, varNum> const &b)
    {
      typedef typename MultiPoly<T, varNum>::Monom_type Monom_type;
      //  int mi = Max(a.Size(), b.Size());
      MultiPoly<T, varNum> res;

      Monom_type ar = 0;
      Monom_type abr;
      for (unsigned int i = 0; i < a.Size(); i++)
      {
	Monom_type br = 0;
	for(unsigned int j = 0; j < b.Size(); j++)
	{
	  for (int k = 0; k<varNum;++k)
	    abr[k] = ar[k] + br[k];
	  res[abr] += a[i] * b[j];
	  
	  b.NextPerm(br);
	}
	a.NextPerm(ar);
      }
      return res;
    }
    template<class T, int varNum, class D>
    MultiPoly<T, varNum>  operator*(MultiPoly<T, varNum> const &a, D const &b)
    {
      MultiPoly<T, varNum> tmp;
      tmp[0] = b;
      return a * tmp;
    }
    template<class T, int varNum, class D>
    MultiPoly<T, varNum>  operator*( D const &b, MultiPoly<T, varNum> const &a)
    {
      return a * b;
    }


    template<class T, int varNum>
    T MultiPoly<T, varNum>::operator ()(typename MultiPoly<T, varNum>::Var_type const & X) const
    {
      typedef typename MultiPoly<T, varNum>::Monom_type Monom_type;
      T res = 0;
      Monom_type r = 0;
      for (unsigned i = 0; i < storage.Size(); i++)
      {
	res += GetMonm(r, X)*storage[i];
	NextPerm(r);
      }
      return res;
    }	

    template<class T, int varNum>
    T MultiPoly<T, varNum>::GetMonm(typename MultiPoly<T, varNum>::Monom_type const &unk, 
				    typename MultiPoly<T, varNum>::Var_type const&X)
    {
      T res = 1;
      for (int i = 0; i < varNum; i++)
	res*=Step(X[i], unk[i]);
      return res;
    }

    template<class T, int varNum>
    int MultiPoly<T, varNum>::NextPerm(typename MultiPoly<T, varNum>::Monom_type &var, unsigned from)
    {
      unsigned int N = varNum - 1;
      unsigned int sum = 0;
      for (int i = from; i < varNum; i++)
	sum += var[i];
      if (var[N] == sum) 
      {
	var[from] = sum+1;
	for (int i = from+1; i <varNum; i++)
	  var[i] = 0;
	return 1;
      }
      else
      {
	var[from] -= NextPerm(var, from+1);
	return 0;
      }
    }

    template<class T, int varNum>
    unsigned MultiPoly<T, varNum>::StepIt(unsigned i, unsigned step)
    {
      unsigned res = 1;
      unsigned plus = 0;
      while(step--)
      {
	res = res* (i+ plus);
	res/= ++plus;
      }
      return res;
    }


    template<class T, int varNum>
    unsigned MultiPoly<T, varNum>::GetInd(typename MultiPoly<T, varNum>::Monom_type const&a)
    {
      unsigned int res = 0;
      int sum = 0;
      for (int i = varNum-1; i>=0; i--)
      {
	sum+=a[i];
	res+=StepIt(sum, varNum-i);
      }
      return res;
    }

  } // Base;
} // My

#endif

