#include "mympicxx.hpp"
/*****************************************************************************/
/*                             static members                                */
/*****************************************************************************/

const void* MPI::BOTTOM = (void*) MPI_BOTTOM;
const void* MPI::IN_PLACE = (void*) MPI_IN_PLACE;

// error-handling specifiers

const MPI::Errhandler  MPI::ERRORS_ARE_FATAL(MPI_ERRORS_ARE_FATAL);
const MPI::Errhandler  MPI::ERRORS_RETURN(MPI_ERRORS_RETURN);

#define ERRORS_THROW_EXCEPTIONS_NUM 3

const MPI::Errhandler  MPI::ERRORS_THROW_EXCEPTIONS(ERRORS_THROW_EXCEPTIONS_NUM);

// elementary datatypes

const MPI::Datatype MPI::CHAR(MPI_CHAR);
const MPI::Datatype MPI::SHORT(MPI_SHORT);
const MPI::Datatype MPI::INT(MPI_INT);
const MPI::Datatype MPI::LONG(MPI_LONG);
const MPI::Datatype MPI::SIGNED_CHAR(MPI_CHAR);
const MPI::Datatype MPI::UNSIGNED_CHAR(MPI_UNSIGNED_CHAR);
const MPI::Datatype MPI::UNSIGNED_SHORT(MPI_UNSIGNED_SHORT);
const MPI::Datatype MPI::UNSIGNED(MPI_UNSIGNED);
const MPI::Datatype MPI::UNSIGNED_LONG(MPI_UNSIGNED_LONG);
const MPI::Datatype MPI::FLOAT(MPI_FLOAT);
const MPI::Datatype MPI::DOUBLE(MPI_DOUBLE);
const MPI::Datatype MPI::LONG_DOUBLE(MPI_LONG_DOUBLE);
const MPI::Datatype MPI::BYTE(MPI_BYTE);
const MPI::Datatype MPI::PACKED(MPI_PACKED);
const MPI::Datatype MPI::WCHAR(MPI_WCHAR);
const MPI::Datatype MPI::LONG_LONG(MPI_LONG_LONG_INT);
const MPI::Datatype MPI::UNSIGNED_LONG_LONG(MPI_UNSIGNED_LONG_LONG);

// datatypes for reductions functions (C / C++)

const MPI::Datatype MPI::FLOAT_INT(MPI_FLOAT_INT);
const MPI::Datatype MPI::DOUBLE_INT(MPI_FLOAT_INT);
const MPI::Datatype MPI::LONG_INT(MPI_LONG_INT);
const MPI::Datatype MPI::TWOINT(MPI_2INT);
const MPI::Datatype MPI::SHORT_INT(MPI_SHORT_INT);
const MPI::Datatype MPI::LONG_DOUBLE_INT(MPI_LONG_DOUBLE);

// elementary datatype (Fortran)

const MPI::Datatype MPI::INTEGER(MPI_INTEGER);
const MPI::Datatype MPI::REAL(MPI_REAL);
const MPI::Datatype MPI::DOUBLE_PRECISION(MPI_DOUBLE_PRECISION);
const MPI::Datatype MPI::F_COMPLEX(MPI_COMPLEX);
const MPI::Datatype MPI::LOGICAL(MPI_LOGICAL);
const MPI::Datatype MPI::CHARACTER(MPI_CHARACTER);

// datatype for reduction functions (Fortran)

const MPI::Datatype MPI::TWOREAL(MPI_2REAL);
const MPI::Datatype MPI::TWODOUBLE_PRECISION(MPI_2DOUBLE_PRECISION);
const MPI::Datatype MPI::TWOINTEGER(MPI_2INTEGER);
const MPI::Datatype MPI::TWOF_COMPLEX(MPI_2COMPLEX);

// optional datatypes (Fortran)

const MPI::Datatype MPI::INTEGER1(MPI_INTEGER1);
const MPI::Datatype MPI::INTEGER2(MPI_INTEGER2);
const MPI::Datatype MPI::INTEGER4(MPI_INTEGER4);
const MPI::Datatype MPI::INTEGER8(MPI_INTEGER8);
const MPI::Datatype MPI::REAL4(MPI_REAL4);
const MPI::Datatype MPI::REAL8(MPI_REAL8);
const MPI::Datatype MPI::REAL16(MPI_REAL16);
const MPI::Datatype MPI::LOGICAL1(MPI_LOGICAL1);
const MPI::Datatype MPI::LOGICAL2(MPI_LOGICAL2);
const MPI::Datatype MPI::LOGICAL4(MPI_LOGICAL4);
const MPI::Datatype MPI::LOGICAL8(MPI_LOGICAL8);
const MPI::Datatype MPI::F_DOUBLE_COMPLEX(MPI_DOUBLE_COMPLEX);
const MPI::Datatype MPI::F_COMPLEX8(MPI_COMPLEX8);
const MPI::Datatype MPI::F_COMPLEX16(MPI_COMPLEX16);
const MPI::Datatype MPI::F_COMPLEX32(MPI_COMPLEX32);

// c++ types

#if 0
const MPI::Datatype MPI::BOOL;
const MPI::Datatype MPI::COMPLEX;
const MPI::Datatype MPI::DOUBLE_COMPLEX;
const MPI::Datatype MPI::LONG_DOUBLE_COMPLEX;
#endif

MPI::Intracomm MPI::COMM_WORLD(MPI_COMM_WORLD);
MPI::Intracomm MPI::COMM_SELF(MPI_COMM_SELF);

// collective operations

const MPI::Op MPI::MAX(MPI_MAX);
const MPI::Op MPI::MIN(MPI_MIN);
const MPI::Op MPI::SUM(MPI_SUM);
const MPI::Op MPI::PROD(MPI_PROD);
const MPI::Op MPI::MAXLOC(MPI_MAXLOC);
const MPI::Op MPI::MINLOC(MPI_MINLOC);
const MPI::Op MPI::BAND(MPI_BAND);
const MPI::Op MPI::BOR(MPI_BOR);
const MPI::Op MPI::BXOR(MPI_BXOR);
const MPI::Op MPI::LAND(MPI_LAND);
const MPI::Op MPI::LOR(MPI_LOR);
const MPI::Op MPI::LXOR(MPI_LXOR);
const MPI::Op MPI::REPLACE(MPI_REPLACE);

/* null handles */
const MPI::Group        MPI::GROUP_NULL(MPI_GROUP_NULL);
const MPI_Comm          MPI::COMM_NULL(MPI_COMM_NULL);
const MPI::Datatype     MPI::DATATYPE_NULL(MPI_DATATYPE_NULL);
const MPI::Request      MPI::REQUEST_NULL(MPI_REQUEST_NULL);
const MPI::Op           MPI::OP_NULL(MPI_OP_NULL);
const MPI::Errhandler   MPI::ERRHANDLER_NULL(MPI_ERRHANDLER_NULL);
const MPI::Info         MPI::INFO_NULL(MPI_INFO_NULL);
const MPI::Win          MPI::WIN_NULL(MPI_WIN_NULL);
MPI::File               MPI::FILE_NULL(MPI_FILE_NULL);

// empty group

const MPI::Group MPI::GROUP_EMPTY(MPI_GROUP_EMPTY);

// special datatypes for contstruction of derived datatypes

const MPI::Datatype MPI::UB(MPI_UB);
const MPI::Datatype MPI::LB(MPI_LB);

// file operation constants:
const MPI::Offset MPI::DISPLACEMENT_CURRENT(MPI_DISPLACEMENT_CURRENT);

// pedefined functions:

MPI::Datarep_conversion_function* MPI::CONVERSION_FN_NULL(MPI_CONVERSION_FN_NULL);
