#include <iostream>
#include <cstdlib>
#include <typeinfo>
#include <iomanip>
#include <cmath>
#include <vector>
#include <string>
#include <sstream>
#include <stdexcept>
#include "mpi.h"



using namespace std;


template<class T, int size>
class SmallVector
{
  T buf[size];
public:

  typedef T Element_type;
  const T& operator[](int i)const {return buf[i];}
  T&operator[](int i)      {return buf[i];}
  int Size()const {return size;}
};

template<class T, int Rows, int Columns>
class SmallMatrix:public SmallVector< SmallVector<T, Columns>, Rows>
{
public:
  typedef T Element_type;
  int Height()const{return Rows;}
  int Width()const {return Columns;}
};

template<class T> class ElementMatrix_type:public SmallMatrix<T, 6,6>{};
template<class T> class ElementRhs_type:public SmallVector<T, 6> {};

template<class Matrix, class Vector, class Int>
void GenerateElementMatrixNCRec(Matrix &mat, Vector &vec, Int I, Int J, Int K, Int MI, Int MJ, Int MK)
{
  for(int i = 0; i < mat.Height();++i){    
    for(int j = 0; j < mat.Width();++j)
      mat[i][j] = -4;
    mat[i][i] = 17;
    vec[i] = 1;
  }

  mat[0][1] = mat[1][0] = 
  mat[2][3] = mat[3][2] = 
  mat[4][5] = mat[5][4] = -1;

  if (K == 0){ // Dirichlet boundary conditions setting // Down
    for(int i = 0; i < mat.Height(); ++i)
      mat[i][4] = mat[4][i] = 0;
    mat[4][4] = 1;
    vec[4] = 0;
  }
  if (K == MK){ // Dirichlet boundary conditions setting // Up
    for(int i = 0; i < mat.Height(); ++i)
      mat[i][5] = mat[5][i] = 0;
    mat[5][5] = 1;
    vec[5] = 0;
  }
  if (I == 0){ // Dirichlet boundary conditions setting // Left
    for(int i = 0; i < mat.Height(); ++i)
      mat[i][0] = mat[0][i] = 0;
    mat[0][0] = 1;
    vec[0] = 0;
  }
  if (I == MI){ // Dirichlet boundary conditions setting // Right
    for(int i = 0; i < mat.Height(); ++i)
      mat[i][1] = mat[1][i] = 0;
    mat[1][1] = 1;
    vec[1] = 0;
  }
  if (J == 0){ // Dirichlet boundary conditions setting // Front
    for(int i = 0; i < mat.Height(); ++i)
      mat[i][2] = mat[2][i] = 0;
    mat[2][2] = 1;
    vec[2] = 0;
  }
  if (J == MJ){ // Dirichlet boundary conditions setting // Back
    for(int i = 0; i < mat.Height(); ++i)
      mat[i][3] = mat[3][i] = 0;
    mat[3][3] = 1;
    vec[3] = 0;
  }
//*/
}


template<class Int> class Index3dBase
{
protected:
  Int CurrentCPos, CurrentbCPos;

  const Int N1stride, N2stride;
  const Int N1bstride, N2bstride;
public:
  const Int N1, N2, N3;
  Index3dBase(Int n1, Int n2, Int n3):N1(n1), N2(n2), N3(n3),
    N2stride(n3+1), N2bstride(n3+3), 
    N1stride((n3+1)*(n2+1)), N1bstride((n3+3)*(n2+3))
  {
//    N2stride = N3+1;
//    N2bstride = N3 +3;

//    N1stride = (N3+1)*(N2+1);
//    N1bstride = (N3+3)*(N2+3);    
  }


  void ToUp(){++CurrentCPos;++CurrentbCPos;}
  void ToDown() {--CurrentCPos; --CurrentbCPos;}
  void ToFront(){CurrentCPos-=N2stride; CurrentbCPos -= N2bstride;}
  void ToBack() {CurrentCPos+=N2stride; CurrentbCPos += N2bstride;}
  void ToRight(){CurrentCPos+=N1stride; CurrentbCPos += N1bstride;}
  void ToLeft() {CurrentCPos-=N1stride; CurrentbCPos -= N1bstride;}
  void SetPos(Int i, Int j, Int k)
  {
    CurrentCPos = k + j *N2stride + i * N1stride;
    CurrentbCPos = k + 1 + (j+1)*N2bstride + (i+1)*N1bstride;    
  }
  
  Index3dBase& operator=(Index3dBase const& a)
  {
    CurrentCPos = a.CurrentCPos;
    CurrentbCPos = a.CurrentbCPos;
  }

  Int Up()const    {return CurrentCPos+1;}
  Int Down()const  {return CurrentCPos-1;}
  Int Left()const  {return CurrentCPos - N1stride;}
  Int Right()const {return CurrentCPos + N1stride;}
  Int Back()const  {return CurrentCPos + N2stride;}
  Int Front()const {return CurrentCPos - N2stride;}
  Int RightFront()const {return CurrentCPos + N1stride - N2stride;}
  Int BackDown()const   {return CurrentCPos + N2stride -1;}
  Int RightDown()const  {return CurrentCPos + N1stride - 1;}

  Int BUp()const   {return CurrentbCPos+1;}
  Int BDown()const {return CurrentbCPos-1;}
  Int BLeft()const {return CurrentbCPos - N1bstride;}
  Int BRight()const{return CurrentbCPos + N1bstride;}
  Int BBack()const {return CurrentbCPos + N2bstride;}
  Int BFront()const{return CurrentbCPos - N2bstride;}
  Int BRightFront()const {return CurrentbCPos + N1bstride - N2bstride;}
  Int BBackDown()const   {return CurrentbCPos + N2bstride -1;}
  Int BRightDown()const  {return CurrentbCPos + N1bstride - 1;}


  Int Ind()const{return CurrentCPos;}
  Int BInd()const{return CurrentbCPos;}

//  Int Size()const{return N1*N2*N3;}wrong
//  Int BSize()const{return (N1+2)*(N2+2)*(N3+2);}wrong

};

template<class Int, class Functor>
void DoForEach(Index3dBase<Int> const &ind, Functor &F, bool dolast=true, Int m1=0)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), K(ind);
  I.SetPos(0, 0, m1);
  
  for(i = ind.N1;i; --i, I.ToRight())
  {
    for(j = ind.N2, J=I; j; --j, J.ToBack())
    { 
      for(k = ind.N3-m1*2, K = J; k; --k, K.ToUp()) // doing middle
      {
        F. DoAll(K);
      }
      if (dolast) F.DoDown(K);                             // top plane
    }
    for(k = ind.N3-m1*2, K = J; k; --k, K.ToUp())  // back plane
    {
      F. DoFront(K);
    }
  }
  for(j = ind.N2, J=I; j; --j, J.ToBack())  // right plane
  { 
    for(k = ind.N3-m1*2, K = J; k; --k, K.ToUp())
    {
      F. DoLeft(K);
    }
  }
}
template<class Int, class Functor>
void DoForEachSpecial(Index3dBase<Int> const &ind, Functor &F, bool dolast=true)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), K(ind);
  I.SetPos(0, 0, 0);
  
  
  for(i = ind.N1;i; --i, I.ToRight())
  {

    for(j = ind.N2, J=I; j; --j, J.ToBack())  // Lefts
    { 
      for(k = ind.N3, K = J; k; --k, K.ToUp())
      {
        F. DoLeft(K);
      }
    }
    F.SRLeftUp(ind.N1 - i);
    for(j = ind.N2, J=I; j; --j, J.ToBack())                  // middle plane
    { 
      for(k = ind.N3, K = J; k; --k, K.ToUp()) 
      {
        F. DoFront(K);
        F. DoDown (K);
      }
      if (dolast) F.DoDown(K);                             // top plane // to be checked!!!!!!!!!!!!!!!!!!!!!!!!111
    }
    for(k = ind.N3, K = J; k; --k, K.ToUp())  // Last Front Column
    {
      F. DoFront(K);
    }
    F.SRFrontDownUp(ind.N1 - i);
  }
  for(j = ind.N2, J=I; j; --j, J.ToBack())  // right plane
  { 
    for(k = ind.N3, K = J; k; --k, K.ToUp())
    {
      F. DoLeft(K);
    }
  }
}

template<class Int, class Functor>
void DoForEachSpecial3(Index3dBase<Int> const &ind, Functor &F, bool dolast=true)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), J1(ind), K(ind);
  I.SetPos(0, 0, 0);
  
  F.NR();
  
  for(i = ind.N1;i; --i, I.ToRight())
  {

    for(j = ind.N2, J=I; j; --j, J.ToBack())  // Lefts
    { 
      for(k = ind.N3-1, K = J; k; --k, K.ToUp())
      {
        F. DoLeft(K);
      }
    }
    J1.SetPos(ind.N1-i, 0, ind.N3-1);
    for(j = ind.N2;j;--j, J1.ToBack())
      F.DoLeft(J1);



    F.SRLeftUp(ind.N1 - i);
    F.SRLeftUpFinish(ind.N1 - i);  //wait

    for(j = ind.N2, J=I, J1=I; j; --j, J.ToBack())                  // middle plane
    { 
      for(k = ind.N3-1, K = J, K.ToUp(); k; --k, K.ToUp()) 
      {
        F. DoFront(K);
        F. DoDown (K);
      }
      if (dolast) F.DoDown(K);                             // top plane // to be checked!!!!!!!!!!!!!!!!!!!!!!!!111
    }
    for(k = ind.N3, K = J; k; --k, K.ToUp())  // Last Front Column
    {
      F. DoFront(K);
    }
    for(j = ind.N2;j;--j, J1.ToBack())
    {
       F.DoFront(J1);
       F.DoDown(J1);
    }
    F.DoFront(J1);
    

    F.SRFrontDownUp(ind.N1 - i);
    F.SRFrontDownUpFinish(ind.N1 -i );
  }

  for(j = ind.N2, J=I; j; --j, J.ToBack())  // right plane
  { 
    for(k = ind.N3, K = J; k; --k, K.ToUp())
    {
      F. DoLeft(K);
    }
  }
}



template<class Int, class Functor>
void DoForEachSpecial2(Index3dBase<Int> const &ind, Functor &F, bool dolast=true)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), K(ind);
  I.SetPos(0, 0, -1);
  
  for(i = ind.N1;i; --i, I.ToRight())
  {

    for(j = ind.N2, J=I; j; --j, J.ToBack())  // Lefts
    { 
      for(k = ind.N3+2, K = J; k; --k, K.ToUp())
      {
        F. DoLeft(K);
      }
    }
    for(j = ind.N2, J=I; j; --j, J.ToBack())                  // middle plane
    { 
      for(k = ind.N3+2, K = J; k; --k, K.ToUp()) 
      {
        F. DoFront(K);
        if (k == ind.N3+2) F. DoDown2 (K);
          else
            F.DoDown(K);
      }
      if (dolast) F.DoDown(K);                           
    }
    for(k = ind.N3+2, K = J; k; --k, K.ToUp())  // Last Front Column
    {
      F. DoFront(K);
    }
  }
  for(j = ind.N2, J=I; j; --j, J.ToBack())  // right plane
  { 
    for(k = ind.N3+2, K = J; k; --k, K.ToUp())
    {
      F. DoLeft(K);
    }
  }
}

template<class Int, class Functor>
void DoForEachSpecialReverse(Index3dBase<Int> const &ind, Functor &F, bool dolast=true)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), J1(ind), K(ind);
  I.SetPos(ind.N1, ind.N2, ind.N3);


  for(j = ind.N2, J=I, J.ToFront(); j; --j, J.ToFront())  // right plane
  { 
    for(k = ind.N3, K = J, K.ToDown(); k; --k, K.ToDown())
    {
      F. DoLeft(K);
    }
  }
  

  for(i = ind.N1, I.ToLeft();i; --i, I.ToLeft())
  {
    F. SRLeftDown(i-1);
    F. SRLeftDownFinish(i-1);

    for(k = ind.N3-1, J = I, K = J, K.ToDown(); k; --k, K.ToDown())       // middle plane 
    {
      F. DoFront(K);
    }
    for(j = ind.N2, J=I, J.ToFront(); j; --j, J.ToFront())
    { 
      if (dolast) F.DoDown(J);                             // top plane
      for(k = ind.N3-1, K = J, K.ToDown(); k; --k, K.ToDown()) 
      {
        F. DoDown(K);
        F. DoFront(K);
      }
    }

    J1.SetPos(i-1, ind.N2, 0);
    F.DoFront(J1);
    for(j = ind.N2, J1.ToFront();j; --j, J1.ToFront())
    {
      F.DoDown(J1);
      F.DoFront(J1);
    }




    F. SRFrontDownDown(i-1);
    F. SRFrontDownDownFinish(i-1);

    for(j = ind.N2, J=I, J.ToFront(); j; --j, J.ToFront())           // Left
    { 
      for(k = ind.N3-1, K = J, K.ToDown(), K.ToDown(); k; --k, K.ToDown()) 
      {
        F. DoLeft(K);
      }
    }
    J1.SetPos(i-1, 0, ind.N3-1);
    for(j = ind.N2;j; --j, J1.ToBack())
      F.DoLeft(J1);

  }
}



template<class Int, class Functor>
void DoForEachExtra(Index3dBase<Int> const &ind, Functor &F)
{

    Int i, j, k;
    Index3dBase<Int> IU(ind), JU(ind);
    Index3dBase<Int> IU1(ind), JU1(ind);

    Index3dBase<Int> ID(ind), JD(ind);
    Index3dBase<Int> ID1(ind), JD1(ind);
    ID.SetPos(0, 0, -1);
    ID1.SetPos(0, 0, 0);
    IU.SetPos(0, 0, ind.N3-1);
    IU1.SetPos(0, 0, ind.N3);
  
    for(i = ind.N1
          ;i; 
        --i, IU.ToRight(), ID.ToRight(), IU1.ToRight(), ID1.ToRight())
    {
      for(j = ind.N2, JD=ID,JU=IU, JU1=IU1, JD1 = ID1
          ; j;
         --j, JD.ToBack(), JU.ToBack(), JU1.ToBack(), JD1.ToBack())
      { 
        F.DoDown(JU1);// 
     
        F.DoAll(JU);
        F.DoAll(JD);
        F.DoAll(JD1);
      }
        F.DoFront(JD);
        F.DoFront(JD1);
        F.DoFront(JU);
    }

      for(j = ind.N2, JD=ID,JU=IU, JU1=IU1, JD1 = ID1
          ; j;
         --j, JD.ToBack(), JU.ToBack(), JU1.ToBack(), JD1.ToBack())
      { 
        F.DoLeft(JD);
        F.DoLeft(JD1);
        F.DoLeft(JU);
      }

}


template<class Int, class Functor>
void CDoForEach(Index3dBase<Int> const &ind, Functor &F)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), K(ind);
  I.SetPos(0, 0, 0);
  
  for(i = ind.N1;i; --i, I.ToRight())
  {

    for(j = ind.N2, J=I; j; --j, J.ToBack())  // Lefts
    { 
      for(k = ind.N3, K = J; k; --k, K.ToUp())
      {
        F. DoLeft(K);
      }
    }
    for(j = ind.N2, J=I; j; --j, J.ToBack())
    { 
      for(k = ind.N3, K = J; k; --k, K.ToUp()) // doing Front Column
      {
        F. DoFront(K);
      }
      for(k = ind.N3, K = J; k; --k, K.ToUp()) // doing middle Column
      {
        F. DoDown(K);
      }
      F.DoDown(K);                             // top plane
    }
    for(k = ind.N3, K = J; k; --k, K.ToUp())  // Last Front Column
    {
      F. DoFront(K);
    }
  }
  for(j = ind.N2, J=I; j; --j, J.ToBack())  // right plane
  { 
    for(k = ind.N3, K = J; k; --k, K.ToUp())
    {
      F. DoLeft(K);
    }
  }
}

template<class Int, class Functor>
void DoForEachReverse(Index3dBase<Int> const &ind, Functor &F)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), K(ind);
  I.SetPos(ind.N1, ind.N2, ind.N3);

  for(j = ind.N2, J=I, J.ToFront(); j; --j, J.ToFront())  // right plane
  { 
    for(k = ind.N3, K = J, K.ToDown(); k; --k, K.ToDown())
    {
      F. DoLeft(K);
    }
  }
  

  for(i = ind.N1, I.ToLeft();i; --i, I.ToLeft())
  {
    for(k = ind.N3, J = I, K = J, K.ToDown(); k; --k, K.ToDown())  // back plane
    {
      F. DoFront(K);
    }
    for(j = ind.N2, J=I, J.ToFront(); j; --j, J.ToFront())
    { 
      F.DoDown(J);                             // top plane
      for(k = ind.N3, K = J, K.ToDown(); k; --k, K.ToDown()) // doing middle
      {
        F. DoAll(K);
      }
    }
  }
}
template<class Int, class Functor>
void CDoForEachReverse(Index3dBase<Int> const &ind, Functor &F)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), K(ind);
  I.SetPos(ind.N1, ind.N2, ind.N3);

  for(j = ind.N2, J=I, J.ToFront(); j; --j, J.ToFront())  // right plane
  { 
    for(k = ind.N3, K = J, K.ToDown(); k; --k, K.ToDown())
    {
      F. DoLeft(K);
    }
  }
  

  for(i = ind.N1, I.ToLeft();i; --i, I.ToLeft())
  {
    for(k = ind.N3, J = I, K = J, K.ToDown(); k; --k, K.ToDown())  // back plane
    {
      F. DoFront(K);
    }

    for(j = ind.N2, J=I, J.ToFront(); j; --j, J.ToFront())
    { 
      F.DoDown(J);                             // top plane
      for(k = ind.N3, K = J, K.ToDown(); k; --k, K.ToDown()) // doing middle
      {
        F. DoDown(K);
      }
      for(k = ind.N3, K = J, K.ToDown(); k; --k, K.ToDown()) // doing middle
      {
        F. DoFront(K);
      }
    }
    for(j = ind.N2, J=I, J.ToFront(); j; --j, J.ToFront())  // right plane
    { 
      for(k = ind.N3, K = J, K.ToDown(); k; --k, K.ToDown())
      {
        F. DoLeft(K);
      }
    }

  }
}

enum Elment{LeftEl, FrontEl, DownEl};


template< class V>
class PrintFunctor
{
V const &vec;
public:
  typedef typename V::int_type Int;
  PrintFunctor(V const & d):vec(d){}
  void DoAll(Index3dBase<Int>const& i)
  {
    DoLeft(i);
    DoFront(i);
    DoDown(i);
  }
  void DoLeft(Index3dBase<Int>const& i)
  {
//    if(vec[i][LeftEl] <=0)
      cout << "l" << vec[i][LeftEl] << endl;
  }
  void DoFront(Index3dBase<Int>const& i)
  {
//    if(vec[i][FrontEl] <=0)
      cout << "f" << vec[i][FrontEl] << endl;
  }
  void DoDown(Index3dBase<Int>const& i)
  {
//    if(vec[i][DownEl] <=0)
      cout << "d" << vec[i][DownEl] << endl;
  }
};

template< class Int>
class TestFunctor
{
public:
  void DoAll(Index3dBase<Int>const& i)
  {
    cout << "l" << i.Ind() << endl;
    cout << "f" << i.Ind() << endl;
    cout << "d" << i.Ind() << endl;
  }
  void DoLeft(Index3dBase<Int>const& i)
  {
    cout << "l" << i.Ind() << endl;
  }
  void DoFront(Index3dBase<Int>const& i)
  {
    cout << "f" << i.Ind() << endl;
  }
  void DoDown(Index3dBase<Int>const& i)
  {
    cout << "d" << i.Ind() << endl;
  }
};


template< class V>
class PrintFunctorMat
{
V const &mat;
public:
  typedef typename V::int_type Int;
  PrintFunctorMat(V const & d):mat(d){}
  void DoAll(Index3dBase<Int>const& i)
  {
     DoLeft(i);
     DoFront(i);
     DoDown(i);
  }
  void DoLeft(Index3dBase<Int>const& i)
  {
    cout << "L" ;
    for(int j = 0; j < 6; ++j)
      cout << setw(10) << mat[i][LeftEl][j];
    cout << endl;
  }
  void DoFront(Index3dBase<Int>const& i)
  {
    cout << "F" ;
    for(int j = 0; j < 6; ++j)
      cout << setw(10) << mat[i][FrontEl][j];
    cout << endl;
  }
  void DoDown(Index3dBase<Int>const& i)
  {
    cout << "D" ;
    for(int j = 0; j < 6; ++j)
      cout << setw(10) << mat[i][DownEl][j];
    cout << endl;
  }
};

template<class BaseType> class Priplet :public SmallVector<BaseType, 3>
{
};
template<class BaseType, class In> class Dom_base
{
public:
  typedef Priplet<BaseType> element_type;
  typedef Priplet<BaseType> const_element_type;
  typedef BaseType base_type;
  typedef In Int;
private:
  element_type *buf;  
  
  Dom_base(const Dom_base&);
public:
  Int Size;
  Dom_base():Size(0)
  {
    buf = 0;
  }
  void SetSize1(Int size)
  {
    Size = size;
    buf = new element_type[size];
  }
  Dom_base(Int size)
  {
    SetSize1(size);
  }
  element_type& Index(Int i)
  { 
    return buf[i];
  }
  const_element_type& Index(Int i)const
  {
    return buf[i];
  }
  ~Dom_base(){delete[] buf;}
  
};
template<class A, class B> class BDom;
template<class BaseType, class Int > class Dom:public Dom_base<BaseType, Int>
{

public:
  typedef Dom_base<BaseType, Int> Base;
  typedef typename Base::element_type element_type;
  typedef typename Base::const_element_type const_element_type;
  typedef typename Base::base_type base_type;
  typedef Int int_type;
  typedef Dom Type;
  typedef BDom<BaseType, Int> BType;

protected:
//using Base::SetSize;
public:

  Int N1, N2, N3;
  Dom(Int n1, Int n2, Int n3) // N1, N2, N3 size of the mesh
    :N1(n1), N2(n2), N3(n3), Base((n1+1)*(n2+1)*(n3+1)){}
  Dom():N1(0), N2(0), N3(0){}
  void SetSize(Int n1, Int n2, Int n3)
  {
    N1 = n1; N2 = n2; N3 = n3;
    Base::SetSize1((n1+1)*(n2+1)*(n3+1));
  }
  
  
  element_type& operator[](const Index3dBase<Int> &i)
  {
    return Base::Index(i.Ind());
  }
  const_element_type& operator[](const Index3dBase<Int> &i) const
  {
    return Base::Index(i.Ind());
  }
  
  element_type& Up(const Index3dBase<Int> &i)
  {
    return Base::Index(i.Up());
  }
  element_type& Down(const Index3dBase<Int> &i)
  {
    return Base::Index(i.Down());
  }
  element_type& Left(const Index3dBase<Int> &i)
  {
    return Base::Index(i.Left());
  }
  element_type& Right(const Index3dBase<Int> &i)
  {
    return Base::Index(i.Right());
  }
  element_type& Back(const Index3dBase<Int> &i)
  {
    return Base::Index(i.Back());
  }
  element_type& Front(const Index3dBase<Int> &i)
  {
    return Base::Index(i.Front());
  }
  element_type& RightFront(const Index3dBase<Int> &i)
  {
    return Base::Index(i.RightFront());
  }
  element_type& BackDown(const Index3dBase<Int> &i)
  {
    return Base::Index(i.BackDown());
  }
  element_type& RightDown(const Index3dBase<Int> &i)
  {
    return Base::Index(i.RightDown());
  }

  const_element_type& Up(const Index3dBase<Int> &i)const
  {
    return Base::Index(i.Up());
  }
  const_element_type& Down(const Index3dBase<Int> &i)const
  {
    return Base::Index(i.Down());
  }
  const_element_type& Left(const Index3dBase<Int> &i)const
  {
    return Base::Index(i.Left());
  }
  const_element_type& Right(const Index3dBase<Int> &i)const
  {
    return Base::Index(i.Right());
  }
  const_element_type& Back(const Index3dBase<Int> &i)const
  {
    return Base::Index(i.Back());
  }
  const_element_type& Front(const Index3dBase<Int> &i)const
  {
    return Base::Index(i.Front());
  }
  const_element_type& RightFront(const Index3dBase<Int> &i)const
  {
    return Base::Index(i.RightFront());
  }
  const_element_type& BackDown(const Index3dBase<Int> &i)const
  {
    return Base::Index(i.BackDown());
  }
  const_element_type& RightDown(const Index3dBase<Int> &i)const
  {
    return Base::Index(i.RightDown());
  }

};

template<class T> class My;

template<> class My<double>
{
  public:
static MPI::Datatype Datatype()
{
  return  MPI::DOUBLE;
}
};
template<> class My<float>
{
  public:
static MPI::Datatype Datatype()
{
  return  MPI::FLOAT;
}
};

template<class BaseType, class Int> class BDom:public Dom_base<BaseType, Int>
{
public:
  typedef Dom_base<BaseType, Int> Base;
  typedef typename Base::element_type element_type;
  typedef typename Base::const_element_type const_element_type;
  typedef typename Base::base_type base_type;
  typedef Int int_type;

protected:
//using Base::SetSize;
std::vector<base_type> SendBufferD, SendBufferU, RecvBufferD, RecvBufferU, SendLine, RecvLine;
MPI::Request rsu, rsd, rru, rrd, lrs, lrr;
public:

  Int N1, N2, N3, N3R;
  typedef Dom<BaseType, Int> Type;
  typedef BDom BType;
  BDom(Int n1, Int n2, Int n3, Int n3r) // N1, N2, N3 size of the mesh
    :N1(n1), N2(n2), N3(n3),N3R(n3r),
     Base((n1+3)*(n2+3)*(n3+3)){    SendLine.resize((N2+1)*2);    RecvLine.resize((N2+1)*2);}
  BDom():N1(0), N2(0), N3(0){}
  void SetSize(Int n1, Int n2, Int n3, Int n3r)
  {
    N1 = n1; N2 = n2; N3 = n3;N3R = n3r;
    Base::SetSize1((n1+3)*(n2+3)*(n3+3));
    SendLine.resize((N2+1)*2);
    RecvLine.resize((N2+1)*2);
  }
  void ClearLine(Int I, Int K, Elment El, Int a = 0)
  {
    Index3dBase<Int> ind(N1, N2, N3);
    ind.SetPos(I, 0, K);
    Int cnt = 0;
    Int J;
    for (J = N2+1; J; --J, ind.ToBack())
    { 
      (*this)[ind][El] = 0;
    }    
  }
  void ClearOthers(bool istop)
  {
    for (int i = 0; i <= N1; i++)
    {
      ClearLine(i, -1, DownEl);
      ClearLine(i, -1, LeftEl);
      ClearLine(i, -1, FrontEl);

      ClearLine(i, N3+1, LeftEl);
      ClearLine(i, N3+1, FrontEl);
      ClearLine(i, N3+1, DownEl);
            
      ClearLine(i, N3, LeftEl);
      ClearLine(i, N3, FrontEl);
      if (!istop)
        ClearLine(i, N3, DownEl);
    }
  }
  void PackLineLeftFront(Int I, Int K)
  {
    Index3dBase<Int> ind(N1, N2, N3);
    ind.SetPos(I, 0, K);
    Int cnt = 0;
    Int J;
    for (J = N2+1; J; --J, ind.ToBack())
    { 
      SendLine[cnt++] = (*this)[ind][LeftEl];
      SendLine[cnt++] = (*this)[ind][FrontEl];
    }        
  }
  void PackLineDown(Int I, Int K)
  {
    Index3dBase<Int> ind(N1, N2, N3);
    ind.SetPos(I, 0, K);
    Int cnt = 0;
    Int J;
    for (J = N2+1; J; --J, ind.ToBack())
    { 
      SendLine[cnt++] = (*this)[ind][DownEl];
//      if ((*this)[ind][DownEl] < 0) 
//        cout << "hhh " << ((*this)[ind][DownEl]) << endl;
    }    
  }
  void UnPackLineLeftFrontAdd(Int I, Int K)
  {
    Index3dBase<Int> ind(N1, N2, N3);
    ind.SetPos(I, 0, K);
    Int cnt = 0;
    Int J;
    for (J = N2+1; J; --J, ind.ToBack())
    { 
      (*this)[ind][LeftEl] += RecvLine[cnt++]  ;
      (*this)[ind][FrontEl]+= RecvLine[cnt++]  ;
    }    
  }
  void UnPackLineLeftFront(Int I, Int K)
  {
    Index3dBase<Int> ind(N1, N2, N3);
    ind.SetPos(I, 0, K);
    Int cnt = 0;
    Int J;
    for (J = N2+1; J; --J, ind.ToBack())
    { 
      (*this)[ind][LeftEl] = RecvLine[cnt++]  ;
      (*this)[ind][FrontEl]= RecvLine[cnt++]  ;
    }        
  }
  void UnPackLineDownAdd(Int I, Int K)
  {
    Index3dBase<Int> ind(N1, N2, N3);
    ind.SetPos(I, 0, K);
    Int cnt = 0;
    Int J;
    for (J = N2+1; J; --J, ind.ToBack())
    { 
      (*this)[ind][DownEl] += RecvLine[cnt++]  ;
//      cout << MPI::COMM_WORLD.Get_rank() << " " << ((*this)[ind][DownEl]) << "  ff " << RecvLine[cnt-1] << endl;
    }        
  }
  void UnPackLineDown(Int I, Int K)
  {
    Index3dBase<Int> ind(N1, N2, N3);
    ind.SetPos(I, 0, K);
    Int cnt = 0;
    Int J;
    for (J = N2+1; J; --J, ind.ToBack())
    { 
      (*this)[ind][DownEl] = RecvLine[cnt++]  ;
    }        
  }

  void SendRecvConstr123(Int I, MPI::Intracomm&comm)
  {
    int size = comm.Get_size();
    int rank = comm.Get_rank();


    int UpP = rank+1;
    int DownP = rank-1;
    if (DownP <0) DownP = MPI::PROC_NULL;
    if (UpP == size) UpP = MPI::PROC_NULL;

    for(int i = 0; i < N2; ++i)
      RecvLine[i] = 0;
    PackLineDown(I, N3); //toppest line
    lrs = comm.Isend(&SendLine[0], N2  , My<base_type>::Datatype(), UpP, 0);
    lrr = comm.Irecv(&RecvLine[0], N2  , My<base_type>::Datatype(), DownP, 0);
    lrs.Wait();
    lrr.Wait();
    UnPackLineDownAdd(I, 0);
  }

  void SendRecvConstr2down(Int I, MPI::Intracomm& comm)
  {
    int size = comm.Get_size();
    int rank = comm.Get_rank();


    int UpP = rank+1;
    int DownP = rank-1;
    if (DownP <0) DownP = MPI::PROC_NULL;
    if (UpP == size) UpP = MPI::PROC_NULL;

    for(int i = 0; i < N2; ++i)
      RecvLine[i] = 0;
    PackLineDown(I, N3); //toppest line
    lrs = comm.Isend(&SendLine[0], N2  , My<base_type>::Datatype(), UpP, 0);
    lrr = comm.Irecv(&RecvLine[0], N2  , My<base_type>::Datatype(), DownP, 0);
    lrs.Wait();
    lrr.Wait();
    UnPackLineDownAdd(I, 0);
    PackLineDown(I, 0);
    lrs = comm.Isend(&SendLine[0], N2  , My<base_type>::Datatype(), DownP, 0);
    lrr = comm.Irecv(&RecvLine[0], N2  , My<base_type>::Datatype(), UpP, 0);
    lrs.Wait();
    lrr.Wait();
    if(rank+1 != size) UnPackLineDown(I, N3);
/*    if (rank +1 == size)
    {
      for(int i = 0; i < N2; ++i)
        if (RecvLine[i] != 0)
          cout << " mmm " << RecvLine[i] << endl;
    }
*/
  }
  void SendRecvConstr2left(Int I, MPI::Intracomm& comm)
  {
    int size = comm.Get_size();
    int rank = comm.Get_rank();


    int UpP = rank+1;
    int DownP = rank-1;
    if (DownP <0) DownP = MPI::PROC_NULL;
    if (UpP == size) UpP = MPI::PROC_NULL;

    for(int i = 0; i < N2*2+2; ++i)
      RecvLine[i] = 0;
    PackLineLeftFront(I, -1); //toppest line
    lrs = comm.Isend(&SendLine[0], 2*N2+2  , My<base_type>::Datatype(), DownP, 0);
    lrr = comm.Irecv(&RecvLine[0], 2*N2+2  , My<base_type>::Datatype(), UpP, 0);
    lrs.Wait();
    lrr.Wait();
    UnPackLineLeftFrontAdd(I, N3-1);
  }

  void SendRecvSolve1down(Int I,MPI::Intracomm& comm)
  {
    int size = comm.Get_size();
    int rank = comm.Get_rank();


    int UpP = rank+1;
    int DownP = rank-1;
    if (DownP <0) DownP = MPI::PROC_NULL;
    if (UpP == size) UpP = MPI::PROC_NULL;

    for(int i = 0; i < N2; ++i)
      RecvLine[i] = 0;
    PackLineDown(I, N3); //toppest line
    lrs = comm.Isend(&SendLine[0], N2  , My<base_type>::Datatype(), UpP, 0);
    lrr = comm.Irecv(&RecvLine[0], N2  , My<base_type>::Datatype(), DownP, 0);
  }
  void SendRecvSolve1downFinish(Int I,MPI::Intracomm& comm)
  {
    lrs.Wait();
    lrr.Wait();
    UnPackLineDownAdd(I, 0);

  }
  void SendRecvSolve1Front(Int I, MPI::Intracomm&comm)
  {
    int size = comm.Get_size();
    int rank = comm.Get_rank();


    int UpP = rank+1;
    int DownP = rank-1;
    if (DownP <0) DownP = MPI::PROC_NULL;
    if (UpP == size) UpP = MPI::PROC_NULL;

    for(int i = 0; i < N2*2+2; ++i)
      RecvLine[i] = 0;
    PackLineLeftFront(I, -1); //toppest line
    lrs = comm.Isend(&SendLine[0], 2*N2+2  , My<base_type>::Datatype(), DownP, 0);
    lrr = comm.Irecv(&RecvLine[0], 2*N2+2  , My<base_type>::Datatype(), UpP, 0);
  }
  void NillRequests()
  {
    lrs = MPI::REQUEST_NULL;
    lrr = MPI::REQUEST_NULL;
    for(int i = 0; i < N2*2+2; ++i)
      RecvLine[i] = 0;
    lrs.Wait();
    lrr.Wait();
  }
  void SendRecvSolve1FrontFinish(Int I, MPI::Intracomm&comm)
  {
    lrs.Wait();
    lrr.Wait();
    UnPackLineLeftFrontAdd(I, N3-1);
  }
  void SendRecvSolve2Down(Int I, MPI::Intracomm&comm)
  {
    int size = comm.Get_size();
    int rank = comm.Get_rank();


    int UpP = rank+1;
    int DownP = rank-1;
    if (DownP <0) DownP = MPI::PROC_NULL;
    if (UpP == size) UpP = MPI::PROC_NULL;

    for(int i = 0; i < N2; ++i)
      RecvLine[i] = 0;
    PackLineDown(I, 0); //toppest line
    lrs = comm.Isend(&SendLine[0], N2  , My<base_type>::Datatype(), DownP, 0);
    lrr = comm.Irecv(&RecvLine[0], N2  , My<base_type>::Datatype(), UpP, 0);
  }
  void SendRecvSolve2DownFinish(Int I, MPI::Intracomm&comm)
  {
    int size = comm.Get_size();
    int rank = comm.Get_rank();
    lrs.Wait();
    lrr.Wait();
    if (rank +1 != size) UnPackLineDown(I, N3);
  } 
  void SendRecvSolve2Left(Int I, MPI::Intracomm&comm)
  {
    int size = comm.Get_size();
    int rank = comm.Get_rank();


    int UpP = rank+1;
    int DownP = rank-1;
    if (DownP <0) DownP = MPI::PROC_NULL;
    if (UpP == size) UpP = MPI::PROC_NULL;

    for(int i = 0; i < N2*2+2; ++i)
      RecvLine[i] = 0;
    PackLineLeftFront(I, N3-1); //toppest line
    lrs = comm.Isend(&SendLine[0], 2*N2+2  , My<base_type>::Datatype(), UpP, 0);
    lrr = comm.Irecv(&RecvLine[0], 2*N2+2  , My<base_type>::Datatype(), DownP, 0);
  }
  void SendRecvSolve2LeftFinish(Int I, MPI::Intracomm&comm)
  {
    lrs.Wait();
    lrr.Wait();
    UnPackLineLeftFront(I, -1);
  }
  
  void Pack()
  {
    Int i, j, k;
    Index3dBase<Int> ind(N1, N2, N3);
    Index3dBase<Int> IU(ind), JU(ind), KU(ind);
    Index3dBase<Int> ID(ind), JD(ind), KD(ind);
    ID.SetPos(0, 0, 0);
    IU.SetPos(0, 0, N3-1);
    Int suk = 0;
    Int sdk = 0;
  
    for(i = N1+1;i; --i, IU.ToRight(), ID.ToRight())
    {
      for(j = N2+1, JD=ID,JU=IU; j; --j, JD.ToBack(), JU.ToBack())
      { 
        SendBufferD[sdk++] = (*this)[JD][DownEl];
        
        SendBufferU[suk++] = (*this)[JU][LeftEl];
        SendBufferU[suk++] = (*this)[JU][FrontEl];
        SendBufferU[suk++] = (*this)[JU][DownEl];
      }
    }

  }
  void UnPack(bool und, bool unu)
  {
    Int i, j, k;
    Index3dBase<Int> ind(N1, N2, N3);
    Index3dBase<Int> IU(ind), JU(ind), KU(ind);
    Index3dBase<Int> ID(ind), JD(ind), KD(ind);
    ID.SetPos(0, 0, -1);
    IU.SetPos(0, 0, N3);
    Int suk = 0;
    Int sdk = 0;
  
    for(i = N1+1;i; --i, IU.ToRight(), ID.ToRight())
    {
      for(j = N2+1, JD=ID,JU=IU; j; --j, JD.ToBack(), JU.ToBack())
      { 
        if (unu) (*this)[JU][DownEl] = RecvBufferU[suk++]  ;
        
        if (und)
          {  
            (*this)[JD][LeftEl] = RecvBufferD[sdk++] ;
            (*this)[JD][FrontEl] = RecvBufferD[sdk++] ;
            (*this)[JD][DownEl] = RecvBufferD[sdk++]  ;
          }
      }
    }
  }
  void StartSendRecv(MPI::Intracomm&comm)
  {
    int size = comm.Get_size();
    int rank = comm.Get_rank();


    int UpP = rank+1;
    int DownP = rank-1;
    if (DownP <0) DownP = MPI::PROC_NULL;
    if (UpP == size) UpP = MPI::PROC_NULL;

    SendBufferD.resize((N1+1)*(N2+1));
    RecvBufferU.resize((N1+1)*(N2+1));
    SendBufferU.resize((N1+1)*(N2+1)*3);
    RecvBufferD.resize((N1+1)*(N2+1)*3);
    Pack(); 
    

//    if (rank == 1)
//    {
//       cout << "send ";
//       for(int i = 0; i < (N1 + 1)*(N2+1) ; ++i)
//         cout << SendBufferD[i] << " ";
//       cout << endl;
//    }

    rsd = comm.Isend(&SendBufferD[0], (N1+1)*(N2+1)  , My<base_type>::Datatype(), DownP, 0);
    rru = comm.Irecv(&RecvBufferU[0], (N1+1)*(N2+1)  , My<base_type>::Datatype(), UpP, 0);

    rsu = comm.Isend(&SendBufferU[0], (N1+1)*(N2+1)*3, My<base_type>::Datatype(), UpP, 1);
    rrd = comm.Irecv(&RecvBufferD[0], (N1+1)*(N2+1)*3, My<base_type>::Datatype(), DownP, 1);    
  }

  void FinishSendRecv(MPI::Intracomm&comm)
  {
    int size = comm.Get_size();
    int rank = comm.Get_rank();
    rsu.Wait();
    rsd.Wait();
    rru.Wait();
    rrd.Wait();
//    rsu.Free();
//    rsd.Free();
//    rru.Free();
//    rrd.Free();
//    if (rank == 0)
//    {
//       cout << "recv ";
//       for(int i = 0; i < (N1 + 1)*(N2+1) ; ++i)
//         cout << RecvBufferU[i] << " ";
//       cout << endl;
//    }
    UnPack(rank !=0, rank+1 != size);
    }
     
  element_type& operator[](const Index3dBase<Int> &i)
  {
    return Index(i.BInd());
  }
  const_element_type& operator[](const Index3dBase<Int> &i) const
  {
    return Index(i.BInd());
  }

  element_type& Front(const Index3dBase<Int> &i)
  {
    return Base::Index(i.BFront());
  }
  element_type& RightFront(const Index3dBase<Int> &i)
  {
    return Base::Index(i.BRightFront());
  }
  element_type& BackDown(const Index3dBase<Int> &i)
  {
    return Base::Index(i.BBackDown());
  }
  element_type& RightDown(const Index3dBase<Int> &i)
  {
    return Base::Index(i.BRightDown());
  }

  const_element_type& Up(const Index3dBase<Int> &i)const
  {
    return Base::Index(i.BUp());
  }
  const_element_type& Down(const Index3dBase<Int> &i)const
  {
    return Base::Index(i.BDown());
  }
  const_element_type& Left(const Index3dBase<Int> &i)const
  {
    return Base::Index(i.BLeft());
  }
  const_element_type& Right(const Index3dBase<Int> &i)const
  {
    return Base::Index(i.BRight());
  }
  const_element_type& Back(const Index3dBase<Int> &i)const
  {
    return Base::Index(i.BBack());
  }
  const_element_type& Front(const Index3dBase<Int> &i)const
  {
    return Base::Index(i.BFront());
  }
  const_element_type& RightFront(const Index3dBase<Int> &i)const
  {
    return Base::Index(i.BRightFront());
  }
  const_element_type& BackDown(const Index3dBase<Int> &i)const
  {
    return Base::Index(i.BBackDown());
  }
  const_element_type& RightDown(const Index3dBase<Int> &i)const
  {
    return Base::Index(i.BRightDown());
  }

};


template<class T, class Int>
void ClearMatrix(BDom<SmallVector<T, 6>, Int > & mat)
{
  for(Int i = 0; i < mat.Size; ++i)
  {
    for(int k = 0; k < 3; ++k)
      for(int j = 0; j < 6; ++j)
        mat.Index(i)[k][j] = T(0);
  }
}

template<class T, class Int, template<class F, class G> class Vector>
void ClearVector(Vector<T, Int> & vec, int AAA = 0)
{
  for(Int i = 0; i < vec.Size; ++i)
  {
    for(int k = 0; k < 3; ++k)
        vec.Index(i)[k] = T(AAA);
  }
}

template<class Matrix, class Vector>
void GenerateMatrix(Matrix &mat, Vector &rhs, 
  typename Matrix::int_type NumberOfLines,
  typename Matrix::int_type  StartLine,
  typename Matrix::int_type LO, 
  typename Matrix::int_type RealN3 )
{
  Index3dBase<typename Matrix::int_type> ind(mat.N1, mat.N2, mat.N3);

  typename Matrix::int_type i, j, k;
  
  ClearMatrix(mat);
  ClearVector(rhs);

  ElementMatrix_type<double> elmat;
  ElementRhs_type<double> elrhs;


  
  for (i = 0; i < mat.N1; ++i)
    for( j = 0; j < mat.N2; ++j)
      for( k = 0; k < NumberOfLines; ++k)
      {
        ind.SetPos(i, j, k+LO);
        GenerateElementMatrixNCRec(elmat, elrhs, i, j, k+StartLine, mat.N1 - 1, mat.N2 - 1, RealN3 - 1);
        
        // vector initialization


        rhs[ind][LeftEl] += elrhs[0];
        rhs.Right(ind)[LeftEl] += elrhs[1];
        rhs[ind][FrontEl] += elrhs[2];
        rhs.Back(ind)[FrontEl] += elrhs[3];
        rhs[ind][DownEl]  += elrhs[4];
        rhs.Up(ind)[DownEl] += elrhs[5];

        // matrix initialization
        
        // Left 
        mat[ind][LeftEl][0]        += elmat[0][0]; // Left
        mat[ind][LeftEl][5]        += elmat[0][1]; // right
        mat[ind][LeftEl][1]        += elmat[0][2]; // front
        mat[ind][LeftEl][4]        += elmat[0][3]; // back;
        mat[ind][LeftEl][2]        += elmat[0][4]; // down
        mat[ind][LeftEl][3]        += elmat[0][5]; // up

        // right        
        mat.Right(ind)[LeftEl][0]  += elmat[1][1]; // right
        mat[ind][FrontEl][5]       += elmat[1][2]; // front
        mat.Back(ind)[FrontEl][4]  += elmat[1][3]; // back;
        mat[ind][DownEl][5]        += elmat[1][4]; // down
        mat.Up(ind)[DownEl][4]     += elmat[1][5]; // up

        // front
        mat[ind][FrontEl][0]       += elmat[2][2]; // front
        mat[ind][FrontEl][3]       += elmat[2][3]; // back;
        mat[ind][FrontEl][1]       += elmat[2][4]; // down
        mat[ind][FrontEl][2]       += elmat[2][5]; // up


        // back
        mat.Back(ind)[FrontEl][0]  += elmat[3][3]; // back;
        mat[ind][DownEl][3]        += elmat[3][4]; // down
        mat.Up(ind)[DownEl][2]     += elmat[3][5]; // up

        // Down
        mat[ind][DownEl][0]        += elmat[4][4]; // down
        mat[ind][DownEl][1]        += elmat[4][5]; // up

        // Up
        mat.Up(ind)[DownEl][0]     += elmat[5][5]; // up


      }
}

template<class Matrix, class SrcVector, class ResVector>
class MultiplyMatVector
{
public:
  const Matrix& mat;
  ResVector &res;
  const SrcVector &src;
  MultiplyMatVector(const Matrix&m, const SrcVector&s, ResVector&r):mat(m), res(r), src(s)
  {
    ClearVector(res);
  }

  void DoLeft(Index3dBase<typename Matrix::int_type> const &ind)
  {
    typename Matrix::base_type &mr = mat[ind][LeftEl];
    res[ind][LeftEl] += 
    mr[0] * src[ind][LeftEl] +
    mr[1] * src[ind][FrontEl] +
    mr[2] * src[ind][DownEl] +
    mr[3] * src.Up(ind)[DownEl] +
    mr[4] * src.Back(ind)[FrontEl] +
    mr[5] * src.Right(ind)[LeftEl];
  
    typename ResVector::base_type x = src[ind][LeftEl];
    
    res[ind][FrontEl]      += mr[1] * x;
    res[ind][DownEl]       += mr[2] * x;
    res.Up(ind)[DownEl]    += mr[3] * x;
    res.Back(ind)[FrontEl] += mr[4] * x;
    res.Right(ind)[LeftEl] += mr[5] * x;
    
  }
  void DoFront(Index3dBase<typename Matrix::int_type> const &ind)
  {
    typename Matrix::base_type &mr = mat[ind][FrontEl];
    res[ind][FrontEl] += 
    mr[0] * src[ind][FrontEl] +
    mr[1] * src[ind][DownEl] +
    mr[2] * src.Up(ind)[DownEl] +
    mr[3] * src.Back(ind)[FrontEl] +
    mr[4] * src.RightFront(ind)[LeftEl] +
    mr[5] * src.Right(ind)[LeftEl];
    
    typename ResVector::base_type x = src[ind][FrontEl];
    
    res[ind][DownEl]            += mr[1] * x;
    res.Up(ind)[DownEl]         += mr[2] * x;
    res.Back(ind)[FrontEl]      += mr[3] * x;
    res.RightFront(ind)[LeftEl] += mr[4] * x;
    res.Right(ind)[LeftEl]      += mr[5] * x;
  }
  void DoDown(Index3dBase<typename Matrix::int_type> const &ind)
  {
    typename Matrix::base_type &mr = mat[ind][DownEl];
    res[ind][DownEl] += 
    mr[0] * src[ind][DownEl] +
    mr[1] * src.Up(ind)[DownEl] +
    mr[2] * src.BackDown(ind)[FrontEl] +
    mr[3] * src.Back(ind)[FrontEl] +
    mr[4] * src.RightDown(ind)[LeftEl] +
    mr[5] * src.Right(ind)[LeftEl];
    
    typename ResVector::base_type x = src[ind][DownEl];
    
    res.Up(ind)[DownEl]         += mr[1] * x;
    res.BackDown(ind)[FrontEl]  += mr[2] * x;
    res.Back(ind)[FrontEl]      += mr[3] * x;
    res.RightDown(ind)[LeftEl]  += mr[4] * x;
    res.Right(ind)[LeftEl]      += mr[5] * x;
  }
  void DoAll(Index3dBase<typename Matrix::int_type> const &ind)
  {
    DoLeft(ind), DoFront(ind), DoDown(ind);
  }
  
};

template<class AVector, class BVector> // A += B*alpha
class VecUpdate
{
public:
  typedef typename AVector::base_type Double;
  AVector& avec;
  const BVector& bvec;
  Double alpha;
  VecUpdate(AVector &Avec, const BVector &Bvec, const Double Al):avec(Avec),bvec(Bvec), alpha(Al)
  {
  }

  void DoLeft(Index3dBase<typename AVector::int_type> const &ind)
  {
    avec[ind][LeftEl] += alpha * bvec[ind][LeftEl];
  }
  void DoFront(Index3dBase<typename AVector::int_type> const &ind)
  {
    avec[ind][FrontEl] += alpha * bvec[ind][FrontEl];
  }
  void DoDown(Index3dBase<typename AVector::int_type> const &ind)
  {
    avec[ind][DownEl] += alpha * bvec[ind][DownEl];
  }
  void DoAll(Index3dBase<typename AVector::int_type> const &ind)
  {
    avec[ind][LeftEl] += alpha * bvec[ind][LeftEl];
    avec[ind][FrontEl] += alpha * bvec[ind][FrontEl];
    avec[ind][DownEl] += alpha * bvec[ind][DownEl];
  }  
};

template<class AVector, class BVector> // A -= B
class VecSub
{
public:
  typedef typename AVector::base_type Double;
  AVector& avec;
  const BVector& bvec;
  VecSub(AVector &Avec, const BVector &Bvec):avec(Avec),bvec(Bvec)
  {
  }

  void DoLeft(Index3dBase<typename AVector::int_type> const &ind)
  {
    avec[ind][LeftEl] -=  bvec[ind][LeftEl];
  }
  void DoFront(Index3dBase<typename AVector::int_type> const &ind)
  {
    avec[ind][FrontEl] -= bvec[ind][FrontEl];
  }
  void DoDown(Index3dBase<typename AVector::int_type> const &ind)
  {
    avec[ind][DownEl] -= bvec[ind][DownEl];
  }
  void DoAll(Index3dBase<typename AVector::int_type> const &ind)
  {
    avec[ind][LeftEl] -= bvec[ind][LeftEl];
    avec[ind][FrontEl] -= bvec[ind][FrontEl];
    avec[ind][DownEl] -= bvec[ind][DownEl];
  }  
};

template<class AVector, class BVector> // A = B
class VecCpy
{
public:
  typedef typename AVector::base_type Double;
  AVector& avec;
  const BVector& bvec;
  VecCpy(AVector &Avec, const BVector &Bvec):avec(Avec),bvec(Bvec)
  {
  }

  void DoLeft(Index3dBase<typename AVector::int_type> const &ind)
  {
    avec[ind][LeftEl] =  bvec[ind][LeftEl];
  }
  void DoFront(Index3dBase<typename AVector::int_type> const &ind)
  {
    avec[ind][FrontEl] = bvec[ind][FrontEl];
  }
  void DoDown(Index3dBase<typename AVector::int_type> const &ind)
  {
    avec[ind][DownEl] = bvec[ind][DownEl];
  }
  void DoAll(Index3dBase<typename AVector::int_type> const &ind)
  {
    avec[ind][LeftEl] = bvec[ind][LeftEl];
    avec[ind][FrontEl] = bvec[ind][FrontEl];
    avec[ind][DownEl] = bvec[ind][DownEl];
  }  
};



template<class AVector, class BVector> // A = -B + A*alpha
class VecUpdate2
{
public:
  typedef typename AVector::base_type Double;
  AVector& avec;
  const BVector& bvec;
  Double alpha;
  VecUpdate2(AVector &Avec, const BVector &Bvec, const Double Al):avec(Avec),bvec(Bvec), alpha(Al)
  {
  }

  void DoLeft(Index3dBase<typename AVector::int_type> const &ind)
  {
    avec[ind][LeftEl] = alpha * avec[ind][LeftEl] - bvec[ind][LeftEl];
  }
  void DoFront(Index3dBase<typename AVector::int_type> const &ind)
  {
    avec[ind][FrontEl] = alpha * avec[ind][FrontEl] - bvec[ind][FrontEl];
  }
  void DoDown(Index3dBase<typename AVector::int_type> const &ind)
  {
    avec[ind][DownEl] = alpha * avec[ind][DownEl] - bvec[ind][DownEl];
  }
  void DoAll(Index3dBase<typename AVector::int_type> const &ind)
  {
    avec[ind][LeftEl] = alpha * avec[ind][LeftEl] - bvec[ind][LeftEl];
    avec[ind][FrontEl] = alpha * avec[ind][FrontEl] - bvec[ind][FrontEl];
    avec[ind][DownEl] = alpha * avec[ind][DownEl] - bvec[ind][DownEl];
  }  
};
template<class AVector, class BVector> // A*B
class VecDot
{
public:
  typedef typename AVector::base_type Double;
  const AVector& avec;
  const BVector& bvec;
  Double sum;
  VecDot(const AVector &Avec, const BVector &Bvec):avec(Avec),bvec(Bvec), sum(0)
  {
  }

  void DoLeft(Index3dBase<typename AVector::int_type> const &ind)
  {
    sum += avec[ind][LeftEl] * bvec[ind][LeftEl];
  }
  void DoFront(Index3dBase<typename AVector::int_type> const &ind)
  {
    sum += avec[ind][FrontEl] * bvec[ind][FrontEl];
  }
  void DoDown(Index3dBase<typename AVector::int_type> const &ind)
  {
    sum += avec[ind][DownEl] * bvec[ind][DownEl];
  }
  void DoAll(Index3dBase<typename AVector::int_type> const &ind)
  {
    sum += avec[ind][LeftEl] * bvec[ind][LeftEl];
    sum += avec[ind][FrontEl] * bvec[ind][FrontEl];
    sum += avec[ind][DownEl] * bvec[ind][DownEl];
  }  

  Double GetSum()
  {
    Double ret = sum;
    sum = 0;
    return ret;
  }
};

template<class Matrix, class Vector>
class PreconCreatorStep1
{
  const Matrix& mat;
  Vector &vec;
  typedef typename Vector::base_type Double;
  Double t;
  typedef typename Matrix::base_type SVec;
  MPI::Intracomm &comm;
  
public:
  typedef typename Matrix::int_type Int;
  void SRLeftUp(Int i)
  {
  }
  void SRFrontDownUp(Int i)
  {
    vec.SendRecvConstr1(i, comm);
  }

  PreconCreatorStep1(const Matrix&m, Vector &v, MPI::Intracomm& c):mat(m), vec(v), comm(c)
  {
    ClearVector(vec);
  }
  void DoLeft(Index3dBase<typename Vector::int_type> const &ind)
  {
    vec[ind][LeftEl] += mat[ind][LeftEl][0];
    vec[ind][LeftEl] += mat[ind][LeftEl][5]; 
    vec.Right(ind)[LeftEl] += mat[ind][LeftEl][5];


  }
  void DoFront(Index3dBase<typename Vector::int_type> const &ind)
  {
    vec[ind][FrontEl] += mat[ind][FrontEl][0];
    vec[ind][FrontEl] += mat[ind][FrontEl][3]; 
    vec.Back(ind)[FrontEl] += mat[ind][FrontEl][3];

    
    vec[ind][FrontEl] += mat[ind][FrontEl][1];
    vec[ind][FrontEl] += mat[ind][FrontEl][2];

    vec[ind][DownEl] += mat[ind][FrontEl][1];
    vec.Up(ind)[DownEl] += mat[ind][FrontEl][2];
    
  }
  void DoDown(Index3dBase<typename Vector::int_type> const &ind)
  {
    vec[ind][DownEl] += mat[ind][DownEl][0];
    vec[ind][DownEl] += mat[ind][DownEl][1];
    vec.Up(ind)[DownEl] += mat[ind][DownEl][1];

    vec[ind][DownEl] += mat[ind][DownEl][2];
    vec[ind][DownEl] += mat[ind][DownEl][3];

    vec.BackDown(ind)[FrontEl] += mat[ind][DownEl][2];
    vec.Back(ind)[FrontEl] += mat[ind][DownEl][3];

  }
  void DoDown2(Index3dBase<typename Vector::int_type> const &ind)
  {
    vec[ind][DownEl] += mat[ind][DownEl][0];
    vec[ind][DownEl] += mat[ind][DownEl][1];
    vec.Up(ind)[DownEl] += mat[ind][DownEl][1];

    vec[ind][DownEl] += mat[ind][DownEl][2];
    vec[ind][DownEl] += mat[ind][DownEl][3];

//    vec.BackDown(ind)[FrontEl] += mat[ind][DownEl][2];
    vec.Back(ind)[FrontEl] += mat[ind][DownEl][3];

  }
  void DoAll(Index3dBase<typename Vector::int_type> const &ind)
  {
    DoLeft(ind);
    DoFront(ind);
    DoDown(ind);
  }  
  
};

template<class Matrix, class Vector>
class PreconCreatorStep1_5
{
  const Matrix& mat;
  Vector &vec;
  typedef typename Vector::base_type Double;
  Double t;
  typedef typename Matrix::base_type SVec;
public:
  Double SumL(const SVec&v)
  {
    return v[1] + v[2] + v[3] + v[4];
  }
  Double SumF(const SVec&v)
  {
//    return v[1] + v[2] + v[4] + v[5];
    return v[4] + v[5];
  }
  Double SumD(const SVec&v)
  {
//    return v[2] + v[3] + v[4] + v[5];
    return  v[4] + v[5];
  }
  PreconCreatorStep1_5(const Matrix&m, Vector &v):mat(m), vec(v)
  {
    t = (vec.N1-0)*(vec.N2-0)*(vec.N3R-0);  // change this
    t = 1/t;
    t = exp(log(t)*(2/3.));
//    cout << "Phy = " << t << endl;
  }
  void DoLeft(Index3dBase<typename Vector::int_type> const &ind)
  {

    Double k = SumL(mat[ind][LeftEl]);
    if (vec[ind][LeftEl]> - 2.* k)
      k = t;
    else
      k = sqrt(t);
    vec[ind][LeftEl] += vec[ind][LeftEl] * k;
  }
  void DoFront(Index3dBase<typename Vector::int_type> const &ind)
  {
    Double k = SumF(mat[ind][FrontEl]);
    if (vec[ind][FrontEl]> - 2.* k)
      k = t;
    else
      k = sqrt(t);
    vec[ind][FrontEl] += vec[ind][FrontEl] * k;
  }
  void DoDown(Index3dBase<typename Vector::int_type> const &ind)
  {
    Double k = SumD(mat[ind][DownEl]);
    if (vec[ind][DownEl]> - 2.* k)
      k = t;
    else
      k = sqrt(t);
    vec[ind][DownEl] += vec[ind][DownEl] * k;
  }
  void DoAll(Index3dBase<typename Vector::int_type> const &ind)
  {
    DoLeft(ind);
    DoFront(ind);
    DoDown(ind);
  }  
  
};


template<class Matrix, class Vector>
class PreconCreatorStep2
{
  const Matrix& mat;
  Vector &vec;
  
  typedef typename Vector::base_type Double;
  typedef typename Matrix::base_type SVec;
  MPI::Intracomm &comm;
public:
  typedef typename Matrix::int_type Int;
  void SRLeftUp(Int i)
  {
      vec.SendRecvConstr2down(i, comm);
  }
  void SRFrontDownUp(Int i)
  {
//      vec.SendRecvConstr2left(i+1, comm);
  }

  
  Double SumL(const SVec&v)
  {
    return v[1] + v[2] + v[3] + v[4];
  }
  Double SumF(const SVec&v)
  {
//    return v[1] + v[2] + v[4] + v[5];
    return  v[4] + v[5];
  }
  Double SumD(const SVec&v)
  {
//    return v[1] + v[2] + v[4] + v[5];
    return v[4] + v[5];
  }

  PreconCreatorStep2(const Matrix&m, Vector &v, MPI::Intracomm &c):mat(m), vec(v), comm(c)
  {
  }
  void DoLeft(Index3dBase<typename Vector::int_type> const &ind)
  {
    Double ss = SumL(mat[ind][LeftEl])/vec[ind][LeftEl];
//    cout << "L" << ss << endl;
    vec[ind][FrontEl] -= mat[ind][LeftEl][1] *ss;
    vec[ind][DownEl]  -= mat[ind][LeftEl][2] *ss;
    vec.Up(ind)[DownEl] -= mat[ind][LeftEl][3] *ss;
    vec.Back(ind)[FrontEl] -= mat[ind][LeftEl][4] *ss;
  }
  void DoFront(Index3dBase<typename Vector::int_type> const &ind)
  {
    Double ss = SumF(mat[ind][FrontEl])/vec[ind][FrontEl];
//    cout << "F" << ss << endl;
//    vec[ind][DownEl] -= mat[ind][FrontEl][1] *ss;
//    vec.Up(ind)[DownEl]  -= mat[ind][FrontEl][2] *ss;
    vec.RightFront(ind)[LeftEl] -= mat[ind][FrontEl][4] *ss;
    vec.Right(ind)[LeftEl] -= mat[ind][FrontEl][5] *ss;
  }
  void DoDown(Index3dBase<typename Vector::int_type> const &ind)
  {

    Double ss = SumD(mat[ind][DownEl])/vec[ind][DownEl];
//    cout << "D" << ss << endl;
//    vec.BackDown(ind)[FrontEl] -= mat[ind][DownEl][2] *ss;
//    vec.Back(ind)[FrontEl]  -= mat[ind][DownEl][3] *ss;
    vec.RightDown(ind)[LeftEl] -= mat[ind][DownEl][4] *ss;
    vec.Right(ind)[LeftEl] -= mat[ind][DownEl][5] *ss;
  }
  void DoAll(Index3dBase<typename Vector::int_type> const &ind)
  {
    DoLeft(ind);
    DoFront(ind);
    DoDown(ind);
  }  
  
};

template<class Matrix, class XVector, class ResVector, class SrcVector>
class PreconSolverStep2
{
  const Matrix& mat;
  const XVector &X;
  ResVector &Res;
  const SrcVector &Src;
  MPI::Intracomm &comm;
public:
  typedef typename XVector::base_type Double;
  PreconSolverStep2(const Matrix&m, const XVector &x, ResVector &res, const SrcVector &src,   MPI::Intracomm &c ):mat(m), X(x), Res(res), Src(src), comm(c)
  {    
  }

  typedef typename XVector::int_type Int;
  
  void  SRFrontDownDown(Int i)
  {
    Res.SendRecvSolve2Down(i, comm);
  }
  void  SRFrontDownDownFinish(Int i)
  {
    Res.SendRecvSolve2DownFinish(i, comm);
  }
  
  void SRLeftDown(Int i)
  {
    Res.SendRecvSolve2Left(i+1, comm);
  }
  void SRLeftDownFinish(Int i)
  {
    Res.SendRecvSolve2LeftFinish(i+1, comm);
  }
  

  void DoLeft(Index3dBase<typename XVector::int_type> const &ind)  
  {
    Double &h = Res[ind][LeftEl];
    typedef typename Matrix::base_type Row;
    Row & ma = mat[ind][LeftEl];
    h -= 
      ma[1] * Res[ind][FrontEl] +
      ma[2] * Res[ind][DownEl] +
      ma[3] * Res.Up(ind)[DownEl] +
      ma[4] * Res.Back(ind)[FrontEl]; 
    h /= X[ind][LeftEl];    

//    Res[ind][LeftEl] = Src[ind][LeftEl];

  }
  void DoFront(Index3dBase<typename XVector::int_type> const &ind)
  {
    Double &h = Res[ind][FrontEl];
    typedef typename Matrix::base_type Row;
    Row & ma = mat[ind][FrontEl];

    h -= 
//      ma[1] * Res[ind][DownEl] +
//      ma[2] * Res.Up(ind)[DownEl] +
      ma[4] * Res.RightFront(ind)[LeftEl] +
      ma[5] * Res.Right(ind)[LeftEl]; 
    h /= X[ind][FrontEl];    

//    Res[ind][FrontEl] = Src[ind][FrontEl];

  }
  void DoDown(Index3dBase<typename XVector::int_type> const &ind)
  {
    Double &h = Res[ind][DownEl];
    typedef typename Matrix::base_type Row;
    Row & ma = mat[ind][DownEl];
    h -= 
//      ma[2] * Res.BackDown(ind)[FrontEl] +
//      ma[3] * Res.Back(ind)[FrontEl] +
      ma[4] * Res.RightDown(ind)[LeftEl] +
      ma[5] * Res.Right(ind)[LeftEl]; 
    h /= X[ind][DownEl];    
    
//    Res[ind][DownEl] = Src[ind][DownEl];
  }
  void DoAll(Index3dBase<typename XVector::int_type> const &ind)
  {
    DoDown(ind);
    DoFront(ind);
    DoLeft(ind);
  }  
  
};


template<class Matrix, class XVector, class ResVector, class SrcVector>
class PreconSolverStep1_5
{
  const Matrix& mat;
  const XVector &X;
  ResVector &Res;
  const SrcVector &Src;
public:
  typedef typename XVector::base_type Double;
  PreconSolverStep1_5(const Matrix&m, const XVector &x, ResVector &res, const SrcVector &src ):mat(m), X(x), Res(res), Src(src)
  {
  }


  void DoLeft(Index3dBase<typename XVector::int_type> const &ind)  
  {
    Double &r = Res[ind][LeftEl];
    r *= X[ind][LeftEl];

  }
  void DoFront(Index3dBase<typename XVector::int_type> const &ind)
  {
    Double &r = Res[ind][FrontEl];
    r *= X[ind][FrontEl];
  }
  void DoDown(Index3dBase<typename XVector::int_type> const &ind)
  {
    Double &r = Res[ind][DownEl];
    r *= X[ind][DownEl];
  }
  void DoAll(Index3dBase<typename XVector::int_type> const &ind)
  {
    DoLeft(ind);
    DoFront(ind);
    DoDown(ind);
  }  
  
};


template<class Matrix, class XVector, class ResVector, class SrcVector>
class PreconSolverStep1
{
  const Matrix& mat;
  const XVector &X;
  ResVector &Res;
  const SrcVector &Src;
  MPI::Intracomm &comm;
public:
  typedef typename XVector::int_type Int;
  typedef typename XVector::base_type Double;
  PreconSolverStep1(const Matrix&m, const XVector &x, ResVector &res, const SrcVector &src,   MPI::Intracomm &c):mat(m), X(x), Res(res), Src(src), comm(c)
  {
    Begin();
  }

  void Begin()
  {
    ClearVector(Res);
  };

  void SRLeftUp(Int i)
  {
    Res.SendRecvSolve1down(i, comm);
  }

  void SRLeftUpFinish(Int i)
  {
    Res.SendRecvSolve1downFinish(i, comm);
  }
  
  void SRFrontDownUp(Int i)
  {
    Res.SendRecvSolve1Front(i+1, comm);
  }

  void SRFrontDownUpFinish(Int i)
  {
    Res.SendRecvSolve1FrontFinish(i+1, comm);
  }

  void NR()
  {
    Res.NillRequests();
  }

  void DoLeft(Index3dBase<typename XVector::int_type> const &ind)  
  {
    Double &r = Res[ind][LeftEl];
    r += Src[ind][LeftEl];
    r /= X[ind][LeftEl];
    typedef typename Matrix::base_type Row;
    Row & ma = mat[ind][LeftEl];

    
    Res[ind][FrontEl]      -=  ma[1] * r;
    Res[ind][DownEl]       -=  ma[2] * r;
    Res.Up(ind)[DownEl]    -=  ma[3] * r;
    Res.Back(ind)[FrontEl] -=  ma[4] * r; 
    
  }
  void DoFront(Index3dBase<typename XVector::int_type> const &ind)
  {
    Double &r = Res[ind][FrontEl];
    r += Src[ind][FrontEl];

    r /= X[ind][FrontEl];
    typedef typename Matrix::base_type Row;
    Row & ma = mat[ind][FrontEl];

//    Res[ind][DownEl]            -= ma[1] * r;
//    Res.Up(ind)[DownEl]         -= ma[2] * r;
    Res.RightFront(ind)[LeftEl] -= ma[4] * r;
    Res.Right(ind)[LeftEl]      -= ma[5] * r; 


  }
  void DoDown(Index3dBase<typename XVector::int_type> const &ind)
  {
    Double &r = Res[ind][DownEl];
    r += Src[ind][DownEl];

    r /= X[ind][DownEl];
    typedef typename Matrix::base_type Row;
    Row & ma = mat[ind][DownEl];

//    Res.BackDown(ind)[FrontEl]  -= ma[2]*r;
//    Res.Back(ind)[FrontEl]      -= ma[3]*r;
    Res.RightDown(ind)[LeftEl]  -= ma[4]*r;
    Res.Right(ind)[LeftEl]      -= ma[5]*r; 

  }
  void DoAll(Index3dBase<typename XVector::int_type> const &ind)
  {
    DoLeft(ind);
    DoFront(ind);
    DoDown(ind);
  }  
  
};






unsigned  GetNum (string s, unsigned def)
{
  istringstream is(s);
  unsigned a;
  is >> a;
  if (a)
    return a;
  return def;
}

typedef double Double;



template<class Double, class Int>
class Worker
{

MPI::Intracomm comm;

int rank, size;

Int N1, N2, N3R, N3L;
Int StartLine;
int Lines;


public:
typedef SmallVector<Double, 6> row_type;
typedef BDom<row_type, Int> matrix_type;
typedef BDom<Double, Int> bvector_type;
typedef Dom<Double, Int> vector_type;
private:
matrix_type mat;
bvector_type rhs;
bvector_type x;
public:


unsigned DoCG(
              const matrix_type &Mat,
              bvector_type &X,
              const bvector_type &RHS,
              Double Eps,
              Index3dBase<Int> &ind)
{
  typedef bvector_type XType;

  Eps *= Eps;
  typedef matrix_type Matrix;
  typedef bvector_type XType;
  typedef bvector_type RHSType;
  typedef typename XType::BType DType;
  typedef typename XType::BType GType;
  typedef typename XType::BType HdType;  

  DType D(X.N1, X.N2, X.N3, X.N3R);
  GType G(X.N1, X.N2, X.N3, X.N3R);
  HdType Hd(X.N1, X.N2, X.N3, X.N3R);
  Double gtg, gtgold, tau, beta, gtgfull;
  Double temp, temp1;
  
  MultiplyMatVector<Matrix, DType, HdType> MulHd(Mat, D, Hd);
  
  VecDot<GType, GType> GtG(G, G);
  VecDot<DType, HdType> DtHd(D, Hd);
  
  
  {
    VecCpy<DType, XType> Cpy(D, X);
    DoForEach(ind, Cpy, rank+1 == size); // D = X;
  }
  
  ClearVector(Hd);
  D.StartSendRecv(comm);
  DoForEach(ind, MulHd, false,Int(1)); // Hd = Ad
  D.FinishSendRecv(comm);
  DoForEachExtra(ind, MulHd);
  


  {
    VecCpy<GType, HdType> Cpy(G, Hd);
    DoForEach(ind, Cpy, rank+1 == size); // G = Hd;
    VecSub<GType, RHSType> vsub(G, RHS);
    DoForEach(ind, vsub, rank+1 == size); // G -= RHS;
  }

  
  ClearVector(D); // D = 0;
  
  {
    VecSub<DType, GType> vsub(D, G);
    DoForEach(ind, vsub, rank+1 == size); // D = -G;
  }


//  ClearVector(G, 1);
  
  DoForEach(ind, GtG, rank+1==size);
  temp = GtG.GetSum(); // GtG
//  cout << rank << " " << temp << endl;
  comm.Allreduce(&temp, &gtg, 1, My<Double>::Datatype(), MPI::SUM);
//  if (rank == 0)
//    cout << gtg << endl;
  gtgfull = gtg;  

  unsigned iter = 0;
//  cout << gtg << endl;
//  return 0;
  while (gtg > Eps*gtgfull)
  {
//    if (rank == 0)
//        cout << iter <<" " << gtg << endl;
    gtgold = gtg;

    ClearVector(Hd);
    D.StartSendRecv(comm);
    DoForEach(ind, MulHd, false, Int(1)); // Hd = Ad
    D.FinishSendRecv(comm);
    DoForEachExtra(ind, MulHd);

//  printing of 
//{
//    PrintFunctor<HdType> pf(D);
//    ForEachInLine(pf, 1);
//}
//{
//    PrintFunctor<HdType> pf(G);
//    ForEachInLine(pf, 1);
//}
//    return 0;

   

    DoForEach(ind, DtHd, rank+1 == size);  // DtHd
    temp = DtHd.GetSum();
    comm.Allreduce(&temp, &temp1, 1, My<Double>::Datatype(), MPI::SUM);

//    if (rank == 0)
//        cout << iter <<" " << temp1 << endl;

    tau = gtg/temp1;
    {
      VecUpdate<XType, DType> vecu1(X, D, tau);
      DoForEach(ind, vecu1, rank+1 == size); // x += d*tau
      VecUpdate<GType, HdType> vecu2(G, Hd, tau);
      DoForEach(ind, vecu2, rank+1 == size); // g += Hd*tau
    }
    DoForEach(ind, GtG, rank+1 == size);
    temp = GtG.GetSum(); // GtG
    comm.Allreduce(&temp, &gtg, 1, My<Double>::Datatype(), MPI::SUM);
    beta = gtg/gtgold;
    {
      VecUpdate2<DType, GType> vecu(D, G, beta);
      DoForEach(ind, vecu, rank+1 == size);
    }
    ++iter;
  }

  return iter;     
}

unsigned DoPCG(
              const matrix_type &Mat,
              bvector_type &X,
              const bvector_type &RHS,
              Double Eps,
              Index3dBase<Int> &ind)
{
  typedef bvector_type XType;

  Eps *= Eps;
  typedef matrix_type Matrix;
  typedef bvector_type XType;
  typedef bvector_type RHSType;
  typedef typename XType::BType DType;
  typedef typename XType::BType GType;
  typedef typename XType::BType HdType;  
  typedef typename XType::BType PType;
  
  
  PType P(X.N1, X.N2, X.N3, X.N3R);
  DType D(X.N1, X.N2, X.N3, X.N3R);
  GType G(X.N1, X.N2, X.N3, X.N3R);
  HdType Hd(X.N1, X.N2, X.N3, X.N3R);
  Double gtg, gtgold, tau, beta, gtgfull, temp, temp1;
  

  PreconCreatorStep1<Matrix, PType> st1(Mat, P, comm);
  DoForEachSpecial2 (ind, st1, rank+1 == size);          // special creator
  PreconCreatorStep1_5<Matrix, PType> st1_5(Mat, P); // everithing's normal
  DoForEach(ind, st1_5, rank+1==size);
  P.ClearOthers(rank+1==size);
  PreconCreatorStep2<Matrix, PType> st2(Mat, P, comm);
  DoForEachSpecial(ind, st2, true);

  
  PreconSolverStep1<Matrix, PType, HdType, GType> pcs1(Mat, P, Hd, G, comm);
  PreconSolverStep2<Matrix, PType, HdType, GType> pcs2(Mat, P, Hd, G, comm);
  PreconSolverStep1_5<Matrix, PType, HdType, GType> pcs1_5(Mat, P, Hd, G);

  MultiplyMatVector<Matrix, DType, HdType> MulHd(Mat, D, Hd);


  
  VecDot<GType, HdType> GtG(G, Hd);
  VecDot<DType, HdType> DtHd(D, Hd);

  comptime = -MPI::Wtime();  
  
  {
    VecCpy<DType, XType> Cpy(D, X);
    DoForEach(ind, Cpy, rank+1==size); // D = X;
  }
  
    ClearVector(Hd);
    D.StartSendRecv(comm);
    DoForEach(ind, MulHd, false, Int(1)); // Hd = Ad
    D.FinishSendRecv(comm);
    DoForEachExtra(ind, MulHd);
  


  {
    VecCpy<GType, HdType> Cpy(G, Hd);
    DoForEach(ind, Cpy, rank+1==size); // G = Hd;
    VecSub<GType, RHSType> vsub(G, RHS);
    DoForEach(ind, vsub, rank+1==size); // G -= RHS;
    pcs1.Begin();


    Hd.ClearOthers(rank+1==size);
    DoForEachSpecial3(ind, pcs1, rank+1==size);
    DoForEach(ind, pcs1_5, rank+1==size);
    DoForEachSpecialReverse(ind, pcs2, rank+1==size);           // Hd = C-1g
  }



  
  
  ClearVector(D); // D = 0;
  
  {
    VecSub<DType, HdType> vsub(D, Hd);
    DoForEach(ind, vsub, rank+1==size); // D = -G;
  }
  
  DoForEach(ind, GtG, rank+1==size);
  temp = GtG.GetSum();
  comm.Allreduce(&temp, &temp1, 1, My<Double>::Datatype(), MPI::SUM);

  gtgfull = gtg = temp1; // GtG
  gtgfull *= Eps;
  unsigned iter = 0;
  while (gtg > gtgfull)
  {
//    if (rank == 0)
  //    cout << iter <<" " << gtg << endl;
    gtgold = gtg;

    ClearVector(Hd);
    D.StartSendRecv(comm);
    D.FinishSendRecv(comm);
    DoForEach(ind, MulHd, false, Int(1)); // Hd = Ad
    DoForEachExtra(ind, MulHd);

   

    DoForEach(ind, DtHd, rank+1==size);  // DtHd
    temp = DtHd.GetSum();
    comm.Allreduce(&temp, &temp1, 1, My<Double>::Datatype(), MPI::SUM);
    tau = gtg/temp1;
    {
      VecUpdate<XType, DType> vecu1(X, D, tau);
      DoForEach(ind, vecu1, rank+1==size); // x += d*tau
      VecUpdate<GType, HdType> vecu2(G, Hd, tau);
      DoForEach(ind, vecu2, rank+1==size); // g += Hd*tau
    }

    pcs1.Begin();
    Hd.ClearOthers(rank+1==size);
    DoForEachSpecial3(ind, pcs1, rank+1==size);
    DoForEach(ind, pcs1_5, rank+1==size);
    DoForEachSpecialReverse(ind, pcs2, rank+1==size);           // Hd = C-1g


    DoForEach(ind, GtG, rank+1==size);
    temp = GtG.GetSum();
    comm.Allreduce(&temp, &temp1, 1, My<Double>::Datatype(), MPI::SUM);
    gtg = temp1;
    beta = gtg/gtgold;
    {
      VecUpdate2<DType, HdType> vecu(D, Hd, beta);
      DoForEach(ind, vecu, rank+1==size);
    }
    ++iter;
  }

  comptime += MPI::Wtime();
  return iter;     
}



  Worker(MPI::Intracomm  & c, Int n1, Int n2, Int n3):N1(n1), N2(n2), N3R(n3),comm(c.Dup()), 
          rank(c.Get_rank()), size(c.Get_size())
  {
//    if (2*size>n3+1)
//        throw runtime_error("Too many processors.");
    StartLine = rank*(N3R+1)/size;
    Lines = (rank+1)*(N3R+1)/size - StartLine;
    if (rank+1 ==size) Lines --;
    mat.SetSize(N1, N2, Lines, N3R);
    rhs.SetSize(N1, N2, Lines, N3R);
    x.SetSize(N1, N2, Lines, N3R);
//    cout << "Lines " << rank << " " << Lines << endl;
  }

template<class Functor>
void ForEachInLine(Functor &F, int line)
{
  if ((line >= StartLine) && (line - StartLine < Lines) || (line == N3R) && (rank == size-1))
  {

    Int i, j, k;
    Index3dBase<Int> ind(N1, N2, Lines);
    Index3dBase<Int> ID(ind), JD(ind), KD(ind);
    ID.SetPos(0, 0, line-StartLine);
    Int suk = 0;
    Int sdk = 0;
    cout <<" Printing " << rank << " " << ind.N1 << " " << ind.N2 <<  endl;
    for(i = ind.N1;i; --i, ID.ToRight())
    {
      for(j = ind.N2, JD=ID; j; --j, JD.ToBack())
      { 
        F.DoLeft(JD);
//        cout << "Mamka mu" << endl;
        F.DoFront(JD);
//        cout << "Mamka mu" << endl;
        F.DoDown(JD);
//        cout << "Mamka mu" << endl;
      }
//        F.DoFront(JD);
    }

      for(j = ind.N2, JD=ID; j; --j, JD.ToBack())
      { 
//        cout << "Mamka mu" << endl;
        F.DoLeft(JD);
      }
  }
}
  double comptime;

  void Work()
  {
//    cout << rank << " " << StartLine << " " << Lines << endl;
    Int NumberOfLines = Lines+2;
    Int SL = StartLine-1;
    Int SO = -1;
    if (rank == 0)
    {
      SL++;
      SO++;
      NumberOfLines--;
    }
    if (rank +1 == size)
    {
      NumberOfLines--;
    }
    ClearMatrix(mat);
    ClearVector(rhs);

//    cout << "generate " << rank << " " << Lines <<  " " << NumberOfLines << " " << SL << " " << SO << endl;

    double t1 = MPI::Wtime();

    GenerateMatrix(mat, rhs, NumberOfLines, SL, SO, N3R);
    ClearVector(x);
    Index3dBase<Int> ind(N1, N2, Lines);
//    return ;
    int iter = DoPCG(mat, x, rhs, 1e-9, ind);
    t1 = MPI::Wtime()-t1;
    if(rank == 0)
        cout << "PCG " <<   iter << ", " << comptime << " sec."<< endl;  

  }
};





int main(int argc, char **argv)
{
  MPI::Init(argc, argv);



  int N1 = 32;
  int N2, N3;

  if (argc>1)
    N1 = GetNum(argv[1], 32);
  if (argc>2)
    N2 = GetNum(argv[2], N1);
  else
    N2 = N1;
  if (argc>3)
    N3 = GetNum(argv[3], N1);
  else
    N3 = N1;


  try{

  Worker<double, ptrdiff_t> wrk(MPI::COMM_WORLD, N1, N2, N3);

  wrk.Work();
  }catch(MPI::Exception &e)
  {
    cout << MPI::COMM_WORLD.Get_size() << " " << MPI::COMM_WORLD.Get_rank() << " " << e.Get_error_string() << endl;
  }catch(std::exception &e)
  {
    cout << MPI::COMM_WORLD.Get_size() << " " << MPI::COMM_WORLD.Get_rank() << " " << e.what() << endl;
  }
  MPI::Finalize();
  return 0;

/*





  

  matrix_type mat(N1, N2, N3);

  bvector_type rhs(N1, N2, N3);
  bvector_type x(N1, N2, N3);



  
  ClearVector(x);

  cout << N1 << "x" << N2 << "x" << N3 << endl;  
  cout << "Pcg " <<   DoPCG(mat, x, rhs, 1e-10, ind) << endl;  
//  ClearVector(x);
*/
}
