#include "optionsparser.hpp"


void PrintHelp()
{
  using namespace std;
  cout << "\t pelast\t[-n=#] ...... [-n1=#] [-n2=#] [-n3=#] \n\t\t[-cg | -m1 | -m2 | -m3 ] [-eps=#] [-f]" << endl;
  cout  << endl;

}

void UpdateMaterials()
{
  Options &opt = Options::GetOptions();
  for(Options::cit_type it = opt.E.begin(); it !=opt.E.end(); ++it)
  {
    int i = it->first;
    double lambda, mu, ni, E;
    ni = opt.ni[i];
    E = opt.E[i];
    if ((ni+E)==0.0)
      continue;
    lambda = E * ni /((1+ni)*(1-2* ni));
    mu = E / (2 *(1+ni));
    opt.lambda[i] = lambda;
    opt.mu[i] = mu;
  }
  for(Options::cit_type it = opt.ni.begin(); it !=opt.ni.end(); ++it)
  {
    int i = it->first;
    double lambda, mu, ni, E;
    ni = opt.ni[i];
    E = opt.E[i];
    if ((ni+E)==0.0)
      continue;
    lambda = E * ni /((1+ni)*(1-2* ni));
    mu = E / (2 *(1+ni));
    opt.lambda[i] = lambda;
    opt.mu[i] = mu;
    //    std::cout << lambda << " " << mu << std::endl;
  }
}

void ReadMaterials(const std::string& fname)
{
  using namespace std;
  ifstream is;
  is.exceptions(ios::badbit|
		ios::eofbit|
		ios::failbit);
  try{
  is.open(fname.c_str(), ios::in);
  std::string a;
  while(true)
  {
    getline(is, a);
    if (a.substr(0,1)=="#")
      continue;
    std::istringstream b;
    b.str(a);
    int ind; double E, ni, lambda, mu;
    b >> ind;
    b >> ni >> E;
    if (b)
    {
      Options &opt= Options::GetOptions();
      if ((E+ni)==0)
      {
	b >> lambda >> mu;
	opt.lambda[ind] = lambda;
	opt.mu[ind] = mu;
	opt.E[ind] = E;
	opt.ni[ind] = ni;
      }
      {
	opt.E[ind] = E;
	opt.ni[ind] = ni;
      }
      //      std::cout << ind << " " << E << " " << ni << std::endl;
    }
  }
  }catch(std::exception &e)
  {
  }
}
bool Help;
bool exithelp;

void Init(MPI::Intracomm &comm, int argc, char **argv)
{
  using namespace std;

  Options &opt = Options::GetOptions();
  int &N = opt.N;
  int &N1 = opt.N1;
  int &N2 = opt.N2;
  int &N3 = opt.N3;
  double &eps = opt.eps;
  //  double zero;
  DiscretizationType &distype  = opt.distype; distype = RP;
  BoundaryConditions &bc = opt.bc; bc = bcNone;//bcLeft|bcRight|bcFront|bcBack|bcUp|bcDown;
  BoundaryConditions &nc = opt.nc; nc = bcNone;//bcLeft|bcRight|bcFront|bcBack|bcUp|bcDown;

  SolutionMethod &method = opt.method; method = smPCG4;

  
  // parameter decoding
  N = GetOption("-n=", argc, argv, 4, "Dimensions if equal.");
  N1 = GetOption("-n1=", argc, argv, N, "First dimension.");
  N2 = GetOption("-n2=", argc, argv, N, "Second dimension.");
  N3 = GetOption("-n3=", argc, argv, N, "Fird dimension.");
  Options::GetOptions().SmallDirichlet = OptionPresent("-sd", argc, argv, "Very little Dirichlet boundary conditions.");
  Options::GetOptions().reuse_rhs = OptionPresent("-rrhs", argc, argv, "Reuse right hand side vecotor.");
  Options::GetOptions().SmallDirichlet2 = OptionPresent("-sd2", argc, argv, "Very little Dirichlet boundary conditions. - only on the matrix - boundary remains the same as with dirichlet");
  Options::GetOptions().PBC = OptionPresent("-pbc", argc, argv, "Periodic domain. Periodic boundary conditions.");
  Options::GetOptions().Reg = GetOption("-reg=", argc, argv, 0.0, "Regularization number");
  Options::GetOptions().debug = OptionPresent("-d", argc, argv, "Debug output.");
  Options::GetOptions().outputprecision = GetOption("-prec=", argc, argv, 7, "Precision for the solution output.");  
  Options::GetOptions().hom = OptionPresent("-hom", argc, argv, "Homogenization mode.");
  Options::GetOptions().homk = GetOption("-hk=", argc, argv, 0, "Homogenization: k.");
  Options::GetOptions().homl = GetOption("-hl=", argc, argv, 0, "Homogenization: l.");
  Options::GetOptions().useplates = !OptionPresent("-np", argc, argv, "No plates (for bone).");
  Options::GetOptions().variant = GetOption("-v=", argc, argv, 0, "Variant.");
  Options::GetOptions().DM1 = GetOption("-dm1=", argc, argv, 1., "Max x.");
  Options::GetOptions().DM2 = GetOption("-dm2=", argc, argv, 1., "Max y.");
  Options::GetOptions().DM3 = GetOption("-dm3=", argc, argv, 1., "Max z.");

  Options::GetOptions().gcg = OptionPresent("-gcg", argc, argv, "Use generalized conjucate gradient method.");
  Options::GetOptions().gcg_m = GetOption("-m=", argc, argv, 1, "GCG iterations.");
  Options::GetOptions().proj = OptionPresent("-proj", argc, argv, "Use projection in the conjucate gradient method. (only for periodic bcs)");
  Options::GetOptions().froj = OptionPresent("-froj", argc, argv, "Use frojection in the conjucate gradient method. (only for periodic bcs)");
  Options::GetOptions().matscale = GetOption("-mc=", argc, argv, 1.0, "MIC(0):scale elementmatrix");
  Options::GetOptions().lam_coeff = GetOption("-lc=", argc, argv, 1.0);

  std::string &namebase = Options::GetOptions().namebase;
  {
    std::stringstream a;
    a.fill('0');
    a << comm.Get_size() << "-" <<setw(4)
      << comm.Get_rank() << "";
    namebase += a.str();
  }
  for(int i = 1; i < argc; ++i)
    namebase += argv[i];

  std::string nb = namebase;
  namebase = "";
  for(unsigned int i = 0; i < nb.length(); ++i)
    if(nb[i]!='/')
      namebase += nb[i];
  //  std::cout << namebase << std::endl;

  Options::GetOptions().outputname = GetOption("-oname=", argc, argv, "put." + (Options::GetOptions().namebase), "Output file name base.");
  
  double maxn = N1;
  if (maxn < N2)
    maxn = N2;
  if (maxn < N3)
    maxn = N3;
  //  DM1 = N1/maxn;
  //  DM2 = N2/maxn;
  //  DM3 = N3/maxn;
  Options::GetOptions().InnerTol = GetOption("-inn=", argc, argv, 1e-3, "Inner tollerance 1e-3");
  Options::GetOptions().pfactor = GetOption("-p=", argc, argv, 1.0, "Perturbation factor 1.");
  Options::GetOptions().ppfactor = GetOption("-pp=", argc, argv, 1.0, "Perturbation factor 2.");
  Options::GetOptions().thres = GetOption("-thres=", argc, argv, .5, "BAMG strong threshold.");
  Options::GetOptions().agr = GetOption("-agr=", argc, argv, 0, "BAMG aggrecive coarsing levels.");
  Options::GetOptions().mit = GetOption("-mit=", argc, argv, 1, "BAMG inner iterations count.");

  Options::GetOptions().HypreCoarsType = GetOption("-hct=", argc, argv,  6, "BAMG coarsening type number.");
  Options::GetOptions().HypreInterpType = GetOption("-hit=", argc, argv, 0,  "BAMG interpolation type number.");
  Options::GetOptions().HyprePMaxElmts = GetOption("-hpm=", argc, argv, 0, "BAMG SetPMaxElmts.");
  Options::GetOptions().HypreRelaxType = GetOption("-hrt=", argc, argv, 6, "BAMG Relaxation type number.");
  Options::GetOptions().HypreOuterWt = GetOption("-how=", argc, argv, 1.0, "BAMG HypreSetOuterWt.");
  Options::GetOptions().HypreNumSweeps = GetOption("-hns=", argc, argv, 1, "BAMG HypreAMGSetNumSweeps");
  
  Options::GetOptions().maxiter = GetOption("-maxiter=", argc, argv, 10101, "Maximum outer iterations count.");
  Options::GetOptions().ct = GetOption("-ct=", argc, argv, 1, "BAMG ct???");
  Options::GetOptions().rwt = GetOption("-rwt=", argc, argv, 1., "BAMG rwt???");
  eps = GetOption("-eps=", argc, argv, 1.0e-6, "Solution precision.");
  Options::GetOptions().usebone = OptionPresent("-b", argc, argv, "External datafile to be used for the domain.");
  Options::GetOptions().zero = GetOption("-z=", argc, argv, 0.0, "How much is zero.");
//   Options::GetOptions().E[0] = GetOption("-E=", argc,argv, Options::GetOptions().E[0]);
//   Options::GetOptions().ni[0] = GetOption("-ni=", argc,argv, Options::GetOptions().ni[0] );
//   Options::GetOptions().E[1] = GetOption("-E1=", argc,argv, Options::GetOptions().E[0]);
//   Options::GetOptions().ni[1] = GetOption("-ni1=", argc,argv, Options::GetOptions().ni[0] );
//   Options::GetOptions().E[255] = GetOption("-E255=", argc,argv, Options::GetOptions().E[1]);
//   Options::GetOptions().ni[255] = GetOption("-ni255=", argc,argv, Options::GetOptions().ni[1] );
  ReadMaterials(GetOption("-mf=", argc, argv, std::string("materials"), "File materials to be read from."));
  GetArrayOption("-ni", argc, argv, Options::GetOptions().ni, "Poisson ratio.");
  GetArrayOption("-E", argc, argv, Options::GetOptions().E, "Elasticity modulus.");
  GetArrayOption("-lambda", argc, argv, Options::GetOptions().lambda, "Lame's lambda.");
  GetArrayOption("-mu", argc, argv, Options::GetOptions().mu, "Lame's mu");
  UpdateMaterials();
  Options::GetOptions().force = GetOption("-f=", argc,argv, Options::GetOptions().force, "Ammount of force applyed.");
  Options::GetOptions().SinglePrecPrecon = OptionPresent("-s", argc, argv, "Single precision preconditioner to be used.");
  Options::GetOptions().scratchpath = GetOption("-sp=", argc, argv, std::string("."), "ScratchPath - where output files to be saved and maybe input to be read from.");
  if (OptionPresent("-cg", argc, argv, "Solution method:CG without preconditioner method.")) method = smCG;
  if (OptionPresent("-jcg", argc, argv, "Solution method:Jacobi preconditioner method.")) method = smJacobi;
  if (OptionPresent("-save", argc, argv, "To save the solution in file")) Options::GetOptions().SaveSolution = true;
  if (OptionPresent("-m1", argc, argv, "Solution method: PCG mic(0) without modification - sequencional.")) method = smPCG1;
  if (OptionPresent("-m2", argc, argv, "Solution method: PCG mic(0) columnwise parallelism.")) method = smPCG2;
  if (OptionPresent("-m3", argc, argv, "Solution method: PCG mic(0) planewise parallelism.")) method = smPCG3;
  if (OptionPresent("-m3p", argc, argv, "Solution method: PCG mic(0) planewise parallelism. Periodic boundary")) method = smPCG3PBC, Options::GetOptions().PBC=true;
  if (OptionPresent("-m4", argc, argv, "Solution method: PCG mic(0) columnwise parallelism. Faster implementation. Allows fewer processors.")) method = smPCG4;
  if (OptionPresent("-hd", argc, argv, "Solution method: PCG DD+BoomerAMG preconditioner.")) method = smBoomerDD;
  if (OptionPresent("-hb", argc, argv, "Solution method: PCG + BoomerAMG (not implemented).")) method = smBoomer;
  
  if (OptionPresent("-mv", argc, argv, "Discretization type: MeanValue (MV).")) distype = MV;
  if (OptionPresent("-mp", argc, argv, "Discretization type: MidPoint (MP).")) distype = MP;
  if (OptionPresent("-rp", argc, argv, "Discretization type: MP, faster")) distype = RP;
  if (OptionPresent("-rv", argc, argv, "Discretization type: MP, faster")) distype = RV;
  Options::GetOptions().Overlap = (OptionPresent("-nolap", argc, argv, "No overlapping of communications and computations.")) ? false :true;
  using namespace std;
  string bound = GetOption("-bc=", argc, argv, string("d"), "Dirichlet boundary conditions. Some of the leters lrudfb - from left, right, up, bottom, front, back.");

  if((method==smPCG4)&&(comm.Get_size()*3 > N3))
  {
    if (comm.Get_rank()==0)
      std::cerr << "switching to method 2" <<std::endl;
    method=smPCG2;
  }
  if((comm.Get_size() > N3+1))
  {
    if (comm.Get_rank()==0)
      std::cerr << "Too many processors. Quitting.." <<std::endl;
    throw "Too many processors. Throwing an exception";
  }

  std::ostringstream bfn;
  bfn << (N/2) << ".bonec.inr";
    Options::GetOptions().bone_fname = GetOption("-bfn=", argc, argv, bfn.str(), "External datafile name to be used for the domain.");
  
  if(bound.find("l")!=string::npos)
    bc = (BoundaryConditions)(bc | bcLeft);
  if(bound.find("r")!=string::npos)
    bc = (BoundaryConditions)(bc | bcRight);
  if(bound.find("d")!=string::npos)
    bc = (BoundaryConditions)(bc | bcDown);
  if(bound.find("u")!=string::npos)
    bc = (BoundaryConditions)(bc | bcUp);
  if(bound.find("f")!=string::npos)
    bc = (BoundaryConditions)(bc | bcFront);
  if(bound.find("b")!=string::npos)
    bc = (BoundaryConditions)(bc | bcBack);
  bound = GetOption("-nc=", argc, argv, string("u"), "Neuman boundary conditions (same format as Dirichlet ones).");
  
  
  if(bound.find("l")!=string::npos)
    nc = (BoundaryConditions)(nc | bcLeft);
  if(bound.find("r")!=string::npos)
    nc = (BoundaryConditions)(nc | bcRight);
  if(bound.find("d")!=string::npos)
    nc = (BoundaryConditions)(nc | bcDown);
  if(bound.find("u")!=string::npos)
    nc = (BoundaryConditions)(nc | bcUp);
  if(bound.find("f")!=string::npos)
    nc = (BoundaryConditions)(nc | bcFront);
  if(bound.find("b")!=string::npos)
    nc = (BoundaryConditions)(nc | bcBack);

  int mhh= max(N1, N2);
  mhh = max(mhh,N3);
  
  //  const int BufSize =  mhh*mhh*8*4*2*2;
  //  std::vector<char> buf;
  //  buf.resize(BufSize);
  //  MPI_Buffer_attach(&buf[0], BufSize);
  //  std::cout <<"bc hmm" <<(int) bc << " " <<(int)nc << std::endl;
  //  if (method==smPCG3PBC);
    
  if (exithelp)
  {
    throw "help";
  }
}

void ParceOptions(int argc, char **argv)
{
  Help = false;
  exithelp = Help = (OptionPresent("--help", argc, argv));
  Help = Help && (MPI::COMM_WORLD.Get_rank()==0);

  if (Help)
    std::cout << "Usage:\n" << argv[0] << " [option]*" << std::endl;
  

  Options::GetOptions().np = GetOption("-np=", argc, argv, 4, "Number of processes to run on (to spawn +1)");
  Options::GetOptions().spawned = OptionPresent("-spawn", argc, argv, "for internal use when spawning");

  
}

