#define MPICH_IGNORE_CXX_SEEK
#include "basic.h"
#include "mpi.h"
// IBM workaround
#ifdef _MY_CPP
#include "mympicxx.hpp"
#endif

#include<typeinfo>
#include <iostream>
#include <cstdlib>
#include <typeinfo>
#include <iomanip>
#include <cmath>
#include <vector>
#include <string>
#include <sstream>
#include <stdexcept>

#ifdef _OPENMP
#include <omp.h>
#endif

#include "elastgen.h"
#include "util.h"
#include "useful.h"
#include "sparcemmat.h"
#include "elastsolver.h"
#include "optionsparser.h"
#include <fenv.h>

template<class Double>
void DoTheWork(MPI::Intracomm &comm)
{
  ElastWorker<Double, int> worker(comm);
  int iterations;
  {
    TimeCalc1("!Total time");
    TimeCalc1p("System generation",2);
    worker.Sysgen();
  }
  {
    TimeCalc1("!Total time");
    iterations=worker.Solve();

    worker.CheckSolution();
    if (Options::GetOptions().PBC) // fixup solution for the periodic boundary
    {
      worker.x.AddBoundarys(0);
    }
    TimeCalc1p("Solution output",2);
    if (Options::GetOptions().SaveSolution)
       worker.WriteSolution();
   }
   ManageTimes::Get("!Total solve") = ManageTimes::Get("!Solution time")
     + ManageTimes::Get("PreconConstruct");
   int rank = comm.Get_rank();
   ManageTimes::Reduce(comm);
   if(rank==0)
   {
     std::ofstream f((Options::GetOptions().scratchpath+"/"+Options::GetOptions().outputname).c_str());
     using namespace std;
     int fcs = 30;
     int scs = 30;
     int proc = 1;
     #ifdef _OPENMP
     proc =    omp_get_max_threads();
     #endif
     f << "Starting  on " 
       << proc << " (omp) processors per rank. Eps is "<< Options::GetOptions().eps<< endl;

     f << setw(fcs) << left << "Problem size "
       << Options::GetOptions().N1 << "x"
       << Options::GetOptions().N2 << "x"
       << Options::GetOptions().N3 << " " << Myf<Double>::ToString() << " "
       << worker.algname <<endl;
     f << setw(fcs) << left  << "Running on "  
       << setw(scs) <<comm.Get_size()  << "procesors." << endl;
     f << setw(fcs) << "Iterations" 
       << setw(scs) << iterations << endl;
     ManageTimes::PrintAll(f,1);
     ManageTimes::PrintAll(f,iterations," pi");
     f <<left<< setw(30) << "Matrix error"
       << setw(30)<< worker.merr << "\n";
     f <<left<< setw(30) << "Solution error"
       << setw(30)<< worker.serr << "\n";

   }
// #pragma omp parallel
// {
//     #pragma omp critical
//     {
// 	std::cout << rank << "x" <<   omp_get_thread_num() << std::endl;
//     }
// }

}


 int main(int argc, char **argv)
 {
   //   feenableexcept(FE_INVALID | FE_OVERFLOW);
   MPI::Init(argc, argv);
   int rank = 0, size=1;
   int exitcode = 0;
   {
     Myf<SmallVector<double,3> > double3;
     Myf<SmallVector<double,3> >::Geta() = &double3;
     Myf<SmallVector<float ,3> > float3;
     Myf<SmallVector<float,3> >::Geta() = &float3;

     Myf<SmallVector<double,6> > double6;
     Myf<SmallVector<double,6> >::Geta() = &double6;
     Myf<SmallVector<float ,6> > float6;
     Myf<SmallVector<float,6> >::Geta() = &float6;

 //     std::vector<char const *> argvn;
 //     argvn.resize(argc+2);
 //     int c = 0;
 //     for(;c<argc-1;c++)
 //       argvn[c] = argv[c+1];
 //     //    argvn[c++] = "-spawn";
 //     argvn[c] = NULL;

     try{    
       ParceOptions(argc, argv);
      MPI::Intracomm comm = MPI::COMM_WORLD;
      Init(comm, argc, argv);
      rank = comm.Get_rank();
      size = comm.Get_size();
//       int ptos = Options::GetOptions().np -1;
//       MPI::Intercomm icomm = MPI::COMM_NULL;
//       icomm = MPI::COMM_WORLD.Get_parent().Dup();
//       Options::GetOptions().spawned = icomm!=MPI::COMM_NULL;
//       MPI::Intracomm comm = MPI::COMM_NULL;
//       if (!Options::GetOptions().spawned)
//       if (ptos >0)
//       {
	
// 	icomm  = MPI::COMM_WORLD.Spawn(argv[0], &argvn[0],
// 						     ptos,
// 						     MPI::INFO_NULL, 
// 						     0
// 				       );
//       }

//       if (icomm!=MPI::COMM_NULL)
// 	icomm.Disconnect();

      //     if (OptionPresent("-c", argc, argv))
      //       DoTheWork<OpCounter<double> >(N1, N2, N3, method, eps, distype, bc, zero, usebone);
      //     else
#ifdef _FLOAT_EXPAND
      if (OptionPresent("-f", argc, argv))
	DoTheWork<float> (comm);
      else
#endif
	DoTheWork<double>(comm);
      //      comm.Disconnect();
      //      icomm.Disconnect();
    }catch(MPI::Exception &e)
    {
      cout << size << " " 
	   << rank << " \n" 
	   << "MPI Exception caught ... \n"
	   << e.Get_error_string() << endl;
      exitcode = 1;
    }catch(std::exception &e)
    {
      cout << size << " " 
	   << rank << " \n"
	   << "std::exception caught ... \n" 
	   << e.what() << endl;
      exitcode = 2;
    }
    catch(const std::string &e)
    {
      cout << size << " " 
	   << rank << " \n"
	   << "string exception cought ... \n" 
	   << e << endl;
      exitcode = 2;
    }
    catch(const char * e)
    {
      cout
	//	<< size << " " 
	//      << rank << " \n"
	//	<< "string* exception cought ... \n" 
	   << e << endl;
      exitcode = 2;
    }
    catch(...)
    {
      cout << size << " " 
	   << rank << " \n"
	   << "Unknown exception caught.... \n" 
	   << "Rethrowing..." << endl;
      exitcode = 3; // probably will be other since we rethrow the except.
      throw;
    }


  }
  MPI::Finalize();
  //   OpCounter<double>::Save();
  //   cout << "adds " << OpCounter<double>::ocnt_add << endl;
  //   cout << "muls " << OpCounter<double>::ocnt_mul << endl;
  //   cout << "divs " << OpCounter<double>::ocnt_div << endl;
  //   cout << "cmps " << OpCounter<double>::ocnt_cmp << endl;
  return exitcode;
}
