#ifndef __precon_hpp__
#define __precon_hpp__

#include "precon.h"
#include "index.h"
#include "visitors.h"
#include "functors.h"
#include "useful.h"
template<class Vector>
void
NoPrecon<Vector>::
PreconSolve(Vector &res, Vector &src)const
{
  VecVecOper<VecCpy>(res, src);   // simple preconditioning
}




template<class Matrix, class Vector, class PreconVector, class PreconMethod>
void
MIC0Precon<Matrix, Vector, PreconVector, PreconMethod>::
Create(Matrix &mat)
{
  TimeCalc1("PreconditionerCreation");
  Mat = &mat;
  P.SetSize(mat.N1, mat.N2, mat.N3, mat.N3R);
  P.SetComm(*mat.comm);
  P.CreateDatatypes();

  PreconCreatorStep1<Matrix, PreconVector, PreconMethod> st1(mat, P);
  PreconCreatorStep1_5<Matrix, PreconVector, PreconMethod> st1_5(mat, P);
  PreconCreatorStep2<Matrix, PreconVector, PreconMethod> st2(mat, P);
  Index3dBase<typename Vector::int_type>
    ind(P.N1, P.N2, P.N3);

  PrintFunctor<PreconVector> pf(P);
  Execute1(ind, st1, P.size, P.rank);
  //  Execute<ForAll>(ind, pf, P.size, P.rank);
  Execute1(ind, st1_5, P.size, P.rank);
  P.ClearOthers();
  Execute1(ind, st2, P.size, P.rank);
}



template<class Matrix, class Vector, class PreconVector, class PreconMethod>
void
MIC0Precon<Matrix, Vector, PreconVector, PreconMethod>::
PreconSolve(Vector &res, Vector &src)const
{
  TimeCalc1("PreconSolve");
  std::string commtimepreconditioner = "PreconSolveComm";
    
  Index3dBase<typename Vector::int_type>
    ind(P.N1, P.N2, P.N3);
  PreconSolverStep1_1_5<Matrix, PreconVector, Vector, Vector, PreconMethod> 
    pcs115(*Mat, P, res, src, commtimepreconditioner);
  PreconSolverStep2<Matrix, PreconVector, Vector, Vector, PreconMethod> 
    pcs2(*Mat, P, res, src, commtimepreconditioner);

  pcs115.Begin();
  Execute1(ind, pcs115, P.size, P.rank);
  res.ClearOthers();
  Execute1(ind, pcs2, P.size, P.rank);
}


template<class Matrix, class Vector, class PreconVector, class PreconMethod>
void
MIC0PreconM4<Matrix, Vector, PreconVector, PreconMethod>::
Create(Matrix &mat)
{
  MIC0Precon<Matrix, Vector, PreconVector, PreconMethod>::Create(mat);
  this->P.StartSendRecv();
  this->P.FinishSendRecv();
}

template<class Matrix, class Vector, class PreconVector, class PreconMethod>
void
MIC0PreconM4<Matrix, Vector, PreconVector, PreconMethod>::
PreconSolve(Vector &res, Vector &src)const
{
  TimeCalc1("PreconSolve");
  std::string commtimepreconditioner = "PreconSolveComm";
  Index3dBase<typename Vector::int_type>
    ind(this->P.N1, this->P.N2, this->P.N3);
  PreconSolverStep1_1_5<Matrix, PreconVector, Vector, Vector, PreconMethod> 
    pcs115(*this->Mat, this->P, res, src, commtimepreconditioner);
  PreconSolverStep2<Matrix, PreconVector, Vector, Vector, PreconMethod> 
    pcs2(*this->Mat, this->P, res, src, commtimepreconditioner);
  src.StartSendRecv();
  src.FinishSendRecv();
  pcs115.Begin();
  Execute1(ind, pcs115, this->P.size, this->P.rank);
  res.ClearOthers1();
  Execute1(ind, pcs2, this->P.size, this->P.rank);
}







template<class Matrix, class Vector, class PreconVector, class PreconMethod>
void
MIC0PreconPBC<Matrix, Vector, PreconVector, PreconMethod>::
Create(Matrix &mat)
{
  TimeCalc1("PreconditionerCreation");
  Mat = &mat;
  P.SetSize(mat.N1, mat.N2, mat.N3, mat.N3R);
  P.SetComm(*mat.comm);
  P.CreateDatatypes();
  P.SetPeriodic(true);

  mat.CreateDatatypes();
  mat.StartSendRecv();
  mat.FinishSendRecv();
  mat.SetPeriodic(true);

  PreconCreatorStep1<Matrix, PreconVector, PreconMethod> st1(mat, P);
  PreconCreatorStep1_5<Matrix, PreconVector, PreconMethod> st1_5(mat, P);
  PreconCreatorStep2<Matrix, PreconVector, PreconMethod> st2(mat, P);
  Index3dBase<typename Vector::int_type>
    ind(P.N1, P.N2, P.N3);


  ind.ON3=P.ON3;

  Execute1(ind, st1, P.size, P.rank);
  P.AddParts();
  P.AddSides(1);
  mat.AddBoundarys(1);
  mat.AddBoundarys(0);
  mat.DoCyclo(0);
  mat.DoCyclo(2);




//   {
//   PrintFunctor<Matrix> Printer(mat);
//   for (int rank =0; rank < P.size; ++rank)
//   {
//     if(P.rank==rank)
//     {
//       std::cout << rank << std:: endl << std::endl;
//       Execute<ForAllPBC>(ind, Printer, P.size, P.rank);
//     }
//     mat.comm->Barrier();
//   }
//    exit(1);
//   }



  
  Execute1(ind, st1_5, P.size, P.rank);
  P.ClearOthers2();
  Execute1(ind, st2, P.size, P.rank);
//   PrintFunctor<PreconVector> Printer(P);
//   for (int rank =0; rank < P.size; ++rank)
//   {
//     if(P.rank==rank)
//     {
//       std::cout << rank << std:: endl << std::endl;
//       Execute<ForAllPBC>(ind, Printer, P.size, P.rank);
//     }
//     mat.comm->Barrier();
//   }
//   exit(1);
}



template<class Matrix, class Vector, class PreconVector, class PreconMethod>
void
MIC0PreconPBC<Matrix, Vector, PreconVector, PreconMethod>::
PreconSolve(Vector &res, Vector &src)const
{
  TimeCalc1("PreconSolve");
  std::string commtimepreconditioner = "PreconSolveComm";

    
  Index3dBase<typename Vector::int_type>
    ind(P.N1, P.N2, P.N3);
  PreconSolverStep1_1_5<Matrix, PreconVector, Vector, Vector, PreconMethod> 
    pcs115(*Mat, P, res, src, commtimepreconditioner);
  PreconSolverStep2<Matrix, PreconVector, Vector, Vector, PreconMethod> 
    pcs2(*Mat, P, res, src, commtimepreconditioner);


  //  ClearVector(src, 1);
  src.SetPeriodic(true);
  res.SetPeriodic(true);
  
  src.AddBoundarys(0);
  src.DoCyclo(0);
  src.ClearOthers2();
  ind.ON3 = src.ON3; 
  
  pcs115.Begin();
  Execute1(ind, pcs115, P.size, P.rank);
  res.ClearOthers2();



  Execute1(ind, pcs2, P.size, P.rank);
//   PrintFunctor<Vector> Printer(res);
//   for (int rank =0; rank < P.size; ++rank)
//   {
//     if(P.rank==rank)
//     {
//       std::cout << rank << std:: endl << std::endl;
//       Execute<ForAllPBC>(ind, Printer, P.size, P.rank);
//     }
//     P.comm->Barrier();
//   }
//   exit(1);
  src.DoCyclo(1);
  res.DoCyclo(1);
  src.SetPeriodic(false);
  res.SetPeriodic(false);
  src.AddBoundarys(0);
  src.AddBoundarys(2);
  res.AddBoundarys(2);
}




#endif
