#include <string>
#include <fstream>
#include <stdexcept>
#include <iostream>

using namespace std;



class EpsWriter:public ofstream
{
	double fheight;
	double xs, xw, ys, yw;
public:
	void WriteHeader()
	{
		*this << "%!PS-Adobe-3.0 EPSF-3.0\n";
		*this << "%%BoundingBox: " << 0 << " " << 0 << " " << xw+xs*2 << " " << yw + ys*2<< "\n";
		*this << "save\n";
	}
	void WriteEnder()
	{
		*this << "restore" << endl;
	}
	EpsWriter (string filename, double width, double height):ofstream(filename.c_str(), ios::binary)
	{
		fheight = 0;
		if (!*this)
			throw runtime_error(string("Cannot open \"") + filename + "\".");
		xs = ys = 10;
		xw = width;
		yw = height;
		WriteHeader();
	}
	void SetFont(const string&fontname, double fontheight)
	{
		*this <<"(" << fontname << ") findfont " << (fheight = fontheight*yw) << " scalefont setfont\n";
	}
	void NewPath()
	{
		*this << "newpath\n";
	}
	void Fill()
	{
	  *this << "fill\n";
	}
	void FillRect(double x1, double y1, double x2, double y2)
	{
	  NewPath();
	  MoveTo(x1, y1);
	  LineTo(x1, y2);
	  LineTo(x2, y2);
	  LineTo(x2, y1);
    ClosePath();
	  Fill();
	}
	void ClosePath()
	{
		*this << "closepath\n";
	}
	void Stroke()
	{
		*this << "stroke\n";
	}
	void MoveTo(double x, double y)
	{
		*this << x*xw + xs << " " << y*yw + ys << " moveto\n";
	}
	void LineTo(double x, double y)
	{
		*this << x*xw + xs << " " << y*yw + ys << " lineto\n";
	}
	void DrawLine(double x1, double y1, double x2, double y2)
	{
		NewPath();
		MoveTo(x1, y1);
		LineTo(x2, y2);
		Stroke();
	}
	void Arc(double x1, double y1, double radius, double sa = 0., double ea = 360.)
	{
		*this << x1*xw+xs << " " << y1*yw+ys << " " << " " << radius*xw << " " << sa << " " << ea << " arc\n";
	}
	void DrawArc(double x1, double y1, double radius, double sa = 0., double ea = 360.)
	{
		NewPath();
		Arc(x1, y1, radius, sa, ea);
		Stroke();
	}

	void DrawString(double x1, double y1, const string &what)
	{
		*this << "(" << what << ") ";
		*this << "dup  stringwidth pop  -2 div " << x1*xw+xs 
			<< " add  " << ys + yw*y1 - fheight/2 << " moveto show\n";
	}

	void SetColor(double r, double g, double b)
	{
		*this << r << " " << g << " " << b << " setrgbcolor\n";
	}

	void LineWidth(double width)
	{
		*this << width << " setlinewidth\n";
	}


	~EpsWriter()
	{
		WriteEnder();
	}
};

void OutputEl(EpsWriter&eps, int i, int j, int N)
{
  double y1 = (double)i/N;
  double x1 = (double)j/N;
  double y2 = (double)(i+1)/N;
  double x2 = (double)(j+1)/N;

  /*  x1*=5;
  x2*=5;
  y1*=5;
  y2*=5;  
  x1-=3;
  x2-=3;
  y1-=3;
  y2-=3;
  if((x1>=0)&&(y2>=0)&&(x2<=1)&&(y1<=1))
  */
  y1 = 1-y1;
  y2 = 1-y2;
  eps.FillRect(x1, y1, x2, y2);
}

void GenAll()
{
	EpsWriter eps("proba.eps", 200, 200);
  int N = 100;
  for(int i = 0; i < N; ++i)
  {
    OutputEl(eps, i, i, N);
    OutputEl(eps, i, i+2, N);
    OutputEl(eps, i+1, i, N);
  }
}

int N1 = 2;
int N2 = 2;
int N3 = 6;

enum Pos{Left , Front , Down};

int GetTotalCount()
{
  return N1*N2*N3*3 + N1*N2 + N2*N3 + N1*N3;
}
int N = GetTotalCount();
int IndFromPos(int i, int j, int k, Pos p)
{
  int ind = 0;
  if(p==Left)
  {
    ind += (N2*N3*3+N2+N3)*i ;
    ind += j*N3;
    ind += k;
    return ind;
  }
  if(p==Front)
  {
    ind += (N2*N3*3+N2+N3)*i+ N2*N3;
    ind += j*N3 + j*(N3+1);
    ind += k;
    return ind;
  }
  if(p==Down)
  {
    ind += (N2*N3*3+N2+N3)*i+ N2*N3;
    ind += j*N3 + j*(N3+1);
    ind += k + N3;
    return ind;
  }
  return ind;
}
void PosFromInd(int ind, int &i, int &j, int &k, Pos &p)
{
  i = ind/(N2*N3*3+N2+N3);
  ind -= i *(N2*N3*3+N2+N3);
  if (ind < (N2*N3))
  {
    p = Left;
    j = ind/N3;
    k = ind%N3;
    return;
  }
  ind -= N2*N3;
  j = ind/(N3+N3+1);
  k = ind%(N3+N3+1);
  if (k>=N3)
  {
    p = Down;
    k-= N3;
  }else
    p = Front;
}

bool IfPosInBounds(int i, int j, int k, Pos p)
{
  if (( i <0) || (j<0) || (k <0))
    return false;
  if ((i>N1) || ((i==N1)&&(p!=Left)))
    return false;
  if ((j>N2) || ((j==N2)&&(p!=Front)))
    return false;
  if ((k>N3) || ((k==N3)&&(p!=Down)))
    return false;
  return true;
}


bool b1 = false;
bool b2 = false;

template<class F>
void Do(int ind, F &f)
{
  int i, j, k;
  
  int i1, j1, k1;
  Pos p, p1;
  double allc = 0;
  double v1olny = 0.8;
  double v2olny = 0.6;
  f.eps.SetColor(allc, allc, allc);
  PosFromInd(ind, i, j, k, p);
  f(ind, ind);
  if (p == Left)
  {
    i1 = i;
    j1 = j;
    k1 = k;
    p1 = Front;
    if (IfPosInBounds(i1, j1, k1, p1))
      f(ind, IndFromPos(i1, j1, k1, p1));
    i1 = i;
    j1 = j;
    k1 = k;
    p1 = Down;
    if (IfPosInBounds(i1, j1, k1, p1))
      f(ind, IndFromPos(i1, j1, k1, p1));
    i1 = i;
    j1 = j+1;
    k1 = k;
    p1 = Front;
    if (IfPosInBounds(i1, j1, k1, p1))
      f(ind, IndFromPos(i1, j1, k1, p1));
    i1 = i;
    j1 = j;
    k1 = k+1;
    p1 = Down;
    if (IfPosInBounds(i1, j1, k1, p1))
      f(ind, IndFromPos(i1, j1, k1, p1));

    i1 = i-1;
    j1 = j;
    k1 = k;
    p1 = Front;
    if (IfPosInBounds(i1, j1, k1, p1))
      f(ind, IndFromPos(i1, j1, k1, p1));
    i1 = i-1;
    j1 = j;
    k1 = k;
    p1 = Down;
    if (IfPosInBounds(i1, j1, k1, p1))
      f(ind, IndFromPos(i1, j1, k1, p1));
    i1 = i-1;
    j1 = j+1;
    k1 = k;
    p1 = Front;
    if (IfPosInBounds(i1, j1, k1, p1))
      f(ind, IndFromPos(i1, j1, k1, p1));
    i1 = i-1;
    j1 = j;
    k1 = k+1;
    p1 = Down;
    if (IfPosInBounds(i1, j1, k1, p1))
      f(ind, IndFromPos(i1, j1, k1, p1));

    if(not b1)
      {
	f.eps.SetColor(v2olny, v2olny, v2olny);
	i1 = i+1;
	j1 = j;
	k1 = k;
	p1 = Left;
	if (IfPosInBounds(i1, j1, k1, p1))
	  f(ind, IndFromPos(i1, j1, k1, p1));
	i1 = i-1;
	j1 = j;
	k1 = k;
	p1 = Left;
	if (IfPosInBounds(i1, j1, k1, p1))
	  f(ind, IndFromPos(i1, j1, k1, p1));
      }
  }

  if (p == Down)
  {
    i1 = i;
    j1 = j;
    k1 = k;
    p1 = Left;
    if (IfPosInBounds(i1, j1, k1, p1))
      f(ind, IndFromPos(i1, j1, k1, p1));
    i1 = i+1;
    j1 = j;
    k1 = k;
    p1 = Left;
    if (IfPosInBounds(i1, j1, k1, p1))
      f(ind, IndFromPos(i1, j1, k1, p1));
    i1 = i;
    j1 = j;
    k1 = k-1;
    p1 = Left;
    if (IfPosInBounds(i1, j1, k1, p1))
      f(ind, IndFromPos(i1, j1, k1, p1));
    i1 = i+1;
    j1 = j;
    k1 = k-1;
    p1 = Left;
    if (IfPosInBounds(i1, j1, k1, p1))
      f(ind, IndFromPos(i1, j1, k1, p1));

    if(not b2)
      {
	f.eps.SetColor(v1olny, v1olny, v1olny);
	i1 = i;
	j1 = j;
	k1 = k;
	p1 = Front;
	if (IfPosInBounds(i1, j1, k1, p1))
	  f(ind, IndFromPos(i1, j1, k1, p1));
	i1 = i;
	j1 = j+1;
	k1 = k;
	p1 = Front;
	if (IfPosInBounds(i1, j1, k1, p1))
	  f(ind, IndFromPos(i1, j1, k1, p1));
	i1 = i;
	j1 = j;
	k1 = k-1;
	p1 = Front;
	if (IfPosInBounds(i1, j1, k1, p1))
	  f(ind, IndFromPos(i1, j1, k1, p1));
	i1 = i;
	j1 = j+1;
	k1 = k-1;
	p1 = Front;
	if (IfPosInBounds(i1, j1, k1, p1))
	  f(ind, IndFromPos(i1, j1, k1, p1));
      }
    if(not b1 and not b2)
      {
	f.eps.SetColor(v2olny, v2olny, v2olny);
	i1 = i;
	j1 = j;
	k1 = k+1;
	p1 = Down;
	if (IfPosInBounds(i1, j1, k1, p1))
	  f(ind, IndFromPos(i1, j1, k1, p1));
	i1 = i;
	j1 = j;
	k1 = k-1;
	p1 = Down;
	if (IfPosInBounds(i1, j1, k1, p1))
	  f(ind, IndFromPos(i1, j1, k1, p1));
      }
  }

  if (p == Front)
  {
    i1 = i;
    j1 = j;
    k1 = k;
    p1 = Left;
    if (IfPosInBounds(i1, j1, k1, p1))
      f(ind, IndFromPos(i1, j1, k1, p1));
    i1 = i+1;
    j1 = j;
    k1 = k;
    p1 = Left;
    if (IfPosInBounds(i1, j1, k1, p1))
      f(ind, IndFromPos(i1, j1, k1, p1));
    i1 = i;
    j1 = j-1;
    k1 = k;
    p1 = Left;
    if (IfPosInBounds(i1, j1, k1, p1))
      f(ind, IndFromPos(i1, j1, k1, p1));
    i1 = i+1;
    j1 = j-1;
    k1 = k;
    p1 = Left;
    if (IfPosInBounds(i1, j1, k1, p1))
      f(ind, IndFromPos(i1, j1, k1, p1));

    if(not b2)
      {
	f.eps.SetColor(v1olny, v1olny, v1olny);
	i1 = i;
	j1 = j;
	k1 = k;
	p1 = Down;
	if (IfPosInBounds(i1, j1, k1, p1))
	  f(ind, IndFromPos(i1, j1, k1, p1));
	i1 = i;
	j1 = j;
	k1 = k+1;
	p1 = Down;
	if (IfPosInBounds(i1, j1, k1, p1))
	  f(ind, IndFromPos(i1, j1, k1, p1));
	i1 = i;
	j1 = j-1;
	k1 = k;
	p1 = Down;
	if (IfPosInBounds(i1, j1, k1, p1))
	  f(ind, IndFromPos(i1, j1, k1, p1));
	i1 = i;
	j1 = j-1;
	k1 = k+1;
	p1 = Down;
	if (IfPosInBounds(i1, j1, k1, p1))
	  f(ind, IndFromPos(i1, j1, k1, p1));
      }

    if(not b1 and not b2)
      {
	f.eps.SetColor(v2olny, v2olny, v2olny);
	i1 = i;
	j1 = j+1;
	k1 = k;
	p1 = Front;
	if (IfPosInBounds(i1, j1, k1, p1))
	  f(ind, IndFromPos(i1, j1, k1, p1));
	i1 = i;
	j1 = j-1;
	k1 = k;
	p1 = Front;
	if (IfPosInBounds(i1, j1, k1, p1))
	  f(ind, IndFromPos(i1, j1, k1, p1));
      }
  }

}

template<class ER>
void DrawSingleLine(ER&eps, int&prev, int count)
{
  int tmp;
  double o;
  o = prev/double(N);
  eps.DrawLine(0,1-o,1,1-o);
  eps.DrawLine(o,0,o,1);
  prev += count;
}
class Write
{
int N;
public:
EpsWriter eps;
public:
  Write():N(GetTotalCount()), eps("proba.eps", 320, 320){}
  bool small;
void operator()(int i, int j)
{
  OutputEl(eps, i, j, N);
  //  OutputEl(eps, j, i, N);
}
  void DrawMesh()
  {
    eps.LineWidth(0.2);
    eps.NewPath();
    eps.MoveTo(0,0);
    eps.LineTo(1,0);
    eps.LineTo(1,1);
    eps.LineTo(0,1);
    eps.LineTo(0,0);
    eps.Stroke();
    int prev = 0;
    //    if(small)
      {
	
	for(int i = 0; i < N1;++i)
	  {
	    for(int j = 0; j < N2; ++j)
	      DrawSingleLine(eps, prev, N3);
	    for(int j = 0; j < N2; ++j)
	      {
		DrawSingleLine(eps, prev, N3);
		DrawSingleLine(eps, prev, N3+1);
	      }
	    DrawSingleLine(eps, prev, N3);
	  }
	for(int j = 0; j < N2; ++j)
	  DrawSingleLine(eps, prev, N3);
      }
      //    else
    eps.LineWidth(0.8);
    prev = 0;
      {
	for(int i = 0; i < N1; ++i)
	  {
	    DrawSingleLine(eps, prev, N3*N2);
	    DrawSingleLine(eps, prev, N3*(N2+1)+N2*(N3+1));
	  }
	DrawSingleLine(eps, prev, N3*N2);
	DrawSingleLine(eps, prev, N3*(N2+1)+N2*(N3+1));
      }
  }
};

int main(int argc, char *argv[])
{
  int N = GetTotalCount();
  
  int i, j, k;
  Pos p;

  Write W;
  b1 = false;
  W.small = true;
  W.DrawMesh();
  
  for (int cnt = 0; cnt < N; ++cnt)
  {
    Do(cnt, W);
  }
  

	return 0;
}

