#ifndef __small_h__
#define __small_h__
#include <iostream>

// template<class T, int size>
// class StackStorage
// {
//  protected:
//   T buf[size];
// public:
//   StackStorage()
//   {
//     for(int i = 0; i < size; ++i)
//       buf[i] = T();
//   }
// private:
//   void  operator = (const StackStorage&);
// };
// template<class T, int size>
// class HeapStorage
// {
// protected:
//   T * const buf;
//   HeapStorage()
//     :buf(new T[size])
//   {
//     for(int i = 0; i < size; ++i)
//       buf[i] = T();
//   }
//   HeapStorage(const HeapStorage &a)
//     :buf(new T[size])
//   {
//     for (int i = size;i--;)
//       buf[i] = a.buf[i];
//   }
//   ~HeapStorage()
//   {
//     delete[] buf;
//   }
// private:
//   void  operator = (const HeapStorage&);
// };
struct StackStorage{};
struct HeapStorage{};

template<class  T, int size, class what>
class StorageHelper;

template<class T, int size>
class StorageHelper<T, size, StackStorage>
{
 protected:
  T buf[size];
public:
  StorageHelper()
  {
    for(int i = 0; i < size; ++i)
      buf[i] = T();
  }
private:
  void  operator = (const StorageHelper&);
};

template<class T, int size>
class StorageHelper<T, size, HeapStorage>
{
protected:
  T * const buf;
  StorageHelper()
    :buf(new T[size])
  {
    for(int i = 0; i < size; ++i)
      buf[i] = T();
  }
  StorageHelper(const StorageHelper &a)
    :buf(new T[size])
  {
    for (int i = size;i--;)
      buf[i] = a.buf[i];
  }
  ~StorageHelper()
  {
    delete[] buf;
  }
private:
  void  operator = (const HeapStorage&);
};

template<class T, int size,  class Storage = StackStorage>
class SmallVector:public StorageHelper<T, size, Storage>
{
 protected:
  // buf is in storage, let's use it
  //  T buf[size];
public:
  using StorageHelper<T, size, Storage>::buf;
  T * GetBuf()
    {
      return buf;
    }
  T const * GetBuf()const
    {
      return buf;
    }
  
  SmallVector(const T& initvalue){
    for(int i = 0; i < size; ++i)
      buf[i] = initvalue;
  }
  SmallVector(const SmallVector &a)
    {
      for(int i = 0; i < size; ++i)
	buf[i] = a.buf[i];
    }
  SmallVector(){}

  void operator =(const T& initvalue)
    {
    for(int i = 0; i < size; ++i)
      buf[i] = initvalue;
    //    return initvalue;
    } 


  SmallVector operator* (const T&val)const
  {
    SmallVector res;
    for(int i = 0; i < size; ++i)
    res[i] = buf[i] * val;
    return res;
  }
  SmallVector operator* (const SmallVector&val)const
  {
    SmallVector res;
    for(int i = 0; i < size; ++i)
    res[i] = buf[i] * val[i];
    return res;
  }
  SmallVector operator- (const SmallVector&val)const
  {
    SmallVector res;
    for(int i = 0; i < size; ++i)
    res[i] = buf[i] - val[i];
    return res;
  }
  
  void operator *= (const SmallVector &v)
    {
      for(int i = 0; i < size; ++i)
	buf[i] *= v.buf[i];
    }
  void operator /= (const SmallVector &v)
    {
      for(int i = 0; i < size; ++i)
	buf[i] /= v.buf[i];
    }
  void operator += (const T& val)
    {
      for(int i = 0; i < size;++i)
	buf[i] += val;
    }

  void operator -= (const T& val)
    {
      *this += -val;
    }
  void operator *= (const T& val)
    {
      for(int i = 0; i < size; ++i)
	buf[i] *= val;
    }

  void operator /= (const T& val)
    {
      *this *= T(1.0/val);
    }

  void operator -= (const SmallVector &v)
    {
      for(int i = 0; i < size; ++i)
	buf[i] -= v.buf[i];
    }
  void operator += (const SmallVector &v)
    {
      for(int i = 0; i < size; ++i)
	buf[i] += v.buf[i];
    }
  void operator = (const SmallVector &v)
    {
      for(int i = 0; i < size; ++i)
	buf[i] = v.buf[i];
    }
  template<class Other, class Bother>
  void operator = (const SmallVector<Other, size, Bother> &v)
  {
      for(int i = 0; i < size; ++i)
	buf[i] = static_cast<T>(v.buf[i]);
  }
  

  typedef T Element_type;
  const T& operator[](int i)const {return buf[i];}
  T&operator[](int i)      {return buf[i];}
  int Size()const {return size;}
};

template<class T, int size, class Storage>
  SmallVector<T, size, Storage> operator * (const T& A, const SmallVector<T, size, Storage>&B)
  {
    SmallVector<T, size, Storage> Res(B);
    Res *= A;
    return Res;
  }
template<class T, int size, class Storage>
  SmallVector<T, size, Storage> operator / (const T& A, const SmallVector<T, size, Storage>&B)
  {
    SmallVector<T, size, Storage> Res(B);
      for(int i = 0; i < size; ++i)
	Res[i] = A/B[i];
    return Res;
  }
template<class T, int size, class Storage, class V>
  SmallVector<T, size, Storage> operator < (const SmallVector<T, size, Storage>&A, const V&B)
  {
    SmallVector<T, size, Storage> Res;
    for(int i = size; i--;)
      Res[i] = A[i] < B;
    return Res;
  }
    
template<class T, int size, class Storage, class V>
  SmallVector<T, size, Storage> operator > (const SmallVector<T, size, Storage>&A, const V&B)
  {
    //    std::cout << "mamka mu 2" << std::endl;
    SmallVector<T, size, Storage> Res;
    for(int i = size; i--;)
      Res[i] = A[i] > B;
    return Res;
  }


template<class T, int size, class Storage>
std::ostream& operator <<(std::ostream &out, SmallVector<T, size, Storage> const &v)
{
  for(int i = 0; i < size; ++i)
    out << v[i] << " " ;
  return out;

}


template<class T, int Rows, int Columns, 
  class Storage = StackStorage>
class SmallMatrix:public SmallVector< SmallVector<T, Columns>, Rows, Storage>
{
  using SmallVector< SmallVector<T, Columns>, Rows, Storage>::buf;
public:
  typedef T Element_type;
  int Height()const{return Rows;}
  int Width()const {return Columns;}
  void operator =(const T& initvalue)
    {
    for(int i = 0; i < Rows; ++i)
      buf[i] = initvalue;
    //    return initvalue;
    } 

  void operator += (const T& val)
    {
      for(int i = 0; i < Rows;++i)
	buf[i] += val;
    }

  void operator -= (const T& val)
    {
      *this += -val;
    }
  void operator *= (const T& val)
    {
      for(int i = 0; i < Rows; ++i)
	buf[i] *= val;
    }

  void operator /= (const T& val)
    {
      *this *= T(1.0/val);
    }

  void operator += (const SmallMatrix &v)
    {
      for(int i = 0; i < Rows; ++i)
	buf[i] += v.buf[i];
    }

  void operator = (const SmallMatrix &v)
    {
      for(int i = 0; i < Rows; ++i)
	buf[i] = v.buf[i];
    }


};





#endif
