#ifndef __solver_hpp__
#define __solver_hpp__
#include "solver.h"
#include "functors.h"
#include <iostream>
#include <algorithm>
#include "projection.hpp"


//PCGSolverBase

template<class Vector>
PCGSolverBase<Vector>::
PCGSolverBase()
{
  eps=1e-3;
}

template<class Vector>
void
PCGSolverBase<Vector>::
SetEps(Double eps_)
{
  eps=eps_;
}
  


//PCGSolver


template<class Vector>
void
PCGSolver<Vector>::
SetSizes(Int N1_, Int N2_, Int N3_, Int N3R_)
{
  Hd.SetSize(N1_, N2_, N3_, N3R_);
  //  h.SetSize(N1_, N2_, N3_, N3R_);
  if (!Options::GetOptions().reuse_rhs)
    D.SetSize(N1_, N2_, N3_, N3R_);
  G.SetSize(N1_, N2_, N3_, N3R_);  
}

template<class Vector>
void
PCGSolver<Vector>::
SetComm(MPI::Intracomm &comm_)
{
  Hd.SetComm(comm_);
  //  h.SetComm(comm_);
  G.SetComm(comm_);
  //  h.CreateDatatypes();
  Hd.CreateDatatypes();
  if (!Options::GetOptions().reuse_rhs)
  {
    D.SetComm(comm_);
    D.CreateDatatypes();
  }
  if (Options::GetOptions().proj)
  {
    proj.SetComm(comm_);
  }
  G.CreateDatatypes();
}


template<class Vector>
unsigned
PCGSolver<Vector>::
Solve(Vector &X, Vector  &RHS)
{
  RHS.ClearZero();
  if (Options::GetOptions().proj)
  {
     // proj.SetProject(RHS);
     // for (int i = 0; i < 6; ++i)
     // {
     //   VecVecOper<VecCpy>(D, proj.P[i]);
     //   mat->Multiply(RHS, D);
       
     //   double gtg = VecVecReduce<VecDot>(RHS, RHS);    //gtg = G^t h
     //   if (!RHS.rank)
     // 	 std::cout <<"proj " << i << " = "<< gtg << std::endl;
     // }
     proj.SetProject(RHS);
     proj.Project(RHS);
  }
  
  // VecVecOper<VecCpy>(X, RHS);       // return rhs;
  // return 333;
  vector_type &h = Hd;
  vector_type &D = Options::GetOptions().reuse_rhs ? RHS : this->D;

  Double norm=0;
  if (Options::GetOptions().froj)
  {
    norm = VecReduceNorm<CNorm>(RHS);
    norm = sqrt(norm);
    norm = sqrt(norm);
    norm = sqrt(norm);
    norm/=3.*X.N1*X.N2*X.N3;
    //    std::cout << "Noooorm " <<  norm << std::endl;
    RandFunctor<Vector> rf(X, norm);
    Index3dBase<typename Vector::int_type>
      ind(X.N1, X.N2, X.N3);
    Execute<ForAll>(ind, rf, X.size, X.rank);
    mat->Multiply(Hd, X);
    //    ClearVector(X, 0);
    VecVecOper<VecSub>(RHS, Hd);
  }

  
  using namespace std;
  Double Eps = eps*eps;
  Double gtgfull, gtg, gtgold;
  gtgfull = 0;
  Double tau, beta, temp;
  unsigned iter = 0;
  int rank = D.rank;
  bool firsttime = true;
  // again:
  //  VecVecOper<VecCpy>(D, X);        // D = X;
  mat->Multiply(Hd, X);            // Hd = H d;
//     if (Options::GetOptions().proj)
//       proj.Project(Hd);

  VecVecOper<VecCpy>(G, Hd);       // G = Hd;

//   if (Options::GetOptions().proj)
//     proj.Project(RHS);  
  
  VecVecOper<VecSub>(G, RHS);      // G -= RHS;

  if (Options::GetOptions().proj)
  {
     proj.SetProject(G);
     proj.Project(G);
  }
   precon->PreconSolve(h, G);        // h = C^{-1} G
//     if (Options::GetOptions().proj)
//       proj.Project(h);

  ClearVector(D);                  // D = 0;
  VecVecOper<VecSub>(D, h);        // D = -h;
  gtg =   VecVecReduce<VecDot>(G, h);    //gtg = G^t h
  //  std::cout << "Res " << gtg << std::endl;

  if (firsttime)
  {
    gtgfull = gtg;
    //    std::cout << "fool = " << gtgfull << std::endl;
    gtgfull *= Eps;
    firsttime = false;
  }
  double oldprn = 1;
  if(sqrt(gtg) > Options::GetOptions().zero) // when residual is really small we do not calculate anything
  while(gtg > gtgfull)
  {

    double prn;
    if (iter==0)
      prn = sqrt(gtg);
    else
      prn = sqrt(gtg/(gtgfull/Eps));
    gtgold = gtg;
    if (Options::GetOptions().debug)
      if (rank==0)
	cerr <<std::setw(6)<< iter << std::setw(16)  <<prn << std::setw(16)<< oldprn/prn << endl;
    oldprn = prn;


    if (iter > Options::GetOptions().maxiter)
      break;

    mat->Multiply(Hd, D);            // Hd = H d;
//     if (Options::GetOptions().proj)
//     {
//       proj.SetProject(D);
//       proj.Project(Hd);
//     }

    temp = VecVecReduce<VecDot>(D, Hd); //temp = D^t Hd

    tau = gtg/temp;

    VecVecOper<VecUpdate>(X, D, tau);  // X += D*tau;
    ++iter;
    VecVecOper<VecUpdate>(G, Hd, tau); // G += Hd*tau;


    if (Options::GetOptions().proj)
    {
      proj.SetProject(G);
      proj.Project(G);
    }
    precon->PreconSolve(h, G);         // h = C^-1 G;
    if (Options::GetOptions().proj)
    {
      proj.SetProject(h);
      proj.Project(h);
    }

    
    gtg = VecVecReduce<VecDot>(G, h); // gtg = G^t h


    
    beta = gtg/gtgold;
    
    VecVecOper<VecUpdate2>(D, h, beta);//D = D*beta -h;
    //    std::cout << gtg << " " <<gtgfull << std::endl;

  }
  if (Options::GetOptions().reuse_rhs)
    VecVecOper<VecCpy>(RHS, G);       // G = Hd; //return residual here!
  if (Options::GetOptions().froj)
  {
    RandFunctor<Vector> rf(G, norm);
    Index3dBase<typename Vector::int_type>
      ind(X.N1, X.N2, X.N3);
    Execute<ForAll>(ind, rf, X.size, X.rank);
    mat->Multiply(Hd, G);
    if (!Options::GetOptions().reuse_rhs)
      VecVecOper<VecAdd>(RHS, Hd);
    VecVecOper<VecAdd>(X, G);
  }

  return iter;
}




























//GPCGSolver

template<class Vector>
GPCGSolver<Vector>::
GPCGSolver():m(Options::GetOptions().gcg_m), Ds(new vector_type[m+1])
{
  taus.resize(m+1);
  sigmas.resize(m+1);
}


template<class Vector>
void
GPCGSolver<Vector>::
SetSizes(Int N1_, Int N2_, Int N3_, Int N3R_)
{
  Hd.SetSize(N1_, N2_, N3_, N3R_);
  h.SetSize(N1_, N2_, N3_, N3R_);
  G.SetSize(N1_, N2_, N3_, N3R_);  
  
  for (int i = 0; i <= m; ++i)
    Ds[i].SetSize(N1_, N2_, N3_, N3R_);
}

template<class Vector>
void
GPCGSolver<Vector>::
SetComm(MPI::Intracomm &comm_)
{
  Hd.SetComm(comm_);
  h.SetComm(comm_);
  G.SetComm(comm_);

  for (int i = 0; i <= m; ++i)
  {
    Ds[i].SetComm(comm_);
    Ds[i].CreateDatatypes();
  }
  if (Options::GetOptions().proj)
  {
    proj.SetComm(comm_);
  }
  G.CreateDatatypes();
  h.CreateDatatypes();
  Hd.CreateDatatypes();
}


template<class Vector>
unsigned
GPCGSolver<Vector>::
Solve(Vector &X, Vector  &RHS)
{
  // VecVecOper<VecCpy>(X, RHS);       // return rhs;
  // return 333;
  RHS.ClearZero();

  vector_type &D = Ds[0];//  to change
  using namespace std;
  Double Eps = eps*eps;
  Double gtgfull, gtg, gtgold;
  gtgfull = 0;
  Double tau, beta, temp;
  int iter = 0;
  int rank = D.rank;
  bool firsttime = true;
  // again:
  //  VecVecOper<VecCpy>(D, X);        // D = X;
  mat->Multiply(Hd, X);            // Hd = H d;

  VecVecOper<VecCpy>(G, RHS);       // G = RHS;

  
  
  VecVecOper<VecSub>(G, Hd);      // G -= Hd;

  if (Options::GetOptions().proj)
  {
    proj.SetProject(G);
    proj.Project(G);
  }
  precon->PreconSolve(h, G);        // h = C^{-1} G

  VecVecOper<VecCpy>(D, h);        //D = h;
  gtg =   VecVecReduce<VecDot>(G, h);    //gtg = G^t h
  sigmas[0] = gtg;
  if (firsttime)
  {
    gtgfull = gtg;
    //    std::cout << "fool = " << gtgfull << std::endl;
    gtgfull *= Eps;
    firsttime = false;
  }
  double oldprn = 1;
  if(sqrt(gtg) > Options::GetOptions().zero) // when residual is really small we do not calculate anything
  while(gtg > gtgfull)
  {
    CircularIndex<int> ic(m+1);
    CircularIndex<int> ic1(m+1);
    ic1 = iter+1;
    ic = iter;
    vector_type &D = Ds[ic];//  to change

    double prn;
    if (iter==0)
      prn = sqrt(gtg);
    else
      prn = sqrt(gtg/(gtgfull/Eps));
    gtgold = gtg;
    if (Options::GetOptions().debug)
      if (rank==0)
	cerr <<std::setw(6)<< iter << std::setw(16)  <<prn << std::setw(16)<< oldprn/prn << endl;
    oldprn = prn;


    if (iter > Options::GetOptions().maxiter)
      break;

    mat->Multiply(Hd, D);            // Hd = H d;

    temp = VecVecReduce<VecDot>(D, Hd); //temp = D^t Hd

    tau = gtg/temp;
    taus[ic] = tau;
    
    VecVecOper<VecUpdate>(X, D, tau);  // X += D*tau;


    VecVecOper<VecUpdate>(G, Hd, -tau); // G += Hd*tau;

    if (Options::GetOptions().proj)
    {
      proj.SetProject(G);
      proj.Project(G);
    }
    precon->PreconSolve(h, G);         // h = C^-1 G;
    if (Options::GetOptions().proj)
    {
      proj.SetProject(h);
      proj.Project(h);
    }
    
    gtg = VecVecReduce<VecDot>(G, h); // gtg = G^t h
    sigmas[ic1] = gtg;
    VecVecOper<VecCpy>(Ds[ic1], h) ;         // D = h;
    for (int k = 1; k <= std::min(iter+1, m);++k)
    {
      CircularIndex<int> kc(m+1);
      CircularIndex<int> kc1(m+1);
      kc = iter+1-k;
      kc1 = iter+2-k ;
      if (k>0)
      {
	mat->Multiply(Hd, Ds[kc]);            // Hd = H d;
      }
	
      Double beta = -taus[kc] * VecVecReduce<VecDot>(h, Hd)/sigmas[kc];
      //      std::cout << k << " " << beta << std::endl
      VecVecOper<VecUpdate>(Ds[ic1], Ds[kc], beta); // G += Hd*tau;
      if (Options::GetOptions().proj)
      {
	proj.SetProject(Ds[ic1]);
	proj.Project(Ds[ic1]);
      }
      
    }
      

    
//      beta = gtg/gtgold;
    
//      VecVecOper<VecUpdate>(Ds[ic1], Ds[ic], beta); //, h, beta);//D = D*beta -h;

    ++iter;

  }
  return iter;
}


#endif
