#include "elastgen.hpp"

#ifndef USE_EXPORT
// some explicit specializations
//double
template class ElastDiscretizator<double>;


//float
#ifdef _FLOAT_EXPAND
template class ElastDiscretizator<float>;
#endif
#else //export

template<class Double>
SmallVector<Double, 3>
FuncRhs(const SmallVector<Double, 3> &p)
{
  Double const &x = p[0];
  Double const &y = p[1];
  Double const &z = p[2];

  SmallVector<Double, 3> res;

  // variant 5;
  res = Double();
  switch (Options::GetOptions().variant)
  {
  case 5:
  case 0:
    break;
  case 4:
  // variant 4;
    res[0] = -(1.5*(6*x*y*z - 3*sin(x + y + z)) +
	       2.5*(6*x*y*z + 3*y*y*z +
		    3*y*z*z - cos(x + y + z) +
		    sin(x + y - z) - sin(x + y + z)));
    res[1] = -(1.5*(6*x*y*z - 3*cos(x + y + z)) +
	       2.5*(3*x*x*z + 6*x*y*z +
		    3*x*z*z - cos(x + y + z) +
		    sin(x + y - z) - sin(x + y + z)));
    res[2] = -(1.5*(6*x*y*z - 3*sin(x + y - z)) +
	       2.5*(3*x*x*y +
		    3*x*y*y + 6*x*y*z -
		    cos(x + y + z) - sin(x + y - z) -
		    sin(x + y + z)));
    break;
  case 3:
    res[0] = -24 * x;
    res[1] = -24 * y;
    res[2] = -24 * z;
    break;
  case 2:
    res[0] = - z * ( 3 * x * x * x + 7.5 * y * y * z + x * y * (24. * y + 15. * z));
    res[1] = - x * ( 3 * y * y * y + 7.5 * x * z * z + y * z * (24. * z + 15. * x));
    res[2] = - y * ( 3 * z * z * z + + 15 * x * y * z + + x * x * (7.5 * y + 24. *z));
    break;
  case 1:
    res[0] = -8;
    res[1] = -8;
    res[2] = -8;
    break;
  }

  return res;
}


template<class Double>
SmallVector<Double, 3>
FuncSolution(const SmallVector<Double, 3> &p)
{
  Double const &x = p[0];
  Double const &y = p[1];
  Double const &z = p[2];

  
  
  SmallVector<Double, 3> res;

  // variant 5;
  res = Double();
  switch (Options::GetOptions().variant)
  {
  case 0:
    break;
  case 5:
    res[0] = x + y + z;
    res[1] = x - y + z;
    res[2] = x + y - z;
    break;
  case 4:
    res[0] = x*x*x*y*z + sin(x + y + z);
    res[1] = x*y*y*y*z + cos(x + y + z);
    res[2] = x*y*z*z*z + sin(x + y - z);
    break;
  case 3:
    res[0] = x * x * x;
    res[1] = y * y * y;
    res[2] = z * z * z;
    break;
  case 2:
    res[0] = x * x * x * y * y * z;
    res[1] = x * y * y * y * z * z;
    res[2] = x * x * y * z * z * z;
    break;
  case 1:
    res[0] = x*x;
    res[1] = y*y;
    res[2] = z*z;
    break;
  }
  return res;
}

template<class Double>
SmallMatrix<Double, 3, 3>
FuncNeuman(const SmallVector<Double, 3> &p) // returns sigma
{
  Double const &x = p[0];
  Double const &y = p[1];
  Double const &z = p[2];


  SmallMatrix<Double, 3, 3> res;

  res = Double();
  switch (Options::GetOptions().variant)
  {
  case 0:
    res[2][2] = - Options::GetOptions().force;
    return res;
    
    if ((x>=(3/8.))&&(x <=(5/8.))&&(y>=(3/8.))&&(y<=(5/8.)))
    {
      res[2][2] = - Options::GetOptions().force;
      return res;
      //      using namespace std;
      //      cout << res[2][2] << endl;
    }
    break;
    break;
  case 5:
    res[0][0] = 2.;
    res[0][1] = 3.;
    res[0][2] = 3.;
    res[1][0] = res[0][1];
    res[1][1] = -4.;
    res[1][2] = 3.;
    res[2][0] = res[0][2];
    res[2][1] = res[1][2];
    res[2][2] = -4.;
    break;
  case 4:
    res[0][0] = x*y*z*(12.*x + 3.*y + 3.*z) - 1.*cos(x + y - z) + 4.*cos(x + y + z) -  1.*sin(x + y + z);
    res[0][1] =  1.5*((x *x * x + y*y*y)*z +  cos(x + y + z) - sin(x + y + z));
    res[0][2] =  1.5*(y*(x * x*x + z*z*z) + 2*cos(x + y)*cos(z));
    res[1][0] = res[0][1];
    res[1][1] =  x*y*z*(3.*x + 12.*y + 3.*z) -  1.*cos(x + y - z) + cos(x + y + z) - 4.*sin(x + y + z);
    res[1][2] =  1.5*(x*(y*y*y + z*z*z) + cos(x + y - z) - sin(x + y + z));
    res[2][0] = res[0][2];
    res[2][1] = res[1][2];
    res[2][2] = x*y*z*(3.*x + 3.*y + 12.*z) - 4.*cos(x + y - z) + cos(x + y + z) - 1.*sin(x + y + z);
    break;
  case 1:
    res[0][0] = 8 * x + 2 * y + 2 * z;
    res[1][1] = 2 * x + 8 * y + 2 * z;
    res[2][2] = 2 * x + 2 * y + 8 * z;
    break;
  }

  return res;
}

#endif
