#ifndef __elastsolver_h__
#define __elastsolver_h__

#include "vectors.h"
#include "small.h"
#include "elastgen.h"
#include "solver.h"

#include "mirorer.h"
#include <map>



template<class Double>
class MacroElementDescriptor
{
 public:
  MacroElementDescriptor(){}
  MacroElementDescriptor(const MacroElementDescriptor&a)
  {
    E = a.E;
    ni = a.ni;
    lambda = a.lambda;
    mu = a.mu;
    mat = a.mat;
    h = a.h;
  }
  typedef typename ElastDiscretizator<Double>::SparceMacroMatrix_type SparceMacroMatrix_type;
  SmallVector<Double,3> h;
  Double E, ni, lambda, mu;
  void SetCoefficients(Double E_, Double ni_)
  {
    E = E_;
    ni = ni_;
    lambda = E * ni /((1+ni)*(1-2* ni));
    mu = E / (2 *(1+ni));
    //    std::cout << E <<" " << ni << " " << lambda << " " << mu << std::endl;
  }
  static SparceMacroMatrix_type m1, m2;

  SparceMacroMatrix_type mat;
  
};
  

template<class Double, class Int>
class ElementwiseElastityMatrix
{
public:
  typedef ElastDiscretizator<Double> Discretizator;
  typedef typename Discretizator::small_vector_type small_vector_type;
  typedef typename Discretizator::SparceMacroMatrix_type SparceMacroMatrix_type;
  
  SmallVector<Double,3> h;
  Int N1, N2, N3;
  Int DoubleLines;
  Int StartLine;
  bool FirstFull;
  Discretizator *pdis;
  //  SparceMacroMatrix_type *lookupmap[256];
  mutable std::map<int, MacroElementDescriptor<Double> > lookupmap;
  vector3d<unsigned short> coeffs;

  void Sysgen(Discretizator * pdis, BoundaryConditions bc, BoundaryConditions nc);
  
  SparceMacroMatrix_type& GetElMatrixAt(Int i, Int j, Int k);
  
  void GenRhsAt(small_vector_type&rhs, small_vector_type const&, 
		Int i, Int j, Int k, Int realk,
		BoundaryConditions bc, BoundaryConditions nc);
  void GenRhsAtd(small_vector_type&rhs, small_vector_type const&, 
		Int i, Int j, Int k, Int realk,
		BoundaryConditions bc, BoundaryConditions nc);
  
};


template<class Vector>
class ElastMatrix:public MatrixBase<Vector>
{
  
public:
  typedef typename Myf<typename Vector::base_type>::base_type Double;
  typedef typename ElastDiscretizator<Double>::SparceMacroMatrix_type SparceMacroMatrix_type;
  typedef typename Vector::Int Int;
  typedef Vector bvector_type;
private:
public:
  Int N1, N2, N3;
  Int DoubleLines;
  bool FirstFull;
  bool FirstFirst;
  bool LastFull;
  bool LastLast;

  Int Lines;
  Int StartLine;
  ElementwiseElastityMatrix<Double, Int> elmentw;

  ElastMatrix();
  ~ElastMatrix();
  virtual void Multiply(Vector &res, Vector  &src)const;
  virtual std::string GetName()const
  {
    return "Elast";
  }

};



// template<class Double, class Int>
// class ElastWorkerBase
// {
// public:
//   ElastDiscretizator<Double> *pdis;
//   MPI::Intracomm comm;
//   int rank,size;
//   Int N1, N2, N3R;
//   ElastWorkerBase(MPI::Intracomm &c,
// 		  Int n1, Int n2, Int n3);
//   ~ElastWorkerBase(){delete pdis;}
// };

template<class Double, class Int, class bVector_type = BDom<SmallVector<Double, 3>,Int > >
class ElastWorker
//:public ElastWorkerBase<Double, Int>
{
 public:

  Int StartLine, Lines;
  Int DoubleLines;
  Int StartDoubleLine;
  bool FirstFull;
  bool FirstFirst;
  bool LastFull;
  bool LastLast;
  Double merr;
  Double serr;

  typedef SmallVector<Double, 3> point_type;
  typedef bVector_type bvector_type;

private:
  bvector_type rhs;
  bvector_type dissol;
  bvector_type x;
public:
  ElastDiscretizator<Double> *pdis;
  MPI::Intracomm comm;
  Int N1, N2, N3R;
  int rank,size;
  std::string algname;
  DiscretizationType distype;
  BoundaryConditions bc;
  ElastMatrix<bvector_type> mat;
  ElastWorker(MPI::Intracomm &c,
		  Int n1, Int n2, Int n3);
 //  ~ElastWorker(){delete pdis;}
  void Sysgen(DiscretizationType distype, BoundaryConditions bc, BoundaryConditions nc);
  int Solve(Double eps, SolutionMethod method);
  void CheckSolution();

  void WriteSolution();

};



#endif
