#ifndef __elastsolver_hpp__
#define __elastsolver_hpp__
#include "elastsolver.h"
#include "precon.h"
#ifdef MY_HYPRE
#include "hypre_precon.h"
#endif
#include "useful.h"

#include "functors.h"

#include <string>
#include <iostream>



template<class Double, class Int>
void
ElementwiseElastityMatrix<Double, Int>::
Sysgen(Discretizator *pdis, BoundaryConditions bc, BoundaryConditions nc)
{
  Discretizator &dis = *pdis;
  this->pdis = pdis;
  SparceMacroMatrix_type &gsm1 = MacroElementDescriptor<Double>::m1,
    &gsm2 = MacroElementDescriptor<Double>::m2, tmp;

  h[0] = Options::GetOptions().DM1/N1;
  h[1] = Options::GetOptions().DM2/N2;
  h[2] = Options::GetOptions().DM3/N3;
  dis.GenElastElMat(gsm1, gsm2, h);

  Options &opt = Options::GetOptions();
  for(unsigned i = 0; i < opt.NumMaterials; ++i)
    if((opt.E[i] + opt.ni[i])!=0)
    {
      //      std::cout << "Material " << i << " " << opt.E[i] << " " << opt.ni[i] << std::endl;
      MacroElementDescriptor<Double> &first = lookupmap[i];
      first.h = h;
      first.SetCoefficients(opt.E[i], opt.ni[i]);
      tmp = gsm1;
      tmp *= first.mu * 2.;
      first.mat = tmp;
      tmp = gsm2;
      tmp *= first.lambda;
      first.mat += tmp;
      //      std::cout << first.mu << " " << first.lambda<< std::endl;

      int bcarr[] = {
	bcRight,
	bcLeft,
	bcUp,
	bcDown,
	bcFront,
	bcBack,
	bcRight | bcUp,
	bcRight | bcDown,
	bcRight | bcFront,
	bcRight | bcBack,
	bcRight | bcUp | bcFront,
	bcRight | bcUp | bcBack,
	bcRight | bcDown |bcFront,
	bcRight | bcDown |bcBack,
	bcLeft | bcUp,
	bcLeft | bcDown,
	bcLeft | bcFront,
	bcLeft | bcBack,
	bcLeft | bcUp | bcFront,
	bcLeft | bcUp | bcBack,
	bcLeft | bcDown |bcFront,
	bcLeft | bcDown |bcBack,
	bcDown |bcFront,
	bcDown |bcBack,
	bcUp |bcFront,
	bcUp |bcBack,
	//	bcDown |bcFront |bcRight | bcUp | bcBack,
	bcLeft | bcDown |bcFront |bcRight | bcUp | bcBack
      };
	
      for(unsigned j = 0; j < sizeof(bcarr)/sizeof(BoundaryConditions); j++)
      {
	int curbc = bcarr[j];
 	if (!(((curbc|bc)==bc)&&((curbc&bc)==curbc)))
 	  continue;
	lookupmap[(curbc <<8) + i].SetCoefficients(first.E, first.ni);
	dis.BoundCondSet(lookupmap[(curbc <<8) + i].mat,
			 first.mat, BoundaryConditions(curbc));
      }
    }

  coeffs.SetSize(N1/2, N2/2, DoubleLines);
  for(unsigned int i = 0; i < coeffs.sizes[0]; i++)
    for(unsigned  int j = 0; j < coeffs.sizes[1]; j++)
      for(unsigned  int k = 0; k < coeffs.sizes[2];k++)
	coeffs(i,j,k) = 255;
  switch (Options::GetOptions().variant)
  {
  case 0:
    for(unsigned int i = 3*coeffs.sizes[0]/8; i< 5*coeffs.sizes[0]/8; ++i)
      for(unsigned int j = 3* coeffs.sizes[1]/8; j< 5*coeffs.sizes[1]/8;++j)
	for(unsigned int k = 0; k < coeffs.sizes[2];k++)
	{
	  Int realk = k*2 + StartLine;
	  if (realk >= N3/2)
	    coeffs(i,j,k) = 0;
	}
  }
  if (Options::GetOptions().usebone)
  {
    std::ostringstream os;
    os << Options::GetOptions().scratchpath << "/"<< N3/2 << ".bonec";

    try{
      coeffs.Read(os.str(), StartLine/2);
    }catch(std::exception&)
    {
      std::cerr << "could not open file" << os.str() << std::endl;
      throw;
    }
    for(unsigned int i = 0; i < coeffs.sizes[0]; i++)
      for(unsigned int j = 0; j < coeffs.sizes[1]; j++)
      {
	int k = 0;
	Int realk = 2*k + StartLine;
	if(realk==0)
	  coeffs(i,j,k) = 1;
	k = coeffs.sizes[2];
	realk = 2*k + StartLine;
	if((realk+2)==N3)
	  coeffs(i,j,k) = 1;
      }
  }
  //boundary conditions

  for(unsigned int i = 0; i < coeffs.sizes[0]; ++i)
    for(unsigned int j = 0; j < coeffs.sizes[1];++j)
    {
      
      if ((bc & bcUp)&&(coeffs.sizes[2])&&(StartLine/2+coeffs.sizes[2]==static_cast<size_t>(N3)/2))
	coeffs(i,j,coeffs.sizes[2]-1)|= unsigned(bcUp)<<8;
      if ((bc & bcDown)&&(coeffs.sizes[2])&&((StartLine/2)==0))
	coeffs(i,j,0)|= unsigned(bcDown)<<8;
    }
  for(unsigned int i = 0; i < coeffs.sizes[0]; ++i)
    for(unsigned int k = 0; k < coeffs.sizes[2];++k)
    {
      if (bc & bcBack)
	coeffs(i,coeffs.sizes[1]-1,k)|= unsigned(bcBack)<<8;
      if (bc & bcFront)
	coeffs(i,0,k)|= unsigned(bcFront)<<8;
    }
  for(unsigned int j = 0; j < coeffs.sizes[1];++j)
    for(unsigned int k = 0; k < coeffs.sizes[2];++k)
    {
      if (bc & bcRight)
	coeffs(coeffs.sizes[0]-1,j,k)|= unsigned(bcRight)<<8;
      if (bc & bcLeft)
	coeffs(0,j,k)|= unsigned(bcLeft)<<8;
    }
}


template<class Double, class Int>
void
ElementwiseElastityMatrix<Double, Int>::
GenRhsAt(small_vector_type&small_rhs, small_vector_type const &discretizedU, Int i, Int j, Int k, Int realk, BoundaryConditions bc, BoundaryConditions nc)
{
  Discretizator &dis = *pdis;
  SmallMatrix<Double, 8, 3> elpoints;
  elpoints[0][0] = h[0]*(Double)(i);
  elpoints[0][1] = h[1]*(Double)j;
  elpoints[0][2] = h[2]*(Double)(realk);

  elpoints[1][0] = h[0]*(Double)(i);
  elpoints[1][1] = h[1]*(Double)j;
  elpoints[1][2] = h[2]*(Double)(realk+1);
	
  elpoints[2][0] = h[0]*(Double)(i);
  elpoints[2][1] = h[1]*(Double)(j+1);
  elpoints[2][2] = h[2]*(Double)(realk);
	
  elpoints[3][0] = h[0]*(Double)(i);
  elpoints[3][1] = h[1]*(Double)(j+1);
  elpoints[3][2] = h[2]*(Double)(realk+1);

  elpoints[4][0] = h[0]*(Double)(i+1);
  elpoints[4][1] = h[1]*(Double)j;
  elpoints[4][2] = h[2]*(Double)(realk);
	
  elpoints[5][0] = h[0]*(Double)(i+1);
  elpoints[5][1] = h[1]*(Double)j;
  elpoints[5][2] = h[2]*(Double)(realk+1);
	
  elpoints[6][0] = h[0]*(Double)(i+1);
  elpoints[6][1] = h[1]*(Double)(j+1);
  elpoints[6][2] = h[2]*(Double)(realk);
	
  elpoints[7][0] = h[0]*(Double)(i+1);
  elpoints[7][1] = h[1]*(Double)(j+1);
  elpoints[7][2] = h[2]*(Double)(realk+1);
  

	
  small_rhs=0.;
  dis.GenRhs(small_rhs, elpoints, h);
	
  // neuman conditions
      
  if ((nc & bcUp)&&
      (coeffs.sizes[2])&&
      (StartLine/2+coeffs.sizes[2]==static_cast<size_t>(N3)/2)&&
      ((static_cast<size_t>(k)+2)==2*coeffs.sizes[2]))
    dis.NoimanBoundCondSet(small_rhs, elpoints[0], h, BUp);
  if ((nc & bcDown)&&(coeffs.sizes[2])&&((StartLine/2)==0)&&(k==0))
    dis.NoimanBoundCondSet(small_rhs, elpoints[0], h, BDown);
	
	
  if (((j+2) == N2)&&(nc & bcBack))
    dis.NoimanBoundCondSet(small_rhs, elpoints[0], h, BBack);
	
  if ((j==0)&&(nc & bcFront))
    dis.NoimanBoundCondSet(small_rhs, elpoints[0], h, BFront);
	
	
  if (((i+2) == N1)&&(nc & bcRight))
    dis.NoimanBoundCondSet(small_rhs, elpoints[0], h, BRight);
	
  if ((i==0)&&(nc & bcLeft))
    dis.NoimanBoundCondSet(small_rhs, elpoints[0], h, BLeft);

  //???????????? maybe k+1
  //  std::cerr << "k " << k << std::endl;
  SparceMacroMatrix_type  &A = lookupmap[coeffs(i/2,j/2,(k+1)/2)&0xff].mat; //we need the matrix without boundary conditions
  
  dis.BoundCondSetRhs(A, small_rhs, discretizedU, BoundaryConditions((coeffs(i/2,j/2,(k+1)/2) >> 8)&0x3f));
}
template<class Double, class Int>
void
ElementwiseElastityMatrix<Double, Int>::
GenRhsAtd(small_vector_type&small_rhs, small_vector_type const &discretizedU, Int i, Int j, Int k, Int realk, BoundaryConditions bc, BoundaryConditions nc)
{
  //int realk = k + StartLine - 1+Int(FirstFull);
  //  Int k = realk - StartLine +1 -Int(FirstFull);
  Discretizator &dis = *pdis;

  //???????????? maybe k+1
  SparceMacroMatrix_type  &A = lookupmap[coeffs(i/2,j/2,(k+1)/2)&0xff].mat; //we need the matrix without boundary conditions
  
  dis.BoundCondSetRhs(A, small_rhs, discretizedU, BoundaryConditions((coeffs(i/2,j/2,(k+1)/2) >> 8)&0x3f));
}
       

















template<class Double, class Int, class BBvec>
ElastWorker<Double, Int, BBvec>::ElastWorker(MPI::Intracomm  & c,
			    Int n1, Int n2, Int n3)
  :  comm(c.Dup()), N1(n1), N2(n2), N3R(n3),
     rank(c.Get_rank()), size(c.Get_size())
{
  pdis = NULL;

  if((N1&1)||(N2&1)||(N3R&1))
    throw std::string("Sizes must be even");
  
  size = comm.Get_size();
  rank = comm.Get_rank();
  StartLine = rank*(N3R+1)/size;
  Lines = (rank+1)*(N3R+1)/size - StartLine;

  DoubleLines = Lines/2;
  
  FirstFirst = rank==0;
  LastLast = rank == (size-1);
  FirstFull = ((StartLine &1) == 0);
  
  DoubleLines = (Lines&1)? (Lines+1)/2-Int(FirstFull) : (Lines)/2;
  //  cout << "rank " << rank << " " << Lines << " " << (Int)FirstFull << " "<< DoubleLines << endl;
  //  exit(0);

  rhs.SetSize(N1, N2, Lines, N3R);
  rhs.SetComm(comm);
  rhs.CreateDatatypes();
  x.SetSize(N1, N2, Lines, N3R);
  x.SetComm(comm);
  //    cout << "Lines " << rank << " " << Lines << endl;
}

template<class Double, class Int, class BBvec>
void
ElastWorker<Double, Int, BBvec>::
Sysgen(DiscretizationType distype, BoundaryConditions bc, BoundaryConditions nc)
{
  this->distype = distype;
  this->bc = bc;
  typedef ElastDiscretizator<Double> Discretizator;
  pdis = new Discretizator(distype);
  //  Discretizator &dis = *pdis;
  typedef typename Discretizator::SparceMacroMatrix_type SparceMacroMatrix_type;
  typedef typename Discretizator::small_vector_type small_vector_type;

  SmallVector<Double, 3> h;
  h[0] = Options::GetOptions().DM1/N1;
  h[1] = Options::GetOptions().DM2/N2;
  h[2] = Options::GetOptions().DM3/N3R;

  ElementwiseElastityMatrix<Double, Int> &elmentw = mat.elmentw;

  elmentw.DoubleLines = DoubleLines;
  elmentw.N1 = N1;
  elmentw.N2 = N2;
  elmentw.N3 = N3R;
  elmentw.StartLine = StartLine;
  elmentw.FirstFull = FirstFull;
  elmentw.Sysgen(pdis, bc, nc);

  
  mat.DoubleLines = DoubleLines;
  mat.FirstFull = FirstFull;
  mat.FirstFirst = FirstFirst;
  mat.LastLast = LastLast;
  mat.LastFull = LastFull;
  
  mat.Lines = Lines;
  mat.StartLine = StartLine;
    
  mat.N1 = N1;
  mat.N2 = N2;
  mat.N3 = Lines;  // ???

  //generating rhs
  ClearVector(rhs, 0);
  dissol.SetSize(N1, N2, Lines, N3R);
  dissol.SetComm(comm);
  dissol.CreateDatatypes();
  ClearVector(dissol,0);
  
  SmallVector<Double, 3> p, pl, pd, pf;
  Int NumberOfLines = Lines+1;
  Int SL = StartLine-1;
  Int SO = -1;
  if (rank == 0)
  {
    SL++;
    SO++;
    NumberOfLines--;
  }
  Index3dBase<Int> ind(N1, N2, Lines);
  for(int i = 0; i <= N1; i++)
  {
    //    cout << "i" << i << endl;
    for(int j = 0; j <= N2; j++)
      for(int k = 0; k < NumberOfLines; k++)
      {
	ind.SetPos(i, j, k + SO);
	
	p[0] = (Double)i *h[0];
	p[1] = (Double)j *h[1];
	p[2] = (Double)(k + SL)*h[2];
	pl = p;
	pd = p;
	pf = p;
	pl[1] += h[1]*0.5;
	pl[2] += h[2]*0.5;
	
	pf[0] += h[0]*0.5;
	pf[2] += h[2]*0.5;
	
	pd[0] += h[0]*0.5;
	pd[1] += h[1]*0.5;

	//	std::cout << i << " " << j << " " << k+SL << " " << pl[0] <<  " " << pl[1] << " " << pl[2] << endl;
	
 	dissol[ind][LeftEl] = FuncSolution(pl);
 	dissol[ind][DownEl] = FuncSolution(pd);
 	dissol[ind][FrontEl] = FuncSolution(pf);
      }
  }
  dissol.StartSendRecv();
  dissol.FinishSendRecv();


  


  for(Int i = 0; i < N1; i+=2)
    for(Int j = 0; j < N2; j+=2)
    {
      for(Int k = 0; k < DoubleLines*2; k+=2)
      {

	int realk = k + StartLine - 1+Int(FirstFull);
	Index3dBase<Int> ind(N1, N2, Lines);
	ind.SetPos(i, j, k - 1 + Int(FirstFull));
	
	small_vector_type small_rhs;
	small_vector_type discretizedU;

	BigToSmallCpy<bvector_type, small_vector_type> b2scopy1(dissol, discretizedU);
	DoForMacroElement(b2scopy1, ind);
	
	elmentw.GenRhsAt(small_rhs, discretizedU, i, j, k, realk, bc, nc);
	
	SmallToBigAdd<bvector_type, small_vector_type> s2badd(rhs, small_rhs);
	DoForMacroElement(s2badd, ind);
      }
    }
  rhs.SendDown = (StartLine & 1) != 0;
  rhs.SendUp   = (((StartLine+Lines-1)&1) != 0);
   rhs.StartSendRecvAdd();
   rhs.FinishSendRecvAdd();
//   rhs.StartSendRecv();
//   rhs.FinishSendRecv();
//   for(Int i = 0; i < N1; i+=2)
//     for(Int j = 0; j < N2; j+=2)
//     {
//       for(Int k = 0; k < DoubleLines*2; k+=2)
//       {
// 	int realk = k + StartLine - 1+Int(FirstFull);
// 	Index3dBase<Int> ind(N1, N2, Lines);
// 	ind.SetPos(i, j, k - 1 + Int(FirstFull));

// 	small_vector_type small_rhs;
// 	small_vector_type discretizedU;

// 	BigToSmallCpy<bvector_type, small_vector_type> b2scopy1(dissol, discretizedU);
// 	BigToSmallCpy<bvector_type, small_vector_type> b2scopy2(rhs, small_rhs);
	
// 	DoForMacroElement(b2scopy1, ind);
// 	DoForMacroElement(b2scopy2, ind);
// 	elmentw.GenRhsAtd(small_rhs, discretizedU, i, j, k, realk, bc, nc);
	
// 	SmallToBigCpy<bvector_type, small_vector_type> s2badd(rhs, small_rhs);
// 	//	DoForMacroElement(s2badd, ind);
//       }
//     }
  //  rhs.StartSendRecvAdd();
  //  rhs.FinishSendRecvAdd(false);
}

template<class Double, class Int, class BBvec>
void
ElastWorker<Double, Int, BBvec>::
WriteSolution()
{
  static bool firsttime = true;
  static int itop[6][8] = {{1,2,3,4,17,18,19,20},
		{17,18,19,20,33,34,35,36},
		{5,6,10,11,21,22,26,27},
		{10,11,15,16,26,27,31,32},
		{7,8,12,13,23,24,28,29},
		{8,9,13,14,24,25,29,30}};
  //  static int cubes[BSize][4];

  if(firsttime)
  {
    firsttime = false;

    for(int i = 0; i < 6; ++i)
      for(int j = 0; j < 8; ++j)
	--itop[i][j];

  }

  std::string fname = Options::GetOptions().scratchpath +"/dss." + Options::GetOptions().namebase; //displacements+strasses+strains
  ofstream fout(fname.c_str());
  fout.precision(15);
  SmallVector<Double, 3> displacements;
  SmallMatrix<Double, 3, 3> stresses;
  SmallMatrix<Double, 3, 3> strains;
  SmallVector<Double, 18> small_x;
  typedef typename ElastDiscretizator<Double>::SparceMacroMatrix_type smt;
  typedef typename smt::vector_type small_vector_type;
  small_vector_type medium_x;
  for(Int k = 0; k < DoubleLines*2; k++)
    for(Int i = 0; i < N1; i++)
      for(Int j = 0; j < N2; j++)
      {
	{
	  Index3dBase<Int> ind(N1, N2, Lines);
	  ind.SetPos(i &~1, j&~1, (k&~1) - 1 +Int(FirstFull));
	  Double lambda = mat.elmentw.lookupmap[mat.elmentw.coeffs(i/2,j/2,k/2)&0x3fff].lambda;
	  Double mu = mat.elmentw.lookupmap[mat.elmentw.coeffs(i/2,j/2,k/2)&0x3fff].mu;
	  
	  BigToSmallCpy<bvector_type, small_vector_type> b2scopy1(x, medium_x);
	  DoForMacroElement(b2scopy1, ind);
	  int cube = (k&1) + (j&1)*2 + (i&1)*4;
	  for(int i = 0; i < 6; ++i)
	  {
	    small_x[0  + i] = medium_x[0 + itop[i][cube]];
	    small_x[6  + i] = medium_x[36 + itop[i][cube]];
	    small_x[12 + i] = medium_x[72 + itop[i][cube]];
	  }
	  pdis->CalculateStressesAndStrains(displacements,
					    stresses,
					    strains, 
					    small_x,
					    lambda,
					    mu);

	  for(int i = 0; i < 3; ++i)
	    fout << displacements[i] << " ";
	  for(int i = 0; i < 3; ++i)
	    for(int j = 0; j < 3; ++j)
	      fout << strains[i][j] << " ";
	  for(int i = 0; i < 3; ++i)
	    for(int j = 0; j < 3; ++j)
	      fout << stresses[i][j] << " ";
	  
	  fout << "\n";

	}
      }
  
}

template<class Double, class Int,class BBvec>
void
ElastWorker<Double, Int, BBvec>::
CheckSolution()
{
  x.CreateDatatypes();
  x.StartSendRecv();
  x.FinishSendRecv();
  VecVecOper<VecSub>(dissol, x);      // G -= RHS;

  serr =  VecReduceNorm<CNorm>(dissol);

  dissol.CreateDatatypes();
  mat.Multiply(dissol, x);
  VecVecOper<VecSub>(dissol, rhs);
  merr =  VecReduceNorm<CNorm>(dissol);
};


template<class Double, class Int, class BBvec>
int
ElastWorker<Double, Int, BBvec>::
Solve(Double eps, SolutionMethod method)
{
  typedef SmallVector<Double, 6> row_type;
  typedef BDom<row_type, Int> matrix_type;
  typedef BDom<Double, Int> plainvector_type;
  typedef SmallVector<float, 6> srow_type;
  typedef BDom<srow_type, Int> smatrix_type;
  typedef BDom<float, Int> splainvector_type;
  typedef BDom<SmallVector<float, 3>, Int>   sbvector_type;
  matrix_type *pmat = NULL;
  smatrix_type *spmat = NULL;

  const int fcs=30;
  const int scs=30;

  PCGSolverBase<bvector_type>
    solver;
  solver.SetEps(eps);
  solver.SetSizes(N1, N2, Lines, N3R);
  solver.SetComm(comm);
  solver.SetMatrix(&mat);
  ClearVector(x);
  
  PreconBase<bvector_type> *precon = NULL;
  PreconBase<bvector_type> *sprecon = NULL;
  typedef SinglePrecon<bvector_type, PreconBase<sbvector_type> > sprecon_type;

  {
    Int NumberOfLines = Lines+1;
    Int SL = StartLine-1;
    Int SO = -1;
    if (rank == 0)
    {
      SL++;
      SO++;
      NumberOfLines--;
    }
    if (rank +1 == size)
    {
      NumberOfLines--;
    }


    coefficientvec.Clear();
    coefficientvec.SetSize(N1, N2, NumberOfLines);
    coefficientvec.SetPars(
 			   Options::GetOptions().zero
 			   ,
			   0, Options::GetOptions().DM1,
			   0, Options::GetOptions().DM2,
			   Options::GetOptions().DM3*(double)SL/N3R,
			   Options::GetOptions().DM3*(double)(NumberOfLines+SL)/N3R);

    int off = FirstFull?1:0;
    if (rank==0)off=0;
    //    std::cout <<rank << " " <<NumberOfLines<<" "<<SL <<" " << DoubleLines<<"-------------------- "<< off <<std::endl;
    for(Int i = 0; i < N1; ++i)
      for(Int j = 0; j < N2; ++j)
	for(Int k  = 0; k<DoubleLines; ++k)
	{
	  coefficientvec(i,j,off+ k*2) = 
	  
	  coefficientvec(i,j,off+ k*2+1) = mat.elmentw.lookupmap[mat.elmentw.coeffs(i/2,j/2,k)&0x3fff].E;
	}

    int UpP = rank+1;
    int DownP = rank-1;
    if (DownP <0) DownP = MPI::PROC_NULL;
    if (UpP == size) UpP = MPI::PROC_NULL;

    MPI::Request rsd = MPI::REQUEST_NULL;
    MPI::Request rsu = rsd;
    int cnt = N1*N2;
    bool SendDown = (StartLine & 1) != 0;
    bool SendUp   = (((StartLine+Lines-1)&1) != 0);
    if ((rank+1)!=size)
      if(SendUp)
      {
	rsu = comm.Isend(&(coefficientvec(0,0,NumberOfLines-1)), cnt,
			 MPI::DOUBLE,
			 UpP, 0);
      }
      else
      {
	rsu = comm.Irecv(&(coefficientvec(0,0,NumberOfLines-1)), cnt,
			 MPI::DOUBLE,
			 UpP, 1);
      }
    

    if(rank)
      if(SendDown)
      {
	rsd = comm.Isend(&(coefficientvec(0,0,0)), cnt,
			 MPI::DOUBLE,
			 DownP, 1);
      }
      else
      {
	rsd = comm.Irecv(&(coefficientvec(0,0,0)), cnt,
			 MPI::DOUBLE,
			 DownP, 0);
      }
    
    rsd.Wait();
    rsu.Wait();
    
    if (Options::GetOptions().SinglePrecPrecon)
    {
      splainvector_type rhsp;
      rhsp.SetSize(N1, N2, Lines, N3R);
      rhsp.SetComm(comm);
      spmat = new smatrix_type;
      spmat->SetSize(N1, N2, Lines, N3R);
      spmat->SetComm(comm);
    
      ClearMatrix(*spmat);
      ClearVector(rhsp);
      DiscretizationType dt;
      if (distype==MP)dt = RP;
      if (distype==MV)dt = RV;
      Discretizator<float> dis(dt);
      TimeCalc1("PreconConstruct");

      if((method==smPCG1)||(method==smPCG2)||(method==smPCG3)||(method==smPCG4))
	dis.GenerateMatrix(*spmat, rhsp, NumberOfLines, SL, SO, N3R, bc);
    }
    else
    {
      plainvector_type rhsp;
      rhsp.SetSize(N1, N2, Lines, N3R);
      rhsp.SetComm(comm);
      pmat = new matrix_type;
      pmat->SetSize(N1, N2, Lines, N3R);
      pmat->SetComm(comm);
    
      ClearMatrix(*pmat);
      ClearVector(rhsp);
      DiscretizationType dt;
      if (distype==MP)dt = RP;
      if (distype==MV)dt = RV;
      Discretizator<Double> dis(dt);
      if((method==smPCG1)||(method==smPCG2)||(method==smPCG3)||(method==smPCG4))
      {
	TimeCalc1("PreconConstruct");
	dis.GenerateMatrix(*pmat, rhsp, NumberOfLines, SL, SO, N3R, bc);
      }
    }
    
  }
    
    if (Options::GetOptions().SinglePrecPrecon)
    {
      typedef SmallVector<float, 6> row_type;
      typedef BDom<srow_type, Int> matrix_type;
      typedef BDom<float, Int> plainvector_type;
      typedef sbvector_type bvector_type;
      PreconBase<bvector_type> *precon = NULL;
      switch(method)
      {
      case smCG:
	precon = new NoPrecon<bvector_type>();
	break;
      
      case smPCG1:
	{
	  MIC0Precon<matrix_type, bvector_type, plainvector_type,  PreconDescriptorM1> *p
	    = new MIC0Precon<matrix_type, bvector_type, plainvector_type, PreconDescriptorM1>;
	  p->Create(*spmat);
	  precon = p;
	}    
	break;
      case smPCG2:
	{
	  MIC0Precon<matrix_type, bvector_type, plainvector_type, PreconDescriptorM2> *p
	    = new MIC0Precon<matrix_type, bvector_type, plainvector_type, PreconDescriptorM2>;
	  p->Create(*spmat);
	  precon = p;
	}    
      
	break;
      case smPCG3:
	{
	  MIC0Precon<matrix_type, bvector_type, plainvector_type,  PreconDescriptorM3> *p
	    = new MIC0Precon<matrix_type, bvector_type, plainvector_type,  PreconDescriptorM3>;
	  p->Create(*spmat);
	  precon = p;
	}    
      
	break;
      case smPCG4:
	{
	  MIC0PreconM4<matrix_type, bvector_type, plainvector_type, PreconDescriptorM4> *p
	    = new MIC0PreconM4<matrix_type, bvector_type, plainvector_type,  PreconDescriptorM4>;
	  p->Create(*spmat);
	  //      std::cout << "m4 " << std::endl;
	  precon = p;
	}
      default:;
      }
      sprecon = new sprecon_type(precon);
    }
    else
    {
#ifdef MY_HYPRE
      /// for hypreprecon discretizator
      Int NumberOfLines = Lines+1;
      Int SL = StartLine-1;
      Int SO = -1;
      if (rank == 0)
      {
	SL++;
	SO++;
	NumberOfLines--;
      }
      if (rank +1 == size)
      {
	NumberOfLines--;
      }
      DiscretizationType dt;
      if (distype==MP)dt = RP;
      if (distype==MV)dt = RV;

      Discretizator<double> dis(dt);
      HypreDDPrecon<bvector_type>* hp = NULL;      
      ///end for it
#endif
      switch(method)
      {
#ifdef MY_HYPRE
      case smBoomerDD:
	{
	  TimeCalc1("PreconConstruct");

	hp = new HypreDDPrecon<bvector_type>;
	hp->Init(comm, N1, N2, N3R);
	dis.GenerateMatrix(hp->mat, hp->hdst, NumberOfLines, SL, SO, N3R, bc);
	hp->Init2();
	precon = hp;
	}
	break;
#endif
      case smCG:
	precon = new NoPrecon<bvector_type>();
	break;
      
      case smPCG1:
	{
	  MIC0Precon<matrix_type, bvector_type, plainvector_type,  PreconDescriptorM1> *p
	    = new MIC0Precon<matrix_type, bvector_type, plainvector_type, PreconDescriptorM1>;
	  p->Create(*pmat);
	  precon = p;
	}    
	break;
      case smPCG2:
	{
	  MIC0Precon<matrix_type, bvector_type, plainvector_type, PreconDescriptorM2> *p
	    = new MIC0Precon<matrix_type, bvector_type, plainvector_type, PreconDescriptorM2>;
	  p->Create(*pmat);
	  precon = p;
	}    
      
	break;
      case smPCG3:
	{
	  MIC0Precon<matrix_type, bvector_type, plainvector_type,  PreconDescriptorM3> *p
	    = new MIC0Precon<matrix_type, bvector_type, plainvector_type,  PreconDescriptorM3>;
	  p->Create(*pmat);
	  precon = p;
	}    
      
	break;
      case smPCG4:
	{
	  MIC0PreconM4<matrix_type, bvector_type, plainvector_type, PreconDescriptorM4> *p
	    = new MIC0PreconM4<matrix_type, bvector_type, plainvector_type,  PreconDescriptorM4>;
	  p->Create(*pmat);
	  //      std::cout << "m4 " << std::endl;
	  precon = p;
	}    
      default:;
      }
    }



    if (sprecon)
      precon = sprecon;
  solver.SetPrecon(precon);

  coefficientvec.Clear();


  if (rank == 0)
  {
    cout << "Starting  on " 
	 << size << " processors. Eps is "<< eps<< endl;
    
    cout << setw(fcs) << left << "Problem size "
	 << N1 << "x" << N2 << "x" << N3R << " " << Myf<Double>::ToString() << " "
	 << algname + solver.GetName() <<endl;
    cout << setw(fcs) << left  << "Running on "  
	 << setw(scs) <<size  << "procesors." << endl;
  }
  algname += solver.GetName();
  int iterations;
  {
    TimeCalc1("!Solution time");
    iterations  = solver.Solve(x, rhs);
  }
  if (rank == 0)
  {
    cout << setw(fcs) << "Iterations" 
	 << setw(scs) << iterations << endl;
  }
  delete precon;
  delete pmat;
  return iterations;
}



template<class Vector>
ElastMatrix<Vector>::
ElastMatrix()
{
}


template<class Vector>
ElastMatrix<Vector>::
~ElastMatrix()
{
}

template<class Vector>
void
ElastMatrix<Vector>::
Multiply(Vector &res, Vector  &src)const
{
  TimeCalc1("MatVec");
  typedef typename SparceMacroMatrix_type:: vector_type vector_type;
  ClearVector(res,0);
  vector_type small_res, small_src;
  res.SendDown = (StartLine & 1) != 0;
  res.SendUp   = (((StartLine+N3-1)&1) != 0);
  typedef typename Vector::Int Int;
  if (DoubleLines>=4)
  {
    Int fk, lk, midk;
    fk = 0;
    lk = (DoubleLines-1)*2;
    midk = DoubleLines;
    midk &= ~Int(1);
    src.StartSendRecv();
    // first middle
    for(Int i = 0; i < N1; i+=2)
      for(Int j = 0; j < N2; j+=2)
      {
	for(Int k = fk+2; k < midk; k+=2)
	{
	  Index3dBase<Int> ind(N1, N2, Lines);
	  ind.SetPos(i, j, k - 1 +Int(FirstFull));
	  //	std::cout << elmentw.coeffs(i/2, j/2, k/2) << endl;
	  SparceMacroMatrix_type  &A = elmentw.lookupmap[elmentw.coeffs(i/2,j/2,k/2)&0x3fff].mat;
	  
	  BigToSmallCpy<bvector_type, vector_type> b2scopy1(src, small_src);
	  BigToSmallCpy<bvector_type, vector_type> b2scopy2(res, small_res);
	  SmallToBigAdd<bvector_type, vector_type> s2badd(res, small_res);
	  
	  // src ->small_src;
	  DoForMacroElement(b2scopy1, ind);
	  // res -> small_res
	  DoForMacroElement(b2scopy2, ind);
	  
	  A.MulVec(small_res, small_src);
	  // small_res -> res	
	  DoForMacroElement(s2badd, ind);
	}
      }
    //    res.StartSendRecvAdd();
    //    res.FinishSendRecvAdd();
    
    src.FinishSendRecv();
    // ends start
    for(Int i = 0; i < N1; i+=2)
      for(Int j = 0; j < N2; j+=2)
      {
	Int k = fk;
	{
	  Index3dBase<Int> ind(N1, N2, Lines);
	  ind.SetPos(i, j, k - 1 +Int(FirstFull));
	  //	std::cout << elmentw.coeffs(i/2, j/2, k/2) << endl;
	  SparceMacroMatrix_type  &A = elmentw.lookupmap[elmentw.coeffs(i/2,j/2,k/2)&0x3fff].mat;
	  
	  BigToSmallCpy<bvector_type, vector_type> b2scopy1(src, small_src);
	  BigToSmallCpy<bvector_type, vector_type> b2scopy2(res, small_res);
	  SmallToBigAdd<bvector_type, vector_type> s2badd(res, small_res);
	  
	  // src ->small_src;
	  DoForMacroElement(b2scopy1, ind);
	  // res -> small_res
	  DoForMacroElement(b2scopy2, ind);
	  
	  A.MulVec(small_res, small_src);
	  // small_res -> res	
	  DoForMacroElement(s2badd, ind);
	}
      }
    for(Int i = 0; i < N1; i+=2)
      for(Int j = 0; j < N2; j+=2)
      {
	Int k = lk;
	{
	  Index3dBase<Int> ind(N1, N2, Lines);
	  ind.SetPos(i, j, k - 1 +Int(FirstFull));
	  //	std::cout << elmentw.coeffs(i/2, j/2, k/2) << endl;
	  SparceMacroMatrix_type  &A = elmentw.lookupmap[elmentw.coeffs(i/2,j/2,k/2)&0x3fff].mat;
	  
	  BigToSmallCpy<bvector_type, vector_type> b2scopy1(src, small_src);
	  BigToSmallCpy<bvector_type, vector_type> b2scopy2(res, small_res);
	  SmallToBigAdd<bvector_type, vector_type> s2badd(res, small_res);
	  
	  // src ->small_src;
	  DoForMacroElement(b2scopy1, ind);
	  // res -> small_res
	  DoForMacroElement(b2scopy2, ind);
	  
	  A.MulVec(small_res, small_src);
	  // small_res -> res	
	  DoForMacroElement(s2badd, ind);
	}
      }
    
    
    // ends end;
    res.StartSendRecvAdd();

    // seccond middle;
    for(Int i = 0; i < N1; i+=2)
      for(Int j = 0; j < N2; j+=2)
      {
	for(Int k = midk; k < lk; k+=2)
	{
	  Index3dBase<Int> ind(N1, N2, Lines);
	  ind.SetPos(i, j, k - 1 +Int(FirstFull));
	  //	std::cout << elmentw.coeffs(i/2, j/2, k/2) << endl;
	  SparceMacroMatrix_type  &A = elmentw.lookupmap[elmentw.coeffs(i/2,j/2,k/2)&0x3fff].mat;
	  
	  BigToSmallCpy<bvector_type, vector_type> b2scopy1(src, small_src);
	  BigToSmallCpy<bvector_type, vector_type> b2scopy2(res, small_res);
	  SmallToBigAdd<bvector_type, vector_type> s2badd(res, small_res);
	  
	  // src ->small_src;
	  DoForMacroElement(b2scopy1, ind);
	  // res -> small_res
	  DoForMacroElement(b2scopy2, ind);
	  
	  A.MulVec(small_res, small_src);
	  // small_res -> res	
	  DoForMacroElement(s2badd, ind);
	}
      }
    
    res.FinishSendRecvAdd();
      
      
  }
  else
  {
    src.StartSendRecv();
    src.FinishSendRecv();
    for(Int i = 0; i < N1; i+=2)
      for(Int j = 0; j < N2; j+=2)
      {
	for(Int k = 0; k < DoubleLines*2; k+=2)
	{
	  Index3dBase<Int> ind(N1, N2, Lines);
	  ind.SetPos(i, j, k - 1 +Int(FirstFull));
	  //	std::cout << elmentw.coeffs(i/2, j/2, k/2) << endl;
	  SparceMacroMatrix_type  &A = elmentw.lookupmap[elmentw.coeffs(i/2,j/2,k/2)&0x3fff].mat;
	  
	  BigToSmallCpy<bvector_type, vector_type> b2scopy1(src, small_src);
	  BigToSmallCpy<bvector_type, vector_type> b2scopy2(res, small_res);
	  SmallToBigAdd<bvector_type, vector_type> s2badd(res, small_res);
	  
	  // src ->small_src;
	  DoForMacroElement(b2scopy1, ind);
	  // res -> small_res
	  DoForMacroElement(b2scopy2, ind);
	  
	  A.MulVec(small_res, small_src);
	  // small_res -> res	
	  DoForMacroElement(s2badd, ind);
	}
      }
    res.StartSendRecvAdd();
    res.FinishSendRecvAdd();
  }

}





#endif

