#ifndef __functors_h__
#define __functors_h__
#include "basic.h"
#include <string>
#include "mmpl.hpp"
#include "useful.h" // for getperturb
#include <math.h>

template<class T> 
struct  DoAller
{
  template<class V>
  void DoAll(V const &i)
  {
    static_cast<T&>(*this).template Do<LeftEl>(i);
    static_cast<T&>(*this).template Do<FrontEl>(i);
    static_cast<T&>(*this).template Do<DownEl>(i);
  }
};

template<class T> 
struct  DoAllerReverse
{
  template<class V>
  void DoAll(V const &i)
  {
    static_cast<T&>(*this).template Do<DownEl>(i);
    static_cast<T&>(*this).template Do<FrontEl>(i);
    static_cast<T&>(*this).template Do<LeftEl>(i);
  }
};

template<class Func1, class Func2>
class DualFunctor:public DoAller<DualFunctor<Func1, Func2> >
{
  Func1 &func1;
  Func2 &func2;
public:
  DualFunctor(Func1 &f1, Func2 &f2):func1(f1), func2(f2){}
  template<Elment El, class T>
  void Do(T const &i)
  {
    func1.template Do<El>(i);
    func2.template Do<El>(i);
  }
};
template<class Func1, class Func2>
class DualFunctorWithMethod1:public Func1
{
  //    Func1 &func1;
  Func2 &func2;
public:
  DualFunctorWithMethod1(Func1 &f1, Func2 &f2):Func1(f1), func2(f2){}
  template<Elment El, class T>
  void Do(T const &i)
  {
    Func1::template Do<El>(i);
    func2.template Do<El>(i);
  }
  template<class T>
  void DoAll(T const &i)
  {
    Func1::DoAll(i);
    func2.DoAll(i);
  }
};
template<class Func1, class Func2>
class DualFunctorWithMethod2:public Func2
{
  Func1 &func1;
  //  Func2 &func2;
public:
  DualFunctorWithMethod2(Func1 &f1, Func2 &f2):func1(f1), Func2(f2){}
  template<Elment El, class T>
  void Do(T const &i)
  {
    func1.template Do<El>(i);
    Func2::template Do<El>(i);
  }
  template<class T>
  void DoAll(T const &i)
  {
    func1.DoAll(i);
    Func2::DoAll(i);
  }

};

template<class Func1, class Func2>
inline DualFunctor<Func1, Func2>
MakeDualFunctor(Func1&f1, Func2&f2)
{
  return DualFunctor<Func1, Func2>(f1, f2);
}


//template<Elment El>struct ElHowder{ElHowder(int){}};






// UpDown functors

template<class Functor, class Index>
class DownFunctor:public DoAller<DownFunctor<Functor, Index> >
{
  Functor& F;
public:
  typedef DownFunctor Type;
  DownFunctor(Functor& f):F(f){}

  template<Elment el, class T = void>
  struct Helper
  {
    static void Do(Functor &f, const Index&i)
    {
    }
  };
  template<class T>
  struct Helper<LeftEl, T>
  {
    static void Do(Functor &f, const Index & i)
    {
      f.template Do<LeftEl>(i);
    }
  };
  template<Elment El>
  void Do(const Index&i)
  {
    Helper<El>::Do(F, i);
  }

};


template<class Functor, class Index> 
class DownFunctorNo:public DoAller<DownFunctorNo<Functor, Index> >
{
  Functor& F;
public:
  typedef DownFunctorNo Type;
  DownFunctorNo(Functor& f):F(f){}
  template<Elment El>
  void Do(const Index&){}
    
};
template<class Functor, class Index>
class UpFunctor:public DoAller<UpFunctor<Functor, Index> >
{
  Functor& F;
public:
  typedef UpFunctor Type;
  UpFunctor(Functor& f):F(f){}


  template<Elment El, class T = void>
  struct Helper
  {
    static void Do(Functor &f, const Index&i)
    {
      f.template Do<El>(i);
    }
  };
  template<class T>
  struct Helper<LeftEl, T>
  {
    static void Do(Functor &f, const Index & i)
    {
    }
  };

  template<Elment El>
  void Do(const Index&i)
  {
    Helper<El>::Do(F, i);
  }

};

template<class Functor, class Index>
class UpFunctorNo:public DoAller<UpFunctorNo<Functor, Index> >{
  Functor& F;
public:
  typedef UpFunctorNo Type;
  UpFunctorNo(Functor& f):F(f){}

  template<Elment el, class T = void>
  struct Helper
  {
    static void Do(Functor &f, const Index&i)
    {
    }
  };
  template<class T>
  struct Helper<DownEl, T>
  {
    static void Do(Functor &f, const Index & i)
    {
      f.template Do<DownEl>(i);
    }
  };

  template<Elment El>
  void Do(const Index&i)
  {
    Helper<El>::Do(F, i);
  }
};



/// UpDown funktors end




/// UpDownInt functors

template<class Functor, class Index, class Int>
class DownFunctorInt//:public DoAller<DownFunctorInt<Functor, Index, Int> >
{
  Functor& F;
public:
  typedef DownFunctorInt Type;
  DownFunctorInt(Functor& f):F(f){}

  template<Elment el, class T = void>
  struct Helper
  {
    static void Do(Functor &f, const Index&i, Int)
    {
    }
  };
  template<class T>
  struct Helper<LeftEl, T>
  {
    static void Do(Functor &f, const Index & i, Int a)
    {
      f.template Do<LeftEl>(i, a);
    }
  };
  template<Elment El>
  void Do(const Index&i, Int a)
  {
    Helper<El>::Do(F, i, a);
  }

};


template<class Functor, class Index, class Int> 
class DownFunctorNoInt//:public DoAller<DownFunctorNoInt<Functor, Index, Int> >
{
  Functor& F;
public:
  typedef DownFunctorNoInt Type;
  DownFunctorNoInt(Functor& f):F(f){}
  template<Elment El>
  void Do(const Index&, Int){}
    
};
template<class Functor, class Index, class Int>
class UpFunctorInt//:public DoAller<UpFunctorInt<Functor, Index, Int> >
{
  Functor& F;
public:
  typedef UpFunctorInt Type;
  UpFunctorInt(Functor& f):F(f){}


  template<Elment El, class T = void>
  struct Helper
  {
    static void Do(Functor &f, const Index&i, Int a)
    {
      f.template Do<El>(i, a);
    }
  };
  template<class T>
  struct Helper<LeftEl, T>
  {
    static void Do(Functor &f, const Index & i, Int)
    {
    }
  };

  template<Elment El>
  void Do(const Index&i, Int a)
  {
    Helper<El>::Do(F, i, a);
  }

};

template<class Functor, class Index, class Int>
class UpFunctorNoInt//:public DoAller<UpFunctorNoInt<Functor, Index, Int> >
{
  Functor& F;
public:
  typedef UpFunctorNoInt Type;
  UpFunctorNoInt(Functor& f):F(f){}

  template<Elment el, class T = void>
  struct Helper
  {
    static void Do(Functor &f, const Index&i, Int)
    {
    }
  };
  template<class T>
  struct Helper<DownEl, T>
  {
    static void Do(Functor &f, const Index & i, Int a)
    {
      f.template Do<DownEl>(i, a);
    }
  };

  template<Elment El>
  void Do(const Index&i, Int a)
  {
    Helper<El>::Do(F, i, a);
  }
};

//// end UpDownInt functors






template<class Index, class Functor1, class Functor2, class Functor3>
void DoWithMacroEllementWorker(const Index &Ibase, Functor1 &f1, Functor2 &f2, Functor3 &f3)
{
  Index I = Ibase;
  Index J = I;
  Index K = J;
  typedef typename Index::Int Int;
  Int i = 0;

  f1.template Do<LeftEl>(K, i++);
  K.ToUp();
  f2.template Do<LeftEl>(K, i++);

  J.ToBack();K = J;
  f1.template Do<LeftEl>(K, i++);
  K.ToUp();
  f2.template Do<LeftEl>(K, i++);

  J = I; K = J;

  f1.template Do<FrontEl>(K, i++);
  K.ToUp();
  f2.template Do<FrontEl>(K, i++);

  K = J;
  f1.template Do<DownEl>(K, i++);
  K.ToUp();
  f2.template Do<DownEl>(K, i++);
  K.ToUp();
  f3.template Do<DownEl>(K, i++);

  J.ToBack(); K = J;

  f1.template Do<FrontEl>(K, i++);
  K.ToUp();
  f2.template Do<FrontEl>(K, i++);

  K = J;
  f1.template Do<DownEl>(K, i++);
  K.ToUp();
  f2.template Do<DownEl>(K, i++);
  K.ToUp();
  f3.template Do<DownEl>(K, i++);

  J.ToBack(); K = J;
  f1.template Do<FrontEl>(K, i++);
  K.ToUp();
  f2.template Do<FrontEl>(K, i++);

  I.ToRight(); J = I; K = J;

  f1.template Do<LeftEl>(K, i++);
  K.ToUp();
  f2.template Do<LeftEl>(K, i++);

  J.ToBack();K = J;
  f1.template Do<LeftEl>(K, i++);
  K.ToUp();
  f2.template Do<LeftEl>(K, i++);

  J = I; K = J;

  f1.template Do<FrontEl>(K, i++);
  K.ToUp();
  f2.template Do<FrontEl>(K, i++);

  K = J;
  f1.template Do<DownEl>(K, i++);
  K.ToUp();
  f2.template Do<DownEl>(K, i++);
  K.ToUp();
  f3.template Do<DownEl>(K, i++);

  J.ToBack(); K = J;

  f1.template Do<FrontEl>(K, i++);
  K.ToUp();
  f2.template Do<FrontEl>(K, i++);

  K = J;
  f1.template Do<DownEl>(K, i++);
  K.ToUp();
  f2.template Do<DownEl>(K, i++);
  K.ToUp();
  f3.template Do<DownEl>(K, i++);

  J.ToBack(); K = J;
  f1.template Do<FrontEl>(K, i++);
  K.ToUp();
  f2.template Do<FrontEl>(K, i++);

  I.ToRight(); J = I; K = J;


  // last phase //plane

  f1.template Do<LeftEl>(K, i++);
  K.ToUp();
  f2.template Do<LeftEl>(K, i++);

  J.ToBack();K = J;
  f1.template Do<LeftEl>(K, i++);
  K.ToUp();
  f2.template Do<LeftEl>(K, i++);
  
}





  


template<class Int, class Functor>
void DoForMacroElement(Functor &f, Index3dBase<Int> const &ind)
{

  DoWithMacroEllementWorker(ind, f, f, f);

}

template<class BigVector, class SmallVector> 
class SmallToBigCpy
{
public:
  //  typedef typename AVector::base_type Double;
  BigVector& big;
  SmallVector& small;
  SmallToBigCpy(BigVector &big_, SmallVector &small_):big(big_),small(small_)
  {
  }

  template<Elment El>
  void Do(Index3dBase<typename BigVector::int_type> const &ind, typename BigVector::int_type i)
  {
    big[ind][El][0] = small[i];
    big[ind][El][1] = small[i+36];
    big[ind][El][2] = small[i+72];
  }
};
template<class BigVector, class SmallVector> 
class SmallToBigAdd
{
public:
  //  typedef typename AVector::base_type Double;
  BigVector& big;
  SmallVector& small;
  SmallToBigAdd(BigVector &big_, SmallVector &small_):big(big_),small(small_)
  {
  }

  template<Elment El>
  void Do(Index3dBase<typename BigVector::int_type> const &ind, typename BigVector::int_type i)
  {
    big[ind][El][0] += small[i];
    big[ind][El][1] += small[i+36];
    big[ind][El][2] += small[i+72];
  }
};

template<class BigVector, class SmallVector> 
class BigToSmallCpy
{
public:
  //  typedef typename AVector::base_type Double;
  BigVector& big;
  SmallVector& small;
  BigToSmallCpy(BigVector &big_, SmallVector &small_):big(big_),small(small_)
  {
  }

  template<Elment El>
  void Do(Index3dBase<typename BigVector::int_type> const &ind, typename BigVector::int_type i)
  {
    small[i]    = big[ind][El][0];
    small[i+36] = big[ind][El][1];
    small[i+72] = big[ind][El][2];
  }
};





template<template<class A, class B, class C, class D> class Function, class Int, class Functor>
inline void
Execute(
	Index3dBase<Int> const &ind, 
	Functor &F
	, int Size
	, int Rank )
{
  typedef DownFunctor<Functor, Index3dBase<Int> > DownT;DownT Down(F);
  typedef DownFunctorNo<Functor, Index3dBase<Int> > NoDownT; NoDownT NoDown(F);
  typedef UpFunctor<Functor, Index3dBase<Int> > UpT;UpT Up(F);
  typedef UpFunctorNo<Functor, Index3dBase<Int> > NoUpT;NoUpT NoUp(F);
  Size--;

  if(Rank)
    if(Rank==Size){
      Function<Int, Functor, DownT, NoUpT> func;
      func(ind, F, Down, NoUp);
    }
    else{
      Function<Int, Functor, DownT, UpT> func;
      func(ind, F, Down, Up);
    }
  else
    if(Rank==Size){
      Function<Int, Functor, NoDownT, NoUpT> func;
      func(ind, F, NoDown, NoUp);
    }
    else{
      Function<Int, Functor, NoDownT, UpT> func;
      func(ind, F, NoDown, Up);
    }
}
template<class Int, class Functor>
inline void Execute1(Index3dBase<Int> const &ind, 
		     Functor &F, int Size, int Rank )
{
  typedef DownFunctor<Functor, Index3dBase<Int> > DownT;DownT Down(F);
  typedef DownFunctorNo<Functor, Index3dBase<Int> > NoDownT; NoDownT NoDown(F);
  typedef UpFunctor<Functor, Index3dBase<Int> > UpT;UpT Up(F);
  typedef UpFunctorNo<Functor, Index3dBase<Int> > NoUpT;NoUpT NoUp(F);
  Size--;

  if(Rank)
    if(Rank==Size){
      typename Functor::template method<Int, Functor, DownT, NoUpT> func;
      func(ind, F, Down, NoUp);
    }
    else{
      typename Functor::template method<Int, Functor, DownT, UpT> func;
      func(ind, F, Down, Up);
    }
  else
    if(Rank==Size){
      typename Functor::template method<Int, Functor, NoDownT, NoUpT> func;
      func(ind, F, NoDown, NoUp);
	    
    }
    else{
      typename Functor::template method<Int, Functor, NoDownT, UpT> func;
      func(ind, F, NoDown, Up);
    }
    
}



template< class V>
class PrintFunctor:public DoAller<PrintFunctor<V> >
{
  V const &vec;
public:
  typedef typename V::int_type Int;
  PrintFunctor(V const & d):vec(d){}
  template<Elment El>
  void Do(Index3dBase<Int>const& i)
  {
    static const char* arr[] = {"L", "F", "D"};
    std::cout << arr[El] << vec[i][El] << std::endl;
  }
};

template<class V>
class InitFunctor:public DoAller<InitFunctor<V> >
{
  V &v;   
  
public:
  typedef typename V::int_type Int;
  Int Count;
  InitFunctor(V &d):v(d), Count(0){}
  
  template<Elment El>
  void Do(Index3dBase<Int>const& i)
  {
    //    cout << "d" << i.Ind() << endl;
    v[i][El] = ++Count;
  }
};

template< class Int>
class TestFunctor
{
public:
  int Count;
  TestFunctor():Count(0){}
  void DoAll(Index3dBase<Int>const& i)
  {
    //    cout << "l" << i.Ind() << endl;
    //    cout << "f" << i.Ind() << endl;
    //    cout << "d" << i.Ind() << endl;
    Count+=3;
  }
  template<Elment El>
  void Do(Index3dBase<Int>const& i)
  {
    //    cout << "d" << i.Ind() << endl;
    Count++;
  }
};

template<class VecList, class Which, 
	 int cur = 0, bool B = cur == ttl::meta::length<Which>::value >
struct Multiplayer
{
  template<class Mat, class SVec, class DVec, class Iter>
  static void Do(Mat const &mat, SVec const &vec, DVec &res, const Iter&iter)
  {
    typedef typename DVec::base_type Double;
    const int ind = ttl::meta::get<Which, cur >::type::value;
    typedef typename ttl::meta::get<VecList, ind >::type VA;
    typedef typename ttl::meta::get<VecList, 0 >::type VA0;
    VA0::Do(res, iter) += VA ::Do(vec, iter) * mat[ind];
    VA ::Do(res, iter) += VA0::Do(vec, iter) * mat[ind];
    Multiplayer<VecList, Which, cur+1>::Do(mat, vec, res, iter);
  }
};
template<class VecList, class Which,
	 int cur>
struct Multiplayer<VecList, Which, cur, true>
{
  template<class Mat, class SVec, class DVec, class Iter>
  static void Do(Mat const &mat, SVec const &vec, DVec &res, const Iter&){}
};

template<class Matrix, class SrcVector, class ResVector>
class MultiplyMatVector:public DoAller<MultiplyMatVector<Matrix, SrcVector, ResVector> >
{
public:
  const Matrix& mat;
  ResVector &res;
  const SrcVector &src;
  MultiplyMatVector(const Matrix&m, const SrcVector&s, ResVector&r):mat(m), res(r), src(s)
  {
    ClearVector(res);
  }


  template<Elment El>
  void Do(Index3dBase<typename Matrix::int_type> const &ind)
  {
    typedef ttl::meta::typelist<
      int_<1>,
      int_<2>,
      int_<3>,
      int_<4>,
      int_<5>
      > All;

    typename Matrix::base_type &mr = mat[ind][El];
    res[ind][El] +=   mr[0] * src[ind][El];
    Multiplayer<typename PreconDescriptorBase::Lists<El>::Func,
      All>::Do(mr, src, res, ind);
  }
};

template<class AVector, class BVector> // A += B*alpha
class VecUpdate:public DoAller<VecUpdate<AVector, BVector> >
{
public:
  typedef typename AVector::base_type Double;
  AVector& avec;
  const BVector& bvec;
  Double alpha;
  bool first;
  VecUpdate(AVector &Avec, const BVector &Bvec, const Double Al):avec(Avec),bvec(Bvec), alpha(Al)
  {
  }
  template<Elment El>
  void Do(Index3dBase<typename AVector::int_type> const &ind)
  {
    avec[ind][El] += alpha * bvec[ind][El];
  }
};


template<class AVector, class BVector> // A += B*alpha
class VecUpdateClear:public DoAller<VecUpdateClear<AVector, BVector> >
{
public:
  typedef typename AVector::base_type Double;
  AVector& avec;
  BVector& bvec;
  Double alpha;
  bool first;
  VecUpdateClear(AVector &Avec, BVector &Bvec, const Double Al):avec(Avec),bvec(Bvec), alpha(Al)
  {
  }
  template<Elment El>
  void Do(Index3dBase<typename AVector::int_type> const &ind)
  {
    avec[ind][El] += alpha * bvec[ind][El];bvec[ind][El] = 0;
  }
};

template<class AVector, class BVector> // A -= B
class VecSub:public DoAller<VecSub<AVector, BVector> >
{
public:
  typedef typename AVector::base_type Double;
  AVector& avec;
  const BVector& bvec;
  VecSub(AVector &Avec, const BVector &Bvec):avec(Avec),bvec(Bvec)
  {
  }
  template<Elment El>
  void Do(Index3dBase<typename AVector::int_type> const &ind)
  {
    avec[ind][El] -=  bvec[ind][El];
  }
};

template<class AVector, class BVector> // A = B
class VecCpy:public DoAller<VecCpy<AVector, BVector> >
{
public:
  typedef typename AVector::base_type Double;
  AVector& avec;
  const BVector& bvec;
  VecCpy(AVector &Avec, const BVector &Bvec):avec(Avec),bvec(Bvec)
  {
  }

  template<Elment El>
  void Do(Index3dBase<typename AVector::int_type> const &ind)
  {
    avec[ind][El] =  bvec[ind][El];
  }
};

template<class AVector, class BVector> // A = -B + A*alpha
class VecUpdate2:public DoAller<VecUpdate2<AVector, BVector> >
{
public:
  typedef typename AVector::base_type Double;
  AVector& avec;
  const BVector& bvec;
  Double alpha;
  VecUpdate2(AVector &Avec, const BVector &Bvec, const Double Al):avec(Avec),bvec(Bvec), alpha(Al)
  {
  }
  template<Elment El>
  void Do(Index3dBase<typename AVector::int_type> const &ind)
  {
    avec[ind][El] = alpha * avec[ind][El] - bvec[ind][El];
  }

};
//#include "linux/prefetch.h"
template<class AVector, class BVector> // A*B
class VecDot1:public DoAller<VecDot1<AVector, BVector> >
{
public:
  typedef typename AVector::base_type Double;
  const AVector& avec;
  const BVector& bvec;
  static const int zero = 0x400;
  int c;
  Double sum, bigS;
  VecDot1(const AVector &Avec, const BVector &Bvec):avec(Avec),bvec(Bvec)
  {
    sum = 0;
    bigS = 0;
    c = 0;
  }
  template<Elment El>
  void Do(Index3dBase<typename AVector::int_type> const &ind)
  {
    Double f = avec[ind][El] * bvec[ind][El];
    sum += f;
    if(!(++c & zero))
    {
      bigS += sum;
      sum = 0;
      c = 0;
    }
  }
  Double GetSum()
  {
    Double ret = sum;
    ret += bigS;
    sum = 0;
    bigS = 0;
    c = 0;
    return ret;
  }
};
template<class AVector, class BVector> // A*B
class VecDot3:public DoAller<VecDot3<AVector, BVector> >
{
public:
  typedef typename AVector::base_type Double;
  const AVector& avec;
  const BVector& bvec;
  Double sum;
  VecDot3(const AVector &Avec, const BVector &Bvec):avec(Avec),bvec(Bvec)
  {
    sum = 0;
  }
  template<Elment El>
  void Do(Index3dBase<typename AVector::int_type> const &ind)
  {
    Double f = avec[ind][El] * bvec[ind][El];
    sum += Myf<Double>::Reduce(f);
  }
  Double GetSum()
  {
    return Double(sum);
  }
};

template<class AVector, class BVector> // A*B
class VecDot2:public DoAller<VecDot2<AVector, BVector> >
{
public:
  typedef typename AVector::base_type Double;
  const AVector& avec;
  const BVector& bvec;

  Double Sum;
  Double ErrSum;

  VecDot2(const AVector &Avec, const BVector &Bvec):avec(Avec),bvec(Bvec)
  {
    Sum = Double(); ErrSum = Double();
  }

  template<Elment El>
  void Do(Index3dBase<typename AVector::int_type> const &ind)
  {
    Double f = avec[ind][El] * bvec[ind][El];
    Double oldsum = Sum;
    Sum += f;
    Double err = f - (Sum - oldsum);
    ErrSum += err;
  }
  Double GetSum()
  {
    Double ret = Sum;
    ret += ErrSum;
    Sum = Double();
    ErrSum = Double();;
    return ret;
  }
};



template<class AVector> // A*B
class CNorm:public DoAller<CNorm<AVector> >
{
public:
  typedef typename AVector::base_type Double;
  const AVector& avec;

  Double max;

  CNorm(const AVector &Avec):avec(Avec)
  {
    max = Double();
  }

  template<Elment El>
  void Do(Index3dBase<typename AVector::int_type> const &ind)
  {
    max[0] = std::max(fabs(max[0]), fabs(avec[ind][El][0]));
    max[1] = std::max(fabs(max[1]), fabs(avec[ind][El][1]));
    max[2] = std::max(fabs(max[2]), fabs(avec[ind][El][2]));
    //    std::cout << avec[ind][El][0] << " " <<avec[ind][El][1] << " "  <<avec[ind][El][2] << std::endl;
  }
  Double GetNorm()
  {
    return max;
  }
};



template<class A, class B>
class VecDot:public VecDot1<A, B>
{
public:
  VecDot(const A &a, const B &b):VecDot1<A,B>(a,b){}
};

template<class VecList, class List, int cur =0, bool F = cur == ttl::meta::length<List>::value>
struct Cycler
{
  template<class Vec, class Ar, class Index>
  static void Do(Vec &vec, Ar const &ar, const Index &index)
  {
    const int ind = ttl::meta::get<List, cur >::type::value;
    typedef typename ttl::meta::get<VecList, 0 >::type VA0;
    typedef typename ttl::meta::get<VecList, ind >::type VA;
    VA0::Do(vec, index) += ar[ind];
    VA::Do (vec, index) += ar[ind];
    Cycler<VecList, List, cur+1>::Do(vec, ar, index);
  }
};
template<class VecList, class List, int cur>
struct Cycler<VecList, List, cur, true>
{
  template<class Vec, class Ar, class Index>
  static void Do(Vec &vec, Ar const &ar, const Index &index){}
};


template<class VecList, class List, int cur =0, bool F = cur == ttl::meta::length<List>::value>
struct Cycler1
{
  template<class Vec, class Ar, class Index>
  static void Do(Vec &vec, Ar const &ar, const Index &index)
  {
    const int ind = ttl::meta::get<List, cur >::type::value;
    typedef typename ttl::meta::get<VecList, 0 >::type VA0;
    typedef typename ttl::meta::get<VecList, ind >::type VA;
    VA0::Do(vec, index) += ar[ind]*(ar[ind]>0);
    VA::Do (vec, index) += ar[ind]*(ar[ind]>0);

    Cycler1<VecList, List, cur+1>::Do(vec, ar, index);
  }
};
template<class VecList, class List, int cur>
struct Cycler1<VecList, List, cur, true>
{
  template<class Vec, class Ar, class Index>
  static void Do(Vec &vec, Ar const &ar, const Index &index){}
};


template<class Matrix, class Vector, 
	 class PD>
class PreconCreatorStep1:public DoAller<PreconCreatorStep1<Matrix, Vector, PD> >
{
  const Matrix& mat;
  Vector &vec;
  typedef typename Vector::base_type Double;
  Double t;
  typedef typename Matrix::base_type SVec;
  
public:

  typedef typename Matrix::int_type Int;
  typedef Index3dBase<typename Vector::int_type> Index;
  template<class A, class B, class C, class D>
  class method:public PD::template Creator1<A, B, C, D> {};

  PreconCreatorStep1(const Matrix&m, Vector &v):mat(m), vec(v)
  {
    ClearVector(vec);
  }
  template<Elment El>
  void Do(Index3dBase<typename Vector::int_type> const &ind)
  {
    vec[ind][El] += mat[ind][El][0];
    Cycler<typename PD::template Lists<El>::Func, 
      typename PD::template Lists<El>::NotIntrestedIn>::Do(vec, mat[ind][El], ind);
    Cycler1<typename PD::template Lists<El>::Func, 
      typename PD::template Lists<El>::IntrestedIn>::Do(vec, mat[ind][El], ind);
  }
  
};

template<class List, int cur =0, bool F = cur == ttl::meta::length<List>::value>
struct Summer
{
  template<class T>  
  static typename T::Element_type Sum(const T& a)
  {
    return a[ttl::meta::get<List, cur >::type::value] + Summer<List, cur+1>::Sum(a);
  }
};
template<class List, int cur>
struct Summer<List, cur, true>
{
  template<class T>
  static typename T::Element_type Sum(const T& a)
  {
    return 0;
  }
};

template<class VecList, class List, int cur =0, bool F = cur == ttl::meta::length<List>::value>
struct Subber
{
  template<class Vec, class Ar, class Double, class Index>
  static void Sub(Vec &vec, Ar const &ar, Double const & s, const Index &index)
  {
    const int ind = ttl::meta::get<List, cur >::type::value;
    typedef typename ttl::meta::get<VecList, ind >::type VA;
    Double a = ar[ind];
    a *= (a < 0);
    VA::Do(vec, index) -= a * s;
    Subber<VecList, List, cur+1>::Sub(vec, ar, s, index);
  }
};
template<class VecList, class List, int cur>
struct Subber<VecList, List, cur, true>
{
  template<class Vec, class Ar, class Double, class Index>
  static void Sub(Vec &vec, Ar const &ar, Double const & s, const Index&)
  {
  }
};


template<class VecList, class List, int cur =0, bool F = cur == ttl::meta::length<List>::value>
struct Subber2
{
  template<class Vec, class Ar, class Double, class Index>
  static void Sub(Vec &vec, Ar const &ar, Double  & s, const Index &index)
  {
    const int ind = ttl::meta::get<List, cur >::type::value;
    typedef typename ttl::meta::get<VecList, ind >::type VA;
    Double a = ar[ind];
    a *= (a <0);
    s -= VA::Do(vec, index) * a;
    Subber2<VecList, List, cur+1>::Sub(vec, ar, s, index);
  }
};
template<class VecList, class List, int cur>
struct Subber2<VecList, List, cur, true>
{
  template<class Vec, class Ar, class Double, class Index>
  static void Sub(Vec &vec, Ar const &ar, Double  & s, const Index &index){}
};

template<class Matrix, class Vector, 
	 class PD>

class PreconCreatorStep1_5:public DoAller<PreconCreatorStep1_5<Matrix, Vector, PD> >
{
  const Matrix& mat;
  Vector &vec;
  typedef typename Vector::base_type Double;
  Double t, ts;
  typedef typename Matrix::base_type SVec;
public:
  typedef typename Matrix::int_type Int;
  template<class A, class B, class C, class D>
  class method:public PD::template Creator1_5<A, B, C, D> {};


  template<Elment El>
  Double Sum(const SVec&v)
  {
    return Summer<typename PD::template Lists<El>::IntrestedIn>::Sum(v);
  }

  PreconCreatorStep1_5(const Matrix&m, Vector &v):mat(m), vec(v)
  {
    t = 1./(std::max(vec.N1, std::max(vec.N2, vec.N3R)));
    t *= GetPerturb<Double>();
    t /= Options::GetOptions().pfactor;
    //    std::cerr << t << std::endl;
    //t /= 5;

    ts = t*t;


  }
  template<Elment El>
  void Do(Index3dBase<typename Vector::int_type> const &ind)
  {

    Double k = Sum<El>(mat[ind][El]);
    if (vec[ind][El]> - 2.* k)
      k = t;
    else
      k = ts;
    vec[ind][El] += vec[ind][El] * k * 2  * Options::GetOptions().ppfactor;
  }
};



template<class Matrix, class Vector, class PD >
class PreconCreatorStep2:public DoAller<PreconCreatorStep2<Matrix, Vector, PD> >
{
  const Matrix& mat;
  Vector &vec;
  
  typedef typename Vector::base_type Double;
  typedef typename Matrix::base_type SVec;
public:
  typedef typename Matrix::int_type Int;

  typedef Index3dBase<typename Vector::int_type> Index;
  template<class A, class B, class C, class D>
  class method:public PD::template Creator2<A, B, C, D> {};

  Double savebuf[2];
  void SaveState(Index const &Id, Index const & Iu)
  {
    savebuf[0] = vec.Up(Id)[DownEl];
    savebuf[1] = vec.Right(Id)[LeftEl];
  }
  void RestoreState(Index const &Id, Index const & Iu)
  {
    vec.Up(Id)[DownEl] = savebuf[0];      
    vec.Right(Id)[LeftEl] = savebuf[1];
  }

  void StartSendLeft(Index const &Id, Index const & Iu)
  {
    vec.StartFwdLeft(Id, Iu);
  }
  void FinishSendLeft(Index const &Id, Index const &Iu)
  {
    vec.FinishFwdLeft(Id, Iu);
  }
  void StartSendM4Up(Index const &Id, Index const & Iu)
  {
    vec.StartFwdM4(Id, Iu);
  }
  void FinishSendM4Up(Index const &Id, Index const &Iu)
  {
    vec.FinishFwdM4(Id, Iu);
  }
  void StartSendFront(Index const &Id, Index const & Iu)
  {
    vec.StartFwdFront(Id, Iu);
  }
  void FinishSendFront(Index const &Id, Index const &Iu)
  {
    vec.FinishFwdFront(Id, Iu);
  }
  void StartSendDown(Index const &Id, Index const & Iu)
  {
    vec.StartFwdDown(Id, Iu);
  }
  void FinishSendDown(Index const &Id, Index const &Iu)
  {
    vec.FinishFwdDown(Id, Iu);
  }


  void StartSendLineLeft(Index const &Id, Index const &Iu)
  {
    vec.StartFwdLineLeft(Id, Iu);
  }
  void FinishSendLineLeft(Index const &Id, Index const &Iu)
  {
    vec.FinishFwdLineLeft(Id, Iu);
  }
  void StartSendLineMiddle(Index const &Id, Index const &Iu)
  {
    vec.StartFwdLineMiddle(Id, Iu);
  }
  void FinishSendLineMiddle(Index const &Id, Index const &Iu)
  {
    vec.FinishFwdLineMiddle(Id, Iu);
  }

  

  PreconCreatorStep2(const Matrix&m, Vector &v):mat(m), vec(v)
  {
  }


  template<Elment El>
  Double Sum(const SVec&v)
  {
    return Summer<typename PD::template Lists<El>::IntrestedIn>::Sum(v);
  }


  template<Elment El>
  void Do(Index3dBase<typename Vector::int_type> const &ind)
  {
    Double ss = Sum<El>(mat[ind][El])/vec[ind][El];
    Subber<typename PD::template Lists<El>::Func, typename PD::template Lists<El>::IntrestedIn>::Sub(vec, mat[ind][El], ss, ind);

    Double &k = vec[ind][El];
    if(k<0.)
    {
      std::cerr << "x < 0 " << k << std::endl;
      //      while (k < 0) k+= Sum<El>(mat[ind][El])/vec[ind][El];
      throw "negativ number exception";
    }
    k = 1./k;
  }
};

template<class Matrix, class XVector, class ResVector, class SrcVector, 
	 class PD>
class PreconSolverStep2:public DoAllerReverse<PreconSolverStep2<Matrix, XVector, ResVector, SrcVector, PD> >
{
  const Matrix& mat;
  const XVector &X;
  ResVector &Res;
  const SrcVector &Src;
  std::string timer;
public:
  typedef typename Matrix::int_type Int;
  typedef typename ResVector::base_type Double;
  typedef Index3dBase<Int> Index;


  template<class A, class B, class C, class D>
  class method:public PD::template Solver2<A, B, C, D> {};

  PreconSolverStep2(const Matrix&m, const XVector &x, ResVector &res, 
		    const SrcVector &src , const std::string &tim)
    :mat(m), X(x), Res(res), Src(src),timer(tim)
  {    
  }
 

  void StartSendLeft(Index const &Id, Index const & Iu)
  {
    TimeCalc(timer);
    Res.StartBwdLeft(Id, Iu);
  }
  void FinishSendLeft(Index const &Id, Index const &Iu)
  {
    TimeCalc(timer);
    Res.FinishBwdLeft(Id, Iu);
  }
  void StartSendFront(Index const &Id, Index const & Iu)
  {
    TimeCalc(timer);
    Res.StartBwdFront(Id, Iu);
  }
  void FinishSendFront(Index const &Id, Index const &Iu)
  {
    TimeCalc(timer);
    Res.FinishBwdFront(Id, Iu);
  }
  void StartSendDown(Index const &Id, Index const & Iu)
  {
    TimeCalc(timer);
    Res.StartBwdDown(Id, Iu);
  }
  void FinishSendDown(Index const &Id, Index const &Iu)
  {
    TimeCalc(timer);
    Res.FinishBwdDown(Id, Iu);
  }

  void StartSendLineLeft(Index const &Id, Index const &Iu)
  {
    TimeCalc(timer);
    Res.StartBwdLineLeft(Id, Iu);
  }
  void FinishSendLineLeft(Index const &Id, Index const &Iu)
  {
    TimeCalc(timer);
    Res.FinishBwdLineLeft(Id, Iu);
  }
  void StartSendLineMiddle(Index const &Id, Index const &Iu)
  {
    TimeCalc(timer);
    Res.StartBwdLineMiddle(Id, Iu);
  }
  void FinishSendLineMiddle(Index const &Id, Index const &Iu)
  {
    TimeCalc(timer);
    Res.FinishBwdLineMiddle(Id, Iu);
  }
  void StartSendM4Down(Index const &Id, Index const & Iu)
  {
    TimeCalc(timer);
    Res.StartBwdM4(Id, Iu);
  }
  void FinishSendM4Down(Index const &Id, Index const &Iu)
  {
    TimeCalc(timer);
    Res.FinishBwdM4(Id, Iu);
  }


  
  /*   void  SRFrontDownDown(Int i) */
  /*   { */
  /*     Res.SendRecvSolve2Down(i, comm); */
  /*   } */
  /*   void  SRFrontDownDownFinish(Int i) */
  /*   { */
  /*     Res.SendRecvSolve2DownFinish(i, comm); */
  /*   } */
  
  /*   void SRLeftDown(Int i) */
  /*   { */
  /*     Res.SendRecvSolve2Left(i+1, comm); */
  /*   } */
  /*   void SRLeftDownFinish(Int i) */
  /*   { */
  /*     Res.SendRecvSolve2LeftFinish(i+1, comm); */
  /*   } */
  
  template<Elment El>
  void Do(Index3dBase<typename XVector::int_type> const &ind)  
  {
    Double &h = Res[ind][El];
    typedef typename Matrix::base_type Row;
    Row & ma = mat[ind][El];

    Subber2<typename PD::template Lists<El>::Func, typename PD::template Lists<El>::IntrestedIn>::Sub(Res, ma, h, ind);
    
    h *= X[ind][El];    
    //Res[ind][El] = Src[ind][El]/mat[ind][El][0];
  }
};



template<class Matrix, class XVector, class ResVector, class SrcVector,
	 class PD>
class PreconSolverStep1_1_5 :public DoAller<PreconSolverStep1_1_5<Matrix, XVector, ResVector, SrcVector, PD> >
{
  const Matrix& mat;
  const XVector &X;
  ResVector &Res;
  const SrcVector &Src;
  std::string timer;
public:
  typedef typename XVector::int_type Int;
  typedef typename ResVector::base_type Double;
  typedef Index3dBase<Int> Index;
  template<class A, class B, class C, class D>
  class method:public PD::template Solver1<A, B, C, D> {};

  PreconSolverStep1_1_5(const Matrix&m, const XVector &x, ResVector &res, const SrcVector &src, const std::string &tim)
    :mat(m), X(x), Res(res), Src(src), timer(tim)
  {
    //    Begin();
  }

  void Begin()
  {
    ClearVector(Res);
  };

  Double savebuf[2];
  void SaveState(Index const &Id, Index const & Iu)
  {
    savebuf[0] = Res.Up(Id)[DownEl];
    savebuf[1] = Res.Right(Id)[LeftEl];
  }
  void RestoreState(Index const &Id, Index const & Iu)
  {
    Res.Up(Id)[DownEl] = savebuf[0];      
    Res.Right(Id)[LeftEl] = savebuf[1];
  }
  void StartSendM4Up(Index const &Id, Index const & Iu)
  {
    TimeCalc(timer);
    Res.StartFwdM4(Id, Iu);
  }
  void FinishSendM4Up(Index const &Id, Index const &Iu)
  {
    TimeCalc(timer);
    Res.FinishFwdM4(Id, Iu);
  }


  void StartSendLeft(Index const &Id, Index const & Iu)
  {
    TimeCalc(timer);
    Res.StartFwdLeft(Id, Iu);
  }
  void FinishSendLeft(Index const &Id, Index const &Iu)
  {
    TimeCalc(timer);
    Res.FinishFwdLeft(Id, Iu);
  }
  void StartSendFront(Index const &Id, Index const & Iu)
  {
    TimeCalc(timer);
    Res.StartFwdFront(Id, Iu);
  }
  void FinishSendFront(Index const &Id, Index const &Iu)
  {
    TimeCalc(timer);
    Res.FinishFwdFront(Id, Iu);
  }
  void StartSendDown(Index const &Id, Index const & Iu)
  {
    TimeCalc(timer);
    Res.StartFwdDown(Id, Iu);
  }
  void FinishSendDown(Index const &Id, Index const &Iu)
  {
    TimeCalc(timer);
    Res.FinishFwdDown(Id, Iu);
  }

  void StartSendLineLeft(Index const &Id, Index const &Iu)
  {
    TimeCalc(timer);
    Res.StartFwdLineLeft(Id, Iu);
  }
  void FinishSendLineLeft(Index const &Id, Index const &Iu)
  {
    TimeCalc(timer);
    Res.FinishFwdLineLeft(Id, Iu);
  }
  void StartSendLineMiddle(Index const &Id, Index const &Iu)
  {
    TimeCalc(timer);
    Res.StartFwdLineMiddle(Id, Iu);
  }
  void FinishSendLineMiddle(Index const &Id, Index const &Iu)
    {
      TimeCalc(timer);
      Res.FinishFwdLineMiddle(Id, Iu);
    }

  template<Elment El>
  void Do(Index3dBase<typename XVector::int_type> const &ind)  
  {
    Double &s = Res[ind][El];
    Double r = s;
    Double  &a = Src[ind][El];
    r += a;
    s = r;
    r *= X[ind][El];
    typedef typename Matrix::base_type Row;
    Row & ma = mat[ind][El];

    Subber<typename PD::template Lists<El>::Func, typename PD::template Lists<El>::IntrestedIn>::Sub(Res, ma, r, ind);
  }
};

EXPORT
template<
  template<class, class> class Functor, 
  class VecType1,
  class VecType2> 
void 
VecVecOper(VecType1 &vec1, VecType2 &vec2);
EXPORT
template<
  template<class, class> class Functor, 
  class VecType1,
  class VecType2,
  class T> 
void 
VecVecOper(VecType1 &vec1, VecType2 &vec2,  T val);
EXPORT
template<
  template<class, class> class Functor, 
  class VecType1,
  class VecType2> 
typename Myf<typename VecType1::base_type>::base_type
VecVecReduce(VecType1 &vec1, VecType2 &vec2);
EXPORT
template<
  template<class> class Functor, 
  class VecType1
  > 
typename Myf<typename VecType1::base_type>::base_type
VecReduceNorm(VecType1 &vec1);


#endif
