#ifndef __functors_hpp__
#define __functors_hpp__

#include <iostream>
#include "visitors.h"

#include "functors.h"
#include "useful.h"


template<
  template<class, class> class Functor, 
  class VecType1,
  class VecType2> 
void 
VecVecOper(VecType1 &vec1, VecType2 &vec2)
{
  TimeCalc1("VecVec");
  Functor<VecType1, VecType2> func(vec1, vec2);
  Index3dBase<typename VecType1::int_type>
    ind(vec1.N1, vec1.N2, vec1.N3);
  Execute<ForAll>(ind, func, vec1.size, vec2.rank);
}

template<
  template<class, class> class Functor, 
  class VecType1,
  class VecType2,
  class T> 
void 
VecVecOper(VecType1 &vec1, VecType2 &vec2,  T val)
{
  TimeCalc1("VecVecAlfa");
  Functor<VecType1, VecType2> func(vec1, vec2, val);
  Index3dBase<typename VecType1::int_type>
    ind(vec1.N1, vec1.N2, vec1.N3);
  Execute<ForAll>(ind, func, vec1.size, vec2.rank);
}

template<
  template<class, class> class Functor, 
  class VecType1,
  class VecType2> 
typename Myf<typename VecType1::base_type>::base_type
VecVecReduce(VecType1 &vec1, VecType2 &vec2)
{
  TimeCalc1("VecDot");
  Functor<VecType1, VecType2> func(vec1, vec2);
  Index3dBase<typename VecType1::int_type>
    ind(vec1.N1, vec1.N2, vec1.N3);
  Execute<ForAll>(ind, func, vec1.size, vec1.rank);
  typedef typename VecType1::base_type Double;
  Double sum = func.GetSum();

  if(!OVERLAP)
    vec1.comm.Barrier();

  {
    TimeCalc1("VecDotComm");
    typename Myf<typename VecType1::base_type>::base_type res, tmp;
    tmp =  Myf<Double>::Reduce(sum);
    vec1.comm.Allreduce(&tmp, &res, 1, Myf<    typename Myf<typename VecType1::base_type>::base_type>::Datatype(), MPI::SUM);

    return res;
  }
}


template<
  template<class> class Functor, 
  class VecType1
  > 
typename Myf<typename VecType1::base_type>::base_type
VecReduceNorm(VecType1 &vec1)
{
  Functor<VecType1> func(vec1);
  Index3dBase<typename VecType1::int_type>
    ind(vec1.N1, vec1.N2, vec1.N3);
  Execute<ForAll>(ind, func, vec1.size, vec1.rank);
  typedef typename VecType1::base_type Double;
  Double sum = func.GetNorm();

  if(!OVERLAP)
    vec1.comm.Barrier();

  {
    TimeCalc1("VecDotComm");
    typename Myf<typename VecType1::base_type>::base_type res, tmp;
    tmp =  Myf<Double>::Max(sum);
    vec1.comm.Allreduce(&tmp, &res, 1, Myf<    typename Myf<typename VecType1::base_type>::base_type>::Datatype(), MPI::MAX);
    return res;
  }

}


#endif
