#ifndef __ntransform_h__
#define __ntransform_h__
//#define for if(false);else for

#include <valarray>
#include "multipoly.h"
#ifdef assert
#undef assert
#endif
#define assert(X)
#include "base.h"
#include "calc.h"





template<class Vec1, class Vec2, class Value>
inline void MapApply(Vec1 &res, const Vec2&src, const Value &val)
{
  for (int i = 0; i < src.Size(); ++i)
    res[i] = src[i](val);
}


template<class ResMatr, class Amat, class Bmat>
inline void MatrixMultiply(ResMatr &Res, Amat const&A, Bmat const &B)
{
  assert((A.Width()==B.Height())&&(Res.Height()==A.Height())&&(Res.Width()==B.Width()));
  for(int i  = 0; i < A.Height(); ++i)
  {
    for(int k = 0; k < Res.Width();k++)
      Res[i][k] = 0.0; // the whole row is 0
    for(int k = 0; k < A.Width();k++)
    {
      for(int j = 0; j < B.Width(); ++j)
	Res[i][j] += A[i][k] * B[k][j];
    }
  }
}
template<class ResMatr, class Amat>
inline void Transpose(ResMatr &Res, const Amat &A)
{
  assert((A.Width()==Res.Height())&&(Res.Width()==A.Height()));
  for(int i = 0; i < A.Height(); ++i)
    for(int j = 0; j < A.Width(); ++j)
      Res[j][i] = A[i][j];
}

template<class Matr, class Xvec,class Rhsvec>
inline void GausJordanSolve(Matr &A, Xvec &X, Rhsvec &Rhs)
{
  int N = Rhs.Size();
  std::vector<int> ind(N);

  typedef typename Xvec::Element_type Double;

  Double max;

  for(int i = 0; i < N; ++i)
    ind[i] = i;
  int imax;
  
  for(int i = 0; i < N; ++i)
  {
    imax = i;
    max = fabs(A[i][ind[i]]);
    //find pivot element on i-th row;
    for(int k = i; k < N; ++k)
      if(fabs(A[i][ind[k]])>max)
      {
	imax = k;
	max = fabs(A[i][ind[k]]);
      }
    std::swap(ind[i], ind[imax]);
      
    max = 1/A[i][ind[i]];
    Rhs[i]*=max;
    for(int k = i; k < N; ++k)
      A[i][ind[k]] *= max;

    for (int j = 0; j < N; ++j)
    {
      if(j != i)
      {
	max = A[j][ind[i]];
	Rhs[j] -= A[j][ind[i]] *Rhs[i] ;
	for(int k = i; k < N; ++k)
	{
	  A[j][ind[k]] -= max*A[i][ind[k]] ;
	}
      }
    }
  }
  for(int i = 0; i < N; i++)
    X[i] = Rhs[ind[i]];
      
}

template<class Matr, class RMatr>
inline void GausJordanInverse(RMatr &R, Matr &A)
{
  int N = A.Size();
  std::vector<int> ind(N);

  typedef typename Matr::Element_type Double;

  Double max, max1;
  for(int i = 0; i < N; ++i)
  {
    ind[i] = i;
    R[i] = 0;
    R[i][i] = 1;
  }
  int imax;
  
  for(int i = 0; i < N; ++i)
  {
    imax = i;
    max = fabs(A[i][ind[i]]);
    //find pivot element on i-th row;
    for(int k = i; k < N; ++k)
      if(fabs(A[i][ind[k]])>max)
      {
	imax = k;
	max = fabs(A[i][ind[k]]);
      }
    std::swap(ind[i], ind[imax]);
      
    max = 1/A[i][ind[i]];
    A[i] *= max;
    R[i] *= max;

    for (int j = 0; j < N; ++j)
    {
      if(j != i)
      {
	max = A[j][ind[i]];
	A[j] -= max * A[i];
	R[j] -= max * R[i];
	// 	      for(int k = 0; k < N; ++k)
	// 		{
	// 		  A[j][ind[k]] -= max*A[i][ind[k]];
	// 		  R[j][ind[k]] -= max*R[i][ind[k]];
	// 		}
      }
    }
  }
  for(int i = 0; i < N; i++)
  {
    if(ind[i]!=i)
    {
      std::swap(A[i], A[ind[i]]);
      std::swap(R[i], R[ind[i]]);
      std::swap(ind[i], ind[ind[i]]);
    }
  }
      
}

template<class Matr>
void DoLU(Matr &mat, std::vector<int>&c)
{
  typedef typename Matr::Element_type Double;
  //  std::vector<int> c;
  int N = mat.Height();
  c.resize(N);
  for(int i = 0; i <N; ++i)c[i]=i;

  for(int j = 0; j<N; j++)
  {
    for(int i = 0; i <=j; ++i)
      for(int k = 0; k <i; ++k)
	mat[c[i]][j] -= mat[c[i]][k] * mat[c[k]][j];
    for(int i = j+1; i <N; ++i)
      for(int k = 0; k <j; ++k)
	mat[c[i]][j] -= mat[c[i]][k] * mat[c[k]][j];

    int maxk = j;
    for(int k = j+1; k < N; k++)
      if(fabs(mat[c[k]][j])>fabs(mat[c[maxk]][j]))
	maxk = k;
    std::swap(c[maxk], c[j]);
    for(int i = j+1; i<N; ++i)
      mat[c[i]][j]/= mat[c[j]][j];
  }	
}

template<class Matr, class Vec>
void BackSubst(Matr const&mat, Vec &vec, Vec &x, std::vector<int>&c)
{
  
  typedef typename Matr::Element_type Double;
  int N = mat.Height();
  for(int i = 0; i <N; ++i)
    for(int j = 0; j < i; j++)
      vec[c[i]] -= mat[c[i]][j] * vec[c[j]];
  //      vec[c[i]]/=1; // beautiful optimization

  for (int i = N; i--;)
  {
    for(int j = N; --j>i;)
      vec[c[i]] -=  mat[c[i]][j]*vec[c[j]];
    vec[c[i]] /=mat[c[i]][i];
  }
  for (int i = 0; i < N; ++i)
    x[i] = vec[c[i]];
}



template<class Mat3x3>
inline typename Mat3x3::Element_type GetAdj(Mat3x3 &mat, int i, int j)
{
  int j1, j2, i1,i2;
  i1 = (i+1)%3;
  j1 = (j+1)%3;
  i2 = 3 - i - i1;
  j2 = 3 - j - j1;
  return mat[i1][j1]*mat[i2][j2] - mat[i1][j2]*mat[i2][j1];
  
}

template<class Double> 
class TrilinearTransform
{
public:
  typedef My::Base::MultiPoly<Double, 3> Poly3d;
  typedef typename Poly3d::Var_type Point3d; 
  typedef SmallMatrix<Double, 8, 3> Points_type;

  SmallMatrix<Poly3d, 1, 3> X;
  SmallMatrix<Poly3d, 3, 8> DxDt;
  SmallMatrix<Double, 3, 8> DxDtPoint;
  

  SmallMatrix<Poly3d, 1, 8> basefuncts;
  
  SmallMatrix<Double, 8, 8> Rmat, LU;
  std::vector<int> permut;
  
  SmallMatrix<Double, 8, 3> coeffs, coeffs1; 
  bool initialized;

  TrilinearTransform(): initialized(false){};

  void SetDest(const Points_type &destpoints)
  {
    Init();
    //    MatrixMultiply(coeffs, Rmat, destpoints);
    Points_type r = destpoints;
    BackSubst(LU, r, coeffs, permut);
 
    //MatrixMultiply(X, basefuncts, coeffs);

  }

  Point3d toX(const Point3d &a)
  {
    SmallVector<Double, 8> fia;
    fia[0] = 1;
    fia[1] = a[0];
    fia[2] = a[1];
    fia[3] = a[2];
    fia[4] = a[0]*a[1];
    fia[5] = a[0]*a[2];
    fia[6] = a[1]*a[2];
    fia[7] = fia[6]*a[0];
    Point3d ret;
    ret = 0.0;
    for(int i = 0; i <8 ;++i)
    {
      ret[0]+= fia[i]*coeffs[i][0];
      ret[1]+= fia[i]*coeffs[i][1];
      ret[2]+= fia[i]*coeffs[i][2];
    }
    return ret;
  }
 

  template<class Mat3x3>
  void GetJacobianAndReverseAtPoint(Mat3x3 &JacobMatI, Double &Det, const Point3d &p)
  {
    SmallMatrix<Double, 3, 3> JacobMat;
    
    for(int i = 0; i < 3; ++i)
      for(int j = 0; j < 8; ++j)
	DxDtPoint[i][j] = DxDt[i][j](p);
    MatrixMultiply(JacobMat, DxDtPoint, coeffs);
    for(int i = 0; i < 3; ++i)
    {
      for(int j = 0; j< 3; ++j)
	JacobMatI[i][j] = GetAdj(JacobMat, i, j);
    }

    Det = 0;
    for(int i = 0; i < 3; i++)
      Det += JacobMat[0][i] * JacobMatI[0][i];
    //    cout << "Det " << Det << endl;

  }

  void Init()
  {
    if(!initialized)
    {
      Poly3d t1;
      Poly3d t2;
      Poly3d t3;
	
      t1[1] = 1;
      t2[2] = 1;  
      t3[3] = 1;

      basefuncts[0][0][0] = 1;
      basefuncts[0][1] = t1;
      basefuncts[0][2] = t2;
      basefuncts[0][3] = t3;
      basefuncts[0][4] = t1*t2;
      basefuncts[0][5] = t1*t3;
      basefuncts[0][6] = t2*t3;
      basefuncts[0][7] = t1*t2*t3;
	
      Points_type srcpoints;

      srcpoints[0][0] = -1; srcpoints[0][1] = -1; srcpoints[0][2] = -1;
      srcpoints[1][0] =  1; srcpoints[1][1] = -1; srcpoints[1][2] = -1;
      srcpoints[2][0] =  1; srcpoints[2][1] =  1; srcpoints[2][2] = -1;
      srcpoints[3][0] = -1; srcpoints[3][1] =  1; srcpoints[3][2] = -1;

      srcpoints[4][0] = -1; srcpoints[4][1] = -1; srcpoints[4][2] =  1;
      srcpoints[5][0] =  1; srcpoints[5][1] = -1; srcpoints[5][2] =  1;
      srcpoints[6][0] =  1; srcpoints[6][1] =  1; srcpoints[6][2] =  1;
      srcpoints[7][0] = -1; srcpoints[7][1] =  1; srcpoints[7][2] =  1;

      SmallMatrix<Double, 8, 8> mat;
      for(int i = 0; i < 8; i++)
	MapApply(mat[i], basefuncts[0], srcpoints[i]);

      LU = mat;
      //	GausJordanInverse(Rmat, mat);
      DoLU(LU, permut);

      using namespace std;


      for(int i = 0; i < 3; ++i)
	for(int j = 0; j < 8; ++j)
	  DxDt[i][j] = basefuncts[0][j].D(i);

      initialized = true;
    }
  }


};


template<class T>
void GetKwadraturePoints(std::valarray<T>&a, const T&left, const T&right)
{
  My::Calc::Vector<T> vec(a.size());
  My::Calc::OrthoRoots(vec, left, right);
  for (unsigned int i = 0; i < a.size(); i++)
    a[i] = vec(i+1);
}



template<class T>
void GetKwadratureCoeff(std::valarray<T> &coeff, const std::valarray<T> &points, const T&left, const T&right)
{
  using namespace My::Calc;
  assert(coeff.size()==points.size() );
  for (unsigned int i = 0; i <coeff.size(); i++)
  {
    Polinom<T> p(0);
    p[0] = 1;
    for (unsigned int j = 0; j < coeff.size(); j++)
      if (i!=j)
      {
	Polinom<T> hh(1);
	hh[1] = 1;
	hh[0] = -points[j];
	p = p*hh;
      }
    p = p*p;
    coeff[i] = T(1)/p(points[i]) * Integrate(p, left, right);
  }

}


template<class Double, int numpoints>
class IntegralCube
{
public:
  typedef SmallVector<Double, 3> Point3d; 

  static const int PointsNumber = numpoints*numpoints*numpoints;
  SmallVector<Point3d, PointsNumber>  points;
  SmallVector<Double, PointsNumber> coeff;


  IntegralCube()
  {
    std::valarray<double> pts(numpoints);
    std::valarray<double> cfs(numpoints);

    GetKwadraturePoints(pts,(double)(-1), (double)(1));
    GetKwadratureCoeff(cfs, pts,  (double)(-1), (double)(1));

    for(int i = 0; i < numpoints; i++)
      for(int j = 0; j < numpoints; j++)
	for(int k = 0; k < numpoints;k++)
	{
	  int ind = i * numpoints*numpoints + j * numpoints + k;
	  points[ind][0] = pts[i];
	  points[ind][1] = pts[j];
	  points[ind][2] = pts[k];
	  coeff[ind] = cfs[i]*cfs[j]*cfs[k];
	}
    
  }
    
  
};

template<class Double, int numpoints>
class IntegralSquare
{
public:
  typedef SmallVector<Double, 2> Point2d; 

  static const int PointsNumber = numpoints*numpoints;
  SmallVector<Point2d, PointsNumber>  points;
  SmallVector<Double, PointsNumber> coeff;


  IntegralSquare()
  {
    std::valarray<double> pts(numpoints);
    std::valarray<double> cfs(numpoints);

    GetKwadraturePoints(pts,(double)(-1), (double)(1));
    GetKwadratureCoeff(cfs, pts,  (double)(-1), (double)(1));

    for(int i = 0; i < numpoints; i++)
      for(int j = 0; j < numpoints; j++)
	{
	  int ind = i * numpoints + j ;
	  points[ind][0] = pts[i];
	  points[ind][1] = pts[j];
	  coeff[ind] = cfs[i]*cfs[j];
	}
    
  }
    
  
};



#endif
