#define MPICH_IGNORE_CXX_SEEK
#include "mpi.h"
#include "basic.h"
// IBM workaround
#ifdef _MY_CPP
#include "mympicxx.hpp"
#endif

#include<typeinfo>
#include <iostream>
#include <cstdlib>
#include <typeinfo>
#include <iomanip>
#include <cmath>
#include <vector>
#include <string>
#include <sstream>
#include <stdexcept>




#include "small.h" 
#include "sysgen.h"
#include "index.h"
#include "cg.h"
#include "useful.h"
#include "util.h"
#include "functors.h"




//using namespace std

//#include "fc.h"

// template<class T>
// size_t OpCounter<T>::cnt_mul;
// template<class T>
// size_t OpCounter<T>:: cnt_div;
// template<class T>
// size_t OpCounter<T>:: cnt_add;
// template<class T>
// size_t OpCounter<T>:: cnt_cmp;
// template<class T>
// size_t OpCounter<T>:: ocnt_mul;
// template<class T>
// size_t OpCounter<T>:: ocnt_div;
// template<class T>
// size_t OpCounter<T>:: ocnt_add;
// template<class T>
// size_t OpCounter<T>:: ocnt_cmp;

EXPORT
template<class Double>
void DoTheWork(int N1, int N2, int N3, SolutionMethod method, Double eps,
	       DiscretizationType distype, BoundaryConditions bc)
{
  int iterations;
  Double serr;
  NewWorkerScalar<Double, int> wrk(MPI::COMM_WORLD, N1, N2, N3);
  {
    TimeCalc1("!Total time");
  {
    TimeCalc1("System generation");
    wrk.Sysgen(distype, bc);
  }
//   typedef OpCounter<double> OP;
//   if(typeid(OP)==typeid(Double))
//     OP::Save();
  {
    TimeCalc1("!Solution time");
    iterations = wrk.Solve(eps, method);
    serr = wrk.CheckSolution();
  }
  {TimeCalc1("Output time");}
  }
  //  wrk.PrintSumary();
//   if(typeid(OP)==typeid(Double))
//   {
//     OP::Save();
//     size_t sum = 0;
//     sum += OP::ocnt_add;
//     sum += OP::ocnt_mul;
//     sum += OP::ocnt_div;
//     sum += OP::ocnt_cmp;
//     cout
//       << "add " << OP::ocnt_add << endl
//       << "mul " << OP::ocnt_mul << endl
//       << "div " << OP::ocnt_div << endl
//       << "cmp " << OP::ocnt_cmp << endl;
//     cout << sum / wrk.comptimesystemsolve /1000 /1000 << endl;
//   }

  if(MPI::COMM_WORLD.Get_rank()==0)
  {
    std::ofstream f(Options::GetOptions().outputname.c_str());
    int fcs = 30;
    int scs = 30;

    f << "Starting  on " 
      << "several? processors. Eps is "<< eps<< endl;
    
    f << setw(fcs) << left << "Problem size "
      << N1 << "x" << N2 << "x" << N3 << " " << Myf<Double>::ToString() << " "
      << wrk.algname <<endl;
    f << setw(fcs) << left  << "Running on "  
      << setw(scs) <<MPI::COMM_WORLD.Get_size()  << "procesors." << endl;
    f << setw(fcs) << "Iterations" 
      << setw(scs) << iterations << endl;
    ManageTimes::PrintAll(f);
    ManageTimes::PrintAll(f,iterations," pi");
    f <<left<< setw(30) << "Matrix error"
      << setw(30)<< 0 << "\n";
    f <<left<< setw(30) << "Solution error"
      << setw(30)<< serr << "\n";
  }

}

void PrintHelp()
{
  using namespace std;
  cout << "Usage: " << endl;
  cout << "\t parpcg\t[-n=#] [-n1=#] [-n2=#] [-n3=#] \n\t\t[-cg | -m1 | -m2 | -m3 ] [-eps=#] [-f] [-p] [-b] [-z=]" << endl;
  cout  << endl;
  cout << "\t\t-n\n\t\t-n1 \n\t\t-n2 \n\t\t-n3 \tDimencions. Default is 31" << endl;
  cout << "\t\t-cg \tConjugate Gradient method." << endl;
  cout << "\t\t-m1 \tPreconditioned CG with MIC(0)." << endl;
  cout << "\t\t-m2 \tPreconditioned CG with MIC(0) parallel on each line. Default." << endl;
  cout << "\t\t-m3 \tPreconditioned CG with MIC(0) parallel on each plane." << endl;
  cout << "\t\t-f \tUse single precision. Default is double" << endl;
  cout << "\t\t-eps\tThe precision. Default is 1e-3" << endl;
}

//typedef OpCounter<double> Double;

int main(int argc, char **argv)
{
  MPI::Init(argc, argv);
  using namespace std;
  int exitcode = 0;

  

  int N;
  int N1;
  int N2, N3;
  double eps;
  DiscretizationType distype = RP;
  BoundaryConditions bc = bcNone;//bcLeft|bcRight|bcFront|bcBack|bcUp|bcDown;

  SolutionMethod method = smPCG4;
  // parameter decoding
  N = GetOption("-n=", argc, argv, 31);
  N1 = GetOption("-n1=", argc, argv, N);
  N2 = GetOption("-n2=", argc, argv, N);
  N3 = GetOption("-n3=", argc, argv, N);
  double maxn = N1;
  if (maxn < N2)
    maxn = N2;
  if (maxn < N3)
    maxn = N3;
  Options::GetOptions().DM1 = N1/maxn;
  Options::GetOptions().DM2 = N2/maxn;
  Options::GetOptions().DM3 = N3/maxn;
  Options::GetOptions().variant = GetOption("-v=", argc, argv, 0);
  Options::GetOptions().pfactor = GetOption("-p=", argc, argv, 1.0);
  Options::GetOptions().ppfactor = GetOption("-pp=", argc, argv, 1.0);
  eps = GetOption("-eps=", argc, argv, 1.0e-3);
  Options::GetOptions().usebone = OptionPresent("-b", argc, argv);
  Options::GetOptions().zero = GetOption("-z=", argc, argv, 0.9);
  if (OptionPresent("-cg", argc, argv)) method = smCG;
  if (OptionPresent("-m1", argc, argv)) method = smPCG1;
  if (OptionPresent("-m2", argc, argv)) method = smPCG2;
  if (OptionPresent("-m3", argc, argv)) method = smPCG3;
  if (OptionPresent("-m4", argc, argv)) method = smPCG4;
  if (OptionPresent("-mv", argc, argv)) distype = MV;
  if (OptionPresent("-mp", argc, argv)) distype = MP;
  if (OptionPresent("-rp", argc, argv)) distype = RP;
  if (OptionPresent("-rv", argc, argv)) distype = RV;
  Options::GetOptions().SinglePrecPrecon = OptionPresent("-s", argc, argv);
  Options::GetOptions().Overlap = (OptionPresent("-nolap", argc, argv)) ? false :true;
  Options::GetOptions().E[0] = GetOption("-E=", argc,argv, Options::GetOptions().E[0]);
  using namespace std;
  string bound = GetStringOption("-bc=", argc, argv, "r");

  if(bound.find("l")!=string::npos)
    bc = (BoundaryConditions)(bc | bcLeft);
  if(bound.find("r")!=string::npos)
    bc = (BoundaryConditions)(bc | bcRight);
  if(bound.find("d")!=string::npos)
    bc = (BoundaryConditions)(bc | bcDown);
  if(bound.find("u")!=string::npos)
    bc = (BoundaryConditions)(bc | bcUp);
  if(bound.find("f")!=string::npos)
    bc = (BoundaryConditions)(bc | bcFront);
  if(bound.find("b")!=string::npos)
    bc = (BoundaryConditions)(bc | bcBack);
  std::string &namebase = Options::GetOptions().namebase;
  {
    std::stringstream a;
    a << MPI::COMM_WORLD.Get_size() << "-" 
      << MPI::COMM_WORLD.Get_rank() << "";
    namebase += a.str();
  }
  for(int i = 1; i < argc; ++i)
    namebase += argv[i];

  Options::GetOptions().outputname = GetOption("-oname=", argc, argv, "put." + (Options::GetOptions().namebase));

  int mhh= max(N1, N2);
  mhh = max(mhh,N3);
  
  const int BufSize =  mhh*mhh*8*4*2*2;
  std::vector<char> buf;
  buf.resize(BufSize);
  MPI_Buffer_attach(&buf[0], BufSize);
  
  if (OptionPresent("--help", argc, argv))
    {
      PrintHelp();
      return 0;
    }

  //  cout << "prec " << eps << endl;
  
  



  try{

//     if (OptionPresent("-c", argc, argv))
//       DoTheWork<OpCounter<double> >(N1, N2, N3, method, eps, distype, bc);
//     else
#ifdef _FLOAT_EXPAND
  if (OptionPresent("-f", argc, argv))
      DoTheWork<float> (N1, N2, N3, method, eps, distype, bc);
  else
#endif
      DoTheWork<double>(N1, N2, N3, method, eps, distype, bc);

  

  }catch(MPI::Exception &e)
  {
    cout << MPI::COMM_WORLD.Get_size() << " " 
	 << MPI::COMM_WORLD.Get_rank() << " \n" 
	 << "MPI Exception caught ... \n"
	 << e.Get_error_string() << endl;
    exitcode = 1;
  }catch(std::exception &e)
  {
    cout << MPI::COMM_WORLD.Get_size() << " " 
	 << MPI::COMM_WORLD.Get_rank() << " \n"
	 << "std::exception caught ... \n" 
	 << e.what() << endl;
    exitcode = 2;
  }catch(const char * &e)
  {
    cout << MPI::COMM_WORLD.Get_size() << " " 
	 << MPI::COMM_WORLD.Get_rank() << " \n"
	 << "string ... \n" 
	 << e<< endl;
    exitcode = 2;
  }catch(...)
  {
    cout << MPI::COMM_WORLD.Get_size() << " " 
	 << MPI::COMM_WORLD.Get_rank() << " \n"
	 << "Unknown exception caught.... \n" 
	 << "Rethrowing..." << endl;
    exitcode = 3; // probably will be other since we rethrow the except.
    throw;
  }
  MPI::Finalize();
//   OpCounter<double>::Save();
//   cout << "adds " << OpCounter<double>::ocnt_add << endl;
//   cout << "muls " << OpCounter<double>::ocnt_mul << endl;
//   cout << "divs " << OpCounter<double>::ocnt_div << endl;
//   cout << "cmps " << OpCounter<double>::ocnt_cmp << endl;
  return exitcode;
  

/*



  

  matrix_type mat(N1, N2, N3);

  bvector_type rhs(N1, N2, N3);
  bvector_type x(N1, N2, N3);



  
  ClearVector(x);

  cout << N1 << "x" << N2 << "x" << N3 << endl;  
  cout << "Pcg " <<   DoPCG(mat, x, rhs, 1e-10, ind) << endl;  
//  ClearVector(x);
*/
}
