#define MPICH_IGNORE_CXX_SEEK
#include "basic.h"
#include "mpi.h"
// IBM workaround
#ifdef _MY_CPP
#include "mympicxx.hpp"
#endif

#include<typeinfo>
#include <iostream>
#include <cstdlib>
#include <typeinfo>
#include <iomanip>
#include <cmath>
#include <vector>
#include <string>
#include <sstream>
#include <stdexcept>




#include "elastgen.h"
#include "util.h"
#include "useful.h"
#include "sparcemmat.h"
#include "elastsolver.h"

template<class Double>
void DoTheWork(int N1, int N2, int N3,
	       SolutionMethod method,
	       DiscretizationType distype,
	       BoundaryConditions bc,
	       BoundaryConditions nc,
	       Double eps
	       )
{
  ElastWorker<Double, int> worker(MPI::COMM_WORLD, N1, N2, N3);
  int iterations;
  {
    TimeCalc1("!Total time");
    TimeCalc1p("System generation",2);
    worker.Sysgen(distype, bc, nc);
  }
  {
    TimeCalc1("!Total time");
    iterations=worker.Solve(eps, method);
    worker.CheckSolution();
    TimeCalc1p("Solution output",2);
    if (Options::GetOptions().SaveSolution)
      worker.WriteSolution();
  }
  //  ManageTimes::Get("!Total solve") = ManageTimes::Get("!Solution time")
  //    + ManageTimes::Get("PreconConstruct");
  ManageTimes::Reduce();
  if(MPI::COMM_WORLD.Get_rank()==0)
  {
    std::ofstream f(Options::GetOptions().outputname.c_str());
    using namespace std;
    int fcs = 30;
    int scs = 30;
    f << "Starting  on " 
      << "several? processors. Eps is "<< eps<< endl;
    
    f << setw(fcs) << left << "Problem size "
      << N1 << "x" << N2 << "x" << N3 << " " << Myf<Double>::ToString() << " "
      << worker.algname <<endl;
    f << setw(fcs) << left  << "Running on "  
      << setw(scs) <<MPI::COMM_WORLD.Get_size()  << "procesors." << endl;
    f << setw(fcs) << "Iterations" 
      << setw(scs) << iterations << endl;
    ManageTimes::PrintAll(f,1);
    ManageTimes::PrintAll(f,iterations," pi");
    f <<left<< setw(30) << "Matrix error"
      << setw(30)<< worker.merr << "\n";
    f <<left<< setw(30) << "Solution error"
      << setw(30)<< worker.serr << "\n";
    
  }
  
}

//using namespace std


void PrintHelp()
{
  using namespace std;
  cout << "\t pelast\t[-n=#] ...... [-n1=#] [-n2=#] [-n3=#] \n\t\t[-cg | -m1 | -m2 | -m3 ] [-eps=#] [-f]" << endl;
  cout  << endl;

}

//typedef OpCounter<double> Double;

int main(int argc, char **argv)
{
  MPI::Init(argc, argv);
  int exitcode = 0;
  {
    Myf<SmallVector<double,3> > double3;
    Myf<SmallVector<double,3> >::Geta() = &double3;
    Myf<SmallVector<float ,3> > float3;
    Myf<SmallVector<float,3> >::Geta() = &float3;
    

  using namespace std;

  int N;
  int N1;
  int N2, N3;
  double eps;
  //  double zero;
  DiscretizationType distype = MP;
  BoundaryConditions bc = bcNone;//bcLeft|bcRight|bcFront|bcBack|bcUp|bcDown;
  BoundaryConditions nc = bcNone;//bcLeft|bcRight|bcFront|bcBack|bcUp|bcDown;

  SolutionMethod method = smPCG4;
  // parameter decoding
  N = GetOption("-n=", argc, argv, 31);
  N1 = GetOption("-n1=", argc, argv, N);
  N2 = GetOption("-n2=", argc, argv, N);
  N3 = GetOption("-n3=", argc, argv, N);

  Options::GetOptions().debug = OptionPresent("-d", argc, argv);
  Options::GetOptions().variant = GetOption("-v=", argc, argv, 0);
  Options::GetOptions().DM1 = GetOption("-dm1=", argc, argv, 1.);
  Options::GetOptions().DM2 = GetOption("-dm2=", argc, argv, 1.);
  Options::GetOptions().DM3 = GetOption("-dm3=", argc, argv, 1.);
  std::string &namebase = Options::GetOptions().namebase;
  {
    std::stringstream a;
    a << MPI::COMM_WORLD.Get_size() << "-" 
      << MPI::COMM_WORLD.Get_rank() << "";
    namebase += a.str();
  }
  for(int i = 1; i < argc; ++i)
    namebase += argv[i];

  Options::GetOptions().outputname = GetOption("-oname=", argc, argv, "put." + (Options::GetOptions().namebase));
  
  double maxn = N1;
  if (maxn < N2)
    maxn = N2;
  if (maxn < N3)
    maxn = N3;
  //  DM1 = N1/maxn;
  //  DM2 = N2/maxn;
  //  DM3 = N3/maxn;
  Options::GetOptions().pfactor = GetOption("-p=", argc, argv, 1.0);
  Options::GetOptions().ppfactor = GetOption("-pp=", argc, argv, 1.0);
  Options::GetOptions().thres = GetOption("-thres=", argc, argv, .5);
  Options::GetOptions().agr = GetOption("-agr=", argc, argv, 0);
  Options::GetOptions().mit = GetOption("-mit=", argc, argv, 1);
  Options::GetOptions().ct = GetOption("-ct=", argc, argv, 1);
  Options::GetOptions().rwt = GetOption("-rwt=", argc, argv, 1.);
  eps = GetOption("-eps=", argc, argv, 1.0e-6);
  Options::GetOptions().usebone = OptionPresent("-b", argc, argv);
  Options::GetOptions().zero = GetOption("-z=", argc, argv, 0.0);
  Options::GetOptions().E[0] = GetOption("-E=", argc,argv, Options::GetOptions().E[0]);
  Options::GetOptions().ni[0] = GetOption("-ni=", argc,argv, Options::GetOptions().ni[0] );
  Options::GetOptions().E[1] = GetOption("-E1=", argc,argv, Options::GetOptions().E[0]);
  Options::GetOptions().ni[1] = GetOption("-ni1=", argc,argv, Options::GetOptions().ni[0] );
  Options::GetOptions().E[255] = GetOption("-E255=", argc,argv, Options::GetOptions().E[1]);
  Options::GetOptions().ni[255] = GetOption("-ni255=", argc,argv, Options::GetOptions().ni[1] );
  Options::GetOptions().force = GetOption("-f=", argc,argv, Options::GetOptions().force );
  Options::GetOptions().SinglePrecPrecon = OptionPresent("-s", argc, argv);
  Options::GetOptions().scratchpath = GetOption("-sp=", argc, argv, std::string("/scratch/hpebgue2"));
  if (OptionPresent("-cg", argc, argv)) method = smCG;
  if (OptionPresent("-save", argc, argv)) Options::GetOptions().SaveSolution = true;
  if (OptionPresent("-m1", argc, argv)) method = smPCG1;
  if (OptionPresent("-m2", argc, argv)) method = smPCG2;
  if (OptionPresent("-m3", argc, argv)) method = smPCG3;
  if (OptionPresent("-m4", argc, argv)) method = smPCG4;
  if (OptionPresent("-hd", argc, argv)) method = smBoomerDD;
  if (OptionPresent("-hb", argc, argv)) method = smBoomer;
  
  if (OptionPresent("-mv", argc, argv)) distype = MV;
  if (OptionPresent("-mp", argc, argv)) distype = MP;
  //  if (OptionPresent("-rp", argc, argv)) distype = RP;
  //  if (OptionPresent("-rv", argc, argv)) distype = RV;
  Options::GetOptions().Overlap = (OptionPresent("-nolap", argc, argv)) ? false :true;
  using namespace std;
  string bound = GetStringOption("-bc=", argc, argv, "d");

  if((method==smPCG4)&&(MPI::COMM_WORLD.Get_size()*3 > N3))
  {
    if (MPI::COMM_WORLD.Get_rank()==0)
      std::cerr << "switching to method 2" <<std::endl;
    method=smPCG2;
  }
  if((MPI::COMM_WORLD.Get_size() > N3+1))
  {
    if (MPI::COMM_WORLD.Get_rank()==0)
      std::cerr << "Too many processors. Quitting.." <<std::endl;
    return 1;
  }
    
  
  if(bound.find("l")!=string::npos)
    bc = (BoundaryConditions)(bc | bcLeft);
  if(bound.find("r")!=string::npos)
    bc = (BoundaryConditions)(bc | bcRight);
  if(bound.find("d")!=string::npos)
    bc = (BoundaryConditions)(bc | bcDown);
  if(bound.find("u")!=string::npos)
    bc = (BoundaryConditions)(bc | bcUp);
  if(bound.find("f")!=string::npos)
    bc = (BoundaryConditions)(bc | bcFront);
  if(bound.find("b")!=string::npos)
    bc = (BoundaryConditions)(bc | bcBack);
  bound = GetStringOption("-nc=", argc, argv, "u");

  
  if(bound.find("l")!=string::npos)
    nc = (BoundaryConditions)(nc | bcLeft);
  if(bound.find("r")!=string::npos)
    nc = (BoundaryConditions)(nc | bcRight);
  if(bound.find("d")!=string::npos)
    nc = (BoundaryConditions)(nc | bcDown);
  if(bound.find("u")!=string::npos)
    nc = (BoundaryConditions)(nc | bcUp);
  if(bound.find("f")!=string::npos)
    nc = (BoundaryConditions)(nc | bcFront);
  if(bound.find("b")!=string::npos)
    nc = (BoundaryConditions)(nc | bcBack);

  if(bc == bcNone)
    bc = bcRight;
  int mhh= max(N1, N2);
  mhh = max(mhh,N3);
  
  //  const int BufSize =  mhh*mhh*8*4*2*2;
  //  std::vector<char> buf;
  //  buf.resize(BufSize);
  //  MPI_Buffer_attach(&buf[0], BufSize);
  
  if (OptionPresent("--help", argc, argv))
    {
      PrintHelp();
      return 0;
    }



  try{

//     if (OptionPresent("-c", argc, argv))
//       DoTheWork<OpCounter<double> >(N1, N2, N3, method, eps, distype, bc, zero, usebone);
//     else
#ifdef _FLOAT_EXPAND
  if (OptionPresent("-f", argc, argv))
    DoTheWork<float> (N1, N2, N3, method, distype, bc, nc, eps);
  else
#endif
    DoTheWork<double>(N1, N2, N3, method, distype, bc, nc, eps);
  }catch(MPI::Exception &e)
  {
    cout << MPI::COMM_WORLD.Get_size() << " " 
	 << MPI::COMM_WORLD.Get_rank() << " \n" 
	 << "MPI Exception caught ... \n"
	 << e.Get_error_string() << endl;
    exitcode = 1;
  }catch(std::exception &e)
  {
    cout << MPI::COMM_WORLD.Get_size() << " " 
	 << MPI::COMM_WORLD.Get_rank() << " \n"
	 << "std::exception caught ... \n" 
	 << e.what() << endl;
    exitcode = 2;
  }
  catch(const std::string &e)
  {
    cout << MPI::COMM_WORLD.Get_size() << " " 
	 << MPI::COMM_WORLD.Get_rank() << " \n"
	 << "string exception cought ... \n" 
	 << e << endl;
    exitcode = 2;
  }
  catch(...)
  {
    cout << MPI::COMM_WORLD.Get_size() << " " 
	 << MPI::COMM_WORLD.Get_rank() << " \n"
	 << "Unknown exception caught.... \n" 
	 << "Rethrowing..." << endl;
    exitcode = 3; // probably will be other since we rethrow the except.
    throw;
  }


}
  MPI::Finalize();
//   OpCounter<double>::Save();
//   cout << "adds " << OpCounter<double>::ocnt_add << endl;
//   cout << "muls " << OpCounter<double>::ocnt_mul << endl;
//   cout << "divs " << OpCounter<double>::ocnt_div << endl;
//   cout << "cmps " << OpCounter<double>::ocnt_cmp << endl;
  return exitcode;
}
