// Poly.h
//

#include <iostream>
#include <vector>
#include <cmath>
#include <ctime>
#include <fstream>



using namespace std;


template<typename t>
class PolinomTemplate{
	vector<t> Coeff;
	PolinomTemplate(){}
public:
	const int step;
	PolinomTemplate(int n):Coeff(n+1), step(n)
	{
	}
	t& operator[](int i)
	{
		return Coeff[i];
	}
	t operator()(t x)
	{
		// direct calculation
		t xx = 1;
		t r = 0;
		for(int i = 0; i <=step;i++)
		{
			r+=xx*Coeff[i];
			xx*=x;
		}
		return r;
/*		// Horner
		t r(Coeff[step]);
		for (int i = step - 1; i>=0; i--)
		{
			r = r*x + Coeff[i];
		}
		return r;
*/
	}
};

template<typename t>
ostream& operator<<(ostream& out, PolinomTemplate<t>& p)
{
	for (int i = p.step; i>0; i--)
		out << p[i] << "*x^" << i << " + ";
	out << p[0];
	return out;
}


template<typename t>
ostream& operator<<( ostream& out, vector<t>& p)
{
	for (int i = 0; i< p.size(); i++)
		out << p[i] << "\n";

	return out;
}

template<class t>
t sqr(t x)
{
	return x*x;
}

template<typename T>
vector<T> CalcDelta(PolinomTemplate<T> &F, vector<T> &X, int k)
{
	int step = F.step;
	vector<T> d(step);
	vector<T> diff(step);
	vector<T> func(step);
	int i;
	int s, j, r;

	for(i = 0; i< step; i++)
	{
		d[i] = T(0);
		func[i] = F(X[i]);
	}
	const T one = 1;
	T rr;
	for(r = 0; r<k; r++)
	{
		for (i = 0; i < step; i++)
			diff[i] = X[i] + d[i];
		for(j = 0; j < step; j++)
		{
			rr = one;
			for(s = 0; s < j; s++)
				rr*=X[j] - diff[s];
			for(s = j+1; s < step; s++)
				rr*=X[j] - diff[s];
			d[j] = -func[j]/rr;
		}
	}
	return d;
}


// endcondition true = |xk+1 - xk| <  eps. false F(xk+1) < eps for ever i
template<typename T, typename REAL_TYPE>
vector<T> FindRoots(PolinomTemplate<T> &F, vector<T> &X, int R, REAL_TYPE eps, int &res, bool endcondition = true)
{
	int step = F.step;
	vector<T> d1(step);
	vector<T> d2(step);
	d1 = X;
	vector<T> *d[2] = {&d1, &d2};
	int b = 0;
	bool be;
	do{
		be = false;
		const vector<T> &delta = CalcDelta(F, *d[b], R + 1);
		for (int i = 0; i < step; i++)
		{
			(*d[1-b])[i] = (*d[b])[i] + delta[i];
			REAL_TYPE hm;
			if (endcondition)
			{
				hm = fabs(((*d[1-b])[i] - (*d[b])[i]));
			}else
			{
				T t = F((*d[1-b])[i]);
				hm = fabs(t);
			}
			if (hm >= eps) be = true;
		}
		b = 1-b;
		res++;
	}while(be && (res!=0));
	return *d[b];
}


