#ifndef __precon_h__
#define __precon_h__
#include "basic.h"
#include "solver.h"
#include "functors.h"
EXPORT
template<class Vector>
class NoPrecon:public PreconBase<Vector>
{
 public:
  
  virtual void PreconSolve(Vector &res, Vector &src)const;
  virtual std::string GetName()const
  {
    return "NoPrecon";
  }
};
EXPORT
template<class Vector, class SinglePreconditioner>
class SinglePrecon:public PreconBase<Vector>
{
  typedef typename SinglePreconditioner::vector_type SVector;
  mutable  SVector sres, ssrc;
  SinglePreconditioner *sp;
 public:

  SinglePrecon(SinglePreconditioner *sp_):sp(sp_){}
  
  virtual void PreconSolve(Vector &res, Vector &src)const
  {
    static bool firsttime = true;
    if(firsttime)
      {
	sres.SetSize(res.N1, res.N2, res.N3, res.N3R);
	ssrc.SetSize(res.N1, res.N2, res.N3, res.N3R);
	sres.SetComm(res.comm);
	ssrc.SetComm(src.comm);
	ssrc.CreateDatatypes();
	sres.CreateDatatypes();
	firsttime = false;
      }
    VecVecOper<VecCpy>(ssrc, src);
    sp->PreconSolve(sres, ssrc);
    VecVecOper<VecCpy>(res, sres);
  };
  virtual std::string GetName()const
  {
    return "Single" + sp->GetName();
  }
  ~SinglePrecon(){delete sp;}
};

EXPORT
template<class Matrix, class Vector, class PreconVector, class PreconMethod>
class MIC0Precon:public PreconBase<Vector>
{
protected:
  PreconVector P;
  Matrix *Mat;
 public:
  virtual void Create(Matrix &mat);
  virtual void PreconSolve(Vector &res, Vector &src)const;
  virtual std::string GetName()const
  {
    return PreconMethod::name();
  }
};

EXPORT
template<class Matrix, class Vector, class PreconVector, class PreconMethod>
class MIC0PreconM4:public MIC0Precon<Matrix, Vector, PreconVector, PreconMethod>
{
 public:
  virtual void Create(Matrix &mat);
  virtual void PreconSolve(Vector &res, Vector &src)const;
};


#endif


