#ifndef __precon_hpp__
#define __precon_hpp__

#include "precon.h"
#include "index.h"
#include "visitors.h"
#include "functors.h"
#include "useful.h"
template<class Vector>
void
NoPrecon<Vector>::
PreconSolve(Vector &res, Vector &src)const
{
  VecVecOper<VecCpy>(res, src);   // simple preconditioning
}




template<class Matrix, class Vector, class PreconVector, class PreconMethod>
void
MIC0Precon<Matrix, Vector, PreconVector, PreconMethod>::
Create(Matrix &mat)
{
  TimeCalc1("PreconditionerCreation");
  Mat = &mat;
  P.SetSize(mat.N1, mat.N2, mat.N3, mat.N3R);
  P.SetComm(mat.comm);
  P.CreateDatatypes();
  PreconCreatorStep1<Matrix, PreconVector, PreconMethod> st1(mat, P);
  PreconCreatorStep1_5<Matrix, PreconVector, PreconMethod> st1_5(mat, P);
  PreconCreatorStep2<Matrix, PreconVector, PreconMethod> st2(mat, P);
  Index3dBase<typename Vector::int_type>
    ind(P.N1, P.N2, P.N3);

  Execute1(ind, st1, P.size, P.rank);
  Execute1(ind, st1_5, P.size, P.rank);
  P.ClearOthers();
  Execute1(ind, st2, P.size, P.rank);
  
}



template<class Matrix, class Vector, class PreconVector, class PreconMethod>
void
MIC0Precon<Matrix, Vector, PreconVector, PreconMethod>::
PreconSolve(Vector &res, Vector &src)const
{
  TimeCalc1("PreconSolve");
  std::string commtimepreconditioner = "PreconSolveComm";
  Index3dBase<typename Vector::int_type>
    ind(P.N1, P.N2, P.N3);
  PreconSolverStep1_1_5<Matrix, PreconVector, Vector, Vector, PreconMethod> 
    pcs115(*Mat, P, res, src, commtimepreconditioner);
  PreconSolverStep2<Matrix, PreconVector, Vector, Vector, PreconMethod> 
    pcs2(*Mat, P, res, src, commtimepreconditioner);

  pcs115.Begin();
  Execute1(ind, pcs115, P.size, P.rank);
  res.ClearOthers();
  Execute1(ind, pcs2, P.size, P.rank);
}


template<class Matrix, class Vector, class PreconVector, class PreconMethod>
void
MIC0PreconM4<Matrix, Vector, PreconVector, PreconMethod>::
Create(Matrix &mat)
{
  MIC0Precon<Matrix, Vector, PreconVector, PreconMethod>::Create(mat);
  this->P.StartSendRecv();
  this->P.FinishSendRecv();
}

template<class Matrix, class Vector, class PreconVector, class PreconMethod>
void
MIC0PreconM4<Matrix, Vector, PreconVector, PreconMethod>::
PreconSolve(Vector &res, Vector &src)const
{
  TimeCalc1("PreconSolve");
  std::string commtimepreconditioner = "PreconSolveComm";
  Index3dBase<typename Vector::int_type>
    ind(this->P.N1, this->P.N2, this->P.N3);
  PreconSolverStep1_1_5<Matrix, PreconVector, Vector, Vector, PreconMethod> 
    pcs115(*this->Mat, this->P, res, src, commtimepreconditioner);
  PreconSolverStep2<Matrix, PreconVector, Vector, Vector, PreconMethod> 
    pcs2(*this->Mat, this->P, res, src, commtimepreconditioner);
  src.StartSendRecv();
  src.FinishSendRecv();
  pcs115.Begin();
  Execute1(ind, pcs115, this->P.size, this->P.rank);
  res.ClearOthers1();
  Execute1(ind, pcs2, this->P.size, this->P.rank);
  //  TimeCalc2("PreconSolveComm", commtimepreconditioner);

}



#endif

