#ifndef __solver_h__
#define __solver_h__
#include "basic.h"
#include "mpi.h"
#include <string>
#include "useful.h"

enum SolutionMethod{smCG, smPCG1, smPCG2, smPCG3, smPCG4, smBoomer, smBoomerDD};


template<class Vector>
class PCGSolverBase;

EXPORT
template<class Vector>
class MatrixBase
{
 public:
  virtual void Multiply(Vector &res, Vector &src)const = 0;
  virtual ~MatrixBase(){};
  virtual std::string GetName()const = 0;
};
EXPORT
template<class Vector>
class PreconBase
{
 public:
  typedef Vector vector_type;
  virtual void PreconSolve(Vector &res, Vector &src)const = 0;
  virtual ~PreconBase(){};
  virtual std::string GetName()const = 0;
};

EXPORT
template<class Vector>
class PCGSolverBase
{
  MatrixBase<Vector> *mat;
  PreconBase<Vector> *precon;
public:
  typedef Vector vector_type;
  typedef typename Myf<typename Vector::base_type>::base_type Double;
  typedef typename Myf<typename Vector::base_type>::type element_type;
  typedef typename Vector::Int Int;

private:
  Double eps;
 public:


  void SetMatrix(MatrixBase<Vector> *m){mat = m;}
  void SetPrecon(PreconBase<Vector> *p){precon = p;}
  
  vector_type Hd, h, D, G;
  PCGSolverBase();
  void SetEps(Double eps_);
  void SetSizes(Int N1_, Int N2_, Int N3_, Int N3R_);
  void SetComm(MPI::Intracomm &comm_);

  //operations Better to be global functions associated with the vectors themselves
  virtual unsigned Solve(Vector &X, Vector const &RHS);
  virtual std::string GetName()const
    {
      return "PCG" + mat->GetName() + precon->GetName();
    }
  virtual ~PCGSolverBase(){}
  
};



#endif
