#ifndef __solver_hpp__
#define __solver_hpp__
#include "solver.h"
#include "functors.h"
#include <iostream>

template<class Vector>
PCGSolverBase<Vector>::
PCGSolverBase()
{
  eps=1e-3;
}

template<class Vector>
void
PCGSolverBase<Vector>::
SetSizes(Int N1_, Int N2_, Int N3_, Int N3R_)
{
  Hd.SetSize(N1_, N2_, N3_, N3R_);
  h.SetSize(N1_, N2_, N3_, N3R_);
  D.SetSize(N1_, N2_, N3_, N3R_);
  G.SetSize(N1_, N2_, N3_, N3R_);  
}

template<class Vector>
void
PCGSolverBase<Vector>::
SetComm(MPI::Intracomm &comm_)
{
  Hd.SetComm(comm_);
  h.SetComm(comm_);
  D.SetComm(comm_);
  G.SetComm(comm_);
  h.CreateDatatypes();
  Hd.CreateDatatypes();
  h.CreateDatatypes();
  D.CreateDatatypes();
  G.CreateDatatypes();
}

template<class Vector>
void
PCGSolverBase<Vector>::
SetEps(Double eps_)
{
  eps=eps_;
}
  

template<class Vector>
unsigned
PCGSolverBase<Vector>::
Solve(Vector &X, Vector const &RHS)
{
  using namespace std;
  Double Eps = eps*eps;
  Double gtgfull, gtg, gtgold;
  Double tau, beta, temp;
  unsigned iter = 0;
  //  int rank = D.rank;
  VecVecOper<VecCpy>(D, X);        // D = X;
  mat->Multiply(Hd, D);            // Hd = H d;

  VecVecOper<VecCpy>(G, Hd);       // G = Hd;
  VecVecOper<VecSub>(G, RHS);      // G -= RHS;

  precon->PreconSolve(h, G);        // h = C^{-1} G

  ClearVector(D);                  // D = 0;
  VecVecOper<VecSub>(D, h);        // D = -h;

  gtgfull = gtg =
    VecVecReduce<VecDot>(G, h);    //gtg = G^t h

  gtgfull *= Eps;
  double oldprn = 1;
  while(gtg > gtgfull)
  {
    if(h.rank==0)
      //    std::cerr << iter << std::endl;
    if (iter >10100)
      break;
    double prn = sqrt(gtg/(gtgfull/Eps));
    if (Options::GetOptions().debug)
      if (MPI::COMM_WORLD.Get_rank()==0)
	cerr << iter << "\t" <<prn << "\t"<< oldprn/prn << endl;
    oldprn = prn;
    gtgold = gtg;

    mat->Multiply(Hd, D);            // Hd = H d;

    temp = VecVecReduce<VecDot>(D, Hd); //temp = D^t Hd

    tau = gtg/temp;

    VecVecOper<VecUpdate>(X, D, tau);  // X += D*tau;
    VecVecOper<VecUpdate>(G, Hd, tau); // G += Hd*tau;
    precon->PreconSolve(h, G);         // h = C^-1 G;

    gtg = VecVecReduce<VecDot>(G, h); // gtg = G^t h
    //    std::cout << iter << " " << gtg<< std::endl;
    beta = gtg/gtgold;

    VecVecOper<VecUpdate2>(D, h, beta);//

    ++iter;    
  }
  return iter;
}


#endif
