#ifndef __sysgen_h__
#define __sysgen_h__
#include "basic.h"
#include "small.h" 
#include "ntransform.h"
#include "mirorer.h"
extern vector3d<double> coefficientvec;

enum DiscretizationType{MV, MP, RV, RP};
enum BoundaryConditions{bcNone  = 0x00,
			bcLeft  = 0x01,
			bcRight = 0x02,
			bcDown  = 0x04,
			bcUp    = 0x08, 
			bcFront = 0x10,
			bcBack  = 0x20,
			bcNeuman= 0x40};
enum {BLeft, BRight, BFront, BBack,  BDown, BUp, BSize};

EXPORT
template<class Stream>
Stream&operator << (Stream &out, const BoundaryConditions&bc);

EXPORT
template<class T>
SmallMatrix<T, 3, 3>GenOne();

EXPORT
template<class T>
SmallMatrix<T, 3, 3>&One();

EXPORT
template<class T>
SmallMatrix<T, 3, 3>&A(const SmallVector<SmallVector<T, 3>, 8> &points);

EXPORT
template<class T>
T Power(T a, int n);
EXPORT
template<class T>
T F(const SmallVector<T, 3> &p);
EXPORT
template<class T>
T U(const SmallVector<T, 3> &p);
EXPORT
template<class Double>
class Discretizator
{
 protected:
  typedef IntegralCube<Double, 3> Integral;
  Integral integral;
  typedef IntegralSquare<Double, 2> Integral2d;
  Integral2d integral2d;


public:
  typedef SmallMatrix<Double, 6, 6> ElementMatrix_type;
  typedef SmallVector<Double, 6>  ElementRhs_type;

  SmallVector<SmallVector<Double, 3>, BSize> normals;
  SmallVector<SmallVector<SmallVector<Double, Integral2d::PointsNumber>, BSize>, 6> disPsiBoundary;
  SmallVector<SmallVector<SmallVector<Double, 3>, Integral2d::PointsNumber>, BSize> pointsOnBoundary;

typedef My::Base::MultiPoly<Double, 3> Poly3d;
  SmallVector<Poly3d, 6> ShapeFun;
  SmallMatrix<Poly3d, 3, 6> DPsiDt;

  std::vector<SmallMatrix<Double, 3, 6> > disDPsiDt;
  std::vector<SmallMatrix<Double, 1, 6> > disShapeFun;
  std::vector<Double> disJdet;
  std::vector<SmallMatrix<Double, 3, 3> >disJrev;
  std::vector<SmallMatrix<Double, 3, 6> > disJrevDPsiDt;

  TrilinearTransform<Double> tri;

  void GenDPsiDt();
  
  void GenShapeFunMP();
  void GenShapeFunMV();
  DiscretizationType distype;
  Discretizator(DiscretizationType t):distype(t)
  {
    if(t == MV) 
      GenShapeFunMV();
    else 
      GenShapeFunMP();
    GenDPsiDt();
  }
EXPORT
template<class Matrix, class Vector, class Int>
  void BoundCondSet(Matrix &mat, Vector &vec, Int I, Int J, Int K, Int MI, Int MJ, Int MK, BoundaryConditions bc);
EXPORT
 template<class Matrix, class Vector>
   void GenerateMatrix(Matrix &mat, Vector &rhs, 
		       typename Matrix::int_type NumberOfLines,
		       typename Matrix::int_type  StartLine,
		       typename Matrix::int_type LO, 
		       typename Matrix::int_type RealN3, 
		       BoundaryConditions bc );

};




#endif
