#ifndef __sysgen_hpp__
#define __sysgen_hpp__

#include "sysgen.h"

#include "useful.h"
#include "index.h"
#include "multipoly.h"
#include "vectors.h"

#include "mirorer.h"

#include "math.h"

template<class Stream>
Stream&operator << (Stream &out, const BoundaryConditions&bc)
{
  if (bc == bcNone)
    out << "None ";
  if (bc & bcLeft)
    out << "left ";
  if (bc & bcRight)
    out << "right ";
  if (bc & bcDown)
    out << "down ";
  if (bc & bcUp)
    out << "up ";
  if (bc & bcFront)
    out << "front ";
  if (bc & bcBack)
    out << "back ";
}

template<class T>
SmallMatrix<T, 3, 3>GenOne()
{
  SmallMatrix<T, 3, 3> a;
  for(int i = 0; i < 3; ++i)
    {
      a[i] = T(0);
      a[i][i] = T(1);
    }
  //  a[0][0] = 0;
  //  a[2][2] = 0;
  return a;
}

template<class T>
SmallMatrix<T, 3, 3>&One()
{
  static SmallMatrix<T, 3, 3> a = GenOne<T>();
  return a;
}
template<class T>
SmallMatrix<T, 3, 3>&A(const SmallVector<T, 3> &pt)
{
  static SmallMatrix<T, 3, 3> a;
  a = One<T>();
  if (Options::GetOptions().variant & 0x10)
  {
    T const E = Options::GetOptions().E[0];
    T const &X = pt[0];
    T const &Y = pt[1];
    T const &Z = pt[2];
#ifndef __MYPI__
    T const M_2PI = atan(1.)*8.;
#endif
    a[0][0] = (1.+E*exp(X+Y+Z));
    a[1][1] = (1.+E*.5*sin(M_2PI*(X+Y+Z)));
    a[2][2] = (1.+E*.5*sin(M_2PI*(X+Y+Z)));
  }else
  {
    T f = coefficientvec.GetAt(pt[0], pt[1], pt[2]); 
  //  std::cout << f << std::endl;
    a *= f;
  //  std::cout << f << endl;
  }
  using namespace std;
  //  cout << a[0][0] << " " << a[1][1] << " " << a[2][2] << endl;
  return a;
  
}
template<class T>
SmallMatrix<T, 3, 3>&A(const SmallVector<SmallVector<T, 3>, 8> &points)
{
  typedef IntegralCube<T, 1> Integral;
  static Integral icube;
  static const int npoints =  Integral::PointsNumber;
  
  if (Options::GetOptions().variant & 0x10)
  {
    SmallVector<T, 3> h, one;
    one = 1;
    h[0] = points[1][0] - points[0][0];
    h[1] = points[2][1] - points[0][1];
    h[2] = points[4][2] - points[0][2];
    static SmallMatrix<T, 3, 3> res;
    res = 0;
    for(int f = 0; f < npoints; ++f)
    {
      SmallVector<T, 3> pt = icube.points[f];
      pt += one;
      pt *= 0.5;
      pt *= h;
      pt += points[0];
      SmallMatrix<T, 3, 3> a = A(pt);
      a *= icube.coeff[f];
      res += a;
    }
    res /=8.;
    return res;
  }
  SmallVector<T, 3> pt;
  pt[0] = pt[1] = pt[2] = T(0);
  for(int i = 0; i < 8; ++i)
    pt += points[i];
  pt/=T(8);
  SmallMatrix<T, 3, 3> &a = A(pt);
  return a;
  
}
template<class T>
T Power(T a, int n)
{
  T res = 1.;
  for( int i = 0; i < n; ++i)
    res*=a;
  return res;
}
template<class T>
T F(const SmallVector<T, 3> &p)
{
  
  T const &x = p[0];
  T const &y = p[1];
  T const &z = p[2];
  //  return 1;
  //  return x*x*x + y*y*y + z*z*z + exp(x)+exp(y)+exp(z);
  //  return - (2*((-1 + x1)*x1*(-1 + x2)*x2 + (-1 + x1)*x1*(-1 + x3)*x3 + (-1 + x2)*x2*(-1 + x3)*x3));
  if (Options::GetOptions().variant & 0x10)
  {
    T const E = Options::GetOptions().E[0];
    T const &X = p[0];
    T const &Y = p[1];
    T const &Z = p[2];
#ifndef __MYPI__
    T const M_2PI = atan(1.)*8.;
#endif
    return (-M_2PI*(E*M_2PI*0.5*cos(M_2PI*Z)*cos(M_2PI*(X+Y+Z))*sin(M_2PI*X)*sin(M_2PI*Y)+sin(M_2PI*Z)*(E*M_2PI*0.5*cos(M_2PI*Y)*cos(M_2PI*(X+Y+Z))*sin(M_2PI*X)+sin(M_2PI*Y)*(exp(X+Y+Z)*E*cos(M_2PI*X)-3*M_2PI*sin(M_2PI*X)-exp(X+Y+Z)*E*M_2PI*sin(M_2PI*X)-E*M_2PI*sin(M_2PI*X)*sin(M_2PI*(X+Y+Z))))));
  }
    
  return x*x*x + y*y*y + z*z*z;
  return -(Power(y,2)*(-3. + 2.*y)*
      Power(z,2)*(-3. + 2.*z) + 
     3.*(-1. + x)*(1. + x)*
      (-1. + 2.*y)*Power(z,2)*
      (-3. + 2.*z) + 
     3.*(-1. + x)*(1. + x)*
      Power(y,2)*(-3. + 2.*y)*
	  (-1. + 2.*z))/18.;

}
template<class T>
T U(const SmallVector<T, 3> &p)
{

  T const &x = p[0];
  T const &y = p[1];
  T const &z = p[2];

  if (Options::GetOptions().variant & 0x10)
  {
    T const twopi = atan(T(1))*8;
    return sin(twopi * x) * sin(twopi*y) * sin(twopi*z);
  }
  //  return  (-1 + x1)*x1*(-1 + x2)*x2*(-1 + x3)*x3;
  //  return (x1 -1)*(x1-1) +2*x1 -2 ;
  // return (-1+x)*(1+x)*(-y*y/2 + y*y*y/3)*(z*z/2 + z*z*z/3);
  return (-1. + x)*(1. + x)*
   (-Power(y,2)/2. + 
     Power(y,3)/3.)*
   (-Power(z,2)/2. + 
    Power(z,3)/3.);
}



template<class Double>
void Discretizator<Double>::
GenDPsiDt()
{
  for(int i = 0; i < 3;++i)
    for(int j = 0; j < 6; ++j)
      DPsiDt[i][j] = ShapeFun[j].D(i);
  disDPsiDt.resize(Integral::PointsNumber);
  disShapeFun.resize(Integral::PointsNumber);
  disJdet.resize(Integral::PointsNumber);
  disJrev.resize(Integral::PointsNumber);
  disJrevDPsiDt.resize(Integral::PointsNumber);
  for(int k = 0; k < Integral::PointsNumber; ++k)
  {
    for(int i = 0; i < 3; ++i)
      for(int j = 0;j <6; ++j)
	disDPsiDt[k][i][j] = DPsiDt[i][j](integral.points[k]);
    for(int j = 0; j < 6; ++j)
      disShapeFun[k][0][j] = ShapeFun[j](integral.points[k]);
  }

  // generating normals
  for(int i = 0; i < BSize; ++i)
    normals[i] = Double();

  normals[BLeft][0] = -1;
  normals[BRight][0] = 1;
  normals[BFront][1] = -1;
  normals[BBack][1] = 1;
  normals[BDown][2] = -1;
  normals[BUp][2] = 1;

  // generating pointsOnBoundary

  for(int i = 0; i < Integral2d::PointsNumber; ++i)
  {
    SmallVector<Double, 2> &pt2d = integral2d.points[i];

    pointsOnBoundary[BLeft][i][0] = -1;
    pointsOnBoundary[BLeft][i][1] = pt2d[0];
    pointsOnBoundary[BLeft][i][2] = pt2d[1];
    pointsOnBoundary[BRight][i][0] = 1;
    pointsOnBoundary[BRight][i][1] = pt2d[0];
    pointsOnBoundary[BRight][i][2] = pt2d[1];

    pointsOnBoundary[BFront][i][0] = pt2d[0];
    pointsOnBoundary[BFront][i][1] = -1;
    pointsOnBoundary[BFront][i][2] = pt2d[1];
    pointsOnBoundary[BBack][i][0] = pt2d[0];
    pointsOnBoundary[BBack][i][1] = 1;
    pointsOnBoundary[BBack][i][2] = pt2d[1];

    pointsOnBoundary[BDown][i][0] = pt2d[0];
    pointsOnBoundary[BDown][i][1] = pt2d[1];
    pointsOnBoundary[BDown][i][2] = -1;
    pointsOnBoundary[BUp][i][0] = pt2d[0];
    pointsOnBoundary[BUp][i][1] = pt2d[1];
    pointsOnBoundary[BUp][i][2] = 1;    
  }
  
  // generating disPsiBoundary

  for(int fun = 0; fun < 6; ++fun)
    for(int boundary = 0; boundary < BSize; ++ boundary)
      for(int i = 0; i < Integral2d::PointsNumber; ++i)
	disPsiBoundary[fun][boundary][i] = ShapeFun[fun](pointsOnBoundary[boundary][i]);
  
}


template<class Double>
void Discretizator<Double>::
GenShapeFunMP()
  {
    	Poly3d t1;
	Poly3d t2;
	Poly3d t3;

	Poly3d one;
	one[0] = 1;

	t1[1] = 1;
	t2[2] = 1;  
	t3[3] = 1;

	ShapeFun[0] = one + (-3.)*t1 + 2. * t1*t1 + (-1.)*t2*t2 + (-1.)*t3*t3;
	ShapeFun[1] = one + (+3.)*t1 + 2. * t1*t1 + (-1.)*t2*t2 + (-1.)*t3*t3;
	ShapeFun[2] = one + (-3.)*t2 + 2. * t2*t2 + (-1.)*t1*t1 + (-1.)*t3*t3;
	ShapeFun[3] = one + (+3.)*t2 + 2. * t2*t2 + (-1.)*t1*t1 + (-1.)*t3*t3;
	ShapeFun[4] = one + (-3.)*t3 + 2. * t3*t3 + (-1.)*t1*t1 + (-1.)*t2*t2;
	ShapeFun[5] = one + (+3.)*t3 + 2. * t3*t3 + (-1.)*t1*t1 + (-1.)*t2*t2;
	
	for(int i = 0; i < 6; ++i)
	  {
	    ShapeFun[i] *= 1./6.;
	  }
	GetPerturb<Double>()=1;
  }


template<class Double>
void Discretizator<Double>::
GenShapeFunMV()
  {
    using namespace My::Base;
    	Poly3d t1;
	Poly3d t2;
	Poly3d t3;

	Poly3d one;
	one[0] = 2; // one == 2!!!

	t1[1] = 1;
	t2[2] = 1;  
	t3[3] = 1;

	ShapeFun[0] = one + (-6.)*t1 + 6. * t1*t1 + (-3.)*t2*t2 + (-3.)*t3*t3;
	ShapeFun[1] = one + (+6.)*t1 + 6. * t1*t1 + (-3.)*t2*t2 + (-3.)*t3*t3;
	ShapeFun[2] = one + (-6.)*t2 + 6. * t2*t2 + (-3.)*t1*t1 + (-3.)*t3*t3;
	ShapeFun[3] = one + (+6.)*t2 + 6. * t2*t2 + (-3.)*t1*t1 + (-3.)*t3*t3;
	ShapeFun[4] = one + (-6.)*t3 + 6. * t3*t3 + (-3.)*t1*t1 + (-3.)*t2*t2;
	ShapeFun[5] = one + (+6.)*t3 + 6. * t3*t3 + (-3.)*t1*t1 + (-3.)*t2*t2;
	
	for(int i = 0; i < 6; ++i)
	  {
	    ShapeFun[i] *= 1./12;
	  }
	GetPerturb<Double>()=1;
  }


template<class Double>
template<class Matrix, class Vector, class Int>
void Discretizator<Double>::
BoundCondSet(Matrix &mat, Vector &vec, Int I, Int J, Int K, Int MI, Int MJ, Int MK, BoundaryConditions bc)
{
  if ((bcDown&bc)&&(K == 0)){ // Dirichlet boundary conditions setting // Down
    for(int i = 0; i < mat.Height(); ++i)
      mat[i][4] = mat[4][i] = 0;
    mat[4][4] = 1;
    vec[4] = 0;
    }
  
  if ((bcUp&bc)&&(K == MK)){  // Dirichlet boundary conditions setting // Up
    for(int i = 0; i < mat.Height(); ++i)
      mat[i][5] = mat[5][i] = 0;
    mat[5][5] = 1;
    vec[5] = 0;
  }
  
  if ((bcLeft&bc)&&(I == 0)){ // Dirichlet boundary conditions setting // Left
    for(int i = 0; i < mat.Height(); ++i)
      mat[i][0] = mat[0][i] = 0;
    mat[0][0] = 1;
    vec[0] = 0;
  }

  if ((bcRight&bc)&&(I == MI)){ // Dirichlet boundary conditions setting // Right
    for(int i = 0; i < mat.Height(); ++i)
      mat[i][1] = mat[1][i] = 0;
    mat[1][1] = 1;
    vec[1] = 0;
  }
  
  if ((bcFront&bc)&&(J == 0)){ // Dirichlet boundary conditions setting // Front
    for(int i = 0; i < mat.Height(); ++i)
      mat[i][2] = mat[2][i] = 0;
    mat[2][2] = 1;
    vec[2] = 0;
  }
  
  if ((bcBack&bc)&&(J == MJ)){ // Dirichlet boundary conditions setting // Back
    for(int i = 0; i < mat.Height(); ++i)
      mat[i][3] = mat[3][i] = 0;
    mat[3][3] = 1;
    vec[3] = 0;
  }
  
}
#ifdef MY_HYPRE
#include "hypre_precon.h"

template<class Elmat, class Elrhs, class Ind>
inline void UpdateMatAndVec(HypreIJMatrix& mat, HypreIJVector& vec,
  const Elmat& elmat,
  const Elrhs& elrhs, Ind &ind,int i, int j, int k, int N1, int N2, int N3)
{
  SmallVector<int, 6> nv;
  CalcIndeces(nv,i,j,k, N1, N2, N3);
  mat.AddTo(elmat, nv);
  vec.AddTo(elrhs, nv);
}
#endif
template<class Mat, class Rhs, class Elmat, class Elrhs, class Ind>
inline void UpdateMatAndVec(Mat& mat, Rhs& rhs,
  const Elmat& elmat,
  const Elrhs& elrhs, Ind &ind,int i, int j, int k, int N1, int N2, int N3)
{
  // vector initialization
  rhs[ind][LeftEl] += elrhs[0];
  rhs.Right(ind)[LeftEl] += elrhs[1];
  rhs[ind][FrontEl] += elrhs[2];
  rhs.Back(ind)[FrontEl] += elrhs[3];
  rhs[ind][DownEl]  += elrhs[4];
  rhs.Up(ind)[DownEl] += elrhs[5];
	     
  // matrix initialization
	     
  // Left 
  mat[ind][LeftEl][0]        += elmat[0][0]; // Left
  mat[ind][LeftEl][5]        += elmat[0][1]; // right
  mat[ind][LeftEl][1]        += elmat[0][2]; // front
  mat[ind][LeftEl][4]        += elmat[0][3]; // back;
  mat[ind][LeftEl][2]        += elmat[0][4]; // down
  mat[ind][LeftEl][3]        += elmat[0][5]; // up
	     
  // right        
  mat.Right(ind)[LeftEl][0]  += elmat[1][1]; // right
  mat[ind][FrontEl][5]       += elmat[1][2]; // front
  mat.Back(ind)[FrontEl][4]  += elmat[1][3]; // back;
  mat[ind][DownEl][5]        += elmat[1][4]; // down
  mat.Up(ind)[DownEl][4]     += elmat[1][5]; // up
	     
  // front
  mat[ind][FrontEl][0]       += elmat[2][2]; // front
  mat[ind][FrontEl][3]       += elmat[2][3]; // back;
  mat[ind][FrontEl][1]       += elmat[2][4]; // down
  mat[ind][FrontEl][2]       += elmat[2][5]; // up
	     
	     
  // back
  mat.Back(ind)[FrontEl][0]  += elmat[3][3]; // back;
  mat[ind][DownEl][3]        += elmat[3][4]; // down
  mat.Up(ind)[DownEl][2]     += elmat[3][5]; // up
	     
  // Down
  mat[ind][DownEl][0]        += elmat[4][4]; // down
  mat[ind][DownEl][1]        += elmat[4][5]; // up
	     
  // Up
  mat.Up(ind)[DownEl][0]     += elmat[5][5]; // up
}


template<class Double>
 template<class Matrix, class Vector>
void Discretizator<Double>::
 GenerateMatrix(Matrix &mat, Vector &rhs, 
		       typename Matrix::int_type NumberOfLines,
		       typename Matrix::int_type  StartLine,
		       typename Matrix::int_type LO, 
		       typename Matrix::int_type RealN3, BoundaryConditions bc )
   {
     Index3dBase<typename Matrix::int_type> ind(mat.N1, mat.N2, mat.N3);
     
     typename Matrix::int_type i, j, k;
     
     ClearMatrix(mat);
     ClearVector(rhs);
     
     ElementMatrix_type elmat;
     ElementMatrix_type mmat;
     ElementRhs_type elrhs, fdis;
     SmallVector<typename Vector::base_type, 3> P;
     
     SmallMatrix<typename Vector::base_type, 8, 3> elpoints;
     //     typename Vector::base_type h1, h2, h3;
     
     SmallVector<typename Vector::base_type, 3> h;
     //     typedef typename Vector::base_type Double;
     h[0] = Options::GetOptions().DM1/mat.N1;
     h[1] = Options::GetOptions().DM2/mat.N2;
     h[2] = Options::GetOptions().DM3/mat.N3R;
     //     std::cout << mat.N1 << " " << mat.N3R << std::endl;
     
     bool first_time = true;
     
     
     elpoints[0][0] = 0;
     elpoints[0][1] = 0;
     elpoints[0][2] = 0;
     
     elpoints[3] = elpoints[0];
     elpoints[3][1] += h[1];
     
     elpoints[1] = elpoints[0];
     elpoints[1][0] += h[0];
     
     elpoints[2] = elpoints[3];
     elpoints[2][0] += h[0];
     
     for(int jj = 0; jj <4; jj++)
       {
	 elpoints[4+jj] = elpoints[jj];
	 elpoints[4+jj][2] += h[2];
       }
     
     tri.SetDest(elpoints);
     if(first_time)
       {
	 //first_time = false;
	 for(int f = 0; f < Integral::PointsNumber; f++)
	   {
	     tri.GetJacobianAndReverseAtPoint(disJrev[f], disJdet[f], integral.points[f]);
	     disJrev[f]/=disJdet[f];
	     MatrixMultiply(disJrevDPsiDt[f], disJrev[f], disDPsiDt[f]);
	   }
	 
       }
     
     for (i = 0; i < mat.N1; ++i)
       for( j = 0; j < mat.N2; ++j)
	 for( k = 0; k < NumberOfLines; ++k)
	   {
	     ind.SetPos(i, j, k+LO);
	     
	     elpoints[0][0] = h[0]*(Double)i;
	     elpoints[0][1] = h[1]*(Double)j;
	     elpoints[0][2] = h[2]*(Double)(k+StartLine);

	     elpoints[1][0] = h[0]*(Double)(i+1);
	     elpoints[1][1] = h[1]*(Double)j;
	     elpoints[1][2] = h[2]*(Double)(k+StartLine);

	     elpoints[2][0] = h[0]*(Double)(i+1);
	     elpoints[2][1] = h[1]*(Double)(j+1);
	     elpoints[2][2] = h[2]*(Double)(k+StartLine);

	     elpoints[3][0] = h[0]*(Double)i;
	     elpoints[3][1] = h[1]*(Double)(j+1);
	     elpoints[3][2] = h[2]*(Double)(k+StartLine);

	     elpoints[4][0] = h[0]*(Double)i;
	     elpoints[4][1] = h[1]*(Double)j;
	     elpoints[4][2] = h[2]*(Double)(k+StartLine+1);

	     elpoints[5][0] = h[0]*(Double)(i+1);
	     elpoints[5][1] = h[1]*(Double)j;
	     elpoints[5][2] = h[2]*(Double)(k+StartLine+1);

	     elpoints[6][0] = h[0]*(Double)(i+1);
	     elpoints[6][1] = h[1]*(Double)(j+1);
	     elpoints[6][2] = h[2]*(Double)(k+StartLine+1);

	     elpoints[7][0] = h[0]*(Double)i;
	     elpoints[7][1] = h[1]*(Double)(j+1);
	     elpoints[7][2] = h[2]*(Double)(k+StartLine+1);

	     SmallMatrix<SmallMatrix<Double, 6,6>, 3, 3> raz;


	     
	     if(distype == RP)
	     {
	       for(int iii = 0; iii < 6; ++iii)
		 for(int j = 0; j < 6; ++j)
		 {
		   for (int I = 0; I < 3; ++I)
		     for(int J = 0; J < 3; ++J)
		     {
		       raz[I][J][iii][j] = 4;
		     }
		 }
	       {
	       for(int i = 0; i < 6; ++i)
	       {
		 raz[0][0][0][i] = -8;
		 raz[0][0][i][0] = -8;
		 raz[0][0][1][i] = -8;
		 raz[0][0][i][1] = -8;

		 raz[1][1][2][i] = -8;
		 raz[1][1][i][2] = -8;
		 raz[1][1][3][i] = -8;
		 raz[1][1][i][3] = -8;

		 raz[2][2][4][i] = -8;
		 raz[2][2][i][4] = -8;
		 raz[2][2][5][i] = -8;
		 raz[2][2][i][5] = -8;		 
	       }}
	       raz[0][0][0][0] = 43;
	       raz[0][0][1][1] = 43;
	       raz[0][0][0][1] = -11;
	       raz[0][0][1][0] = -11;

	       raz[1][1][2][2] = 43;
	       raz[1][1][3][3] = 43;
	       raz[1][1][2][3] = -11;
	       raz[1][1][3][2] = -11;

	       raz[2][2][4][4] = 43;
	       raz[2][2][5][5] = 43;
	       raz[2][2][4][5] = -11;
	       raz[2][2][5][4] = -11;
	       
	       for(int ii = 0; ii < elmat.Height();++ii){
		 for(int j = 0; j < elmat.Width();++j)
		 {
		   elmat[ii][j] = -4;
		   mmat[ii][j] = 22;
		 }
		 elmat[ii][ii] = 17;
		 mmat[ii][ii] = 136;
		 elrhs[ii] = 0;
	       }
	       elmat[0][1] = elmat[1][0] =
		 elmat[2][3] = elmat[3][2] =
		 elmat[4][5] = elmat[5][4] = -1;	    
	       mmat[0][1] = mmat[1][0] =
		 mmat[2][3] = mmat[3][2] =
		 mmat[4][5] = mmat[5][4] = -44;	    
	       //		 Double J = mat.N1;
	       //		 elmat /= J;
	       //		 elrhs *= J*J*J;
	       //mmat *= 0.2; // /5 (divided by five)
	       //		 mmat /= 27;
	       //		 elmat *= h[2];
	       elmat *= 5;
	       //		 elmat /= 27;
	       elmat *= 3;
	       SmallMatrix<Double, 3, 3> &aaa = A(elpoints);
	       {
		 elmat = 0;
		 for(int i = 0; i < 3; ++i)
		 {
		   SmallMatrix<Double, 6, 6> p;
		   p = raz[i][i];
		   p *= aaa[i][i];
		   elmat += p;
		 }
		 elmat *=5;
	       }
	       //	       elmat *= (aaa[0][0]+aaa[1][1]+aaa[2][2])/3;
	       //		 elmat /= 1.2256; //eigenvalue
	       mmat *= h[0] * h[1] /8. ;//* h[2]/8 ;

	       P[0] = h[0]*(Double)i;
	       P[1] = h[1]*(Double)(j + 0.5);
	       P[2] = h[2]*(Double)(k+StartLine + 0.5);
	       fdis[0] = F(P);
	       P[0] = h[0]*(Double)(i+1);
	       P[1] = h[1]*(Double)(j + 0.5);
	       P[2] = h[2]*(Double)(k+StartLine + 0.5);
	       fdis[1] = F(P);
	       P[0] = h[0]*(Double)(i+0.5);
	       P[1] = h[1]*(Double)(j + 0.0);
	       P[2] = h[2]*(Double)(k+StartLine + 0.5);
	       fdis[2] = F(P);
	       P[0] = h[0]*(Double)(i+0.5);
	       P[1] = h[1]*(Double)(j + 1.0);
	       P[2] = h[2]*(Double)(k+StartLine + 0.5);
	       fdis[3] = F(P);
	       P[0] = h[0]*(Double)(i+0.5);
	       P[1] = h[1]*(Double)(j + 0.5);
	       P[2] = h[2]*(Double)(k+StartLine + 0.0);
	       fdis[4] = F(P);
	       P[0] = h[0]*(Double)(i+0.5);
	       P[1] = h[1]*(Double)(j + 0.5);
	       P[2] = h[2]*(Double)(k+StartLine + 1.0);
	       fdis[5] = F(P);

	       for(int i = 0; i < 6; i++)
		 for(int j = 0; j < 6; j++)
		   elrhs[i] += fdis[j] * mmat[i][j];

		 
// 	       elmat /= h[0]*h[1]*h[2];
// 	       elrhs /= h[0]*h[1]*h[2];


	     }
	     else
	       if(distype == RV)
	       {
		 {
		   for(int i = 0; i < 6; ++i)
		 for(int j = 0; j < 6; ++j)
		 {
		   for (int I = 0; I < 3; ++I)
		     for(int J = 0; J < 3; ++J)
		     {
		       raz[I][J][i][j] = 1;
		     }
		 }
		   
	       for(int i = 0; i < 6; ++i)
	       {
		 raz[0][0][0][i] = -2;
		 raz[0][0][i][0] = -2;
		 raz[0][0][1][i] = -2;
		 raz[0][0][i][1] = -2;

		 raz[1][1][2][i] = -2;
		 raz[1][1][i][2] = -2;
		 raz[1][1][3][i] = -2;
		 raz[1][1][i][3] = -2;

		 raz[2][2][4][i] = -2;
		 raz[2][2][i][4] = -2;
		 raz[2][2][5][i] = -2;
		 raz[2][2][i][5] = -2;		 
	       }}
	       raz[0][0][0][0] = 7;
	       raz[0][0][1][1] = 7;
	       raz[0][0][0][1] = 1;
	       raz[0][0][1][0] = 1;

	       raz[1][1][2][2] = 7;
	       raz[1][1][3][3] = 7;
	       raz[1][1][2][3] = 1;
	       raz[1][1][3][2] = 1;

	       raz[2][2][4][4] = 7;
	       raz[2][2][5][5] = 7;
	       raz[2][2][4][5] = 1;
	       raz[2][2][5][4] = 1;
	       
		 for(int ii = 0; ii < elmat.Height();++ii){
		   for(int j = 0; j < elmat.Width();++j)
		   {
		     elmat[ii][j] = -1;
		     mmat[ii][j] = 4;
		   }
		   elmat[ii][ii] = 3;
		   mmat[ii][ii] = 52 ;
		   elrhs[ii] = 0;
		 }
		 elmat[0][1] = elmat[1][0] =
		   elmat[2][3] = elmat[3][2] =
		   elmat[4][5] = elmat[5][4] = 1;	    
		 mmat[0][1] = mmat[1][0] =
		   mmat[2][3] = mmat[3][2] =
		   mmat[4][5] = mmat[5][4] = -8;	    
		 //		 mmat /= 45;
		 elmat *= 45.;
		 SmallMatrix<Double, 3, 3> &aaa = A(elpoints);
		 elmat *= aaa[0][0];//+aaa[1][1]+aaa[2][2])/3;

		 {
		   elmat = 0;
		   for(int i = 0; i < 3; ++i)
		   {
		     SmallMatrix<Double, 6, 6> p;
		     p = raz[i][i];
		     p *= aaa[i][i];
		     elmat += p;
		   }
		   elmat *=15;
		 }
		 mmat *= h[0] * h[1] /8. ;//* h[2]/8 ;

		 P[0] = h[0]*(Double)i;
		 P[1] = h[1]*(Double)(j + 0.5);
		 P[2] = h[2]*(Double)(k+StartLine + 0.5);
		 fdis[0] = F(P);
		 P[0] = h[0]*(Double)(i+1);
		 P[1] = h[1]*(Double)(j + 0.5);
		 P[2] = h[2]*(Double)(k+StartLine + 0.5);
		 fdis[1] = F(P);
		 P[0] = h[0]*(Double)(i+0.5);
		 P[1] = h[1]*(Double)(j + 0.0);
		 P[2] = h[2]*(Double)(k+StartLine + 0.5);
		 fdis[2] = F(P);
		 P[0] = h[0]*(Double)(i+0.5);
		 P[1] = h[1]*(Double)(j + 1.0);
		 P[2] = h[2]*(Double)(k+StartLine + 0.5);
		 fdis[3] = F(P);
		 P[0] = h[0]*(Double)(i+0.5);
		 P[1] = h[1]*(Double)(j + 0.5);
		 P[2] = h[2]*(Double)(k+StartLine + 0.0);
		 fdis[4] = F(P);
		 P[0] = h[0]*(Double)(i+0.5);
		 P[1] = h[1]*(Double)(j + 0.5);
		 P[2] = h[2]*(Double)(k+StartLine + 1.0);
		 fdis[5] = F(P);

		 for(int i = 0; i < 6; i++)
		   for(int j = 0; j < 6; j++)
		     elrhs[i] += fdis[j] * mmat[i][j];

		 
		 


	       }
	       else
	       {
		 tri.SetDest(elpoints);
		 
		 //if(first_time)
		 {
		   first_time = false;
		   for(int f = 0; f < Integral::PointsNumber; f++)
		   {
		     tri.GetJacobianAndReverseAtPoint(disJrev[f], disJdet[f], integral.points[f]);
		     disJrev[f]/=disJdet[f];
		     MatrixMultiply(disJrevDPsiDt[f], disJrev[f], disDPsiDt[f]);
		   }
		     
		 }
		 
		 SmallMatrix<typename Vector::base_type, 6, 3> dPsiDtJrevTran;
		 SmallMatrix<typename Vector::base_type, 6, 3> tmp;
		 SmallMatrix<typename Vector::base_type, 6, 6> tmp2;
		 //		 SmallMatrix<typename Vector::base_type, 3, 3> &aa = A(elpoints);
		 SmallMatrix<typename Vector::base_type, 3, 3> a;
		 SmallVector<typename Vector::base_type, Integral::PointsNumber> disF;
		 
		 elmat = 0.0;
		 elrhs = 0.0;
		 
		 SmallMatrix<typename Vector::base_type, 6, 6> elx, ely, elz;
		 elx = 0;
		 ely = 0;
		 elz = 0;
		 for(int f = 0; f < Integral::PointsNumber; f++)
		 {
		   //a = aa;
		   a = A(tri.toX(integral.points[f]));
		   Transpose(dPsiDtJrevTran, disJrevDPsiDt[f]);
		   //	    a /= disJdet[f];
		   //	    a *= integral.coeff[f];

// 		   printmat
// 		   for(int i = 0; i < 6; ++i)
// 		     for(int j = 0; j < 6; ++j)
// 		     {
// 		       elx[i][j] += dPsiDtJrevTran[i][0] * dPsiDtJrevTran[j][0] * integral.coeff[f]*disJdet[f];
// 		       ely[i][j] += dPsiDtJrevTran[i][1] * dPsiDtJrevTran[j][1] * integral.coeff[f]*disJdet[f];
// 		       elz[i][j] += dPsiDtJrevTran[i][2] * dPsiDtJrevTran[j][2] * integral.coeff[f]*disJdet[f];
// 		     }
		       
		   MatrixMultiply(tmp, dPsiDtJrevTran, a);
		   MatrixMultiply(tmp2, tmp, disJrevDPsiDt[f]);
		   tmp2 *= integral.coeff[f]*disJdet[f];
		   elmat += tmp2;
		   disF[f] = F(tri.toX(integral.points[f]))*integral.coeff[f] * disJdet[f];
		     
		 }
		 // elmatprn
// 		 double konst = (distype == MP)?27:3;

// 		 for(int i = 0; i < 6; ++i, cout << endl)
// 		   for(int j = 0; j < 6; ++j)
// 		     cout << setw(10) <<konst/h[0]*elx[i][j] ;
// 		 cout << endl;
// 		 for(int i = 0; i < 6; ++i, cout << endl)
// 		   for(int j = 0; j < 6; ++j)
// 		     cout << setw(10) <<konst/h[0]*ely[i][j] ;
// 		 cout << endl;
// 		 for(int i = 0; i < 6; ++i, cout << endl)
// 		   for(int j = 0; j < 6; ++j)
// 		     cout << setw(10) <<konst*elz[i][j]/h[0] ;
// 		 exit(1);

		 SmallVector<typename Vector::base_type, 3> p;
		 p[0] = -1;
		 p[1] = 1;
		 p[2] = 0;
		 //	cout << tri.toX(p) << endl;
	    
		 
		 for(int f = 0; f < Integral::PointsNumber; f++)
		   for(int of = 0; of < 6; of++)
		     elrhs[of] += disF[f] * disShapeFun[f][0][of];
		 
	       }
	     
/* 		 	for (int i = 0; i < 6; i ++) */
/* 		 	  { */
/* 		 	    for(int j = 0; j < 6; j++) */
/* 		 	      cout << setw(15) <<  elmat[i][j] ; */
/* 		 	    cout << endl; */
/* 		 	  } */
/* 		 	cout << endl; */
/* 		 	return ; */
	     
	     BoundCondSet(elmat, elrhs, i, j, k+StartLine, mat.N1 - 1, mat.N2 - 1, RealN3 - 1, bc);
	     
	     
	     UpdateMatAndVec(mat, rhs,
			     elmat, elrhs,
			     ind, i,j,k+StartLine, mat.N1, mat.N2, RealN3);
	     
	     
	   }
   }






vector3d<double> coefficientvec;

#endif
