#ifndef __useful_h__
#define __useful_h__
// IBM workaround
#ifdef _MY_CPP
#include "mympicxx.hpp"
#endif

#include "mpi.h"


#include <string>
#include <sstream>
#include <map>
#include <iostream>
#include <iomanip>


#include "small.h"

template<class T>
T& GetPerturb()
{
  static T a = 0;
  return a;
}

class ManageTimes
{
 public:
  typedef std::map<std::string, double> map_type;
  typedef map_type::iterator map_iterator;
  static map_type& GetMap()
  {
    static map_type map;
    return map;
  }
  static double & Get(const std::string &str)
  {
    return GetMap()[str];
  }
  static void Reduce()
  {
    for(map_iterator it = GetMap().begin(); it!=GetMap().end();++it)
    {
      double time1 = it->second;
      double time;
      MPI::COMM_WORLD.Allreduce(&time1, &time, 1, MPI::DOUBLE, MPI::MAX);
      it->second=time;
    }
  }
  template<class Ostream>
  static void PrintAll(Ostream &out, int iter=1, std::string su="")
  {
    for(map_iterator it = GetMap().begin(); it!=GetMap().end();++it)
    {
      double time = it->second;
      using namespace std;
      if (MPI::COMM_WORLD.Get_rank()==0)
	out << left<< setw(30) << it->first + su
	    << setw(30)<< time/iter << "\n";
    }
  }
  
};

class Timer
{
    std::string time;
 public:
    Timer(const std::string &tm):time(tm)
	{
	    ManageTimes::Get(time) -= MPI::Wtime();
	}
    ~Timer()
	{
	  ManageTimes::Get(time) += MPI::Wtime();
	}
};



class Options
{
public:
  bool Overlap;
  bool SaveSolution;
  bool debug;
  int variant;
  int agr;
  int ct;
  int mit;
  double pfactor;
  double rwt;
  double ppfactor;
  bool SinglePrecPrecon;
  double zero;
  double thres;
  double DM1, DM2, DM3;
  bool usebone;
  double E[256];
  double ni[256];
  unsigned NumMaterials;
  double force;
  std::string namebase;
  std::string outputname;
  std::string scratchpath;
  Options()
    {
      mit = 1;
      debug = false;
      ct = 1;
      agr = 0;
      rwt = 0.5;
      thres = 0.5;
      SaveSolution = false;
      SinglePrecPrecon = false;
      force = 0;
      NumMaterials=256;
      for(unsigned int i = 0; i < NumMaterials; ++i)
	E[i] = ni[i] = 0;
      E[0] = 10e6;
      ni[0] = 0.35;
      E[255] = 10e6;
      ni[255] = 0.35;
      Overlap = true;
      pfactor = 1.;
      ppfactor = 1.;
      usebone = false;
      zero = 1.;
      DM1 = DM2 = DM3 = 1.;
      variant=0;
    }

  
  static Options& GetOptions()
  {
    static Options opts;
    return opts;
  }
};

#define OVERLAP (Options::GetOptions().Overlap)


#define TimeCalc1(what) Timer t_m_p## 1 (what)
#define TimeCalc(what) Timer t_m_p ## 1 (what)
#define TimeCalc1p(what,ln) Timer t_m_p## ln (what)
#define TimeCalcp(what,ln) Timer t_m_p ## ln (what)
#define TimeCalc2(what, howmany) ManageTimes::Get(what)+=(howmany)
template<class T> class Myf;

template<> class Myf<double>
{
  public:
  typedef double type;
  typedef double base_type;
  typedef volatile float single;
static MPI::Datatype Datatype()
{
  return  MPI::DOUBLE;
}
static base_type Reduce(type& a)
{
  return a;
}
static std::string ToString()
   {
     return "d";
   }
};
template<> class Myf<float>
{
  public:
  typedef float type;
  typedef float base_type;
  typedef volatile float single;
static MPI::Datatype Datatype()
{
  return  MPI::FLOAT;
}
static base_type Reduce(type& a)
{
  return a;
}
static std::string ToString()
{
  return "f";
}
};

template<class T>
class Myf<SmallVector<T,3> >
{
public:
  MPI::Datatype mpitype;
  Myf() // Here we construct the datatype
  {
    MPI::Datatype bdt = Myf<T>::Datatype();
    mpitype = bdt.Create_contiguous(3);
    mpitype.Commit();
  }
  ~Myf() // Here we destruct it;
  {
    mpitype.Free();
  }
  typedef SmallVector<T, 3> type;
  typedef T base_type;
  typedef SmallVector<volatile float, 3> single;
  static Myf*& Geta()
  {
    static Myf *a;
    return a;
  }
  static MPI::Datatype Datatype()
  {

    return Geta()->mpitype;
  }
  static std::string ToString()
  {
    return "3" + Myf<T>::ToString();
  }
static base_type Reduce(type& a)
{
  return a[0]+a[1]+a[2];
}
static base_type Max(type& a)
{
  return std::max(a[0],std::max(a[1],a[2]));
}
};


template<class T> class Myf
{
public:
  static MPI::Datatype Datatype()
    {
      throw "losh tip";
      //      return  MPI::FLOAT;
    }
  static std::string ToString()
    {
      return "fucked";
    }
};






inline unsigned  GetNum (std::string s, unsigned def)
{
  std::istringstream is(s);
  unsigned a;
  is >> a;
  if (a)
    return a;
  return def;
}


#endif
