#ifndef __visitors_hpp__
#define __visitors_hpp__
#include "visitors.h"
#include <iostream>

template<class Int, class Functor, class FunctorDown, class FunctorUp>
void
ForAll<Int, Functor, FunctorDown, FunctorUp>::
operator()(Index3dBase<Int> const &ind, 
	    Functor &F, FunctorDown &FDown, FunctorUp &FUp)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), K(ind);
  I.SetPos(0, 0, -1);

  //  std::cout << "height is " << ind.N3 << std::endl;
  
  for(i = ind.N1;i; --i, I.ToRight())
  {
    for(j = ind.N2, J=I; j; --j, J.ToBack())
    { 
      FDown.template Do<LeftEl>(J);
      for(k = ind.N3-1, K = J, K.ToUp(); k>0; --k, K.ToUp()) // doing middle
      {
        F. DoAll(K);
      }
      FUp.template Do<FrontEl>(K);
      FUp.template Do<DownEl>(K);                             // top plane
    }
    for(k = ind.N3-1, K = J, K.ToUp(); k>0; --k, K.ToUp())  // back plane
    {
      F. template Do<FrontEl>(K);
    }
    FUp.template Do<FrontEl>(K);
  }
  for(j = ind.N2, J=I; j; --j, J.ToBack())  // right plane
  { 
    FDown.template Do<LeftEl>(J);
    for(k = ind.N3-1, K = J, K.ToUp(); k>0; --k, K.ToUp())
    {
      F. template Do<LeftEl>(K);
    }
  }
}


template<class Int, class Functor, class FunctorDown, class FunctorUp>
void
ForAllReverse<Int, Functor, FunctorDown, FunctorUp>::
operator()(Index3dBase<Int> const &ind, 
	    Functor &F, FunctorDown &FDown, FunctorUp &FUp)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), K(ind);
  I.SetPos(ind.N1, ind.N2, ind.N3-1);


  for(j = ind.N2, J=I, J.ToFront(); j; --j, J.ToFront())  // right plane
  { 
    for(k = ind.N3-1, K = J, K.ToDown(); k; --k, K.ToDown())
    {
      F. template Do<LeftEl>(K);
    }
  }
  

  for(i = ind.N1, I.ToLeft();i; --i, I.ToLeft())
  {
    for(k = ind.N3-1, J = I, K = J, K.ToDown(); k; --k, K.ToDown())  // back plane
    {
      F. template Do<FrontEl>(K);
    }
    for(j = ind.N2, J=I, J.ToFront(); j; --j, J.ToFront())
    { 
      F.template Do<DownEl>(J);                             // top plane
      for(k = ind.N3-1, K = J, K.ToDown(); k; --k, K.ToDown()) // doing middle
      {
        F. DoAll(K);
      }
    }
  }

}

template<class Int, class Functor, class FunctorDown, class FunctorUp>
void ForAllWithoutDown<Int, Functor, FunctorDown, FunctorUp>::
operator()(Index3dBase<Int> const &ind, 
	    Functor &F, FunctorDown &FDown, FunctorUp &FUp)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), K(ind);
  I.SetPos(0, 0, -1);

//  std::cout << "height is " << ind.N3 << std::endl;
  
  for(i = ind.N1;i; --i, I.ToRight())
  {
    for(j = ind.N2, J=I; j; --j, J.ToBack())
    { 
      //      FDown.template Do<LeftEl>(J); // without down
      for(k = ind.N3-1, K = J, K.ToUp(); k>0; --k, K.ToUp()) // doing middle
      {
        F. DoAll(K);
      }
      FUp.template Do<DownEl>(K);                             // top plane
      FUp.template Do<FrontEl>(K);
    }
    for(k = ind.N3-1, K = J, K.ToUp(); k>0; --k, K.ToUp())  // back plane
    {
      F. template Do<FrontEl>(K);
    }
    FUp.template Do<FrontEl>(K);
  }
  for(j = ind.N2, J=I; j; --j, J.ToBack())  // right plane
  { 
    //FDown.template Do<LeftEl>(J); // without down
    for(k = ind.N3-1, K = J, K.ToUp(); k>0; --k, K.ToUp())
    {
      F. template Do<LeftEl>(K);
    }
  }
}


template<class Int, class Functor, class FunctorDown, class FunctorUp>
void ForDownOnly<Int, Functor, FunctorDown, FunctorUp>::
operator()(Index3dBase<Int> const &ind, 
	    Functor &F, FunctorDown &FDown, FunctorUp &FUp)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), K(ind);
  I.SetPos(0, 0, -1);

//  std::cout << "height is " << ind.N3 << std::endl;
  
  for(i = ind.N1+1;i; --i, I.ToRight())
  {
    for(j = ind.N2, J=I; j; --j, J.ToBack())
    { 
      FDown.template Do<LeftEl>(J); // without down
    }
  }
}


template<class Int, class Functor, class FunctorDown, class FunctorUp>
void ForAllM3Up<Int, Functor, FunctorDown, FunctorUp>::
operator()(Index3dBase<Int> const &ind, 
	    Functor &F, FunctorDown &FDown, FunctorUp &FUp)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), K(ind), Iu(ind), Id(ind);
  I.SetPos(0, 0, -1);

  
  for(i = ind.N1;i; --i, I.ToRight())
  {
    Iu.SetPos(ind.N1-i, 0, ind.N3-1);
    Id.SetPos(ind.N1-i, 0, -1);
    //left plane
    for(j = ind.N2, J=I; j; --j, J.ToBack())
    { 
      FDown.template Do<LeftEl>(J);
    }


    F.StartSendLineLeft(Id, Iu);
    for(j = ind.N2, J=I, J.ToUp(); j; --j, J.ToBack())
      for(k = ind.N3-1, K = J; k>0; --k, K.ToUp())
      {
        F.template Do<LeftEl>(K);
      }
    
    F.FinishSendLineLeft(Id, Iu);


    for(j = ind.N2, J=Iu; j; --j, J.ToBack())
      { 
	FUp.template Do<DownEl>(J);
	FUp.template Do<FrontEl>(J);
      }
    FUp.template Do<FrontEl>(J);
    
    Id.ToRight();Iu.ToRight();
    
    F.StartSendLineMiddle(Id, Iu);
    for(j = ind.N2, J=I, J.ToUp(); j; --j, J.ToBack())
      for(k = ind.N3-1, K = J; k>0; --k, K.ToUp())
	{
	  F.template Do<DownEl>(K);
	  F.template Do<FrontEl>(K);
	}    
    for(k = ind.N3-1, K = J; k>0; --k, K.ToUp())
      {
	F.template Do<FrontEl>(K);
      }
    F.FinishSendLineMiddle(Id, Iu);
    
  }
  for(j = ind.N2, J=I; j; --j, J.ToBack())  // right plane
  { 
    FDown.template Do<LeftEl>(J);
    for(k = ind.N3-1, K = J, K.ToUp(); k>0; --k, K.ToUp())
    {
      F. template Do<LeftEl>(K);
    }
  }

}


template<class Int, class Functor, class FunctorDown, class FunctorUp>
void ForAllM3Down<Int, Functor, FunctorDown, FunctorUp>::
operator()(Index3dBase<Int> const &ind, 
	    Functor &F, FunctorDown &FDown, FunctorUp &FUp)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), K(ind), Iu(ind), Id(ind);
  I.SetPos(ind.N1, 0, -1);

  
  for(i = ind.N1;i;)
  {
    Iu.SetPos(i, 0, ind.N3-1);
    Id.SetPos(i, 0, -1);
    //left plane
    for(j = ind.N2, J=I; j; --j, J.ToBack())
    { 
      FDown.template Do<LeftEl>(J);
    }
    F.StartSendLineLeft(Id, Iu);
    for(j = ind.N2, J=I, J.ToUp(); j; --j, J.ToBack())
      for(k = ind.N3-1, K = J; k>0; --k, K.ToUp())
      {
        F.template Do<LeftEl>(K);
      }    
    F.FinishSendLineLeft(Id, Iu);

    --i, I.ToLeft();
    Id.ToLeft();Iu.ToLeft();

    for(j = ind.N2, J=Iu; j; --j, J.ToBack())
      { 
	FUp.template Do<DownEl>(J);
	FUp.template Do<FrontEl>(J);
      }
    FUp.template Do<FrontEl>(J);
        
    F.StartSendLineMiddle(Id, Iu);
    for(j = ind.N2, J=I, J.ToUp(); j; --j, J.ToBack())
      for(k = ind.N3-1, K = J; k>0; --k, K.ToUp())
	{
	  F.template Do<DownEl>(K);
	  F.template Do<FrontEl>(K);
	}    
    for(k = ind.N3-1, K = J; k>0; --k, K.ToUp())
      {
	F.template Do<FrontEl>(K);
      }
    F.FinishSendLineMiddle(Id, Iu);
    
  }
  for(j = ind.N2, J=I; j; --j, J.ToBack())  // right plane
  { 
    FDown.template Do<LeftEl>(J);
    for(k = ind.N3-1, K = J, K.ToUp(); k>0; --k, K.ToUp())
    {
      F. template Do<LeftEl>(K);
    }
  }

}
template<class Int, class Functor, class FunctorDown, class FunctorUp>
void ForAllMore<Int, Functor, FunctorDown, FunctorUp>::
operator()(Index3dBase<Int> const &ind, 
	    Functor &F, FunctorDown &FDown, FunctorUp &FUp)
{
  //  std::cout << "AllMore " << std::endl;

  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), K(ind);
  I.SetPos(0, 0, -1);

  for(i = ind.N1;i; --i, I.ToRight())
  {
    for(j = ind.N2, J=I; j; --j, J.ToBack())
    { 
      for(k = ind.N3+2, K = J; k>0; --k, K.ToUp()) // doing middle
      {
        F. DoAll(K);
      }
    }
    for(k = ind.N3+2, K = J; k>0; --k, K.ToUp())  // back plane
    {
      F. template Do<FrontEl>(K);
    }
  }
  for(j = ind.N2, J=I; j; --j, J.ToBack())  // right plane
  { 
    for(k = ind.N3+2, K = J; k>0; --k, K.ToUp())
    {
      F. template Do<LeftEl>(K);
    }
  }

}



template<class Int, class Functor, class FunctorDown, class FunctorUp>
void ForAllColumnwiseM2Up<Int, Functor, FunctorDown, FunctorUp>::
operator()(Index3dBase<Int> const &ind, 
	    Functor &F, FunctorDown &FDown, FunctorUp &FUp)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), K(ind), UpInd(ind), DownInd(ind);
  I.SetPos(0, 0, -1);
  
  
  //  std::cout << "M2UP " << std::endl;

    for(j = ind.N2, J=I; j; --j, J.ToBack())
    { 
      FDown.template Do<LeftEl>(J);
      for(k = ind.N3-1, K=J, K.ToUp();k>0; --k, K.ToUp()) 
	F.template Do<LeftEl>(K);    
    }
  
  for(i = ind.N1;i;)
    {
      for(j = ind.N2, J=I; j; --j, J.ToBack())
	{
	  UpInd.SetPos(ind.N1 - i, ind.N2 - j, ind.N3 -1);
	  DownInd.SetPos(ind.N1 - i, ind.N2 - j, -1);
	  F.StartSendFront(DownInd, UpInd);
	  for(k = ind.N3-1, K = J, K.ToUp(); k>0; --k, K.ToUp())
	    F. template Do<FrontEl>(K);
	  F.FinishSendFront(DownInd, UpInd);
	  FUp.template Do<FrontEl>(UpInd);
	  

	  UpInd.SetPos(ind.N1 - i, ind.N2 - j, ind.N3);
	  DownInd.SetPos(ind.N1 - i, ind.N2 - j, 0);
	  F.StartSendDown(DownInd, UpInd);
	  for(k = ind.N3-1, K = J, K.ToUp(), K.ToUp(); k>0; --k, K.ToUp())
	    F.template Do<DownEl>(K);
	  F.FinishSendDown(DownInd, UpInd);
       	  F.template Do<DownEl>(DownInd);
	}
      UpInd.SetPos(ind.N1 - i, ind.N2 - j, ind.N3 -1);
      DownInd.SetPos(ind.N1 - i, ind.N2 - j, -1);
      F.StartSendFront(DownInd, UpInd);
      for(k = ind.N3-1, K = J, K.ToUp(); k>0; --k, K.ToUp())
	F. template Do<FrontEl>(K);
      F.FinishSendFront(DownInd, UpInd);
      FUp.template Do<FrontEl>(K);


      I.ToRight();--i;

      UpInd.SetPos(ind.N1 - i, 0, ind.N3 -1);
      DownInd.SetPos(ind.N1 - i,0, -1);
      F.StartSendLeft(DownInd, UpInd);
      for(j = ind.N2, J=I; j; --j, J.ToBack())  // right plane
	{ 
	  //	  FDown.template Do<LeftEl>(J); // we don't need this now, we'll compute this later
	  for(k = ind.N3-1, K = J, K.ToUp(); k>0; --k, K.ToUp())
	    {
	      F. template Do<LeftEl>(K);
	    }
	}
      F.FinishSendLeft(DownInd, UpInd);
      for(j = ind.N2, J=I; j; --j, J.ToBack())  // bottom line, this is the later computation itself
	FDown.template Do<LeftEl>(J);
    }
}
template<class Int, class Functor, class FunctorDown, class FunctorUp>
void ForAllColumnwiseM2Down<Int, Functor, FunctorDown, FunctorUp>::
 operator()(Index3dBase<Int> const &ind, 
	    Functor &F, FunctorDown &FDown, FunctorUp &FUp)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), K(ind), UpInd(ind), DownInd(ind), Jtmp(ind);
  I.SetPos(ind.N1, ind.N2, ind.N3-1);
  
  for(i = ind.N1;i;)
    {

      // left columns
      J.SetPos(i,ind.N2-1, -1);
      DownInd.SetPos(i, 0, -1);
      UpInd.SetPos(i, 0, ind.N3-1);
      for(j = ind.N2; j; --j, J.ToFront())  
	FDown.template Do<LeftEl>(J);
      F.StartSendLeft(DownInd, UpInd);
      for(j = ind.N2, J=I, J.ToFront(); j; --j, J.ToFront())  
	{ 
	  for(k = ind.N3-1, K = J, K.ToDown(); k>0; --k, K.ToDown())
	    {
	      F. template Do<LeftEl>(K);
	    }
	  // FDown.template Do<LeftEl>(K); We do not whant this, we've computed it earlyer
	}
      F.FinishSendLeft(DownInd, UpInd);

      I.ToLeft();--i;


      // Back plane
      UpInd.SetPos(i, ind.N2, ind.N3-1);
      DownInd.SetPos(i, ind.N2, -1);
      J=I;
      FUp.template Do<FrontEl>(I);
      F.StartSendFront(DownInd, UpInd);
      for(k = ind.N3-1, K = J, K.ToDown(); k>0; --k, K.ToDown())
	F. template Do<FrontEl>(K);
      F.FinishSendFront(DownInd, UpInd);
      J.ToFront();
      
      for(j = ind.N2; j; --j, J.ToFront())
	{
	  UpInd.SetPos  (i, j-1, ind.N3-1);
	  DownInd.SetPos(i,j-1, -1);
	  K = J;
	  // middle column

	  FUp.template Do<DownEl>(J);
	  if(ind.N3 >1)
	    {
	      Jtmp.SetPos(i, j-1, 0);
	      F.template Do<DownEl>(Jtmp);
	    }
	  F.StartSendDown(DownInd, UpInd);
	  for(k = ind.N3-2, K.ToDown(); k>0; --k, K.ToDown())
	    F.template Do<DownEl>(K);
	  F.FinishSendDown(DownInd, UpInd);
	  K = J;
	  // Front column;


	  FUp.template Do<FrontEl>(J);
	  F.StartSendFront(DownInd, UpInd);
	  for(k = ind.N3-1, K.ToDown(); k>0; --k, K.ToDown())
	    F. template Do<FrontEl>(K);
	  F.FinishSendFront(DownInd, UpInd);
	}
      
    }

  for(j = ind.N2, J=I, J.ToFront(); j; --j, J.ToFront())  // right plane
    { 
      for(k = ind.N3-1, K = J, K.ToDown(); k>0; --k, K.ToDown())
	{
	  F. template Do<LeftEl>(K);
	}
      FDown.template Do<LeftEl>(K);
    }

}

template<class Int, class Functor, class FunctorDown, class FunctorUp>
void ForAllColumnwiseM4Up<Int, Functor, FunctorDown, FunctorUp>::
 operator()(Index3dBase<Int> const &ind, 
	    Functor &F, FunctorDown &FDown, FunctorUp &FUp)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), K(ind), UpInd(ind), DownInd(ind);
  I.SetPos(0, 0, -1);
  
  
  for(i = ind.N1;i;I.ToRight(),--i)
    {
      for(j = ind.N2, J=I; j; --j, J.ToBack())
	{
	  UpInd.SetPos(ind.N1 - i, ind.N2 - j, ind.N3 -2);
	  DownInd.SetPos(ind.N1 - i, ind.N2 - j, 0);

	  F.template Do<LeftEl>(DownInd);F.template Do<FrontEl>(DownInd);

	  F.template Do<LeftEl>(UpInd);
	  F.template Do<FrontEl>(UpInd);
	  UpInd.ToUp();
	  F.DoAll(UpInd);


	  K = DownInd;

	  DownInd.ToDown();

	  F.StartSendM4Up(DownInd, UpInd);

	  for(k = ind.N3-3,K.ToUp(); k>0; --k, K.ToUp())
	      F.DoAll(K);
	  F.template Do<DownEl>(K);


	  F.FinishSendM4Up(DownInd, UpInd);
	  F.SaveState(DownInd, UpInd);
	  F.template Do<LeftEl>(DownInd);
	  F.RestoreState(DownInd, UpInd);

	  UpInd.ToUp();
	  F.template Do<DownEl>(UpInd);

	  
	  DownInd.ToUp();
	  F.template Do<DownEl>(DownInd);

	}
      // back plane
      UpInd.SetPos(ind.N1 - i, ind.N2 - j, ind.N3 -1);
      DownInd.SetPos(ind.N1 - i, ind.N2 - j, -1);
      FUp.template Do<FrontEl>(UpInd);
      F.StartSendM4Up(DownInd, UpInd);
      for(k = ind.N3-1, K = J, K.ToUp(); k>0; --k, K.ToUp())
       	F.template Do<FrontEl>(K);
      F.FinishSendM4Up(DownInd, UpInd);
    }
  // right plane
  for(j = ind.N2, J=I; j; --j, J.ToBack())
    for(k = ind.N3+1, K = J; k>0; --k, K.ToUp())
      F.template Do<LeftEl>(K);
}


template<class Int, class Functor, class FunctorDown, class FunctorUp>
void ForAllColumnwise<Int, Functor, FunctorDown, FunctorUp>::
operator()(Index3dBase<Int> const &ind, 
	    Functor &F, FunctorDown &FDown, FunctorUp &FUp)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), K(ind);
  I.SetPos(0, 0, -1);
  
  //  std::cout << "height is " << ind.N3 << std::endl;
  
  for(i = ind.N1;i; --i, I.ToRight())
    {
      //left side
      for(j = ind.N2, J=I; j; --j, J.ToBack())
	{ 
	  FDown.template Do<LeftEl>(J);
	  for(k = ind.N3-1, K=J, K.ToUp();k>0; --k, K.ToUp()) 
	    F.template Do<LeftEl>(K);    
	}
      for(j = ind.N2, J=I; j; --j, J.ToBack())
	{
	  for(k = ind.N3-1, K = J, K.ToUp(); k>0; --k, K.ToUp())
	    F. template Do<FrontEl>(K);
	  FUp.template Do<FrontEl>(K);
	  for(k = ind.N3-1, K = J, K.ToUp(); k>0; --k, K.ToUp())
	    F.template Do<DownEl>(K);
	  FUp.template Do<DownEl>(K);
	}
      for(k = ind.N3-1, K = J, K.ToUp(); k>0; --k, K.ToUp())
	F. template Do<FrontEl>(K);
      FUp.template Do<FrontEl>(K);
    }
  for(j = ind.N2, J=I; j; --j, J.ToBack())  // right plane
    { 
      FDown.template Do<LeftEl>(J);
      for(k = ind.N3-1, K = J, K.ToUp(); k>0; --k, K.ToUp())
	{
	  F. template Do<LeftEl>(K);
	}
    }
}

template<class Int, class Functor, class FunctorDown, class FunctorUp>
void ForAllColumnwiseReverse<Int, Functor, FunctorDown, FunctorUp>::
operator()(Index3dBase<Int> const &ind, 
	    Functor &F, FunctorDown &FDown, FunctorUp &FUp)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), K(ind);
  I.SetPos(ind.N1, ind.N2, ind.N3-1);
  
  for(j = ind.N2, J=I, J.ToFront(); j; --j, J.ToFront())  // right plane
    { 
      for(k = ind.N3-1, K = J, K.ToDown(); k>0; --k, K.ToDown())
	{
	  F. template Do<LeftEl>(K);
	}
      FDown.template Do<LeftEl>(K);
    }
  I.ToLeft();
  for(i = ind.N1;i; --i, I.ToLeft())
    {
      J=I;
      // Back plane
      FUp.template Do<FrontEl>(I);
      for(k = ind.N3-1, K = J, K.ToDown(); k>0; --k, K.ToDown())
	F. template Do<FrontEl>(K);
      J.ToFront();
      
      for(j = ind.N2; j; --j, J.ToFront())
	{
	  K = J;
	  // middle column
	  FUp.template Do<DownEl>(J);
	  for(k = ind.N3-1, K.ToDown(); k>0; --k, K.ToDown())
	    F.template Do<DownEl>(K);
	  K = J;
	  // Front column;
	  FUp.template Do<FrontEl>(J);
	  for(k = ind.N3-1, K.ToDown(); k>0; --k, K.ToDown())
	    F. template Do<FrontEl>(K);
	}
      
      // left columns
      for(j = ind.N2, J=I, J.ToFront(); j; --j, J.ToFront())  // right plane
	{ 
	  for(k = ind.N3-1, K = J, K.ToDown(); k>0; --k, K.ToDown())
	    {
	      F. template Do<LeftEl>(K);
	    }
	  FDown.template Do<LeftEl>(K);
	}
    }
}


template<class Int, class Functor, class FunctorDown, class FunctorUp>
void ForAllColumnwiseM4Down<Int, Functor, FunctorDown, FunctorUp>::
operator()(Index3dBase<Int> const &ind, 
	    Functor &F, FunctorDown &FDown, FunctorUp &FUp)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), K(ind), UpInd(ind), DownInd(ind);
  I.SetPos(ind.N1, ind.N2, ind.N3);
  
  for(j = ind.N2, J=I, J.ToFront(); j; --j, J.ToFront())  // right plane
    { 
      for(k = ind.N3+1, K = J, K.ToDown(); k>0; --k, K.ToDown())
	{
	  F. template Do<LeftEl>(K);
	}
    }
  I.ToLeft();
  for(i = ind.N1;i; --i, I.ToLeft())
    {
      J=I;
      // Back plane
      UpInd.SetPos(i-1, ind.N2, ind.N3-1);
      DownInd.SetPos(i-1, ind.N2, -1);

      FUp.template Do<FrontEl>(UpInd);
      F.StartSendM4Down(DownInd, UpInd);
      
      for(k = ind.N3-1, K = UpInd, K.ToDown(); k>0; --k, K.ToDown())
	F. template Do<FrontEl>(K);
      J.ToFront();
      F.FinishSendM4Down(DownInd, UpInd);

      for(j = ind.N2; j; --j, J.ToFront())
	{
	  UpInd.SetPos(i-1, j-1, ind.N3-1);
	  DownInd.SetPos(i-1, j-1, 0);

	  F.template Do<DownEl>(UpInd);
	  F.template Do<DownEl>(DownInd);DownInd.ToDown();

	  F.StartSendM4Down(DownInd, UpInd);
	  for(k = ind.N3-2, K = UpInd, K.ToDown(); k>0; --k, K.ToDown())
	    F. DoAll(K);
	  F.template Do<FrontEl>(K);F.template Do<LeftEl>(K);

	  F.FinishSendM4Down(DownInd, UpInd);
	  F.template Do<FrontEl>(UpInd);
	  F.template Do<FrontEl>(DownInd);
	  F.template Do<LeftEl>(DownInd);
	  F.template Do<LeftEl>(UpInd);
	  
	}
    }
}


//<Int, Functor, FunctorDown, FunctorUp>::
template<class Int, class Functor>
void ForAllMiddle(Index3dBase<Int> const &ind, 
	    Functor &F)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), K(ind);
  I.SetPos(0, 0, 1);

  for(i = ind.N1;i; --i, I.ToRight())
  {
    for(j = ind.N2, J=I; j; --j, J.ToBack())
    { 
      for(k = ind.N3-2, K = J; k>0; --k, K.ToUp()) // doing middle
      {
        F.DoAll(K);
      }
    }
    for(k = ind.N3-2, K = J; k>0; --k, K.ToUp())  // back plane
    {
      F.template Do<FrontEl>(K);
    }
  }
  for(j = ind.N2, J=I; j; --j, J.ToBack())  // right plane
  { 
    for(k = ind.N3-2, K = J; k>0; --k, K.ToUp())
    {
      F.template Do<LeftEl>(K);
    }
  }

}
template<class Int, class Functor>
void ForAllNonMiddle(Index3dBase<Int> const &ind,
	    Functor &F)
{
  Int i, j;//, k;
  Index3dBase<Int> I(ind), J(ind);
  Index3dBase<Int> Iu(ind),Ju(ind);
  I.SetPos(0, 0, -1);
  Iu.SetPos(0,0,0);
  

  for(i = ind.N1;i; --i, I.ToRight(),Iu.ToRight())
  {
    for(j = ind.N2, J=I,Ju=Iu; j; --j, J.ToBack(),Ju.ToBack())
    {
	F.DoAll(J);
	F.DoAll(Ju);
    }
    F.template Do<FrontEl>(J);
    F.template Do<FrontEl>(Ju);
  }
  for(j = ind.N2, J=I, Ju = Iu; j; --j,
	  J.ToBack(), Ju.ToBack())  // right plane
  {
    F.template Do<LeftEl>(J);
    F.template Do<LeftEl>(Ju);
  }

  if(ind.N3>1)
  {
      I.SetPos(0, 0, ind.N3-1);
      Iu.SetPos(0,0,ind.N3);
      
      
      for(i = ind.N1;i; --i, I.ToRight(),Iu.ToRight())
      {
	  for(j = ind.N2, J=I,Ju=Iu; j; --j, J.ToBack(),Ju.ToBack())
	  {
	      F.DoAll(J);
	      F.DoAll(Ju);
	  }
	  F.template Do<FrontEl>(J);
	  //	  F.template Do<FrontEl>(Ju);
      }
      for(j = ind.N2, J=I, Ju = Iu; j; --j,
	      J.ToBack(), Ju.ToBack())  // right plane
      {
	  F.template Do<LeftEl>(J);
	  //	  F.template Do<LeftEl>(Ju);
      }
  }else
  {
      I.SetPos(0, 0, ind.N3);
      
      
      for(i = ind.N1;i; --i, I.ToRight())
      {
	  for(j = ind.N2, J=I; j; --j, J.ToBack())
	  {
	      F.DoAll(J);
	  }
	  F.template Do<FrontEl>(J);
      }
      for(j = ind.N2, J=I; j; --j, J.ToBack())  // right plane
      {
	  F.template Do<LeftEl>(J);

      }
  }

}


template<class Int, class Functor>
void DoForEach(Index3dBase<Int> const &ind, Functor &F, bool dolast, Int m1)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), K(ind);
  I.SetPos(0, 0, m1);
  
  for(i = ind.N1;i; --i, I.ToRight())
  {
    for(j = ind.N2, J=I; j; --j, J.ToBack())
    {
      for(k = ind.N3-m1*2, K = J; k; --k, K.ToUp()) // doing middle
      {
        F. DoAll(K);
      }
      if (dolast) F.template Do<DownEl>(K);                             // top plane
    }
    for(k = ind.N3-m1*2, K = J; k; --k, K.ToUp())  // back plane
    {
      F. template Do<FrontEl>(K);
    }
  }
  for(j = ind.N2, J=I; j; --j, J.ToBack())  // right plane
  {
    for(k = ind.N3-m1*2, K = J; k; --k, K.ToUp())
    {
      F. template Do<LeftEl>(K);
    }
  }
}
template<class Int, class Functor>
void DoForEachSpecial(Index3dBase<Int> const &ind, Functor &F, bool dolast)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), K(ind);
  I.SetPos(0, 0, 0);
  
  
  for(i = ind.N1;i; --i, I.ToRight())
  {

    for(j = ind.N2, J=I; j; --j, J.ToBack())  // Lefts
    {
      for(k = ind.N3, K = J; k; --k, K.ToUp())
      {
        F. template Do<LeftEl>(K);
      }
    }
    F.SRLeftUp(ind.N1 - i);
    for(j = ind.N2, J=I; j; --j, J.ToBack())                  // middle plane
    {
      for(k = ind.N3, K = J; k; --k, K.ToUp())
      {
        F. template Do<FrontEl>(K);
        F. template Do<DownEl> (K);
      }
      if (dolast) F.template Do<DownEl>(K);                             // top plane // to be checked!!!!!!!!!!!!!!!!!!!!!!!!111
    }
    for(k = ind.N3, K = J; k; --k, K.ToUp())  // Last Front Column
    {
      F. template Do<FrontEl>(K);
    }
    F.SRFrontDownUp(ind.N1 - i);
  }
  for(j = ind.N2, J=I; j; --j, J.ToBack())  // right plane
  {
    for(k = ind.N3, K = J; k; --k, K.ToUp())
    {
      F. template Do<LeftEl>(K);
    }
  }
}

template<class Int, class Functor>
void DoForEachSpecial3(Index3dBase<Int> const &ind, Functor &F, bool dolast)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), J1(ind), K(ind);
  I.SetPos(0, 0, 0);
  
  F.NR();
  
  for(i = ind.N1;i; --i, I.ToRight())
  {

    for(j = ind.N2, J=I; j; --j, J.ToBack())  // Lefts
    {
      for(k = ind.N3-1, K = J; k; --k, K.ToUp())
      {
        F. template Do<LeftEl>(K);
      }
    }
    F.SRFrontDownUpFinish(ind.N1 -(i +1));
    J1.SetPos(ind.N1-i, 0, ind.N3-1);
    for(j = ind.N2;j;--j, J1.ToBack())
      F.template Do<LeftEl>(J1);



    F.SRLeftUp(ind.N1 - i);

    for(j = ind.N2, J=I, J1=I; j; --j, J.ToBack())                  // middle plane
    {
      for(k = ind.N3-1, K = J, K.ToUp(); k; --k, K.ToUp())
      {
        F. template Do<FrontEl>(K);
        F. template Do<DownEl> (K);
      }
      if (dolast) F.template Do<DownEl>(K);                             // top plane // to be checked!!!!!!!!!!!!!!!!!!!!!!!!111
    }
    for(k = ind.N3, K = J; k; --k, K.ToUp())  // Last Front Column
    {
      F. template Do<FrontEl>(K);
    }
    F.SRLeftUpFinish(ind.N1 - i);  //wait
    for(j = ind.N2;j;--j, J1.ToBack())
    {
       F.template Do<FrontEl>(J1);
       F.template Do<DownEl>(J1);
    }
    F.template Do<FrontEl>(J1);
    

    F.SRFrontDownUp(ind.N1 - i);
  }
    F.SRFrontDownUpFinish(ind.N1 -(i +1));
  for(j = ind.N2, J=I; j; --j, J.ToBack())  // right plane
  {
    for(k = ind.N3, K = J; k; --k, K.ToUp())
    {
      F. template Do<LeftEl>(K);
    }
  }
}



template<class Int, class Functor>
void DoForEachSpecial2(Index3dBase<Int> const &ind, Functor &F, bool dolast)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), K(ind);
  I.SetPos(0, 0, -1);
  
  for(i = ind.N1;i; --i, I.ToRight())
  {

    for(j = ind.N2, J=I; j; --j, J.ToBack())  // Lefts
    {
      for(k = ind.N3+2, K = J; k; --k, K.ToUp())
      {
        F. template Do<LeftEl>(K);
      }
    }
    for(j = ind.N2, J=I; j; --j, J.ToBack())                  // middle plane
    {
      for(k = ind.N3+2, K = J; k; --k, K.ToUp())
      {
        F. template Do<FrontEl>(K);
        if (k == ind.N3+2) F. template Do<DownEl> (K);
          else
            F.template Do<DownEl>(K);
      }
      if (dolast) F.template Do<DownEl>(K);
    }
    for(k = ind.N3+2, K = J; k; --k, K.ToUp())  // Last Front Column
    {
      F. template Do<FrontEl>(K);
    }
  }
  for(j = ind.N2, J=I; j; --j, J.ToBack())  // right plane
  {
    for(k = ind.N3+2, K = J; k; --k, K.ToUp())
    {
      F. template Do<LeftEl>(K);
    }
  }
}

template<class Int, class Functor>
void DoForEachSpecialReverse(Index3dBase<Int> const &ind, Functor &F, bool dolast)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), J1(ind), K(ind);
  I.SetPos(ind.N1, ind.N2, ind.N3);


  for(j = ind.N2, J=I, J.ToFront(); j; --j, J.ToFront())  // right plane
  {
    for(k = ind.N3, K = J, K.ToDown(); k; --k, K.ToDown())
    {
      F. template Do<LeftEl>(K);
    }
  }
  

  for(i = ind.N1, I.ToLeft();i; --i, I.ToLeft())
  {
    F. SRLeftDown(i-1);
    F. SRLeftDownFinish(i-1);

    for(k = ind.N3-1, J = I, K = J, K.ToDown(); k; --k, K.ToDown())       // middle plane
    {
      F. template Do<FrontEl>(K);
    }
    for(j = ind.N2, J=I, J.ToFront(); j; --j, J.ToFront())
    {
      if (dolast) F.template Do<DownEl>(J);                             // top plane
      for(k = ind.N3-1, K = J, K.ToDown(); k; --k, K.ToDown())
      {
        F. template Do<DownEl>(K);
        F. template Do<FrontEl>(K);
      }
    }

    J1.SetPos(i-1, ind.N2, 0);
    F.template Do<FrontEl>(J1);
    for(j = ind.N2, J1.ToFront();j; --j, J1.ToFront())
    {
      F.template Do<DownEl>(J1);
      F.template Do<FrontEl>(J1);
    }




    F. SRFrontDownDown(i-1);

    for(j = ind.N2, J=I, J.ToFront(); j; --j, J.ToFront())           // Left
    {
      for(k = ind.N3-1, K = J, K.ToDown(), K.ToDown(); k; --k, K.ToDown())
      {
        F. template Do<LeftEl>(K);
      }
    }
    F. SRFrontDownDownFinish(i-1);
    J1.SetPos(i-1, 0, ind.N3-1);
    for(j = ind.N2;j; --j, J1.ToBack())
      F.template Do<LeftEl>(J1);

  }
}



template<class Int, class Functor>
void DoForEachExtra(Index3dBase<Int> const &ind, Functor &F)
{

    Int i, j, k;
    Index3dBase<Int> IU(ind), JU(ind);
    Index3dBase<Int> IU1(ind), JU1(ind);

    Index3dBase<Int> ID(ind), JD(ind);
    Index3dBase<Int> ID1(ind), JD1(ind);
    ID.SetPos(0, 0, -1);
    ID1.SetPos(0, 0, 0);
    IU.SetPos(0, 0, ind.N3-1);
    IU1.SetPos(0, 0, ind.N3);
  
    for(i = ind.N1
          ;i;
        --i, IU.ToRight(), ID.ToRight(), IU1.ToRight(), ID1.ToRight())
    {
      for(j = ind.N2, JD=ID,JU=IU, JU1=IU1, JD1 = ID1
          ; j;
         --j, JD.ToBack(), JU.ToBack(), JU1.ToBack(), JD1.ToBack())
      {
        F.template Do<DownEl>(JU1);//
     
        F.DoAll(JU);
        F.DoAll(JD);
        F.DoAll(JD1);
      }
        F.template Do<FrontEl>(JD);
        F.template Do<FrontEl>(JD1);
        F.template Do<FrontEl>(JU);
    }

      for(j = ind.N2, JD=ID,JU=IU, JU1=IU1, JD1 = ID1
          ; j;
         --j, JD.ToBack(), JU.ToBack(), JU1.ToBack(), JD1.ToBack())
      {
        F.template Do<LeftEl>(JD);
        F.template Do<LeftEl>(JD1);
        F.template Do<LeftEl>(JU);
      }

}


template<class Int, class Functor>
void CDoForEach(Index3dBase<Int> const &ind, Functor &F)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), K(ind);
  I.SetPos(0, 0, 0);
  
  for(i = ind.N1;i; --i, I.ToRight())
  {

    for(j = ind.N2, J=I; j; --j, J.ToBack())  // Lefts
    {
      for(k = ind.N3, K = J; k; --k, K.ToUp())
      {
        F. template Do<LeftEl>(K);
      }
    }
    for(j = ind.N2, J=I; j; --j, J.ToBack())
    {
      for(k = ind.N3, K = J; k; --k, K.ToUp()) // doing Front Column
      {
        F. template Do<FrontEl>(K);
      }
      for(k = ind.N3, K = J; k; --k, K.ToUp()) // doing middle Column
      {
        F. template Do<DownEl>(K);
      }
      F.template Do<DownEl>(K);                             // top plane
    }
    for(k = ind.N3, K = J; k; --k, K.ToUp())  // Last Front Column
    {
      F. template Do<FrontEl>(K);
    }
  }
  for(j = ind.N2, J=I; j; --j, J.ToBack())  // right plane
  {
    for(k = ind.N3, K = J; k; --k, K.ToUp())
    {
      F. template Do<LeftEl>(K);
    }
  }
}

template<class Int, class Functor>
void DoForEachReverse(Index3dBase<Int> const &ind, Functor &F)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), K(ind);
  I.SetPos(ind.N1, ind.N2, ind.N3);

  for(j = ind.N2, J=I, J.ToFront(); j; --j, J.ToFront())  // right plane
  {
    for(k = ind.N3, K = J, K.ToDown(); k; --k, K.ToDown())
    {
      F. template Do<LeftEl>(K);
    }
  }
  

  for(i = ind.N1, I.ToLeft();i; --i, I.ToLeft())
  {
    for(k = ind.N3, J = I, K = J, K.ToDown(); k; --k, K.ToDown())  // back plane
    {
      F. template Do<FrontEl>(K);
    }
    for(j = ind.N2, J=I, J.ToFront(); j; --j, J.ToFront())
    {
      F.template Do<DownEl>(J);                             // top plane
      for(k = ind.N3, K = J, K.ToDown(); k; --k, K.ToDown()) // doing middle
      {
        F. DoAll(K);
      }
    }
  }
}
template<class Int, class Functor>
void CDoForEachReverse(Index3dBase<Int> const &ind, Functor &F)
{
  Int i, j, k;
  Index3dBase<Int> I(ind), J(ind), K(ind);
  I.SetPos(ind.N1, ind.N2, ind.N3);

  for(j = ind.N2, J=I, J.ToFront(); j; --j, J.ToFront())  // right plane
  {
    for(k = ind.N3, K = J, K.ToDown(); k; --k, K.ToDown())
    {
      F. template Do<LeftEl>(K);
    }
  }
  

  for(i = ind.N1, I.ToLeft();i; --i, I.ToLeft())
  {
    for(k = ind.N3, J = I, K = J, K.ToDown(); k; --k, K.ToDown())  // back plane
    {
      F. template Do<FrontEl>(K);
    }

    for(j = ind.N2, J=I, J.ToFront(); j; --j, J.ToFront())
    {
      F.template Do<DownEl>(J);                             // top plane
      for(k = ind.N3, K = J, K.ToDown(); k; --k, K.ToDown()) // doing middle
      {
        F. template Do<DownEl>(K);
      }
      for(k = ind.N3, K = J, K.ToDown(); k; --k, K.ToDown()) // doing middle
      {
        F. template Do<FrontEl>(K);
      }
    }
    for(j = ind.N2, J=I, J.ToFront(); j; --j, J.ToFront())  // right plane
    {
      for(k = ind.N3, K = J, K.ToDown(); k; --k, K.ToDown())
      {
        F. template Do<LeftEl>(K);
      }
    }

  }
}

#endif
