#define NDEBUG
#define FAST_POLY1
#include <string>
#include <sstream>
#include <iostream>
#include <fstream>
#include <cstdlib>
#include "elastgen.hpp"
#include "sysgen.hpp"
#include "sparcemmat.hpp"
#include "small.h"
#include "mirorer.h"

#include "sl.h"
#include "queue.h"
#include "mpt.cpp"
string ofname;

typedef double Double;
typedef  ElastDiscretizator<double>::Poly3d Poly3d;


int I, J;
int what; //0,1,2 - dis strejns stresses
int pts;
int K;
int N1, N2, N3;

ElastDiscretizator<double> *pdis;


void DoLine(const std::string &line, int bi, int bj, std::vector<double> &vec)
{
  SmallVector<Double, 3> displacements;
  SmallMatrix<Double, 3, 3> stresses;
  SmallMatrix<Double, 3, 3> strains;
  SmallVector<Double, 18> small_x;
  istringstream is(line);
  double lambda, mu;
  is >> lambda >> mu;
  for(int i = 0; i < 18; ++i)
  {
    is >> small_x[i];
    if (isnan(small_x[i]))
      cout << "nanann\n" << std::endl;
  }
  for (int ii = 0; ii < K; ++ii)
    for(int jj = 0; jj < K; ++jj)
      for(int kk = 0; kk <K; ++kk)
      {
	SmallVector<double, 3> p;
	p[0] = p[1] = p[2] = 2./K;
	p[0] = -1 + p[0]*0.5 + ii*p[0];
	p[1] = -1 + p[1]*0.5 + jj*p[1];
	p[2] = -1 + p[2]*0.5 + kk*p[2];
	pdis->CalculateStressesAndStrains(displacements,
					  stresses,
					  strains, 
					  small_x,
					  lambda,
					  mu,p);
	double val;
	if (what ==0)
	  val = displacements[I];
	if (what ==1)
	  val = strains[I][J];
	if (what ==2)
	  val = stresses[I][J];
	if (isnan(val))
	  cout << "nanv\n" << std::endl;
	  
	vec[(bi*K + ii)*N2*K + ( bj*K + jj)*K + kk] = val;
      }
}

struct Slice
{
  std::vector<double> vec;
  int slicen;
  atomic_count num_to_be_ready;

  Slice(int s, long n):slicen(s),num_to_be_ready(n),vec(N1*N2*K*K*K)
  {
  }
  bool Done()
  {
    //    std::cout << "ntbr " << num_to_be_ready << std::endl;
    return --num_to_be_ready == 0;
  }
};

struct PointPortion
{
  std::string string;
  int i, j;
};


struct Portion
{
  std::vector<PointPortion> vec;
  Slice *slice;
};

mpt::Waiter<int, Slice*> waiter;
mpt::TSQueue<Portion*> que(200);


void WorkerThread(int num)
{

  //  std::vector<double> vec1(N1*N2*K*K*K);
  while(true)
  {
    Portion *p ;
    que.Pop(p);
    Slice *s = p->slice;
    for(int i = 0; i < N2; ++i)
    {
      DoLine(p->vec[i].string, p->vec[i].i, p->vec[i].j, s->vec);
    }
    delete p;
    if (s->Done())
    {
      waiter.Signal(s->slicen, s);
    }
  }  
}

void OutputThread()
{
  ofstream of(ofname.c_str(), ios::binary);
  for (int k = 0; k < N3; ++k)    
  {
    //    std::cout << "Waiting " << k << std::endl;
    Slice * s = waiter.Wait(k);
    for(int kk = 0; kk < K; ++kk)
      for(int i = 0; i < N1*K; ++i)
	for(int j = 0; j < N2*K; ++j)
	{
	  of.write((char*)&(s->vec[i*N2*K+ j*K +kk]), sizeof(double));
	}
    delete s;
  }
}

void InputThread(ifstream &file)
{
  const int TC = 12;
  mpt::Thread wt[TC], ot;
  ot = mpt::StartThread(std::tr1::bind(OutputThread));
  for(int i = 0; i < TC; ++i)
    wt[i] = mpt::StartThread(std::tr1::bind(WorkerThread,i));
  
  for(int k = 0; k < N3; ++k)
  {
    Slice *s = new Slice(k, N1);
    for(int i = 0; i <N1; ++i)
    {
      Portion *p = new Portion;
      p->slice = s;
      p->vec.resize(N2);
      for(int j = 0; j < N2; ++j)
      {
	std::string &str = p->vec[j].string;
	p->vec[j].j = j;
	p->vec[j].i = i;
	do{
	  getline(file, str);
	}
	while((str.length()>0)&&(str[0]=='#'));
      }
      //      std::cout << "i " << i << std::endl;
      que.Push(p);
    }
  }
  ot.Join();
  for(int i = 0; i < TC; ++i)
    wt[i].Stop();    
}





void DoLine1(const std::string &line, Poly3d &p)
{
  SmallVector<Poly3d, 3> displacements;
  SmallMatrix<Poly3d, 3, 3> stresses;
  SmallMatrix<Poly3d, 3, 3> strains;
  SmallVector<Double, 18> small_x;
  
  istringstream is(line);
  double lambda, mu;
  is >> lambda >> mu;
  for(int i = 0; i < 18; ++i)
  {
    is >> small_x[i];
    if (isnan(small_x[i]))
      cout << "nanann\n" << std::endl;
  }
  pdis->CalculateStressesAndStrains(displacements,
					  stresses,
					  strains, 
					  small_x,
					  lambda,
					  mu);
  Poly3d &val = p;
  if (what ==0)
    val = displacements[I];
  if (what ==1)
    val = strains[I][J];
  if (what ==2)
    val = stresses[I][J];
}

struct Portion1
{
  std::vector<std::string> strings;
  std::vector<Poly3d> polys;
  Portion1(size_t n):strings(n), polys(n){}
  int i,  k;
};
mpt::Waiter<int, Portion1*> waiter1;
mpt::TSQueue<Portion1*> que1(200);

void WorkerThread1(int i)
{
  while(1)
  {
    Portion1 *p;
    que1.Pop(p);
    for(int j = 0; j < N2; ++j)
    {
      DoLine1(p->strings[j], p->polys[j]);
    }
    waiter1.Signal(p->k*N1 + p->i, p);    
  }
}

void OutputThread1()
{
  ofstream of(ofname.c_str());
  of.precision(16);
  of.setf(ios::scientific);
  if (K==0)
    of << "#" << N1 << "\t" << N2 << "\t" << N3 << "\n";
  for (int k = 0; k < N3; ++k)    
    for(int i = 0; i < N1; ++i)
    {
      Portion1 *p = waiter1.Wait(N1*k + i);
      for(int j = 0; j < N2; ++j)
      {
	of << p->polys[j].Size();
	for(int i =0; i < p->polys[j].Size(); ++i)
	  of << " " << p->polys[j][i] ;
	of << "\n";
	
      }
      delete p;
    }
}



void InputThread1(ifstream &file)
{
  const int TC = 3;
  mpt::Thread wt[TC], ot;
  ot = mpt::StartThread(std::tr1::bind(OutputThread1));
  for(int i = 0; i < TC; ++i)
    wt[i] = mpt::StartThread(std::tr1::bind(WorkerThread1,i));
  
  for(int k = 0; k < N3; ++k)
  {
    for(int i = 0; i <N1; ++i)
    {
      Portion1 *p = new Portion1(N2);
      p->i = i;
      p->k = k;
      for(int j = 0; j < N2; ++j)
      {
	std::string &str = p->strings[j];
	do{
	  getline(file, str);
	}
	while((str.length()>0)&&(str[0]=='#'));
      }
      que1.Push(p);
    }
  }
  ot.Join();
  for(int i = 0; i < TC; ++i)
    wt[i].Stop();    
}




int main(int argc, char **argv)
{
  if (argc<2)
    return 1;
  std::string fname = argv[1];
  K = 1;
  what = 2;
  I = J = 2;
  if (argc>=3)
    K = std::atol(argv[2]);
  if (argc>=4)
    what = std::atol(argv[3]);
  if (argc>=5)
    I = std::atol(argv[4]);
  if (argc>=6)
    J = std::atol(argv[5]);

  ifstream file(fname.c_str());
  std::string str;
  getline(file, str);
  //  std::cout << str << std::endl;
  str = str.substr(1);
  istringstream is(str);
  int DT;
  is >> N1 >> N2 >> N3 >>DT ;
  pdis = new ElastDiscretizator<double>((DT==0)?MP:MV);

  
  ostringstream os;
  os << N1;
  if ((N1!=N2)||(N2!=N3)||(N1!=N3))
    os << "x" << N2 << "x" << N3;
  os << ((K>0)?".f." : ".g.")<< K << "." << what << "." << I <<J;
  ofname = os.str();
  if (K>0)
    InputThread(file);
  else
    InputThread1(file);
}
