#include "../small.h"
#include <vector>

extern "C" void efulrc_(double *A, double *EIGVEC, double *EIGVAL, double *WS,int *IWS,
	    const int &NP, const int &NR, const int &N, const int &NWS,
	     const int &IOPT1, const int &IOPT2, const int &IOPT3);

template<int I, int J>
void CalcEigen(SmallMatrix<double, I, J> const &mat,
	       SmallVector<double, I> &Vals,
	       SmallMatrix<double, I, J> &Vecs)
{
  const int NR = I;
  const int N = J;
  const int NWS = 4;
  const int NP = 4;
  std::vector<double> WS(NWS*NR);
  std::vector<int> IWS(N);
  std::vector<double> A(NR*N);
  std::vector<double> vecs(NP*NR*N);
  std::vector<double> vals(2*N);
  int IOPT1, IOPT2, IOPT3;

  IOPT1 = 3;
  IOPT2 = 2;
  IOPT3 = 1;
  //copy the matrix;
  for (int i = 0; i < I; ++i)
    for (int j = 0; j < J; ++j)
      A[j*I+i] = mat[i][j];
  efulrc_(&A[0], &vecs[0], &vals[0],
	 &WS[0], &IWS[0],
	  NP, NR, N, NWS, IOPT1, IOPT2, IOPT3);
  for (int i = 0; i < I; ++i)
    Vals[i] = vals[i*2];
  SmallVector<double, I> v;
  // for (int i = 0; i < NP*N*NR; ++i)
  //   std::cout << i << " " << vecs[i] << std::endl;
  for (int i = 0; i < NR; ++i)
  {
    for (int j = 0; j < N; ++j)
      Vecs[i][j] = vecs[2 + NP*( i +j*NR)];
  }      
    
}
// C
// C This subroutine computes the eigenvalues and eigenvectors of a
// C real, full, general matrix, using the QR algorithm of Francis,
// C as described by Martin, Peters, and Wilkinson in the Handbook
// C for Automatic Computation-Linear Algebra, volume II,
// C articles 14 and 15 (HACLA II/14 and HACLA II/15).  
// C
// C The eigenvalues and eigenvectors may be real or complex. 
// C Both left and right eigenvectors may be computed.  Each
// C eigenvector is normalized to have unit modulus, with the
// C real part of its first element nonnegative.  The eigenvalues
// C are sorted by their real parts; complex conjugate pairs
// C are ordered so that the eigenvalue with the negative imaginary
// C part comes first.  The vectors are ordered to correspond with
// C the eigenvalues.
// C
// C The matrix A is first transformed to upper Hessenberg form by a
// C direct transformation, as in HACLA II/13.  Then successive double QR
// C transformations are performed, using Householder transformations and
// C origin shifts.  The "active matrix" (matrix of as yet undetermined
// C eigenvalues) is deflated (reduced in size) as eigenvalues are
// C found in the lower right corner.  The eigenvalues occur either as the
// C last element on the diagonal of the active matrix, or as the two
// C real or complex conjugate eigenvalues of the last 2 X 2 sub-matrix
// C along the diagonal.
// C
// C Once the eigenvalues have been computed, the matrix is in nearly
// C triangular form (with at most block 2 X 2 matrices along the
// C principal diagonal), and the right eigenvectors are computed by
// C back substitution.  The eigenvectors of the nearly triangular matrix
// C are then multiplied by the accumulated transformation matrix and
// C thus transformed into the eigenvectors of the input matrix.
// C
// C Left eigenvectors of the input matrix are computed as the right
// C eigenvectors of the transpose of the input matrix.  Condition
// C numbers may then be computed from the two sets of eigenvectors.
// C
// C An accuracy test of the eigenvalues may also be performed by 
// C comparing the eigenvalues of the original matrix with those of its
// C transpose.  Theoretically the two sets should be the same.  In
// C practice, the difference between two theoretically identical
// C eigenvalues is about the same as the error in either of them.
// C
// C CALLING FORMAT:
// C
// C      CALL EFULRC(A,EIGVEC,EIGVAL,WS,IWS,NP,NR,N,NWS,
// C     1            IOPT1,IOPT2,IOPT3)
// C
// C PARAMETER LIST:
// C
// C      A  (Input)  Array of row dimension NR and column dimension at
// C                  least N, containing the N by N input matrix whose
// C                  eigenvalues (and, optionally, eigenvectors) are to
// C                  be computed.
// C         (Output) The N X N accumulated transformation matrix 
// C                  S which reduces the matrix B to nearly triangular
// C                  form T, by the transformation
// C                   -1
// C                  S * B * S = T, or B * S = S * T,
// C
// C                  if eigenvectors are computed.  (Otherwise the
// C                  contents of A have no particular significance.)
// C                  If IOPT1 = 1, B is the input matrix A; 
// C                  if IOPT1 = 2 or 3, B is the transpose of A.
// C
// C EIGVEC  (Output) Array of dimensions (NP,NR,N), containing the
// C                  eigenvectors, if eigenvectors are computed.  See 
// C                  the description below.
// C
// C EIGVAL  (Output) REAL*8 Array of row dimension 2 and column 
// C                  dimension at least N, or COMPLEX*16 vector of
// C                  length N, which contains the eigenvalues computed by
// C                  this subroutine.  If EIGVAL is a REAL*8 array, then
// C                  element EIGVAL(1,I) contains the real part of the 
// C                  I'th eigenvalue, while element EIGVAL(2,I) contains
// C                  the imaginary part of the I'th eigenvalue.  If 
// C                  EIGVAL is a COMPLEX*16 array, then element EIGVAL(I)
// C                  contains the I'th eigenvalue in COMPLEX*16 format.
// C
// C     WS  (Input)  Work space array of row dimension NR (exactly) and 
// C                  column dimension NWS.
// C         (Output) If IOPT1 = 3, or IOPT3 = 1, then WS(I,1) is the 
// C                  relative accuracy of the I'th eigenvalue.
// C                  If IOPT1 = 3 and IOPT3 = 1, then WS(I,2) is the 
// C                  condition number (sensitivity) of the I'th
// C                  eigenvalue.
// C
// C    IWS  (Input)  Integer array of length at least N, used as 
// C                  temporary work space.
// C
// C     NP  (Input)  First (page) dimension of EIGVEC, exactly as
// C                  specified in the calling program.  It must be
// C                  at least as large the values given in the table
// C                  below.
// C
// C     NR  (Input)  Row dimension of A, second dimension of EIGVEC, and
// C                  row dimension of WS exactly as specified in the
// C                  calling program.
// C                  NR must be greater than or equal to N.
// C
// C      N  (Input)  Number of rows and columns in the input matrix.
// C
// C    NWS  (Input)  Second dimension of WS.  It must be at least as
// C                  large as the value given in the table below.
// C
// C  IOPT1  (Input)  Eigenvalue-eigenvector option parameter.  The
// C                  value of IOPT1 determines what calculation is
// C                  to be performed, as described below.
// C
// C  IOPT2  (Input)  Eigenvector storage parameter.  The storage
// C                  format of the output vectors in the eigenvector
// C                  array (compressed or uncompressed) is specified by
// C                  this parameter, as described below.
// C
// C  IOPT3  (Input)  Accuracy parameter.  This parameter determines which
// C                  of the accuracy tests (estimated errors or condition
// C                  numbers) are to be performed, as described below.
// C
// C======================================================================
// C        Table of Input Parameters and Calculations Performed         |
// C_____________________________________________________________________|
// C                                                                     |
// C                               Eigen-    Relative   Condition        |
// C IOPT1  IOPT2  IOPT3  NP  NWS  vectors*  Accuracy   Numbers          |
// C_____________________________________________________________________|
// C                                                                     |
// C   0      0      0    0    2   None        No          No            |
// C   0      0      1    1    4   None        Yes         No            |
// C                                                                     |
// C   1      1      0    1    2   Right C     No          No            |
// C   1      1      1    1    4   Right C     Yes         No            |
// C   1      2      0    2    2   Right U     No          No            |
// C   1      2      1    2    4   Right U     Yes         No            |
// C                                                                     |
// C   2      1      0    1    2   Left C      No          No            |
// C   2      1      1    1    4   Left C      Yes         No            |
// C   2      2      0    2    2   Left U      No          No            |
// C   2      2      1    2    4   Left U      Yes         No            |
// C                                                                     |
// C   3      1      0    2    4   Both C      Yes         No            |
// C   3      1      1    2    4   Both C      Yes         Yes           |
// C   3      2      0    4    4   Both U      Yes         No            |
// C   3      2      1    4    4   Both U      Yes         Yes           |
// C                                                                     |
// C======================================================================
// C
// C * The C or U refer to vectors stored in compressed or uncompressed
// C   format, respectively.  The type of vector(s) computed (none, right,
// C   left, or both) is as indicated.
// C
// C No vectors:      If eigenvectors and the accuracy test are not to be
// C                  computed, set NP = 0, so that no storage space is
// C                  allocated for the vector array.
// C
// C Either Left or   If all eigenvalues (and hence the eigenvectors) 
// C Right Vectors:   are known to be real, then space need
// C                  not be allocated for the imaginary parts of the 
// C                  eigenvectors, and we may set IOPT2 = 1.  Then
// C                  column J of EIGVEC contains the J'th eigenvector;
// C                  i.e., element EIGVEC(1,I,J) is the I'th element
// C                  of the J'th eigenvector.
// C
// C                  If any of the eigenvalues is complex, however, and
// C                  sufficient memory is available, then IOPT2 should be
// C                  set to 2, so the program will have sufficient space
// C                  for normal complex eigenvector storage.  In the
// C                  normal "uncompressed" format, elements EIGVEC(1,I,J)
// C                  and EIGVEC(2,I,J) are the real and imaginary parts
// C                  of the I'th element of the J'th eigenvector.  Note
// C                  that complex eigenvalues and vectors always occur
// C                  in complex conjugate pairs for real input matrices.
// C
// C                  If IOPT2 = 1 and some of the eigenvalues are
// C                  complex, then the complex eigenvectors are
// C                  stored in a compressed mode.  Real eigenvector
// C                  J is in column J of EIGVEC, and its I'th element
// C                  is EIGVEC(1,I,J).  If the J'th vector is the 
// C                  first of a complex conjugate pair,
// C                  however, then the real part of eigenvector J
// C                  is in column J of EIGVEC, and its imaginary 
// C                  part is in column J + 1.  Then complex
// C                  eigenvector J + 1 is the complex conjugate of
// C                  eigenvector J, and is not stored separately.
// C                  Thus the real and imaginary parts of the J'th
// C                  eigenvector are EIGVEC(1,I,J) and EIGVEC(1,I,J+1),
// C                  while the real and imaginary parts of the J+1'st
// C                  eigenvector are EIGVEC(1,I,J) and -EIGVEC(1,I,J+1).
// C
// C                  The choice IOPT2 = 2 is always safe, though it 
// C                  may waste space if the eigenvectors are real.
// C
// C Both Left and    If IOPT2 = 1, the left vectors are stored in
// C Right Vectors:   compressed format as elements EIGVEC(1,I,J),
// C                  while the right vectors are stored in compressed
// C                  format as elements EIGVEC(2,I,J).
// C
// C                  If IOPT2 = 2, the left vectors are uncompressed,
// C                  with real parts in EIGVEC(1,I,J) and imaginary
// C                  parts EIGVEC(2,I,J).  Similarly, the uncompressed
// C                  right vectors have their real and imaginary parts
// C                  stored as EIGVEC(3,I,J) and EIGVEC(4,I,J).
// C
// C Required routines: EBALNF, EINTF, EHESSF, EQRF, EHOUSF, EVECRF, 
// C                    EBAKRF, CDIV, EBAKBF, ENORMF, VSORT2,
// C                    UERCTL, UERDMP, UERPRT, NACOM
// C
// C Revision history:
// C
// C  Kevin Thompson     November 26, 1987     Version 3.0
// C
// C  Kevin Thompson     June 8, 1991          Version 3.0
// C
// C   Replaced numerous loops by calls to BLAS routines.  Also changed
// C   call to EVECRF to match new parameter list, using WS for scratch
// C   space.
// C
// C
// C**********************************************************************
// C
//       IMPLICIT REAL*8 (A-H,O-Z)
//       DIMENSION A(NR,N),EIGVEC(NP,NR,N),EIGVAL(2,N),WS(NR,NWS),IWS(N)
//       DIMENSION W(3)
//       LOGICAL GOOD,TRANSP,COMVEC
// C *** INCLUDE file for compilers which use the IEEE number format:
// C *** All IBM PC/compatibles and the Apple Macintosh
//       REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
//       INTEGER IWARN,IFATAL,ISYSTM
//       CHARACTER MES*60,EFNAME*32
