#define NDEBUG
#define FAST_POLY1
#include "mpi.h"
#include <string>
#include <sstream>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <cstdlib>
#include <stdexcept>
#include <algorithm>
#include <math.h>

#include "small.h"

#include "sl.h"
#include "queue.h"
#include "mpt.cpp"
#include "multipoly.h"
#include "../calc.h"
#include "../small.h"
#include "../smallops.h"
#include "eig1test.h"



int N1, N2, N3;
double Dim[3];
double h[3];
double vol;

template<class Matr, class RMatr>
inline void GausJordanInverse(RMatr &R, Matr &A)
{
  int N = A.Size();
  std::vector<int> ind(N);

  typedef typename Matr::Element_type Double;

  Double max, max1;
  for(int i = 0; i < N; ++i)
  {
    ind[i] = i;
    R[i] = 0;
    R[i][i] = 1;
  }
  int imax;
  
  for(int i = 0; i < N; ++i)
  {
    imax = i;
    max = fabs(A[i][ind[i]]);
    //find pivot element on i-th row;
    for(int k = i; k < N; ++k)
      if(fabs(A[i][ind[k]])>max)
      {
	imax = k;
	max = fabs(A[i][ind[k]]);
      }
    std::swap(ind[i], ind[imax]);
      
    max = 1/A[i][ind[i]];
    A[i] *= max;
    R[i] *= max;

    for (int j = 0; j < N; ++j)
    {
      if(j != i)
      {
	max = A[j][ind[i]];
	A[j] -= max * A[i];
	R[j] -= max * R[i];
	// 	      for(int k = 0; k < N; ++k)
	// 		{
	// 		  A[j][ind[k]] -= max*A[i][ind[k]];
	// 		  R[j][ind[k]] -= max*R[i][ind[k]];
	// 		}
      }
    }
  }
  for(int i = 0; i < N; i++)
  {
    if(ind[i]!=i)
    {
      std::swap(A[i], A[ind[i]]);
      std::swap(R[i], R[ind[i]]);
      std::swap(ind[i], ind[ind[i]]);
    }
  }
      
}


typedef double Double;
typedef My::Base::MultiPoly<Double, 3> Poly3d;


std::string bonename;
std::ifstream streams[6];

typedef std::map<int, double> coeffarray_type;
typedef coeffarray_type::const_iterator cit_type;
coeffarray_type E;
coeffarray_type ni;
coeffarray_type lambda;
coeffarray_type mu;

int rank, size;

typedef SmallMatrix<SmallMatrix<double, 3, 3>, 3, 3> Coeff_type;
typedef std::map<int, Coeff_type> Coeff_ar_type;

Coeff_ar_type Es;

SmallVector<Poly3d, 6> ShapeFun;

void
GenShapeFunMP()
{
  Poly3d t1;
  Poly3d t2;
  Poly3d t3;

  Poly3d one;
  one[0] = 1;

  t1[1] = 1;
  t2[2] = 1;  
  t3[3] = 1;

  ShapeFun[0] = one + (-3.)*t1 + 2. * t1*t1 + (-1.)*t2*t2 + (-1.)*t3*t3;
  ShapeFun[1] = one + (+3.)*t1 + 2. * t1*t1 + (-1.)*t2*t2 + (-1.)*t3*t3;
  ShapeFun[2] = one + (-3.)*t2 + 2. * t2*t2 + (-1.)*t1*t1 + (-1.)*t3*t3;
  ShapeFun[3] = one + (+3.)*t2 + 2. * t2*t2 + (-1.)*t1*t1 + (-1.)*t3*t3;
  ShapeFun[4] = one + (-3.)*t3 + 2. * t3*t3 + (-1.)*t1*t1 + (-1.)*t2*t2;
  ShapeFun[5] = one + (+3.)*t3 + 2. * t3*t3 + (-1.)*t1*t1 + (-1.)*t2*t2;
	
  for(int i = 0; i < 6; ++i)
  {
    ShapeFun[i] *= 1./6.;
  }
}


void
GenShapeFunMV()
{
  using namespace My::Base;
  Poly3d t1;
  Poly3d t2;
  Poly3d t3;

  Poly3d one;
  one[0] = 2; // one == 2!!!

  t1[1] = 1;
  t2[2] = 1;  
  t3[3] = 1;

  ShapeFun[0] = one + (-6.)*t1 + 6. * t1*t1 + (-3.)*t2*t2 + (-3.)*t3*t3;
  ShapeFun[1] = one + (+6.)*t1 + 6. * t1*t1 + (-3.)*t2*t2 + (-3.)*t3*t3;
  ShapeFun[2] = one + (-6.)*t2 + 6. * t2*t2 + (-3.)*t1*t1 + (-3.)*t3*t3;
  ShapeFun[3] = one + (+6.)*t2 + 6. * t2*t2 + (-3.)*t1*t1 + (-3.)*t3*t3;
  ShapeFun[4] = one + (-6.)*t3 + 6. * t3*t3 + (-3.)*t1*t1 + (-3.)*t2*t2;
  ShapeFun[5] = one + (+6.)*t3 + 6. * t3*t3 + (-3.)*t1*t1 + (-3.)*t2*t2;
	
  for(int i = 0; i < 6; ++i)
  {
    ShapeFun[i] *= 1./12;
  }
}




template<class Etype, class Vtype>
void
copy(Etype &diss, const Vtype &b)
{
  for(int i = 0; i < 3; ++i)
    for(int j = 0; j < 3; ++j)
      for(int k = 0; k < 3; ++k)
	for(int l = 0; l < 3; ++l)
	  diss[i][j][k][l]=b[i][j][k][l];
}




template<class Etype>
void
zero(Etype &diss)
{
  for(int i = 0; i < 3; ++i)
    for(int j = 0; j < 3; ++j)
      for(int k = 0; k < 3; ++k)
	for(int l = 0; l < 3; ++l)
	  diss[i][j][k][l]=0;
}
template<class Etype, class Double>
void
multiply(Etype &diss, const Double &val)
{
  for(int i = 0; i < 3; ++i)
    for(int j = 0; j < 3; ++j)
      for(int k = 0; k < 3; ++k)
	for(int l = 0; l < 3; ++l)
	  diss[i][j][k][l]*=val;
}

template<class Etype>
void
print(Etype &diss)
{
  for(int i = 0; i < 3; ++i)
    for(int j = 0; j < 3; ++j)
      for(int k = 0; k < 3; ++k)
	for(int l = 0; l < 3; ++l)
	  std::cout<<i<<j<<k<<l<< " " <<diss[i][j][k][l] << std::endl;
  std::cout << std::endl;
}

template<class Etype>
void
printmat(Etype &diss)
{
  int ii[] = {0, 1, 2, 1, 0, 0};
  int jj[] = {0, 1, 2, 2, 2, 1};
  for(int i = 0; i < 6; ++i)
  {
    std::cout.precision(16);
    std::cout.flags(std::ios::scientific);
    for(int j = 0; j < 6; ++j)
      std::cout << std::setw(24) <<
	diss
	[ii[i]]
	[jj[i]]
	[ii[j]]
	[jj[j]];
    std::cout << std::endl;
  }
}

template<class Etype>
void
printTenzor(Etype &a)
{
  std::cout.precision(15);
  for (int i = 0; i < 3; ++i)
    for (int j = 0; j < 3; ++j)
    {
      for (int k = 0; k < 3; ++k)
	for (int l = 0; l < 3; ++l)
	  std::cout << std::setw(23) << a[i][j][k][l];
      std::cout << std::endl;
    }					  
}
template<class Etype>
void
printbulk(Etype &diss)
{
  for (int i = 0; i < 3; ++i)
  {
    for (int j = 0; j < 3; ++j)
    {
      double Sum = 0;
      for (int k = 0; k < 3; ++k)
	Sum += diss[i][j][k][k];
      std::cout << std::setw(24) <<
	Sum;
    }
    std::cout << std::endl;
  }
}

template<class T, class Etype>
void
CalcBulk(SmallMatrix<T, 3, 3> &bulk, Etype &diss)
{
  for (int i = 0; i < 3; ++i)
  {
    for (int j = 0; j < 3; ++j)
    {
      double Sum = 0;
      for (int k = 0; k < 3; ++k)
	Sum += diss[i][j][k][k];
      bulk[i][j] = Sum;
    }
  }
  // for (int i = 0; i < 3; ++i)
  // {
  //   for (int j =i+1; j < 3; ++j)
  //   {
  //     bulk[i][j] = bulk[j][i];
  //   }
  // }

}

template<class T, class Etype>
void
CalcStress(SmallMatrix<T, 3, 3> &bulk, Etype &diss)
{
  for (int i = 0; i < 3; ++i)
  {
    for (int j = 0; j < 3; ++j)
    {
      double Sum = 0;
      for (int k = 0; k < 3; ++k)
	Sum += diss[i][j][k][k];
      bulk[i][j] = Sum;
    }
  }
  // for (int i = 0; i < 3; ++i)
  // {
  //   for (int j =i+1; j < 3; ++j)
  //   {
  //     bulk[i][j] = bulk[j][i];
  //   }
  // }

}


template<class T, int N>
bool Permute(T (&mas)[N], int max)
{
  for (int i = 0; i < N; ++i)
  {
    mas[i]++;
    if (mas[i] < max)
      return true;
    mas[i] = 0;
  }
  return false;
}

template<class T, int N>
SmallVector<T, N> proj(SmallVector<T, N>const &u, SmallVector<T, N> const &v)
{
  SmallVector<T, N> res = u;
  res *= VecDot(u,v)/VecDot(u,u);
  return res;
}

template<class T, int N>
void OrthoVecs(std::vector<SmallVector<T, N> > &v)
{
  for (int i = 1; i < v.size();++i)
    for (int j = 0; j < i; ++j)
    {
      SmallVector<T, N> p = proj(v[j], v[i]);
      v[i]-=p;
    }
  for (int i = 0; i < v.size(); ++i)
  {
    v[i]/=sqrt(VecDot(v[i], v[i]));
  }
}
template<class T, int N>
void OrthoVecs(SmallMatrix<T, N, N> &mat)
{
  std::vector<SmallVector<T, N> > v;
  for (int i = 0; i < N; ++i)
  {
    SmallVector<T, N> vec;
    for (int j = 0; j < N; ++j)
      vec[j] = mat[i][j];
    v.push_back(vec);
  }
  OrthoVecs(v);
  for (int i = 0; i < N; ++i)
    for (int j = 0; j < N; ++j)
      mat[i][j] = v[i][j];  
}



template<class T, class Etype>
void
Transform(SmallMatrix<T, 3, 3> &Tr, Etype const &src, Etype &dst)
{
  zero(dst);
  int pp[8]={};
  
  const int &k = pp[0];
  const int &l = pp[1];
  const int &s = pp[2];
  const int &t = pp[3];
  const int &m = pp[4];
  const int &n = pp[5];
  const int &p = pp[6];
  const int &r = pp[7];
  int cnt = 0;
  do {
    cnt++;
    dst[k][l][s][t] += src[m][n][p][r] * Tr[k][m]*Tr[l][n]*Tr[s][p]*Tr[t][r];
    //dst[k][l][s][t] += src[m][n][p][r] * Tr[m][k]*Tr[n][l]*Tr[p][s]*Tr[r][t];
    // for (int i = 0; i < 8; ++i)
    //   std::cout << std::setw(2) << pp[i];
    // std::cout << std::endl;
  }while(Permute(pp, 3));
  std::cout << "Cycle rolled " << cnt << " times\n";
}
template<class T, int N>
void transpose(SmallMatrix<T, N, N> &mat)
{
  for (int i = 0; i < N; ++i)
    for (int j = i+1; j<N; ++j)
      std::swap(mat[i][j], mat[j][i]);
}
template<class T, int N>
void SortEigen(SmallVector<T, N> &vals, SmallMatrix<T, N, N> &vecs)
{
  
  for (int i = 0; i <  N; ++i)
  {
    int maxind = i;
    T maxval = vals[maxind];
    for (int j=i+1; j < N; ++j)
      if (vals[j]>maxval)
      {
	maxind = j;
	maxval = vals[j];
      }
    std::swap(vals[i], vals[maxind]);
    for (int k = 0; k < N; k++)
      std::swap(vecs[k][i], vecs[k][maxind]);
  }
  std::cout << "Eigenvals \n" <<std::setw(26) <<  vals << std::endl;
  std::cout << "Eigenvecs \n";
  transpose(vecs);
  OrthoVecs(vecs);
  for (int i = 0; i < N; ++i)
  {
    for (int j = 0; j < N; ++j)
      std::cout << std::setw(26) << vecs[i][j];
    std::cout << std::endl;
  }    
}


template<class Mat, class Etype>
void
copymat(Mat &A, Etype &diss)
{
  int ii[] = {0, 1, 2, 1, 0, 0};
  int jj[] = {0, 1, 2, 2, 2, 1};
  for(int i = 0; i < 6; ++i)
    for(int j = 0; j < 6; ++j)
      A[i][j] = 
	diss
	[ii[i]]
	[jj[i]]
	[ii[j]]
	[jj[j]];
}

void
printind()
{
  int ii1[] = {0, 1, 2, 1, 0, 0};
  int jj1[] = {0, 1, 2, 2, 2, 1};
  int ii2[] = {0, 1, 2, 1, 0, 0};
  int jj2[] = {0, 1, 2, 2, 2, 1};
  std::cout << "$$\n\\left[\n" "\\begin{array}{cccccc}\n";
  for(int i = 0; i < 6; ++i)
  {
    std::cout.precision(4);
    std::cout.flags(std::ios::scientific);
    for(int j = 0; j < 6; ++j)
      std::cout << 
	"c_{" << 
	1+ii1[i] << 
	1+jj1[i] << 
	1+ii2[j] << 
	1+jj2[j] << ((j<5)? "} & ": "}\\\\\n");
  }
  std::cout << "\\end{array}\n\\right]\n$$\n";
}

template<class Etype, class Double>
inline static void CalcE(Etype &E, Double lambda, Double mu)
{
  zero(E);
  double nu = lambda/(2.*(lambda+mu));
  double Y = mu*(3.*lambda + 2.*mu)/(lambda +mu);

  double A = Y * (1.-nu)/((1.+nu)*(1-2.*nu));
  double B = nu/(1.-nu);
  double C = (1-2.*nu)/(2.*(1-nu));
    
  for(int i = 0; i < 3; ++i)
    for(int j = 0; j < 3; ++j)
      E[i][j][i][j] = E[i][j][j][i] = A*C;
  for(int i = 0; i < 3; ++i)
    for(int j = 0; j < 3; ++j)
      E[i][i][j][j] = A*B;
  for(int i = 0; i < 3; ++i)
    E[i][i][i][i] = A;
  //   for(int i = 0; i < 3; ++i)
  //     for(int j = 0; j < 3; ++j)
  //     {
  //       for( int k = 0; k < 3; ++k)
  // 	for( int l = 0; l < 3; ++l)
  // 	  std::cout << E[i][j][k][l] << " ";
  //       std::cout << endl;
  //     }
}
void ReadMaterials(const std::string& fname)
{
  using namespace std;
  ifstream is;
  is.exceptions(ios::badbit|
		ios::eofbit|
		ios::failbit);
  try{
    is.open(fname.c_str(), ios::in);
    std::string a;
    while(true)
    {
      getline(is, a);
      if (a.substr(0,1)=="#")
	continue;
      std::istringstream b;
      b.str(a);
      int ind; double E_, ni_, lambda_, mu_;
      b >> ind;
      b >> ni_ >> E_;
      if (b)
      {
	if ((E_+ni_)==0)
	{
	  b >> lambda_ >> mu_;
	  lambda[ind] = lambda_;
	  mu[ind] = mu_;
	}
	{
	  E[ind] = E_;
	  ni[ind] = ni_;
	  lambda_ = E_ * ni_ /((1+ni_)*(1-2* ni_));
	  mu_ = E_ / (2 *(1+ni_));
	  lambda[ind] = lambda_;
	  mu[ind] = mu_;
	}
	CalcE(Es[ind], lambda_, mu_);
	std::cout << "Material #" << ind << endl;
	printmat(Es[ind]);
	  
	//      std::cout << ind << " " << E << " " << ni << std::endl;
      }
    }
  }catch(std::exception &e)
  {
  }
}

int GetIndex(int k, int l)
{
  if ((k<0)||(l<0)||(k>2)||(l>2)||(k>l))
    throw std::runtime_error("wrong index");
  if (k==l)
    return k;
  if(k==0)
    return 2+l;
  else
    return 5; //l should be 2!!!
}

void GetSizeRankKL(const std::string &s, int &size, int &rank, int &k, int &l)
{
  typedef std::string str;
  typedef str::size_type s_t;
  s_t pos;
  pos = s.find_last_of('.');
  if ((pos<=0)&&(pos==str::npos)) //skip l
    throw std::runtime_error("bad filename 1");
  pos = s.find_last_of('.', pos-1);
  if ((pos<=0)&&(pos==str::npos)) //skip k
    throw std::runtime_error("bad filename 2");
  pos = s.find_last_of('.', pos-1);
  if ((pos<=0)&&(pos==str::npos)) //skip rank
    throw std::runtime_error("bad filename 3");
  pos = s.find_last_of('.', pos-1);
  if ((pos<=0)&&(pos==str::npos)) //skip size
    throw std::runtime_error("bad filename 4");
  std::string ss = s.substr(pos+1);
  std::replace_if(ss.begin(), ss.end(), std::bind2nd(std::equal_to<char>(), '.'), ' '); // replace dots with spaces
  std::istringstream iss(ss+"\n");
  using std::ios;
  iss.exceptions(ios::failbit | ios::eofbit | ios::badbit);
  iss >> size >> rank >> k >> l;  
}
      

void OpenFiles(const std::string &fname)
{
  std::ifstream file;
  using std::ios;
  int openedfiles=0;
  file.exceptions(ios::failbit | ios::eofbit | ios::badbit);
  file.open(fname.c_str());
  //  getline(file, bonename);
  std::string matname;
//   getline(file, matname);
//   ReadMaterials(matname);

  while(openedfiles<6)
  {
    std::string name;
    std::getline(file, name);
    if (!file)
      break;

    int ran, siz, k, l;
    GetSizeRankKL(name, siz, ran, k, l);

    if(siz!=size)
      throw std::runtime_error("Wrong number of processors.");
    if (rank ==ran)
    {
      
      int i = GetIndex(k,l);
      
      streams[i].exceptions(ios::failbit | ios::eofbit | ios::badbit);
      streams[i].open(name.c_str(), ios::in);
      ++openedfiles;
    }
  }
  if (openedfiles!=6)
    throw "wrong number of opened files";
}

// int I, J;
// int what; //0,1,2 - dis strejns stresses
// int pts;
// int K;
// int N1, N2, N3;

// ElastDiscretizator<double> *pdis;



//indexes  //diss[ klpq]
// klpq
template<class Etype>
void
CalcDerivatives(Etype &diss,//SmallMatrix< SmallMatrix<Double, 3, 3>, 3, 3> & diss,
		SmallVector<SmallVector<Double, 18>, 6> &x
		)
{
  static bool firsttime = true;
  static SmallVector<SmallVector<double , 3>, 6> NablaUd;
  static SmallVector<SmallMatrix<double ,3, 3>, 6> Nd;
  if (firsttime)
  {
    SmallVector<SmallVector<Double, 3>, 6 >  pts;
    pts = 0;
    // pts[0][0] = -1;
    // pts[1][0] = 1;
    // pts[2][1] = -1;
    // pts[3][1] = 1;
    // pts[4][2] = -1;
    // pts[5][2] = 1;

    SmallVector<SmallVector<Poly3d, 3>, 6> U;
    for(int which = 0; which< 6; ++which)
      for(int j = 0; j < 3; ++j)
	{
	  U[which][j] = ShapeFun[which].D(j) ;
	  NablaUd[which][j]  = U[which][j](pts[which])*(2./h[j]);
	}
    firsttime = false;
  }
  for(int i = 0; i < 3; ++i)
    for(int j = 0; j < 3; ++j)
      for(int k = 0; k < 6; ++k)
	Nd[k][i][j] = 0;
  for(int k = 0; k < 6; ++k)
    for(int i = 0; i < 3; ++i)
      for(int l = 0; l < 6; ++l)
	for(int j = 0; j < 3; ++j)
	  Nd[k][i][j] += NablaUd[l][j] * x[k][i*6+l];
  zero(diss);
  for(int p = 0; p < 3; ++p)
    for(int q = 0; q < 3; ++q)
    {
      diss[0][0][p][q] = Nd[0][p][q];
      diss[1][1][p][q] = Nd[1][p][q];
      diss[2][2][p][q] = Nd[2][p][q];
      diss[0][1][p][q] = Nd[3][p][q];
      diss[0][2][p][q] = Nd[4][p][q];
      diss[1][2][p][q] = Nd[5][p][q];
      diss[1][0][p][q] = diss[0][1][p][q];
      diss[2][0][p][q] = diss[0][2][p][q];
      diss[2][1][p][q] = diss[1][2][p][q];
    }
//   std::cerr <<"pt" <<   NablaU[0][0][0](pt) << std::endl;
//   std::cerr <<"ptu" <<   U[0][0](pt) << std::endl;

//   zero(diss);

//   SmallVector<SmallVector<Poly3d, 3>, 6> U;
//   SmallVector<SmallMatrix<Poly3d, 3,3>, 6> NablaU;
//   SmallMatrix<Poly3d, 3,3> NablaDotUI;
//   for(int which = 0; which< 6; ++which)
//     for(int i = 0; i < 3; ++i)
//       U[which][i] = 0;
//   for(int which = 0; which< 6; ++which)
//     for(int j = 0; j < 3; ++j)
//       for(int i = 0; i < 6; ++i)
//       {
// 	U[which][j] += ShapeFun[i] * x[which][6*j+i];
// 	//	std::cerr << i<<j<<"X\t"<< 6*j+i << "\t" << x[0][6*j+i] << std::endl;
//       }
  
//   for(int which = 0; which< 6; ++which)
//     for(int i = 0; i < 3; ++i)
//       for(int j = 0; j < 3; ++j)
// 	NablaU[which][i][j] = U[which][i].D(j);
//   for(int p = 0; p < 3; ++p)
//     for(int q = 0; q < 3; ++q)
//     {
//       diss[0][0][p][q] = NablaU[0][p][q](pt);
//       diss[1][1][p][q] = NablaU[1][p][q](pt);
//       diss[2][2][p][q] = NablaU[2][p][q](pt);
//       diss[0][1][p][q] = NablaU[3][p][q](pt);
//       diss[0][2][p][q] = NablaU[4][p][q](pt);
//       diss[1][2][p][q] = NablaU[5][p][q](pt);
//       diss[1][0][p][q] = diss[0][1][p][q];
//       diss[2][0][p][q] = diss[0][2][p][q];
//       diss[2][1][p][q] = diss[1][2][p][q];
//     }
//   std::cerr <<"pt" <<   NablaU[0][0][0](pt) << std::endl;
//   std::cerr <<"ptu" <<   U[0][0](pt) << std::endl;
// }
}

bool ReadLineX(SmallVector<SmallVector<Double, 18>, 6> &x, double &lambda, double &mu)
{
  try{
    std::string str;
    for(int i = 0; i < 6; ++i)
    {
      do{
	getline(streams[i], str);
      }
      while((str.length()>0)&&(str[0]=='#'));
      std::istringstream is(str);
      is >> lambda>>mu;
      for(int j = 0; j < 18; ++j)
      {
	is >> x[i][j];
	//      x[i][j]/=16;
	//      std::cerr << "read" << i<<j<< "\t" << x[i][j]<< std::endl;
      }
    }
    return true;
  }catch(std::exception &e)
  {
  }
  return false;
}

template<class T>
class Sumator
{
  T possum, negsum;
  T poserr, negerr;
    
  static  void AddTo(T &sum, T&err, const T& what)
  {
    T oldsum = sum;
    sum += what;
    T tmperr = what - (sum - oldsum);
    err += tmperr;
  }
public:
  Sumator(int a= 0):possum(T(a)), poserr(T()), negsum(T()), negerr(T()){}
  void operator +=(const T&what)
  {
    Add(what);
  }
  void Add(const T&what)
  {
    if (what>=0)
      AddTo(possum, poserr, what);
    else
      AddTo(negsum, negerr, what);
  }
  operator T const () const
  {
    return (possum+negsum)+(poserr+negerr);
  }
  
};

double Currs[3][3][3][3];
double //Sumator<double>
 EH[3][3][3][3];
double EE[3][3][3][3];




template<class EHt, class Et, class Dt>
void
Update(EHt &EH, const Et &E, const Dt& der, double vol)
{
//   for(int i = 0; i < 3; ++i)
//     for(int j = 0; j < 3; ++j)
//       for(int k = 0; k < 3; ++k)
// 	for (int l = 0; l < 3; ++l)
  for(int i = 3; i--;)
    for(int j = 3; j--;)
      for(int k = 3; k--;)
	for (int l = 3; l--;)
	{
	  double sum = 0.0;//	  Sumator<double> sum = 0.0;
	  
// 	  for(int p = 0; p < 3; ++p)
// 	    for(int q = 0; q < 3; ++q)
	  for(int p = 3; p--;)
	    for(int q = 3; q--;)
	    {
	      sum += E[i][j][p][q] * der[k][l][p][q];
	      //	      std::cerr <<p<<q<<"\t"<<  E[i][j][p][q] << "\t" << der[k][l][p][q] << std::endl;
	    }
	  //	  std::cerr << i << j << k << l << "\t" << E[i][j][k][l] << "\t" << sum << std::endl;
	  //	  sum = 0;
	  EH[i][j][k][l] += (E[i][j][k][l] - sum)*vol;
	}
}

template<class T, class V>
void PrintDataAndCalcBulk(T &t2, V&bulk)
{
      std::cout << "EH " << std::endl;
      printmat(t2);
      std::cout << "Bulk modulus tensor" << std::endl;
      printbulk(t2);
      double lambda =
	t2[0][0][1][1] +
	t2[0][0][2][2] +
	t2[1][1][2][2];
      double mu =
	t2[0][1][0][1] +
	t2[0][2][0][2] +
	t2[1][2][1][2];
      double sum =
	t2[0][0][0][0] +
	t2[1][1][1][1] +
	t2[2][2][2][2];
      SmallMatrix<double, 6, 6> mat, imat;
      copymat(mat, t2);
      mat[3][3]*=2;
      mat[4][4]*=2;
      mat[5][5]*=2;
      GausJordanInverse(imat, mat);
      sum/=3;
      lambda/=3;
      mu/=3;
      double C = sum;
      double A = lambda;
      double B = mu;
      double Y = (C+B-A)/3.;
      double X = (B + 2.*C - 5.*Y)/2.;
      
      double mu1 = (sum -lambda)/2;
      mu=(mu1+mu)/2.;
      lambda = X;
      mu = Y;
      std::cout << lambda << " " << mu << std::endl;
      double nu = lambda/(2.*(lambda+mu));
      double E = mu* ((3.*lambda+2.*mu)/(lambda+mu));
      std::cout << "Isotropic values : " << std::endl;
      std::cout << "v = " << nu << std::endl;
      std::cout << "E = " << E << std::endl;
      double Ex, Ey, Ez, vyx, vzx, vzy, vxy, vxz, vyz,  muxy, muxz, muyz;
      Ex = 1./imat[0][0];
      Ey = 1./imat[1][1];
      Ez = 1./imat[2][2];
      vxy = - imat[1][0]*Ex;
      vxz = - imat[2][0]*Ex;
      vyz = - imat[2][1]*Ey;
      vyx = - imat[0][1]*Ey;
      vzx = - imat[0][2]*Ez;
      vzy = - imat[1][2]*Ez;
      std::cout << "Orthotropic values : " << std::endl;
      std::cout << "E_x = " << Ex << std::endl
		<< "E_y = " << Ey << std::endl
		<< "E_z = " << Ez << std::endl
		<< "v_xy = " << vxy << std::endl
		<< "v_xz = " << vxz << std::endl
		<< "v_yx = " << vyx << std::endl
		<< "v_yz = " << vyz << std::endl
		<< "v_zx = " << vzx << std::endl
		<< "v_zy = " << vzy << std::endl;
      CalcBulk(bulk, t2);
}

int main(int argc, char **argv)
{
  MPI::Init(argc, argv);
  MPI::Intracomm &comm = MPI::COMM_WORLD;
  rank = comm.Get_rank();
  size = comm.Get_size();
  try{
    if (argc<2)
      throw (ReadMaterials("materials"),printind(),1);

    OpenFiles(argv[1]);
    //    std::cout << rank<< "opened "<< std::endl;

    std::string str;
    getline(streams[0], str);
    //    std::cout << rank << "read << " << str;
    str = str.substr(1);
    std::istringstream is(str);
    int DT;
    int dup;
    is >> N1 >> N2 >> N3 >>DT >>dup >> dup >>Dim[0] >> Dim[1] >> Dim[2];
    if (DT==0) GenShapeFunMP(); else GenShapeFunMV();
    SmallVector<SmallVector<Double, 18>, 6> x;
      
    h[0] = Dim[0]/N1;
    h[1] = Dim[1]/N2;
    h[2] = Dim[2]/N3;
    vol = h[0]*h[1]*h[2];
    zero(EH);
    double lambda, mu;
    double total = (double)N1*N2*N3/size;
    unsigned long  count = 0;
    while (ReadLineX(x, lambda, mu))
    {
      
      //      for(int f = 0; f < 6; ++f)
      CalcDerivatives(Currs, x);
      CalcE(EE, lambda, mu);
      Update(EH, EE, Currs, vol);
      count += 1;
      if ((rank==0)&&(!(count&127)))
	std::cerr <<"\r"  <<(int) round(100.0*(count/total)) << "%     " <<std::flush;
    }
    double bigvol = Dim[0]*Dim[1]*Dim[2];
    double t1[3][3][3][3], t2[3][3][3][3];
    double tt[3][3][3][3];
    copy(t1, EH);
    comm.Reduce(&t1[0][0][0][0], &t2[0][0][0][0], 3*3*3*3, MPI::DOUBLE, MPI::SUM, 0);
    multiply(t2, 1./bigvol);
    if (rank == 0)
    {
      std::cerr << "        \r           \r" << std::flush;
      std::cout << vol << " " << bigvol  << std::endl;

    }
    if (rank == 0)
    {
      SmallMatrix<double, 3, 3> bulk;
      SmallMatrix<double, 3, 3>  v;
      SmallVector<double, 3> llambda;
      
      PrintDataAndCalcBulk(t2, bulk);

      CalcEigen(bulk, llambda, v);
      //      My::Calc::REigenStuff(bulk, llambda, v);
      SortEigen(llambda, v);
      Transform(v, t2, tt);
      std::cout << "Transformed \n" ;
      PrintDataAndCalcBulk(tt, bulk);

      CalcEigen(bulk, llambda, v);
      SortEigen(llambda, v);
      Transform(v, tt, t2);
      std::cout << "Transformed again\n" ;
      PrintDataAndCalcBulk(t2, bulk);
    }
  }
  catch(std::exception&e)
  {
    //    if (rank==0)
      std::cerr << "Exception throuwn " << e.what() <<"." << rank << ":" << size << std::endl;
  }
  catch(...)
  {
  }
  MPI::Finalize();
}

