#ifndef __EPS_WRITER_H_INCLUDED__
#define __EPS_WRITER_H_INCLUDED__

class EpsWriter:public ofstream
{
	double fheight;
	double xs, xw, ys, yw;
public:
	void WriteHeader()
	{
		*this << "%!PS-Adobe-3.0 EPSF-3.0\n";
		*this << "%%BoundingBox: " << 0 << " " << 0 << " " << xw+xs*2 << " " << yw + ys*2<< "\n";
		*this << "save\n";
	}
	void WriteEnder()
	{
		*this << "restore" << endl;
	}
	EpsWriter (string filename, double width, double height):ofstream(filename.c_str(), ios::binary)
	{
		fheight = 0;
		if (!*this)
			throw runtime_error(string("Cannot open \"") + filename + "\".");
		xs = ys = 10;
		xw = width;
		yw = height;
		WriteHeader();
	}
	void SetFont(const string&fontname, double fontheight)
	{
		*this <<"(" << fontname << ") findfont " << (fheight = fontheight*yw) << " scalefont setfont\n";
	}
	void NewPath()
	{
		*this << "newpath\n";
	}
	void Fill()
	{
	  *this << "fill\n";
	}
	void FillRect(double x1, double y1, double x2, double y2)
	{
	  NewPath();
	  MoveTo(x1, y1);
	  LineTo(x1, y2);
	  LineTo(x2, y2);
	  LineTo(x2, y1);
    ClosePath();
	  Fill();
	}
	void ClosePath()
	{
		*this << "closepath\n";
	}
	void Stroke()
	{
		*this << "stroke\n";
	}
	void MoveTo(double x, double y)
	{
		*this << x*xw + xs << " " << y*yw + ys << " moveto\n";
	}
	void LineTo(double x, double y)
	{
		*this << x*xw + xs << " " << y*yw + ys << " lineto\n";
	}
	void DrawLine(double x1, double y1, double x2, double y2)
	{
		NewPath();
		MoveTo(x1, y1);
		LineTo(x2, y2);
		Stroke();
	}
	void Arc(double x1, double y1, double radius, double sa = 0., double ea = 360.)
	{
		*this << x1*xw+xs << " " << y1*yw+ys << " " << " " << radius*xw << " " << sa << " " << ea << " arc\n";
	}
	void DrawArc(double x1, double y1, double radius, double sa = 0., double ea = 360.)
	{
		NewPath();
		Arc(x1, y1, radius, sa, ea);
		Stroke();
	}

	void DrawString(double x1, double y1, const string &what)
	{
		*this << "(" << what << ") ";
		*this << "dup  stringwidth pop  -2 div " << x1*xw+xs 
			<< " add  " << ys + yw*y1 - fheight/2 << " moveto show\n";
	}

	void SetColor(double r, double g, double b)
	{
		*this << r << " " << g << " " << b << " setrgbcolor\n";
	}

	void LineWidth(double width)
	{
		*this << width << " setlinewidth\n";
	}


	~EpsWriter()
	{
		WriteEnder();
	}
};

#endif
