//  config.hpp
//
//  Copyright (c) 2003 Eugene Gladyshev
//
//  Permission to copy, use, modify, sell and distribute this software
//  is granted provided this copyright notice appears in all copies.
//  This software is provided "as is" without express or implied
//  warranty, and with no claim as to its suitability for any purpose.
//

#ifndef __ttl_config__hpp
#define __ttl_config__hpp

#include <new>
#include <stdexcept>
#include <functional>
#include <memory>


#if defined(_MSC_VER)
#	define TTL_MAX_TEMPLATE_PARAMS 25
#elif defined(__GNUC__)
#	define TTL_MAX_TEMPLATE_PARAMS 25
#else
#	define TTL_MAX_TEMPLATE_PARAMS 25
#endif

#define TTL_MAX_TYPELIST_PARAMS TTL_MAX_TEMPLATE_PARAMS


namespace ttl
{
	struct empty_type {};
	
namespace impl
{
	struct data_holder_base {};
	
	template< typename T >
	struct data_holder : data_holder_base
	{
		typedef T type;
		typedef T& reference;
		typedef T* pointer;
		
		type d_;
		
		data_holder( const T& l ) : d_(l) 
		{
		}
	};

	template< typename T >
	struct data_holder<T&> : data_holder_base
	{
		typedef T& type;
		typedef T& reference;
		typedef T* pointer;
		
		type d_;
		
		data_holder( const T& l ) : d_((type)l)
		{
		}
	};
	
	template< typename T >
	struct data_holder<const T&> : data_holder_base
	{
		typedef const T& type;
		typedef const T& reference;
		typedef const T* pointer;
		
		type d_;
		
		data_holder( const T& l ) : d_(l)
		{
		}
	};
	
	template< typename T >
	struct data_holder<T*> : data_holder_base
	{
		typedef T* type;
		typedef T& reference;
		typedef T* pointer;
		
		type d_;
		
		data_holder( T* l ) : d_(l) {}
	};
	
	template< typename T >
	struct data_holder<const T*> : data_holder_base
	{
		typedef const T* type;
		typedef const T& reference;
		typedef const T* pointer;
		
		type d_;
		
		data_holder( const T* l ) : d_(l) {}
	};
	
	template< typename T >
	struct is_reference
	{
		enum { value = 0 };
	};
	
	template< typename T >
	struct is_reference<T&>
	{
		enum { value = 1 };
	};
	
	template< typename T >
	struct is_reference<const T&>
	{
		enum { value = 1 };
	};
}
}


#endif //__ttl_config__hpp
