#ifndef __useful_h__
#define __useful_h__
// IBM workaround
#ifdef _MY_CPP
#include "mympicxx.hpp"
#endif

#include "mpi.h"


#include <string>
#include <sstream>
#include <map>
#include <iostream>
#include <iomanip>
#include <typeinfo>

#include "small.h"
#include <math.h>

enum DiscretizationType{MV, MP, RV, RP};
enum BoundaryConditions{bcNone  = 0x00,
			bcLeft  = 0x01,
			bcRight = 0x02,
			bcDown  = 0x04,
			bcUp    = 0x08, 
			bcFront = 0x10,
			bcBack  = 0x20,
			bcNeuman= 0x40};
enum SolutionMethod{smCG, smJacobi, smPCG1, smPCG2, smPCG3, smPCG3PBC, smPCG4, smBoomer, smBoomerDD};


template<class T>
T& GetPerturb()
{
  static T a = 1;
  return a;
}

class ManageTimes
{
 public:
  typedef std::map<std::string, double> map_type;
  typedef map_type::iterator map_iterator;
  static map_type& GetMap()
  {
    static map_type map;
    return map;
  }
  static double & Get(const std::string &str)
  {
    return GetMap()[str];
  }
  static void Reduce(MPI::Intracomm &comm)
  {
    int c = 0;
    for(map_iterator it = GetMap().begin(); it!=GetMap().end();++it)
    {
      comm.Barrier();
      double time1 = it->second;
      double time;
      comm.Allreduce(&time1, &time, 1, MPI::DOUBLE, MPI::MAX);
      it->second=time;
      c++;
    }
  }
  template<class Ostream>
    static void PrintAll(Ostream &out, int iter=1, std::string su="")
  {
    for(map_iterator it = GetMap().begin(); it!=GetMap().end();++it)
    {
      double time = it->second;
      using namespace std;
      out << left<< setw(30) << it->first + su
	  << setw(30)<< time/iter << "\n";
    }
  }
  
};

class Timer
{
    std::string time;
 public:
    Timer(const std::string &tm):time(tm)
	{
	    ManageTimes::Get(time) -= MPI::Wtime();
	}
    ~Timer()
	{
	  ManageTimes::Get(time) += MPI::Wtime();
	}
};



class Options
{
public:
  double lam_coeff;
  bool Overlap;
  bool SaveSolution;
  bool debug;
  bool PBC;
  bool spawned;
  double Reg;
  int variant;
  int agr;
  int ct;
  int mit;
  int np;
  int maxiter;
  double pfactor;
  double rwt;
  double ppfactor;
  bool SinglePrecPrecon;
  double zero;
  double thres;
  double DM1, DM2, DM3;
  double eps;
  bool gcg;
  bool proj;
  bool froj;
  int gcg_m;
  int N, N1, N2, N3;
  bool useplates;
  bool reuse_rhs;
  DiscretizationType distype;
  BoundaryConditions bc;
  BoundaryConditions nc;
  SolutionMethod method;
  bool SmallDirichlet;
  bool SmallDirichlet2;
  double matscale;

  double InnerTol;
  int HypreCoarsType;
  int HypreInterpType;
  int HyprePMaxElmts;
  int HypreRelaxType;
  double HypreOuterWt;
  int HypreNumSweeps;
  int homk, homl;
  bool hom;

  bool usebone;
  std::string bone_fname;
  typedef std::map<int, double> coeffarray_type;
  typedef coeffarray_type::const_iterator cit_type;
  coeffarray_type E;
  coeffarray_type ni;
  coeffarray_type lambda;
  coeffarray_type mu;

  //  unsigned NumMaterials;
  double force;
  std::string namebase;
  std::string outputname;
  std::string scratchpath;
  int outputprecision;
  Options()
    {
      outputprecision = 7;
      mit = 1;
      debug = false;
      ct = 1;
      agr = 0;
      rwt = 0.5;
      thres = 0.5;
      SaveSolution = false;
      SinglePrecPrecon = false;
      force = 0;
      matscale=1;
/*       NumMaterials=256; */
/*       for(unsigned int i = 0; i < NumMaterials; ++i) */
/* 	E[i] = ni[i] = 0; */
/*       E[0] = 10e6; */
/*       ni[0] = 0.35; */
/*       E[255] = 10e6; */
/*       ni[255] = 0.35; */
      Overlap = true;
      pfactor = 1.;
      ppfactor = 1.;
      usebone = false;
      zero = 1.;
      DM1 = DM2 = DM3 = 1.;
      variant=0;
    }

  
  static Options& GetOptions()
  {
    static Options opts;
    return opts;
  }
};

#define OVERLAP (Options::GetOptions().Overlap)


#define TimeCalc1(what) Timer t_m_p## 1 (what)
#define TimeCalc(what) Timer t_m_p ## 1 (what)
#define TimeCalc1p(what,ln) Timer t_m_p## ln (what)
#define TimeCalcp(what,ln) Timer t_m_p ## ln (what)
#define TimeCalc2(what, howmany) ManageTimes::Get(what)+=(howmany)
template<class T> class Myf;

template<> class Myf<double>
{
public:
  typedef double type;
  typedef double base_type;
  typedef volatile float single;
  static void  Max(type &max, const type &a)
  {
    max = std::max(max, fabs(a));
  }
  static type  Max(const type &max)
  {
    return max;
  }
  static MPI::Datatype Datatype()
  {
    return  MPI::DOUBLE;
  }
  static base_type Reduce(type& a)
  {
    return a;
  }
  static std::string ToString()
  {
    return "d";
  }
  static void  Zero(double &f)
  {
    if (fabs(f) < Options::GetOptions().zero)
      {
	// if (fabs(f)>0) 
	//   std::cout << " ziroema " <<std::scientific << f <<" " << Options::GetOptions().zero<<  std::endl ; 
	f = 0;
      }
  }
  static void Onify(double &){}
};
template<> class Myf<float>
{
  public:
  typedef float type;
  typedef float base_type;
  typedef volatile float single;
  static void  Max(type &max, const type &a)
  {
    max = std::max(max, fabsf(a));
  }
  static type  Max(const type &max)
  {
    return max;
  }
static MPI::Datatype Datatype()
{
  return  MPI::FLOAT;
}
static base_type Reduce(type& a)
{
  return a;
}
static std::string ToString()
{
  return "f";
}
static void  Zero(float &f)
{
  if (fabs(f) < Options::GetOptions().zero)
    f = 0;
}
static void Onify(float &){}
};

template<class T, int I3>
class Myf<SmallVector<T,I3> >
{
public:
  typedef SmallVector<T, I3> type;
  typedef T base_type;
  typedef SmallVector<volatile float, I3> single;
  MPI::Datatype mpitype;
  static void  Max(type &max, const type &a)
  {
    for(int i=0; i < I3; ++i)
      Myf<base_type>::Max(max[i], a[i]);
  }
  Myf() // Here we construct the datatype
  {
    MPI::Datatype bdt = Myf<T>::Datatype();
    mpitype = bdt.Create_contiguous(I3);
    mpitype.Commit();
  }
  ~Myf() // Here we destruct it;
  {
    mpitype.Free();
  }
  static Myf*& Geta()
  {
    static Myf *a;
    return a;
  }
  static MPI::Datatype Datatype()
  {

    return Geta()->mpitype;
  }
  static std::string ToString()
  {
    return "v3" + Myf<T>::ToString();
  }
  static base_type Reduce(type& a)
  {
    base_type r;
    r = 0;
    for (int i = 0; i < I3; ++i)
      r += a[i];
    return r;
  }
  static base_type Max(type& a)
  {
    base_type r;
    r = 0;
    for (int i = 0; i < I3; ++i)
      r = std::max(r, a[i]);
    return r;
  }
  static void Zero(SmallVector<T,I3> &a)
  {
    //  std::cout << "Called" << std::endl;
    for (int i = 0; i < I3; i++)
      Myf<T>::Zero(a[i]);
  }
  static void Onify(SmallVector<T,I3> &a)
  {
    a[0]=1;
    for(int i =1; i < I3; ++i) 
      a[i]=0; 
  }
};


template<class T> class Myf
{
public:
  /* typedef typename T::base_type base_type; */
  /* typedef typename T::type type; */
  /* typedef typename T::single single; */

  static MPI::Datatype Datatype()
  {
    throw "losh tip";
    //      return  MPI::FLOAT;
  }
  static std::string ToString()
  {
    return "fucked";
  }
  static void Zero(T &a)
  {
    std::cout << typeid(T).name() << std::endl;
  }
  static void Onify(T &){}
};






inline unsigned  GetNum (std::string s, unsigned def)
{
  std::istringstream is(s);
  unsigned a;
  is >> a;
  if (a)
    return a;
  return def;
}


#endif
