#ifndef __vectors_hpp__
#define __vectors_hpp__

#include "vectors.h"
#include "useful.h"
#include <cstdlib>




template<class BaseType, class Int>
void
BDom<BaseType, Int>::
CreateDatatypes()
{
  if(Dt_Created)
      return;
  Index3dBase<Int> ind(N1, N2, N3);
  MPI::Datatype bdt = Myf<BaseType>::Datatype();
  int lengths[3] = {1,1,1};
  MPI::Aint displacements[3];
  MPI::Aint lb, extent,     stride;
    //    pripletextent,
  MPI::Datatype datatypes[3];
  bdt.Get_extent(lb, extent);
  /*
    MPI::Datatype Left, Down, Front;

    MPI::Datatype Left_line, Down_line, Front_line;

    MPI::Datatype Left_down, Front_down;
    
    MPI::Datatype Up_unfinished;
  */
  displacements[0] = 0;
  displacements[1] = MPI::Get_address(&(buf[0][LeftEl])) -
    MPI::Get_address(buf);
  displacements[2] = MPI::Get_address(buf+1)-MPI::Get_address(buf);
    
  datatypes[0] = MPI::LB;
  datatypes[1] = bdt;
  datatypes[2] = MPI::UB;



  DtLeft = MPI::Datatype::Create_struct(3, lengths, displacements,
					datatypes);
  displacements[1] = MPI::Get_address(&(buf[0][FrontEl])) -
    MPI::Get_address(buf);


  DtFront = MPI::Datatype::Create_struct(3, lengths, displacements,
					 datatypes);
  displacements[1] = MPI::Get_address(&(buf[0][DownEl])) -
    MPI::Get_address(buf);


  DtDown = MPI::Datatype::Create_struct(3, lengths, displacements,
					datatypes);

  ind.SetPos(0,0,0);
  stride = MPI::Get_address(&(*this)[ind]);
  ind.SetPos(0,1,0);
  stride = MPI::Get_address(&(*this)[ind])-stride;


  DtLeft_line = DtLeft.Create_hvector(ind.N2, 1, stride);

  DtDown_line = DtDown.Create_hvector(ind.N2, 1, stride);

  DtFront_line = DtFront.Create_hvector(ind.N2+1, 1, stride);

  ind.SetPos(0,0,0);
  stride = MPI::Get_address(&(*this)[ind]);
  ind.SetPos(1,0,0);
  stride = MPI::Get_address(&(*this)[ind])-stride;

  DtDown_left_plane = DtLeft_line.Create_hvector(ind.N1+1, 1, stride);
    
  displacements[0] = 0;

  ind.SetPos(0,0,0);
  stride = MPI::Get_address(&(*this)[ind]);
  ind.SetPos(0,0,1);
  stride = MPI::Get_address(&(*this)[ind]) - stride;
  displacements[1] = stride;
  ind.SetPos(0,0,0);
  stride = MPI::Get_address(&(*this)[ind]);
  ind.SetPos(1,0,0);
  stride = MPI::Get_address(&(*this)[ind]) - stride;
  displacements[2] = stride;
  datatypes[0] = DtLeft_line;
  datatypes[1] = DtDown_line;
  datatypes[2] = MPI::UB;
  DtLeft_down = MPI::Datatype::Create_struct(3, lengths, 
					     displacements,
					     datatypes);
  datatypes[0] = DtFront_line;
  datatypes[1] = DtDown_line;
  displacements[1] = 0;
  displacements[2] = stride;
  DtFront_down = MPI::Datatype::Create_struct(3, lengths,
					      displacements,
					      datatypes);

    
  DtUp_unfinished = DtLeft_down.Create_contiguous(ind.N1);
  DtDown_plane = DtFront_down.Create_contiguous(ind.N1);
    
  DtLeft_down.Get_extent(lb, extent);



  datatypes[0] = DtUp_unfinished;
  datatypes[1] = DtLeft_line;
  displacements[0] = 0;

  ind.SetPos(0,0,0);
  stride = MPI::Get_address(&(*this)[ind]);
  ind.SetPos(ind.N1,0,0);
  stride = MPI::Get_address(&(*this)[ind]) - stride;
  displacements[1] = stride;

  DtUp_plane = MPI::Datatype::Create_struct(2, lengths,
					    displacements,
					    datatypes);


  DtDown_plane.Commit();
  DtUp_plane.Commit();

  DtLeft.Commit();DtLeft_line.Commit();
  DtFront.Commit();DtFront_line.Commit();
  DtDown.Commit();DtDown_line.Commit();
  DtLeft_down.Commit();
  DtFront_down.Commit();
  DtUp_unfinished.Commit();
  DtDown_left_plane.Commit();
  Dt_Created = true;
    
}

template<class BaseType, class Int>
BDom<BaseType, Int>::
BDom(Int n1, Int n2, Int n3, Int n3r) // N1, N2, N3 size of the mesh
  : Base((n1+3)*(n2+3)*(n3+3)),N1(n1), N2(n2), N3(n3),N3R(n3r)  
{    
  //      SendLine.resize((N2+1)*2);    
  //      RecvLine.resize((N2+1)*2);
    Dt_refcnt++;
  comm = NULL;
}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
SetComm(MPI::Intracomm  &co)
{
  comm = &co;
  rank = comm->Get_rank();
  size = comm->Get_size();
  NextP = UpP = rank+1;
  PrevP = DownP = rank-1;
  if (DownP <0)
  {
    DownP = MPI::PROC_NULL;
    PrevP = size-1;
  }
  if (UpP == size)
  {
    UpP = MPI::PROC_NULL;
    NextP=0;
  }

  PUpP = UpP;
  PDownP = DownP;
  using namespace std;
  //      cerr << Rank << " " << Size << endl;
}

template<class BaseType, class Int>
BDom<BaseType, Int>::
BDom():N1(0), N2(0), N3(0){Dt_refcnt++;SetComm(MPI::COMM_SELF);}


template<class BaseType, class Int>
BDom<BaseType, Int>::
~BDom()
{
  if((!--Dt_refcnt) && Dt_Created)
  {
    DtUp_plane.Free();
    DtDown_plane.Free();
    DtLeft.Free();DtLeft_line.Free();
    DtFront.Free();DtFront_line.Free();
    DtDown.Free();DtDown_line.Free();
    DtLeft_down.Free();
    DtFront_down.Free();
    DtUp_unfinished.Free();
    DtDown_left_plane.Free();
  }
}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
SetSize(Int n1, Int n2, Int n3, Int n3r)
{
  N1 = n1; N2 = n2; ON3 = N3 = n3;N3R = n3r;
  Base::SetSize1((n1+3)*(n2+3)*(n3+3));
  SetPeriodic(false);
  //    SendLine.resize((N2+1)*2);
  //    RecvLine.resize((N2+1)*2);
}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
ClearLine(Int I, Int K, Elment El, Int a)
{
  Index3dBase<Int> ind(N1, N2, N3);
  ind.SetPos(I, -1, K);
  //  Int cnt = 0;
  Int J;
  for (J = 0; J<N2+3; J++, ind.ToBack())
  { 
    (*this)[ind][El] = a;
  }    
}





template<class BaseType, class Int>
void
BDom<BaseType, Int>::
ClearZero()
{
  //  std::cout << "size is" << size << std::endl;
  for (int i = 0; i < this->Size; ++i)
    for(int j = 0; j < 3; ++j)
      Myf<BaseType>::Zero(buf[i][j]);
}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
ClearOthers()
{
  using namespace std;
  //    cout << "n3 is " << N3 << endl;
  for (Int i = -1; i <= N1+1; i++)
  {
    ClearLine(i, -1, DownEl);
    ClearLine(i, -1, FrontEl);
    if (rank == 0)
      ClearLine(i, -1, LeftEl);

    ClearLine(i, N3-1, LeftEl);
    if(rank+1 == size)
      ClearLine(i, N3-1, FrontEl);
    ClearLine(i, N3, DownEl); 
    ClearLine(i, N3, LeftEl);
    ClearLine(i, N3, FrontEl);
    ClearLine(i, N3+1, DownEl);

  }
  for (Int i = -1; i <= N3+1; ++i)
  {
    ClearLine(N1, i, DownEl);
    ClearLine(N1, i, FrontEl);
    ClearLine(N1+1, i, LeftEl);
    ClearLine(N1+1, i, DownEl);
    ClearLine(N1+1, i, FrontEl);
  }
    
  for (Int i = -1; i <= N1+1; ++i)
    for (Int j = -1; j<= N3+1; ++j)
    {
      Index3dBase<Int> ind(N1, N2, N3);
      ind.SetPos(i, N2, j);
      (*this)[ind][LeftEl] = 0;
      (*this)[ind][DownEl] = 0;
      ind.ToBack();
      (*this)[ind][FrontEl] = 0;
      (*this)[ind][LeftEl] = 0;
      (*this)[ind][DownEl] = 0;

      ind.SetPos(i, -1, j);
      (*this)[ind][LeftEl] = 0;
      (*this)[ind][DownEl] = 0;
      (*this)[ind][FrontEl] = 0;
    }
}


template<class BaseType, class Int>
void
BDom<BaseType, Int>::
ClearOthers1()
{
  using namespace std;
  //    cout << "n3 is " << N3 << endl;
  for (Int i = -1; i <= N1+1; i++)
  {
    ClearLine(i, -1, DownEl);
    ClearLine(i, -1, FrontEl);
    //      if (rank == 0)
    //      ClearLine(i, -1, LeftEl);

    //      ClearLine(i, N3-1, LeftEl);
    if(rank+1 == size)
      ClearLine(i, N3-1, FrontEl);
    ClearLine(i, N3, DownEl); 
    ClearLine(i, N3, LeftEl);
    ClearLine(i, N3, FrontEl);
    ClearLine(i, N3+1, DownEl);

  }
  for (Int i = -1; i <= N3+1; ++i)
  {
    ClearLine(N1, i, DownEl);
    ClearLine(N1, i, FrontEl);
    ClearLine(N1+1, i, LeftEl);
    ClearLine(N1+1, i, DownEl);
    ClearLine(N1+1, i, FrontEl);
  }
    
  for (Int i = -1; i <= N1+1; ++i)
    for (Int j = -1; j<= N3+1; ++j)
    {
      Index3dBase<Int> ind(N1, N2, N3);
      ind.SetPos(i, N2, j);
      (*this)[ind][LeftEl] = 0;
      (*this)[ind][DownEl] = 0;
      ind.ToBack();
      (*this)[ind][FrontEl] = 0;
      (*this)[ind][LeftEl] = 0;
      (*this)[ind][DownEl] = 0;

      ind.SetPos(i, -1, j);
      (*this)[ind][LeftEl] = 0;
      (*this)[ind][DownEl] = 0;
      (*this)[ind][FrontEl] = 0;
    }
}
template<class BaseType, class Int>
void
BDom<BaseType, Int>::
ClearOthers2()
{
  using namespace std;
  //  cout << "n3 is " << N3 << ON3 << endl;
  for (Int i = -1; i <= N1+1; i++)
  {
    ClearLine(i, -1, DownEl);
    ClearLine(i, -1, FrontEl);

    ClearLine(i, N3-1, LeftEl);
    ClearLine(i, N3, DownEl); 
    ClearLine(i, N3, LeftEl);
    ClearLine(i, N3, FrontEl);
    ClearLine(i, N3+1, DownEl);

    ClearLine(i, ON3-1, LeftEl);
    ClearLine(i, ON3, DownEl); 
    ClearLine(i, ON3, LeftEl);
    ClearLine(i, ON3, FrontEl);
    ClearLine(i, ON3+1, DownEl); 

  }
  for (Int i = -1; i <= N3+1; ++i)
  {
    ClearLine(N1, i, LeftEl);
    ClearLine(N1, i, DownEl);
    ClearLine(N1, i, FrontEl);
    ClearLine(N1+1, i, LeftEl);
    ClearLine(N1+1, i, DownEl);
    ClearLine(N1+1, i, FrontEl);
  }

  for (Int i = -1; i <= N1+1; ++i)
    for (Int j = -1; j<= N3+1; ++j)
    {
      Index3dBase<Int> ind(N1, N2, N3);
      ind.SetPos(i, N2, j);
      (*this)[ind][LeftEl] = 0;
      (*this)[ind][DownEl] = 0;
      ind.ToBack();
      (*this)[ind][FrontEl] = 0;
      (*this)[ind][LeftEl] = 0;
      (*this)[ind][DownEl] = 0;

      ind.SetPos(i, -1, j);
      (*this)[ind][LeftEl] = 0;
      (*this)[ind][DownEl] = 0;
      (*this)[ind][FrontEl] = 0;
    }
}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
StartFwdM4(Index3dBase<Int>const & Id, Index3dBase<Int>const & Iu)
{
  Index3dBase<Int> I(Id);
  senddata[0] = this->Up(Id)[DownEl];
  senddata[1] = this->Up(Iu)[DownEl]; 
  senddata[2] = this->RightFront(Iu)[LeftEl];
  Recvdata[0] = 0;
  Recvdata[1] = 0;
  Recvdata[2] = 0;


  rsd = comm->Isend(senddata + 0,  1, Myf<base_type>::Datatype(), DownP, 0);
  rru = comm->Irecv(Recvdata + 0,  1, Myf<base_type>::Datatype(), UpP,   0);
  rsu = comm->Isend(senddata + 1,  2, Myf<base_type>::Datatype(), UpP,   1);
  rrd = comm->Irecv(Recvdata + 1,  2, Myf<base_type>::Datatype(), DownP, 1);

  if(!OVERLAP)
  {
    rsd.Wait();
    rsu.Wait();
    rru.Wait();
    rrd.Wait();
  }

}
  
template<class BaseType, class Int>
void
BDom<BaseType, Int>::
FinishFwdM4(Index3dBase<Int>const & Id, Index3dBase<Int>const & Iu)
{
  if(OVERLAP)
  {
    rsd.Wait();
    rsu.Wait();
    rru.Wait();
    rrd.Wait();
  }

  this->Up(Id)[DownEl]    += Recvdata[1];
  this->RightFront(Id)[LeftEl] = Recvdata[2];
      
  this->Up(Iu)[DownEl] += Recvdata[0];

}
  
template<class BaseType, class Int>
void
BDom<BaseType, Int>::
StartBwdM4(Index3dBase<Int>const & Id, Index3dBase<Int>const & Iu)
{
  Index3dBase<Int> I(Id);
  senddata[0] = this->Up(Id)[DownEl];
  senddata[1] = (*this)[Iu][DownEl]; 
  senddata[2] = (*this)[Iu][FrontEl];
  Recvdata[0] = 0;
  Recvdata[1] = 0;
  Recvdata[2] = 0;


  rsd = comm->Isend(senddata + 0,  1, Myf<base_type>::Datatype(), DownP, 0);
  rru = comm->Irecv(Recvdata + 0,  1, Myf<base_type>::Datatype(), UpP,   0);
  rsu = comm->Isend(senddata + 1,  2, Myf<base_type>::Datatype(), UpP,   1);
  rrd = comm->Irecv(Recvdata + 1,  2, Myf<base_type>::Datatype(), DownP, 1);

  if(!OVERLAP)
  {
    rsd.Wait();
    rsu.Wait();
    rru.Wait();
    rrd.Wait();
  }
}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
FinishBwdM4(Index3dBase<Int>const & Id, Index3dBase<Int>const & Iu)
{
  if(OVERLAP)
  {
    rsd.Wait();
    rsu.Wait();
    rru.Wait();
    rrd.Wait();
  }

  this->Up(Iu)[DownEl] = Recvdata[0];

  (*this)[Id][DownEl] = Recvdata[1];
  (*this)[Id][FrontEl] = Recvdata[2];
}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
StartFwdFront(Index3dBase<Int>const & Id, Index3dBase<Int>const & Iu)
{
  Recvdata[0] = 0;Recvdata[1] = 0;
  senddata[0] = (*this)[Id][FrontEl];
  senddata[1] = (*this)[Id][DownEl];
  rsd = comm->Isend(senddata, 2, Myf<base_type>::Datatype(), DownP, 0);
  rru = comm->Irecv(Recvdata, 2, Myf<base_type>::Datatype(), UpP,0);
  if(!OVERLAP)
  {
    rsd.Wait();
    rru.Wait();
  }
      
}
  
template<class BaseType, class Int>
void
BDom<BaseType, Int>::
FinishFwdFront(Index3dBase<Int>const & Id, Index3dBase<Int>const & Iu)
{
  if(OVERLAP)
  {
    rsd.Wait();
    rru.Wait();
  }
  (*this)[Iu][FrontEl] += Recvdata[0];
  (*this)[Iu][DownEl] += Recvdata[1];
} 

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
StartFwdDown(Index3dBase<Int> const & Id, Index3dBase<Int> const & Iu)
{
  Recvdata[0] = 0;
  using namespace std;
	
  rsu = comm->Isend(&((*this)[Iu]), 1, DtDown, UpP, 0);
  rrd = comm->Irecv(Recvdata, 1, Myf<base_type>::Datatype(), DownP,0);
  if(!OVERLAP)
  {
    rsu.Wait();
    rrd.Wait();
  }
}
  
template<class BaseType, class Int>
void
BDom<BaseType, Int>::
FinishFwdDown(Index3dBase<Int> Id, Index3dBase<Int> Iu)
{
  if(OVERLAP)
  {
    rsu.Wait();
    rrd.Wait();
  }
  
  (*this)[Id][DownEl] += Recvdata[0];
}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
StartFwdLeft(Index3dBase<Int> const & Id, Index3dBase<Int> const & Iu)
{
      
  if(RecvdataLong.size()<static_cast<size_t>(N2))
    RecvdataLong.resize(N2);
  std::fill(RecvdataLong.begin(), RecvdataLong.end(), base_type(0));
  rsu = comm->Isend(&((*this)[Iu]), 1, DtLeft_line, UpP, 0);
  rrd = comm->Irecv(&RecvdataLong[0], N2, Myf<base_type>::Datatype(), DownP,0);
  if(!OVERLAP)
  {
    rsu.Wait();
    rrd.Wait();
  }
}
  
template<class BaseType, class Int>
void
BDom<BaseType, Int>::
FinishFwdLeft(Index3dBase<Int> const & Id, Index3dBase<Int> const & Iu)
{
  if(OVERLAP)
  {
    rsu.Wait();
    rrd.Wait();
  }
  
  element_type *d = &(*this)[Id];
  base_type *s = &RecvdataLong[0];
  ptrdiff_t step = &this->Back(Id) - d;
  for(Int i = N2; i; --i, ++s, d+=step)
  {
    //	  if(rank == 1)
    //std::cout <<"Left " << (*d)[LeftEl] << " " << *s << std::endl;
    (*d)[LeftEl] += *s;
  }
}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
StartBwdLeft(Index3dBase<Int> const& Id, Index3dBase<Int> const & Iu)
{
  rsd = comm->Isend(&(*this)[Id], 1, DtLeft_line, DownP, 0);
  rru = comm->Irecv(&(*this)[Iu], 1, DtLeft_line, UpP, 0);
  if(!OVERLAP)
  {
    rsd.Wait();
    rru.Wait();
  }
}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
FinishBwdLeft(Index3dBase<Int> const& Id, Index3dBase<Int> const &Iu)
{
  if(OVERLAP)
  {
    rsd.Wait();
    rru.Wait();
  }
  
}
  
template<class BaseType, class Int>
void
BDom<BaseType, Int>::
StartBwdFront(Index3dBase<Int> const& Id, Index3dBase<Int> const &Iu)
{
  rsu = comm->Isend(&(*this)[Iu], 1, DtFront, UpP, 0);
  rrd = comm->Irecv(&(*this)[Id], 1, DtFront, DownP, 0);
  if(!OVERLAP)
  {
    rsu.Wait();
    rrd.Wait();
  }
}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
FinishBwdFront(Index3dBase<Int> const& Id, Index3dBase<Int> const&Iu)
{
  if(OVERLAP)
  {
    rsu.Wait();
    rrd.Wait();
  }
}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
StartBwdDown(Index3dBase<Int> const &Id, Index3dBase<Int> const&Iu)
{
  rsd = comm->Isend(&this->Up(Id), 1, DtDown, DownP, 0);
  rru = comm->Irecv(&this->Up(Iu), 1, DtDown, UpP, 0);
  rsu = comm->Isend(&(*this)[Iu],  1, DtDown, UpP,   1);
  rrd = comm->Irecv(&(*this)[Id],  1, DtDown, DownP, 1);
  if(!OVERLAP)
  {
    rsd.Wait();
    rsu.Wait();
    rru.Wait();
    rrd.Wait();
  }
}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
FinishBwdDown(Index3dBase<Int> const& Id, Index3dBase<Int> const&Iu)
{
  if(OVERLAP)
  {
    rsd.Wait();
    rsu.Wait();
    rru.Wait();
    rrd.Wait();
  }
}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
StartFwdLineLeft(Index3dBase<Int> const &Id, Index3dBase<Int> const&Iu)
{
  if(RecvdataLong2.size()<static_cast<size_t>(2*N2 + 1))
    RecvdataLong2.resize(2*N2+1);
  //  std::cout << "srend " << PDownP << " " << PUpP << std::endl;
  rsd = comm->Isend(&(*this)[Id], 1, DtFront_down, PDownP, 0);
  rru = comm->Irecv(&RecvdataLong2[0], 2*N2+1, Myf<base_type>::Datatype(), PUpP, 0);
  if(!OVERLAP)
  {
    rsd.Wait();
    rru.Wait();
  }
}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
FinishFwdLineLeft(Index3dBase<Int> Id, Index3dBase<Int> Iu) // we need these variables so we pass by value
{
  if (OVERLAP)
  {
    rsd.Wait();
    rru.Wait();
  }
  int j;
  for(j = 0; j<N2; ++j, Iu.ToBack())
  {
    (*this)[Iu][DownEl]  += RecvdataLong2[j + N2 +1];        // if we change DTFront_down we'll have to change here as well
    (*this)[Iu][FrontEl] += RecvdataLong2[j];
  }
  (*this)[Iu][FrontEl] += RecvdataLong2[j];
//   for (int i  =0; i < 2*N2+1; ++i)
//     std::cout << RecvdataLong2[i] << " ";
//   std::cout <<std::endl;
}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
StartFwdLineMiddle(Index3dBase<Int> const &Id, Index3dBase<Int> const&Iu)
{
  if(RecvdataLong.size()<static_cast<size_t>(N2))
    RecvdataLong.resize(N2);

  rsu = comm->Isend(&(*this)[Iu], 1, DtLeft_line, PUpP, 1);
  rrd = comm->Irecv(&RecvdataLong[0], N2, Myf<base_type>::Datatype(), PDownP, 1);
  if(!OVERLAP)
  {
    rsu.Wait();
    rrd.Wait();
  }
}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
FinishFwdLineMiddle(Index3dBase<Int> Id, Index3dBase<Int> Iu) // we need these variables so we pass by value
{
  if(OVERLAP)
  {
    rsu.Wait();
    rrd.Wait();
  }
  for(int j = 0; j<N2; ++j, Id.ToBack())
  {
    (*this)[Id][LeftEl]  += RecvdataLong[j];
  }
}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
StartBwdLineLeft(Index3dBase<Int> const &Id, Index3dBase<Int> const&Iu)
{
  rsd = comm->Isend(&(*this)[Id], 1, DtLeft_line, PDownP, 0);
  rru = comm->Irecv(&(*this)[Iu], 1, DtLeft_line, PUpP, 0);
  if(!OVERLAP)
  {
    rsd.Wait();
    rru.Wait();
  }
}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
FinishBwdLineLeft(Index3dBase<Int> const &Id, Index3dBase<Int> const&Iu)
{
  if(OVERLAP)
  {
    rsd.Wait();
    rru.Wait();
  }
}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
StartBwdLineMiddle(Index3dBase<Int> const &Id, Index3dBase<Int> const&Iu)
{
  rsu = comm->Isend(&(*this)[Iu], 1, DtFront_down, PUpP, 1);
  rrd = comm->Irecv(&(*this)[Id], 1, DtFront_down, PDownP, 1);
  if(!OVERLAP)
  {
    rsu.Wait();
    rrd.Wait();
  }
}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
FinishBwdLineMiddle(Index3dBase<Int> const &Id, Index3dBase<Int> const&Iu)
{
  if(OVERLAP)
  {
    rsu.Wait();
    rrd.Wait();
  }
}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
StartDownLeftPlane()
{
  Index3dBase<Int> Id(N1, N2, N3), Iu(N1, N2, N3);
      
      
  Id.SetPos(0,0,-1);
  Iu.SetPos(0,0,N3-1);

  rsd = comm->Isend(&((*this)[Id]), 1, DtLeft_line, DownP, 0);
  rru = comm->Irecv(&((*this)[Iu]), 1, DtLeft_line, UpP,0);
  if(!OVERLAP)
  {
    rsd.Wait();
    rru.Wait();
  }
}



template<class BaseType, class Int>
void
BDom<BaseType, Int>::
FinishDownLeftPlane()
{
  if(OVERLAP)
  {
    rsd.Wait();
    rru.Wait();
  }
}
template<class BaseType, class Int>
void
BDom<BaseType, Int>::
DoCyclo(int dir)
{

  //  dir == 0 -  N -> 0
  //  dir == 1 -  0 -> N;
  Index3dBase<Int> Id(N1, N2, N3), Iu(N1, N2, N3);
  rsd = MPI::REQUEST_NULL;
  rsu = rsd;
  Id.SetPos(0,0,-1);
  Iu.SetPos(0,0,ON3-1);

  if (rank == 0)
  {
    if  (dir == 0)
      rru = comm->Irecv(&((*this)[Id]), 1, DtUp_plane, size-1,0);
    if (dir == 1)
      rru = comm->Isend(&((*this)[Id]), 1, DtUp_plane, size-1,0);
    if (dir == 2)
      rru = comm->Irecv(&((*this)[Id]), 1, DtDown_plane, size-1,0);
  }

  if (rank == (size-1))
  {
    if  (dir == 0)
      rsd = comm->Isend(&((*this)[Iu]), 1, DtUp_plane, 0, 0);
    if (dir == 1)
      rsd = comm->Irecv(&((*this)[Iu]), 1, DtUp_plane, 0, 0);
    if (dir == 2)
      rsd = comm->Isend(&((*this)[Iu]), 1, DtDown_plane, 0,0);
  }

  rsd.Wait();
  rru.Wait();

}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
StartSendRecvh1()
{
  Index3dBase<Int> Id(N1, N2, N3), Iu(N1, N2, N3);
    

  Id.SetPos(0,0,-1);
  Iu.SetPos(0,0,N3-1);
  rsd = comm->Isend(&((*this)[Id]), 1, DtDown_left_plane, DownP, 0);
  rru = comm->Irecv(&((*this)[Iu]), 1, DtDown_left_plane, UpP,0);
  if(!OVERLAP)
  {
    rsd.Wait();
    rru.Wait();
  }
}
template<class BaseType, class Int>
void
BDom<BaseType, Int>::
FinishSendRecvh1()
{
  if(OVERLAP)
  {
    rsd.Wait();
    rru.Wait();
  }
}


template<class BaseType, class Int>
void
BDom<BaseType, Int>::
StartSendRecvh2()
{
  Index3dBase<Int> Id(N1, N2, N3), Iu(N1, N2, N3);
    

  Id.SetPos(0,0,-1);
  Iu.SetPos(0,0,N3-1);
  rsd = comm->Isend(&((*this)[Iu]), 1, DtDown_left_plane, UpP, 0);
  rru = comm->Irecv(&((*this)[Id]), 1, DtDown_left_plane, DownP,0);
  if(!OVERLAP)
  {
    rsd.Wait();
    rru.Wait();
  }
}
template<class BaseType, class Int>
void
BDom<BaseType, Int>::
FinishSendRecvh2()
{
  if(OVERLAP)
  {
    rsd.Wait();
    rru.Wait();
  }
}







template<class BaseType, class Int>
void
BDom<BaseType, Int>::
StartSendRecv()
{
  Index3dBase<Int> Id(N1, N2, N3), Iu(N1, N2, N3);
    

  Id.SetPos(0,0,-1);
  Iu.SetPos(0,0,N3-1);

  rsd = comm->Isend(&((*this)[Id]), 1, DtUp_plane, DownP, 0);
  rru = comm->Irecv(&((*this)[Iu]), 1, DtUp_plane, UpP,0);
  rsu = comm->Isend(&((*this)[Iu]), 1, DtDown_plane, UpP, 1);
  rrd = comm->Irecv(&((*this)[Id]), 1, DtDown_plane, DownP, 1);
  if(!OVERLAP)
  {
    rsu.Wait();
    rsd.Wait();
    rru.Wait();
    rrd.Wait();
  }
}
template<class BaseType, class Int>
void
BDom<BaseType, Int>::
SendRecvPBC()
{
  Index3dBase<Int> Id(N1, N2, N3), Iu(N1, N2, N3);
    

  Id.SetPos(0,0,-1);
  Iu.SetPos(0,0,ON3-1);

  rsd = comm->Isend(&((*this)[Id]), 1, DtUp_plane, PDownP, 0);
  rru = comm->Irecv(&((*this)[Iu]), 1, DtUp_plane, PUpP,0);
  rsu = comm->Isend(&((*this)[Iu]), 1, DtDown_plane, PUpP, 1);
  rrd = comm->Irecv(&((*this)[Id]), 1, DtDown_plane, PDownP, 1);
  rsu.Wait();
  rsd.Wait();
  rru.Wait();
    rrd.Wait();
}
template<class BaseType, class Int>
void
BDom<BaseType, Int>::
FinishSendRecv()
{
  if(OVERLAP)
  {
    rsu.Wait();
    rsd.Wait();
    rru.Wait();
    rrd.Wait();
  }
}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
Pack()
{
  if((rank+1)!=size)
    if (SendUp)
    {
      Index3dBase<Int> I1(N1, N2, N3), I2(N1, N2, N3);
      Index3dBase<Int> J1(N1, N2, N3), J2(N1, N2, N3);
      int ind = 0;
      unsigned int count = N1*N2 + (N1+1)*N2;
      if (BufUp.size()<count)
	BufUp.resize(count);
    
      I1.SetPos(0,0,N3-1);
      I2.SetPos(0,0,N3);
      int i,  j;
      for (i = 0;i < N1; ++i, I1.ToRight(),I2.ToRight())
      {
	for(j = 0, J1 = I1, J2 = I2; j<N2; ++j, J1.ToBack(), J2.ToBack())
	{
	  BufUp[ind++] = (*this)[J1][LeftEl];
	  BufUp[ind++] = (*this)[J2][DownEl];
	}
      }
      for(j = 0, J1 = I1; j<N2; ++j, J1.ToBack())
      {
	BufUp[ind++] = (*this)[J1][LeftEl];
      }
    }else
    {
      unsigned int count = N1*N2 + N1*(N2+1);
      if (BufUp.size()<count)
	BufUp.resize(count);
    }
  if (rank)
    if (SendDown)
    {
      Index3dBase<Int> I1(N1, N2, N3);
      Index3dBase<Int> J1(N1, N2, N3);
      int ind = 0;
      unsigned int count = N1*N2 + N1*(N2+1);
      if (BufDown.size()<count)
	BufDown.resize(count);
    
      I1.SetPos(0,0,-1);
      int i,  j;
      for (i = 0;i < N1; ++i, I1.ToRight())
      {
	for(j = 0, J1 = I1; j<N2; ++j, J1.ToBack())
	{
	  BufDown[ind++] = (*this)[J1][FrontEl];
	  BufDown[ind++] = (*this)[J1][DownEl];
	}
	BufDown[ind++] = (*this)[J1][FrontEl];
      }
    }else
    {
      unsigned int count = N1*N2 + (N1+1)*N2;
      if (BufDown.size()<count)
	BufDown.resize(count);
    }
}


template<class BaseType, class Int>
void
BDom<BaseType, Int>::
UnPackAdd()
{
  if((rank+1)!=size)
    if (!SendUp) // receive from up
    {
      Index3dBase<Int> I1(N1, N2, N3);
      Index3dBase<Int> J1(N1, N2, N3);
      int ind = 0;
      I1.SetPos(0,0,N3-1);
      int i,  j;
      for (i = 0;i < N1; ++i, I1.ToRight())
      {
	for(j = 0, J1 = I1; j<N2; ++j, J1.ToBack())
	{
	  (*this)[J1][FrontEl] += BufUp[ind++];
	  (*this)[J1][DownEl]  += BufUp[ind++];
	}
	(*this)[J1][FrontEl] += BufUp[ind++];
      }
    }
  
  if(rank)
    if (!SendDown) // receive from down
    {
      Index3dBase<Int> I1(N1, N2, N3), I2(N1, N2, N3);
      Index3dBase<Int> J1(N1, N2, N3), J2(N1, N2, N3);
      int ind = 0;
    
      I1.SetPos(0,0,-1);
      I2.SetPos(0,0,0);
      int i,  j;
      for (i = 0;i < N1; ++i, I1.ToRight(),I2.ToRight())
      {
	for(j = 0, J1 = I1, J2 = I2; j<N2; ++j, J1.ToBack(), J2.ToBack())
	{
	  (*this)[J1][LeftEl] += BufDown[ind++];
	  (*this)[J2][DownEl] += BufDown[ind++];
	}
      }
      for(j = 0, J1 = I1; j<N2; ++j, J1.ToBack())
      {
	(*this)[J1][LeftEl] += BufDown[ind++];
      }
    }
}






template<class BaseType, class Int>
void
BDom<BaseType, Int>::
Pack2()
{
  unsigned int count = N1*N2;
  if (BufUp.size()<count)
    BufUp.resize(count);
  if (BufDown.size()<count)
    BufDown.resize(count);
  if (BufUpR.size()<count)
    BufUpR.resize(count);
  if (BufDownR.size()<count)
    BufDownR.resize(count);

  Index3dBase<Int> I1(N1, N2, N3);
  Index3dBase<Int> J1(N1, N2, N3);
  I1.SetPos(0,0,N3-1);
  int i,  j;
  int ind = 0;
  if ((rank+1)==size)
    for (i = 0;i < N1; ++i, I1.ToRight())
    {
      for(j = 0, J1 = I1; j<N2; ++j, J1.ToBack())
      {
	BufUp[ind++] = (*this)[J1][DownEl];
      }
    }
  I1.SetPos(0,0,0);

  ind = 0;
  if (rank==0)
    for (i = 0;i < N1; ++i, I1.ToRight())
    {
      for(j = 0, J1 = I1; j<N2; ++j, J1.ToBack())
      {
	BufDown[ind++] = (*this)[J1][DownEl];
      }
    }
}
template<class BaseType, class Int>
void
BDom<BaseType, Int>::
CopyLeft(int v)
{
  Index3dBase<Int> I1(N1, N2, N3), I2(N1, N2, N3);
  Index3dBase<Int> J1(N1, N2, N3), J2(N1, N2, N3);
  Index3dBase<Int> K1(N1, N2, N3), K2(N1, N2, N3);
  I1.SetPos(0,0,0);
  I2.SetPos(0,N2,0);
  int k, j;
  J1.SetPos(0,0,-1);
  J2.SetPos(N1,0,-1);
  for (j = 0;j < N2; ++j, J1.ToBack(),J2.ToBack())
  {
    for(k = 0, K1 = J1, K2 = J2; k<=N3; ++k, K1.ToUp(), K2.ToUp())
    {
      if(v==0)
	(*this)[K2][LeftEl] = (*this)[K1][LeftEl];
      if(v==2)
	(*this)[K2][LeftEl] = 0;//(*this)[K1][LeftEl];
    }
  }


  I1.SetPos(0,0,0);
  I2.SetPos(0,N2,0);
  int i;
  for (i = 0;i < N1; ++i, I1.ToRight(),I2.ToRight())
  {
    for(k = 0, K1 = I1, K2 = I2; k<N3; ++k, K1.ToUp(), K2.ToUp())
    {
      (*this)[K2][FrontEl] = 0;
    }
  }

}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
AddSides(int add)
{
  //  std::cout << N1 << " " << N2 << " " << N3 << std::endl;
  Index3dBase<Int> I1(N1, N2, N3), I2(N1, N2, N3);
  Index3dBase<Int> J1(N1, N2, N3), J2(N1, N2, N3);
  Index3dBase<Int> K1(N1, N2, N3), K2(N1, N2, N3);
  I1.SetPos(0,0,0);
  I2.SetPos(0,N2,0);
  int  k;
  BaseType a,b,c;
  int i;
  // l, r previos // this is the two opposite sides 
  // L, R result
  // 0 : L = l; R = l;
  // 1 : L = l+r; R = 0;
  // 2 : L = l; R = 0;
  // 3 : // matrix fix L = l+r; R = 1 on the diagonal, 0 else
  // 4 :  F= f; B = 0; L=l R=R;
  for (i = 0;i < N1; ++i, I1.ToRight(),I2.ToRight())
  {
    for(k = 0, K1 = I1, K2 = I2; k<N3; ++k, K1.ToUp(), K2.ToUp())
    {
      if (add&1)
      {
	
	b = (*this)[K1][FrontEl];
	c = (*this)[K2][FrontEl];
	a = b;
	a += c;
	(*this)[K1][FrontEl] = a;
	if (add==1)
	  (*this)[K2][FrontEl] = 0;
	if (add==3)
	  Myf<BaseType>::Onify((*this)[K2][FrontEl]);     // fix for the matrix
	  
      }else if(add==0)
      {
	(*this)[K2][FrontEl] = 	(*this)[K1][FrontEl];
	a = (*this)[K2][FrontEl];
	//std::cout <<  a << std::endl;
      }else if((add==2)||(add==4))
      {
	(*this)[K2][FrontEl] = 0;
      }
    }
  }
  int j;
  J1.SetPos(0,0,-1);
  J2.SetPos(N1,0,-1);
  for (j = 0;j < N2; ++j, J1.ToBack(),J2.ToBack())
  {
    for(k = 0, K1 = J1, K2 = J2; k<=N3; ++k, K1.ToUp(), K2.ToUp())
    {
      if (add&1)
      {
	b =  (*this)[K1][LeftEl];
	c = (*this)[K2][LeftEl];
	a = b;
	a += c;
	(*this)[K1][LeftEl] = a;
	if (add==1)
	  (*this)[K2][LeftEl] = 0;
	if (add==3)
	  Myf<BaseType>::Onify( (*this)[K2][LeftEl]);
      }else if (add==0)
      {
	(*this)[K2][LeftEl] = (*this)[K1][LeftEl];
      }
      else if(add==2)
	(*this)[K2][LeftEl] = 0;
    }
  }

  
//   I1.SetPos(0,0,0);
//   I2.SetPos(0,0,N3-1);
//   for (i = 0;i < N1; ++i, I1.ToRight(),I2.ToRight())
//   {
//     for (j = 0,J1=I1,J2=I2;j < N2; ++j, J1.ToBack(),J2.ToBack())
//     {
//       if (add==1)
//       {
// 	a = (*this)[J1][DownEl];
// 	a+= (*this)[J2][DownEl];
// 	(*this)[J1][DownEl] = a;
// 	(*this)[J2][DownEl] = 0;
//       }else if (add==0)
//       {
// 	(*this)[J2][DownEl] = (*this)[J1][DownEl];
//       }
//       else if(add==2)
// 	(*this)[J2][DownEl] = 0;
//     }
//   }

}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
AddSidesM0(int add)
{
  //  std::cout << N1 << " " << N2 << " " << N3 << std::endl;
  Index3dBase<Int> I1(N1, N2, N3), I2(N1, N2, N3);
  Index3dBase<Int> J1(N1, N2, N3), J2(N1, N2, N3);
  Index3dBase<Int> K1(N1, N2, N3), K2(N1, N2, N3);
  I1.SetPos(0,-1,0);
  I2.SetPos(0,N2-1,0);
  int  k;
  BaseType a,b,c;

  int i;
  for (i = 0;i < N1; ++i, I1.ToRight(),I2.ToRight())
  {
    for(k = 0, K1 = I1, K2 = I2; k<N3; ++k, K1.ToUp(), K2.ToUp())
    {
      (*this)[K1] = (*this)[K2];
    }
  }
  int j;
  J1.SetPos(-1,0,-1);
  J2.SetPos(N1-1,0,-1);
  for (j = 0;j < N2; ++j, J1.ToBack(),J2.ToBack())
  {
    for(k = 0, K1 = J1, K2 = J2; k<=N3; ++k, K1.ToUp(), K2.ToUp())
    {
      (*this)[K1] = (*this)[K2];
    }
  }

}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
AddBoundarys(int add)
{
  StartSendRecvAdd2();
  AddSides(add);
  FinishSendRecvAdd2(add);
  
}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
AddBoundarysM0(int add)
{
  StartSendRecvAddM0();
  AddSidesM0(add);
  FinishSendRecvAddM0(add);
  
}
template<class BaseType, class Int>
void
BDom<BaseType, Int>::
UnPackAdd2(int add)
{
  Index3dBase<Int> I1(N1, N2, N3);
  Index3dBase<Int> J1(N1, N2, N3);
  I1.SetPos(0,0,N3-1);
  int i,  j;
  int ind = 0;
  if ((rank+1)==size)
    for (i = 0;i < N1; ++i, I1.ToRight())
    {
      for(j = 0, J1 = I1; j<N2; ++j, J1.ToBack())
      {
	if((add&1)||(add==2))
	{
	  if (add!=3)
	    (*this)[J1][DownEl]= 0;
	    if (add==3)
	      Myf<BaseType>::Onify((*this)[J1][DownEl]);
	}
	else if (add==0)
	  (*this)[J1][DownEl] = BufUpR[ind++];
      }
    }
  I1.SetPos(0,0,0);
  ind = 0;
  if (rank==0)
    for (i = 0;i < N1; ++i, I1.ToRight())
    {
      for(j = 0, J1 = I1; j<N2; ++j, J1.ToBack())
      {
	if (add&1) //1 or 3
	{
	  (*this)[J1][DownEl]+=BufDownR[ind++];
	}//else if ((add==0)||(add==2))	  ;
      }
    }
}

// template<class BaseType, class Int>
// void
// BDom<BaseType, Int>::
// CopyBoundary( BDom&a)
// {
// //   if (Options::GetOptions().PBC)
// //   {
// //     a.StartSendRecvAdd2();
// //     a.FinishSendRecvAdd2(2);
// //     Index3dBase<Int> I1(N1, N2, N3);
// //     Index3dBase<Int> J1(N1, N2, N3);
// //     int i,  j;
// //     I1.SetPos(0,0,0);
// //     if (rank==0)
// //       for (i = 0;i < N1; ++i, I1.ToRight())
// //       {
// // 	for(j = 0, J1 = I1; j<N2; ++j, J1.ToBack())
// // 	{
// // 	  (*this)[J1][DownEl]=a[J1][DownEl];
// // 	}
// //       }
// //   }
// }
  








template<class BaseType, class Int>
void
BDom<BaseType, Int>::
UnPackNoAdd()
{
  if((rank+1)!=size)
    if (!SendUp) // receive from up
    {
      Index3dBase<Int> I1(N1, N2, N3);
      Index3dBase<Int> J1(N1, N2, N3);
      int ind = 0;
      I1.SetPos(0,0,N3-1);
      int i,  j;
      for (i = 0;i < N1; ++i, I1.ToRight())
      {
	for(j = 0, J1 = I1; j<N2; ++j, J1.ToBack())
	{
	  (*this)[J1][FrontEl] = BufUp[ind++];
	  (*this)[J1][DownEl]  = BufUp[ind++];
	}
	(*this)[J1][FrontEl] = BufUp[ind++];
      }
    }
  
  if(rank)
    if (!SendDown) // receive from down
    {
      Index3dBase<Int> I1(N1, N2, N3), I2(N1, N2, N3);
      Index3dBase<Int> J1(N1, N2, N3), J2(N1, N2, N3);
      int ind = 0;
    
      I1.SetPos(0,0,-1);
      I2.SetPos(0,0,0);
      int i,  j;
      for (i = 0;i < N1; ++i, I1.ToRight(),I2.ToRight())
      {
	for(j = 0, J1 = I1, J2 = I2; j<N2; ++j, J1.ToBack(), J2.ToBack())
	{
	  (*this)[J1][LeftEl] = BufDown[ind++];
	  (*this)[J2][DownEl] = BufDown[ind++];
	}
      }
      for(j = 0, J1 = I1; j<N2; ++j, J1.ToBack())
      {
	(*this)[J1][LeftEl] = BufDown[ind++];
      }
    }
}

template<class BaseType, class Int>
void
BDom<BaseType, Int>::
AddParts()
{
  int countfrontdown = N1*N2 + N1*(N2+1);
  int countleftdown = N1*N2 + (N1+1)*N2;
  {
    Index3dBase<Int> I1(N1, N2, N3), I2(N1, N2, N3);
    Index3dBase<Int> J1(N1, N2, N3), J2(N1, N2, N3);
    int ind = 0;
    int count =countleftdown;
    if (BufUp.size()<count)
      BufUp.resize(count);
    if (BufDownR.size()<count)
      BufDownR.resize(count);
    
    I1.SetPos(0,0,ON3-1);
    I2.SetPos(0,0,ON3);
    int i,  j;
    for (i = 0;i < N1; ++i, I1.ToRight(),I2.ToRight())
    {
      for(j = 0, J1 = I1, J2 = I2; j<N2; ++j, J1.ToBack(), J2.ToBack())
      {
	BufUp[ind++] = (*this)[J1][LeftEl];
	BufUp[ind++] = (*this)[J2][DownEl];
      }
    }
    for(j = 0, J1 = I1; j<N2; ++j, J1.ToBack())
    {
      BufUp[ind++] = (*this)[J1][LeftEl];
    }
  }
  {
    Index3dBase<Int> I1(N1, N2, N3);
    Index3dBase<Int> J1(N1, N2, N3);
    int ind = 0;
    unsigned int count = countfrontdown;
    if (BufDown.size()<count)
      BufDown.resize(count);
    if (BufUpR.size()<count)
      BufUpR.resize(count);
    
    I1.SetPos(0,0,-1);
    int i,  j;
    for (i = 0;i < N1; ++i, I1.ToRight())
    {
      for(j = 0, J1 = I1; j<N2; ++j, J1.ToBack())
      {
	BufDown[ind++] = (*this)[J1][FrontEl];
	BufDown[ind++] = (*this)[J1][DownEl];
      }
	BufDown[ind++] = (*this)[J1][FrontEl];
    }
  }


  rsu = comm->Isend(&(BufUp[0]), countleftdown,
		       Myf<base_type>::Datatype(),
		       NextP, 0);

  rru = comm->Irecv(&(BufUpR[0]), countfrontdown,
		       Myf<base_type>::Datatype(),
		       NextP, 1);

  rsd = comm->Isend(&(BufDown[0]), countfrontdown,
		       Myf<base_type>::Datatype(),
		       PrevP, 1);
  rrd = comm->Irecv(&(BufDownR[0]), countleftdown,
		       Myf<base_type>::Datatype(),
		       PrevP, 0);

  rsu.Wait();
  rsd.Wait();
  rru.Wait();
  rrd.Wait();


  {
    Index3dBase<Int> I1(N1, N2, N3), I2(N1, N2, N3);
    Index3dBase<Int> J1(N1, N2, N3), J2(N1, N2, N3);
    int ind=0;
    
    I1.SetPos(0,0,-1);
    I2.SetPos(0,0,0);
    int i,  j;
    for (i = 0;i < N1; ++i, I1.ToRight(),I2.ToRight())
    {
      for(j = 0, J1 = I1, J2 = I2; j<N2; ++j, J1.ToBack(), J2.ToBack())
      {
	(*this)[J1][LeftEl] += BufDownR[ind++];
	(*this)[J2][DownEl] += BufDownR[ind++];
	//	std::cout << BufDownR[ind-1] << std::endl;
      }
    }
    for(j = 0, J1 = I1; j<N2; ++j, J1.ToBack())
    {
      (*this)[J1][LeftEl] += BufDownR[ind++];
    }
  }
  {
    Index3dBase<Int> I1(N1, N2, N3);
    Index3dBase<Int> J1(N1, N2, N3);
    int ind = 0;
    
    I1.SetPos(0,0,ON3-1);
    int i,  j;
    for (i = 0;i < N1; ++i, I1.ToRight())
    {
      for(j = 0, J1 = I1; j<N2; ++j, J1.ToBack())
      {
	(*this)[J1][FrontEl] += BufUpR[ind++];
	(*this)[J1][DownEl]  += BufUpR[ind++];
      }
	(*this)[J1][FrontEl] += BufUpR[ind++];
    }
  }
}


template<class BaseType, class Int>
void
BDom<BaseType, Int>::
StartSendRecvAdd()
{
  rsd = MPI::REQUEST_NULL;
  rsu = rsd;
  Pack();
  int todowncount = N1*N2 + N1*(N2+1);
  int toupcount = N1*N2 + (N1+1)*N2;
  if ((rank+1)!=size)
    if(SendUp)
      rsu = comm->Isend(&(BufUp[0]), toupcount,
		       Myf<base_type>::Datatype(),
		       UpP, 0);
    else
      rsu = comm->Irecv(&(BufUp[0]), todowncount,
		       Myf<base_type>::Datatype(),
		       UpP, 1);

  if(rank)
    if(SendDown)
      rsd = comm->Isend(&(BufDown[0]), todowncount,
		       Myf<base_type>::Datatype(),
		       DownP, 1);
    else
      rsd = comm->Irecv(&(BufDown[0]), toupcount,
		       Myf<base_type>::Datatype(),
		       DownP, 0);
  if(!OVERLAP)
  {
    rsu.Wait();
    rsd.Wait();
  }
}
template<class BaseType, class Int>
void
BDom<BaseType, Int>::
FinishSendRecvAdd(bool add)
{
  if(OVERLAP)
  {
    rsu.Wait();
    rsd.Wait();
  }
  if (add)
    UnPackAdd();
  else
    UnPackNoAdd();
}
template<class BaseType, class Int>
void
BDom<BaseType, Int>::
StartSendRecvAdd2()
{
  Pack2();
  rsu = rsd = rru = rrd = MPI::REQUEST_NULL;
  if (rank==0)
  {
      rsu = comm->Isend(&(BufDown[0]), N1*N2,
		       Myf<base_type>::Datatype(),
		       size-1, 1);
      rru = comm->Irecv(&(BufDownR[0]), N1*N2,
		       Myf<base_type>::Datatype(),
		       size-1, 0);
  }
  if (rank==(size-1))
  {
      rsd = comm->Isend(&(BufUp[0]), N1*N2,
		       Myf<base_type>::Datatype(),
		       0, 0);
      rrd = comm->Irecv(&(BufUpR[0]), N1*N2,
		       Myf<base_type>::Datatype(),
		       0, 1);
  }
  if(!OVERLAP)
  {
    rsu.Wait();
    rsd.Wait();
    rru.Wait();
    rrd.Wait();
  }
}
template<class BaseType, class Int>
void
BDom<BaseType, Int>::
FinishSendRecvAdd2(int add)
{
  if(OVERLAP)
  {
    rsu.Wait();
    rsd.Wait();
    rru.Wait();
    rrd.Wait();
  }
  UnPackAdd2(add);
}


template<class BaseType, class Int>
void
BDom<BaseType, Int>::
StartSendRecvAddM0()
{
  Int i,j;
  rsu = rsd = rru = rrd = MPI::REQUEST_NULL;
  if (rank==(size-1))
  {
    Index3dBase<Int> I1(N1, N2, N3);
    Index3dBase<Int> J1(N1, N2, N3);
    I1.SetPos(0,0,N3-1);
    Int ind = 0;
    if (BufUp.size()<N1*N2*3)
      BufUp.resize(N1*N2*3);
    for (i = 0;i < N1; ++i, I1.ToRight())
    {
      for(j = 0, J1 = I1; j<N2; ++j, J1.ToBack())
      {
	BufUp[ind++] = (*this)[J1][DownEl];
	BufUp[ind++] = (*this)[J1][LeftEl];
	BufUp[ind++] = (*this)[J1][FrontEl];
      }
    }
    rsd = comm->Isend(&(BufUp[0]), N1*N2*3,
		      Myf<base_type>::Datatype(),
		      0, 0);
  }
  if (rank==0)
  {
    if (BufDownR.size()<N1*N2*3)
      BufDownR.resize(N1*N2*3);
      rru = comm->Irecv(&(BufDownR[0]), N1*N2*3,
		       Myf<base_type>::Datatype(),
		       size-1, 0);
  }
  
  if(!OVERLAP)
  {
    rsu.Wait();
    rsd.Wait();
    rru.Wait();
    rrd.Wait();
  }
}
template<class BaseType, class Int>
void
BDom<BaseType, Int>::
FinishSendRecvAddM0(int add)
{
  if(OVERLAP)
  {
    rsu.Wait();
    rsd.Wait();
    rru.Wait();
    rrd.Wait();
  }
  Int i,j;
  Index3dBase<Int> I1(N1, N2, N3);
  Index3dBase<Int> J1(N1, N2, N3);
  I1.SetPos(0,0,-1);
  Int ind = 0;
  if (rank==0)
    for (i = 0;i < N1; ++i, I1.ToRight())
    {
      for(j = 0, J1 = I1; j<N2; ++j, J1.ToBack())
      {
	(*this)[J1][DownEl]+=BufDownR[ind++];
	(*this)[J1][LeftEl]+=BufDownR[ind++];
	(*this)[J1][FrontEl]+=BufDownR[ind++];
      }
    }
  
}




template<class T, class Int>
 void ClearMatrix(BDom<SmallVector<T, 6>, Int > & mat)
{
  for(Int i = 0; i < mat.Size; ++i)
  {
    for(Int k = 0; k < 3; ++k)
      for(Int j = 0; j < 6; ++j)
        mat.Index(i)[k][j] = T(0);
  }
}

template<class T, class Int, template<class F, class G> class Vector>
 void ClearVector(Vector<T, Int> & vec, Int AAA )
{
  for(Int i = 0; i < vec.Size; ++i)
  {
    for(Int k = 0; k < 3; ++k)
      vec.Index(i)[k] = AAA;
  }
}

template<class T, class Int, template<class F, class G> class Vector>
 void ClearVectorRand(Vector<T, Int> & vec, Int AAA )
{
  for(Int i = 0; i < vec.Size; ++i)
  {
    for(Int k = 0; k < 3; ++k)
      vec.Index(i)[k] = (double)std::rand()/RAND_MAX;
  }
}





//vector static members
template<class BaseType, class Int> 
bool
BDom<BaseType, Int>::Dt_Created=false;
template<class BaseType, class Int> 
int
BDom<BaseType, Int>::Dt_refcnt=0;

template<class BaseType, class Int> 
MPI::Datatype 
BDom<BaseType, Int>::DtUp_plane;

template<class BaseType, class Int> 
MPI::Datatype 
BDom<BaseType, Int>::DtDown_plane;

template<class BaseType, class Int> 
MPI::Datatype 
BDom<BaseType, Int>::DtLeft;

template<class BaseType, class Int> 
MPI::Datatype 
BDom<BaseType, Int>::DtDown;

template<class BaseType, class Int> 
MPI::Datatype 
BDom<BaseType, Int>::DtFront;

template<class BaseType, class Int> 
MPI::Datatype 
BDom<BaseType, Int>::DtLeft_line;

template<class BaseType, class Int> 
MPI::Datatype 
BDom<BaseType, Int>::DtDown_line;

template<class BaseType, class Int> 
MPI::Datatype 
BDom<BaseType, Int>::DtFront_line;

template<class BaseType, class Int> 
MPI::Datatype 
BDom<BaseType, Int>::DtLeft_down;

template<class BaseType, class Int> 
MPI::Datatype 
BDom<BaseType, Int>::DtFront_down;

template<class BaseType, class Int> 
MPI::Datatype 
BDom<BaseType, Int>::DtUp_unfinished;

template<class BaseType, class Int> 
MPI::Datatype 
BDom<BaseType, Int>::DtDown_left_plane;





#endif
