c$$$      PROGRAM TEFULRC
c$$$      CALL SEFULRC
c$$$      END
C
      SUBROUTINE SEFULRC
C
C SEFULRC is a test subroutine for the EFULRC subroutine,
C which computes the eigenvalues and eigenvectors of
C a full, general, real matrix.
C It reads sample problem data from the file SEFULRC.Inn,
C writes solution data to the file SEFULRC.Onn, and writes
C PASS / FAIL diagnostic information to the file SEFULRC.Dnn.
C
C Required Numerical Analyst routines: DASUM, DCOPY, EFULRC
C
C Other required routines: ESETUP, EVALCOM, EVECCOM, 
C                          INTEST, MACON, TSTMSG
C
C Revision history:
C
C  Kevin W. Thompson     December 8, 1991     First version
C
C***********************************************************************
C
      IMPLICIT REAL*8 (A-H,O-Z)
      PARAMETER (NP=4,NWS=4,NR=50)
      REAL*8 A(NR,NR),ACOPY(NR,NR),EIGVAL(2,NR),EIGVEC(NP,NR,NR),
     1 WS(NR,NWS),EVALCH(2,NR),RELER1(NR),RELER2(NR),RELER3(NR),
     2 VALFAC(NR),EVALPAS(NR)
      INTEGER IWS(NR)
      CHARACTER*60 MES1(0:3),MES2(0:2),MES3(0:2)
C
C Define a character variable for error messages.
      CHARACTER*60 MES
C
C File unit numbers.
      INTEGER INFILE,OUTFILE,DIAGFILE
C
C Pass-fail indicators.
      LOGICAL PASS,FAIL
C
C Initialize relative error arrays to zero.
      DATA RELER1/NR*1.D0/,RELER2/NR*1.D0/,RELER3/NR*1.D0/
C
C Initialize option message arrays.
      DATA MES1/
     1 'IOPT1=0: Compute no eigenvectors.                          ',
     2 'IOPT1=1: Compute right eigenvectors only.                  ',
     3 'IOPT1=2: Compute left eigenvectors only.                   ',
     4 'IOPT1=3: Compute both right and left eigenvectors.         '/
      DATA MES2/
     1 'IOPT2=0: Store no eigenvectors.                            ',
     2 'IOPT2=1: Store eigenvectors in compressed form.            ',
     3 'IOPT2=2: Store eigenvectors in uncompressed form.          '/
      DATA MES3/
     1 'IOPT3=0: No error analysis.                                ',
     2 'IOPT3=1, IOPT1 < 3: Estimate relative error, no condition #',
     3 'IOPT3=1, IOPT1=3: Estimate relative error and condition #. '/
C
C Initialize the testing procedure and get the file unit numbers.
      CALL INTEST(INFILE,OUTFILE,DIAGFILE)
C
C Initialize PASS to .TRUE. and FAIL to .FALSE.
      PASS=.TRUE.
      FAIL=.FALSE.
C
C Read in the remaining data from the file.
      READ(INFILE,*)VECFAC
      READ(INFILE,*)RERBAS
      READ(INFILE,*)IOPT1
      READ(INFILE,*)IOPT2
      READ(INFILE,*)IOPT3
      READ(INFILE,*)N
      READ(INFILE,*)ITEST
      DO 10 K=1,N
        READ(INFILE,*)VALFAC(K)
10    CONTINUE
C Get the machine dependent constants.
      CALL MACON(PRECSN,RMIN,RMAX)
C
C Check the array dimensions and option parameters. 
C Generate a FAIL message and terminate if any are invalid.
      IF(N .GT. NR) THEN
        MES=
     1  'N is greater than NR.                                       '
        CALL TSTMSG(FAIL,DIAGFILE,MES)
      ENDIF
      IF(IOPT1 .LT. 0 .OR. IOPT1 .GT. 3) THEN
        MES=
     1  'IOPT1 is invalid                                            '
        CALL TSTMSG(FAIL,DIAGFILE,MES)
      ENDIF
      IF(IOPT2 .LT. 0 .OR. IOPT2 .GT. 2) THEN
        MES=
     1  'IOPT2 is invalid                                            '
        CALL TSTMSG(FAIL,DIAGFILE,MES)
      ENDIF
      IF(IOPT3 .LT. 0 .OR. IOPT3 .GT. 1) THEN
        MES=
     1  'IOPT3 is invalid                                            '
        CALL TSTMSG(FAIL,DIAGFILE,MES)
      ENDIF
C
C
C Print out the options selected for this calculation.
      INDEX3=IOPT3*(1+IOPT1/3)
      WRITE(OUTFILE,20)MES1(IOPT1),MES2(IOPT2),MES3(INDEX3)
20    FORMAT(' Eigenvalue-eigenvector test calculation.'//
     1 ' The options selected are:'/
     2 1X,A60/1X,A60/1X,A60/)
C Select the test according to the value of ITEST.
C
C Initialize the matrix A, and the known eigenvalues in EVALCH.
      CALL ESETUP(ITEST,A,ACOPY,EVALCH,NR,N)
C
C Compute the eigenvalues.
      CALL EFULRC(A,EIGVEC,EIGVAL,WS,IWS,NP,NR,N,NWS,
     1            IOPT1,IOPT2,IOPT3)
C
C Compare the computed eigenvalues to the known eigenvalues.
      CALL EVALCOM(N,EIGVAL,EVALCH,RELER1,AVGABS,AVGREL,RELMX1)
C
C Test the accuracy of the combined eigenvalue-eigenvector
C calculation.
C
      RELMX2=1.D0
      RELMX3=1.D0
      IF(IOPT1 .EQ. 1 .OR. IOPT1 .EQ. 3) THEN
C Right eigenvectors:
        IVEC=0
        CALL EVECCOM(NR,N,ACOPY,EIGVEC,EIGVAL,RELER2,RELMX2,IOPT1,IOPT2,
     1              NP,IVEC)
      ENDIF
      IF(IOPT1 .EQ. 2 .OR. IOPT1 .EQ. 3) THEN
C Left eigenvectors.
        IVEC=1
        CALL EVECCOM(NR,N,ACOPY,EIGVEC,EIGVAL,RELER3,RELMX3,IOPT1,IOPT2,
     1               NP,IVEC)
      ENDIF
C
C Compute the allowed relative errors for the eigenvalues.
      DO 190 K=1,N
        EVALPAS(K)=DMAX1(RERBAS,VALFAC(K)*PRECSN)
190   CONTINUE
C
C Print out the eigenvalues, iteration counts, and errors in the
C eigenvalues.
C
      WRITE(OUTFILE,149)
149   FORMAT(' Real parts of eigenvalues (sorted), computed and',
     1 ' known:'/)
      WRITE(OUTFILE,150)(I,EIGVAL(1,I),EVALCH(1,I),I=1,N)
      WRITE(OUTFILE,151)
151   FORMAT(/' Imaginary parts of eigenvalues (sorted), computed',
     1 ' and known:'/)
      WRITE(OUTFILE,150)(I,EIGVAL(2,I),EVALCH(2,I),I=1,N)
150   FORMAT(' Eigenvalue #', I3,' = ',1PE22.15,1X,1PE22.15)
      WRITE(OUTFILE,148)
148   FORMAT(/' Accuracy of Eigenvalues:'//
     1 7X,'Allowed',5X,'Observed',4X,'Estimated',3X,'Condition'/
     2 3X,'I',3X,'Error',7X,'Error',7X,'Error',7X,'Number')
      WRITE(OUTFILE,152)(I,EVALPAS(I),RELER1(I),WS(I,1),WS(I,2),
     1 I=1,N)
152   FORMAT(1X, I3,1P4E12.3)
      WRITE(OUTFILE,153)AVGABS,AVGREL,RELMX1
153   FORMAT(/' Average absolute error =',1PE10.3/
     1        ' Average relative error =',1PE10.3/
     2        ' Maximum relative error =',1PE10.3)
C
C Print out the error information for the eigenvectors.
C
      WRITE(OUTFILE,166)
166   FORMAT(/' Complex eigenvalues (R,I) and relative error in',
     1 ' right and left eigenvectors:'/
     2 3X,'I',20X,'Eigenvalues',22X,'Right Error',1X,'Left Error')
      WRITE(OUTFILE,162)(I,EIGVAL(1,I),EIGVAL(2,I),RELER2(I),RELER3(I),
     1                   I=1,N)
162   FORMAT(1X,I3,' (',1PE22.15,' , ',1PE22.15,' )',1P2E12.4)
      WRITE(OUTFILE,163)RELMX2,RELMX3
163   FORMAT(/' Maximum eigenvector error -- Right:',1PE10.3,
     1 '  Left:',1PE10.3)
C
C Test the relative error of the eigenvalues for acceptability.
      DO 210 K=1,N
        RERPAS=DMAX1(RERBAS,VALFAC(K)*PRECSN)
C Generate a FAIL message and terminate if the relative error is
C too large.
        IF(RELER1(K) .GT. RERPAS) THEN
          MES=
     1  'Relative error in an eigenvalue exceeds the allowed limit.  '
          CALL TSTMSG(FAIL,DIAGFILE,MES)
        ENDIF
210   CONTINUE
C
C Test to see if Ar = \lambda r, or A^T l = \lambda l, closely enough.
C
C Right eigenvectors:
C
      IF(IOPT1 .EQ. 1 .OR. IOPT1 .EQ. 3) THEN
        DO 220 K=1,N
          RERPAS=VECFAC*PRECSN
C Generate a FAIL message and terminate if the relative error is
C too large.
          IF(RELER2(K) .GT. RERPAS) THEN
            MES=
     1  'Relative error in a right vector exceeds the allowed limit. '
            CALL TSTMSG(FAIL,DIAGFILE,MES)
          ENDIF
220     CONTINUE
      ENDIF
C
C Left eigenvectors.
C
      IF(IOPT1 .EQ. 2 .OR. IOPT1 .EQ. 3) THEN
C If IOPT1 = 2, the vectors should match the eigenvalues to roundoff
C error, while if IOPT2 = 3, then the error is governed by the error
C in the eigenvalues.
        DO 230 K=1,N
          RERPAS=((IOPT1-2)*DMAX1(VALFAC(K),VECFAC)
     1           +(3-IOPT1)*VECFAC)*PRECSN
C Generate a FAIL message and terminate if the relative error is
C too large.
          IF(RELER3(K) .GT. RERPAS) THEN
            MES=
     1  'Relative error in a left vector exceeds the allowed limit.  '
            CALL TSTMSG(FAIL,DIAGFILE,MES)
          ENDIF
230     CONTINUE
      ENDIF
C
C Generate a PASS message and quit.
      CALL TSTMSG(PASS,DIAGFILE,MES)
      RETURN
      END
      SUBROUTINE TSTMSG(PASS,UNIT,MES)
C
C TSTMSG prints out the success or failure message for the
C calling program.  It terminates after printing the message.
C
C Calling format:
C
C      CALL TSTMSG(PASS,UNIT,MES)
C
C Parameter list:
C
C     PASS (Input)  LOGICAL variable.  If PASS is .TRUE.,
C                   print the success message and terminate.
C                   Otherwise print the failure message
C                   and the description given in MES.
C
C     UNIT (Input)  INTEGER unit number for the diagnostic file.
C
C      MES (Input)  CHARACTER*60 variable containing the descriptive
C                   message to be printed if PASS is .FALSE.
C
C Required routines: None
C
C Revision history:
C
C  Kevin W. Thompson     July 29, 1991     First version
C
C************************************************************************
C
      LOGICAL PASS
      INTEGER UNIT
      CHARACTER*60 MES
C
C Test for success or failure.
      IF(PASS) THEN
C Print a success message.
        WRITE(UNIT,10)
10      FORMAT(' PASS')
      ELSE
C Print a failure message and description.
        WRITE(UNIT,20)MES
20      FORMAT(' FAIL: ',A60)
      ENDIF
      STOP
      RETURN
      END
      real*8 function dasum(n,dx,incx)
c
c     takes the sum of the absolute values.
c     jack dongarra, linpack, 3/11/78.
c
      implicit real*8 (a-h,o-z)
      integer i,incx,m,mp1,n,nincx
      dimension dx(*)
C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
      REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
      INTEGER IWARN,IFATAL,ISYSTM
      CHARACTER MES*60,EFNAME*32
      DATA RMIN/2.23D-308/,RMAX/1.79D308/
      DATA RMNHAF/1.50D-154/,RMXHAF/1.33D154/
      DATA PRECSN/2.22D-16/,PRCHAF/1.5D-8/
      DATA IWARN/1/,IFATAL/2/,ISYSTM/3/
      DATA MES/
     1 'MAGUS Numerical Analyst 3.0.  Copyright (C) 1990 by MAGUS.  '/
C *** End of INCLUDE file.
c
      dasum = 0.0d0
      dtemp = 0.0d0
      if(n.le.0)return
      if(incx.eq.1)go to 20
c
c        code for increment not equal to 1
c
      nincx = n*incx
      do 10 i = 1,nincx,incx
        dtemp = dtemp + dabs(dx(i))
   10 continue
      dasum = dtemp
      return
c
c        code for increment equal to 1
c
c
   20 CONTINUE
c
C*********************************************************************
C
C Modifications by Kevin Thompson, 12/30/89
C Put in alternate loop for use on vector machines.  To vectorize
C more efficiently, remove comment from DO 30 through DASUM=... below,
C and comment out all subsequent lines before the RETURN statement.
C     DO 30 I=1,N
C       DTEMP=DTEMP+DABS(DX(I))
C30   CONTINUE
C     DASUM=DTEMP
C
C*********************************************************************
C
c        clean-up loop
c
      m = mod(n,6)
      if( m .eq. 0 ) go to 40
      do 30 i = 1,m
        dtemp = dtemp + dabs(dx(i))
   30 continue
      if( n .lt. 6 ) go to 60
   40 mp1 = m + 1
      do 50 i = mp1,n,6
        dtemp = dtemp + dabs(dx(i)) + dabs(dx(i + 1)) + dabs(dx(i + 2))
     *  + dabs(dx(i + 3)) + dabs(dx(i + 4)) + dabs(dx(i + 5))
   50 continue
   60 dasum = dtemp
      return
      end
      subroutine  dcopy(n,dx,incx,dy,incy)
c
c     copies a vector, x, to a vector, y.
c     uses unrolled loops for increments equal to one.
c     jack dongarra, linpack, 3/11/78.
c
      implicit real*8 (a-h,o-z)
      integer i,incx,incy,ix,iy,m,mp1,n
      dimension dx(*),dy(*)
C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
      REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
      INTEGER IWARN,IFATAL,ISYSTM
      CHARACTER MES*60,EFNAME*32
      DATA RMIN/2.23D-308/,RMAX/1.79D308/
      DATA RMNHAF/1.50D-154/,RMXHAF/1.33D154/
      DATA PRECSN/2.22D-16/,PRCHAF/1.5D-8/
      DATA IWARN/1/,IFATAL/2/,ISYSTM/3/
      DATA MES/
     1 'MAGUS Numerical Analyst 3.0.  Copyright (C) 1990 by MAGUS.  '/
C *** End of INCLUDE file.
c
      if(n.le.0)return
      if(incx.eq.1.and.incy.eq.1)go to 20
c
c        code for unequal increments or equal increments
c          not equal to 1
c
      ix = 1
      iy = 1
      if(incx.lt.0)ix = (-n+1)*incx + 1
      if(incy.lt.0)iy = (-n+1)*incy + 1
      do 10 i = 1,n
        dy(iy) = dx(ix)
        ix = ix + incx
        iy = iy + incy
   10 continue
      return
c
c        code for both increments equal to 1
c
c
   20 CONTINUE
C*********************************************************************
C
C Modifications by Kevin Thompson, 12/30/89
C Put in alternate loop for use on vector machines.  To vectorize
C more efficiently, remove comment from DO 30 through 30 CONTINUE 
C below, and comment out all subsequent lines before the RETURN 
C statement.
C     DO 30 I=1,N
C       DY(I)=DX(I)
C30   CONTINUE
C
C*********************************************************************
C
c        clean-up loop
c
      m = mod(n,7)
      if( m .eq. 0 ) go to 40
      do 30 i = 1,m
        dy(i) = dx(i)
   30 continue
      if( n .lt. 7 ) return
   40 mp1 = m + 1
      do 50 i = mp1,n,7
        dy(i) = dx(i)
        dy(i + 1) = dx(i + 1)
        dy(i + 2) = dx(i + 2)
        dy(i + 3) = dx(i + 3)
        dy(i + 4) = dx(i + 4)
        dy(i + 5) = dx(i + 5)
        dy(i + 6) = dx(i + 6)
   50 continue
      return
      end
      SUBROUTINE EFULRC(A,EIGVEC,EIGVAL,WS,IWS,NP,NR,N,NWS,
     1                  IOPT1,IOPT2,IOPT3)
C
C This subroutine computes the eigenvalues and eigenvectors of a
C real, full, general matrix, using the QR algorithm of Francis,
C as described by Martin, Peters, and Wilkinson in the Handbook
C for Automatic Computation-Linear Algebra, volume II,
C articles 14 and 15 (HACLA II/14 and HACLA II/15).  
C
C The eigenvalues and eigenvectors may be real or complex. 
C Both left and right eigenvectors may be computed.  Each
C eigenvector is normalized to have unit modulus, with the
C real part of its first element nonnegative.  The eigenvalues
C are sorted by their real parts; complex conjugate pairs
C are ordered so that the eigenvalue with the negative imaginary
C part comes first.  The vectors are ordered to correspond with
C the eigenvalues.
C
C The matrix A is first transformed to upper Hessenberg form by a
C direct transformation, as in HACLA II/13.  Then successive double QR
C transformations are performed, using Householder transformations and
C origin shifts.  The "active matrix" (matrix of as yet undetermined
C eigenvalues) is deflated (reduced in size) as eigenvalues are
C found in the lower right corner.  The eigenvalues occur either as the
C last element on the diagonal of the active matrix, or as the two
C real or complex conjugate eigenvalues of the last 2 X 2 sub-matrix
C along the diagonal.
C
C Once the eigenvalues have been computed, the matrix is in nearly
C triangular form (with at most block 2 X 2 matrices along the
C principal diagonal), and the right eigenvectors are computed by
C back substitution.  The eigenvectors of the nearly triangular matrix
C are then multiplied by the accumulated transformation matrix and
C thus transformed into the eigenvectors of the input matrix.
C
C Left eigenvectors of the input matrix are computed as the right
C eigenvectors of the transpose of the input matrix.  Condition
C numbers may then be computed from the two sets of eigenvectors.
C
C An accuracy test of the eigenvalues may also be performed by 
C comparing the eigenvalues of the original matrix with those of its
C transpose.  Theoretically the two sets should be the same.  In
C practice, the difference between two theoretically identical
C eigenvalues is about the same as the error in either of them.
C
C CALLING FORMAT:
C
C      CALL EFULRC(A,EIGVEC,EIGVAL,WS,IWS,NP,NR,N,NWS,
C     1            IOPT1,IOPT2,IOPT3)
C
C PARAMETER LIST:
C
C      A  (Input)  Array of row dimension NR and column dimension at
C                  least N, containing the N by N input matrix whose
C                  eigenvalues (and, optionally, eigenvectors) are to
C                  be computed.
C         (Output) The N X N accumulated transformation matrix 
C                  S which reduces the matrix B to nearly triangular
C                  form T, by the transformation
C                   -1
C                  S * B * S = T, or B * S = S * T,
C
C                  if eigenvectors are computed.  (Otherwise the
C                  contents of A have no particular significance.)
C                  If IOPT1 = 1, B is the input matrix A; 
C                  if IOPT1 = 2 or 3, B is the transpose of A.
C
C EIGVEC  (Output) Array of dimensions (NP,NR,N), containing the
C                  eigenvectors, if eigenvectors are computed.  See 
C                  the description below.
C
C EIGVAL  (Output) REAL*8 Array of row dimension 2 and column 
C                  dimension at least N, or COMPLEX*16 vector of
C                  length N, which contains the eigenvalues computed by
C                  this subroutine.  If EIGVAL is a REAL*8 array, then
C                  element EIGVAL(1,I) contains the real part of the 
C                  I'th eigenvalue, while element EIGVAL(2,I) contains
C                  the imaginary part of the I'th eigenvalue.  If 
C                  EIGVAL is a COMPLEX*16 array, then element EIGVAL(I)
C                  contains the I'th eigenvalue in COMPLEX*16 format.
C
C     WS  (Input)  Work space array of row dimension NR (exactly) and 
C                  column dimension NWS.
C         (Output) If IOPT1 = 3, or IOPT3 = 1, then WS(I,1) is the 
C                  relative accuracy of the I'th eigenvalue.
C                  If IOPT1 = 3 and IOPT3 = 1, then WS(I,2) is the 
C                  condition number (sensitivity) of the I'th
C                  eigenvalue.
C
C    IWS  (Input)  Integer array of length at least N, used as 
C                  temporary work space.
C
C     NP  (Input)  First (page) dimension of EIGVEC, exactly as
C                  specified in the calling program.  It must be
C                  at least as large the values given in the table
C                  below.
C
C     NR  (Input)  Row dimension of A, second dimension of EIGVEC, and
C                  row dimension of WS exactly as specified in the
C                  calling program.
C                  NR must be greater than or equal to N.
C
C      N  (Input)  Number of rows and columns in the input matrix.
C
C    NWS  (Input)  Second dimension of WS.  It must be at least as
C                  large as the value given in the table below.
C
C  IOPT1  (Input)  Eigenvalue-eigenvector option parameter.  The
C                  value of IOPT1 determines what calculation is
C                  to be performed, as described below.
C
C  IOPT2  (Input)  Eigenvector storage parameter.  The storage
C                  format of the output vectors in the eigenvector
C                  array (compressed or uncompressed) is specified by
C                  this parameter, as described below.
C
C  IOPT3  (Input)  Accuracy parameter.  This parameter determines which
C                  of the accuracy tests (estimated errors or condition
C                  numbers) are to be performed, as described below.
C
C======================================================================
C        Table of Input Parameters and Calculations Performed         |
C_____________________________________________________________________|
C                                                                     |
C                               Eigen-    Relative   Condition        |
C IOPT1  IOPT2  IOPT3  NP  NWS  vectors*  Accuracy   Numbers          |
C_____________________________________________________________________|
C                                                                     |
C   0      0      0    0    2   None        No          No            |
C   0      0      1    1    4   None        Yes         No            |
C                                                                     |
C   1      1      0    1    2   Right C     No          No            |
C   1      1      1    1    4   Right C     Yes         No            |
C   1      2      0    2    2   Right U     No          No            |
C   1      2      1    2    4   Right U     Yes         No            |
C                                                                     |
C   2      1      0    1    2   Left C      No          No            |
C   2      1      1    1    4   Left C      Yes         No            |
C   2      2      0    2    2   Left U      No          No            |
C   2      2      1    2    4   Left U      Yes         No            |
C                                                                     |
C   3      1      0    2    4   Both C      Yes         No            |
C   3      1      1    2    4   Both C      Yes         Yes           |
C   3      2      0    4    4   Both U      Yes         No            |
C   3      2      1    4    4   Both U      Yes         Yes           |
C                                                                     |
C======================================================================
C
C * The C or U refer to vectors stored in compressed or uncompressed
C   format, respectively.  The type of vector(s) computed (none, right,
C   left, or both) is as indicated.
C
C No vectors:      If eigenvectors and the accuracy test are not to be
C                  computed, set NP = 0, so that no storage space is
C                  allocated for the vector array.
C
C Either Left or   If all eigenvalues (and hence the eigenvectors) 
C Right Vectors:   are known to be real, then space need
C                  not be allocated for the imaginary parts of the 
C                  eigenvectors, and we may set IOPT2 = 1.  Then
C                  column J of EIGVEC contains the J'th eigenvector;
C                  i.e., element EIGVEC(1,I,J) is the I'th element
C                  of the J'th eigenvector.
C
C                  If any of the eigenvalues is complex, however, and
C                  sufficient memory is available, then IOPT2 should be
C                  set to 2, so the program will have sufficient space
C                  for normal complex eigenvector storage.  In the
C                  normal "uncompressed" format, elements EIGVEC(1,I,J)
C                  and EIGVEC(2,I,J) are the real and imaginary parts
C                  of the I'th element of the J'th eigenvector.  Note
C                  that complex eigenvalues and vectors always occur
C                  in complex conjugate pairs for real input matrices.
C
C                  If IOPT2 = 1 and some of the eigenvalues are
C                  complex, then the complex eigenvectors are
C                  stored in a compressed mode.  Real eigenvector
C                  J is in column J of EIGVEC, and its I'th element
C                  is EIGVEC(1,I,J).  If the J'th vector is the 
C                  first of a complex conjugate pair,
C                  however, then the real part of eigenvector J
C                  is in column J of EIGVEC, and its imaginary 
C                  part is in column J + 1.  Then complex
C                  eigenvector J + 1 is the complex conjugate of
C                  eigenvector J, and is not stored separately.
C                  Thus the real and imaginary parts of the J'th
C                  eigenvector are EIGVEC(1,I,J) and EIGVEC(1,I,J+1),
C                  while the real and imaginary parts of the J+1'st
C                  eigenvector are EIGVEC(1,I,J) and -EIGVEC(1,I,J+1).
C
C                  The choice IOPT2 = 2 is always safe, though it 
C                  may waste space if the eigenvectors are real.
C
C Both Left and    If IOPT2 = 1, the left vectors are stored in
C Right Vectors:   compressed format as elements EIGVEC(1,I,J),
C                  while the right vectors are stored in compressed
C                  format as elements EIGVEC(2,I,J).
C
C                  If IOPT2 = 2, the left vectors are uncompressed,
C                  with real parts in EIGVEC(1,I,J) and imaginary
C                  parts EIGVEC(2,I,J).  Similarly, the uncompressed
C                  right vectors have their real and imaginary parts
C                  stored as EIGVEC(3,I,J) and EIGVEC(4,I,J).
C
C Required routines: EBALNF, EINTF, EHESSF, EQRF, EHOUSF, EVECRF, 
C                    EBAKRF, CDIV, EBAKBF, ENORMF, VSORT2,
C                    UERCTL, UERDMP, UERPRT, NACOM
C
C Revision history:
C
C  Kevin Thompson     November 26, 1987     Version 3.0
C
C  Kevin Thompson     June 8, 1991          Version 3.0
C
C   Replaced numerous loops by calls to BLAS routines.  Also changed
C   call to EVECRF to match new parameter list, using WS for scratch
C   space.
C
C
C**********************************************************************
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION A(NR,N),EIGVEC(NP,NR,N),EIGVAL(2,N),WS(NR,NWS),IWS(N)
      DIMENSION W(3)
      LOGICAL GOOD,TRANSP,COMVEC
C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
      REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
      INTEGER IWARN,IFATAL,ISYSTM
      CHARACTER MES*60,EFNAME*32
      DATA RMIN/2.23D-308/,RMAX/1.79D308/
      DATA RMNHAF/1.50D-154/,RMXHAF/1.33D154/
      DATA PRECSN/2.22D-16/,PRCHAF/1.5D-8/
      DATA IWARN/1/,IFATAL/2/,ISYSTM/3/
      DATA MES/
     1 'MAGUS Numerical Analyst 3.0.  Copyright (C) 1990 by MAGUS.  '/
C *** End of INCLUDE file.
C
C Check for errors from preceding subroutine calls.
      CALL UERGET(LEVERR,IFEN,IWEN)
      IF(LEVERR .GT. IWARN) RETURN
C Test for valid option parameters.
C
      GOOD=(IOPT3 .EQ. 0 .OR. IOPT3 .EQ. 1) .AND.
     1     ((IOPT1 .EQ. 0 .AND. IOPT2 .EQ. 0) .OR.
     2      ((IOPT1 .GE. 1 .AND. IOPT1 .LE. 3) .AND.
     3       (IOPT2 .EQ. 1 .OR. IOPT2 .EQ. 2)))
      IF(.NOT. GOOD) THEN
        LEVERR=IFATAL
        MES=
     1  'Invalid option: IOPT parameter was given an incorrect value '
        CALL UERCTL('EFULRC',5111,MES,LEVERR)
        RETURN
      ENDIF
C
C Test for valid array dimensions (NR, N, NWS, and NP).
C
C NR,N
      IF(N .GT. NR) THEN
        LEVERR=IFATAL
        MES=
     1  'Number of rows in matrix exceeds array row dimension        '
        CALL UERCTL('EFULRC',5206,MES,LEVERR)
        RETURN
      ENDIF
      IF(NR .LE. 0 .OR. N .LE. 0) THEN
        LEVERR=IFATAL
        MES=
     1  'Number of elements in array can not be negative or zero     '
        CALL UERCTL('EFULRC',5203,MES,LEVERR)
        RETURN
      ENDIF
C NWS
      GOOD=(IOPT1 .LT. 3 .AND. ((IOPT3 .EQ. 0 .AND. NWS .GE. 2) .OR.
     1                          (IOPT3 .EQ. 1 .AND. NWS .GE. 4))) .OR.
     2     (IOPT1 .EQ. 3 .AND. NWS .GE. 4)
      IF(.NOT. GOOD) THEN
        LEVERR=IFATAL
        MES=
     1  'The work space array is too small: need larger NWS          '
        CALL UERCTL('EFULRC',5215,MES,LEVERR)
        RETURN
      ENDIF
C NP
      GOOD=(IOPT1 .EQ. 0 .AND. NP .GE. IOPT3) .OR.
     1     ((IOPT1 .EQ. 1 .OR. IOPT1 .EQ. 2) .AND. NP .GE. IOPT2) .OR.
     2     (IOPT1 .EQ. 3 .AND. NP .GE. 2*IOPT2)
      IF(.NOT. GOOD) THEN
        LEVERR=IFATAL
        MES=
     1  'The eigenvector array is too small: need larger NP          '
        CALL UERCTL('EFULRC',5214,MES,LEVERR)
        RETURN
      ENDIF
C
C Test to see if the input matrix will need to be transposed before
C entering the master loop.
      TRANSP=((IOPT1 .EQ. 0 .OR. IOPT1 .EQ. 1) .AND. IOPT3 .EQ. 1) 
     1       .OR. (IOPT1 .EQ. 2 .AND. IOPT3 .EQ. 0)
C
C Define the master loop index, which determines how many
C complete eigenvalue-eigenvectors calculations are to be performed.
C
      IF(IOPT1 .EQ. 3 .OR. IOPT3 .EQ. 1) THEN
        NCALC=2
      ELSE
        NCALC=1
      ENDIF
C
C Make a copy of the input matrix, if needed for two calculations
C or transposition.
      IF(TRANSP .OR. NCALC .EQ. 2) THEN
C
C Copy the matrix to the top page of EIGVEC.
C Alliant directive allowing concurrent calls to DCOPY:
cvd$  cncall
        DO 10 J=1,N
          CALL DCOPY(N,A(1,J),1,EIGVEC(NP,1,J),NP)
10      CONTINUE
      ENDIF
C
C Transpose the input matrix, if needed.
C Alliant directive allowing concurrent calls to DCOPY:
cvd$  cncall
      IF(TRANSP) THEN
        DO 20 J=1,N
          CALL DCOPY(N,EIGVEC(NP,1,J),1,A(J,1),NR)
20      CONTINUE
      ENDIF
C
C Now begin the master loop.
C
      DO 1000 ICALC=1,NCALC
C
C Define a logical variable, COMVEC, which is true if eigenvectors are
C to be computed in this loop, and false if they are not.
C
        COMVEC=IOPT1 .EQ. 3 .OR. (IOPT1 .NE. 0 .AND. ICALC .EQ. NCALC)
C
C Do the eigenvalue-eigenvector calculation.
C
C Balance the input matrix, and determine the central sub-matrix
C limits (ILOW and IHI), with subroutine EBALNF.  The first column
C of array WS will hold the cumulative balancing matrix elements
C and pivot indices.
        NBASE=2
        CALL EBALNF(A,NR,N,NBASE,WS,ILOW,IHI)
C
C Transform the matrix to upper Hessenberg form, provided that
C the active area is larger than 2 x 2.  Store the transformation
C matrix in EIGVEC.
        IF((IHI-ILOW) .GT. 1) THEN
          CALL EHESSF(A,EIGVEC,EIGVAL,IWS,NP,NR,N,ILOW,IHI,COMVEC)
        ELSEIF(COMVEC) THEN
C Initialize the transformation matrix in EIGVEC to the unit matrix.
          DO 5 J=1,N
            DO 15 I=1,N
              EIGVEC(1,I,J)=0.D0
15          CONTINUE
5         CONTINUE
          DO 30 I=1,N
            EIGVEC(1,I,I)=1.D0
30        CONTINUE
        ENDIF
C
C Perform the QR calculation to triangularize the matrix and
C compute the eigenvalues.  Compute the transformation matrix
C only if eigenvectors are to be determined on this pass.
C
C
        CALL EQRF(A,EIGVEC,EIGVAL,NP,NR,N,ILOW,IHI,COMVEC)
C
C Now that the eigenvalues have been determined, compute the
C eigenvectors.  Omit the vector calculation if the current
C loop is to determine eigenvalues only.
C
        IF(COMVEC) THEN
C
C First get the right eigenvectors of the nearly triangular matrix
C stored in A.
C
          CALL EVECRF(A,EIGVAL,WS(1,2),NR,N)
C
C Perform the back transformation on the compressed eigenvectors
C stored in A, overwriting A with the new compressed vectors.
C
          CALL EBAKRF(A,EIGVEC,EIGVAL,WS(1,2),NP,NR,N,ILOW,IHI)
C
C Reverse the balancing procedure on the current eigenvectors, yielding
C the eigenvectors of the original input matrix.
C
          CALL EBAKBF(A,1,NR,N,WS,ILOW,IHI,N)
C
C Accumulate the balancing transformation with the others in EIGVEC.
C
          CALL EBAKBF(EIGVEC,NP,NR,N,WS,ILOW,IHI,N)
C
C Normalize the eigenvectors to unit modulus.
C
          CALL ENORMF(A,EIGVAL,WS,NR,N)
C
C Sort the eigenvalues by their real parts, and sort the
C eigenvectors accordingly.
C
C First copy the real parts of the eigenvalues to the first colulmn
C of WS.
          CALL DCOPY(N,EIGVAL(1,1),2,WS(1,1),1)
C Sort the eigenvectors according to the eigenvalues, using the second
C column of WS for scratch space.
          CALL VSORT2(WS(1,1),WS(1,2),IWS,N,A,N,NR,0,1)
C Copy the eigenvalues to array WS, in the right order.
          DO 850 I=1,N
            WS(I,1)=EIGVAL(1,IWS(I))
            WS(I,2)=EIGVAL(2,IWS(I))
850       CONTINUE
C Put them back in EIGVAL in the right order.
          CALL DCOPY(N,WS(1,1),1,EIGVAL(1,1),2)
          CALL DCOPY(N,WS(1,2),1,EIGVAL(2,1),2)
C
C ********************************************************************
C
C Put the eigenvectors into the EIGVEC array for output.
C First interchange the transformation matrix and compressed
C vectors.
C
C Alliant directive allowing concurrent calls to DSWAP:
cvd$  cncall
          DO 950 J=1,N
            CALL DSWAP(N,A(1,J),1,EIGVEC(1,1,J),NP)
950       CONTINUE
C
C Finally, check for storage allocation for complex eigenvectors.
C If sufficient storage exists to decompress the eigenvectors, 
C then do so; otherwise retain the compressed vectors currently
C in EIGVEC.
C
          IF(IOPT2 .EQ. 2) THEN
C Sufficient storage exists to decompress the eigenvectors, so
C do so.
C
C Copy (and decompress complex) eigenvectors from A to EIGVEC,
C and copy the transformation matrix from EIGVEC to A.
C
            J=1
990         CONTINUE
              IF(EIGVAL(2,J) .EQ. 0.D0) THEN
C Real vector.
                DO 910 I=1,N
                  EIGVEC(2,I,J)=0.D0
910             CONTINUE
C Increment J and continue.
                J=J+1
              ELSE
C Complex vector--decompress it and store two complex vectors.
                DO 930 I=1,N
                  TEMPR=EIGVEC(1,I,J)
                  TEMPI=EIGVEC(1,I,J+1)
C                 EIGVEC(1,I,J)=TEMPR
                  EIGVEC(1,I,J+1)=TEMPR
                  EIGVEC(2,I,J)=TEMPI
                  EIGVEC(2,I,J+1)=-TEMPI
930             CONTINUE
C Increment J and continue.
                J=J+2
              ENDIF
            IF(J .LE. N)GO TO 990
C
          ENDIF
C
        ELSE
C
C ********************************************************************
C
C Only the eigenvalues are wanted.  Sort them by their real parts.
C
C First copy the real parts of the eigenvalues to the first colulmn
C of WS.
          CALL DCOPY(N,EIGVAL(1,1),2,WS(1,1),1)
C Sort the eigenvalues, using the second column of WS for scratch
C space.
          CALL VSORT2(WS(1,1),WS(1,2),IWS,N,EIGVAL,N,2,0,1)
        ENDIF
C
C If there will be a second pass through this loop, then set up the
C arrays properly for it.
C
        IF(ICALC .EQ. 1 .AND. NCALC .EQ. 2) THEN
C
C Copy the current eigenvalues to WS(I,3) and WS(I,4), for later
C comparison with the eigenvalues of the next loop.
          CALL DCOPY(N,EIGVAL(1,1),2,WS(1,3),1)
          CALL DCOPY(N,EIGVAL(2,1),2,WS(1,4),1)
C Copy the input matrix, or its transpose, from the highest page of
C EIGVEC to the array A.  The second pass uses the transpose of 
C the matrix which was used in the first pass.
          IF(TRANSP) THEN
C Copy the input matrix into the array A.
C Alliant directive allowing concurrent calls to DCOPY:
cvd$  cncall
            DO 80 J=1,N
              CALL DCOPY(N,EIGVEC(NP,1,J),NP,A(1,J),1)
80          CONTINUE
          ELSE
C Copy the transpose of the input matrix into the array A.
C Alliant directive allowing concurrent calls to DCOPY:
cvd$  cncall
            DO 65 J=1,N
              CALL DCOPY(N,EIGVEC(NP,1,J),NP,A(J,1),NR)
65          CONTINUE
          ENDIF
C If both left and right eigenvectors of A are to be computed, then
C copy the right eigenvectors of A into the highest page(s) of EIGVEC
C to free up page 1 for the transformation matrix.
          IF(IOPT1 .EQ. 3) THEN
            IF(IOPT2 .EQ. 1) THEN
C Copy one page.
C Alliant directive allowing concurrent calls to DCOPY:
cvd$  cncall
              DO 40 J=1,N
                CALL DCOPY(N,EIGVEC(1,1,J),NP,EIGVEC(2,1,J),NP)
40            CONTINUE
            ELSE
C Copy two pages.
C Alliant directive allowing concurrent calls to DCOPY:
cvd$  cncall
              DO 50 J=1,N
                CALL DCOPY(N,EIGVEC(2,1,J),NP,EIGVEC(4,1,J),NP)
                CALL DCOPY(N,EIGVEC(1,1,J),NP,EIGVEC(3,1,J),NP)
50            CONTINUE
            ENDIF
          ENDIF
C
        ENDIF
C
C ******************** End of master loop ****************************
C
1000  CONTINUE
C
C ********************************************************************
C
C
C Perform an accuracy test by computing the difference of the
C eigenvalues of the original input matrix and those of its
C transpose.
C
      IF(NCALC .EQ. 2) THEN
        DO 70 J=1,N
          DIFFR=WS(J,3)-EIGVAL(1,J)
          DIFFI=WS(J,4)-EIGVAL(2,J)
C Absolute error:
          EPSABS=DSQRT(DIFFR**2+DIFFI**2)
C Magnitude (use the smaller as the reference):
          EIMAG1=DSQRT(WS(J,3)**2+WS(J,4)**2)
          EIMAG2=DSQRT(EIGVAL(1,J)**2+EIGVAL(2,J)**2)
          EIMAG=DMIN1(EIMAG1,EIMAG2)
C Relative error:
          IF(EIMAG .NE. 0.D0) THEN
            WS(J,1)=DMAX1(EPSABS/EIMAG,PRECSN)
          ELSEIF(DMAX1(EIMAG1,EIMAG2) .EQ. 0.D0) THEN
C Both eigenvalues are zero: probably exact.
            WS(J,1)=0.D0
          ELSE
C One eigenvalue is zero, and one is not; return a unit error
C since we can't do anything else.
            WS(J,1)=1.D0
          ENDIF
70      CONTINUE
C The eigenvalues currently stored in EIGVAL are consistent with the
C most recent eigenvector calculation.  When both sets of eigenvectors
C are computed, consistency with both is not possible, and the
C eigenvalues output are made consistent with the right eigenvectors.
        IF(IOPT1 .EQ. 3) THEN
          CALL DCOPY(N,WS(1,3),1,EIGVAL(1,1),2)
          CALL DCOPY(N,WS(1,4),1,EIGVAL(2,1),2)
        ENDIF
      ENDIF
C
C Perform the condition number calculation, if desired.
C
      IF(IOPT1 .EQ. 3 .AND. IOPT3 .EQ. 1) THEN
        J=1
100     CONTINUE
C Check for complex eigenvalues.
          IF(EIGVAL(2,J) .LT. 0.D0) THEN
C First of a complex conjugate pair of vectors.
C Check for storage format.
            IF(IOPT2 .EQ. 1) THEN
C Compressed format.
              SUMR=DDOT(N,EIGVEC(1,1,J),NP,EIGVEC(2,1,J),NP)
     1            -DDOT(N,EIGVEC(1,1,J+1),NP,EIGVEC(2,1,J+1),NP)
              SUMI=DDOT(N,EIGVEC(1,1,J),NP,EIGVEC(2,1,J+1),NP)
     1            +DDOT(N,EIGVEC(1,1,J+1),NP,EIGVEC(2,1,J),NP)
            ELSE
C Uncompressed format.
              SUMR=DDOT(N,EIGVEC(1,1,J),NP,EIGVEC(3,1,J),NP)
     1            -DDOT(N,EIGVEC(2,1,J),NP,EIGVEC(4,1,J),NP)
              SUMI=DDOT(N,EIGVEC(1,1,J),NP,EIGVEC(4,1,J),NP)
     1            +DDOT(N,EIGVEC(2,1,J),NP,EIGVEC(3,1,J),NP)
            ENDIF
C Compute the condition numbers for this pair.  Put a ceiling on them
C of 1/PRECSN, since any larger numbers can only arise through
C meaningless, accidental cancellations.
            PRDMAG=DSQRT(SUMR**2+SUMI**2)
            C=1.D0/DMAX1(PRDMAG,PRECSN)
            WS(J,2)=C
            WS(J+1,2)=C
C Increment the index.
            J=J+2
          ELSE
C Real vectors.
C Define a base index for the real page of the right vector.
            NBASE=IOPT2+1
            SUM=DDOT(N,EIGVEC(1,1,J),NP,EIGVEC(NBASE,1,J),NP)
C Compute the condition number.  Put a ceiling on it of 1/PRECSN, since
C any larger numbers can only arise through meaningless, accidental
C cancellations.
            PRDMAG=DABS(SUM)
            C=1.D0/DMAX1(PRDMAG,PRECSN)
            WS(J,2)=C
C Increment the index.
            J=J+1
          ENDIF
C Go back and do the next one.
        IF(J .LE. N) GO TO 100
      ENDIF
C Reset the error counter.
      CALL UERCTL('EFULRC',0,MES,LEVERR)
      RETURN
      END
      SUBROUTINE ESETUP(ITEST,A,ACOPY,EVALCH,NR,N)
C
C This subroutine is called by the SEFULRC routines to set up
C the matrix A, and the known eigenvalues of A in the
C array EVALCH, according to the test selected by ITEST.
C
C The test problems are taken from the following sources:
C
C Martin, Peters, and Wilkinson test problems:
C   R. S. Martin, G. Peters, and J. H. Wilkinson,
C   "The QR Algorithm for Real Hessenberg Matrices",
C   Numerische Mathematik 14, 219-231 (1970)
C
C Gregory and Karney test problems:
C   R. T. Gregory and D. L. Karney,
C   A Collection of Matrices for Testing Computational Algorithms,
C   Robert E. Krieger Publishing Company, Huntington, New York (1978)
C
C Peters and Wilkinson HACLA II/15 test problems:
C   G. Peters and J. H. Wilkinson,
C   "Eigenvectors of Real and Complex Matrices by
C    LR and QR triangularizations",
C   Numerische Mathematik 16, 181-204 (1970)
C
C Parlett and Reinsch paper:
C   B. N. Parlett and C. Reinsch,
C   "Balancing a Matrix for Calculation of Eigenvalues and Eigenvectors",
C   Numerische Mathematik 13, 293-304 (1960)
C
C All other problems were devised by Kevin W. Thompson.
C
C
C Calling format:
C
C      CALL ESETUP(ITEST,A,ACOPY,EVALCH,NR,N)
C
C Parameter list:
C
C  ITEST  (Input)  Test number, in the range from 1 to 27.
C
C      A  (Output) REAL*8 array of row dimension NR and column
C                  dimension at least N, containing the matrix
C                  whose eigenvalues are to be computed.
C
C  ACOPY  (Output) REAL*8 array of row dimension NR and column
C                  dimension at least N, containing the matrix
C                  whose eigenvalues are to be computed.  (ACOPY
C                  is a copy of the A array).
C
C EVALCH  (Output) REAL*8 array of row dimension 2 and column 
C                  dimension at least N, which contains the 
C                  known eigenvalues of the matrix in A.
C                  Element EIGVAL(1,I) contains the real part of the 
C                  I'th eigenvalue, while element EIGVAL(2,I) contains
C                  the imaginary part of the I'th eigenvalue.
C
C     NR  (Input)  The row dimension of arrays A and ACOPY.
C
C      N  (Output) The number of rows in the test matrix, and
C                  the number of eigenvalues stored in EVALCH.
C
C
C
C Required Numerical Analyst routines: DCOPY
C
C Other required routines: None
C
C Revision history:
C
C  Kevin W. Thompson     December 7, 1991     First version
C
C***********************************************************************
C
      IMPLICIT REAL*8 (A-H,O-Z)
      REAL*8 A(NR,NR),ACOPY(NR,NR),EVALCH(2,NR),B(10,10),C(5,5)
C
C Initialize all matrix elements to zero.
      DO 1001 I=1,NR
        DO 1002 J=1,NR
          A(I,J)=0.D0
1002    CONTINUE
1001  CONTINUE
C
C Set up the problem selected.
C
      GO TO (1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,
     1 22,23,24,25,26,27) ITEST
C
C ********** Martin, Peters, and Wilkinson test problems **************
C
C===================================================  Test Problem 1
1     CONTINUE
C This test is example 1 on page 227.
        DO 1003 J=1,7
          A(J+1,J)=1.D0
1003    CONTINUE
        A(1,8)=1.D0
C
        N=8
C Store the eigenvalues.
        EVALCH(1,1)=-1.D0
        EVALCH(2,1)=0.D0
        EVALCH(1,2)=-1.D0/DSQRT(2.D0)
        EVALCH(2,2)=-1.D0/DSQRT(2.D0)
        EVALCH(1,3)=-1.D0/DSQRT(2.D0)
        EVALCH(2,3)=1.D0/DSQRT(2.D0)
        EVALCH(1,4)=0.D0
        EVALCH(2,4)=-1.D0
        EVALCH(1,5)=0.D0
        EVALCH(2,5)=1.D0
        EVALCH(1,6)=1.D0/DSQRT(2.D0)
        EVALCH(2,6)=-1.D0/DSQRT(2.D0)
        EVALCH(1,7)=1.D0/DSQRT(2.D0)
        EVALCH(2,7)=1.D0/DSQRT(2.D0)
        EVALCH(1,8)=1.D0
        EVALCH(2,8)=0.D0
C Jump to the end.
      GO TO 1000
C
C
C===================================================  Test Problem 2
2     CONTINUE
C This test is example 2 on page 228.
        A(1,1)=3.0D0
        A(1,2)=2.0D0
        A(1,3)=1.0D0
        A(1,4)=2.0D0
        A(1,5)=1.0D0
        A(1,6)=4.0D0
        A(1,7)=1.0D0
        A(1,8)=2.0D0
C
        A(2,1)=2.0D0
        A(2,2)=1.0D0
        A(2,3)=3.0D0
        A(2,4)=1.0D0
        A(2,5)=2.0D0
        A(2,6)=2.0D0
        A(2,7)=1.0D0
        A(2,8)=4.0D0
C
        A(3,2)=3.0D0
        A(3,3)=1.0D0
        A(3,4)=2.0D0
        A(3,5)=1.0D0
        A(3,6)=2.0D0
        A(3,7)=1.0D0
        A(3,8)=3.0D0
C
        A(4,3)=1.0D0
        A(4,4)=1.0D0
        A(4,5)=2.0D0
        A(4,6)=1.0D0
        A(4,7)=3.0D0
        A(4,8)=1.0D0
C
        A(5,4)=1.0D-7
        A(5,5)=3.0D0
        A(5,6)=1.0D0
        A(5,7)=4.0D0
        A(5,8)=2.0D0
C
        A(6,5)=1.0D-6
        A(6,6)=2.0D0
        A(6,7)=1.0D0
        A(6,8)=4.0D0
C
        A(7,6)=1.0D0
        A(7,7)=2.0D0
        A(7,8)=3.0D0
C
        A(8,7)=3.0D0
        A(8,8)=2.0D0
C 
        N=8
C Store the eigenvalues.
        EVALCH(1,1)=-2.42951247406D0
        EVALCH(1,2)=1.16665894308D-6
        EVALCH(1,3)=3.54247302818D-1
        EVALCH(1,4)=8.66367190136D-1
        EVALCH(1,5)=1.94495965031D0
        EVALCH(1,6)=3.D0
        EVALCH(1,7)=5.61818555238D0
        EVALCH(1,8)=5.64575162584D0
        DO 201 I=1,N
          EVALCH(2,I)=0.D0
201     CONTINUE
C Jump to the end.
      GO TO 1000
C
C
C===================================================  Test Problem 3
3     CONTINUE
C This test is example 3 on page 228.
        A(1,1)=1.0D0
        A(1,2)=2.0D0
        A(1,3)=3.0D0
        A(1,4)=4.0D0
C
        A(2,1)=1.0D-3
        A(2,2)=3.0D-3
        A(2,3)=2.0D-3
        A(2,4)=1.0D-3
C
        A(3,2)=1.0D-6
        A(3,3)=3.0D-6
        A(3,4)=2.0D-6
C
        A(4,3)=1.0D-9
        A(4,4)=2.0D-9
C 
        N=4
C Store the eigenvalues.
        EVALCH(1,1)=7.49765325656D-10
        EVALCH(1,2)=4.00529353526D-6
        EVALCH(1,3)=9.96994963685D-4
        EVALCH(1,4)=1.00200200099D0
        DO 31 I=1,N
          EVALCH(2,I)=0.D0
31      CONTINUE
C Jump to the end.
      GO TO 1000
C
C
C===================================================  Test Problem 4
4     CONTINUE
C This test is example 4 on page 228.
        A(1,1)=1.0D0
        A(1,2)=6.25D-2
        A(1,3)=3.66210937500D-4
        A(1,4)=9.53674316406D-7
C
        A(2,1)=3.2D-2
        A(2,2)=3.0D-3
        A(2,3)=7.8125D-6
        A(2,4)=7.62939453126D-9
C
        A(3,2)=2.56D-4
        A(3,3)=3.0D-6
        A(3,4)=3.90625D-9
C
        A(4,3)=5.12D-7
        A(4,4)=2.0D-9
C 
        N=4
C Store the eigenvalues.
        EVALCH(1,1)=7.49765325656D-10
        EVALCH(1,2)=4.00529353526D-6
        EVALCH(1,3)=9.96994963685D-4
        EVALCH(1,4)=1.00200200099D0
        DO 41 I=1,N
          EVALCH(2,I)=0.D0
41      CONTINUE
C Jump to the end.
      GO TO 1000
C
C
C===================================================  Test Problem 5
5     CONTINUE
C This test is example 5 on page 230.
        DO 501 I=1,13
          DO 502 J=I,13
            A(I,J)=14-J
502       CONTINUE
501     CONTINUE
        DO 503 J=1,12
          A(J+1,J)=13-J
503     CONTINUE
C
C 
        N=13
C Store the eigenvalues.
        EVALCH(1,13)=3.56138612027D1
        EVALCH(1,12)=2.30375322906D1
        EVALCH(1,11)=1.46297821337D1
        EVALCH(1,10)=8.75377956658D0
        EVALCH(1,9)=4.76730653278D0
        EVALCH(1,8)=2.29891123271D0
        EVALCH(1,7)=1.D0
        EVALCH(1,6)=1.D0/EVALCH(1,8)
        EVALCH(1,5)=1.D0/EVALCH(1,9)
        EVALCH(1,4)=1.D0/EVALCH(1,10)
        EVALCH(1,3)=1.D0/EVALCH(1,11)
        EVALCH(1,2)=1.D0/EVALCH(1,12)
        EVALCH(1,1)=1.D0/EVALCH(1,13)
        DO 51 I=1,N
          EVALCH(2,I)=0.D0
51      CONTINUE
C Jump to the end.
      GO TO 1000
C
C
C ********** Gregory and Karney test problems *************************
C
C===================================================  Test Problem 6
6     CONTINUE
C This test is example 5.1 on page 81
        A(1,1)=33.0D0
        A(1,2)=16.0D0
        A(1,3)=72.0D0
C
        A(2,1)=-24.0D0
        A(2,2)=-10.0D0
        A(2,3)=-57.0D0
C
        A(3,1)=-8.0D0
        A(3,2)=-4.0D0
        A(3,3)=-17.0D0
C
C 
        N=3
C Store the eigenvalues.
        DO 61 I=1,N
          EVALCH(1,I)=I
          EVALCH(2,I)=0.D0
61      CONTINUE
C Jump to the end.
      GO TO 1000
C
C
C===================================================  Test Problem 7
7     CONTINUE
C This test is example 5.2 on page 82
        A(1,1)=4.0D0
        A(1,2)=1.0D0
        A(1,3)=1.0D0
C
        A(2,1)=2.0D0
        A(2,2)=4.0D0
        A(2,3)=1.0D0
C
        A(3,1)=0.0D0
        A(3,2)=1.0D0
        A(3,3)=4.0D0
C
C 
        N=3
C Store the eigenvalues.
        EVALCH(1,1)=3.D0
        EVALCH(2,1)=0.D0
        EVALCH(1,2)=3.D0
        EVALCH(2,2)=0.D0
        EVALCH(1,3)=6.D0
        EVALCH(2,3)=0.D0
C Jump to the end.
      GO TO 1000
C
C
C===================================================  Test Problem 8
8     CONTINUE
C This test is example 5.5 on page 85
        A(1,1)=-2.0D0
        A(1,2)=2.0D0
        A(1,3)=2.0D0
        A(1,4)=2.0D0
C
        A(2,1)=-3.0D0
        A(2,2)=3.0D0
        A(2,3)=2.0D0
        A(2,4)=2.0D0
C
        A(3,1)=-2.0D0
        A(3,2)=0.0D0
        A(3,3)=4.0D0
        A(3,4)=2.0D0
C
        A(4,1)=-1.0D0
        A(4,2)=0.0D0
        A(4,3)=0.0D0
        A(4,4)=5.0D0
C
C 
        N=4
C Store the eigenvalues.
        DO 81 I=1,N
          EVALCH(1,I)=I
          EVALCH(2,I)=0.D0
81      CONTINUE
C Jump to the end.
      GO TO 1000
C
C
C===================================================  Test Problem 9
9     CONTINUE
C This test is example 5.9 on page 88
        A(1,1)=122.0D0
        A(1,2)=41.0D0
        A(1,3)=40.0D0
        A(1,4)=26.0D0
        A(1,5)=25.0D0
C
        A(2,1)=40.0D0
        A(2,2)=170.0D0
        A(2,3)=25.0D0
        A(2,4)=14.0D0
        A(2,5)=24.0D0
C
        A(3,1)=27.0D0
        A(3,2)=26.0D0
        A(3,3)=172.0D0
        A(3,4)=7.0D0
        A(3,5)=3.0D0
C
        A(4,1)=32.0D0
        A(4,2)=22.0D0
        A(4,3)=9.0D0
        A(4,4)=106.0D0
        A(4,5)=6.0D0
C
        A(5,1)=31.0D0
        A(5,2)=28.0D0
        A(5,3)=-2.0D0
        A(5,4)=-1.0D0
        A(5,5)=165.0D0
C
C 
        N=5
C Store the eigenvalues.
        EVALCH(1,1)=77.6971911963D0
        EVALCH(1,2)=112.154193247D0
        EVALCH(1,3)=134.686463320D0
        EVALCH(1,4)=167.484878917D0
        EVALCH(1,5)=242.977273320D0
        DO 91 I=1,N
          EVALCH(2,I)=0.D0
91      CONTINUE
C Jump to the end.
      GO TO 1000
C
C
C===================================================  Test Problem 10
10    CONTINUE
C This test is example 5.14 on page 92.
        N=12
C Put in the nonzero elements.
        DO 101 J=1,12
          DO 102 I=J,N
            A(I,J)=13-I
102       CONTINUE
101     CONTINUE
        DO 103 I=1,N-1
          A(I,I+1)=12-I
103     CONTINUE
C Store the eigenvalues.
        EVALCH(1,1)=3.1028060644010D-2
        EVALCH(1,2)=4.9507429185278D-2
        EVALCH(1,3)=8.1227659240405D-2
        EVALCH(1,4)=1.43646519769220D-1
        EVALCH(1,5)=2.84749720558478D-1
        EVALCH(1,6)=6.435053190048555D-1
        EVALCH(1,7)=1.55398870913210790D0
        EVALCH(1,8)=3.511855948580757194D0
        EVALCH(1,9)=6.961533085567122113D0
        EVALCH(1,10)=12.311077408868526120D0
        EVALCH(1,11)=20.198988645877079428D0
        EVALCH(1,12)=32.228891501572160750D0
        DO 104 I=1,N
          EVALCH(2,I)=0.D0
104     CONTINUE
C Jump to the end.
      GO TO 1000
C
C
C===================================================  Test Problem 11
11    CONTINUE
C This test is example 5.16 on page 97.
        N=16
C Fill the C matrix.
        C(1,1)=-2.D0
        C(1,2)=2.D0
        C(1,3)=2.D0
        C(1,4)=2.D0
        C(2,1)=-3.D0
        C(2,2)=3.D0
        C(2,3)=2.D0
        C(2,4)=2.D0
        C(3,1)=-2.D0
        C(3,2)=0.D0
        C(3,3)=4.D0
        C(3,4)=2.D0
        C(4,1)=-1.D0
        C(4,2)=0.D0
        C(4,3)=0.D0
        C(4,4)=5.D0
C Fill the B matrix.
        DO 111 J=1,4
          DO 112 I=1,4
            B(I,J)=5.D0*C(I,J)
            B(I+4,J)=5.D0*C(I,J)
            B(I,J+4)=-C(I,J)
            B(I+4,J+4)=C(I,J)
112       CONTINUE
111     CONTINUE
C
C Finally, fill the A matrix.
        DO 113 J=1,8
          DO 114 I=1,8
            A(I,J)=B(I,J)
            A(I+8,J)=4.D0*B(I,J)
            A(I,J+8)=2.D0*B(I,J)
            A(I+8,J+8)=3.D0*B(I,J)
114       CONTINUE
113     CONTINUE
C Store the eigenvalues.
        EVALCH(1,1)=-12.D0
        EVALCH(2,1)=-4.D0
        EVALCH(1,2)=-12.D0
        EVALCH(2,2)=4.D0
        EVALCH(1,3)=-9.D0
        EVALCH(2,3)=-3.D0
        EVALCH(1,4)=-9.D0
        EVALCH(2,4)=3.D0
        EVALCH(1,5)=-6.D0
        EVALCH(2,5)=-2.D0
        EVALCH(1,6)=-6.D0
        EVALCH(2,6)=2.D0
        EVALCH(1,7)=-3.D0
        EVALCH(2,7)=-1.D0
        EVALCH(1,8)=-3.D0
        EVALCH(2,8)=1.D0
        EVALCH(1,9)=15.D0
        EVALCH(2,9)=-5.D0
        EVALCH(1,10)=15.D0
        EVALCH(2,10)=5.D0
        EVALCH(1,11)=30.D0
        EVALCH(2,11)=-10.D0
        EVALCH(1,12)=30.D0
        EVALCH(2,12)=10.D0
        EVALCH(1,13)=45.D0
        EVALCH(2,13)=-15.D0
        EVALCH(1,14)=45.D0
        EVALCH(2,14)=15.D0
        EVALCH(1,15)=60.D0
        EVALCH(2,15)=-20.D0
        EVALCH(1,16)=60.D0
        EVALCH(2,16)=20.D0
C Jump to the end.
      GO TO 1000
C
C
C===================================================  Test Problem 12
12    CONTINUE
C This test is example 5.24 on page 103, with epsilon = 1.
C
C Define the size of the matrix.
        N=10
C Create the nonzero diagonal.
        DO 121 J=2,N
          A(J-1,J)=1.D0
121     CONTINUE
        A(N,1)=1.D0
C Store the eigenvalues.
        PI=3.141592653589793D0
C Check for an even or odd number.
        IF(MOD(N,2) .EQ. 0) THEN
          EVALCH(1,1)=-1.D0
          EVALCH(2,1)=0.D0
          ISHIFT=1
        ELSE
          ISHIFT=0
        ENDIF
C Get the upper loop index.
        INDMAX=(N-1)/2
        DO 122 I=1,INDMAX
          ARG=2.D0*(INDMAX-I+1)*PI/N
          COSARG=DCOS(ARG)
          SINARG=DSIN(ARG)
          EVALCH(1,2*I-1+ISHIFT)=COSARG
          EVALCH(2,2*I-1+ISHIFT)=-SINARG
          EVALCH(1,2*I+ISHIFT)=COSARG
          EVALCH(2,2*I+ISHIFT)=SINARG
122     CONTINUE
        EVALCH(1,N)=1.D0
        EVALCH(2,N)=0.D0
C Jump to the end.
      GO TO 1000
C
C
C===================================================  Test Problem 13
13    CONTINUE
C This test is example 5.26 on page 108.
        N=6
C Fill the matrix.
        A(1,1)=-9.D0
        A(1,2)=21.D0
        A(1,3)=-15.D0
        A(1,4)=4.D0
        A(1,5)=2.D0
        A(1,6)=0.D0
        A(2,1)=-10.D0
        A(2,2)=21.D0
        A(2,3)=-14.D0
        A(2,4)=4.D0
        A(2,5)=2.D0
        A(2,6)=0.D0
        A(3,1)=-8.D0
        A(3,2)=16.D0
        A(3,3)=-11.D0
        A(3,4)=4.D0
        A(3,5)=2.D0
        A(3,6)=0.D0
        A(4,1)=-6.D0
        A(4,2)=12.D0
        A(4,3)=-9.D0
        A(4,4)=3.D0
        A(4,5)=3.D0
        A(4,6)=0.D0
        A(5,1)=-4.D0
        A(5,2)=8.D0
        A(5,3)=-6.D0
        A(5,4)=0.D0
        A(5,5)=5.D0
        A(5,6)=0.D0
        A(6,1)=-2.D0
        A(6,2)=4.D0
        A(6,3)=-3.D0
        A(6,4)=0.D0
        A(6,5)=1.D0
        A(6,6)=3.D0
C Store the eigenvalues.
        EVALCH(1,1)=1.D0
        EVALCH(2,1)=0.D0
        EVALCH(1,2)=1.D0
        EVALCH(2,2)=0.D0
        EVALCH(1,3)=2.D0
        EVALCH(2,3)=-1.D0
        EVALCH(1,4)=2.D0
        EVALCH(2,4)=1.D0
        EVALCH(1,5)=3.D0
        EVALCH(2,5)=0.D0
        EVALCH(1,6)=3.D0
        EVALCH(2,6)=0.D0
C Jump to the end.
      GO TO 1000
C
C
C********* Peters and Wilkinson HACLA II/15 test problems *************
C
C===================================================  Test Problem 14
14    CONTINUE
C This test is example 5 on page 394.
        N=4
C Fill the matrix.
        A(1,1)=3.D0
        A(1,2)=1.D0
        A(1,3)=2.D0
        A(1,4)=5.D0
        A(2,1)=2.D0
        A(2,2)=1.D0
        A(2,3)=3.D0
        A(2,4)=-1.D0
        A(3,1)=0.D0
        A(3,2)=4.D0
        A(3,3)=1.D0
        A(3,4)=1.D0
        A(4,1)=0.D0
        A(4,2)=0.D0
        A(4,3)=2.D0
        A(4,4)=1.D0
C Store the eigenvalues.
        EVALCH(1,1)=-3.19552891925D0
        EVALCH(2,1)=0.D0
        EVALCH(1,2)=1.55881150332D0
        EVALCH(2,2)=-2.00151592083D0
        EVALCH(1,3)=EVALCH(1,2)
        EVALCH(2,3)=-EVALCH(2,2)
        EVALCH(1,4)=6.07790591265D0
        EVALCH(2,4)=0.D0
C Jump to the end.
      GO TO 1000
C
C
C===================================================  Test Problem 15
15    CONTINUE
C This test is example 6 on page 395.
        N=3
C Fill the matrix.
        A(1,1)=2.D0
        A(1,2)=4.D0
        A(1,3)=4.D0
        A(2,1)=0.D0
        A(2,2)=3.D0
        A(2,3)=1.D0
        A(3,1)=0.D0
        A(3,2)=1.D0
        A(3,3)=3.D0
C Store the eigenvalues.
        EVALCH(1,1)=2.D0
        EVALCH(2,1)=0.D0
        EVALCH(1,2)=2.D0
        EVALCH(2,2)=0.D0
        EVALCH(1,3)=4.D0
        EVALCH(2,3)=0.D0
C Jump to the end.
      GO TO 1000
C
C
C===================================================  Test Problem 16
16    CONTINUE
C This test is example 7 on page 395.
        N=4
C Fill the matrix.
        A(1,1)=2.D0
        A(1,2)=-3.D0
        A(1,3)=2.D0
        A(1,4)=-1.D0
        A(2,1)=1.D0
        A(2,2)=0.D0
        A(2,3)=0.D0
        A(2,4)=0.D0
        A(3,1)=0.D0
        A(3,2)=1.D0
        A(3,3)=0.D0
        A(3,4)=0.D0
        A(4,1)=0.D0
        A(4,2)=0.D0
        A(4,3)=1.D0
        A(4,4)=0.D0
C Store the eigenvalues.
        ROOT32=DSQRT(3.D0)/2.D0
        EVALCH(1,1)=0.5D0
        EVALCH(2,1)=-ROOT32
        EVALCH(1,2)=0.5D0
        EVALCH(2,2)=ROOT32
        EVALCH(1,3)=0.5D0
        EVALCH(2,3)=-ROOT32
        EVALCH(1,4)=0.5D0
        EVALCH(2,4)=ROOT32
C Jump to the end.
      GO TO 1000
C
C
C********************  Parlett and Reinsch paper *********************
C
C
C===================================================  Test Problem 17
17    CONTINUE
C This test is example 1
        N=7
C Fill the matrix.
        A(1,1)=6.D0
        A(1,6)=1.D0
        A(2,2)=4.D0
        A(2,4)=3.D-4
        A(2,5)=1.D-2
        A(2,6)=2.D-2
        A(2,7)=1.D-1
        A(3,1)=1.D0
        A(3,2)=1.D2
        A(3,3)=7.D0
        A(3,6)=-2.D0
        A(3,7)=2.D1
        A(4,2)=2.D4
        A(4,4)=1.D0
        A(4,5)=-4.D2
        A(4,6)=3.D2
        A(4,7)=-4.D3
        A(5,1)=-2.D0
        A(5,2)=-3.D2
        A(5,4)=1.D-2
        A(5,5)=2.D0
        A(5,6)=2.D0
        A(5,7)=4.D1
        A(7,2)=1.D1
        A(7,4)=4.D-3
        A(7,5)=1.D-1
        A(7,6)=-2.D-1
        A(7,7)=3.D0
C Store the eigenvalues.
        EVALCH(1,1)=-0.0771557967D0
        EVALCH(2,1)=0.D0
        EVALCH(1,2)=0.D0
        EVALCH(2,2)=0.D0
        EVALCH(1,3)=2.7438084239D0
        EVALCH(2,3)=-3.8811589838D0
        EVALCH(1,4)=2.7438084239D0
        EVALCH(2,4)=3.8811589838D0
        EVALCH(1,5)=4.5895389489D0
        EVALCH(2,5)=0.D0
        EVALCH(1,6)=6.D0
        EVALCH(2,6)=0.D0
        EVALCH(1,7)=7.D0
        EVALCH(2,7)=0.D0
C Jump to the end.
      GO TO 1000
C
C
C**************************** Gregory and Karney **********************
C
C
C===================================================  Test Problem 18
18    CONTINUE
C This test is example 5.5 on page 85 of Gregory and Karney, but
C scaled up to a norm of approximately MAX( R, 1/R**2 ).
        R=1.D20
C
        A(1,1)=-2.0D0
        A(1,2)=2.0D0*R
        A(1,3)=2.0D0/R
        A(1,4)=2.0D0
C
        A(2,1)=-3.0D0/R
        A(2,2)=3.0D0
        A(2,3)=2.0D0/R**2
        A(2,4)=2.0D0/R
C
        A(3,1)=-2.0D0*R
        A(3,2)=0.0D0
        A(3,3)=4.0D0
        A(3,4)=2.0D0*R
C
        A(4,1)=-1.0D0
        A(4,2)=0.0D0
        A(4,3)=0.0D0
        A(4,4)=5.0D0
C
        N=4
C Store the eigenvalues.
        DO 181 I=1,N
          EVALCH(1,I)=I
          EVALCH(2,I)=0.D0
181     CONTINUE
C Jump to the end.
      GO TO 1000
C
C
C************ Triangular or Nearly Triangular Matrices ***************
C
C
C===================================================  Test Problem 19
19    CONTINUE
C This test is a matrix of order N, with N distinct eigenvalues and
C eigenvectors.
C
C Define the size of the matrix.
        N=10
        DO 1013 J=1,N
          DO 1014 I=1,J
            A(I,J)=J
1014      CONTINUE
          EVALCH(1,J)=J
          EVALCH(2,J)=0.D0
1013    CONTINUE
C Jump to the end.
      GO TO 1000
C
C
C===================================================  Test Problem 20
20    CONTINUE
C This test is a matrix of order 3, with 2 degenerate eigenvalues and
C eigenvectors.
        N=3
C
        A(1,1)=2.D0
        A(1,2)=6.D0
        A(1,3)=9.D0
        A(2,2)=4.D0
        A(2,3)=3.D0
        A(3,3)=2.D0
C
        EVALCH(1,1)=2.D0
        EVALCH(2,1)=0.D0
        EVALCH(1,2)=2.D0
        EVALCH(2,2)=0.D0
        EVALCH(1,3)=4.D0
        EVALCH(2,3)=0.D0
C Jump to the end.
      GO TO 1000
C
C
C===================================================  Test Problem 21
21    CONTINUE
C This test is a matrix of order 3, with 2 defective eigenvalues and
C eigenvectors.
        N=3
C
        A(1,1)=2.D0
        A(1,2)=6.D0
        A(1,3)=3.D0
        A(2,2)=4.D0
        A(2,3)=3.D0
        A(3,3)=2.D0
C
        EVALCH(1,1)=2.D0
        EVALCH(2,1)=0.D0
        EVALCH(1,2)=2.D0
        EVALCH(2,2)=0.D0
        EVALCH(1,3)=4.D0
        EVALCH(2,3)=0.D0
C Jump to the end.
      GO TO 1000
C
C
C===================================================  Test Problem 22
22    CONTINUE
C This test is a matrix of order N, with N defective eigenvalues and
C eigenvectors.
C
C Define the size of the matrix.
        N=10
        ROOT2=DSQRT(2.D0)
        DO 1043 I=1,N-1
          A(I,I)=ROOT2
          A(I,I+1)=1.D0
          EVALCH(1,I)=ROOT2
          EVALCH(2,I)=0.D0
1043    CONTINUE
        A(N,N)=ROOT2
        EVALCH(1,N)=ROOT2
        EVALCH(2,N)=0.D0
C Jump to the end.
      GO TO 1000
C
C
C===================================================  Test Problem 23
23    CONTINUE
C This test is a matrix of order 2, with 2 complex eigenvectors.
        N=2
C
        A(1,1)=1.D0
        A(1,2)=1.D0
        A(2,1)=-4.D0
        A(2,2)=1.D0
C
        EVALCH(1,1)=1.D0
        EVALCH(2,1)=-2.D0
        EVALCH(1,2)=1.D0
        EVALCH(2,2)=2.D0
C Jump to the end.
      GO TO 1000
C
C
C===================================================  Test Problem 24
24    CONTINUE
C This test is a matrix of order 4, with 2 degenerate, complex eigenvalues
C and eigenvectors (i.e., two repeated complex conjugate pairs of
C eigenvalues, but four independent eigenvectors).
        N=4
C
        A(1,1)=1.D0
        A(1,2)=1.D0
        A(2,1)=-4.D0
        A(2,2)=1.D0
        A(2,3)=0.D0
        A(2,4)=0.D0
        A(3,3)=1.D0
        A(3,4)=1.D0
        A(4,3)=-4.D0
        A(4,4)=1.D0
C
        EVALCH(1,1)=1.D0
        EVALCH(2,1)=-2.D0
        EVALCH(1,2)=1.D0
        EVALCH(2,2)=2.D0
        EVALCH(1,3)=1.D0
        EVALCH(2,3)=-2.D0
        EVALCH(1,4)=1.D0
        EVALCH(2,4)=2.D0
C Jump to the end.
      GO TO 1000
C
C
C===================================================  Test Problem 25
25    CONTINUE
C This test is a matrix of order 4, with 4 defective complex eigenvalues
C and eigenvectors (i.e., two repeated complex conjugate pairs of
C eigenvalues, but only two independent eigenvectors).
        N=4
C
        A(1,1)=1.D0
        A(1,2)=1.D0
        A(2,1)=-4.D0
        A(2,2)=1.D0
        A(2,3)=1.D0
        A(2,4)=1.D0
        A(3,3)=1.D0
        A(3,4)=1.D0
        A(4,3)=-4.D0
        A(4,4)=1.D0
C
        EVALCH(1,1)=1.D0
        EVALCH(2,1)=-2.D0
        EVALCH(1,2)=1.D0
        EVALCH(2,2)=2.D0
        EVALCH(1,3)=1.D0
        EVALCH(2,3)=-2.D0
        EVALCH(1,4)=1.D0
        EVALCH(2,4)=2.D0
C Jump to the end.
      GO TO 1000
C
C
C===================================================  Test Problem 26
26    CONTINUE
C This test is a matrix of order N (even only), with only two 
C independent eigenvectors, and two complex conjugate eigenvalues
C repeated N/2 times.
C
C Define the size of the matrix.
        N=10
        ROOT2=DSQRT(2.D0)
        ROOT3=DSQRT(3.D0)
C
        DO 2601 I=1,N-1,2
          A(I,I)=ROOT2
          A(I+1,I)=-ROOT3
          A(I+1,I+1)=ROOT2
2601    CONTINUE
        DO 2602 I=1,N-1
          A(I,I+1)=1.D0
2602    CONTINUE
C
C Store the eigenvalues.
        TEMP=DSQRT(ROOT3)
        DO 2603 I=1,N-1,2
          EVALCH(1,I)=ROOT2
          EVALCH(2,I)=-TEMP
          EVALCH(1,I+1)=ROOT2
          EVALCH(2,I+1)=TEMP
2603    CONTINUE
C Jump to the end.
      GO TO 1000
C
C
C===================================================  Test Problem 27
27    CONTINUE
C This test is a matrix of order 12, with two degenerate and one
C defective real eigenvalues (all the same), and two each (in pairs)
C of two degenerate and one defective complex eigenvalues, one
C distinct real eigenvalue, and two distinct complex conjugate
C eigenvalues.  In short, everything.
        N=12
C
        ROOT2=DSQRT(2.D0)
        ROOT5=DSQRT(5.D0)
        A(1,1)=1.D0
        A(2,2)=2.D0
        A(2,3)=1.D0
        A(3,2)=-5.D0
        A(3,3)=2.D0
        A(4,4)=1.D0
        A(4,5)=ROOT2
        A(5,5)=2.D0
        A(5,6)=1.D0
        A(6,5)=-5.D0
        A(6,6)=2.D0
        A(6,7)=ROOT2
        A(7,7)=1.D0
        A(8,8)=2.D0
        A(8,9)=1.D0
        A(9,8)=-5.D0
        A(9,9)=2.D0
        A(10,10)=7.D0
        A(11,11)=ROOT2
        A(11,12)=1.D0
        A(12,11)=-2.D0
        A(12,12)=ROOT2
C
        EVALCH(1,1)=1.D0
        EVALCH(2,1)=0.D0
        EVALCH(1,2)=1.D0
        EVALCH(2,2)=0.D0
        EVALCH(1,3)=1.D0
        EVALCH(2,3)=0.D0
        EVALCH(1,4)=ROOT2
        EVALCH(2,4)=-ROOT2
        EVALCH(1,5)=ROOT2
        EVALCH(2,5)=ROOT2
        EVALCH(1,6)=2.D0
        EVALCH(2,6)=-ROOT5
        EVALCH(1,7)=2.D0
        EVALCH(2,7)=ROOT5
        EVALCH(1,8)=2.D0
        EVALCH(2,8)=-ROOT5
        EVALCH(1,9)=2.D0
        EVALCH(2,9)=ROOT5
        EVALCH(1,10)=2.D0
        EVALCH(2,10)=-ROOT5
        EVALCH(1,11)=2.D0
        EVALCH(2,11)=ROOT5
        EVALCH(1,12)=7.D0
        EVALCH(2,12)=0.D0
C Jump to the end.
      GO TO 1000
C
C
C *********************************************************************
C
1000  CONTINUE
C Copy the input matrix to ACOPY.
      DO 1004 J=1,N
        CALL DCOPY(N,A(1,J),1,ACOPY(1,J),1)   
1004  CONTINUE
      RETURN
      END
      SUBROUTINE EVALCOM(N,EIGVAL,EVALCH,RELERR,AVGABS,
     1                   AVGREL,RELMAX)
C
C EVALCOM compares the N computed eigenvalues in EIGVAL with the known
C eigenvalues in EVALCH.  It computes the relative error for each
C eigenvalue, the average absolute error, the average relative error,
C and the maximum relative error.  
C
C Calling format:
C
C      CALL EVALCOM(N,EIGVAL,EVALCH,RELERR,AVGABS,
C     1             AVGREL,RELMAX)
C
C Parameter list:
C
C      N  (Input)  Number of eigenvalues to be examined.
C
C EIGVAL  (Input)  REAL*8 Array of row dimension 2 and column 
C                  dimension at least N, which contains the eigenvalues 
C                  to be examined by this subroutine.  
C                  Element EIGVAL(1,I) contains the real part of the 
C                  I'th eigenvalue, while element EIGVAL(2,I) contains
C                  the imaginary part of the I'th eigenvalue.
C
C EVALCH  (Input)  REAL*8 Array of row dimension 2 and column 
C                  dimension at least N, which contains the known
C                  eigenvalues against which the elements of
C                  EIGVAL are to be compared.
C
C RELERR  (Output) REAL*8 vector of dimension at least N.
C                  RELERR(I) contains the relative error of the
C                  I'th eigenvalue, defined as
C
C                               || EIGVAL(I) - EVALCH(I) ||
C                  RELERR(I) =  ---------------------------
C                                     || EVALCH(I) ||
C
C AVGABS  (Output) The average absolute error.
C
C AVGREL  (Output) The average relative error.
C
C RELMAX  (Output) The maximum relative error.
C
C
C Required routines: None
C
C Revision history:
C
C  Kevin Thompson     December 6, 1991     First version
C
C**********************************************************************
C
      IMPLICIT REAL*8 (A-H,O-Z)
      REAL*8 EIGVAL(2,N),EVALCH(2,N),RELERR(N)
C Compute the errors.
      SUM1=0.D0
      SUM2=0.D0
      RELMAX=0.D0
      DO 10 I=1,N
C
C Error in eigenvalue:
        DELTA=DABS(EIGVAL(1,I)-EVALCH(1,I))
     1       +DABS(EIGVAL(2,I)-EVALCH(2,I))
C
C Norm of eigenvalue:
        RMAG=DABS(EVALCH(1,I))+DABS(EVALCH(2,I))
C
C Accumulate errors and norms:
        SUM1=SUM1+DELTA
        SUM2=SUM2+RMAG
C
C Relative error in eigenvalue:
        IF(RMAG .NE. 0.D0) THEN
          RELERR(I)=DELTA/RMAG
        ELSE
          RELERR(I)=DELTA
        ENDIF
C Largest relative error.
        RELMAX=DMAX1(RELMAX,RELERR(I))
10    CONTINUE
C Absolute absolute error:
      AVGABS=SUM1/N
C Relative relative error:
      DENOM=SUM2/N
      IF(DENOM .NE. 0.D0) THEN
        AVGREL=AVGABS/DENOM
      ELSE
        AVGREL=AVGABS
      ENDIF
      RETURN
      END
      SUBROUTINE EVECCOM(NR,N,A,EIGVEC,EIGVAL,RELERR,RELMAX,IOPT1,IOPT2,
     1                   NP,IVEC)
C
C EVECCOM compares the N computed eigenvectors in EIGVEC with the matrix
C product of A and the eigenvectors.  It computes the relative
C error and the maximum relative error from the absolute error
C
C      Error(j) = A * x(j) - lambda(j) * x(j),                      (1)
C
C which would be zero if lambda(j) and x(j) were exact.  Here lambda(j)
C and x(j) are the j'th eigenvalue and eigenvector, respectively.  The
C first term in (1) is actually a matrix product, and therefore a sum.
C
C The relative error is obtained by dividing the 1-norm of E(j) by the
C 1-norm of Vmag(j),
C
C      Vmag(j) = pos sum [ A * x(j) ] + | lambda(j) * x(j) |,       (2)
C
C where "pos sum" means that the sum in brackets has all terms in
C absolute value.  
C
C When lambda(j) and x(j) are complex, each element in E(j) is the
C square root of the sum of the squares of the real and imaginary
C parts of the error vector elements Error(j).  However, the real
C and imaginary parts of Vmag(j) are computed separately and then
C summed to get a measure of its magnitude for array V(j).
C
C The relative error in the I'th eigenvector is returned as the 
C I'th element of RELERR.  The definitions in (1) and (2) give a
C relative error which is independent of the scaling of A or x.
C
C The IVEC switch selects right eigenvectors (IVEC=0)
C or left eigenvectors (IVEC=1).  In the latter case, the 
C A array must contain the transpose of the original input
C matrix.  (The necessary transposition is performed as needed 
C in this subroutine.)
C
C Calling format:
C
C      CALL EVECCOM(NR,N,A,EIGVEC,EIGVAL,RELERR,RELMAX,IOPT1,IOPT2,
C    1              NP,IVEC)
C
C Parameter list:
C
C     NR  (Input)  Row dimension of A, second dimension of EIGVEC, and
C                  row dimension of WS exactly as specified in the
C                  calling program.
C                  NR must be greater than or equal to N.
C
C      N  (Input)  Number of rows and columns in the input matrix.
C                  Must not exceed 100.
C
C      A  (Input)  Array of row dimension NR and column dimension at
C                  least N, containing the N by N input matrix whose
C                  eigenvalues and eigenvectors have been computed.
C
C EIGVEC  (Input)  Array of dimensions (NP,NR,N), containing the
C                  eigenvectors.  
C
C EIGVAL  (Input)  REAL*8 array of row dimension 2 and column 
C                  dimension at least N, which contains the eigenvalues.
C                  Element EIGVAL(1,I) contains the real part of the 
C                  I'th eigenvalue, while element EIGVAL(2,I) contains
C                  the imaginary part of the I'th eigenvalue.
C
C RELERR  (Output) REAL*8 vector of dimension at least N, containing
C                  the relative error in the eigenvectors.
C                  RELERR(j) = Error(j)/Vmag(j), with Error(j)
C                  and Vmag(j) as defined in Equations (1) and (2).
C
C RELMAX  (Output) The largest relative error.
C
C   IOPT1 (Input)  Eigenvector option parameter
C
C   IOPT2 (Input)  Eigenvector storage option parameter.
C
C      NP (Input)  First (page) dimension of EIGVEC, exactly as
C                  specified in the calling program.  It must be
C                  at least as large the values given in the table
C                  below.
C
C    IVEC (Input)  INTEGER parameter which determines whether the
C                  error analysis is to be performed for the right
C                  or left eigenvectors.
C                     IVEC=0: Analyze the right eigenvectors.
C                     IVEC=1: Analyze the left eigenvectors.
C
C
C Required routines: VASUM
C
C Revision history:
C
C  Kevin Thompson     December 6, 1991     First version
C
C**********************************************************************
C
      IMPLICIT REAL*8 (A-H,O-Z)
      REAL*8 A(NR,N),EIGVEC(NP,NR,N),EIGVAL(2,N),RELERR(N)
      REAL*8 V(100),E(100)
C
C Compute the errors.
      RELMAX=0.D0
C First transpose the matrix in A, if left vectors are to be
C checked.
      IF(IVEC .EQ. 1) THEN
        DO 35 J=1,N
          DO 25 I=J+1,N
            TEMP=A(I,J)
            A(I,J)=A(J,I)
            A(J,I)=TEMP
25        CONTINUE
35      CONTINUE
      ENDIF
C Check for storage allocation.
      IF(IOPT2 .EQ. 2) THEN
C Uncompressed vectors.
C Define a base index for the current real eigenvector part page.
        IF(IVEC .EQ. 0) THEN
          NBASE=IOPT1
        ELSE
          NBASE=1
        ENDIF
C
        DO 10 J=1,N
C Compute the product of A and x(j).
          DO 20 I=1,N
            SUMR=0.D0
            SUMI=0.D0
            ABSUMR=0.D0
            ABSUMI=0.D0
            DO 30 K=1,N
              TERMR=A(I,K)*EIGVEC(NBASE,K,J)
              TERMI=A(I,K)*EIGVEC(NBASE+1,K,J)
              SUMR=SUMR+TERMR
              SUMI=SUMI+TERMI
              ABSUMR=ABSUMR+DABS(TERMR)
              ABSUMI=ABSUMI+DABS(TERMI)
30          CONTINUE
C Store the I'th component of E(j) = A * x(j) - lambda(j) * x(j)
C and V(j).
            TERMR=EIGVAL(1,J)*EIGVEC(NBASE,I,J)
     1         -EIGVAL(2,J)*EIGVEC(NBASE+1,I,J)
            TERMI=EIGVAL(1,J)*EIGVEC(NBASE+1,I,J)
     1         +EIGVAL(2,J)*EIGVEC(NBASE,I,J)
            SUMR=SUMR-TERMR
            SUMI=SUMI-TERMI
            ABSUMR=ABSUMR+DABS(TERMR)
            ABSUMI=ABSUMI+DABS(TERMI)
C 
            E(I)=DABS(SUMR)+DABS(SUMI)
            V(I)=ABSUMR+ABSUMI
20        CONTINUE
C Now compute the norms of E(j) and V(j).
          ENORM=VASUM(E,N,1,1)
          VNORM=VASUM(V,N,1,1)
C The relative error is then
          IF(VNORM .NE. 0.D0) THEN
            RELERR(J)=ENORM/VNORM
          ELSE
            RELERR(J)=ENORM
          ENDIF
C Largest relative error.
          RELMAX=DMAX1(RELMAX,RELERR(J))
10      CONTINUE
C
C*********************************************************************
C
      ELSE
C Compressed vectors.
C Define a base index for the current eigenvector page.
        IF(IVEC .EQ. 0) THEN
          NBASE=(IOPT1-1)/2+1
        ELSE
          NBASE=1
        ENDIF
C
        J=1
1000    CONTINUE
C Check for a real or complex vector.
          IF(EIGVAL(2,J) .LT. 0.D0) THEN
C Complex conjugate pair.
C Compute the product of A and x(j).
            DO 40 I=1,N
              SUMR=0.D0
              SUMI=0.D0
              ABSUMR=0.D0
              ABSUMI=0.D0
              DO 50 K=1,N
                TERMR=A(I,K)*EIGVEC(NBASE,K,J)
                TERMI=A(I,K)*EIGVEC(NBASE,K,J+1)
                SUMR=SUMR+TERMR
                SUMI=SUMI+TERMI
                ABSUMR=ABSUMR+DABS(TERMR)
                ABSUMI=ABSUMI+DABS(TERMI)
50            CONTINUE
C Store the I'th component of E(j) = A * x(j) - lambda(j) * x(j)
C and V(j).
              TERMR=EIGVAL(1,J)*EIGVEC(NBASE,I,J)
     1           -EIGVAL(2,J)*EIGVEC(NBASE,I,J+1)
              TERMI=EIGVAL(1,J)*EIGVEC(NBASE,I,J+1)
     1           +EIGVAL(2,J)*EIGVEC(NBASE,I,J)
              SUMR=SUMR-TERMR
              SUMI=SUMI-TERMI
              ABSUMR=ABSUMR+DABS(TERMR)
              ABSUMI=ABSUMI+DABS(TERMI)
C 
              E(I)=DABS(SUMR)+DABS(SUMI)
              V(I)=ABSUMR+ABSUMI
40          CONTINUE
C Now compute the norms of E(j) and V(j).
            ENORM=VASUM(E,N,1,1)
            VNORM=VASUM(V,N,1,1)
C The relative error is then
            IF(VNORM .NE. 0.D0) THEN
              RELERR(J)=ENORM/VNORM
            ELSE
              RELERR(J)=ENORM
            ENDIF
            RELERR(J+1)=RELERR(J)
C Largest relative error.
            RELMAX=DMAX1(RELMAX,RELERR(J))
C Increment the index.
            J=J+2
          ELSE
C Real vector.
C Compute the product of A and x(j).
            DO 60 I=1,N
              SUMR=0.D0
              ABSUMR=0.D0
              DO 70 K=1,N
                TERMR=A(I,K)*EIGVEC(NBASE,K,J)
                SUMR=SUMR+TERMR
                ABSUMR=ABSUMR+DABS(TERMR)
70            CONTINUE
C Store the I'th component of E(j) = A * x(j) - lambda(j) * x(j)
C and V(j).
              TERMR=EIGVAL(1,J)*EIGVEC(NBASE,I,J)
              SUMR=SUMR-TERMR
              ABSUMR=ABSUMR+DABS(TERMR)
C 
              E(I)=DABS(SUMR)
              V(I)=ABSUMR
60          CONTINUE
C Now compute the norms of E(j) and V(j).
            ENORM=VASUM(E,N,1,1)
            VNORM=VASUM(V,N,1,1)
C The relative error is then
            IF(VNORM .NE. 0.D0) THEN
              RELERR(J)=ENORM/VNORM
            ELSE
              RELERR(J)=ENORM
            ENDIF
C Largest relative error.
            RELMAX=DMAX1(RELMAX,RELERR(J))
C Increment the index.
            J=J+1
          ENDIF
C Do the next vector.
        IF(J .LE. N) GO TO 1000
C
C*********************************************************************
C
      ENDIF
C Put the matrix back in its original form, if it has been transposed.
      IF(IVEC .EQ. 1) THEN
        DO 45 J=1,N
          DO 55 I=J+1,N
            TEMP=A(I,J)
            A(I,J)=A(J,I)
            A(J,I)=TEMP
55        CONTINUE
45      CONTINUE
      ENDIF
      RETURN
      END
      SUBROUTINE INTEST(INFILE,OUTFILE,DIAGFILE)
C
C INTEST performs a standard initialization process for the 
C Numerical Analyst automated testing programs.  All such
C programs call INTEST to generate output file names and
C unit numbers, open output files, write out beginning-of-test
C information, and skip past the input file header for
C subsequent reads.
C
C Calling format:
C        
C        CALL INTEST(INFILE,OUTFILE,DIAGFILE)
C
C Parameter list:
C
C    INFILE (Output) INTEGER unit number for input file.
C   OUTFILE (Output) INTEGER unit number for output file.
C  DIAGFILE (Output) INTEGER unit number for diagnostic file.
C
C  Required routines: TSTMSG
C
C Revision history:
C
C  Kevin W. Thompson     August 31, 1991     First version
C
C************************************************************************
C
C Define character variables for file names.
      CHARACTER*60 MES
      CHARACTER*12 INNAME,OUTNAME,DIAGNAME
      CHARACTER*8 TESTNAME
      CHARACTER*1 EHEAD
C
      LOGICAL PASS
C
      INTEGER INFILE,OUTFILE,DIAGFILE
      INTEGER IDOT,LENSTR
C Define file unit numbers.  INFILE is normally standard input
C (corresponding to keyboard).  OUTFILE and DIAGFILE should
C not correspond to compiler-preset unit numbers.
      INFILE=5
      OUTFILE=7
      DIAGFILE=8
C
C Initialize name test.
      PASS=.TRUE.
C
C Read the input file name from the input file.  It has the format
C S<subname>[m].Inn, where <subname> is the name of the subroutine
C to be tested (up to six characters), m is an optional number
C indicating which of multiple test programs for this subroutine
C the input file corresponds to, and nn is a one or two digit
C number indicating which of multiple test problems is stored
C in this file.
C
      READ(INFILE,10)INNAME
10    FORMAT(A12)
C
C Find the dot in the input file name.
      LENSTR=LEN(INNAME)
      DO 20 I=2,LENSTR
        IF(INNAME(I:I) .EQ. '.') THEN
          IDOT=I
          GO TO 21
        ENDIF
20    CONTINUE
C The name is invalid.
        PASS=.FALSE.
21    CONTINUE
C
C Define the other file names, replacing the I after the dot by O or D.
      OUTNAME=INNAME(1:IDOT)//'O'//INNAME(IDOT+2:)
      DIAGNAME=INNAME(1:IDOT)//'D'//INNAME(IDOT+2:)
C The test module name omits the extension from the input file name.
      TESTNAME=INNAME(1:IDOT-1)
C Open the output files.
      OPEN(OUTFILE,FILE=OUTNAME,STATUS='UNKNOWN')
      OPEN(DIAGFILE,FILE=DIAGNAME,STATUS='UNKNOWN')
C Print the starting message to the diagnostic file.
      WRITE(DIAGFILE,30)TESTNAME,INNAME,OUTNAME
30    FORMAT(' START: Test module ',A,
     1 ', Input file ',A,', Output file ',A)
C
C Quit if the file name is bad.
      IF(.NOT. PASS) THEN
        MES=
     1  'Invalid input file name.                                    '
        CALL TSTMSG(PASS,DIAGFILE,MES)
      ENDIF
C
C If the test program aborts, some of the information written to
C the diagnostic file may be buffered for a later write and may
C be lost.  Close the diagnostic file to force any buffered
C output to be written now.  (If the compiler does not support
C the ACCESS=APPEND parameter in the OPEN statement, then the
C CLOSE and OPEN statements should be commented out, or the
C file contents will be overwritten after OPENing.)
      CLOSE(DIAGFILE)
C
C Reopen the diagnostic file for further writes.
      OPEN(DIAGFILE,FILE=DIAGNAME,STATUS='OLD',ACCESS='APPEND')
C
C Read past the input file header, which is ended by
C a number symbol (#) in the first column.
40    CONTINUE
        READ(INFILE,42)EHEAD
42      FORMAT(A1)
      IF(EHEAD .NE. '#') GO TO 40
C
      RETURN
      END
      SUBROUTINE MACON(PRECSN,RMIN,RMAX)
C
C MACON provides values for the machine-dependent constants
C pertaining to precision and range of floating point numbers.
C
C Parameter list:
C
C  PRECSN (Output) The machine precision.
C  RMIN   (Output) The largest representable number.
C  RMAX   (Output) The smallest representable number.
C
C Required Numerical Analyst routines: None
C
C Revision history:
C
C  Kevin W. Thompson    November 11, 1991     First  version
C
C***********************************************************************
C
C *** This data comes from IEEE.INC, the
C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
C
      REAL*8 RMIN,RMAX,PRECSN
      PRECSN=2.22D-16
      RMIN=2.23D-308
      RMAX=1.79D308
C
      RETURN
      END
      real*8 function ddot(n,dx,incx,dy,incy)
c
c     forms the dot product of two vectors.
c     uses unrolled loops for increments equal to one.
c     jack dongarra, linpack, 3/11/78.
c
      implicit real*8 (a-h,o-z)
      integer i,incx,incy,ix,iy,m,mp1,n
      dimension dx(*),dy(*)
C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
      REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
      INTEGER IWARN,IFATAL,ISYSTM
      CHARACTER MES*60,EFNAME*32
      DATA RMIN/2.23D-308/,RMAX/1.79D308/
      DATA RMNHAF/1.50D-154/,RMXHAF/1.33D154/
      DATA PRECSN/2.22D-16/,PRCHAF/1.5D-8/
      DATA IWARN/1/,IFATAL/2/,ISYSTM/3/
      DATA MES/
     1 'MAGUS Numerical Analyst 3.0.  Copyright (C) 1990 by MAGUS.  '/
C *** End of INCLUDE file.
c
      ddot = 0.0d0
      dtemp = 0.0d0
      if(n.le.0)return
      if(incx.eq.1.and.incy.eq.1)go to 20
c
c        code for unequal increments or equal increments
c          not equal to 1
c
      ix = 1
      iy = 1
      if(incx.lt.0)ix = (-n+1)*incx + 1
      if(incy.lt.0)iy = (-n+1)*incy + 1
      do 10 i = 1,n
        dtemp = dtemp + dx(ix)*dy(iy)
        ix = ix + incx
        iy = iy + incy
   10 continue
      ddot = dtemp
      return
c
c        code for both increments equal to 1
c
c
   20 CONTINUE
C*********************************************************************
C
C Modifications by Kevin Thompson, 12/30/89
C Put in alternate loop for use on vector machines.  To vectorize
C more efficiently, remove comment from DO 30 through DDOT=... below,
C and comment out all subsequent lines before the RETURN statement.
C     DO 30 I=1,N
C       DTEMP=DTEMP+DX(I)*DY(I)
C30   CONTINUE
C     DDOT=DTEMP
C
C*********************************************************************
C
c        clean-up loop
c
      m = mod(n,5)
      if( m .eq. 0 ) go to 40
      do 30 i = 1,m
        dtemp = dtemp + dx(i)*dy(i)
   30 continue
      if( n .lt. 5 ) go to 60
   40 mp1 = m + 1
      do 50 i = mp1,n,5
        dtemp = dtemp + dx(i)*dy(i) + dx(i + 1)*dy(i + 1) +
     *   dx(i + 2)*dy(i + 2) + dx(i + 3)*dy(i + 3) + dx(i + 4)*dy(i + 4)
   50 continue
   60 ddot = dtemp
      return
      end
      subroutine  dswap (n,dx,incx,dy,incy)
c
c     interchanges two vectors.
c     uses unrolled loops for increments equal one.
c     jack dongarra, linpack, 3/11/78.
c
      implicit real*8 (a-h,o-z)
      integer i,incx,incy,ix,iy,m,mp1,n
      dimension dx(*),dy(*)
C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
      REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
      INTEGER IWARN,IFATAL,ISYSTM
      CHARACTER MES*60,EFNAME*32
      DATA RMIN/2.23D-308/,RMAX/1.79D308/
      DATA RMNHAF/1.50D-154/,RMXHAF/1.33D154/
      DATA PRECSN/2.22D-16/,PRCHAF/1.5D-8/
      DATA IWARN/1/,IFATAL/2/,ISYSTM/3/
      DATA MES/
     1 'MAGUS Numerical Analyst 3.0.  Copyright (C) 1990 by MAGUS.  '/
C *** End of INCLUDE file.
c
      if(n.le.0)return
      if(incx.eq.1.and.incy.eq.1)go to 20
c
c       code for unequal increments or equal increments not equal
c         to 1
c
      ix = 1
      iy = 1
      if(incx.lt.0)ix = (-n+1)*incx + 1
      if(incy.lt.0)iy = (-n+1)*incy + 1
      do 10 i = 1,n
        dtemp = dx(ix)
        dx(ix) = dy(iy)
        dy(iy) = dtemp
        ix = ix + incx
        iy = iy + incy
   10 continue
      return
c
c       code for both increments equal to 1
c
c
   20 CONTINUE
C*********************************************************************
C
C Modifications by Kevin Thompson, 12/30/89
C Put in alternate loop for use on vector machines.  To vectorize
C more efficiently, remove comment from DO 30 through 30 CONTINUE 
C below, and comment out all subsequent lines before the RETURN 
C statement.
C     DO 30 I=1,N
C       DTEMP=DX(I)
C       DX(I)=DY(I)
C       DY(I)=DTEMP
C30   CONTINUE
C
C*********************************************************************
C
c       clean-up loop
c
      m = mod(n,3)
      if( m .eq. 0 ) go to 40
      do 30 i = 1,m
        dtemp = dx(i)
        dx(i) = dy(i)
        dy(i) = dtemp
   30 continue
      if( n .lt. 3 ) return
   40 mp1 = m + 1
      do 50 i = mp1,n,3
        dtemp = dx(i)
        dx(i) = dy(i)
        dy(i) = dtemp
        dtemp = dx(i + 1)
        dx(i + 1) = dy(i + 1)
        dy(i + 1) = dtemp
        dtemp = dx(i + 2)
        dx(i + 2) = dy(i + 2)
        dy(i + 2) = dtemp
   50 continue
      return
      end
      SUBROUTINE EBAKBF(A,NP,NR,N,D,ILOW,IHI,IR)
C
C This subroutine backward transforms a set of right hand eigenvectors
C of a balanced matrix into the eigenvectors of the original matrix,
C from which the original matrix was derived by a call to EBALNF.
C The algorigthm used is that of Parlett and Reinsch in 
C Numerische Mathematik, vol. 13, p. 293-304 (1969), and HACLA II/11.
C
C CALLING FORMAT:
C
C      CALL EBAKBF(A,NP,NR,N,D,ILOW,IHI,IR)
C
C PARAMETER LIST:
C
C      A  (Input)  Array of dimensions (NP,NR,IR), containing the
C                  IR eigenvectors or matrix columns to be
C                  transformed.  
C
C                  The vectors are stored as columns in A.
C                  Real and imaginary parts, stored as separate
C                  columns, are transformed in exactly the same
C                  fashion and do not require separate treatment.
C                  Element I of column J is stored as A(1,I,J),
C                  independent of the value of NP.  Thus only
C                  elements with a first index of 1 are actually
C                  transformed.
C
C                  The matrix columns have their elements (I,J)
C                  stored as A(1,I,J), independent of the value of NP.
C                  Thus only elements with a first index of 1 are
C                  actually transformed.
C
C         (Output) The transformed eigenvectors or matrix.
C     NP  (Input)  The first dimension of A, exactly as specified in
C                  the calling program.  If NP = 1, then A may be
C                  represented as a two dimensional array in the
C                  calling program.
C     NR  (Input)  Row dimension of A exactly as specified in the
C                  calling program.  NR must be greater than or equal
C                  to N.
C      N  (Input)  Number of elements in the input eigenvectors.
C      D  (Input)  A vector of length N, containing the diagonal
C                  elements of the diagonal transformation matrix
C                  which was used to balance the original matrix in 
C                  EBALNF, in positions ILOW through IHI.  
C
C                  Outside the range from ILOW to IHI, the value
C                  of D(I) is the the index for the row and column
C                  which were interchanged with those of index I.
C   ILOW  (Input)  The initial row and column of the central submatrix.
C    IHI  (Input)  The final row and column of the central submatrix.
C     IR  (Input)  The number of eigenvectors or matrix columns to
C                  be transformed.
C
C Required routines: DSWAP
C
C Revision history:
C
C Kevin Thompson     October 31, 1987     Version 3.0
C
C Kevin Thompson     March 16, 1991       Version 3.0
C   Modified DO 10 loop to minimize memory strides and maximize
C   locality and cache efficiency.  Rewrote DO 30 and DO 50 loops
C   to use the BLAS-1 DSWAP routine for swapping rows and columns.
C
C**********************************************************************
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION A(NP,NR,IR),D(N)
C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
      REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
      INTEGER IWARN,IFATAL,ISYSTM
      CHARACTER MES*60,EFNAME*32
      DATA RMIN/2.23D-308/,RMAX/1.79D308/
      DATA RMNHAF/1.50D-154/,RMXHAF/1.33D154/
      DATA PRECSN/2.22D-16/,PRCHAF/1.5D-8/
      DATA IWARN/1/,IFATAL/2/,ISYSTM/3/
      DATA MES/
     1 'MAGUS Numerical Analyst 3.0.  Copyright (C) 1990 by MAGUS.  '/
C *** End of INCLUDE file.
C Perform the backward transformation
      DO 10 J=1,IR
        DO 20 I=ILOW,IHI
          A(1,I,J)=D(I)*A(1,I,J)
20      CONTINUE
10    CONTINUE
C
C Left eigenvectors are back transformed if the above statement 
C divides by D(I) instead of multiplying by it.
C
C Now perform the row interchanges.
C
      INCA=NP*NR
      DO 30 I=ILOW-1,1,-1
        K=D(I)
        IF(K .NE. I) THEN
          CALL DSWAP(IR,A(1,I,1),INCA,A(1,K,1),INCA)
        ENDIF
30    CONTINUE
      DO 50 I=IHI+1,N
        K=D(I)
        IF(K .NE. I) THEN
          CALL DSWAP(IR,A(1,I,1),INCA,A(1,K,1),INCA)
        ENDIF
50    CONTINUE
      RETURN
      END
      SUBROUTINE EBAKRF(A,TRANS,EIGVAL,WS,NP,NR,N,ILOW,IHI)
C
C This subroutine performs the back transformation of the
C eigenvectors of a nearly triangular matrix to those of a
C full matrix.
C
C CALLING FORMAT:
C
C      CALL EBAKRF(A,TRANS,EIGVAL,WS,NP,NR,N,ILOW,IHI)
C
C PARAMETER LIST:
C
C      A  (Input)  Array of row dimension NR and column dimension at
C                  least N, containing the N right eigenvectors.
C                  The eigenvectors are stored in compressed format
C                  in the columns of A.
C
C         (Output) The transformed eigenvectors, also compressed.
C
C  TRANS  (Input/  Array of dimensions (NP,NR,N), containing the
C          Output) cumulative transformation matrix elements as
C                  TRANS(1,I,J).
C
C EIGVAL  (Output) REAL*8 Array of row dimension 2 and column 
C                  dimension at least N, which contains the eigenvalues
C                  computed by this subroutine.  Element EIGVAL(1,I)
C                  contains the real part of the I'th eigenvalue,
C                  while element EIGVAL(2,I) contains the imaginary
C                  part of the I'th eigenvalue.
C
C     WS           Vector of dimension at least N, used for scratch
C                  space.
C
C     NP  (Input)  NP is the first dimension of the transformation
C                  matrix array, TRANS.
C
C     NR  (Input)  Row dimension of A and second dimension of TRANS
C                  exactly as specified in the calling program.  NR
C                  must be greater than or equal to N.
C
C      N  (Input)  Number of rows and columns in the input matrix.
C
C   ILOW  (Input)  Starting index for the active central sub-matrix.
C
C    IHI  (Input)  Ending index for the active central sub-matrix.
C
C Required routines: None
C
C Revision history:
C
C Kevin Thompson     October 31, 1987     Version 3.0
C
C Kevin Thompson     June 8, 1991         Version 3.0
C   Replaced the original matrix-vector product by a call to the
C   BLAS-1 routine DDOT.  Also replaced a copy loop with a call
C   to the BLAS-1 DCOPY.
C
C**********************************************************************
C
      IMPLICIT REAL*8 (A-H,O-Z)
      REAL*8 A(NR,N),TRANS(NP,NR,N),EIGVAL(2,N),WS(N)
C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
      REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
      INTEGER IWARN,IFATAL,ISYSTM
      CHARACTER MES*60,EFNAME*32
      DATA RMIN/2.23D-308/,RMAX/1.79D308/
      DATA RMNHAF/1.50D-154/,RMXHAF/1.33D154/
      DATA PRECSN/2.22D-16/,PRCHAF/1.5D-8/
      DATA IWARN/1/,IFATAL/2/,ISYSTM/3/
      DATA MES/
     1 'MAGUS Numerical Analyst 3.0.  Copyright (C) 1990 by MAGUS.  '/
C *** End of INCLUDE file.
C
C Perform the back transformation on the compressed eigenvectors
C stored in A, overwriting A with the new compressed vectors.  Only
C the elements which change in the transformation are explicitly
C transformed.
C
      LDA=NP*NR
C
      DO 800 J=ILOW,N
C Define an upper limit for this loop which works for both the real and
C imaginary parts.
        IF(EIGVAL(2,J) .GE. 0.D0) THEN
          ITOP=J
        ELSE
          ITOP=J+1
        ENDIF
C Copy this column of A to WS.
        CALL DCOPY(ITOP-ILOW+1,A(ILOW,J),1,WS(ILOW),1)
C Transform this column.
C Alliant directive allowing concurrent calls to DDOT:
cvd$  cncall
        DO 820 I=ILOW,IHI
          A(I,J)=DDOT(ITOP-ILOW+1,TRANS(1,I,ILOW),LDA,WS(ILOW),1)
820     CONTINUE
800   CONTINUE
      RETURN
      END
      SUBROUTINE EBALNF(A,NR,N,NBASE,D,ILOW,IHI)
C
C This subroutine balances a full matrix, using the method Parlett
C and Reinsch in Numerische Mathematik, vol. 13, p. 293-304 (1969),
C and HACLA II/11.
C The balancing is done to improve the accuracy of subsequent
C eigenvalue and eigenvector calculations performed by EFULRC,
C by reducing the norm of the matrix.  The reduction of the norm
C is obtained by a particular sequence of diagonal similarity
C transformations, and is exact (free of rounding errors) if NBASE 
C is the base of the computer's numbering system.
C
C The rows and columns of the input matrix may contain a sufficient
C number of zero elements that rearranging them will isolate
C eigenvalues along the diagonal.  If this is so, then EBALNF will
C interchange rows and columns so that such eigenvalues are positioned
C along the diagonal, above and below a central non-triangular 
C submatrix.  The initial and final rows (and columns) of the central
C submatrix will be ILOW and IHI.
C
C
C CALLING FORMAT:
C
C      CALL EBALNF(A,NR,N,NBASE,D,ILOW,IHI)
C
C PARAMETER LIST:
C
C      A  (Input)  Array of row dimension NR and column dimension at
C                  least N, containing the N by N input matrix to be
C                  balanced.
C         (Output) The balanced matrix.
C     NR  (Input)  Row dimension of A exactly as specified in the
C                  calling program.  NR must be greater than or equal
C                  to N.
C      N  (Input)  Number of rows and columns in the input matrix.
C  NBASE  (Input)  The numbering system base assumed for this computer
C                  and language implementation.
C      D  (Output) A vector of length N, containing the diagonal
C                  elements of the diagonal transformation matrix
C                  which was used to balance the matrix in A, in
C                  positions ILOW through IHI.
C
C                  Outside the range from ILOW to IHI, the value
C                  of D(I) is the the index for the row and column
C                  which were interchanged with those of index I.
C   ILOW  (Output) The initial row and column of the central submatrix.
C    IHI  (Output) The final row and column of the central submatrix.
C
C Required routines: EINTF
C
C Revision history:
C
C Kevin Thompson     November 11, 1987     Version 3.0
C
C Kevin Thompson     May 25, 1991          Version 3.0
C  Replaced the scaling loops with calls to DSCAL.
C
C
C**********************************************************************
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION A(NR,N),D(N)
      LOGICAL NOCONV
C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
      REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
      INTEGER IWARN,IFATAL,ISYSTM
      CHARACTER MES*60,EFNAME*32
      DATA RMIN/2.23D-308/,RMAX/1.79D308/
      DATA RMNHAF/1.50D-154/,RMXHAF/1.33D154/
      DATA PRECSN/2.22D-16/,PRCHAF/1.5D-8/
      DATA IWARN/1/,IFATAL/2/,ISYSTM/3/
      DATA MES/
     1 'MAGUS Numerical Analyst 3.0.  Copyright (C) 1990 by MAGUS.  '/
C *** End of INCLUDE file.
C Initialize some variables.  BASE2 is the square of the number system
C base, ILOW is the lower index limit, and IHI is the upper index
C limit.
      BASE=NBASE
      BASE2=BASE*BASE
      ILOW=1
      IHI=N
      NEWLOW=ILOW
      NEWHI=IHI
C Search for rows which isolate an eigenvalue, and push them down.
1000  CONTINUE
      IHI=NEWHI
      DO 10 I=IHI,ILOW,-1
C Compute partial row norms, excluding the diagonal element and those
C elements which are outside the current sub-matrix from ILOW to IHI.
        IP1=MIN0(I+1,IHI)
        R=DASUM(I-1,A(I,1),NR)+DASUM(IHI-I,A(I,IP1),NR)
        IF(R .EQ. 0.D0) THEN
C Interchange rows I and IHI, then columns I and IHI.  Store the
C interchange index in D(IHI), and decrement IHI.
          CALL EINTF(A,NR,N,ILOW,IHI,I,IHI)
          D(IHI)=I
          NEWHI=IHI-1
          GO TO 1000
        ENDIF
10    CONTINUE
C Next search for columns which isolate an eigenvalue, and push them
C left.
2000  CONTINUE
      ILOW=NEWLOW
      DO 40 J=ILOW,IHI
C Compute partial column norms.
        JP1=MIN0(J+1,IHI)
        C=DASUM(J-ILOW,A(ILOW,J),1)+DASUM(IHI-J,A(JP1,J),1)
        IF(C .EQ. 0.D0) THEN
C Interchange rows J and ILOW, then columns J and ILOW.  Store the
C interchange index in D(ILOW), and increment ILOW.
          CALL EINTF(A,NR,N,ILOW,IHI,ILOW,J)
          D(ILOW)=J
          NEWLOW=ILOW+1
          GO TO 2000
        ENDIF
40    CONTINUE
C
C Now balance the submatrix in rows ILOW through IHI.
C Initialize the transformation matrix elements to the identity matrix.
      DO 70 I=ILOW,IHI
        D(I)=1.D0
70    CONTINUE
C
      IF(IHI .GT. ILOW) THEN
5000    CONTINUE
        NOCONV=.FALSE.
C Balancing loop.
        DO 80 I=ILOW,IHI
          IP1=MIN0(I+1,IHI)
          C=DASUM(I-ILOW,A(ILOW,I),1)+DASUM(IHI-I,A(IP1,I),1)
          R=DASUM(I-ILOW,A(I,ILOW),NR)+DASUM(IHI-I,A(I,IP1),NR)
          S=C+R
C Compute the balancing factor.
          RATIO=R/C
          F=1.D0
C Test for the proper balancing factor.
          IF(RATIO .GT. BASE) THEN
C Increase F by factors of BASE.
            X=RATIO/BASE
6000        CONTINUE
              F=F*BASE
              X=X/BASE
            IF(X .GT. F) GO TO 6000
          ELSEIF(1.D0 .GE. BASE*RATIO) THEN
C Decrease F by factors of BASE.
            X=RATIO*BASE
7000        CONTINUE
              F=F/BASE
              X=X*BASE
            IF(F .GE. X) GO TO 7000
          ENDIF
          IF((C*F+R/F) .LT. 0.95*S) THEN
C Do the transformation.
            G=1.D0/F
            D(I)=F*D(I)
            NOCONV=.TRUE.
            CALL DSCAL(N-ILOW+1,G,A(I,ILOW),NR)
            CALL DSCAL(IHI,F,A(1,I),1)
          ENDIF
80      CONTINUE
C Check for convergence, and repeat if not achieved.
        IF(NOCONV) GO TO 5000
      ENDIF
      RETURN
      END
      SUBROUTINE EHESSF(A,TRANS,WS,IWS,NP,NR,N,ILOW,IHI,COMVEC)
C
C This subroutine performs a similarity transformation on a matrix A,
C converting it directly to upper Hessenberg form.  If we represent
C the Hessenberg matrix by H, and the transformation matrix by T, then
C the transformation can be written
C
C               -1
C          H = T  A T,     or equivalently     A T = T H.     (1)
C
C The H matrix has the same eigenvalues as A.  The upper triangle,
C main diagonal, and uppermost lower diagonal are generally nonzero.
C All other elements are zero.
C
C The T matrix is in lower triangular form, with ones on the principal
C diagonal, and all elements in the first column (except the top one)
C zero.  Thus equation (1) may be written
C
C  _                               _   _                          _
C | A(1,1) A(1,2) A(1,3) A(1,4) ... | | 1                          |
C | A(2,1) A(2,2) A(2,3) A(2,4) ... | | 0 1                        |
C | A(3,1) A(3,2) A(3,3) A(3,4) ... | | 0 T(3,2) 1                 |
C | A(4,1) A(4,2) A(4,3) A(4,4) ... | | 0 T(4,2) T(4,3) 1          |
C | ...    ...    ...    ...    ... | | 0 ...    ...    ...    1   |
C |_                               _| |_                          _|
C
C                                                                   (2)
C     _                         _   _                               _
C    | 1                         | | H(1,1) H(1,2) H(1,3) H(1,4) ... |
C    | 0 1                       | | H(2,1) H(2,2) H(2,3) H(2,4) ... |
C  = | 0 T(3,2) 1                | | 0      H(3,2) H(3,3) H(3,4) ... | .
C    | 0 T(4,2) T(4,3) 1         | | 0      0      H(4,3) H(4,4) ... |
C    | 0 ...    ...    ...    1  | | ...    ...    ...    ...    ... |
C    |_                         _| |_                               _|
C
C The non-trivial elements of T, together with the nonzero elements 
C of H, may be stored conveniently (during the calculation) in the 
C array A, which originally contained the input matrix, in the form
C             _                               _
C            | H(1,1) H(1,2) H(1,3) H(1,4) ... |
C            | H(2,1) H(2,2) H(2,3) H(2,4) ... |
C            | T(3,2) H(3,2) H(3,3) H(3,4) ... |
C   A(I,J) = | T(4,2) T(4,3) H(4,3) H(4,4) ... | .        (3)
C            | T(5,2) T(5,3) T(5,4) H(5,4) ... |
C            | ...    ...    ...    ...    ... |
C            |_                               _|
C
C This is the contents of the new array A--in the absence of
C pivoting.  It is necessary to perform simultaneous row and
C column interchanges (pivoting) in order to avoid the buildup
C of roundoff error and possible division by zero.  Fortunately,
C the actual computed array has the above form even after 
C pivoting.  The array IWS holds the pivot indices.
C
C After the transformation is complete, the rows and
C columns have the form of Equation (3).  However, the
C elements of T, stored in this triangular form, do
C not transform A to H.  Instead, they transform a matrix
C A' to H, where A' is that matrix which is obtained from
C A by performing the sequence of row and column interchanges
C described above on A; i.e., it is the permuted matrix,
C equivalent to A, which may be transformed to Hessenberg
C form without a need for pivoting.
C
C The transformation matrix which connects the original (non-
C permuted) matrix A and the Hessenberg matrix H is not lower
C triangular, but can be obtained by interchanging rows of the
C T matrix.  The resulting transformation matrix, TRANS, does
C satisfy A * TRANS = TRANS * H, where A is the original matrix, and
C is generated on output, if desired.
C
C The transformation matrix elements remaining in the array A are
C then overwritten with zeroes so that the output array contains
C only the Hessenberg matrix.
C
C CALLING FORMAT:
C
C      CALL EHESSF(A,TRANS,WS,IWS,NP,NR,N,ILOW,IHI,COMVEC)
C
C PARAMETER LIST:
C
C      A  (Input)  Array of row dimension NR and column dimension at
C                  least N, containing the N by N input matrix to be 
C                  converted into Hessenberg form.
C         (Output) On output, A contains the Hessenberg matrix, as 
C                  described above.
C  TRANS  (Output) Array of dimensions (NP,NR,N) which contains the
C                  transformation matrix TRANS such that 
C                          A * TRANS = TRANS * H,
C                  with A the original input matrix and H the output
C                  upper Hessenberg matrix.  The transformation
C                  element (I,J) is stored as TRANS(1,I,J).  The
C                  calculation of TRANS is performed only if
C                  COMVEC .EQ. .TRUE.
C     WS           Array of row dimension 2 and column dimension at 
C                  least N, used as a temporary work space.
C    IWS           Integer array of length at least N, used to store
C                  the pivot indices.
C     NP  (Input)  NP is the first dimension of the transformation
C                  array, TRANS.  If COMVEC = .FALSE., then NP should 
C                  be set to zero to prevent wasted space.  Otherwise
C                  COMVEC = .TRUE. and NP is either 1, 2, or 4.
C                  Note that only elements TRANS(1,I,J) are defined
C                  by this subroutine, even if NP is 2.
C     NR  (Input)  Row dimension of A and TRANS exactly as specified
C                  in the calling program.  NR must be greater than or
C                  equal to N.
C      N  (Input)  Number of rows and columns in the input matrix.
C   ILOW  (Input)  The beginning index for the (sub)matrix to be
C                  converted to Hessenberg form.
C    IHI  (Input)  The ending index for the (sub)matrix to be
C                  converted.
C COMVEC  (Input)  Transformation matrix calculation parameter.  The
C                  value of COMVEC determines whether the 
C                  transformation matrix is to be computed, as 
C                  described below.  The transformation matrix is
C                  needed to compute the eigenvectors later.
C
C                  COMVEC = .FALSE.: Do not compute the transformation
C                                    matrix.
C                  COMVEC = .TRUE. : Compute the transformation matrix.
C
C
C Required routines: DDOT, DASUM, DSCAL, EINTF, IDAMAX
C
C Revision history:
C
C Kevin Thompson     November 13, 1987     Version 3.0
C
C Kevin Thompson     March 16, 1991        Version 3.0
C   Replaced many loops with calls to BLAS routines to improve
C   efficiency.  Modified initialization of TRANS to allow
C   vectorization and concurrency.
C
C**********************************************************************
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION A(NR,N),TRANS(NP,NR,N),WS(2,N),IWS(N)
      LOGICAL COMVEC
C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
      REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
      INTEGER IWARN,IFATAL,ISYSTM
      CHARACTER MES*60,EFNAME*32
      DATA RMIN/2.23D-308/,RMAX/1.79D308/
      DATA RMNHAF/1.50D-154/,RMXHAF/1.33D154/
      DATA PRECSN/2.22D-16/,PRCHAF/1.5D-8/
      DATA IWARN/1/,IFATAL/2/,ISYSTM/3/
      DATA MES/
     1 'MAGUS Numerical Analyst 3.0.  Copyright (C) 1990 by MAGUS.  '/
C *** End of INCLUDE file.
C
C Begin by setting up the index array.
      DO 5 I=1,N
        IWS(I)=I
5     CONTINUE
C
C If the input matrix is already in Hessenberg form, then we
C should omit this calculation entirely.  Otherwise we should
C skip over any leading columns which are in Hessenberg form.
      J=ILOW-1
20    CONTINUE
        J=J+1
C Get the norm of the sub-Hessenberg elements in this column.
        SUM=DASUM(IHI-J-1,A(J+2,J),1)
      IF(SUM .EQ. 0.D0 .AND. J .LT. IHI-1) GO TO 20
      JSTART=J
C Skip the transformation if JSTART indicates a Hessenberg matrix.
      IF(JSTART .LT. IHI-1) THEN
C
C The first nontrivial element of the first nontrivial column in the
C Hessenberg matrix, H(JSTART,JSTART), is simply A(JSTART,JSTART),
C which is therefore left unaltered.
C
C To get the second (and final) element of this column of H,
C H(JSTART+1,JSTART), first find the maximum element of A(I,JSTART), 
C for I .GE. JSTART+1.
C
        IMAX=IDAMAX(IHI-JSTART,A(JSTART+1,JSTART),1)+JSTART
        WMAX=A(IMAX,JSTART)
C
C Interchange rows JSTART+1 and IMAX, and columns JSTART+1 and IMAX, to
C select the largest element as H(JSTART+1,JSTART).  Store the new
C pivot indices.
C
        CALL EINTF(A,NR,N,ILOW,IHI,JSTART+1,IMAX)
C  
        ITEMP=IWS(IMAX)
        IWS(IMAX)=IWS(JSTART+1)
        IWS(JSTART+1)=ITEMP
C
C Now H(JSTART+1,JSTART) is the new A(JSTART+1,JSTART), so it need not be
C computed explicitly.
C
C Compute the next column of the transformation matrix, T(I,JSTART+1),
C and store in A(I,JSTART).  Set T(I,JSTART+1) = 0 if
C H(JSTART+1,JSTART) = 0.
        IF(WMAX .EQ. 0.D0) THEN
          HINV=0.D0
        ELSE
          HINV=1.D0/WMAX
        ENDIF
C
        CALL DSCAL(IHI-JSTART-1,HINV,A(JSTART+2,JSTART),1)
C
C Now loop over the remaining columns, JSTART+1 through N.
C
        DO 40 J=JSTART+1,N
C  
C Compute H(I,J) for I .LE. J, if J .LE. IHI (central sub-matrix);
C and for JSTART .LE. I .LE. IHI, if J .GT. IHI (right central
C sub-matrix).
          IF(J .LE. IHI) THEN
            ISTART=1
            ISTOP=J
          ELSE
            ISTART=JSTART
            ISTOP=IHI
          ENDIF
C
          JP1=MIN0(J+1,IHI)
          DO 50 I=ISTART,ISTOP
            SUM1=DDOT(IHI-J,A(I,JP1),NR,A(JP1,J-1),1)
            SUM2=DDOT(I-JSTART-1,A(I,JSTART),NR,A(JSTART+1,J),1)
C
            A(I,J)=A(I,J)+SUM1-SUM2
50        CONTINUE
C
C Compute possible values of H(J+1,J) and find the largest.
C
          WMAX=0.D0
          IMAX=J+1
          DO 80 I=J+1,IHI
            SUM1=DDOT(IHI-J,A(I,J+1),NR,A(J+1,J-1),1)
            SUM2=DDOT(J-JSTART,A(I,JSTART),NR,A(JSTART+1,J),1)
            WS(1,I)=A(I,J)+SUM1-SUM2
80        CONTINUE
C
C If J is less than IHI, interchange rows J+1 and IMAX, and columns 
C J+1 and IMAX, to make H(J+1,J) as large as possible.
C
          IF(J .LT. IHI) THEN
            IMAX=IDAMAX(IHI-J,WS(1,J+1),2)+J
            WMAX=WS(1,IMAX)
C
            CALL EINTF(A,NR,N,ILOW,IHI,J+1,IMAX)
C
            TEMP=WS(1,J+1)
            WS(1,J+1)=WS(1,IMAX)
            WS(1,IMAX)=TEMP
C
            ITEMP=IWS(IMAX)
            IWS(IMAX)=IWS(J+1)
            IWS(J+1)=ITEMP
C Assign the largest value to H(J+1,J).
C
            A(J+1,J)=WMAX
          ENDIF
C  
C Having pivoted to maximize H(J+1,J), solve for the next column (J+1)
C of the transformation matrix.  If WMAX = 0, all numerators and the
C denominator H(J+1,J) are zero, and the transformation matrix elements
C are arbitrary, and most conveniently set to zero.
C
          IF(WMAX .EQ. 0.D0) THEN
            HINV=0.D0
          ELSE
            HINV=1.D0/WMAX
          ENDIF
C
          DO 110 I=J+2,IHI
            A(I,J)=HINV*WS(1,I)
110       CONTINUE
C
C End the loop and go to the next column.
C
40      CONTINUE
      ENDIF
C
C The transformation is complete.  Now set up the permuted 
C transformation matrix which transforms A to H, by interchanging
C rows of the triangular transformation matrix already stored.
C
      IF(COMVEC) THEN
C Initialize TRANS to the permuted identity matrix.
C First set the whole array to zero:
        DO 310 J=1,N
          DO 320 I=1,N
            TRANS(1,I,J)=0.D0
320       CONTINUE
310     CONTINUE
C Then set the permuted diagonal elements to one:
C
C No two elements of IWS are the same.
C Alliant directive stating the elements of IWS are permuted:
cvd$  permutation(IWS)
        DO 325 I=1,N
          TRANS(1,IWS(I),I)=1.D0
325     CONTINUE
C Put in the remaining non-zero elements, suitably permuted.
C Alliant directive stating the elements of IWS are permuted:
cvd$  permutation(IWS)
        DO 330 J=JSTART+1,IHI-1
          DO 340 I=J+1,IHI
            TRANS(1,IWS(I),J)=A(I,J-1)
340       CONTINUE
330     CONTINUE
      ENDIF
C
C Put zeroes in the sub-Hessenberg elements of the output matrix.
      DO 260 J=JSTART,IHI-2
        DO 270 I=J+2,IHI
          A(I,J)=0.D0
270     CONTINUE
260   CONTINUE
      RETURN
      END
      SUBROUTINE ENORMF(A,EIGVAL,WS,NR,N)
C
C This subroutine normalizes the right eigenvectors stored
C in compressed form in the columns of array A.  The phase (sign)
C convention chosen is such that the real part of the first
C element of each vector is non-negative.
C
C CALLING FORMAT:
C
C      CALL ENORMF(A,EIGVAL,WS,NR,N)
C
C PARAMETER LIST:
C
C      A  (Input)  Array of row dimension NR and column dimension at
C                  least N, containing the N right eigenvectors.
C                  The eigenvectors are stored in compressed format
C                  in the columns of A.
C         (Output) The normalized eigenvectors, also compressed.
C EIGVAL  (Output) REAL*8 Array of row dimension 2 and column 
C                  dimension at least N, which contains the eigenvalues
C                  computed by this subroutine.  Element EIGVAL(1,I)
C                  contains the real part of the I'th eigenvalue,
C                  while element EIGVAL(2,I) contains the imaginary
C                  part of the I'th eigenvalue.
C     WS           REAL*8 Array of dimension at least N, used
C                  to store the eigenvector normalizations.
C     NR  (Input)  Row dimension of A exactly as specified in the
C                  calling program.  NR must be greater than or
C                  equal to N.
C      N  (Input)  Number of eigenvectors to be normalized.
C
C Required routines: DNRM2, DSCAL
C
C Revision history:
C
C Kevin Thompson     October 12, 1987     Version 3.0
C
C Kevin Thompson     October 10, 1990     Version 3.0
C   Replaced loops which compute the sum-of-squares of a 
C   vector's elements with calls to DNRM2.  Replaced
C   scaling loops with calls to DSCAL.
C
C Kevin Thompson     March 16, 1991       Version 3.0
C   Optimized calculations for concurrency, by treating all
C   column vectors identically.  An initial pass computes the
C   norm of each column.  A second scalar loop of O(N) operations
C   then computes the eigenvector normalizations from the column
C   norms.  The third loop is again concurrent and renormalizes
C   all columns appropriately.  This revised routine requires new
C   scratch space in the WS array to store vector norms, and
C   modifies the calling format.
C
C**********************************************************************
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION A(NR,N),EIGVAL(2,N),WS(N)
C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
      REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
      INTEGER IWARN,IFATAL,ISYSTM
      CHARACTER MES*60,EFNAME*32
      DATA RMIN/2.23D-308/,RMAX/1.79D308/
      DATA RMNHAF/1.50D-154/,RMXHAF/1.33D154/
      DATA PRECSN/2.22D-16/,PRCHAF/1.5D-8/
      DATA IWARN/1/,IFATAL/2/,ISYSTM/3/
      DATA MES/
     1 'MAGUS Numerical Analyst 3.0.  Copyright (C) 1990 by MAGUS.  '/
C *** End of INCLUDE file.
C
C Compute the magnitude (2-norm) of each column of A.
C
C Alliant directive allowing concurrent calls to DNRM2:
cvd$  cncall
      DO 10 J=1,N
        WS(J)=DNRM2(N,A(1,J),1)
10    CONTINUE
C Compute the normalizing factors for the eigenvectors.  We
C normalize the eigenvectors to unit magnitude.  The phase (sign)
C convention chosen is such that the real part of the first
C element of each vector is non-negative.
C
      J=1
100   CONTINUE
        IF(EIGVAL(2,J) .EQ. 0.D0) THEN
C Real vector.
          VMAG=DSIGN(WS(J),A(1,J))
C Normalizing factor:
          WS(J)=1.D0/VMAG
C Increment the index.
          J=J+1
        ELSE
C Complex vector.
          VMAG=DSIGN(DSQRT(WS(J)*WS(J)+WS(J+1)*WS(J+1)),A(1,J))
C Normalizing factor (same for real and imaginary parts):
          WS(J)=1.D0/VMAG
          WS(J+1)=WS(J)
C Increment the index.
          J=J+2
        ENDIF
C Do the next one.
      IF(J .LE. N) GO TO 100 
C
C Now normalize all the columns.
C
C Alliant directive allowing concurrent calls to DSCAL:
cvd$  cncall
      DO 20 J=1,N
        CALL DSCAL(N,WS(J),A(1,J),1)
20    CONTINUE
      RETURN
      END
      SUBROUTINE EQRF(A,TRANS,EIGVAL,NP,NR,N,ILOW,IHI,COMVEC)
C
C This subroutine computes the eigenvalues of a full, general matrix,
C using the QR algorithm of Francis, as described by Martin, Peters,
C and Wilkinson in the Handbook for Automatic Computation-Linear
C Algebra, volume II, articles 14 and 15 (HACLA II/14 and HACLA II/15).
C
C The input matrix A is in upper Hessenberg form.  Successive double QR
C transformations are performed, using Householder transformations and
C origin shifts.  The "active matrix" (matrix of as yet undetermined
C eigenvalues) is deflated (reduced in size) as eigenvalues are
C found in the lower right corner.  The eigenvalues occur either as the
C last element on the diagonal of the active matrix, or as the two
C real or complex conjugate eigenvalues of the last 2 X 2 sub-matrix
C along the diagonal.
C
C If eigenvectors are to be computed later, then the cumulative
C transformation matrix is stored in TRANS.
C
C CALLING FORMAT:
C
C      CALL EQRF(A,TRANS,EIGVAL,NP,NR,N,ILOW,IHI,COMVEC)
C
C PARAMETER LIST:
C
C      A  (Input)  Array of row dimension NR and column dimension at
C                  least N, containing the N by N input matrix whose
C                  eigenvalues (and, optionally, eigenvectors) are to
C                  be computed.
C         (Output) The N X N nearly triangular matrix T.
C  TRANS  (Input/  Array of dimensions (NP,NR,N), containing the
C          Output) cumulative transformation matrix elements as
C                  TRANS(1,I,J), if needed.
C EIGVAL  (Output) REAL*8 Array of row dimension 2 and column 
C                  dimension at least N, which contains the eigenvalues
C                  computed by this subroutine.  Element EIGVAL(1,I)
C                  contains the real part of the I'th eigenvalue,
C                  while element EIGVAL(2,I) contains the imaginary
C                  part of the I'th eigenvalue.
C     NP  (Input)  NP is the first dimension of the transformation
C                  matrix array, TRANS.
C     NR  (Input)  Row dimension of A and second dimension of TRANS
C                  exactly as specified in the calling program.  NR
C                  must be greater than or equal to N.
C      N  (Input)  Number of rows and columns in the input matrix.
C   ILOW  (Input)  Starting index for the active central sub-matrix.
C    IHI  (Input)  Ending index for the active central sub-matrix.
C COMVEC  (Input)  Transformation matrix calculation parameter.  The
C                  value of COMVEC determines whether or not the
C                  transformation matrix is to be computed.
C                  If COMVEC = .FALSE., the transformation matrix is
C                  not computed.  If COMVEC = .TRUE., then it is
C                  computed.
C
C Required routines: DROT, EHOUSF, UERCTL, UERDMP, UERPRT, UERMSG
C
C Revision history:
C
C Kevin Thompson     November 26, 1987     Version 2.1
C
C Kevin Thompson     March 16, 1991        Version 3.0
C   Replaced Fortran-coded Givens rotations with BLAS calls.
C
C**********************************************************************
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION A(NR,N),TRANS(NP,NR,N),EIGVAL(2,N)
      DIMENSION W(3)
      LOGICAL COMVEC,TEST
C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
      REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
      INTEGER IWARN,IFATAL,ISYSTM
      CHARACTER MES*60,EFNAME*32
      DATA RMIN/2.23D-308/,RMAX/1.79D308/
      DATA RMNHAF/1.50D-154/,RMXHAF/1.33D154/
      DATA PRECSN/2.22D-16/,PRCHAF/1.5D-8/
      DATA IWARN/1/,IFATAL/2/,ISYSTM/3/
      DATA MES/
     1 'MAGUS Numerical Analyst 3.0.  Copyright (C) 1990 by MAGUS.  '/
C *** End of INCLUDE file.
C
C Perform the sequence of QR decompositions, via Householder
C transformations.
C
C Store the eigenvalues for isolated values.
      DO 10 I=1,ILOW-1
        EIGVAL(1,I)=A(I,I)
        EIGVAL(2,I)=0.D0
10    CONTINUE
      DO 20 I=IHI+1,N
        EIGVAL(1,I)=A(I,I)
        EIGVAL(2,I)=0.D0
20    CONTINUE
C Initialize the remaining eigenvalues to zero.
      DO 15 I=ILOW,IHI
        EIGVAL(1,I)=0.D0
        EIGVAL(2,I)=0.D0
15    CONTINUE
C
C Initialize the sum of diagonal matrix elements subtracted from
C the matrix A during exceptional shifts.
      DIASUM=0.D0
C Define the top of the master loop, and initialize the number of
C eigenvalues remaining to be determined (NLEFT).
      NLEFT=IHI
100   CONTINUE
C
C Test for a completed calculation.
      IF(NLEFT .NE. ILOW-1) THEN
C Perform up to 100 iterations to find each eigenvalue.
        DO 200 ITS=1,100
C Copy frequently used array elements to local variables for quicker
C access.
          ANN=A(NLEFT,NLEFT)
          IF(NLEFT .GT. ILOW) THEN
            ANMNM=A(NLEFT-1,NLEFT-1)
            ANMN=A(NLEFT-1,NLEFT)
            ANNM=A(NLEFT,NLEFT-1)
            PROD=ANMN*ANNM
          ENDIF
C Look for a single small sub-diagonal element.
C
          DO 201 L=NLEFT,ILOW+1,-1
            TEST1=DABS(A(L,L-1))
            TEST2=DABS(A(L-1,L-1))+DABS(A(L,L))
            IF((TEST1+TEST2) .EQ. TEST2) GO TO 300
201       CONTINUE
          L=ILOW
C
300       CONTINUE
C
C Check for 1 or 2 eigenvalues in the last 2 X 2 sub-matrix.
C
          IF(L .EQ. NLEFT) THEN
C One eigenvalue (real).
C Add the diagonal sum matrix back in.
            ANN=ANN+DIASUM
            A(NLEFT,NLEFT)=ANN
C Store the eigenvalue.
            EIGVAL(1,NLEFT)=ANN
            EIGVAL(2,NLEFT)=0.D0
C Adjust the counter.
            NLEFT=NLEFT-1
            GO TO 100
          ELSEIF(L .EQ. NLEFT-1) THEN
C Two eigenvalues (real or complex).
            DIFF=0.5D0*(ANMNM-ANN)
C Add the diagonal sum matrix back in.
            ANN=ANN+DIASUM
            ANMNM=ANMNM+DIASUM
            A(NLEFT,NLEFT)=ANN
            A(NLEFT-1,NLEFT-1)=ANMNM
C
C Compute the trace of the last 2 X 2 sub-matrix.
            TRACE=ANMNM+ANN
C Use explicit formulas for the eigenvalues of the 2 X 2 sub-matrix.
            DSCRMT=DIFF**2+PROD
            ROOT=DSQRT(DABS(DSCRMT))
C Check for real or complex case.
            IF(DSCRMT .LT. 0.D0) THEN
C Complex case:
              EIGVAL(1,NLEFT)=0.5D0*TRACE
              EIGVAL(2,NLEFT)=ROOT
              EIGVAL(1,NLEFT-1)=0.5D0*TRACE
              EIGVAL(2,NLEFT-1)=-ROOT
            ELSE
C Real case:
              IF(DIFF .GT. 0.D0) THEN
                SUM=DIFF+ROOT
              ELSE
                SUM=DIFF-ROOT
              ENDIF
              EIGVAL(1,NLEFT-1)=ANN+SUM
              EIGVAL(2,NLEFT-1)=0.D0
              EIGVAL(1,NLEFT)=ANN-PROD/SUM
              EIGVAL(2,NLEFT)=0.D0
C
C Now perform a similarity transformation to triangularize the rows and
C columns which define the real eigenvalues just computed.  A Givens
C plane rotation is used.  This transformation is performed only if
C the eigenvectors are to be computed later.
C
              IF(COMVEC) THEN
                DENOM=DSQRT(ANNM**2+SUM**2)
                SINX=ANNM/DENOM
                COSX=SUM/DENOM
C Row modification:
                CALL DROT(N-NLEFT+2,A(NLEFT-1,NLEFT-1),NR,
     1           A(NLEFT,NLEFT-1),NR,COSX,SINX)
C Column modification:
                CALL DROT(NLEFT,A(1,NLEFT-1),1,A(1,NLEFT),1,COSX,SINX)
C
C Accumulate this transformation with the others.
                CALL DROT(IHI-ILOW+1,TRANS(1,ILOW,NLEFT-1),NP,
     1           TRANS(1,ILOW,NLEFT),NP,COSX,SINX)
              ENDIF
C
            ENDIF
C Adjust the counter.
            NLEFT=NLEFT-2
            GO TO 100
          ELSE
C No eigenvalues were found, so continue.
C
C Every 10 iterations, if the calculation has not converged, two
C steps are taken to enhance convergence.  First, the diagonal
C matrix of elements A(NLEFT,NLEFT) is subtracted from the matrix
C A.  The accumulated total of such subtractions is stored as
C DIASUM, and is added back in after convergence is achieved and
C the eigenvalues are stored.  Second, different shifts of origin
C are introduced into the product matrix calculation to alter what
C might be an invariant matrix under the standard transformations.
C
            TEST=MOD(ITS,10) .NE. 0
            IF(TEST) THEN
C This calculation uses shifts of origin whose sum and product are
C given by the trace and determinant of the last 2 X 2 sub-matrix.
              ANN2=ANN
              ANMNM2=ANMNM
              PROD2=PROD
            ELSE
C Subtract the diagonal matrix of elements A(NLEFT,NLEFT).
              DIASUM=DIASUM+ANN
              DO 730 I=ILOW,NLEFT
                A(I,I)=A(I,I)-ANN
730           CONTINUE
              ANMNM=ANMNM-ANN
              ANN=0.D0
C Compute the equivalent matrix elements of the exceptional shifts.
              ABSUM=DABS(ANNM)+DABS(A(NLEFT-1,NLEFT-2))
              ANN2=0.75D0*ABSUM
              ANMNM2=ANN2
              PROD2=-0.4375D0*ABSUM**2
            ENDIF
C
C Look for two consecutive small elements.
C
            DO 301 M=NLEFT-2,L,-1
C Compute the product matrix elements for the matrix starting in 
C row and column M, omitting a common factor of A(M+1,M).
C
              AMM=A(M,M)
              DIFF1=ANN2-AMM
              DIFF2=ANMNM2-AMM
C
              EMMM=(DIFF1*DIFF2-PROD2)/A(M+1,M)+A(M,M+1)
              EMM1M=(A(M+1,M+1)-AMM)-DIFF1-DIFF2
              EMM2M=A(M+2,M+1)
C Normalize the sub-diagonal elements and store them in W.
              WSUM=DABS(EMMM)+DABS(EMM1M)+DABS(EMM2M)
              IF(WSUM .NE. 0.D0) THEN
                W(1)=EMMM/WSUM
                W(2)=EMM1M/WSUM
                W(3)=EMM2M/WSUM
              ELSE
                W(1)=0.D0
                W(2)=0.D0
                W(3)=0.D0
              ENDIF
              IF(M .EQ. L) THEN
                GO TO 400
              ELSE
C Perform the test for small elements.
                TEST1=DABS(A(M,M-1))*(DABS(W(2))+DABS(W(3)))
                TEST2=DABS(W(1))*(DABS(A(M-1,M-1))+DABS(AMM)
     1                        +DABS(A(M+1,M+1)))
                IF((TEST1+TEST2) .EQ. TEST2) GO TO 400
              ENDIF
301         CONTINUE
400         CONTINUE
C
C Now perform the double QR step on the matrix in rows and columns 
C L to NLEFT.  The first transformation is based on the product
C matrix starting at location (M,M), and creates nonzero (but
C negligible) sub-Hessenberg elements in column M - 1.  The
C sub-Hessenberg elements in column M are then eliminated by
C subsequent transformations.
C
            CALL EHOUSF(A,TRANS,NP,NR,N,NLEFT,ILOW,IHI,L,M,W,WSUM,
     1                  COMVEC)
          ENDIF
200     CONTINUE
C
C The sequence has failed to converge to the eigenvalues.  Give a
C warning message.
        LEVERR=IWARN
        MES=
     1  'Convergence failure: routine failed to give an answer       '
        CALL UERCTL('EFULRC',5411,MES,LEVERR)
        IF(LEVERR .GT. IWARN) RETURN
      ENDIF
C The eigenvalue calculation is complete.
C
      RETURN
      END
      SUBROUTINE EVECRF(A,EIGVAL,WS,NR,N)
C
C This subroutine computes the right eigenvectors of a triangular
C or nearly triangular matrix, stored in full storage format.
C In the latter case the eigenvectors are complex, and the matrix
C departs from upper triangular form in that there are 2 X 2
C sub-matrices along the main diagonal, which always correspond
C to complex eigenvalues.
C
C The eigenvectors are stored in array A, overwriting the input
C matrix.  If the j'th eigenvalue is real, then column j contains
C the j'th eigenvector.  If eigenvalues j and j + 1 are complex
C conjugate values, then eigenvector j + 1 is the complex conjugate
C of eigenvector j.  In this case the real part of eigenvector j
C is stored in column j of the array, while its imaginary part is
C stored in column j + 1.  Eigenvector j + 1 may then be computed
C from this information.
C
C CALLING FORMAT:
C
C      CALL EVECRF(A,EIGVAL,WS,NR,N)
C
C PARAMETER LIST:
C
C      A  (Input)  Array of row dimension NR and column dimension at
C                  least N, containing the N by N nearly upper 
C                  triangular input matrix whose eigenvectors are to 
C                  be computed.
C
C         (Output) The right eigenvectors of the input matrix. If the
C                  j'th eigenvalue is real, then column j contains the
C                  j'th eigenvector.  If eigenvalues j and j + 1 are
C                  complex conjugate values, then eigenvector j + 1 is
C                  the complex conjugate of eigenvector j.  In this
C                  case the real part of eigenvector j is stored in
C                  column j of the array, while its imaginary part is
C                  stored in column j + 1.  Eigenvector j + 1 may then
C                  be computed from this information.
C
C EIGVAL  (Input)  REAL*8 Array of row dimension 2 and column 
C                  dimension at least N, which contains the eigenvalues
C                  input to this subroutine.  Element EIGVAL(1,I)
C                  contains the real part of the I'th eigenvalue, while
C                  element EIGVAL(2,I) contains the imaginary part of
C                  the I'th eigenvalue.
C
C     WS           Vector of dimension at least N, used for scratch
C                  space.
C
C     NR  (Input)  Row dimension of A exactly as specified
C                  in the calling program.  NR must be greater than or
C                  equal to N.
C
C      N  (Input)  Number of rows and columns in the input matrix,
C                  and of eigenvectors to be obtained.
C
C Required routines: CDIV
C
C Revision history:
C
C Kevin Thompson     September 3, 1987     Version 3.0
C
C Kevin Thompson     June 15, 1991         Version 3.0
C   Rewrote numerous loops to call BLAS-1 routines.  Also changed
C   calling format to put WS scratch array in parameter list, for
C   use in vector loops.
C
C**********************************************************************
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION A(NR,N),EIGVAL(2,N),WS(N)
C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
      REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
      INTEGER IWARN,IFATAL,ISYSTM
      CHARACTER MES*60,EFNAME*32
      DATA RMIN/2.23D-308/,RMAX/1.79D308/
      DATA RMNHAF/1.50D-154/,RMXHAF/1.33D154/
      DATA PRECSN/2.22D-16/,PRCHAF/1.5D-8/
      DATA IWARN/1/,IFATAL/2/,ISYSTM/3/
      DATA MES/
     1 'MAGUS Numerical Analyst 3.0.  Copyright (C) 1990 by MAGUS.  '/
C *** End of INCLUDE file.
C
C Compute the norm of the matrix.
C
C First compute the norm of each column.
C
C Alliant directive allowing concurrent calls to DASUM:
cvd$  cncall
      DO 10 J=1,N
        NITEMS=MIN0(J+1,N)
        WS(J)=DASUM(NITEMS,A(1,J),1)
10    CONTINUE
C
C Sum the column norms.
      RNORM=DASUM(N,WS,1)
      EPSNRM=PRECSN*RNORM
C
C Loop over all eigenvalues.  Start with the last one, and work
C toward the first.
C
      J=N
2000  CONTINUE
C Check for a real or complex eigenvalue.
        IF(EIGVAL(2,J) .EQ. 0.D0) THEN
C This eigenvalue is real.  Only real arithmetic is needed.
C
C Set real elements J+1 to N to zero.
          DO 2010 I=J+1,N
            A(I,J)=0.D0
2010      CONTINUE
C
C Set element J to one.
          A(J,J)=1.D0
C
C Perform the back substitution to get the remaining elements.
C
          I=J-1
2020      CONTINUE
            IF(I .GT. 0) THEN
C Check for the presence of a 2 X 2 sub-matrix.
              IF(EIGVAL(2,I) .EQ. 0.D0) THEN
C No 2 X 2 sub-matrix.  Use the triangular solution for element I.
C
C Compute the sum.
                SIGMA=DDOT(J-I,A(I,I+1),NR,A(I+1,J),1)
C
C Test for repeated eigenvalues and take appropriate action.
C
                DIFF=EIGVAL(1,J)-EIGVAL(1,I)
                IF(DIFF .NE. 0.D0) THEN
C Standard eigenvector calculation: nonzero denominator.
                  A(I,J)=SIGMA/DIFF
                ELSEIF(SIGMA .NE. 0.D0) THEN
C Numerically defective eigenvector: nonzero numerator, zero
C denominator.  Replace the denominator by PRECSN * RNORM.
                  A(I,J)=SIGMA/EPSNRM
                ELSE
C Numerically degenerate: zero numerator and denominator.
                  A(I,J)=1.D0
                ENDIF
              ELSE
                I=I-1
C There is a 2 X 2 sub-matrix here, which determines elements I and
C I + 1.
C
C Compute the two sums.
                SIGI=DDOT(J-I-1,A(I,I+2),NR,A(I+2,J),1)
                SIGIP1=DDOT(J-I-1,A(I+1,I+2),NR,A(I+2,J),1)
C Get the determinant of the coefficient matrix, and solve for the
C unknowns.
C
                DIFFI=EIGVAL(1,J)-A(I,I)
                DIFFIP=EIGVAL(1,J)-A(I+1,I+1)
                DET=DIFFI*DIFFIP-A(I,I+1)*A(I+1,I)
C
C No test for repeated eigenvalues is performed, since the J'th 
C eigenvalue is real and the 2 X 2 sub-matrix has complex eigenvalues.
C
C Solve the system.
C
                A(I,J)=(DIFFIP*SIGI+A(I,I+1)*SIGIP1)/DET
                A(I+1,J)=(DIFFI*SIGIP1+A(I+1,I)*SIGI)/DET
C
              ENDIF
C
C If the eigenvalue is defective and has a large multiplicity, repeated
C divisions by EPSNRM could lead to numerical overflow.  Even if it is
C not defective, the leading element could become large due to the
C structure of the matrix.  Check the leading element of the
C eigenvector and rescale all computed elements if it is too large.
              IF(DABS(A(I,J)) .GT. RMXHAF) THEN
                SCALE=1.D0/RMXHAF
                CALL DSCAL(J-I+1,SCALE,A(I,J),1)
              ENDIF
C Decrement the element index and begin the next element calculation.
              I=I-1
            ENDIF
          IF(I .GT. 0)GO TO 2020
C
C
        ELSE
          J=J-1
C
C This eigenvalue is complex.  The resulting eigenvector will be
C complex as well.  First compute vector J explicitly, then set
C vector J + 1 to be the complex conjugate of vector J.
C
C Set elements J + 2 through N to zero.
          DO 3010 I=J+2,N
            A(I,J)=0.D0
            A(I,J+1)=0.D0
3010      CONTINUE
C Store local elements so they don't get overwritten before use.
          AJJ=A(J,J)
          AJPJ=A(J+1,J)
          AJJP=A(J,J+1)
          AJPJP=A(J+1,J+1)
C Set element J + 1 to the real number 1 (1 = 1 + 0i).
          A(J+1,J)=1.D0
          A(J+1,J+1)=0.D0
C Compute element J.  Use the calculation with the largest matrix
C element in the denominator.
          IF(DABS(AJPJ) .GT. DABS(AJJP)) THEN
            A(J,J)=(EIGVAL(1,J)-AJPJP)/AJPJ
            A(J,J+1)=EIGVAL(2,J)/AJPJ
          ELSE
            CALL CDIV(AJJP,0.D0,EIGVAL(1,J)-AJJ,EIGVAL(2,J),
     1                A(J,J),A(J,J+1))
          ENDIF
C Perform the back substitution to get the remaining elements.
          I=J-1
3020      CONTINUE
            IF(I .GT. 0) THEN
C Check for a 2 X 2 sub-matrix.
              IF(EIGVAL(2,I) .EQ. 0.D0) THEN
C No 2 X 2 sub-matrix.  Use the triangular solution for element I.
C
C Compute the sum.
                SIGMAR=DDOT(J-I+1,A(I,I+1),NR,A(I+1,J),1)
                SIGMAI=DDOT(J-I+1,A(I,I+1),NR,A(I+1,J+1),1)
C No test for repeated eigenvalues is performed, since the J'th
C eigenvalue is complex and this triangular equation corresponds to a
C real eigenvalue.
                CALL CDIV(SIGMAR,SIGMAI,EIGVAL(1,J)-A(I,I),EIGVAL(2,J),
     1                    A(I,J),A(I,J+1))
              ELSE
                I=I-1
C There is a 2 X 2 sub-matrix here, which determines elements I
C and I + 1.
C
C Compute the two sums.
                SIGIR=DDOT(J-I,A(I,I+2),NR,A(I+2,J),1)
                SIGII=DDOT(J-I,A(I,I+2),NR,A(I+2,J+1),1)
                SIGIPR=DDOT(J-I,A(I+1,I+2),NR,A(I+2,J),1)
                SIGIPI=DDOT(J-I,A(I+1,I+2),NR,A(I+2,J+1),1)
C Compute the determinant of the coefficient matrix, and its 1-norm.
C The 1-norm of a complex number provides a simple way to test for
C whether or not the number is zero.
                DIFFER=EIGVAL(1,I)-EIGVAL(1,J)
                DIFFEI=EIGVAL(2,I)-EIGVAL(2,J)
                SUMEI=EIGVAL(2,I)+EIGVAL(2,J)
C
                DETR=DIFFER**2+DIFFEI*SUMEI
                DETI=-2.D0*DIFFER*EIGVAL(2,J)
                DNORM=DABS(DETR)+DABS(DETI)
C Now calculate the numerators of the formulas, and their 1-norms.
                DIFFI=EIGVAL(1,J)-A(I,I)
                DIFFIP=EIGVAL(1,J)-A(I+1,I+1)
                TOPIR=DIFFIP*SIGIR-EIGVAL(2,J)*SIGII+A(I,I+1)*SIGIPR
                TOPII=DIFFIP*SIGII+EIGVAL(2,J)*SIGIR+A(I,I+1)*SIGIPI
                TINRM=DABS(TOPIR)+DABS(TOPII)
                TOPIPR=DIFFI*SIGIPR-EIGVAL(2,J)*SIGIPI+A(I+1,I)*SIGIR
                TOPIPI=DIFFI*SIGIPI+EIGVAL(2,J)*SIGIPR+A(I+1,I)*SIGII
                TIPNRM=DABS(TOPIPR)+DABS(TOPIPI)
C Test for repeated eigenvalues and take appropriate action.
C
                IF(DNORM .NE. 0.D0) THEN
C Standard eigenvector calculation: nonzero denominator.
                  CALL CDIV(TOPIR,TOPII,DETR,DETI,
     1                      A(I,J),A(I,J+1))
                  CALL CDIV(TOPIPR,TOPIPI,DETR,DETI,
     1                      A(I+1,J),A(I+1,J+1))
                ELSE
C The denominator is zero, so take special action.
C Determine whether this case is degenerate or defective.
                  IF(TINRM .EQ. 0.D0 .AND. TIPNRM .EQ. 0.D0) THEN
C Numerically degenerate: zero numerator and denominator.
C Set element I + 1 to the real number 1 (1 = 1 + 0i).
                    A(I+1,J)=1.D0
                    A(I+1,J+1)=0.D0
C Compute element I.  Use the calculation with the largest matrix
C element in the denominator.
                    IF(DABS(A(I+1,I)) .GT. DABS(A(I,I+1))) THEN
                      A(I,J)=(EIGVAL(1,J)-A(I+1,I+1))/A(I+1,I)
                      A(I,J+1)=EIGVAL(2,J)/A(I+1,I)
                    ELSE
                      CALL CDIV(A(I,I+1),0.D0,EIGVAL(1,J)-A(I,I),
     1                  EIGVAL(2,J),A(I,J),A(I,J+1))
                    ENDIF
                  ELSE
C Numerically defective eigenvector: nonzero numerator, zero
C denominator.  Replace the denominator by PRECSN * RNORM.
                    CALL CDIV(TOPIR,TOPII,EPSNRM,0.D0,
     1                        A(I,J),A(I,J+1))
                    CALL CDIV(TOPIPR,TOPIPI,EPSNRM,0.D0,
     1                        A(I+1,J),A(I+1,J+1))
                  ENDIF
                ENDIF
              ENDIF
C
C If the eigenvalue is defective and has a large multiplicity, repeated
C divisions by EPSNRM could lead to numerical overflow.  Even if it is
C not defective, the leading element could become large due to the
C structure of the matrix.  Check the leading element of the
C eigenvector and rescale all computed elements if it is too large.
C
              ONENRM=DABS(A(I,J))+DABS(A(I,J+1))
              IF(ONENRM .GT. RMXHAF) THEN
                SCALE=1.D0/RMXHAF
                CALL DSCAL(J-I+2,SCALE,A(I,J),1)
                CALL DSCAL(J-I+2,SCALE,A(I,J+1),1)
              ENDIF
C Decrement the element index and do the next element(s).
              I=I-1
            ENDIF
          IF(I .GT. 0) GO TO 3020
C
        ENDIF
C
C Decrement the vector index and do the next vector.
        J=J-1
C
      IF(J .GT. 0) GO TO 2000
C All done.
      RETURN
      END
      SUBROUTINE UERCTL(NAME,NUMERR,ERRMES,LEVERR)
C
C     UERCTL, the error control routine, is called by a Numerical
C     Analyst library routine when the routine recognizes an error
C     condition.  UERCTL reports the error condition as indicated
C     in the NASYS1 common block, and then either aborts or returns
C     control as indicated by error severity and parameters set in
C     the NASYS1 common block.
C
C     A subroutine issues a reset command to UERCTL by calling it
C     with a NUMERR of 0.  The reset command causes UERCTL to
C     report counts for all warning errors which occurred during
C     the execution of the routine and to reset the warning
C     error counters to 0.
C
C
C     CALLING SEQUENCE
C
C       CALL UERCTL(NAME,NUMERR,ERRMES,LEVERR)
C
C     PARAMETER LIST
C
C        NAME (INPUT)    Name of the routine in which the error
C                        occurred.  (CHARACTER*6)
C
C      NUMERR (INPUT)    Number which uniquely identifies the error
C                        condition detected.
C
C                        Note:  NUMERR = 0 indicates a reset command.
C
C      ERRMES (INPUT)    Error Message.
C
C      LEVERR (INPUT)    Error severity level.
C             (OUTPUT)   1 = warning error
C                        2 = fatal error
C                        3 = system error
C
C     PROGRAMMING NOTES
C
C     The parameter NWARN is the total number of unique warning
C     errors that a subroutine can encounter before overflowing
C     the arrays which keep track of the count of warning error
C     messages.
C
C     -  Current value is 10.
C
C     The parameter NRECR is the total number of error checking
C     recursion levels permitted within the library.  An error
C     checking recursion level is a call from one error trapping
C     library routine to another.
C
C     -  Current value is 5.
C
C     UERCTL will be linked in to every program written using
C     Numerical Analyst.  To save RAM, some integer arrays have
C     been declared INTEGER*2.
C
C     -  This is possible because of the numbering scheme used for
C        errors.  The first two digits of the five digit error
C        error number represent a letter of the alphabet.  Therefore
C        the error number will never exceed 26999 - a number which
C        will fit in an INTEGER*2 variable.
C
C     If the Fatal Error Abort indicator (NAFEA) is set to no,
C     the calling routine will receive control back from UERCTL
C     even after a fatal error.  Therefore, the calling routine
C     must recognize that continued processing would be meaningless
C     and immediately return control to its calling program or
C     subroutine.
C
C     -  It is possible for UERCTL to "upgrade" a warning error
C        to a fatal or system error.  For example, if the warning
C        error causes any of the recursion tables to overflow,
C        UERCTL will disregard the warning error and create a
C        fatal one.  After return from UERCTL, the calling program
C        must call UERCHK to check for error severity level.  The
C        calling program can not count on the error severity level
C        being the same that it passed to UERCTL.
C
C     -  A user who sets NAFEA to no must accept responsiblity to
C        call UERCHK to check for fatal errors after any call to a
C        library routine.
C
C     -  Any library routine that calls another library routine which
C        calls UERCTL must accept responsibility to call UERCHK to
C        check for fatal errors upon return from the called library
C        routine.
C
C     REQUIRED ROUTINES
C
C       UERPRT, UERDMP, NACOM
C
C     REVISION HISTORY
C
C       Richard Lahey       10/31/87      Initial release
C       Richard Lahey       10/26/88      first revision
C       Richard Lahey       02/25/89      version 3.0
C       Richard Lahey       11/19/89      minor bug fix
C
C********************************************************************

      IMPLICIT REAL*8 (A-H,O-Z)

C     NWARN:  The total number of warning errors in the library.
C
C     NRECR:  The total number of recursion levels permitted.
C
C     NREWAR: The number of recursion levels multiplied by the
C             total number of warning errors.
C
C
      PARAMETER (NWARN = 10, NRECR = 5, NREWAR = NWARN*NRECR)

C     INO:  Code for no.
      PARAMETER (INO = 0)

C     IYES:  Code for yes.
      PARAMETER (IYES = 1)

      CHARACTER*6 NAME
      CHARACTER*60 ERRMES

C     The NASYS common blocks
      CHARACTER*6 NAFER,NAWER
      CHARACTER*32 FNAME
      CHARACTER*60 NAFMES, NAWMES
      COMMON /NASYS1/ NAESI, NAFEN, NAWEC, NAWEN, NAFEA,
     1                NAWEG, NAEMS, NAEMF, NASUN, NAFUN
      COMMON /NASYS2/ NAFER, NAFMES, NAWER, NAWMES, FNAME

C     RTBL: The recursion table.
      CHARACTER*6 RTBL(NRECR)

C     IERRN:  An array which contains all warning error numbers
C     encountered for a recursion level.
      INTEGER*2 IERRN(NRECR,NWARN)

C     IERRC:  An array which contains a count of all warning
C     error numbers encountered for a recursion level.
      INTEGER*2 IERRC(NRECR,NWARN)

C     NUQERR:  The number of unique errors for a recursion level.
      INTEGER*2 NUQERR(NRECR)

C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
      REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
      INTEGER IWARN,IFATAL,ISYSTM
      CHARACTER MES*60,EFNAME*32
      DATA RMIN/2.23D-308/,RMAX/1.79D308/
      DATA RMNHAF/1.50D-154/,RMXHAF/1.33D154/
      DATA PRECSN/2.22D-16/,PRCHAF/1.5D-8/
      DATA IWARN/1/,IFATAL/2/,ISYSTM/3/
      DATA MES/
     1 'MAGUS Numerical Analyst 3.0.  Copyright (C) 1990 by MAGUS.  '/
C *** End of INCLUDE file.

C     Initializations for variables which must retain their
C     values until they are explicitly cleared - even if the
C     clearing occurs after UERCTL has returned control and been
C     re-CALLed with a reset command:
C     - IERRN    Warning Error Numbers for a Recursion Level
C     - IERRC    Warning Error Number Count for a Recursion Level
C     - IRLEV    Recursion Level
C     - NUQERR   Number of Unique Errors for a Recursion Level
C     - RTBL     Recursion Table

      DATA IERRN/NREWAR*0/
      DATA IERRC/NREWAR*0/
      DATA IRLEV/0/
      DATA NUQERR/NRECR*0/
      DATA RTBL/NRECR*'      '/

100   CONTINUE
*********************************************************************
*                                                                   *
*                    Paragraph 100:  Mainline                       *
*                                                                   *
*********************************************************************

C     If NUMERR is zero, respond to a reset command.
      IF (NUMERR .EQ. 0) THEN

200     CONTINUE
C       *************************************************************
C       *                                                           *    
C       *            PARAGRAPH 200 - RESET COMMAND                  *
C       *                                                           *    
C       *   The calling subroutine is about to return control       *
C       *   to its own calling program or subroutine.  If any       *
C       *   warning error messages have occurred during the         *
C       *   execution of the subroutine issuing the reset command,  *
C       *   they are reported according to parameters in the        *
C       *   NASYS1 and NASYS2 common blocks and then cleared from   *
C       *   the warning error stack.                                *
C       *                                                           *
C       *************************************************************

C       If IRLEV = 0, no warning errors are in the recursion stack.
C       No processing is necessary.  Fall through to subroutine end.
        IF (IRLEV .EQ. 0) THEN
201       CONTINUE

C       If the last routine name in the recursion stack is not NAME,
C       this recursion of NAME encountered no warning errors.
C       No processing is necessary.  Fall through to subroutine end.
        ELSEIF (RTBL(IRLEV) .NE. NAME) THEN
202       CONTINUE

C       Else,
C       - report the warning error count for NAME
C       - reset all counters at the current recursion level
C       - pop the recursion level
C       - go to subroutine end
        ELSE

          CALL UERDMP(IRLEV,RTBL,IERRN,IERRC,NUQERR,NRECR)

          RTBL(IRLEV) = '      '
          NUQERR(IRLEV) = 0
          DO 219 I=1,NWARN
            IERRN(IRLEV,I) = 0
            IERRC(IRLEV,I) = 0
219       CONTINUE
          IRLEV = IRLEV - 1
        ENDIF

C     Otherwise, an error has occurred.
      ELSE

C     Error Case 1:  It is a warning error
      IF (LEVERR .EQ. IWARN) THEN

300     CONTINUE
C       *************************************************************
C       *                                                           *
C       *           Paragraph 300:  Warning Errors                  *
C       *                                                           *
C       *************************************************************

 
C       Assume (until proven otherwise) that this is a unique error,
C       that is, that it has not occurred before for this recursion
C       of NAME.
        IUNIQ = IYES

C       Is this the first error encountered by the first subroutine
C       in a chain?  If so,
C       1.  Set NASYS information
C           -  NAESI     Error Status Indicator.
C           -  NAWER     Warning Error Routine name. (CHARACTER*6)
C           -  NAWMES    Warning Error Message. (CHARACTER*60)
C           -  NAWEN     Warning Error Number.
C       2.  Start the Recursion Level and Recursion Table
C           -  IRLEV     Recursion Level
C           -  RTBL      Recursion Table
C
        IF (IRLEV .EQ. 0) THEN
          NAESI = LEVERR
          NAWER = NAME
          NAWMES = ERRMES
          NAWEN = NUMERR
          IRLEV = 1
          RTBL(IRLEV) = NAME

C       Else, if NAME is not the last entry in the recursion table,
C       increment the recursion level, and add NAME to the table.

        ELSEIF (RTBL(IRLEV) .NE. NAME) THEN

          IF (IRLEV .LT. NRECR) THEN
            IRLEV = IRLEV + 1
            RTBL(IRLEV) = NAME
     
C         Else too many recursion levels.
C         Throw out the warning error and create a fatal error for
C         too many recursion levels.  Fall through this paragraph,
C         and pick up again in Paragraph 400, where we test for
C         fatal errors.
          ELSE

C           Error 502 - Warning errors nested too deeply.
C           Chapter U, Utilities (21)
            NAME = 'UERCTL'
            NUMERR = 21502
            ERRMES =
     1 'Warning errors nested too deeply                            '
            LEVERR = ISYSTM
            GO TO 399
          ENDIF

C       Else, since NAME is already the last entry in the recursion
C       table, search IERRN, the warning error table, for the error
C       number.  If found, increment the count for that error number
C       and set IUNIQ to no.

        ELSE

          DO 319 I=1,NUQERR(IRLEV)
            IF (IERRN(IRLEV,I) .EQ. NUMERR) THEN
              IERRC(IRLEV,I) = IERRC(IRLEV,I)+1
              IUNIQ = INO       
            ENDIF
319       CONTINUE
        ENDIF

C       If the error is unique for this recursion of NAME,
C       and we haven't thrown it out because it exceeded
C       the number of recursion levels:
C       - increment NAWEC
C       - increment NUQERR
C       - put NUMERR in IERRN  (the array of error numbers
C         encountered)
C       - set the count for NUMERR in IERRC (the array of
C         counts of error numbers encountered) to 1
C       - report the error according to parameters set in NASYS

        IF (IUNIQ .EQ. IYES) THEN

C         Check to see if there are too many unique warning
C         errors.  If there are, throw out the latest warning
C         error and create a fatal error for too many unique
C         warning errors.  Then fall through this paragraph,
C         and pick up again in Paragraph 400, where we test for
C         fatal errors.

          IF (NUQERR(IRLEV) + 1 .GT. NWARN) THEN
C           Error 501 - Too many unique warning errors
C           Chapter U, Utilities (21)
            NAME = 'UERCTL'
            NUMERR = 21501
            ERRMES =
     1 'Too many unique warning errors encountered by one routine   '
            LEVERR = IFATAL
            GO TO 399

          ELSE
            NAWEC = NAWEC + 1
            NUQERR(IRLEV) = NUQERR(IRLEV)+1
            IERRN(IRLEV,NUQERR(IRLEV)) = NUMERR
            IERRC(IRLEV,NUQERR(IRLEV)) = 1

C           If warning error generation is on, print warning errors
            IF (NAWEG .EQ. IYES) CALL UERPRT(NAME,NUMERR,ERRMES,LEVERR)
          ENDIF
        ENDIF
      ENDIF

C     Pick up here if we had to throw out a warning error and
C     replace it with a fatal or system error.
399   CONTINUE
C     Error Case 2:  It is a fatal or system error
      IF (LEVERR .EQ. IFATAL .OR. LEVERR .EQ. ISYSTM) THEN

400     CONTINUE
C       *************************************************************
C       *                                                           *
C       *        Paragraph 400:  Fatal and System Errors            *
C       *                                                           *
C       *************************************************************

C       First, report the error.
        CALL UERPRT(NAME,NUMERR,ERRMES,LEVERR)

C       Next, report all warning errors for each level of recursion.
        DO 419, I=IRLEV,1,-1
          CALL UERDMP(I,RTBL,IERRN,IERRC,NUQERR,NRECR)
419     CONTINUE

C       If fatal-error-abort is on, stop everything.
        IF (NAFEA .EQ. IYES) THEN
          STOP

C       Else, set NASYS1 and NASYS2, clear all warning error
C       counters, and return control.
        ELSE

          NAESI = LEVERR
          NAFER = NAME
          NAFEN = NUMERR
          NAFMES = ERRMES
          DO 429 I=1,IRLEV
            RTBL(I) = '      '
            NUQERR(I) = 0
            DO 425 J=1,NWARN
              IERRN(I,J) = 0
              IERRC(I,J) = 0
425         CONTINUE
429       CONTINUE
          IRLEV = 0
        ENDIF
      ENDIF
      ENDIF

900   CONTINUE
*********************************************************************
*                                                                   *
*                   Paragraph 900:  Subroutine End                  *
*                                                                   *
*********************************************************************

      RETURN
      END

      BLOCK DATA NACOM
C
C     NACOM contains the default values for the Numerical Analyst
C     common blocks.
C
C     PROGRAMMING NOTES
C
C       1.  A pathname may be included in FNAME.  However, specifying 
C           a nonexistant pathname will cause an operating system
C           abort when the file is opened.
C
C       2.  This block data subroutine must be included in the same
C           source file as a callable subroutine.  Otherwise, the
C           common block is not correctly initialized.
C
C     REQUIRED ROUTINES
C
C       none
C
C     REVISION HISTORY
C
C       Richard Lahey       10/10/87      initial release
C
C       Richard Lahey        9/26/88      Ver 3.0
C
C       Kevin Thompson      12/28/89      NASUN, NAFUN values for i860
C
C********************************************************************

      IMPLICIT REAL*8 (A-H,O-Z)

C     The NASYS common blocks
      CHARACTER*6 NAFER,NAWER
      CHARACTER*32 FNAME
      CHARACTER*60 NAFMES, NAWMES
      COMMON /NASYS1/ NAESI, NAFEN, NAWEC, NAWEN, NAFEA,
     1                NAWEG, NAEMS, NAEMF, NASUN, NAFUN
      COMMON /NASYS2/ NAFER, NAFMES, NAWER, NAWMES, FNAME

C************************************************************
C
C       NAESI     Error Status indicator.
C                 0 = no error
C                 1 = only warning errors have occurred.
C                 2 = a fatal error has occurred.
C                 Initialized to zero.
C
        DATA NAESI/0/

C************************************************************
C
C       NAFEN     Fatal Error Number.
C                 Initialized to zero.
C
        DATA NAFEN/0/

C************************************************************
C
C       NAWEC     Warning Error Count.
C                 The number of different warning errors which
C                 have occurred.
C                 Initialized to zero.
C
        DATA NAWEC/0/

C************************************************************
C
C       NAWEN     Warning Error Number for the FIRST warning
C                 error which has occurred.
C                 Initialized to zero.
C
        DATA NAWEN/0/

C************************************************************
C
C       NAFEA     Fatal Error Abort indicator.
C                 0 = no 
C                 1 = yes (default)
C
        DATA NAFEA/1/

C************************************************************
C
C       NAWEG     Warning Error Generation indicator.
C                 0 = no 
C                 1 = yes (default)
C
        DATA NAWEG/1/

C************************************************************
C
C       NAEMS     Error Message to Screen indicator.
C                 0 = no 
C                 1 = yes (default)
C
        DATA NAEMS/1/

C************************************************************
C
C       NAEMF     Error Message to File indicator.
C                 0 = no  (default)
C                 1 = yes
C
        DATA NAEMF/0/

C************************************************************
C
C       NASUN     Screen Unit Number (STDOUT).
C
C  Allowed range:
C
C  -32767 to 32767 for Microsoft Fortran, IBM PC
C       0 to    99 for Green Hills Fortran, i860 machines
C
C  Default value:
C    0 or 6 for Microsoft Fortran, IBM PC
C         6 for Green Hills Fortran, i860 machines
C                 [integer]
C
        DATA NASUN/6/
C************************************************************
C
C       NAFUN     (Error) File Unit Number.
C
C  Allowed range:
C
C  -32767 to 32767 for Microsoft Fortran, IBM PC
C       0 to    99 for Green Hills Fortran, i860 machines
C
C  Default value:
C      None for Microsoft Fortran, IBM PC
C      None for Green Hills Fortran, i860 machines
C                 [INTEGER]
C
        DATA NAFUN/91/

C************************************************************
C
C       NAFER     Fatal Error Routine name.
C                 The name of the routine in which the fatal
C                 error occurred.
C                 Initialized to spaces.
C
        DATA NAFER/'      '/

C************************************************************
C
C       NAFMES    Fatal Error Message
C                 The error message associated with the fatal
C                 error.
C                 Initialized to spaces.
C
        DATA NAFMES/
     1 '                                                            '/

C************************************************************
C
C       NAWER     Warning Error Routine name.
C                 The name of the FIRST routine in which a
C                 warning error occurred.
C                 Initialized to spaces.
C
        DATA NAWER/'      '/

C************************************************************
C
C       NAWMES    Warning Error Message
C                 The error message associated with the FIRST
C                 warning error which occurred.
C                 Initialized to spaces.
C
        DATA NAWMES/
     1 '                                                            '/

C************************************************************
C
C       FNAME     (Error) File Name.
C                 'NUMAN.ERR                      ' (default)
C                 [CHARACTER*32 expression]

      DATA FNAME/'NUMAN.ERR                       '/

C     End of block data subroutine
      END
      SUBROUTINE UERGET(LEVERR,IFEN,IWEN)
C
C UERGET reads the NASYS1 and NASYS2 common blocks and returns
C error status information.  UERGET is intended for internal use;
C it returns the minimum amount of error information needed for
C library routines to react intelligently to error conditions.
C UERGET's parameters are a subset of UERCHK's parameters.  UERCHK
C is a user-callable error statusing routine.
C
C CALLING SEQUENCE
C
C       CALL UERGET(LEVERR,IFEN,IWEN)
C
C PARAMETER LIST
C
C  LEVERR (OUTPUT)     Level of Error
C                      0 = no error
C                      1 = warning error(s)
C                      2 = fatal error
C                      3 = system error
C
C    IFEN (OUTPUT)     Fatal Error Number.
C                      The number which uniquely identifies a fatal
C                      or system error.
C
C    IWEN (OUTPUT)     Warning Error Number.
C                      The number of the FIRST warning error
C                      encountered.
C
C       
C PROGRAMMING NOTES
C
C       None
C
C REQUIRED ROUTINES
C
C       NACOM
C
C REVISION HISTORY
C
C       Richard Lahey       02/25/89      initial release (ver 3.0)
C     
C********************************************************************

      IMPLICIT REAL*8 (A-H,O-Z)

      CHARACTER*6 FERNAM, WERNAM
      CHARACTER*60 FERMES, WERMES

C     The NASYS common blocks
      CHARACTER*6 NAFER,NAWER
      CHARACTER*32 FNAME
      CHARACTER*60 NAFMES, NAWMES
      COMMON /NASYS1/ NAESI, NAFEN, NAWEC, NAWEN, NAFEA,
     1                NAWEG, NAEMS, NAEMF, NASUN, NAFUN
      COMMON /NASYS2/ NAFER, NAFMES, NAWER, NAWMES, FNAME

C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
      REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
      INTEGER IWARN,IFATAL,ISYSTM
      CHARACTER MES*60,EFNAME*32
      DATA RMIN/2.23D-308/,RMAX/1.79D308/
      DATA RMNHAF/1.50D-154/,RMXHAF/1.33D154/
      DATA PRECSN/2.22D-16/,PRCHAF/1.5D-8/
      DATA IWARN/1/,IFATAL/2/,ISYSTM/3/
      DATA MES/
     1 'MAGUS Numerical Analyst 3.0.  Copyright (C) 1990 by MAGUS.  '/
C *** End of INCLUDE file.

      LEVERR = NAESI
      IFEN = NAFEN
      IWEN = NAWEN
      RETURN
      END
      REAL*8 FUNCTION VASUM(V,N,I1,ISKIP)
C
C VASUM RETURNS THE SUM OF THE ABSOLUTE VALUES IN A VECTOR.  GIVEN
C VECTOR V OF LENGTH N, VASUM ADDS THE ABSOLUTE VALUES OF ELEMENTS
C I1, I1 + ISKIP, I1 + 2 * ISKIP, ETC., AND RETURNS THE SUM.   
C
C CALLING SEQUENCE:
C
C       X = VASUM(V,N,I1,ISKIP)       
C
C PARAMETER LIST:
C
C       V = VECTOR CONTAINING THE ABSOLUTE VALUES TO BE SUMMED.
C                                                               
C       N = NUMBER OF ELEMENTS IN VECTOR V.                         
C
C      I1 = FIRST ELEMENT IN VECTOR V WHOSE ABSOLUTE VALUE IS TO
C           BE ADDED.
C
C   ISKIP = NUMBER OF ELEMENTS IN VECTOR V TO BE SKIPPED BETWEEN
C           ADDITIONS.
C
C PROGRAMMING NOTES:
C
C       N   IS AN INTEGER GREATER THAN ZERO.
C     
C      I1   IS AN INTEGER GREATER THAN ZERO AND LESS THAN OR
C           EQUAL TO N.
C
C   ISKIP   IS AN INTEGER GREATER THAN ZERO.                       
C
C REQUIRED ROUTINES: UERGET, UERCTL, UERDMP, UERPRT, NACOM
C
C REVISION HISTORY:
C      RICHARD LAHEY      SEPTMEBER 1, 1984    INITIAL RELEASE
C      Richard Lahey      March 15, 1989       version 3.0
C
C**********************************************************************
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION V(N)

C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
      REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
      INTEGER IWARN,IFATAL,ISYSTM
      CHARACTER MES*60,EFNAME*32
      DATA RMIN/2.23D-308/,RMAX/1.79D308/
      DATA RMNHAF/1.50D-154/,RMXHAF/1.33D154/
      DATA PRECSN/2.22D-16/,PRCHAF/1.5D-8/
      DATA IWARN/1/,IFATAL/2/,ISYSTM/3/
      DATA MES/
     1 'MAGUS Numerical Analyst 3.0.  Copyright (C) 1990 by MAGUS.  '/
C *** End of INCLUDE file.

C
C  Check for errors from previous subroutine calls
C
      CALL UERGET (LEVERR, IFEN, IWEN)
      IF (LEVERR .GT. IWARN) RETURN

C  
C  TEST FOR VALID N, I1, AND ISKIP                       
C 
      IF (N .LT. 1) THEN
        LEVERR = IFATAL
        MES =
     1  'Number of elements in array can not be negative or zero     '
        CALL UERCTL ('VASUM ',22203,MES,LEVERR)
        RETURN
      ENDIF

      IF (I1 .LT. 1  .OR. I1 .GT. N) THEN
        LEVERR = IFATAL
        MES =
     1  'Invalid starting index:  must lie between 1 and N           '
        CALL UERCTL ('VASUM ',22120,MES,LEVERR)
        RETURN
      ENDIF

      IF (ISKIP .LT. 1) THEN
        LEVERR = IFATAL
        MES =
     1  'Skip interval can not be negative or zero                   '
        CALL UERCTL ('VASUM ',22106,MES,LEVERR)
        RETURN
      ENDIF

C
C  CALCULATE THE SUM OF THE ABSOLUTE VALUES IN THE ARRAY
C                                                                
100   CONTINUE
      SUM = DABS(V(I1))
      DO 199 I=I1+ISKIP,N,ISKIP
      SUM = SUM + DABS(V(I))
199   CONTINUE
C
C  CLOSE OF FUNCTION
C                             
200   CONTINUE                                      
      VASUM = SUM
      RETURN
      END
      SUBROUTINE VSORT2(V,WS,IWS,N,A,M,NR,IOPT1,IOPT2)
C
C VSORT2 USES AN ADDRESS CALCULATION ALGORITHM TO SORT THE VALUES
C IN VECTOR V.  GIVEN VECTOR V, VSORT2 RETURNS V WITH ITS VALUES
C SORTED IN EITHER ASCENDING OR DESCENDING ORDER.  VSORT2 ALSO
C SORTS ARRAY A TO CORRESPOND WITH VECTOR V.
C
C CALLING SEQUENCE:
C
C       CALL VSORT2(V,WS,IWS,N,A,M,NR,IOPT1,IOPT2)
C
C PARAMETER LIST:
C
C       V = REAL*8 VECTOR OF LENGTH N, CONTAINING THE VALUES
C             TO BE SORTED.
C                                                               
C      WS = REAL*8 VECTOR OF LENGTH OF LENGTH N, WORKSPACE.
C
C     IWS = INTEGER*4 VECTOR OF LENGTH N, WORKSPACE.
C           ON OUTPUT, IWS(K) REPRESENTS THE NEW POSITION IN THE ARRAY
C           V OCCUPIED BY THE NUMBER THAT WAS IN POSITION K IN THE
C           INPUT ARRAY V.  THE USER NEED TAKE NO ACTION WITH THIS
C           INFORMATION, BUT MAY USE IT TO SORT SUBSEQUENT A ARRAYS
C           CORRESPONDING TO THE SAME VECTOR V.
C
C       N = NUMBER OF ELEMENTS IN VECTORS V, WS, AND IWS.
C
C       A = REAL*8 ARRAY CONTAINING VALUES TO BE SORTED TO
C             CORRESPOND WITH VECTOR V.
C
C       M = THE NUMBER OF COLUMNS IN ARRAY A.  ZERO OR NEGATIVE
C             IF THERE IS NO ARRAY TO BE SORTED TO CORRESPOND
C             WITH VECTOR V.  IF THERE IS AN ARRAY A, M MUST
C             BE > = N IF IOPT2 = 1, (I.E. IF COLUMN SORT OF ARRAY A.)
C
C      NR = THE NUMBER OF ROWS IN ARRAY A, EXACTLY AS SPECIFIED IN
C             THE CALLING PROGRAM.  IF THERE IS AN ARRAY A (I.E. IF
C             M > 0), NR MUST BE > = N IF IOPT2 = 0 (I.E. IF ROW SORT
C             OF ARRAY A.)
C
C   IOPT1 = 0 FOR ASCENDING;  1 FOR DESCENDING SORT ORDER.
C
C   IOPT2 = 0 FOR ROW SORT OF ARRAY A;  1 FOR COLUMN SORT OF ARRAY A.
C
C PROGRAMMING NOTES:
C
C   VSORT2 IS INEFFICIENT FOR SMALL N (I.E. N LESS THAN 1000).
C   VSORT1, AN INTERCHANGE SORT, MAY BE PREFERABLE FOR SMALL N.  
C
C   AS IN ALL ADDRESS CALCULATION SORTS, THE HIGHER THE PROPORTION
C   OF IDENTICAL (OR NEARLY IDENTICAL) VALUES IN V, THE MORE VSORT2
C   PERFORMANCE WILL DEGRADE.
C
C REQUIRED ROUTINES: VSORTU, UERGET, UERCTL, UERDMP, UERPRT, NACOM
C
C REVISION HISTORY:
C      RICHARD LAHEY      NOVEMBER 14, 1984    INITIAL RELEASE
C      RICHARD LAHEY      OCTOBER 3, 1985      VERSION 2.0
C      RICHARD LAHEY      JANUARY 6, 1986      ROW AND COLUMN SORTING
C      Richard Lahey      December 29, 1988    Change Log Entry 1
C      Richard Lahey      March 17, 1989       version 3.0
C      Richard Lahey      March 17, 1989       Change Log Entry 2
C
C*******************************************************************
C
C  Change Log Entry 1  12/29/88 RAL
C
C  Two changes:
C
C  1.  Changed the initial setting of IOLDND.  The way it was
C      previously initialized caused an error when N was very small
C      and all numbers hashed to the same subtable.  (The error was
C      overflowing array bounds.)  I fixed the problem by making the
C      initial value of IOLDND depend on N.
C
C  2.  Changed the hashing algorithm to increase efficiency.  The
C      original algorithm resulted in many more collisions than were
C      necessary.
C
C  Change Log Entry 2  03/17/89 RAL
C
C  One Change:
C
C  1.  Commented out the code that checks for degradation.  Left in 
C      the commented code in case it is needed during some future
C      further refinement of the algorithm.
C
C********************************************************************

C  A PARTIAL LIST OF VARIABLES AND PARAMETERS USED IN THIS SUBROUTINE:
C      
C       V:   THE VECTOR TO BE SORTED.
C  
C     IWS:   IN THE FIRST PART OF THE SUBROUTINE, FOR EACH
C              VALUE IN V, HOLDS
C                1) THE LOG OF V, IF V > 2.8
C                2) MINUS THE ABSOLUTE VALUE OF THE LOG OF V,
C                     IF V < -2.8
C                3) ZERO IF V LIES BETWEEN -2.8 AND +2.8
C            LATER IN THE SUBROUTINE, HOLDS THE INDEX OF V IN
C              SORTED ORDER.
C
C   IWSP1:   DIMENSIONED WITH ONE ELEMENT FOR THE LOG (OR MINUS
C              THE LOG OF THE ABSOLUTE VALUE IN THE CASE OF
C              NEGATIVE NUMBERS, ZERO IN THE CASE OF ZERO) OF
C              EVERY NUMBER THAT CAN BE CONTAINED IN A REAL*8
C              VARIABLE.
C            IN THE FIRST PART OF THE SUBROUTINE, IWSP1(I) HOLDS
C              THE COUNT OF ELEMENTS IN V WHOSE LOG VALUE (AS
C              REPRESENTED IN IWS) LIES BETWEEN I AND I + 1.
C              (EXCEPT FOR NUMBERS NEAR ZERO, WHOSE COUNT IS HELD
C              IN IPOS AND INEG.)
C            LATER IN THE SUBROUTINE, IWSP1(I) HOLDS THE START OF
C              THE SUB-TABLE TO WHICH ELEMENTS IN V WHOSE LOG
C              LIES BETWEEN I AND I + 1 WILL HASH.
C
C   IWSP2:   DIMENSIONED TO CORRESPOND WITH IWSP1, IWSP2(I) HOLDS
C              THE LENGTH OF THE SUB-TABLE TO WHICH ELEMENTS IN V
C              WHOSE LOG VALUE (AS REPRESENTED IN IWS) LIES BETWEEN
C              I AND I + 1 WILL HASH.  (EXCEPT FOR NUMBERS NEAR ZERO,
C              WHOSE SUB-TABLE LENGTH IS HELD IN IPOS AND INEG.)
C 
C   IWSP3:   DIMENSIONED TO CORRESPOND WITH IWSP1, IWSP3(I) HOLDS
C              THE INDEX IN V OF THE LARGEST NUMBER IN THE SUB-TABLE
C              TO WHICH ELEMENTS IN V WHOSE LOG VALUE (AS REPRESENTED
C              IN IWS) LIES BETWEEN I AND I + 1 WILL HASH.  (EXCEPT
C              FOR NUMBERS NEAR ZERO, WHOSE LARGEST NUMBER INDEX
C              IS HELD IN IPOS3 AND INEG3.)
C
C   IWSP4:   DIMENSIONED TO CORRESPOND WITH IWSP1, IWSP4(I) HOLDS
C              THE INDEX IN V OF THE SMALLEST NUMBER IN THE SUB-TABLE
C              TO WHICH ELEMENTS IN V WHOSE LOG VALUE (AS REPRESENTED
C              IN IWS) LIES BETWEEN I AND I + 1 WILL HASH.  (EXCEPT
C              FOR NUMBERS NEAR ZERO, WHOSE SMALLEST NUMBER INDEX
C              IS HELD IN IPOS4 AND INEG4.)
C
C     WS:   IN THE FIRST PART OF THE SUB-ROUTINE, LOGICALLY AN
C              INTEGER*2 ARRAY OF N ROWS AND 4 COLUMNS.
C            LATER IN THE SUBROUTINE, A REAL*8 VECTOR WHICH RECEIVES
C              THE VALUES IN V AS ORDERED BY THE INDEX IWS.
C
C    IPOS:   AN INTEGER*4 VARIABLE WHICH IN THE BEGINNING OF THE
C              SUBROUTINE HOLDS THE COUNT OF THE ELEMENTS IN V WHICH
C              LIE BETWEEN ZERO AND 2.8.  LATER IN THE SUBROUTINE
C              HOLDS THE SUB-TABLE LENGTH FOR ELEMENTS IN V WHICH LIE
C              BETWEEN ZERO AND 2.8.
C
C    INEG:   AN INTEGER*4 VARIABLE WHICH IN THE BEGINNING OF THE
C              SUBROUTINE HOLDS THE COUNT OF THE ELEMENTS IN V WHICH
C              LIE BETWEEN -2.8 AND ZERO.  LATER IN THE SUBROUTINE
C              HOLDS THE SUB-TABLE LENGTH FOR ELEMENTS IN V WHICH LIE
C              BETWEEN -2.8 AND ZERO.
C
C  ILDISP:   LOG DISPLACEMENT FROM ZERO.  A PARAMETER WHICH
C              REPRESENTS THE LOG OF A NUMBER ONE GREATER THAN
C              CAN BE CONTAINED IN A REAL*8 VARIABLE.  REPRESENTS
C              THE POSITION IN IWSP1 AND IWSP2 WHICH CORRESPONDS
C              TO ZERO.
C
C  IIDISP:   INTEGER DISPLACEMENT FROM ZERO.  A PARAMETER WHICH
C              REPRESENTS A NUMBER ONE GREATER THAN CAN BE CONTAINED
C              IN AN INTEGER*2 VARIABLE.  BY ALWAYS DISPLACING
C              THE INDEX OF A VALUE IN V BY IIDISP, THE INTEGER*2
C              ARRAY WITHIN WS CAN POINT TO 64,000 POSITIONS RATHER
C              THAN 32,000.

      IMPLICIT REAL*8 (A-H,O-Z)
      INTEGER*2 IMOVER, IWSP3, IWSP4, INEG3, INEG4, IPOS3, IPOS4
      PARAMETER (ILDISP=711,IIDISP=32768,IZERO=0)
      DIMENSION V(N), A(NR,M), IWS(N), WS(N), IMOVER(4)
      DIMENSION IWSP1(1421), IWSP2(1421), IWSP3(1421), IWSP4(1421)
      REAL*8 RMOVER
      EQUIVALENCE (IMOVER(1),RMOVER)

C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
      REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
      INTEGER IWARN,IFATAL,ISYSTM
      CHARACTER MES*60,EFNAME*32
      DATA RMIN/2.23D-308/,RMAX/1.79D308/
      DATA RMNHAF/1.50D-154/,RMXHAF/1.33D154/
      DATA PRECSN/2.22D-16/,PRCHAF/1.5D-8/
      DATA IWARN/1/,IFATAL/2/,ISYSTM/3/
      DATA MES/
     1 'MAGUS Numerical Analyst 3.0.  Copyright (C) 1990 by MAGUS.  '/
C *** End of INCLUDE file.

C
C  Check for errors from previous subroutine calls
C
      CALL UERGET (LEVERR, IFEN, IWEN)
      IF (LEVERR .GT. IWARN) RETURN

C  
C  TEST FOR VALID N, NR, IOPT1, IOPT2
C 
100   CONTINUE

      IF (N .LT. 1) THEN
        LEVERR = IFATAL
        MES =
     1  'Number of elements in array can not be negative or zero     '
        CALL UERCTL ('VSORT2',22203,MES,LEVERR)
        RETURN
      ENDIF

      IF (IOPT1 .NE. 0 .AND. IOPT1 .NE. 1) THEN
        LEVERR = IFATAL
        MES =
     1  'Invalid option: IOPT1 parameter was given an incorrect value'
        CALL UERCTL ('VSORT2',22112,MES,LEVERR)
        RETURN
      ENDIF

      IF (IOPT2 .NE. 0 .AND. IOPT2 .NE. 1) THEN
        LEVERR = IFATAL
        MES =
     1  'Invalid option: IOPT2 parameter was given an incorrect value'
        CALL UERCTL ('VSORT2',22113,MES,LEVERR)
        RETURN
      ENDIF

      IF (IOPT2 .EQ. 0 .AND. M .GT. 0 .AND. NR .LT. N) THEN
        LEVERR = IFATAL
        MES =
     1  'Too few rows in Array A for requested row sort of A         '
        CALL UERCTL ('VSORT2',22217,MES,LEVERR)
        RETURN
      ENDIF

      IF (IOPT2 .EQ. 1 .AND. M .LT. N) THEN
        LEVERR = IFATAL
        MES =
     1  'Too few columns in Array A for requested column sort of A   '
        CALL UERCTL ('VSORT2',22216,MES,LEVERR)
        RETURN
      ENDIF

      IF (M .LT. 0) M = 0

      IF (N .GT. 65534) THEN
        LEVERR = IFATAL
        MES =
     1  'Length of input vector exceeds maximum sort size (65,534)   '
        CALL UERCTL ('VSORT2',22409,MES,LEVERR)
        RETURN
      ENDIF

C                          
C  SET UP THE INDEX ARRAYS
C                                                                
110   CONTINUE
      DO 119, I = 1,ILDISP+ILDISP-1
        IWSP1(I)=IZERO
        IWSP2(I)=IZERO
        IWSP3(I)=IZERO - IIDISP
        IWSP4(I)=IZERO - IIDISP
119   CONTINUE
120   CONTINUE
      DO 129, I = 1,N
        IWS(I)=IZERO
129   CONTINUE
130   CONTINUE
      DO 139, I = 1,N
        DO 134, J = 1,4
          IMOVER(J) = IZERO - IIDISP
134     CONTINUE
        WS(I) = RMOVER
139   CONTINUE
C
C  THIS SECTION DETERMINES THE DISTRIBUTION OF ELEMENTS IN V,
C  PLACES IN IWSP3 THE INDEX OF THE LARGEST NUMBER FOR EACH
C  SUB-TABLE, AND PLACES IN IWSP4 THE INDEX OF THE SMALLEST
C  NUMBER FOR EACH SUB-TABLE.
C
200   CONTINUE
      INEG = IZERO
      IPOS = IZERO
      INEG3 = IZERO - IIDISP
      IPOS3 = IZERO - IIDISP
      INEG4 = IZERO - IIDISP
      IPOS4 = IZERO - IIDISP
      DO 299, I = 1,N 
        IF (V(I) .LT. -2.8D0) THEN
          RHASH = DLOG(DABS(V(I)))
          IWS(I) = 0.0D0 - RHASH
        ELSEIF (V(I) .GT. 2.8D0) THEN
          RHASH = DLOG(V(I))
          IWS(I) = RHASH
        ELSEIF (V(I) .LT. 0.D0) THEN
          INEG = INEG + 1
          IWS(I) = 0
          IF (INEG3 .EQ. IZERO-IIDISP) THEN
            INEG3 = I - IIDISP
          ELSEIF (V(I) .LT. V(INEG3 + IIDISP)) THEN
            INEG3 = I - IIDISP
          ENDIF
          IF (INEG4 .EQ. IZERO-IIDISP) THEN
            INEG4 = I - IIDISP
          ELSEIF (V(I) .GT. V(INEG4 + IIDISP)) THEN
            INEG4 = I - IIDISP
          ENDIF
        ELSE
          IPOS = IPOS + 1
          IWS(I) = 0
          IF (IPOS3 .EQ. IZERO-IIDISP) THEN
            IPOS3 = I - IIDISP
          ELSEIF (V(I) .GT. V(IPOS3 + IIDISP)) THEN
            IPOS3 = I - IIDISP
          ENDIF
          IF (IPOS4 .EQ. IZERO-IIDISP) THEN
            IPOS4 = I - IIDISP
          ELSEIF (V(I) .LT. V(IPOS4 + IIDISP)) THEN
            IPOS4 = I - IIDISP
          ENDIF
        ENDIF
        IHASH = ILDISP + IWS(I)
        IWSP1(IHASH) = IWSP1(IHASH)+1
        IF (IWSP3(IHASH) .EQ. IZERO-IIDISP) THEN
          IWSP3(IHASH) = I - IIDISP
        ELSEIF (DABS(V(I)) .GT. DABS(V(IWSP3(IHASH)+IIDISP))) THEN
          IWSP3(IHASH) = I - IIDISP
        ENDIF
        IF (IWSP4(IHASH) .EQ. IZERO-IIDISP) THEN
          IWSP4(IHASH) = I - IIDISP
        ELSEIF (DABS(V(I)) .LT. DABS(V(IWSP4(IHASH)+IIDISP))) THEN
          IWSP4(IHASH) = I - IIDISP
        ENDIF
299   CONTINUE
C
C  THIS SECTION PUTS START OF SORT SUB-TABLE IN IWSP1 AND SUB-TABLE
C  LENGTH IN IWSP2.                          
C
300   CONTINUE
C
C  Change 12/29/88 RAL [Bug Fix]
C
C  Change the method for setting the initial value of IOLDND.
C  (IOLDND is the "old table end" value.)  The first time through
C  it should be 1/2 N.  This is because the work space available
C  is 4*N.  The tables take 3*N.  That leaves 1*N for overflow.
C  1/2 N should be at the beginning.  The remaining 1/2 end will
C  automatically be located at the end, with the 3*N table space
C  in the middle.
C
      IOLDND = N/2
C
C  THIS SUB-SECTION IDENTIFIES THE START OF SORT SUB-TABLE 
C  AND SUB-TABLE LENGTH FOR NEGATIVE NUMBERS
C
320   CONTINUE
      DO 329, I=1,ILDISP-1
      IF (IWSP1(I) .NE. 0) THEN
        IWSP2(I) = IWSP1(I) + IWSP1(I) + IWSP1(I)
        IOLDND = IOLDND + IWSP2(I)
        IWSP1(I) = IOLDND
      ENDIF
329   CONTINUE
C
C  THIS SUB-SECTION IDENTIFIES THE START OF SORT SUB-TABLE 
C  AND SUB-TABLE LENGTH FOR NUMBERS NEAR ZERO
C
330   CONTINUE
      I = ILDISP
      IF (IWSP1(I) .NE. 0) THEN
        INEG = INEG + INEG + INEG
        IPOS = IPOS + IPOS + IPOS
        IWSP2(I) = INEG + IPOS 
        IWSP1(I) = IOLDND + INEG
        IOLDND = IOLDND + IWSP2(I)
      ENDIF
C
C  THIS SUB-SECTION IDENTIFIES THE START OF SORT SUB-TABLE 
C  AND SUB-TABLE LENGTH FOR POSITIVE NUMBERS.
C
340   CONTINUE
      DO 349, I = ILDISP+1,ILDISP+ILDISP-1
      IF (IWSP1(I) .NE. 0) THEN
        IWSP2(I) = IWSP1(I) + IWSP1(I) + IWSP1(I)
        IWSP1(I) = IOLDND + 1
        IOLDND = IOLDND + IWSP2(I)
      ENDIF
349   CONTINUE
399   CONTINUE
C
C  THIS SECTION HASHES THE INDEX OF EACH ELEMENT IN V TO AN INTEGER*2
C  INDEX (IMOVER) WHICH MAKES TEMPORARY USE OF THE SPACE ALLOCATED
C  TO WS.
C
600   CONTINUE
      I610 = 0
      IDGRD = N
      DO 699, I = 1,N
C
C  THIS SUB-SECTION DETERMINES THE HASH NUMBER.  THE HASHING
C  ALGORITHM IS:  EACH NUMBER HASHES TO THE START OF ITS SUB-TABLE
C  PLUS ITS DISPLACEMENT FROM THE START.  THE START OF ITS SUB-TABLE
C  IS DETERMINED BY ITS LOG.  ITS DISPLACEMENT FROM THE START IS
C  DETERMINED SUCH THAT ITS POSITION IN THE SUB-TABLE IS THE
C  INTERPOLATION OF ITS POSITION BETWEEN THE LARGEST AND SMALLEST
C  VALUES WHICH HASHED TO THE SAME SUB-TABLE.
C
C  Change 12/29/88 RAL  [Enhanced Hash Algorithm]
C  Allow for RANGE (difference between the largest and the smallest)
C  to be 0.  If RANGE is 0, make RATIO 0 for positive numbers and -1
C  for negative numbers.  Otherwise, calculation of RATIO involves
C  division by 0.  The rationale for this is:  Since every number
C  in the subtable is the same, all of them will inevitably collide.
C  Therefore, select the hash algorithm which will minimize the
C  impact of the collisions.  Since the collision algorithm bumps
C  forward any number that is greater than or equal to the number it
C  collides with, positive numbers should hash to the beginning of
C  their subtables, and negative numbers should hash to the end of
C  theirs.  (That is, both should hash to the smallest value.)
C
        ISTART = IWSP1(ILDISP+IWS(I))
        IF (IWS(I) .NE. 0) THEN
          DIF = DABS(V(I) - V(IWSP4(ILDISP+IWS(I))+IIDISP))
          RANGE = DABS(V(IWSP3(ILDISP+IWS(I))+IIDISP)
     1               - V(IWSP4(ILDISP+IWS(I))+IIDISP))
          IF (RANGE .EQ. 0) THEN
            RATIO = (DSIGN(1.D0,V(I)) - 1.D0) / 2.D0
          ELSE
            RATIO = DSIGN(DIF/RANGE,V(I))
          ENDIF
          IHASH = ISTART + (RATIO * (IWSP2(ILDISP+IWS(I)) - 1 ))
        ELSEIF (V(I) .LT. 0.0D0) THEN
          DIF = DABS(V(I) - V(INEG4+IIDISP))
          RANGE = DABS(V(INEG3+IIDISP) - V(INEG4+IIDISP))
          IF (RANGE .EQ. 0) THEN
            RATIO = -1.D0
          ELSE
            RATIO = DSIGN(DIF/RANGE,V(I))
          ENDIF
          IHASH = ISTART + (RATIO * (INEG - 1))
        ELSE
          DIF = V(I) - V(IPOS4+IIDISP)
          RANGE =      V(IPOS3+IIDISP) - V(IPOS4+IIDISP)
          IF (RANGE .EQ. 0) THEN
            RATIO = 0.D0
          ELSE
            RATIO = DIF/RANGE
          ENDIF
          IHASH = ISTART + (RATIO * (IPOS - 1 ))
        ENDIF
        ITEMP = I
        I610 = I610 - 1
610     CONTINUE
C
C  CHECK FOR DEGRADATION
C
        I610 = I610 + 1
        IF (I610 .GT. IDGRD) THEN
C  Change 03/17/89 RAL  [Commented out degrade warning error]
C         LEVERR = IWARN
C         MES =
C    1  'Performance degraded due to many (nearly) identical numbers '
C         CALL UERCTL ('VSORT2',7,MES,LEVERR)
C         IF (LEVERR .GT. IWARN) RETURN
          I610 = 0
        ENDIF
C
C  THIS SUB-SECTION CONVERTS THE HASH NUMBER TO AN ARRAY ADDRESS
C  IN LOGICAL ARRAY WS. 
C
        IHASHV = (IHASH/4) + 1
        IHASHI = MOD(IHASH,4) + 1
        RMOVER = WS(IHASHV)
C
C  THIS SUB-SECTION ATTEMPTS THE HASH.  IF THERE IS A COLLISION,
C  THE VALUE OF V TO BE HASHED IS COMPARED TO THE VALUE WHICH HAS
C  ALREADY HASHED TO THIS ADDRESS.  IF THE NEW VALUE IS GREATER
C  THAN THE ALREADY HASHED VALUE, THE HASH ADDRESS IS INCREMENTED
C  BY 1 AND THE HASH IS ATTEMPTED AGAIN.  IF THE NEW VALUE IS LESS
C  THAN THE ALREADY HASHED VALUE, THE NEW VALUE IS HASHED TO THE
C  ADDRESS.  THE VALUE WHICH HAD HASHED TO THAT ADDRESS BECOMES THE
C  NUMBER TO BE HASHED.  THE HASH ADDRESS IS INCREMENTED BY 1 AND
C  THE HASH IS ATTEMPTED AGAIN.
C
        IF (IMOVER(IHASHI) .EQ. IZERO - IIDISP) THEN
          IMOVER(IHASHI) = ITEMP - IIDISP
          WS(IHASHV) = RMOVER
        ELSEIF (V(I) .GE. V(IMOVER(IHASHI)+IIDISP)) THEN
          IHASH = IHASH + 1
          GOTO 610
        ELSEIF (V(I) .LT. V(IMOVER(IHASHI)+IIDISP)) THEN
          ISWAP = IMOVER(IHASHI)+IIDISP
          IMOVER(IHASHI) = ITEMP-IIDISP
          WS(IHASHV) = RMOVER
          ITEMP = ISWAP
          IHASH = IHASH + 1
          GOTO 610
        ENDIF
C
C  THE FOLLOWING WRITE/FORMAT STATEMENTS CAN BE USED FOR DEBUGGING
C  THE HASHING ALGORITHM
C
C         WRITE(*,611)V(I)
C 611     FORMAT(' NUMBER TO BE SORTED ',1PE11.3)
C         WRITE(*,612)ISTART
C 612     FORMAT(' TABLE START',I6)
C         WRITE(*,614)I610
C 614     FORMAT(' THE NUMBER OF TIMES THROUGH 610           ',I6)
C         WRITE(*,615)V(ITEMP)
C 615     FORMAT(' THE FINAL NUMBER TO HASH',1PE11.3)
C         WRITE(*,616)IHASH
C 616     FORMAT(' WHERE IT HASHED TO',I6)
C         WRITE(*,617)
C 617     FORMAT('  ')
699   CONTINUE
C
C  THIS SECTION TRANSFERS INDEXES FROM THE LOGICAL INDEX ARRAY
C  WITHIN WS TO THE INDEX VECTOR IWS.
C
700   CONTINUE
      ILOC = 1
      DO 799, I = 1,N
        RMOVER = WS(I)
        DO 749, J = 1,4
          IF (IMOVER(J) .NE. IZERO-IIDISP) THEN
            IWS(ILOC) = IMOVER(J)+IIDISP
            ILOC = ILOC+1
          ENDIF
749     CONTINUE
799   CONTINUE
C
C  CALL VSORTU TO RESEQUENCE VECTOR V AND ARRAY A ACCORDING TO
C  INDEX IWS.
C
800   CONTINUE
      CALL VSORTU(V,WS,IWS,N,A,M,NR,IOPT1,IOPT2)
C
C  Call UERCTL to reset warning error count to zero.
C  Change 03/17/89 RAL
C    Commented out the reset call to UERCTL.  Since the only
C    warning error has been commented out above, there is no need
C    to call UERCTL with a reset command.
C
900   CONTINUE
C     CALL UERCTL('VSORT2',0,MES,LEVERR)
C                 
C  CLOSE OF SUBROUTINE 
C
5000  CONTINUE                                      
      RETURN            
      END
      SUBROUTINE CDIV(XR,XI,YR,YI,ZR,ZI)
C
C This subroutine computes the complex ratio
C
C     ZR + i * ZI = ( XR + i * XI ) / ( YR + i * YI )
C
C as described in section II, chapter 13 of the Handbook for 
C Automatic Computation, volume II, edited by Wilkinson
C and Reinsch-HACLA II/13 for short.
C
C YR and YI may not both be zero.
C
C CALLING FORMAT:
C
C      CALL CDIV(XR,XI,YR,YI,ZR,ZI)
C
C PARAMETER LIST:
C
C     XR  (Input)  Real part of numerator.
C     XI  (Input)  Imaginary part of numerator.
C     YR  (Input)  Real part of denominator.
C     YI  (Input)  Imaginary part of denominator.
C     ZR  (Output) Real part of ratio.
C     ZI  (Output) Imaginary part of ratio.
C
C Required routines: None
C
C Revision history:
C
C      Kevin Thompson     February 21, 1987     Version 3.0
C
C**********************************************************************
C
      IMPLICIT REAL*8 (A-H,O-Z)
C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
      REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
      INTEGER IWARN,IFATAL,ISYSTM
      CHARACTER MES*60,EFNAME*32
      DATA RMIN/2.23D-308/,RMAX/1.79D308/
      DATA RMNHAF/1.50D-154/,RMXHAF/1.33D154/
      DATA PRECSN/2.22D-16/,PRCHAF/1.5D-8/
      DATA IWARN/1/,IFATAL/2/,ISYSTM/3/
      DATA MES/
     1 'MAGUS Numerical Analyst 3.0.  Copyright (C) 1990 by MAGUS.  '/
C *** End of INCLUDE file.
C Compute the ratio.
      IF(DABS(YR) .GT. DABS(YI)) THEN
        RATIO=YI/YR
        DENOM=RATIO*YI+YR
        ZR=(XR+RATIO*XI)/DENOM
        ZI=(XI-RATIO*XR)/DENOM
      ELSE
        RATIO=YR/YI
        DENOM=RATIO*YR+YI
        ZR=(RATIO*XR+XI)/DENOM
        ZI=(RATIO*XI-XR)/DENOM
      ENDIF
      RETURN
      END
      real*8 function dnrm2 ( n, dx, incx)
      implicit real*8 (a-h,o-z)
      integer next
      dimension dx(*)
c
c     euclidean norm of the n-vector stored in dx() with storage
c     increment incx .
c     if    n .le. 0 return with result = 0.
c     if n .ge. 1 then incx must be .ge. 1
c
c           c.l.lawson, 1978 jan 08
c
c     four phase method     using two built-in constants that are
c     hopefully applicable to all machines.
c         cutlo = maximum of  dsqrt(u/eps)  over all known machines.
c         cuthi = minimum of  dsqrt(v)      over all known machines.
c     where
c         eps = smallest no. such that eps + 1. .gt. 1.
c         u   = smallest positive no.   (underflow limit)
c         v   = largest  no.            (overflow  limit)
c
c     brief outline of algorithm..
c
c     phase 1    scans zero components.
c     move to phase 2 when a component is nonzero and .le. cutlo
c     move to phase 3 when a component is .gt. cutlo
c     move to phase 4 when a component is .ge. cuthi/m
c     where m = n for x() real and m = 2*n for complex.
c
c     values for cutlo and cuthi..
c     from the environmental parameters listed in the imsl converter
c     document the limiting values are as follows..
c     cutlo, s.p.   u/eps = 2**(-102) for  honeywell.  close seconds are
c                   univac and dec at 2**(-103)
c                   thus cutlo = 2**(-51) = 4.44089e-16
c     cuthi, s.p.   v = 2**127 for univac, honeywell, and dec.
c                   thus cuthi = 2**(63.5) = 1.30438e19
c     cutlo, d.p.   u/eps = 2**(-67) for honeywell and dec.
c                   thus cutlo = 2**(-33.5) = 8.23181d-11
c     cuthi, d.p.   same as s.p.  cuthi = 1.30438d19
c     data cutlo, cuthi / 8.232d-11,  1.304d19 /
c     data cutlo, cuthi / 4.441e-16,  1.304e19 /
c
C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
      REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
      INTEGER IWARN,IFATAL,ISYSTM
      CHARACTER MES*60,EFNAME*32
      DATA RMIN/2.23D-308/,RMAX/1.79D308/
      DATA RMNHAF/1.50D-154/,RMXHAF/1.33D154/
      DATA PRECSN/2.22D-16/,PRCHAF/1.5D-8/
      DATA IWARN/1/,IFATAL/2/,ISYSTM/3/
      DATA MES/
     1 'MAGUS Numerical Analyst 3.0.  Copyright (C) 1990 by MAGUS.  '/
C *** End of INCLUDE file.
c
      data   zero, one /0.0d0, 1.0d0/
C*********************************************************************
C
C Modifications by Kevin Thompson, 12/30/89
C Put in alternate loop for use on vector machines.  To vectorize
C more efficiently, remove comment from DTEMP=... through 
C DNRM2=... below, and comment out all subsequent lines before the
C RETURN statement.
C
C     DTEMP=0.D0
C Find the index of the element of largest absolute magintude.
C     INDEX=IDAMAX(N,DX,INCX)
C If the largest element is nonzero, then define a scale factor 
C as the reciprocal of the largest element (in magnitude);
C otherwise skip the norm calculation, as the norm is zero.
C     XMAX=DABS(DX(1+(INDEX-1)*INCX))
C     IF(XMAX .GT. 0.D0) THEN
C       SCALE=1.D0/XMAX
C Sum the scaled, squared elements.  Use different loops for
C INCX=1 and INCX .ne. 1.
C       IF(INCX .EQ. 1) THEN
C         DO 25 I=1,N
C           XSCALE=SCALE*DX(I)
C           DTEMP=DTEMP+XSCALE*XSCALE
C25       CONTINUE
C       ELSE
C         DO 30 I=1,N
C           XSCALE=SCALE*DX(1+(I-1)*INCX)
C           DTEMP=DTEMP+XSCALE*XSCALE
C30       CONTINUE
C       ENDIF
C Take the square root to get the norm of the scaled vector.
C       DTEMP=DSQRT(DTEMP)
C     ENDIF
C Return the input vector norm.
C     DNRM2=XMAX*DTEMP
C
C*********************************************************************
C
      CUTLO = RMNHAF/PRCHAF
      CUTHI = RMXHAF
      if(n .gt. 0) go to 10
         dnrm2  = zero
         go to 300
c
   10 next=1
      sum = zero
      nn = n * incx
c                                                 begin main loop
      i = 1
   20 go to (30, 50, 70, 110),next
   30 if( dabs(dx(i)) .gt. cutlo) go to 85
      next=2
      xmax = zero
c
c                        phase 1.  sum is zero
c
   50 if( dx(i) .eq. zero) go to 200
      if( dabs(dx(i)) .gt. cutlo) go to 85
c
c                                prepare for phase 2.
      next=3
      go to 105
c
c                                prepare for phase 4.
c
  100 i = j
      next=4
      sum = (sum / dx(i)) / dx(i)
  105 xmax = dabs(dx(i))
      go to 115
c
c                   phase 2.  sum is small.
c                             scale to avoid destructive underflow.
c
   70 if( dabs(dx(i)) .gt. cutlo ) go to 75
c
c                     common code for phases 2 and 4.
c                     in phase 4 sum is large.  scale to avoid overflow.
c
  110 if( dabs(dx(i)) .le. xmax ) go to 115
         sum = one + sum * (xmax / dx(i))**2
         xmax = dabs(dx(i))
         go to 200
c
  115 sum = sum + (dx(i)/xmax)**2
      go to 200
c
c
c                  prepare for phase 3.
c
   75 sum = (sum * xmax) * xmax
c
c
c     for real or d.p. set hitest = cuthi/n
c     for complex      set hitest = cuthi/(2*n)
c
   85 hitest = cuthi/float( n )
c
c                   phase 3.  sum is mid-range.  no scaling.
c
      do 95 j =i,nn,incx
      if(dabs(dx(j)) .ge. hitest) go to 100
   95    sum = sum + dx(j)**2
      dnrm2 = dsqrt( sum )
      go to 300
c
  200 continue
      i = i + incx
      if ( i .le. nn ) go to 20
c
c              end of main loop.
c
c              compute square root and adjust for scaling.
c
      dnrm2 = xmax * dsqrt(sum)
  300 continue
      return
      end
      subroutine  drot (n,dx,incx,dy,incy,c,s)
c
c     applies a plane rotation.
c     jack dongarra, linpack, 3/11/78.
c
      implicit real*8 (a-h,o-z)
      integer i,incx,incy,ix,iy,n
      dimension dx(*),dy(*)
C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
      REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
      INTEGER IWARN,IFATAL,ISYSTM
      CHARACTER MES*60,EFNAME*32
      DATA RMIN/2.23D-308/,RMAX/1.79D308/
      DATA RMNHAF/1.50D-154/,RMXHAF/1.33D154/
      DATA PRECSN/2.22D-16/,PRCHAF/1.5D-8/
      DATA IWARN/1/,IFATAL/2/,ISYSTM/3/
      DATA MES/
     1 'MAGUS Numerical Analyst 3.0.  Copyright (C) 1990 by MAGUS.  '/
C *** End of INCLUDE file.
c
      if(n.le.0)return
      if(incx.eq.1.and.incy.eq.1)go to 20
c
c       code for unequal increments or equal increments not equal
c         to 1
c
      ix = 1
      iy = 1
      if(incx.lt.0)ix = (-n+1)*incx + 1
      if(incy.lt.0)iy = (-n+1)*incy + 1
      do 10 i = 1,n
        dtemp = c*dx(ix) + s*dy(iy)
        dy(iy) = c*dy(iy) - s*dx(ix)
        dx(ix) = dtemp
        ix = ix + incx
        iy = iy + incy
   10 continue
      return
c
c       code for both increments equal to 1
c
   20 do 30 i = 1,n
        dtemp = c*dx(i) + s*dy(i)
        dy(i) = c*dy(i) - s*dx(i)
        dx(i) = dtemp
   30 continue
      return
      end
      subroutine  dscal(n,da,dx,incx)
c
c     scales a vector by a constant.
c     uses unrolled loops for increment equal to one.
c     jack dongarra, linpack, 3/11/78.
c
      implicit real*8 (a-h,o-z)
      integer i,incx,m,mp1,n,nincx
      dimension dx(*)
C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
      REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
      INTEGER IWARN,IFATAL,ISYSTM
      CHARACTER MES*60,EFNAME*32
      DATA RMIN/2.23D-308/,RMAX/1.79D308/
      DATA RMNHAF/1.50D-154/,RMXHAF/1.33D154/
      DATA PRECSN/2.22D-16/,PRCHAF/1.5D-8/
      DATA IWARN/1/,IFATAL/2/,ISYSTM/3/
      DATA MES/
     1 'MAGUS Numerical Analyst 3.0.  Copyright (C) 1990 by MAGUS.  '/
C *** End of INCLUDE file.
c
      if(n.le.0)return
      if(incx.eq.1)go to 20
c
c        code for increment not equal to 1
c
      nincx = n*incx
      do 10 i = 1,nincx,incx
        dx(i) = da*dx(i)
   10 continue
      return
c
c        code for increment equal to 1
c
c
   20 CONTINUE
C*********************************************************************
C
C Modifications by Kevin Thompson, 12/30/89
C Put in alternate loop for use on vector machines.  To vectorize
C more efficiently, remove comment from DO 30 through 30 CONTINUE 
C below, and comment out all subsequent lines before the RETURN 
C statement.
C     DO 30 I=1,N
C       DX(I)=DA*DX(I)
C30   CONTINUE
C
C*********************************************************************
C
c        clean-up loop
c
      m = mod(n,5)
      if( m .eq. 0 ) go to 40
      do 30 i = 1,m
        dx(i) = da*dx(i)
   30 continue
      if( n .lt. 5 ) return
   40 mp1 = m + 1
      do 50 i = mp1,n,5
        dx(i) = da*dx(i)
        dx(i + 1) = da*dx(i + 1)
        dx(i + 2) = da*dx(i + 2)
        dx(i + 3) = da*dx(i + 3)
        dx(i + 4) = da*dx(i + 4)
   50 continue
      return
      end
      SUBROUTINE EHOUSF(A,TRANS,NP,NR,N,NLEFT,ILOW,IHI,L,M,W,WSUM,
     1                  COMVEC)
C
C This subroutine performs the double step QR algorithm of Francis.
C Each individual step can be expressed as a pre- and post-
C multiplication of the Hessenberg matrix A - SHIFT * I by the 
C Householder transformation matrices, where I is the identity matrix
C and SHIFT is a shift of origin.  The N - 1 pre-multiplications of
C one step produce an upper triangular matrix, while the N - 1
C post-multiplications restore the Hessenberg form.  The new Hessenberg
C matrix should have smaller sub-diagonal elements than the previous
C one.  The diagonal matrix SHIFT * I is added to produce the final
C matrix.
C
C The double step combines two steps to operate on the matrix A with 
C the product of two Householder transformations.  The initial (Mth)
C pre-multiplication uses a transformation which would eliminate the 
C sub-diagonal elements in the first column of the product matrix
C
C           MPROD = (A - SHIFT1 * I) * (A - SHIFT2 * I),
C
C if applied to MPROD, but is instead applied to A. This pre- and post-
C multiplication leave A in a non-Hessenberg form.  The remaining 
C Householder transformations (M + 1 through N - 1) restore A to
C Hessenberg form, with smaller sub-diagonal elements.  Only the
C sum and product of the shifts appear, which means that only real
C numbers are encountered, as the two shifts are either real numbers
C or complex conjugate numbers, whose sum and product are real. 
C
C CALLING FORMAT:
C
C      CALL EHOUSF(A,TRANS,NP,NR,N,NLEFT,ILOW,IHI,L,M,W,WSUM,COMVEC)
C
C PARAMETER LIST:
C
C      A  (Input)  Array of row dimension NR and column dimension at
C                  least N, containing the N by N input Hessenberg 
C                  matrix to be transformed.
C         (Output) The new Hessenberg matrix.
C  TRANS  (Output) Array of dimensions (NP,NR,N), containing the
C                  accumulated transformation matrix in elements
C                  TRANS(1,I,J).  This array is later used in
C                  EFULL to store the eigenvectors.
C     NP  (Input)  First dimension of the transformation matrix array,
C                  TRANS.  It must be nonzero of the transformation
C                  matrix is to be computed.
C     NR  (Input)  Row dimension of A exactly as specified in the
C                  calling program.  NR must be greater than or equal 
C                  to N.
C      N  (Input)  Number of rows and columns in the input matrix.
C  NLEFT  (Input)  Index of the current row and column limit for the
C                  eigenvalue(s) being determined.  The number of
C                  eigenvalues remaining to be determined is
C                  NLEFT - ILOW + 1.
C   ILOW  (Input)  Lower index of the non-trivial sub-matrix being
C                  transformed.
C    IHI  (Input)  Upper index of the non-trivial sub-matrix being
C                  transformed.
C      L  (Input)  The starting point of the sub-matrix in rows L
C                  through N to be transformed.
C      M  (Input)  The starting transformation, of QR transformations
C                  M through N-1.
C      W  (Input)  The normalized elements of the product matrix 
C                  starting in column M.
C   WSUM  (Input)  The sum of the absolute values of the elements in
C                  the product matrix starting in column M (used to
C                  normalize W).
C COMVEC  (Input)  Transformation matrix option parameter.  The
C                  value of COMVEC determines what calculation is
C                  to be performed, as described below.
C
C                  COMVEC = .FALSE.: Do not compute the transformation
C                                    matrix.
C                  COMVEC = .TRUE.:  Compute the transformation matrix.
C
C
C Many short loops involving small vectors (of length 2 or 3) have
C been eliminated (unrolled) in favor of straight-through code.  The
C vectors U, V, and W are represented as local variables, rather than
C arrays, to avoid the overhead of loop setup time and array
C fetching time for these small loops.  The penalty for this choice
C is less elegant looking code, with many IF blocks inserted to
C avoid fetching elements outside the A array.  The only place
C W appears as an array is in the parameter list, where it passes
C information from the calling program.
C
C Note that the first element of the V vector, V1, is always equal
C to 1, and therefore never appears in the calculations.
C
C Required routines: None
C
C Revision history:
C
C Kevin Thompson     October 31, 1987     Version 3.0
C
C Kevin Thompson     June 8, 1991         Version 3.0
C   Set upper index of the DO 100 loop to NLEFT-2 and omitted
C   the tests for K = NLEFT -1 in this loop.  Duplicated code
C   for the case K = NLEFT - 1 after the loop and handled the
C   omission of out-of-bounds elements in this code.  The
C   result should cause inner loops to vectorize throughout.
C
C**********************************************************************
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION A(NR,N),TRANS(NP,NR,N),W(3)
      LOGICAL COMVEC
C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
      REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
      INTEGER IWARN,IFATAL,ISYSTM
      CHARACTER MES*60,EFNAME*32
      DATA RMIN/2.23D-308/,RMAX/1.79D308/
      DATA RMNHAF/1.50D-154/,RMXHAF/1.33D154/
      DATA PRECSN/2.22D-16/,PRCHAF/1.5D-8/
      DATA IWARN/1/,IFATAL/2/,ISYSTM/3/
      DATA MES/
     1 'MAGUS Numerical Analyst 3.0.  Copyright (C) 1990 by MAGUS.  '/
C *** End of INCLUDE file.
C
C If the column sum WSUM is zero, then the sub-diagonal elements of the
C product matrix are zero.  The transformation is then skipped, because
C it is not needed.
      IF(WSUM .NE. 0.D0) THEN
        W1=W(1)
        W2=W(2)
        W3=W(3)
        COLMAG=DSQRT(W1**2+W2**2+W3**2)
C Transfer the sign of W1 (i.e., of the first element of the product
C matrix, EMMM, computed in the calling routine) to the column 
C magnitude.
        IF(W1 .GE. 0.D0) THEN
          SCOL=COLMAG
        ELSE
          SCOL=-COLMAG
        ENDIF
        SUM=SCOL+W1
C Compute the vector elements.
C
        U1=SUM/SCOL
        U2=W2/SCOL
        U3=W3/SCOL
C       V1=1.D0
        V2=W2/SUM
        V3=W3/SUM
C
C Next perform a Householder transformation on A, using the vector 
C elements just computed.
C
C Pre-multiplication:
C
C The first column (J = M - 1 for M > L) to be transformed is treated
C separately to avoid cancellation errors in the calculation of the
C matrix elements.  It is altered by the pre-multiplication, but
C not the post-multiplication, since the sub-Hessenberg elements of
C the product matrix (EMM1M and EMM2M) are negligible.  
C If M = L, column L - 1 does not enter the calculation in any way
C and is not altered.
C
        IF(M .NE. L) THEN
          A(M,M-1)=-A(M,M-1)
        ENDIF
C Now do the rest of the elements. If COMVEC = .FALSE., only the active
C sub-matrix need be transformed, and the remainder may be ignored.
C Otherwise elements outside the active sub-matrix must be transformed
C also.
        IF(COMVEC) THEN
          IMIN=1
          JMAX=N
        ELSE
          IMIN=L
          JMAX=NLEFT
        ENDIF
        DO 10 J=M,JMAX
          C=A(M,J)+V2*A(M+1,J)+V3*A(M+2,J)
          A(M,J)=A(M,J)-C*U1
          A(M+1,J)=A(M+1,J)-C*U2
          A(M+2,J)=A(M+2,J)-C*U3
10      CONTINUE
C
C Post-multiplication:
C
        DO 20 I=IMIN,MIN0(M+3,NLEFT)
          C=U1*A(I,M)+U2*A(I,M+1)+U3*A(I,M+2)
          A(I,M)=A(I,M)-C
          A(I,M+1)=A(I,M+1)-C*V2
          A(I,M+2)=A(I,M+2)-C*V3
20      CONTINUE
      ENDIF
C
C Accumulate the transformations in TRANS, if needed.
C
      IF(COMVEC) THEN         
        DO 30 I=ILOW,IHI
          C=U1*TRANS(1,I,M)+U2*TRANS(1,I,M+1)+U3*TRANS(1,I,M+2)
          TRANS(1,I,M)=TRANS(1,I,M)-C
          TRANS(1,I,M+1)=TRANS(1,I,M+1)-C*V2
          TRANS(1,I,M+2)=TRANS(1,I,M+2)-C*V3
30      CONTINUE
      ENDIF
C
C Now the A matrix has sub-Hessenberg elements A(M+2,M), A(M+3,M),
C and A(M+3,M+1).  Householder transformations M + 1 through N - 1 will
C restore A to Hessenberg form.
C
      DO 100 K=M+1,NLEFT-2
C Compute the Householder vector elements.
C
C To prevent overflows in the calculation of the vectors (due to
C products of matrix elements), normalize the leading sub-diagonal
C elements and store them in W.
C
        WSUM=DABS(A(K,K-1))+DABS(A(K+1,K-1))+DABS(A(K+2,K-1))
C
C If the column sum WSUM is zero, then the sub-diagonal elements of the
C matrix are zero.  The transformation is then skipped, because
C it is not needed.
        IF(WSUM .NE. 0.D0) THEN
          W1=A(K,K-1)/WSUM
          W2=A(K+1,K-1)/WSUM
          W3=A(K+2,K-1)/WSUM
C First the magnitude of the sub-diagonal column:
          COLMAG=DSQRT(W1**2+W2**2+W3**2)
C Transfer the sign of A(K,K-1) to the column magnitude.
          IF(W1 .GE. 0.D0) THEN
            SCOL=COLMAG
          ELSE
            SCOL=-COLMAG
          ENDIF
          SUM=SCOL+W1
C
C Now get the vector elements.
C
          U1=SUM/SCOL
          U2=W2/SCOL
          U3=W3/SCOL
C         V1=1.D0
          V2=W2/SUM
          V3=W3/SUM
C
C Perform the Householder transformation.
C
C Pre-multiplication:
C
C The first column (J = K - 1) to be transformed is treated
C separately to avoid cancellation errors in the calculation of the
C matrix elements.  It is altered by the pre-multiplication, but
C not the post-multiplication, since the sub-Hessenberg elements
C are eliminated by the pre-multiplication.
C
          A(K,K-1)=-SCOL*WSUM
          A(K+1,K-1)=0.D0
          A(K+2,K-1)=0.D0
C Now do the remaining columns.  If COMVEC = .FALSE., only the active
C sub-matrix need be transformed, and the remainder may be ignored.
C Otherwise elements outside the active sub-matrix must be transformed
C also.
          DO 140 J=K,JMAX
C Get the C coefficient and do the transformation.
            C=A(K,J)+V2*A(K+1,J)+V3*A(K+2,J)
            A(K+2,J)=A(K+2,J)-C*U3
            A(K+1,J)=A(K+1,J)-C*U2
            A(K,J)=A(K,J)-C*U1
140       CONTINUE
C
C Post-multiplication:
C
          DO 170 I=IMIN,MIN0(K+3,NLEFT)
C Get the C coefficient and do the transformation.
            C=U1*A(I,K)+U2*A(I,K+1)+U3*A(I,K+2)
            A(I,K+2)=A(I,K+2)-C*V3
            A(I,K+1)=A(I,K+1)-C*V2
            A(I,K)=A(I,K)-C
170       CONTINUE
C
C Accumulate the transformations in TRANS, if needed.
C
          IF(COMVEC) THEN         
            DO 180 I=ILOW,IHI
              C=U1*TRANS(1,I,K)+U2*TRANS(1,I,K+1)+U3*TRANS(1,I,K+2)
              TRANS(1,I,K+2)=TRANS(1,I,K+2)-C*V3
              TRANS(1,I,K+1)=TRANS(1,I,K+1)-C*V2
              TRANS(1,I,K)=TRANS(1,I,K)-C
180         CONTINUE
          ENDIF
C
C The Kth Householder transformation is complete.
C
        ENDIF
100   CONTINUE
C
C
C Handle the case K = NLEFT - 1.
C
      K=NLEFT-1
C
C Compute the Householder vector elements.
C
C To prevent overflows in the calculation of the vectors (due to
C products of matrix elements), normalize the leading sub-diagonal
C elements and store them in W.
C
      WSUM=DABS(A(K,K-1))+DABS(A(K+1,K-1))
C
C If the column sum WSUM is zero, then the sub-diagonal elements of the
C matrix are zero.  The transformation is then skipped, because
C it is not needed.
      IF(WSUM .NE. 0.D0) THEN
        W1=A(K,K-1)/WSUM
        W2=A(K+1,K-1)/WSUM
C First the magnitude of the sub-diagonal column:
        COLMAG=DSQRT(W1**2+W2**2)
C Transfer the sign of A(K,K-1) to the column magnitude.
        IF(W1 .GE. 0.D0) THEN
          SCOL=COLMAG
        ELSE
          SCOL=-COLMAG
        ENDIF
        SUM=SCOL+W1
C
C Now get the vector elements.
C
        U1=SUM/SCOL
        U2=W2/SCOL
C       V1=1.D0
        V2=W2/SUM
C
C Perform the Householder transformation.
C
C Pre-multiplication:
C
C The first column (J = K - 1) to be transformed is treated
C separately to avoid cancellation errors in the calculation of the
C matrix elements.  It is altered by the pre-multiplication, but
C not the post-multiplication, since the sub-Hessenberg elements
C are eliminated by the pre-multiplication.
C
        A(K,K-1)=-SCOL*WSUM
        A(K+1,K-1)=0.D0
C Now do the remaining columns.  If COMVEC = .FALSE., only the active
C sub-matrix need be transformed, and the remainder may be ignored.
C Otherwise elements outside the active sub-matrix must be transformed
C also.
        DO 240 J=K,JMAX
C Get the C coefficient and do the transformation.
          C=A(K,J)+V2*A(K+1,J)
          A(K+1,J)=A(K+1,J)-C*U2
          A(K,J)=A(K,J)-C*U1
240     CONTINUE
C
C Post-multiplication:
C
        DO 270 I=IMIN,NLEFT
C Get the C coefficient and do the transformation.
          C=U1*A(I,K)+U2*A(I,K+1)
          A(I,K+1)=A(I,K+1)-C*V2
          A(I,K)=A(I,K)-C
270     CONTINUE
C
C Accumulate the transformations in TRANS, if needed.
C
        IF(COMVEC) THEN         
          DO 280 I=ILOW,IHI
            C=U1*TRANS(1,I,K)+U2*TRANS(1,I,K+1)
            TRANS(1,I,K+1)=TRANS(1,I,K+1)-C*V2
            TRANS(1,I,K)=TRANS(1,I,K)-C
280       CONTINUE
        ENDIF
C
C The NLEFT - 1st Householder transformation is complete.
C
      ENDIF
C
C
C The transformation back to Hessenberg form is complete.
      RETURN
      END
      SUBROUTINE EINTF(A,NR,N,ILOW,IHI,I1,I2)
C
C This subroutine exchanges rows and columns I1 and I2 of the matrix
C A, and is used by the EBALNF subroutine to isolate eigenvalues in
C the input matrix.
C
C CALLING FORMAT:
C
C      CALL EINTF(A,NR,N,ILOW,IHI,I1,I2)
C
C PARAMETER LIST:
C
C      A  (Input/  Array of row dimension NR and column dimension at
C          Output) least N, containing the N by N input matrix whose
C                  eigenvalues are to be computed.  Rows I1 and I2 of
C                  this matrix are exchanged, as are columns I1 and I2.
C     NR  (Input)  Row dimension of A exactly as specified in the
C                  calling program.  NR must be greater than or equal
C                  to N.
C      N  (Input)  Number of rows and columns in the input matrix.
C   ILOW  (Input)  Current lower limit.
C    IHI  (Input)  Current upper limit.
C     I1  (Input)  First row and column number.
C     I2  (Input)  Second row and column number.
C
C Required routines: DSWAP
C
C Revision history:
C
C Kevin Thompson     September 19, 1987     Version 3.0
C
C Kevin Thompson     March 16, 1991       Version 3.0
C   Replaced loops with calls to DSWAP to take advantage of
C   more efficient BLAS routines.
C
C**********************************************************************
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION A(NR,N)
C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
      REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
      INTEGER IWARN,IFATAL,ISYSTM
      CHARACTER MES*60,EFNAME*32
      DATA RMIN/2.23D-308/,RMAX/1.79D308/
      DATA RMNHAF/1.50D-154/,RMXHAF/1.33D154/
      DATA PRECSN/2.22D-16/,PRCHAF/1.5D-8/
      DATA IWARN/1/,IFATAL/2/,ISYSTM/3/
      DATA MES/
     1 'MAGUS Numerical Analyst 3.0.  Copyright (C) 1990 by MAGUS.  '/
C *** End of INCLUDE file.
C Test for a need to interchange.
      IF(I1 .NE. I2) THEN
C Interchange row I1 with row I2.
        CALL DSWAP(N-ILOW+1,A(I1,ILOW),NR,A(I2,ILOW),NR)
C Interchange column I1 with column I2.
        CALL DSWAP(IHI,A(1,I1),1,A(1,I2),1)
      ENDIF
      RETURN
      END
      integer function IDAMAX(n,dx,incx)
c
c     finds the index of element having max. absolute value.
c     jack dongarra, linpack, 3/11/78.
c
      implicit real*8 (a-h,o-z)
      integer i,incx,ix,n
      dimension dx(*)
C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
      REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
      INTEGER IWARN,IFATAL,ISYSTM
      CHARACTER MES*60,EFNAME*32
      DATA RMIN/2.23D-308/,RMAX/1.79D308/
      DATA RMNHAF/1.50D-154/,RMXHAF/1.33D154/
      DATA PRECSN/2.22D-16/,PRCHAF/1.5D-8/
      DATA IWARN/1/,IFATAL/2/,ISYSTM/3/
      DATA MES/
     1 'MAGUS Numerical Analyst 3.0.  Copyright (C) 1990 by MAGUS.  '/
C *** End of INCLUDE file.
c
      IDAMAX = 0
      if( n .lt. 1 ) return
      IDAMAX = 1
      if(n.eq.1)return
      if(incx.eq.1)go to 20
c
c        code for increment not equal to 1
c
      ix = 1
      dmax = dabs(dx(1))
      ix = ix + incx
      do 10 i = 2,n
         if(dabs(dx(ix)).le.dmax) go to 5
         IDAMAX = i
         dmax = dabs(dx(ix))
    5    ix = ix + incx
   10 continue
      return
c
c        code for increment equal to 1
c
   20 dmax = dabs(dx(1))
      do 30 i = 2,n
         if(dabs(dx(i)).le.dmax) go to 30
         IDAMAX = i
         dmax = dabs(dx(i))
   30 continue
      return
      end
      SUBROUTINE UERDMP(IRLEV,RTBL,IERRN,IERRC,NUQERR,NRECR)
C
C     UERDMP reports the total number of each kind of warning error
C     which has occurred at the error recursion level passed in the
C     variable IRLEV.
C
C     CALLING SEQUENCE:
C
C       CALL UERDMP(IRLEV,RTBL,IERRN,IERRC,NUQERR,NRECR)
C
C     PARAMETER LIST:
C
C      IRLEV (INPUT)   The error recursion level.
C
C       RTBL (INPUT)   The recursion table, a character array
C                      containing the names of all the routines
C                      which have encountered warning errors
C                      in the current error chain.
C
C      IERRN (INPUT)   An INTEGER*2 array containing a list of all
C                      the warning errors encountered at recursion
C                      level IRLEV.
C
C      IERRC (INPUT)   An INTEGER*2 array containing error
C                      occurrence counts which correspond to each
C                      warning error number in IERRN.
C
C     NUQERR (INPUT)   An INTEGER*2 array containing the number of
C                      different (unique) warning errors encountered
C                      at recursion level IRLEV.
C
C      NRECR (INPUT)   Number of recursions possible, required for
C                      dimensioning the above arrays.
C
C     PROGRAMMING NOTES:
C
C       This routine will be linked in to every program written using
C       Numerical Analyst.  To save RAM, some integer arrays have
C       been declared INTEGER*2.
C
C     REQUIRED ROUTINES:
C
C       NACOM
C
C     REVISION HISTORY:
C
C       Kevin Thompson     July 7, 1984        initial release
C       Richard Lahey      October 25, 1987    first revision
C       Richard Lahey      October 28, 1988    second revision
C       Richard Lahey      February 25, 1989   version 3.0
C
C**********************************************************************

      IMPLICIT REAL*8(A-H,O-Z)

C     IYES:  Code for yes.
      PARAMETER (IYES = 1)

C     The NASYS common blocks
      CHARACTER*6 NAFER,NAWER
      CHARACTER*32 FNAME
      CHARACTER*60 NAFMES, NAWMES
      COMMON /NASYS1/ NAESI, NAFEN, NAWEC, NAWEN, NAFEA,
     1                NAWEG, NAEMS, NAEMF, NASUN, NAFUN
      COMMON /NASYS2/ NAFER, NAFMES, NAWER, NAWMES, FNAME

      CHARACTER*47 TOTMES
      CHARACTER*18 ENDMES
      INTEGER*2 K
      LOGICAL OPEN, THERE


C     IERRN:  An array which contains all warning error numbers
C     encountered for a recursion level.
      INTEGER*2 IERRN(NRECR,*)

C     IERRC:  An array which contains a count of all warning
C     error numbers encountered for a recursion level.
      INTEGER*2 IERRC(NRECR,*)

C     NUQERR:  The number of unique errors for a recursion level.
      INTEGER*2 NUQERR(NRECR)

C     RTBL: The recursion table.
      CHARACTER*6 RTBL(NRECR)

C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
      REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
      INTEGER IWARN,IFATAL,ISYSTM
      CHARACTER MES*60,EFNAME*32
      DATA RMIN/2.23D-308/,RMAX/1.79D308/
      DATA RMNHAF/1.50D-154/,RMXHAF/1.33D154/
      DATA PRECSN/2.22D-16/,PRCHAF/1.5D-8/
      DATA IWARN/1/,IFATAL/2/,ISYSTM/3/
      DATA MES/
     1 'MAGUS Numerical Analyst 3.0.  Copyright (C) 1990 by MAGUS.  '/
C *** End of INCLUDE file.

      DATA TOTMES/' *** The total warning error count for routine '/
      DATA ENDMES/' is the following:'/

C     If warning error generation is on
      IF (NAWEG .EQ. IYES) THEN

C       If error messages are to be printed to the screen
        IF (NAEMS .EQ. IYES) THEN
          WRITE(NASUN,10)TOTMES,RTBL(IRLEV),ENDMES
10        FORMAT(/A,A,A)
          WRITE(NASUN,20)(IERRN(IRLEV,K),IERRC(IRLEV,K),
     1                K=1,NUQERR(IRLEV))
20        FORMAT('     Error number ',I5,' occurred ',I5,' time(s)')
        ENDIF

C       If error messages are to be written to the error file
        IF (NAEMF .EQ. IYES) THEN
          INQUIRE (FILE=FNAME, EXIST=THERE, OPENED=OPEN)
          IF (.NOT. THERE) THEN
            OPEN (NAFUN,FILE=FNAME,STATUS='NEW')               
          ELSEIF (.NOT. OPEN) THEN
            OPEN (NAFUN,FILE=FNAME,STATUS='OLD')
          ENDIF
          WRITE(NAFUN,10)TOTMES,RTBL(IRLEV),ENDMES
          WRITE(NAFUN,20)(IERRN(IRLEV,K),IERRC(IRLEV,K),
     1                K=1,NUQERR(IRLEV))
        ENDIF
      ENDIF

      RETURN
      END
      SUBROUTINE UERPRT (NAME,NUMERR,ERRMES,LEVERR)
C
C     UERPRT, prints error messages to the screen and to the
C     error file as indicated by parameters in NASYS1.
C
C     CALLING SEQUENCE
C
C       CALL UERPRT (NAME,NUMERR,ERRMES,LEVERR)
C
C     PARAMETER LIST
C
C        NAME (INPUT)   Name of the routine in which the error
C                       occurred.  (CHARACTER*6)
C
C      NUMERR (INPUT)   Number which uniquely identifies the error
C                       condition detected.
C
C      ERRMES (INPUT)   Error message.
C
C      LEVERR (INPUT)   Severity level of the error.
C
C
C     PROGRAMMING NOTES
C
C       None
C
C     REQUIRED ROUTINES
C
C       NACOM
C
C     REVISION HISTORY
C
C       Richard Lahey       10/31/87      initial release
C       Richard Lahey       10/28/88      first revision
C       Richard Lahey       02/25/89      version 3.0
C       Richard Lahey       11/19/89      change system error message
C     
C********************************************************************

      IMPLICIT REAL*8 (A-H,O-Z)

C     IYES:  Code for yes.
      PARAMETER (IYES = 1)

      CHARACTER*6 NAME
      CHARACTER*60 ERRMES

      CHARACTER*7 WARMES,FATMES,SYSMES,GENMES
      LOGICAL OPEN, THERE

C     The NASYS common blocks
      CHARACTER*6 NAFER,NAWER
      CHARACTER*32 FNAME
      CHARACTER*60 NAFMES, NAWMES
      COMMON /NASYS1/ NAESI, NAFEN, NAWEC, NAWEN, NAFEA,
     1                NAWEG, NAEMS, NAEMF, NASUN, NAFUN
      COMMON /NASYS2/ NAFER, NAFMES, NAWER, NAWMES, FNAME

C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
      REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
      INTEGER IWARN,IFATAL,ISYSTM
      CHARACTER MES*60,EFNAME*32
      DATA RMIN/2.23D-308/,RMAX/1.79D308/
      DATA RMNHAF/1.50D-154/,RMXHAF/1.33D154/
      DATA PRECSN/2.22D-16/,PRCHAF/1.5D-8/
      DATA IWARN/1/,IFATAL/2/,ISYSTM/3/
      DATA MES/
     1 'MAGUS Numerical Analyst 3.0.  Copyright (C) 1990 by MAGUS.  '/
C *** End of INCLUDE file.

      DATA WARMES/'Warning'/ 
      DATA FATMES/'Fatal  '/ 
      DATA SYSMES/'System '/

100   CONTINUE

      IF (LEVERR .EQ. IWARN) THEN
        GENMES = WARMES
      ELSEIF (LEVERR .EQ. IFATAL) THEN
        GENMES = FATMES
      ELSE
        GENMES = SYSMES
      ENDIF


C     If error messages are to be printed to the screen
      IF (NAEMS .EQ. IYES) THEN
        WRITE(NASUN,110)GENMES,NUMERR,NAME
110     FORMAT(/' *** ',A7,' error ',I5,
     1          ' from Numerical Analyst routine ',A6)
        WRITE(NASUN,120)ERRMES
120     FORMAT(' *** ',A60)
        IF (LEVERR .GT. IFATAL) THEN
          WRITE(NASUN,130)
130       FORMAT('     If this error occurs, please report it to'
     1          ,' MAGUS.')
        ENDIF
      ENDIF

C     If error messages are to be written to the error file
      IF (NAEMF .EQ. IYES) THEN
        INQUIRE (FILE=FNAME, EXIST=THERE, OPENED=OPEN)
        IF (.NOT. THERE) THEN
          OPEN (NAFUN,FILE=FNAME,STATUS='NEW')               
        ELSEIF (.NOT. OPEN) THEN
          OPEN (NAFUN,FILE=FNAME,STATUS='OLD')
        ENDIF
        WRITE(NAFUN,110)GENMES,NUMERR,NAME
        WRITE(NAFUN,120)ERRMES
        IF (LEVERR .GT. IFATAL) WRITE(NAFUN,130)
      ENDIF
C
      RETURN
      END
      SUBROUTINE VSORTU(V,WS,IWS,N,A,M,NR,IOPT1,IOPT2)
C
C VSORTU IS A UTILITY PROGRAM CALLED BY VSORT1 AND VSORT2.       
C VSORTU USES THE INDEX CREATED BY VSORT1 OR VSORT2 TO RESEQUENCE
C VECTOR V AND ARRAY A.
C
C CALLING SEQUENCE:
C
C       CALL VSORTU(V,WS,IWS,N,A,M,NR,IOPT1,IOPT2)
C
C PARAMETER LIST:
C
C       V = REAL*8 VECTOR OF LENGTH N, CONTAINING THE VALUES
C             TO BE SORTED.
C                                                               
C      WS = REAL*8 VECTOR OF LENGTH OF LENGTH N, WORKSPACE.
C
C     IWS = INTEGER*4 VECTOR OF LENGTH N, CONTAINS THE NEW SEQUENCE
C             FOR VECTOR V AND ARRAY A.
C
C       N = NUMBER OF ELEMENTS IN VECTORS V, WS, AND IWS.
C
C       A = REAL*8 ARRAY CONTAINING VALUES TO BE SORTED TO
C             CORRESPOND WITH VECTOR V.
C
C       M = THE NUMBER OF COLUMNS IN ARRAY A.  ZERO OR NEGATIVE
C             IF THERE IS NO ARRAY TO BE SORTED TO CORRESPOND
C             WITH VECTOR V.  IF THERE IS AN ARRAY A, M MUST
C             BE > = N IF IOPT2 = 1 (I.E. IF COLUMN SORT OF ARRAY A.)
C
C      NR = THE NUMBER OF ROWS IN ARRAY A, EXACTLY AS SPECIFIED IN
C             THE CALLING PROGRAM.  IF THERE IS AN ARRAY A (I.E. IF
C             M > 0), NR MUST BE > = N IF IOPT2 = 0 (I.E. IF ROW
C             SORT OF ARRAY A.)
C
C   IOPT1 = 0 FOR ASCENDING;  1 FOR DESCENDING SORT ORDER.
C
C   IOPT2 = 0 FOR ROW SORT OF ARRAY A;  1 FOR COLUMN SORT OF ARRAY A.
C
C PROGRAMMING NOTES: NONE
C
C REQUIRED ROUTINES: NONE
C
C REVISION HISTORY:
C      RICHARD LAHEY      DECEMBER 22, 1984    INITIAL RELEASE
C                         JANUARY 6, 1986      ROW AND COLUMN SORTING
C      Richard Lahey      March 17, 1989       version 3.0
C
C**********************************************************************
C
      IMPLICIT REAL*8 (A-H,O-Z)
      PARAMETER (ILDISP=711,IIDISP=32768,IZERO=0)
      PARAMETER (IASCND=0,IDCND=1,IRSRT=0,ICSRT=1)
      DIMENSION V(N), A(NR,M), IWS(N), WS(N)

C *** INCLUDE file for compilers which use the IEEE number format:
C *** All IBM PC/compatibles and the Apple Macintosh
      REAL*8 RMIN,RMAX,RMNHAF,RMXHAF,PRECSN,PRCHAF
      INTEGER IWARN,IFATAL,ISYSTM
      CHARACTER MES*60,EFNAME*32
      DATA RMIN/2.23D-308/,RMAX/1.79D308/
      DATA RMNHAF/1.50D-154/,RMXHAF/1.33D154/
      DATA PRECSN/2.22D-16/,PRCHAF/1.5D-8/
      DATA IWARN/1/,IFATAL/2/,ISYSTM/3/
      DATA MES/
     1 'MAGUS Numerical Analyst 3.0.  Copyright (C) 1990 by MAGUS.  '/
C *** End of INCLUDE file.

C
C  THE FOLLOWING SECTION SORTS VECTOR V ACCORDING TO ITS INDEX
C
1000  CONTINUE
      IF(IOPT1 .EQ. IASCND) THEN
C  ASCENDING SORT
        DO 1019,I=1,N
          WS(I) = V(IWS(I))
1019    CONTINUE
C  DESCENDING SORT
      ELSE
        DO 1029,I=1,N
          WS(N+1-I)=V(IWS(I))
1029    CONTINUE
      ENDIF
1100  CONTINUE
      DO 1199,I=1,N
        V(I) = WS(I)
1199  CONTINUE
C
C  THE FOLLOWING SECTION SORTS ARRAY A TO CORRESPOND WITH V
C
2000  CONTINUE
      IF (IOPT2 .EQ. IRSRT) THEN
C  ROW SORT OF ARRAY A
        DO 2059,J=1,M
          IF (IOPT1 .EQ. IASCND) THEN
C  ASCENDING SORT
            DO 2019,I=1,N
              WS(I)=A(IWS(I),J)
2019        CONTINUE
          ELSE
C  DESCENDING SORT
            DO 2029,I=1,N
              WS(N+1-I)=A(IWS(I),J)
2029        CONTINUE
          ENDIF
2030      CONTINUE
          DO 2039,I=1,N
            A(I,J)=WS(I)
2039      CONTINUE
2059    CONTINUE
      ELSE
C  COLUMN SORT OF ARRAY A
        DO 2099,J=1,NR
          IF (IOPT1 .EQ. IASCND) THEN
C  ASCENDING SORT
            DO 2069,I=1,N
              WS(I)=A(J,IWS(I))
2069        CONTINUE
          ELSE
C  DESCENDING SORT
            DO 2079,I=1,N
              WS(N+1-I)=A(J,IWS(I))
2079        CONTINUE
          ENDIF
2080      CONTINUE
          DO 2089,I=1,N
            A(J,I)=WS(I)
2089      CONTINUE
2099    CONTINUE
      ENDIF
C                 
C  CLOSE OF SUBROUTINE 
C                             
5000  CONTINUE                                      
      RETURN            
      END
